"""
Defines Bodo's compiler pipeline.
"""
import os
import warnings
from collections import namedtuple
import numba
from numba.core import ir, ir_utils
from numba.core.compiler import DefaultPassBuilder
from numba.core.compiler_machinery import AnalysisPass, FunctionPass, register_pass
from numba.core.inline_closurecall import inline_closure_call
from numba.core.ir_utils import build_definitions, find_callname, get_definition, guard
from numba.core.registry import CPUDispatcher
from numba.core.typed_passes import DumpParforDiagnostics, InlineOverloads, IRLegalization, NopythonTypeInference, ParforPass, PreParforPass
from numba.core.untyped_passes import MakeFunctionToJitFunction, ReconstructSSA, WithLifting
import bodo
import bodo.hiframes.dataframe_indexing
import bodo.hiframes.datetime_datetime_ext
import bodo.hiframes.datetime_timedelta_ext
import bodo.io
import bodo.libs
import bodo.libs.array_kernels
import bodo.libs.int_arr_ext
import bodo.libs.re_ext
import bodo.libs.spark_extra
import bodo.transforms
import bodo.transforms.series_pass
import bodo.transforms.untyped_pass
import bodo.utils
import bodo.utils.typing
from bodo.transforms.series_pass import SeriesPass
from bodo.transforms.table_column_del_pass import TableColumnDelPass
from bodo.transforms.typing_pass import BodoTypeInference
from bodo.transforms.untyped_pass import UntypedPass
from bodo.utils.utils import is_assign, is_call_assign, is_expr
numba.core.config.DISABLE_PERFORMANCE_WARNINGS = 1
from numba.core.errors import NumbaExperimentalFeatureWarning, NumbaPendingDeprecationWarning
warnings.simplefilter('ignore', category=NumbaExperimentalFeatureWarning)
warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)
inline_all_calls = False


class BodoCompiler(numba.core.compiler.CompilerBase):

    def define_pipelines(self):
        return self._create_bodo_pipeline(distributed=True,
            inline_calls_pass=inline_all_calls)

    def _create_bodo_pipeline(self, distributed=True, inline_calls_pass=
        False, udf_pipeline=False):
        mbf__oug = 'bodo' if distributed else 'bodo_seq'
        mbf__oug = mbf__oug + '_inline' if inline_calls_pass else mbf__oug
        pm = DefaultPassBuilder.define_nopython_pipeline(self.state, mbf__oug)
        if inline_calls_pass:
            pm.add_pass_after(InlinePass, WithLifting)
        if udf_pipeline:
            pm.add_pass_after(ConvertCallsUDFPass, WithLifting)
        add_pass_before(pm, BodoUntypedPass, ReconstructSSA)
        replace_pass(pm, BodoTypeInference, NopythonTypeInference)
        remove_pass(pm, MakeFunctionToJitFunction)
        add_pass_before(pm, BodoSeriesPass, PreParforPass)
        if distributed:
            pm.add_pass_after(BodoDistributedPass, ParforPass)
        else:
            pm.add_pass_after(LowerParforSeq, ParforPass)
            pm.add_pass_after(LowerBodoIRExtSeq, LowerParforSeq)
        add_pass_before(pm, BodoTableColumnDelPass, IRLegalization)
        pm.add_pass_after(BodoDumpDistDiagnosticsPass, DumpParforDiagnostics)
        pm.finalize()
        return [pm]


def add_pass_before(pm, pass_cls, location):
    assert pm.passes
    pm._validate_pass(pass_cls)
    pm._validate_pass(location)
    for xkf__huobg, (wlskp__gtnu, fctd__nwu) in enumerate(pm.passes):
        if wlskp__gtnu == location:
            break
    else:
        raise bodo.utils.typing.BodoError('Could not find pass %s' % location)
    pm.passes.insert(xkf__huobg, (pass_cls, str(pass_cls)))
    pm._finalized = False


def replace_pass(pm, pass_cls, location):
    assert pm.passes
    pm._validate_pass(pass_cls)
    pm._validate_pass(location)
    for xkf__huobg, (wlskp__gtnu, fctd__nwu) in enumerate(pm.passes):
        if wlskp__gtnu == location:
            break
    else:
        raise bodo.utils.typing.BodoError('Could not find pass %s' % location)
    pm.passes[xkf__huobg] = pass_cls, str(pass_cls)
    pm._finalized = False


def remove_pass(pm, location):
    assert pm.passes
    pm._validate_pass(location)
    for xkf__huobg, (wlskp__gtnu, fctd__nwu) in enumerate(pm.passes):
        if wlskp__gtnu == location:
            break
    else:
        raise bodo.utils.typing.BodoError('Could not find pass %s' % location)
    pm.passes.pop(xkf__huobg)
    pm._finalized = False


@register_pass(mutates_CFG=True, analysis_only=False)
class InlinePass(FunctionPass):
    _name = 'inline_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        assert state.func_ir
        inline_calls(state.func_ir, state.locals)
        state.func_ir.blocks = ir_utils.simplify_CFG(state.func_ir.blocks)
        return True


def _convert_bodo_dispatcher_to_udf(rhs, func_ir):
    vwks__hut = guard(get_definition, func_ir, rhs.func)
    if isinstance(vwks__hut, (ir.Global, ir.FreeVar, ir.Const)):
        kgt__yzycc = vwks__hut.value
    else:
        ahecw__qrk = guard(find_callname, func_ir, rhs)
        if not (ahecw__qrk and isinstance(ahecw__qrk[0], str) and
            isinstance(ahecw__qrk[1], str)):
            return
        func_name, func_mod = ahecw__qrk
        try:
            import importlib
            ycop__cmu = importlib.import_module(func_mod)
            kgt__yzycc = getattr(ycop__cmu, func_name)
        except:
            return
    if isinstance(kgt__yzycc, CPUDispatcher) and issubclass(kgt__yzycc.
        _compiler.pipeline_class, BodoCompiler
        ) and kgt__yzycc._compiler.pipeline_class != BodoCompilerUDF:
        kgt__yzycc._compiler.pipeline_class = BodoCompilerUDF
        kgt__yzycc.recompile()


@register_pass(mutates_CFG=True, analysis_only=False)
class ConvertCallsUDFPass(FunctionPass):
    _name = 'inline_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        assert state.func_ir
        for block in state.func_ir.blocks.values():
            for oqveo__ifvqj in block.body:
                if is_call_assign(oqveo__ifvqj):
                    _convert_bodo_dispatcher_to_udf(oqveo__ifvqj.value,
                        state.func_ir)
        return True


@register_pass(mutates_CFG=True, analysis_only=False)
class BodoUntypedPass(FunctionPass):
    _name = 'bodo_untyped_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        assert state.func_ir
        xlxbq__ihm = UntypedPass(state.func_ir, state.typingctx, state.args,
            state.locals, state.metadata, state.flags)
        xlxbq__ihm.run()
        return True


def _update_definitions(func_ir, node_list):
    mzf__njpna = ir.Loc('', 0)
    cxjq__xigq = ir.Block(ir.Scope(None, mzf__njpna), mzf__njpna)
    cxjq__xigq.body = node_list
    build_definitions({(0): cxjq__xigq}, func_ir._definitions)


_series_inline_attrs = {'values', 'shape', 'size', 'empty', 'name', 'index',
    'dtype'}
_series_no_inline_methods = {'to_list', 'tolist', 'rolling', 'to_csv',
    'count', 'fillna', 'to_dict', 'map', 'apply', 'pipe', 'combine',
    'bfill', 'ffill', 'pad', 'backfill', 'mask', 'where'}
_series_method_alias = {'isnull': 'isna', 'product': 'prod', 'kurtosis':
    'kurt', 'is_monotonic': 'is_monotonic_increasing', 'notnull': 'notna'}
_dataframe_no_inline_methods = {'apply', 'itertuples', 'pipe', 'to_parquet',
    'to_sql', 'to_csv', 'to_json', 'assign', 'to_string', 'query',
    'rolling', 'mask', 'where'}
TypingInfo = namedtuple('TypingInfo', ['typingctx', 'targetctx', 'typemap',
    'calltypes', 'curr_loc'])


def _inline_bodo_getattr(stmt, rhs, rhs_type, new_body, func_ir, typingctx,
    targetctx, typemap, calltypes):
    from bodo.hiframes.pd_dataframe_ext import DataFrameType
    from bodo.hiframes.pd_series_ext import SeriesType
    from bodo.utils.transform import compile_func_single_block
    if isinstance(rhs_type, SeriesType) and rhs.attr in _series_inline_attrs:
        ijd__wrjpy = 'overload_series_' + rhs.attr
        vrc__qrk = getattr(bodo.hiframes.series_impl, ijd__wrjpy)
    if isinstance(rhs_type, DataFrameType) and rhs.attr in ('index', 'columns'
        ):
        ijd__wrjpy = 'overload_dataframe_' + rhs.attr
        vrc__qrk = getattr(bodo.hiframes.dataframe_impl, ijd__wrjpy)
    else:
        return False
    func_ir._definitions[stmt.target.name].remove(rhs)
    syz__qovx = vrc__qrk(rhs_type)
    qsl__vadea = TypingInfo(typingctx, targetctx, typemap, calltypes, stmt.loc)
    ymo__iij = compile_func_single_block(syz__qovx, (rhs.value,), stmt.
        target, qsl__vadea)
    _update_definitions(func_ir, ymo__iij)
    new_body += ymo__iij
    return True


def _inline_bodo_call(rhs, i, func_mod, func_name, pass_info, new_body,
    block, typingctx, targetctx, calltypes, work_list):
    from bodo.hiframes.pd_dataframe_ext import DataFrameType
    from bodo.hiframes.pd_series_ext import SeriesType
    from bodo.utils.transform import replace_func, update_locs
    func_ir = pass_info.func_ir
    typemap = pass_info.typemap
    if isinstance(func_mod, ir.Var) and isinstance(typemap[func_mod.name],
        SeriesType) and func_name not in _series_no_inline_methods:
        if func_name in _series_method_alias:
            func_name = _series_method_alias[func_name]
        if (func_name in bodo.hiframes.series_impl.explicit_binop_funcs or 
            func_name.startswith('r') and func_name[1:] in bodo.hiframes.
            series_impl.explicit_binop_funcs):
            return False
        rhs.args.insert(0, func_mod)
        uufp__ldmrb = tuple(typemap[ubvo__peesn.name] for ubvo__peesn in
            rhs.args)
        yga__sdhy = {mbf__oug: typemap[ubvo__peesn.name] for mbf__oug,
            ubvo__peesn in dict(rhs.kws).items()}
        syz__qovx = getattr(bodo.hiframes.series_impl, 'overload_series_' +
            func_name)(*uufp__ldmrb, **yga__sdhy)
    elif isinstance(func_mod, ir.Var) and isinstance(typemap[func_mod.name],
        DataFrameType) and func_name not in _dataframe_no_inline_methods:
        if func_name in _series_method_alias:
            func_name = _series_method_alias[func_name]
        rhs.args.insert(0, func_mod)
        uufp__ldmrb = tuple(typemap[ubvo__peesn.name] for ubvo__peesn in
            rhs.args)
        yga__sdhy = {mbf__oug: typemap[ubvo__peesn.name] for mbf__oug,
            ubvo__peesn in dict(rhs.kws).items()}
        syz__qovx = getattr(bodo.hiframes.dataframe_impl, 
            'overload_dataframe_' + func_name)(*uufp__ldmrb, **yga__sdhy)
    else:
        return False
    ipi__hrl = replace_func(pass_info, syz__qovx, rhs.args, pysig=numba.
        core.utils.pysignature(syz__qovx), kws=dict(rhs.kws))
    block.body = new_body + block.body[i:]
    sccj__zfyhv, fctd__nwu = inline_closure_call(func_ir, ipi__hrl.glbls,
        block, len(new_body), ipi__hrl.func, typingctx=typingctx, targetctx
        =targetctx, arg_typs=ipi__hrl.arg_types, typemap=typemap, calltypes
        =calltypes, work_list=work_list)
    for mfq__intmi in sccj__zfyhv.values():
        mfq__intmi.loc = rhs.loc
        update_locs(mfq__intmi.body, rhs.loc)
    return True


def bodo_overload_inline_pass(func_ir, typingctx, targetctx, typemap, calltypes
    ):
    hgfvh__eno = namedtuple('PassInfo', ['func_ir', 'typemap'])
    pass_info = hgfvh__eno(func_ir, typemap)
    igad__rkxc = func_ir.blocks
    work_list = list((loz__anasw, igad__rkxc[loz__anasw]) for loz__anasw in
        reversed(igad__rkxc.keys()))
    while work_list:
        ghc__nqyj, block = work_list.pop()
        new_body = []
        zqvak__qpozp = False
        for i, stmt in enumerate(block.body):
            if is_assign(stmt) and is_expr(stmt.value, 'getattr'):
                rhs = stmt.value
                rhs_type = typemap[rhs.value.name]
                if _inline_bodo_getattr(stmt, rhs, rhs_type, new_body,
                    func_ir, typingctx, targetctx, typemap, calltypes):
                    continue
            if is_call_assign(stmt):
                rhs = stmt.value
                ahecw__qrk = guard(find_callname, func_ir, rhs, typemap)
                if ahecw__qrk is None:
                    new_body.append(stmt)
                    continue
                func_name, func_mod = ahecw__qrk
                if _inline_bodo_call(rhs, i, func_mod, func_name, pass_info,
                    new_body, block, typingctx, targetctx, calltypes, work_list
                    ):
                    zqvak__qpozp = True
                    break
            new_body.append(stmt)
        if not zqvak__qpozp:
            igad__rkxc[ghc__nqyj].body = new_body
    func_ir.blocks = ir_utils.simplify_CFG(func_ir.blocks)


@register_pass(mutates_CFG=True, analysis_only=False)
class BodoDistributedPass(FunctionPass):
    _name = 'bodo_distributed_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        assert state.func_ir
        from bodo.transforms.distributed_pass import DistributedPass
        ypamu__kzboy = DistributedPass(state.func_ir, state.typingctx,
            state.targetctx, state.typemap, state.calltypes, state.
            return_type, state.metadata, state.flags)
        state.return_type = ypamu__kzboy.run()
        return True


@register_pass(mutates_CFG=True, analysis_only=False)
class BodoSeriesPass(FunctionPass):
    _name = 'bodo_series_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        assert state.func_ir
        gwfhr__rmuxc = SeriesPass(state.func_ir, state.typingctx, state.
            targetctx, state.typemap, state.calltypes, state.locals)
        gwfhr__rmuxc.run()
        gwfhr__rmuxc.run()
        gwfhr__rmuxc.run()
        return True


@register_pass(mutates_CFG=False, analysis_only=True)
class BodoDumpDistDiagnosticsPass(AnalysisPass):
    _name = 'bodo_dump_diagnostics_pass'

    def __init__(self):
        AnalysisPass.__init__(self)

    def run_pass(self, state):
        szg__uqpvb = 0
        xgufi__kxzco = 'BODO_DISTRIBUTED_DIAGNOSTICS'
        try:
            szg__uqpvb = int(os.environ[xgufi__kxzco])
        except:
            pass
        if szg__uqpvb > 0 and 'distributed_diagnostics' in state.metadata:
            state.metadata['distributed_diagnostics'].dump(szg__uqpvb,
                state.metadata)
        return True


class BodoCompilerSeq(BodoCompiler):

    def define_pipelines(self):
        return self._create_bodo_pipeline(distributed=False,
            inline_calls_pass=inline_all_calls)


class BodoCompilerUDF(BodoCompiler):

    def define_pipelines(self):
        return self._create_bodo_pipeline(distributed=False, udf_pipeline=True)


@register_pass(mutates_CFG=False, analysis_only=True)
class LowerParforSeq(FunctionPass):
    _name = 'bodo_lower_parfor_seq_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        bodo.transforms.distributed_pass.lower_parfor_sequential(state.
            typingctx, state.func_ir, state.typemap, state.calltypes, state
            .metadata)
        return True


@register_pass(mutates_CFG=False, analysis_only=True)
class LowerBodoIRExtSeq(FunctionPass):
    _name = 'bodo_lower_ir_ext_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        from bodo.transforms.distributed_pass import distributed_run_extensions
        from bodo.transforms.table_column_del_pass import remove_dead_table_columns
        state.func_ir._definitions = build_definitions(state.func_ir.blocks)
        qsl__vadea = TypingInfo(state.typingctx, state.targetctx, state.
            typemap, state.calltypes, state.func_ir.loc)
        remove_dead_table_columns(state.func_ir, state.typemap, qsl__vadea)
        for block in state.func_ir.blocks.values():
            new_body = []
            for oqveo__ifvqj in block.body:
                if type(oqveo__ifvqj) in distributed_run_extensions:
                    avzaz__wkish = distributed_run_extensions[type(
                        oqveo__ifvqj)]
                    rrk__ufptl = avzaz__wkish(oqveo__ifvqj, None, state.
                        typemap, state.calltypes, state.typingctx, state.
                        targetctx)
                    new_body += rrk__ufptl
                elif is_call_assign(oqveo__ifvqj):
                    rhs = oqveo__ifvqj.value
                    ahecw__qrk = guard(find_callname, state.func_ir, rhs)
                    if ahecw__qrk == ('gatherv', 'bodo') or ahecw__qrk == (
                        'allgatherv', 'bodo'):
                        oqveo__ifvqj.value = rhs.args[0]
                    new_body.append(oqveo__ifvqj)
                else:
                    new_body.append(oqveo__ifvqj)
            block.body = new_body
        return True


@register_pass(mutates_CFG=False, analysis_only=True)
class BodoTableColumnDelPass(AnalysisPass):
    _name = 'bodo_table_column_del_pass'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        dpey__zbwa = TableColumnDelPass(state.func_ir, state.typingctx,
            state.targetctx, state.typemap, state.calltypes)
        return dpey__zbwa.run()


def inline_calls(func_ir, _locals, work_list=None, typingctx=None,
    targetctx=None, typemap=None, calltypes=None):
    if work_list is None:
        work_list = list(func_ir.blocks.items())
    iwsk__zpzqm = set()
    while work_list:
        ghc__nqyj, block = work_list.pop()
        iwsk__zpzqm.add(ghc__nqyj)
        for i, jxbcz__dtf in enumerate(block.body):
            if isinstance(jxbcz__dtf, ir.Assign):
                ymjiv__riljp = jxbcz__dtf.value
                if isinstance(ymjiv__riljp, ir.Expr
                    ) and ymjiv__riljp.op == 'call':
                    vwks__hut = guard(get_definition, func_ir, ymjiv__riljp
                        .func)
                    if isinstance(vwks__hut, (ir.Global, ir.FreeVar)
                        ) and isinstance(vwks__hut.value, CPUDispatcher
                        ) and issubclass(vwks__hut.value._compiler.
                        pipeline_class, BodoCompiler):
                        zhka__ltg = vwks__hut.value.py_func
                        arg_types = None
                        if typingctx:
                            jki__ltvl = dict(ymjiv__riljp.kws)
                            dstg__ixj = tuple(typemap[ubvo__peesn.name] for
                                ubvo__peesn in ymjiv__riljp.args)
                            rqn__fysgl = {raef__lqmou: typemap[ubvo__peesn.
                                name] for raef__lqmou, ubvo__peesn in
                                jki__ltvl.items()}
                            fctd__nwu, arg_types = (vwks__hut.value.
                                fold_argument_types(dstg__ixj, rqn__fysgl))
                        fctd__nwu, snd__paape = inline_closure_call(func_ir,
                            zhka__ltg.__globals__, block, i, zhka__ltg,
                            typingctx=typingctx, targetctx=targetctx,
                            arg_typs=arg_types, typemap=typemap, calltypes=
                            calltypes, work_list=work_list)
                        _locals.update((snd__paape[raef__lqmou].name,
                            ubvo__peesn) for raef__lqmou, ubvo__peesn in
                            vwks__hut.value.locals.items() if raef__lqmou in
                            snd__paape)
                        break
    return iwsk__zpzqm


def udf_jit(signature_or_function=None, **options):
    amjf__dvg = {'comprehension': True, 'setitem': False, 'inplace_binop': 
        False, 'reduction': True, 'numpy': True, 'stencil': False, 'fusion':
        True}
    return numba.njit(signature_or_function, parallel=amjf__dvg,
        pipeline_class=bodo.compiler.BodoCompilerUDF, **options)


def is_udf_call(func_type):
    return isinstance(func_type, numba.core.types.Dispatcher
        ) and func_type.dispatcher._compiler.pipeline_class == BodoCompilerUDF


def is_user_dispatcher(func_type):
    return isinstance(func_type, numba.core.types.functions.ObjModeDispatcher
        ) or isinstance(func_type, numba.core.types.Dispatcher) and issubclass(
        func_type.dispatcher._compiler.pipeline_class, BodoCompiler)


@register_pass(mutates_CFG=False, analysis_only=True)
class DummyCR(FunctionPass):
    _name = 'bodo_dummy_cr'

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        state.cr = (state.func_ir, state.typemap, state.calltypes, state.
            return_type)
        return True


def remove_passes_after(pm, location):
    assert pm.passes
    pm._validate_pass(location)
    for xkf__huobg, (wlskp__gtnu, fctd__nwu) in enumerate(pm.passes):
        if wlskp__gtnu == location:
            break
    else:
        raise bodo.utils.typing.BodoError('Could not find pass %s' % location)
    pm.passes = pm.passes[:xkf__huobg + 1]
    pm._finalized = False


class TyperCompiler(BodoCompiler):

    def define_pipelines(self):
        [pm] = self._create_bodo_pipeline()
        remove_passes_after(pm, InlineOverloads)
        pm.add_pass_after(DummyCR, InlineOverloads)
        pm.finalize()
        return [pm]


def get_func_type_info(func, arg_types, kw_types):
    typingctx = numba.core.registry.cpu_target.typing_context
    targetctx = numba.core.registry.cpu_target.target_context
    uvsz__bgqg = None
    qumr__moxwp = None
    _locals = {}
    ydut__oxd = numba.core.utils.pysignature(func)
    args = bodo.utils.transform.fold_argument_types(ydut__oxd, arg_types,
        kw_types)
    hkqn__exhh = numba.core.compiler.Flags()
    gai__hka = {'comprehension': True, 'setitem': False, 'inplace_binop': 
        False, 'reduction': True, 'numpy': True, 'stencil': False, 'fusion':
        True}
    pspqb__lyby = {'nopython': True, 'boundscheck': False, 'parallel': gai__hka
        }
    numba.core.registry.cpu_target.options.parse_as_flags(hkqn__exhh,
        pspqb__lyby)
    gpzd__axjzz = TyperCompiler(typingctx, targetctx, uvsz__bgqg, args,
        qumr__moxwp, hkqn__exhh, _locals)
    return gpzd__axjzz.compile_extra(func)
