"""Support distributed deep learning with Horovod
"""
import time
import numba
import numpy as np
from mpi4py import MPI
import bodo
from bodo.libs.distributed_api import create_subcomm_mpi4py, get_host_ranks, get_nodes_first_ranks
dl_status = None


def assert_dl_initialized():
    assert dl_status is not None, 'Horovod has not been initialized. Call bodo.dl.start() first'


class DLStatus(object):

    def __init__(self, framework, gpu_ranks):
        self.framework = framework
        self.gpu_ranks = gpu_ranks


def get_num_gpus(framework):
    if framework == 'torch':
        import torch
        return torch.cuda.device_count()
    elif framework == 'tensorflow':
        import tensorflow as tf
        return len(tf.config.experimental.list_physical_devices('GPU'))
    else:
        raise RuntimeError('Framework {} not recognized'.format(framework))


def get_gpu_ranks(framework):
    xue__ardb = MPI.COMM_WORLD
    ojsck__zvwi = xue__ardb.Get_rank()
    uyhr__vdu = get_host_ranks()
    ntw__juwck = get_nodes_first_ranks()
    if ojsck__zvwi in ntw__juwck:
        try:
            mtc__jaq = get_num_gpus(framework)
        except Exception as gjs__ndr:
            mtc__jaq = gjs__ndr
        jbe__suc = create_subcomm_mpi4py(ntw__juwck)
        qqwf__qir = jbe__suc.gather(mtc__jaq)
        if ojsck__zvwi == 0:
            gpu_ranks = []
            zij__jvl = None
            for ndurj__qxm, pglyt__wsee in enumerate(uyhr__vdu.values()):
                mirx__hjrth = qqwf__qir[ndurj__qxm]
                if isinstance(mirx__hjrth, Exception):
                    zij__jvl = mirx__hjrth
                    break
                if mirx__hjrth == 0:
                    continue
                vcwho__poc = len(pglyt__wsee) // mirx__hjrth
                for ggpj__ohioi, kzf__rcgl in enumerate(pglyt__wsee):
                    if ggpj__ohioi % vcwho__poc == 0:
                        iecwn__erc = ggpj__ohioi / vcwho__poc
                        if iecwn__erc < mirx__hjrth:
                            gpu_ranks.append(kzf__rcgl)
            if zij__jvl:
                xue__ardb.bcast(zij__jvl)
                raise zij__jvl
            else:
                xue__ardb.bcast(gpu_ranks)
    if ojsck__zvwi != 0:
        gpu_ranks = xue__ardb.bcast(None)
        if isinstance(gpu_ranks, Exception):
            gjs__ndr = gpu_ranks
            raise gjs__ndr
    return gpu_ranks


def is_cuda_available():
    assert_dl_initialized()
    return len(dl_status.gpu_ranks) > 0


def initialize_horovod(framework):
    global dl_status
    if dl_status is not None:
        assert dl_status.framework == framework, 'Attempted to initialize Horovod with different DL frameworks'
        return np.array(dl_status.gpu_ranks, dtype=np.int32)
    gpu_ranks = get_gpu_ranks(framework)
    if framework == 'torch':
        import horovod.torch as hvd
        import torch
        torch.set_num_threads(1)
    elif framework == 'tensorflow':
        import horovod.tensorflow as hvd
        import tensorflow as tf
    else:
        raise RuntimeError('Framework {} not recognized'.format(framework))
    rmaqm__avd = MPI.COMM_WORLD.rank
    if len(gpu_ranks) > 0:
        jbe__suc = MPI.COMM_WORLD.Split(color=0 if rmaqm__avd in gpu_ranks else
            MPI.UNDEFINED, key=rmaqm__avd)
        if jbe__suc != MPI.COMM_NULL:
            hvd.init(comm=jbe__suc)
            if framework == 'torch':
                torch.cuda.set_device(hvd.local_rank())
            elif framework == 'tensorflow':
                gvd__nnx = tf.config.experimental.list_physical_devices('GPU')
                for ruc__mlall in gvd__nnx:
                    tf.config.experimental.set_memory_growth(ruc__mlall, True)
                tf.config.experimental.set_visible_devices(gvd__nnx[hvd.
                    local_rank()], 'GPU')
    else:
        if rmaqm__avd == 0:
            print('[BODO-DL]: No GPUs found in cluster. Using CPUs')
        hvd.init()
    dl_status = DLStatus(framework, np.array(gpu_ranks, dtype=np.int32))


@numba.njit
def start(framework):
    with numba.objmode:
        initialize_horovod(framework)


@numba.njit
def end():
    with numba.objmode:
        end_py()


def end_py():
    if is_cuda_available():
        nxole__pnjbd = 17
        xue__ardb = MPI.COMM_WORLD
        momt__ujk = MPI.Get_processor_name()
        pvvzd__kwj = get_host_ranks()[momt__ujk]
        assert_dl_initialized()
        if bodo.get_rank() == pvvzd__kwj[0]:
            assert bodo.get_rank() in dl_status.gpu_ranks
            for ojsck__zvwi in pvvzd__kwj[1:]:
                xue__ardb.isend(1, dest=ojsck__zvwi, tag=nxole__pnjbd)
        else:
            while True:
                lkgzw__rzhje = MPI.Status()
                jro__ssqm = xue__ardb.Iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG,
                    lkgzw__rzhje)
                if jro__ssqm:
                    assert lkgzw__rzhje.source == pvvzd__kwj[0]
                    assert lkgzw__rzhje.tag == nxole__pnjbd
                    xue__ardb.recv(source=0, tag=nxole__pnjbd)
                    break
                time.sleep(1.0)
    else:
        bodo.barrier()


def _prepare_data_get_gpu_ranks():
    assert_dl_initialized()
    return dl_status.gpu_ranks


@numba.njit
def prepare_data(data):
    with numba.objmode(gpu_ranks='int32[:]'):
        gpu_ranks = _prepare_data_get_gpu_ranks()
    if len(gpu_ranks) > 0:
        data = bodo.rebalance(data, dests=list(gpu_ranks), parallel=True)
    else:
        data = bodo.rebalance(data, parallel=True)
    return data
