"""
Boxing and unboxing support for DataFrame, Series, etc.
"""
import datetime
import decimal
import warnings
from enum import Enum
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.ir_utils import GuardException, guard
from numba.core.typing import signature
from numba.cpython.listobj import ListInstance
from numba.extending import NativeValue, box, intrinsic, typeof_impl, unbox
from numba.np import numpy_support
from numba.typed.typeddict import Dict
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.datetime_timedelta_ext import datetime_timedelta_array_type
from bodo.hiframes.pd_categorical_ext import PDCategoricalDtype
from bodo.hiframes.pd_dataframe_ext import DataFramePayloadType, DataFrameType, check_runtime_cols_unsupported, construct_dataframe
from bodo.hiframes.pd_index_ext import BinaryIndexType, CategoricalIndexType, DatetimeIndexType, NumericIndexType, PeriodIndexType, RangeIndexType, StringIndexType, TimedeltaIndexType
from bodo.hiframes.pd_series_ext import HeterogeneousSeriesType, SeriesType
from bodo.hiframes.split_impl import string_array_split_view_type
from bodo.libs import hstr_ext
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.decimal_arr_ext import Decimal128Type, DecimalArrayType
from bodo.libs.int_arr_ext import IntDtype, IntegerArrayType, typeof_pd_int_dtype
from bodo.libs.map_arr_ext import MapArrayType
from bodo.libs.str_arr_ext import string_array_type, string_type
from bodo.libs.str_ext import string_type
from bodo.libs.struct_arr_ext import StructArrayType, StructType
from bodo.libs.tuple_arr_ext import TupleArrayType
from bodo.utils.cg_helpers import is_ll_eq
from bodo.utils.typing import BodoError, BodoWarning, dtype_to_array_type, get_overload_const_bool, get_overload_const_int, get_overload_const_str, is_overload_constant_bool, is_overload_constant_int, is_overload_constant_str, raise_bodo_error, to_nullable_type
ll.add_symbol('array_size', hstr_ext.array_size)
ll.add_symbol('array_getptr1', hstr_ext.array_getptr1)
TABLE_FORMAT_THRESHOLD = 20


def _set_bodo_meta_in_pandas():
    if '_bodo_meta' not in pd.Series._metadata:
        pd.Series._metadata.append('_bodo_meta')
    if '_bodo_meta' not in pd.DataFrame._metadata:
        pd.DataFrame._metadata.append('_bodo_meta')


_set_bodo_meta_in_pandas()


@typeof_impl.register(pd.DataFrame)
def typeof_pd_dataframe(val, c):
    from bodo.transforms.distributed_analysis import Distribution
    ezvlr__vks = tuple(val.columns.to_list())
    sobni__ynl = get_hiframes_dtypes(val)
    if (len(val.index) == 0 and val.index.dtype == np.dtype('O') and
        hasattr(val, '_bodo_meta') and val._bodo_meta is not None and 
        'type_metadata' in val._bodo_meta and val._bodo_meta[
        'type_metadata'] is not None and len(val._bodo_meta['type_metadata'
        ][1]) == len(val.columns) and val._bodo_meta['type_metadata'][0] is not
        None):
        bjl__gihd = _dtype_from_type_enum_list(val._bodo_meta[
            'type_metadata'][0])
    else:
        bjl__gihd = numba.typeof(val.index)
    leu__jboae = Distribution(val._bodo_meta['dist']) if hasattr(val,
        '_bodo_meta') and val._bodo_meta is not None else Distribution.REP
    jxak__ksulg = len(sobni__ynl) >= TABLE_FORMAT_THRESHOLD
    return DataFrameType(sobni__ynl, bjl__gihd, ezvlr__vks, leu__jboae,
        is_table_format=jxak__ksulg)


@typeof_impl.register(pd.Series)
def typeof_pd_series(val, c):
    from bodo.transforms.distributed_analysis import Distribution
    leu__jboae = Distribution(val._bodo_meta['dist']) if hasattr(val,
        '_bodo_meta') and val._bodo_meta is not None else Distribution.REP
    if (len(val.index) == 0 and val.index.dtype == np.dtype('O') and
        hasattr(val, '_bodo_meta') and val._bodo_meta is not None and 
        'type_metadata' in val._bodo_meta and val._bodo_meta[
        'type_metadata'] is not None and val._bodo_meta['type_metadata'][0]
         is not None):
        ued__tjs = _dtype_from_type_enum_list(val._bodo_meta[
            'type_metadata'][0])
    else:
        ued__tjs = numba.typeof(val.index)
    return SeriesType(_infer_series_dtype(val), index=ued__tjs, name_typ=
        numba.typeof(val.name), dist=leu__jboae)


@unbox(DataFrameType)
def unbox_dataframe(typ, val, c):
    check_runtime_cols_unsupported(typ, 'Unboxing')
    lfnss__arxf = c.pyapi.object_getattr_string(val, 'index')
    ywjle__pwsb = c.pyapi.to_native_value(typ.index, lfnss__arxf).value
    c.pyapi.decref(lfnss__arxf)
    if typ.is_table_format:
        bbphf__ysy = cgutils.create_struct_proxy(typ.table_type)(c.context,
            c.builder)
        bbphf__ysy.parent = val
        for fesqc__gxvti, bew__vgcvu in typ.table_type.type_to_blk.items():
            hpc__bjet = c.context.get_constant(types.int64, len(typ.
                table_type.block_to_arr_ind[bew__vgcvu]))
            rqrh__gznfg, dgdbi__xmu = ListInstance.allocate_ex(c.context, c
                .builder, types.List(fesqc__gxvti), hpc__bjet)
            dgdbi__xmu.size = hpc__bjet
            setattr(bbphf__ysy, f'block_{bew__vgcvu}', dgdbi__xmu.value)
        jvryi__xalb = c.pyapi.call_method(val, '__len__', ())
        aeptw__spk = c.pyapi.long_as_longlong(jvryi__xalb)
        c.pyapi.decref(jvryi__xalb)
        bbphf__ysy.len = aeptw__spk
        pgw__cyi = c.context.make_tuple(c.builder, types.Tuple([typ.
            table_type]), [bbphf__ysy._getvalue()])
    else:
        tkxxt__agl = [c.context.get_constant_null(fesqc__gxvti) for
            fesqc__gxvti in typ.data]
        pgw__cyi = c.context.make_tuple(c.builder, types.Tuple(typ.data),
            tkxxt__agl)
    okwem__gxjf = construct_dataframe(c.context, c.builder, typ, pgw__cyi,
        ywjle__pwsb, val, None)
    return NativeValue(okwem__gxjf)


def get_hiframes_dtypes(df):
    if (hasattr(df, '_bodo_meta') and df._bodo_meta is not None and 
        'type_metadata' in df._bodo_meta and df._bodo_meta['type_metadata']
         is not None and len(df._bodo_meta['type_metadata'][1]) == len(df.
        columns)):
        ohjt__okslu = df._bodo_meta['type_metadata'][1]
    else:
        ohjt__okslu = [None] * len(df.columns)
    edfh__tihak = [dtype_to_array_type(_infer_series_dtype(df.iloc[:, i],
        array_metadata=ohjt__okslu[i])) for i in range(len(df.columns))]
    return tuple(edfh__tihak)


class SeriesDtypeEnum(Enum):
    Int8 = 0
    UInt8 = 1
    Int32 = 2
    UInt32 = 3
    Int64 = 4
    UInt64 = 7
    Float32 = 5
    Float64 = 6
    Int16 = 8
    UInt16 = 9
    STRING = 10
    Bool = 11
    Decimal = 12
    Datime_Date = 13
    NP_Datetime64ns = 14
    NP_Timedelta64ns = 15
    Int128 = 16
    LIST = 18
    STRUCT = 19
    BINARY = 21
    ARRAY = 22
    PD_nullable_Int8 = 23
    PD_nullable_UInt8 = 24
    PD_nullable_Int16 = 25
    PD_nullable_UInt16 = 26
    PD_nullable_Int32 = 27
    PD_nullable_UInt32 = 28
    PD_nullable_Int64 = 29
    PD_nullable_UInt64 = 30
    PD_nullable_bool = 31
    CategoricalType = 32
    NoneType = 33
    Literal = 34
    IntegerArray = 35
    RangeIndexType = 36
    DatetimeIndexType = 37
    NumericIndexType = 38
    PeriodIndexType = 39
    IntervalIndexType = 40
    CategoricalIndexType = 41
    StringIndexType = 42
    BinaryIndexType = 43
    TimedeltaIndexType = 44
    LiteralType = 45


_one_to_one_type_to_enum_map = {types.int8: SeriesDtypeEnum.Int8.value,
    types.uint8: SeriesDtypeEnum.UInt8.value, types.int32: SeriesDtypeEnum.
    Int32.value, types.uint32: SeriesDtypeEnum.UInt32.value, types.int64:
    SeriesDtypeEnum.Int64.value, types.uint64: SeriesDtypeEnum.UInt64.value,
    types.float32: SeriesDtypeEnum.Float32.value, types.float64:
    SeriesDtypeEnum.Float64.value, types.NPDatetime('ns'): SeriesDtypeEnum.
    NP_Datetime64ns.value, types.NPTimedelta('ns'): SeriesDtypeEnum.
    NP_Timedelta64ns.value, types.bool_: SeriesDtypeEnum.Bool.value, types.
    int16: SeriesDtypeEnum.Int16.value, types.uint16: SeriesDtypeEnum.
    UInt16.value, types.Integer('int128', 128): SeriesDtypeEnum.Int128.
    value, bodo.hiframes.datetime_date_ext.datetime_date_type:
    SeriesDtypeEnum.Datime_Date.value, IntDtype(types.int8):
    SeriesDtypeEnum.PD_nullable_Int8.value, IntDtype(types.uint8):
    SeriesDtypeEnum.PD_nullable_UInt8.value, IntDtype(types.int16):
    SeriesDtypeEnum.PD_nullable_Int16.value, IntDtype(types.uint16):
    SeriesDtypeEnum.PD_nullable_UInt16.value, IntDtype(types.int32):
    SeriesDtypeEnum.PD_nullable_Int32.value, IntDtype(types.uint32):
    SeriesDtypeEnum.PD_nullable_UInt32.value, IntDtype(types.int64):
    SeriesDtypeEnum.PD_nullable_Int64.value, IntDtype(types.uint64):
    SeriesDtypeEnum.PD_nullable_UInt64.value, bytes_type: SeriesDtypeEnum.
    BINARY.value, string_type: SeriesDtypeEnum.STRING.value, bodo.bool_:
    SeriesDtypeEnum.Bool.value, types.none: SeriesDtypeEnum.NoneType.value}
_one_to_one_enum_to_type_map = {SeriesDtypeEnum.Int8.value: types.int8,
    SeriesDtypeEnum.UInt8.value: types.uint8, SeriesDtypeEnum.Int32.value:
    types.int32, SeriesDtypeEnum.UInt32.value: types.uint32,
    SeriesDtypeEnum.Int64.value: types.int64, SeriesDtypeEnum.UInt64.value:
    types.uint64, SeriesDtypeEnum.Float32.value: types.float32,
    SeriesDtypeEnum.Float64.value: types.float64, SeriesDtypeEnum.
    NP_Datetime64ns.value: types.NPDatetime('ns'), SeriesDtypeEnum.
    NP_Timedelta64ns.value: types.NPTimedelta('ns'), SeriesDtypeEnum.Int16.
    value: types.int16, SeriesDtypeEnum.UInt16.value: types.uint16,
    SeriesDtypeEnum.Int128.value: types.Integer('int128', 128),
    SeriesDtypeEnum.Datime_Date.value: bodo.hiframes.datetime_date_ext.
    datetime_date_type, SeriesDtypeEnum.PD_nullable_Int8.value: IntDtype(
    types.int8), SeriesDtypeEnum.PD_nullable_UInt8.value: IntDtype(types.
    uint8), SeriesDtypeEnum.PD_nullable_Int16.value: IntDtype(types.int16),
    SeriesDtypeEnum.PD_nullable_UInt16.value: IntDtype(types.uint16),
    SeriesDtypeEnum.PD_nullable_Int32.value: IntDtype(types.int32),
    SeriesDtypeEnum.PD_nullable_UInt32.value: IntDtype(types.uint32),
    SeriesDtypeEnum.PD_nullable_Int64.value: IntDtype(types.int64),
    SeriesDtypeEnum.PD_nullable_UInt64.value: IntDtype(types.uint64),
    SeriesDtypeEnum.BINARY.value: bytes_type, SeriesDtypeEnum.STRING.value:
    string_type, SeriesDtypeEnum.Bool.value: bodo.bool_, SeriesDtypeEnum.
    NoneType.value: types.none}


def _dtype_from_type_enum_list(typ_enum_list):
    nev__upu, typ = _dtype_from_type_enum_list_recursor(typ_enum_list)
    if len(nev__upu) != 0:
        raise_bodo_error(
            f"""Unexpected Internal Error while converting typing metadata: Dtype list was not fully consumed.
 Input typ_enum_list: {typ_enum_list}.
Remainder: {nev__upu}. Please file the error here: https://github.com/Bodo-inc/Feedback"""
            )
    return typ


def _dtype_from_type_enum_list_recursor(typ_enum_list):
    if len(typ_enum_list) == 0:
        raise_bodo_error('Unable to infer dtype from empty typ_enum_list')
    elif typ_enum_list[0] in _one_to_one_enum_to_type_map:
        return typ_enum_list[1:], _one_to_one_enum_to_type_map[typ_enum_list[0]
            ]
    elif typ_enum_list[0] == SeriesDtypeEnum.IntegerArray.value:
        zrb__kajfk, typ = _dtype_from_type_enum_list_recursor(typ_enum_list[1:]
            )
        return zrb__kajfk, IntegerArrayType(typ)
    elif typ_enum_list[0] == SeriesDtypeEnum.ARRAY.value:
        zrb__kajfk, typ = _dtype_from_type_enum_list_recursor(typ_enum_list[1:]
            )
        return zrb__kajfk, dtype_to_array_type(typ)
    elif typ_enum_list[0] == SeriesDtypeEnum.Decimal.value:
        xwmq__muw = typ_enum_list[1]
        ctzea__ixeu = typ_enum_list[2]
        return typ_enum_list[3:], Decimal128Type(xwmq__muw, ctzea__ixeu)
    elif typ_enum_list[0] == SeriesDtypeEnum.STRUCT.value:
        rxi__xdedc = typ_enum_list[1]
        ngo__jdr = tuple(typ_enum_list[2:2 + rxi__xdedc])
        xwk__zrzi = typ_enum_list[2 + rxi__xdedc:]
        ltb__stii = []
        for i in range(rxi__xdedc):
            xwk__zrzi, rgrjo__wqfn = _dtype_from_type_enum_list_recursor(
                xwk__zrzi)
            ltb__stii.append(rgrjo__wqfn)
        return xwk__zrzi, StructType(tuple(ltb__stii), ngo__jdr)
    elif typ_enum_list[0] == SeriesDtypeEnum.Literal.value:
        if len(typ_enum_list) == 1:
            raise_bodo_error(
                f"Unexpected Internal Error while converting typing metadata: Encountered 'Literal' internal enum value with no value following it. Please file the error here: https://github.com/Bodo-inc/Feedback"
                )
        fmvkl__ocp = typ_enum_list[1]
        xwk__zrzi = typ_enum_list[2:]
        return xwk__zrzi, fmvkl__ocp
    elif typ_enum_list[0] == SeriesDtypeEnum.LiteralType.value:
        if len(typ_enum_list) == 1:
            raise_bodo_error(
                f"Unexpected Internal Error while converting typing metadata: Encountered 'LiteralType' internal enum value with no value following it. Please file the error here: https://github.com/Bodo-inc/Feedback"
                )
        fmvkl__ocp = typ_enum_list[1]
        xwk__zrzi = typ_enum_list[2:]
        return xwk__zrzi, numba.types.literal(fmvkl__ocp)
    elif typ_enum_list[0] == SeriesDtypeEnum.CategoricalType.value:
        xwk__zrzi, tbmew__ezgv = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        xwk__zrzi, nuxkt__lrph = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        xwk__zrzi, sdlng__leug = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        xwk__zrzi, mzl__suxn = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        xwk__zrzi, uhxq__lier = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        return xwk__zrzi, PDCategoricalDtype(tbmew__ezgv, nuxkt__lrph,
            sdlng__leug, mzl__suxn, uhxq__lier)
    elif typ_enum_list[0] == SeriesDtypeEnum.DatetimeIndexType.value:
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        return xwk__zrzi, DatetimeIndexType(nbx__iraa)
    elif typ_enum_list[0] == SeriesDtypeEnum.NumericIndexType.value:
        xwk__zrzi, dtype = _dtype_from_type_enum_list_recursor(typ_enum_list
            [1:])
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        xwk__zrzi, mzl__suxn = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        return xwk__zrzi, NumericIndexType(dtype, nbx__iraa, mzl__suxn)
    elif typ_enum_list[0] == SeriesDtypeEnum.PeriodIndexType.value:
        xwk__zrzi, zsyht__yhpr = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        return xwk__zrzi, PeriodIndexType(zsyht__yhpr, nbx__iraa)
    elif typ_enum_list[0] == SeriesDtypeEnum.CategoricalIndexType.value:
        xwk__zrzi, mzl__suxn = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(xwk__zrzi)
        return xwk__zrzi, CategoricalIndexType(mzl__suxn, nbx__iraa)
    elif typ_enum_list[0] == SeriesDtypeEnum.RangeIndexType.value:
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        return xwk__zrzi, RangeIndexType(nbx__iraa)
    elif typ_enum_list[0] == SeriesDtypeEnum.StringIndexType.value:
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        return xwk__zrzi, StringIndexType(nbx__iraa)
    elif typ_enum_list[0] == SeriesDtypeEnum.BinaryIndexType.value:
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        return xwk__zrzi, BinaryIndexType(nbx__iraa)
    elif typ_enum_list[0] == SeriesDtypeEnum.TimedeltaIndexType.value:
        xwk__zrzi, nbx__iraa = _dtype_from_type_enum_list_recursor(
            typ_enum_list[1:])
        return xwk__zrzi, TimedeltaIndexType(nbx__iraa)
    else:
        raise_bodo_error(
            f'Unexpected Internal Error while converting typing metadata: unable to infer dtype for type enum {typ_enum_list[0]}. Please file the error here: https://github.com/Bodo-inc/Feedback'
            )


def _dtype_to_type_enum_list(typ):
    return guard(_dtype_to_type_enum_list_recursor, typ)


def _dtype_to_type_enum_list_recursor(typ, upcast_numeric_index=True):
    if typ.__hash__ and typ in _one_to_one_type_to_enum_map:
        return [_one_to_one_type_to_enum_map[typ]]
    if isinstance(typ, (dict, int, list, tuple, str, bool, bytes, float)):
        return [SeriesDtypeEnum.Literal.value, typ]
    elif typ is None:
        return [SeriesDtypeEnum.Literal.value, typ]
    elif is_overload_constant_int(typ):
        jnsge__cmox = get_overload_const_int(typ)
        if numba.types.maybe_literal(jnsge__cmox) == typ:
            return [SeriesDtypeEnum.LiteralType.value, jnsge__cmox]
    elif is_overload_constant_str(typ):
        jnsge__cmox = get_overload_const_str(typ)
        if numba.types.maybe_literal(jnsge__cmox) == typ:
            return [SeriesDtypeEnum.LiteralType.value, jnsge__cmox]
    elif is_overload_constant_bool(typ):
        jnsge__cmox = get_overload_const_bool(typ)
        if numba.types.maybe_literal(jnsge__cmox) == typ:
            return [SeriesDtypeEnum.LiteralType.value, jnsge__cmox]
    elif isinstance(typ, IntegerArrayType):
        return [SeriesDtypeEnum.IntegerArray.value
            ] + _dtype_to_type_enum_list_recursor(typ.dtype)
    elif bodo.utils.utils.is_array_typ(typ, False):
        return [SeriesDtypeEnum.ARRAY.value
            ] + _dtype_to_type_enum_list_recursor(typ.dtype)
    elif isinstance(typ, StructType):
        bvo__jyj = [SeriesDtypeEnum.STRUCT.value, len(typ.names)]
        for bvp__xai in typ.names:
            bvo__jyj.append(bvp__xai)
        for rznh__agj in typ.data:
            bvo__jyj += _dtype_to_type_enum_list_recursor(rznh__agj)
        return bvo__jyj
    elif isinstance(typ, bodo.libs.decimal_arr_ext.Decimal128Type):
        return [SeriesDtypeEnum.Decimal.value, typ.precision, typ.scale]
    elif isinstance(typ, PDCategoricalDtype):
        ilr__yqas = _dtype_to_type_enum_list_recursor(typ.categories)
        lttsj__gpfz = _dtype_to_type_enum_list_recursor(typ.elem_type)
        kzz__nxe = _dtype_to_type_enum_list_recursor(typ.ordered)
        zujpd__mmkm = _dtype_to_type_enum_list_recursor(typ.data)
        gfdq__ahp = _dtype_to_type_enum_list_recursor(typ.int_type)
        return [SeriesDtypeEnum.CategoricalType.value
            ] + ilr__yqas + lttsj__gpfz + kzz__nxe + zujpd__mmkm + gfdq__ahp
    elif isinstance(typ, DatetimeIndexType):
        return [SeriesDtypeEnum.DatetimeIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.name_typ)
    elif isinstance(typ, NumericIndexType):
        if upcast_numeric_index:
            if isinstance(typ.dtype, types.Float):
                jmjqc__pgkj = types.float64
                spz__hdols = types.Array(jmjqc__pgkj, 1, 'C')
            elif typ.dtype in {types.int8, types.int16, types.int32, types.
                int64}:
                jmjqc__pgkj = types.int64
                spz__hdols = types.Array(jmjqc__pgkj, 1, 'C')
            elif typ.dtype in {types.uint8, types.uint16, types.uint32,
                types.uint64}:
                jmjqc__pgkj = types.uint64
                spz__hdols = types.Array(jmjqc__pgkj, 1, 'C')
            elif typ.dtype == types.bool_:
                jmjqc__pgkj = typ.dtype
                spz__hdols = typ.data
            else:
                raise GuardException('Unable to convert type')
            return [SeriesDtypeEnum.NumericIndexType.value
                ] + _dtype_to_type_enum_list_recursor(jmjqc__pgkj
                ) + _dtype_to_type_enum_list_recursor(typ.name_typ
                ) + _dtype_to_type_enum_list_recursor(spz__hdols)
        else:
            return [SeriesDtypeEnum.NumericIndexType.value
                ] + _dtype_to_type_enum_list_recursor(typ.dtype
                ) + _dtype_to_type_enum_list_recursor(typ.name_typ
                ) + _dtype_to_type_enum_list_recursor(typ.data)
    elif isinstance(typ, PeriodIndexType):
        return [SeriesDtypeEnum.PeriodIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.freq
            ) + _dtype_to_type_enum_list_recursor(typ.name_typ)
    elif isinstance(typ, CategoricalIndexType):
        return [SeriesDtypeEnum.CategoricalIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.data
            ) + _dtype_to_type_enum_list_recursor(typ.name_typ)
    elif isinstance(typ, RangeIndexType):
        return [SeriesDtypeEnum.RangeIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.name_typ)
    elif isinstance(typ, StringIndexType):
        return [SeriesDtypeEnum.StringIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.name_typ)
    elif isinstance(typ, BinaryIndexType):
        return [SeriesDtypeEnum.BinaryIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.name_typ)
    elif isinstance(typ, TimedeltaIndexType):
        return [SeriesDtypeEnum.TimedeltaIndexType.value
            ] + _dtype_to_type_enum_list_recursor(typ.name_typ)
    else:
        raise GuardException('Unable to convert type')


def _infer_series_dtype(S, array_metadata=None):
    if S.dtype == np.dtype('O'):
        if len(S.values) == 0:
            if (hasattr(S, '_bodo_meta') and S._bodo_meta is not None and 
                'type_metadata' in S._bodo_meta and S._bodo_meta[
                'type_metadata'][1] is not None):
                jjj__nydj = S._bodo_meta['type_metadata'][1]
                return _dtype_from_type_enum_list(jjj__nydj)
            elif array_metadata != None:
                return _dtype_from_type_enum_list(array_metadata).dtype
        return numba.typeof(S.values).dtype
    if isinstance(S.dtype, pd.core.arrays.integer._IntegerDtype):
        return typeof_pd_int_dtype(S.dtype, None)
    elif isinstance(S.dtype, pd.CategoricalDtype):
        return bodo.typeof(S.dtype)
    elif isinstance(S.dtype, pd.StringDtype):
        return string_type
    elif isinstance(S.dtype, pd.BooleanDtype):
        return types.bool_
    if isinstance(S.dtype, pd.DatetimeTZDtype):
        raise BodoError('Timezone-aware datetime data type not supported yet')
    try:
        return numpy_support.from_dtype(S.dtype)
    except:
        raise BodoError(
            f'data type {S.dtype} for column {S.name} not supported yet')


def _get_use_df_parent_obj_flag(builder, context, pyapi, parent_obj, n_cols):
    if n_cols is None:
        return context.get_constant(types.bool_, False)
    kzid__pgzx = cgutils.is_not_null(builder, parent_obj)
    shpcj__lilgo = cgutils.alloca_once_value(builder, context.get_constant(
        types.int64, 0))
    with builder.if_then(kzid__pgzx):
        gwo__ddtyg = pyapi.object_getattr_string(parent_obj, 'columns')
        jvryi__xalb = pyapi.call_method(gwo__ddtyg, '__len__', ())
        builder.store(pyapi.long_as_longlong(jvryi__xalb), shpcj__lilgo)
        pyapi.decref(jvryi__xalb)
        pyapi.decref(gwo__ddtyg)
    use_parent_obj = builder.and_(kzid__pgzx, builder.icmp_unsigned('==',
        builder.load(shpcj__lilgo), context.get_constant(types.int64, n_cols)))
    return use_parent_obj


def _get_df_columns_obj(c, builder, context, pyapi, df_typ, dataframe_payload):
    if df_typ.has_runtime_cols:
        rwysq__tbwgl = df_typ.runtime_colname_typ
        context.nrt.incref(builder, rwysq__tbwgl, dataframe_payload.columns)
        return pyapi.from_native_value(rwysq__tbwgl, dataframe_payload.
            columns, c.env_manager)
    if all(isinstance(c, int) for c in df_typ.columns):
        machj__xaf = np.array(df_typ.columns, 'int64')
    elif all(isinstance(c, str) for c in df_typ.columns):
        machj__xaf = pd.array(df_typ.columns, 'string')
    else:
        machj__xaf = df_typ.columns
    cnyxb__gkhme = numba.typeof(machj__xaf)
    vvjml__pddyz = context.get_constant_generic(builder, cnyxb__gkhme,
        machj__xaf)
    emhyj__gggo = pyapi.from_native_value(cnyxb__gkhme, vvjml__pddyz, c.
        env_manager)
    return emhyj__gggo


def _create_initial_df_object(builder, context, pyapi, c, df_typ, obj,
    dataframe_payload, res, use_parent_obj):
    with c.builder.if_else(use_parent_obj) as (use_parent, otherwise):
        with use_parent:
            pyapi.incref(obj)
            kgnkr__ukil = context.insert_const_string(c.builder.module, 'numpy'
                )
            ufdev__ttd = pyapi.import_module_noblock(kgnkr__ukil)
            if df_typ.has_runtime_cols:
                ohzxl__rcmm = 0
            else:
                ohzxl__rcmm = len(df_typ.columns)
            djiav__lpjus = pyapi.long_from_longlong(lir.Constant(lir.
                IntType(64), ohzxl__rcmm))
            bcfvn__yjm = pyapi.call_method(ufdev__ttd, 'arange', (
                djiav__lpjus,))
            pyapi.object_setattr_string(obj, 'columns', bcfvn__yjm)
            pyapi.decref(ufdev__ttd)
            pyapi.decref(bcfvn__yjm)
            pyapi.decref(djiav__lpjus)
        with otherwise:
            context.nrt.incref(builder, df_typ.index, dataframe_payload.index)
            dxmcv__fshl = c.pyapi.from_native_value(df_typ.index,
                dataframe_payload.index, c.env_manager)
            kgnkr__ukil = context.insert_const_string(c.builder.module,
                'pandas')
            ufdev__ttd = pyapi.import_module_noblock(kgnkr__ukil)
            df_obj = pyapi.call_method(ufdev__ttd, 'DataFrame', (pyapi.
                borrow_none(), dxmcv__fshl))
            pyapi.decref(ufdev__ttd)
            pyapi.decref(dxmcv__fshl)
            builder.store(df_obj, res)


@box(DataFrameType)
def box_dataframe(typ, val, c):
    from bodo.hiframes.table import box_table
    context = c.context
    builder = c.builder
    pyapi = c.pyapi
    dataframe_payload = bodo.hiframes.pd_dataframe_ext.get_dataframe_payload(c
        .context, c.builder, typ, val)
    qtfad__iokky = cgutils.create_struct_proxy(typ)(context, builder, value=val
        )
    n_cols = len(typ.columns) if not typ.has_runtime_cols else None
    obj = qtfad__iokky.parent
    res = cgutils.alloca_once_value(builder, obj)
    use_parent_obj = _get_use_df_parent_obj_flag(builder, context, pyapi,
        obj, n_cols)
    _create_initial_df_object(builder, context, pyapi, c, typ, obj,
        dataframe_payload, res, use_parent_obj)
    if typ.is_table_format:
        trdv__qgfq = typ.table_type
        bbphf__ysy = builder.extract_value(dataframe_payload.data, 0)
        context.nrt.incref(builder, trdv__qgfq, bbphf__ysy)
        hpu__pghho = box_table(trdv__qgfq, bbphf__ysy, c, builder.not_(
            use_parent_obj))
        with builder.if_else(use_parent_obj) as (then, orelse):
            with then:
                kxs__pfzn = pyapi.object_getattr_string(hpu__pghho, 'arrays')
                ooadw__kjyl = c.pyapi.make_none()
                if n_cols is None:
                    jvryi__xalb = pyapi.call_method(kxs__pfzn, '__len__', ())
                    hpc__bjet = pyapi.long_as_longlong(jvryi__xalb)
                    pyapi.decref(jvryi__xalb)
                else:
                    hpc__bjet = context.get_constant(types.int64, n_cols)
                with cgutils.for_range(builder, hpc__bjet) as loop:
                    i = loop.index
                    wjeep__usy = pyapi.list_getitem(kxs__pfzn, i)
                    xlwh__lneff = c.builder.icmp_unsigned('!=', wjeep__usy,
                        ooadw__kjyl)
                    with builder.if_then(xlwh__lneff):
                        wgts__jquxt = pyapi.long_from_longlong(i)
                        df_obj = builder.load(res)
                        pyapi.object_setitem(df_obj, wgts__jquxt, wjeep__usy)
                        pyapi.decref(wgts__jquxt)
                pyapi.decref(kxs__pfzn)
                pyapi.decref(ooadw__kjyl)
            with orelse:
                df_obj = builder.load(res)
                dxmcv__fshl = pyapi.object_getattr_string(df_obj, 'index')
                wjzc__kxywg = c.pyapi.call_method(hpu__pghho, 'to_pandas',
                    (dxmcv__fshl,))
                builder.store(wjzc__kxywg, res)
                pyapi.decref(df_obj)
                pyapi.decref(dxmcv__fshl)
        pyapi.decref(hpu__pghho)
    else:
        uwo__tcb = [builder.extract_value(dataframe_payload.data, i) for i in
            range(n_cols)]
        zoxp__oqoi = typ.data
        for i, kvjm__qzv, plea__pgzfp in zip(range(n_cols), uwo__tcb,
            zoxp__oqoi):
            cta__mklf = cgutils.alloca_once_value(builder, kvjm__qzv)
            hqtf__mbtfu = cgutils.alloca_once_value(builder, context.
                get_constant_null(plea__pgzfp))
            xlwh__lneff = builder.not_(is_ll_eq(builder, cta__mklf,
                hqtf__mbtfu))
            pyls__lgpt = builder.or_(builder.not_(use_parent_obj), builder.
                and_(use_parent_obj, xlwh__lneff))
            with builder.if_then(pyls__lgpt):
                wgts__jquxt = pyapi.long_from_longlong(context.get_constant
                    (types.int64, i))
                context.nrt.incref(builder, plea__pgzfp, kvjm__qzv)
                arr_obj = pyapi.from_native_value(plea__pgzfp, kvjm__qzv, c
                    .env_manager)
                df_obj = builder.load(res)
                pyapi.object_setitem(df_obj, wgts__jquxt, arr_obj)
                pyapi.decref(arr_obj)
                pyapi.decref(wgts__jquxt)
    df_obj = builder.load(res)
    emhyj__gggo = _get_df_columns_obj(c, builder, context, pyapi, typ,
        dataframe_payload)
    pyapi.object_setattr_string(df_obj, 'columns', emhyj__gggo)
    pyapi.decref(emhyj__gggo)
    _set_bodo_meta_dataframe(c, df_obj, typ)
    c.context.nrt.decref(c.builder, typ, val)
    return df_obj


def get_df_obj_column_codegen(context, builder, pyapi, df_obj, col_ind,
    data_typ):
    ooadw__kjyl = pyapi.borrow_none()
    qcn__ibcx = pyapi.unserialize(pyapi.serialize_object(slice))
    mpvxe__usme = pyapi.call_function_objargs(qcn__ibcx, [ooadw__kjyl])
    tdz__tojbv = pyapi.long_from_longlong(col_ind)
    qpary__vuab = pyapi.tuple_pack([mpvxe__usme, tdz__tojbv])
    zpju__cutor = pyapi.object_getattr_string(df_obj, 'iloc')
    ose__bidy = pyapi.object_getitem(zpju__cutor, qpary__vuab)
    ffb__ymv = pyapi.object_getattr_string(ose__bidy, 'values')
    if isinstance(data_typ, types.Array):
        umsq__ozbmh = context.insert_const_string(builder.module, 'numpy')
        ouo__wnal = pyapi.import_module_noblock(umsq__ozbmh)
        arr_obj = pyapi.call_method(ouo__wnal, 'ascontiguousarray', (ffb__ymv,)
            )
        pyapi.decref(ffb__ymv)
        pyapi.decref(ouo__wnal)
    else:
        arr_obj = ffb__ymv
    pyapi.decref(qcn__ibcx)
    pyapi.decref(mpvxe__usme)
    pyapi.decref(tdz__tojbv)
    pyapi.decref(qpary__vuab)
    pyapi.decref(zpju__cutor)
    pyapi.decref(ose__bidy)
    return arr_obj


@intrinsic
def unbox_dataframe_column(typingctx, df, i=None):
    assert isinstance(df, DataFrameType) and is_overload_constant_int(i)

    def codegen(context, builder, sig, args):
        pyapi = context.get_python_api(builder)
        c = numba.core.pythonapi._UnboxContext(context, builder, pyapi)
        df_typ = sig.args[0]
        col_ind = get_overload_const_int(sig.args[1])
        data_typ = df_typ.data[col_ind]
        qtfad__iokky = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=args[0])
        arr_obj = get_df_obj_column_codegen(context, builder, pyapi,
            qtfad__iokky.parent, args[1], data_typ)
        dinbt__mej = _unbox_series_data(data_typ.dtype, data_typ, arr_obj, c)
        c.pyapi.decref(arr_obj)
        dataframe_payload = (bodo.hiframes.pd_dataframe_ext.
            get_dataframe_payload(c.context, c.builder, df_typ, args[0]))
        if df_typ.is_table_format:
            bbphf__ysy = cgutils.create_struct_proxy(df_typ.table_type)(c.
                context, c.builder, builder.extract_value(dataframe_payload
                .data, 0))
            bew__vgcvu = df_typ.table_type.type_to_blk[data_typ]
            lgjjd__pzff = getattr(bbphf__ysy, f'block_{bew__vgcvu}')
            pduuo__mrz = ListInstance(c.context, c.builder, types.List(
                data_typ), lgjjd__pzff)
            hncz__xgqo = context.get_constant(types.int64, df_typ.
                table_type.block_offsets[col_ind])
            pduuo__mrz.inititem(hncz__xgqo, dinbt__mej.value, incref=False)
        else:
            dataframe_payload.data = builder.insert_value(dataframe_payload
                .data, dinbt__mej.value, col_ind)
        uioz__ozhu = DataFramePayloadType(df_typ)
        rwxaw__rblj = context.nrt.meminfo_data(builder, qtfad__iokky.meminfo)
        lvs__hjujm = context.get_value_type(uioz__ozhu).as_pointer()
        rwxaw__rblj = builder.bitcast(rwxaw__rblj, lvs__hjujm)
        builder.store(dataframe_payload._getvalue(), rwxaw__rblj)
    return signature(types.none, df, i), codegen


@unbox(SeriesType)
def unbox_series(typ, val, c):
    ffb__ymv = c.pyapi.object_getattr_string(val, 'values')
    if isinstance(typ.data, types.Array):
        umsq__ozbmh = c.context.insert_const_string(c.builder.module, 'numpy')
        ouo__wnal = c.pyapi.import_module_noblock(umsq__ozbmh)
        arr_obj = c.pyapi.call_method(ouo__wnal, 'ascontiguousarray', (
            ffb__ymv,))
        c.pyapi.decref(ffb__ymv)
        c.pyapi.decref(ouo__wnal)
    else:
        arr_obj = ffb__ymv
    dvym__yewea = _unbox_series_data(typ.dtype, typ.data, arr_obj, c).value
    dxmcv__fshl = c.pyapi.object_getattr_string(val, 'index')
    ywjle__pwsb = c.pyapi.to_native_value(typ.index, dxmcv__fshl).value
    modm__pft = c.pyapi.object_getattr_string(val, 'name')
    jzg__dax = c.pyapi.to_native_value(typ.name_typ, modm__pft).value
    exvav__dnb = bodo.hiframes.pd_series_ext.construct_series(c.context, c.
        builder, typ, dvym__yewea, ywjle__pwsb, jzg__dax)
    c.pyapi.decref(arr_obj)
    c.pyapi.decref(dxmcv__fshl)
    c.pyapi.decref(modm__pft)
    return NativeValue(exvav__dnb)


def _unbox_series_data(dtype, data_typ, arr_obj, c):
    if data_typ == string_array_split_view_type:
        ebxy__eqrnw = c.context.make_helper(c.builder,
            string_array_split_view_type)
        return NativeValue(ebxy__eqrnw._getvalue())
    return c.pyapi.to_native_value(data_typ, arr_obj)


@box(HeterogeneousSeriesType)
@box(SeriesType)
def box_series(typ, val, c):
    kgnkr__ukil = c.context.insert_const_string(c.builder.module, 'pandas')
    mmlq__sbd = c.pyapi.import_module_noblock(kgnkr__ukil)
    vxv__jtad = bodo.hiframes.pd_series_ext.get_series_payload(c.context, c
        .builder, typ, val)
    c.context.nrt.incref(c.builder, typ.data, vxv__jtad.data)
    c.context.nrt.incref(c.builder, typ.index, vxv__jtad.index)
    c.context.nrt.incref(c.builder, typ.name_typ, vxv__jtad.name)
    arr_obj = c.pyapi.from_native_value(typ.data, vxv__jtad.data, c.env_manager
        )
    dxmcv__fshl = c.pyapi.from_native_value(typ.index, vxv__jtad.index, c.
        env_manager)
    modm__pft = c.pyapi.from_native_value(typ.name_typ, vxv__jtad.name, c.
        env_manager)
    if isinstance(typ, HeterogeneousSeriesType) and isinstance(typ.data,
        bodo.NullableTupleType):
        dtype = c.pyapi.unserialize(c.pyapi.serialize_object(object))
    else:
        dtype = c.pyapi.make_none()
    res = c.pyapi.call_method(mmlq__sbd, 'Series', (arr_obj, dxmcv__fshl,
        dtype, modm__pft))
    c.pyapi.decref(arr_obj)
    c.pyapi.decref(dxmcv__fshl)
    c.pyapi.decref(modm__pft)
    if isinstance(typ, HeterogeneousSeriesType) and isinstance(typ.data,
        bodo.NullableTupleType):
        c.pyapi.decref(dtype)
    _set_bodo_meta_series(res, c, typ)
    c.pyapi.decref(mmlq__sbd)
    c.context.nrt.decref(c.builder, typ, val)
    return res


def type_enum_list_to_py_list_obj(pyapi, context, builder, env_manager,
    typ_list):
    ete__hrd = []
    for cgyk__ozjxt in typ_list:
        if isinstance(cgyk__ozjxt, int) and not isinstance(cgyk__ozjxt, bool):
            gxx__bai = pyapi.long_from_longlong(lir.Constant(lir.IntType(64
                ), cgyk__ozjxt))
        else:
            acz__bvx = numba.typeof(cgyk__ozjxt)
            nribd__azzh = context.get_constant_generic(builder, acz__bvx,
                cgyk__ozjxt)
            gxx__bai = pyapi.from_native_value(acz__bvx, nribd__azzh,
                env_manager)
        ete__hrd.append(gxx__bai)
    vtp__ewli = pyapi.list_pack(ete__hrd)
    for val in ete__hrd:
        pyapi.decref(val)
    return vtp__ewli


def _set_bodo_meta_dataframe(c, obj, typ):
    pyapi = c.pyapi
    context = c.context
    builder = c.builder
    mnpdl__xkbhm = not typ.has_runtime_cols and (not typ.is_table_format or
        len(typ.columns) < TABLE_FORMAT_THRESHOLD)
    qonn__ghbn = 2 if mnpdl__xkbhm else 1
    kddnv__eofy = pyapi.dict_new(qonn__ghbn)
    dmee__altq = pyapi.long_from_longlong(lir.Constant(lir.IntType(64), typ
        .dist.value))
    pyapi.dict_setitem_string(kddnv__eofy, 'dist', dmee__altq)
    pyapi.decref(dmee__altq)
    if mnpdl__xkbhm:
        jkj__fctjc = _dtype_to_type_enum_list(typ.index)
        if jkj__fctjc != None:
            lnsrg__vylf = type_enum_list_to_py_list_obj(pyapi, context,
                builder, c.env_manager, jkj__fctjc)
        else:
            lnsrg__vylf = pyapi.make_none()
        emzft__ukdxw = []
        for dtype in typ.data:
            typ_list = _dtype_to_type_enum_list(dtype)
            if typ_list != None:
                vtp__ewli = type_enum_list_to_py_list_obj(pyapi, context,
                    builder, c.env_manager, typ_list)
            else:
                vtp__ewli = pyapi.make_none()
            emzft__ukdxw.append(vtp__ewli)
        itwa__mxmlw = pyapi.list_pack(emzft__ukdxw)
        sivhr__fyiqj = pyapi.list_pack([lnsrg__vylf, itwa__mxmlw])
        for val in emzft__ukdxw:
            pyapi.decref(val)
        pyapi.dict_setitem_string(kddnv__eofy, 'type_metadata', sivhr__fyiqj)
    pyapi.object_setattr_string(obj, '_bodo_meta', kddnv__eofy)
    pyapi.decref(kddnv__eofy)


def get_series_dtype_handle_null_int_and_hetrogenous(series_typ):
    if isinstance(series_typ, HeterogeneousSeriesType):
        return None
    if isinstance(series_typ.dtype, types.Number) and isinstance(series_typ
        .data, IntegerArrayType):
        return IntDtype(series_typ.dtype)
    return series_typ.dtype


def _set_bodo_meta_series(obj, c, typ):
    pyapi = c.pyapi
    context = c.context
    builder = c.builder
    kddnv__eofy = pyapi.dict_new(2)
    dmee__altq = pyapi.long_from_longlong(lir.Constant(lir.IntType(64), typ
        .dist.value))
    jkj__fctjc = _dtype_to_type_enum_list(typ.index)
    if jkj__fctjc != None:
        lnsrg__vylf = type_enum_list_to_py_list_obj(pyapi, context, builder,
            c.env_manager, jkj__fctjc)
    else:
        lnsrg__vylf = pyapi.make_none()
    dtype = get_series_dtype_handle_null_int_and_hetrogenous(typ)
    if dtype != None:
        typ_list = _dtype_to_type_enum_list(dtype)
        if typ_list != None:
            peoyl__zko = type_enum_list_to_py_list_obj(pyapi, context,
                builder, c.env_manager, typ_list)
        else:
            peoyl__zko = pyapi.make_none()
    else:
        peoyl__zko = pyapi.make_none()
    gnf__pwn = pyapi.list_pack([lnsrg__vylf, peoyl__zko])
    pyapi.dict_setitem_string(kddnv__eofy, 'type_metadata', gnf__pwn)
    pyapi.decref(gnf__pwn)
    pyapi.dict_setitem_string(kddnv__eofy, 'dist', dmee__altq)
    pyapi.object_setattr_string(obj, '_bodo_meta', kddnv__eofy)
    pyapi.decref(kddnv__eofy)
    pyapi.decref(dmee__altq)


@typeof_impl.register(np.ndarray)
def _typeof_ndarray(val, c):
    try:
        dtype = numba.np.numpy_support.from_dtype(val.dtype)
    except NotImplementedError as jtur__lwghp:
        dtype = types.pyobject
    if dtype == types.pyobject:
        return _infer_ndarray_obj_dtype(val)
    jbk__jbq = numba.np.numpy_support.map_layout(val)
    nss__pua = not val.flags.writeable
    return types.Array(dtype, val.ndim, jbk__jbq, readonly=nss__pua)


def _infer_ndarray_obj_dtype(val):
    if not val.dtype == np.dtype('O'):
        raise BodoError('Unsupported array dtype: {}'.format(val.dtype))
    i = 0
    while i < len(val) and (pd.api.types.is_scalar(val[i]) and pd.isna(val[
        i]) or not pd.api.types.is_scalar(val[i]) and len(val[i]) == 0):
        i += 1
    if i == len(val):
        warnings.warn(BodoWarning(
            'Empty object array passed to Bodo, which causes ambiguity in typing. This can cause errors in parallel execution.'
            ))
        return string_array_type
    tnxd__lupr = val[i]
    if isinstance(tnxd__lupr, str):
        return string_array_type
    elif isinstance(tnxd__lupr, bytes):
        return binary_array_type
    elif isinstance(tnxd__lupr, bool):
        return bodo.libs.bool_arr_ext.boolean_array
    elif isinstance(tnxd__lupr, (int, np.int32, np.int64)):
        return bodo.libs.int_arr_ext.IntegerArrayType(numba.typeof(tnxd__lupr))
    elif isinstance(tnxd__lupr, (dict, Dict)) and all(isinstance(mwte__bfd,
        str) for mwte__bfd in tnxd__lupr.keys()):
        ngo__jdr = tuple(tnxd__lupr.keys())
        zert__iyxlt = tuple(_get_struct_value_arr_type(v) for v in
            tnxd__lupr.values())
        return StructArrayType(zert__iyxlt, ngo__jdr)
    elif isinstance(tnxd__lupr, (dict, Dict)):
        hve__fexb = numba.typeof(_value_to_array(list(tnxd__lupr.keys())))
        hzs__kqem = numba.typeof(_value_to_array(list(tnxd__lupr.values())))
        return MapArrayType(hve__fexb, hzs__kqem)
    elif isinstance(tnxd__lupr, tuple):
        zert__iyxlt = tuple(_get_struct_value_arr_type(v) for v in tnxd__lupr)
        return TupleArrayType(zert__iyxlt)
    if isinstance(tnxd__lupr, (list, np.ndarray, pd.arrays.BooleanArray, pd
        .arrays.IntegerArray, pd.arrays.StringArray)):
        if isinstance(tnxd__lupr, list):
            tnxd__lupr = _value_to_array(tnxd__lupr)
        fxrz__plela = numba.typeof(tnxd__lupr)
        return ArrayItemArrayType(fxrz__plela)
    if isinstance(tnxd__lupr, datetime.date):
        return datetime_date_array_type
    if isinstance(tnxd__lupr, datetime.timedelta):
        return datetime_timedelta_array_type
    if isinstance(tnxd__lupr, decimal.Decimal):
        return DecimalArrayType(38, 18)
    raise BodoError('Unsupported object array with first value: {}'.format(
        tnxd__lupr))


def _value_to_array(val):
    assert isinstance(val, (list, dict, Dict))
    if isinstance(val, (dict, Dict)):
        val = dict(val)
        return np.array([val], np.object_)
    tyx__sslp = val.copy()
    tyx__sslp.append(None)
    kvjm__qzv = np.array(tyx__sslp, np.object_)
    if len(val) and isinstance(val[0], float):
        kvjm__qzv = np.array(val, np.float64)
    return kvjm__qzv


def _get_struct_value_arr_type(v):
    if isinstance(v, (dict, Dict)):
        return numba.typeof(_value_to_array(v))
    if isinstance(v, list):
        return dtype_to_array_type(numba.typeof(_value_to_array(v)))
    if pd.api.types.is_scalar(v) and pd.isna(v):
        warnings.warn(BodoWarning(
            'Field value in struct array is NA, which causes ambiguity in typing. This can cause errors in parallel execution.'
            ))
        return string_array_type
    plea__pgzfp = dtype_to_array_type(numba.typeof(v))
    if isinstance(v, (int, bool)):
        plea__pgzfp = to_nullable_type(plea__pgzfp)
    return plea__pgzfp
