"""
Implementation of DataFrame attributes and methods using overload.
"""
import operator
import re
import warnings
from collections import namedtuple
from typing import Tuple
import llvmlite.llvmpy.core as lc
import numba
import numpy as np
import pandas as pd
from numba.core import cgutils, ir, types
from numba.core.imputils import RefType, impl_ret_borrowed, impl_ret_new_ref, iternext_impl, lower_builtin
from numba.core.ir_utils import mk_unique_var, next_label
from numba.core.typing import signature
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import lower_getattr, models, overload, overload_attribute, overload_method, register_model, type_callable
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.datetime_timedelta_ext import _no_input, datetime_timedelta_array_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType
from bodo.hiframes.pd_dataframe_ext import DataFrameType, check_runtime_cols_unsupported, handle_inplace_df_type_change
from bodo.hiframes.pd_index_ext import DatetimeIndexType, RangeIndexType, StringIndexType, is_pd_index_type
from bodo.hiframes.pd_multi_index_ext import MultiIndexType
from bodo.hiframes.pd_series_ext import SeriesType, if_series_to_array_type
from bodo.hiframes.pd_timestamp_ext import pd_timestamp_type
from bodo.hiframes.rolling import is_supported_shift_array_type
from bodo.hiframes.split_impl import string_array_split_view_type
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type
from bodo.libs.bool_arr_ext import BooleanArrayType, boolean_array, boolean_dtype
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.interval_arr_ext import IntervalArrayType
from bodo.libs.map_arr_ext import MapArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.utils.transform import bodo_types_with_params, gen_const_tup, no_side_effect_call_tuples
from bodo.utils.typing import BodoError, BodoWarning, check_unsupported_args, dtype_to_array_type, ensure_constant_arg, ensure_constant_values, get_index_data_arr_types, get_index_names, get_literal_value, get_nullable_and_non_nullable_types, get_overload_const_bool, get_overload_const_int, get_overload_const_list, get_overload_const_str, get_overload_const_tuple, get_overload_constant_dict, get_overload_constant_series, is_common_scalar_dtype, is_literal_type, is_overload_bool, is_overload_bool_list, is_overload_constant_bool, is_overload_constant_dict, is_overload_constant_int, is_overload_constant_list, is_overload_constant_series, is_overload_constant_str, is_overload_constant_tuple, is_overload_false, is_overload_int, is_overload_none, is_overload_true, is_overload_zero, parse_dtype, raise_bodo_error, raise_const_error, unliteral_val
from bodo.utils.utils import is_array_typ


@overload_attribute(DataFrameType, 'index', inline='always')
def overload_dataframe_index(df):
    return lambda df: bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)


def generate_col_to_index_func_text(col_names: Tuple):
    if all(isinstance(a, str) for a in col_names) or all(isinstance(a,
        bytes) for a in col_names):
        pcnrm__qsdln = f'bodo.utils.conversion.coerce_to_array({col_names})'
        return (
            f'bodo.hiframes.pd_index_ext.init_binary_str_index({pcnrm__qsdln})\n'
            )
    elif all(isinstance(a, (int, float)) for a in col_names):
        arr = f'bodo.utils.conversion.coerce_to_array({col_names})'
        return f'bodo.hiframes.pd_index_ext.init_numeric_index({arr})\n'
    else:
        return f'bodo.hiframes.pd_index_ext.init_heter_index({col_names})\n'


@overload_attribute(DataFrameType, 'columns', inline='always')
def overload_dataframe_columns(df):
    check_runtime_cols_unsupported(df, 'DataFrame.columns')
    visq__yelo = 'def impl(df):\n'
    iylt__hhsz = bodo.hiframes.dataframe_impl.generate_col_to_index_func_text(
        df.columns)
    visq__yelo += f'  return {iylt__hhsz}'
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@overload_attribute(DataFrameType, 'values')
def overload_dataframe_values(df):
    check_runtime_cols_unsupported(df, 'DataFrame.values')
    if not is_df_values_numpy_supported_dftyp(df):
        raise_bodo_error(
            'DataFrame.values: only supported for dataframes containing numeric values'
            )
    potqq__ublqz = len(df.columns)
    ebj__klgg = set(i for i in range(potqq__ublqz) if isinstance(df.data[i],
        IntegerArrayType))
    data_args = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}){}'.
        format(i, '.astype(float)' if i in ebj__klgg else '') for i in
        range(potqq__ublqz))
    visq__yelo = 'def f(df):\n'.format()
    visq__yelo += '    return np.stack(({},), 1)\n'.format(data_args)
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'np': np}, zrt__acbb)
    vsd__jskm = zrt__acbb['f']
    return vsd__jskm


@overload_method(DataFrameType, 'to_numpy', inline='always', no_unliteral=True)
def overload_dataframe_to_numpy(df, dtype=None, copy=False, na_value=_no_input
    ):
    check_runtime_cols_unsupported(df, 'DataFrame.to_numpy()')
    if not is_df_values_numpy_supported_dftyp(df):
        raise_bodo_error(
            'DataFrame.to_numpy(): only supported for dataframes containing numeric values'
            )
    hwwj__nvjxf = {'dtype': dtype, 'na_value': na_value}
    hriw__fson = {'dtype': None, 'na_value': _no_input}
    check_unsupported_args('DataFrame.to_numpy', hwwj__nvjxf, hriw__fson,
        package_name='pandas', module_name='DataFrame')

    def impl(df, dtype=None, copy=False, na_value=_no_input):
        return df.values
    return impl


@overload_attribute(DataFrameType, 'ndim', inline='always')
def overload_dataframe_ndim(df):
    return lambda df: 2


@overload_attribute(DataFrameType, 'size')
def overload_dataframe_size(df):
    if df.has_runtime_cols:

        def impl(df):
            t = bodo.hiframes.pd_dataframe_ext.get_dataframe_table(df)
            oljbj__gfab = bodo.hiframes.table.compute_num_runtime_columns(t)
            return oljbj__gfab * len(t)
        return impl
    ncols = len(df.columns)
    return lambda df: ncols * len(df)


@lower_getattr(DataFrameType, 'shape')
def lower_dataframe_shape(context, builder, typ, val):
    impl = overload_dataframe_shape(typ)
    return context.compile_internal(builder, impl, types.Tuple([types.int64,
        types.int64])(typ), (val,))


def overload_dataframe_shape(df):
    if df.has_runtime_cols:

        def impl(df):
            t = bodo.hiframes.pd_dataframe_ext.get_dataframe_table(df)
            oljbj__gfab = bodo.hiframes.table.compute_num_runtime_columns(t)
            return len(t), oljbj__gfab
        return impl
    ncols = len(df.columns)
    return lambda df: (len(df), types.int64(ncols))


@overload_attribute(DataFrameType, 'dtypes')
def overload_dataframe_dtypes(df):
    check_runtime_cols_unsupported(df, 'DataFrame.dtypes')
    visq__yelo = 'def impl(df):\n'
    data = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}).dtype\n'
         for i in range(len(df.columns)))
    tbot__julg = ',' if len(df.columns) == 1 else ''
    index = f'bodo.hiframes.pd_index_ext.init_heter_index({df.columns})'
    visq__yelo += f"""  return bodo.hiframes.pd_series_ext.init_series(({data}{tbot__julg}), {index}, None)
"""
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@overload_attribute(DataFrameType, 'empty')
def overload_dataframe_empty(df):
    check_runtime_cols_unsupported(df, 'DataFrame.empty')
    if len(df.columns) == 0:
        return lambda df: True
    return lambda df: len(df) == 0


@overload_method(DataFrameType, 'assign', no_unliteral=True)
def overload_dataframe_assign(df, **kwargs):
    check_runtime_cols_unsupported(df, 'DataFrame.assign()')
    raise_bodo_error('Invalid df.assign() call')


@overload_method(DataFrameType, 'insert', no_unliteral=True)
def overload_dataframe_insert(df, loc, column, value, allow_duplicates=False):
    check_runtime_cols_unsupported(df, 'DataFrame.insert()')
    raise_bodo_error('Invalid df.insert() call')


def _get_dtype_str(dtype):
    if isinstance(dtype, types.Function):
        if dtype.key[0] == str:
            return "'str'"
        elif dtype.key[0] == float:
            return 'float'
        elif dtype.key[0] == int:
            return 'int'
        elif dtype.key[0] == bool:
            return 'bool'
        else:
            raise BodoError(f'invalid dtype: {dtype}')
    if isinstance(dtype, types.DTypeSpec):
        dtype = dtype.dtype
    if isinstance(dtype, types.functions.NumberClass):
        return f"'{dtype.key}'"
    if isinstance(dtype, types.PyObject) or dtype in (object, 'object'):
        return "'object'"
    if dtype in (bodo.libs.str_arr_ext.string_dtype, pd.StringDtype()):
        return 'str'
    return f"'{dtype}'"


@overload_method(DataFrameType, 'astype', inline='always', no_unliteral=True)
def overload_dataframe_astype(df, dtype, copy=True, errors='raise',
    _bodo_nan_to_str=True, _bodo_object_typeref=None):
    check_runtime_cols_unsupported(df, 'DataFrame.astype()')
    hwwj__nvjxf = {'copy': copy, 'errors': errors}
    hriw__fson = {'copy': True, 'errors': 'raise'}
    check_unsupported_args('df.astype', hwwj__nvjxf, hriw__fson,
        package_name='pandas', module_name='DataFrame')
    if dtype == types.unicode_type:
        raise_bodo_error(
            "DataFrame.astype(): 'dtype' when passed as string must be a constant value"
            )
    extra_globals = None
    if _bodo_object_typeref is not None:
        assert isinstance(_bodo_object_typeref, types.TypeRef
            ), 'Bodo schema used in DataFrame.astype should be a TypeRef'
        gzlcp__mdcah = _bodo_object_typeref.instance_type
        assert isinstance(gzlcp__mdcah, DataFrameType
            ), 'Bodo schema used in DataFrame.astype is only supported for DataFrame schemas'
        extra_globals = {}
        erm__iwq = {}
        for i, name in enumerate(gzlcp__mdcah.columns):
            arr_typ = gzlcp__mdcah.data[i]
            if isinstance(arr_typ, IntegerArrayType):
                ppag__fuoa = bodo.libs.int_arr_ext.IntDtype(arr_typ.dtype)
            elif arr_typ == boolean_array:
                ppag__fuoa = boolean_dtype
            else:
                ppag__fuoa = arr_typ.dtype
            extra_globals[f'_bodo_schema{i}'] = ppag__fuoa
            erm__iwq[name] = f'_bodo_schema{i}'
        data_args = ', '.join(
            f'bodo.utils.conversion.fix_arr_dtype(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), {erm__iwq[ghpd__ovmzd]}, copy, nan_to_str=_bodo_nan_to_str, from_series=True)'
             if ghpd__ovmzd in erm__iwq else
            f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})' for
            i, ghpd__ovmzd in enumerate(df.columns))
    elif is_overload_constant_dict(dtype) or is_overload_constant_series(dtype
        ):
        qyo__tlxg = get_overload_constant_dict(dtype
            ) if is_overload_constant_dict(dtype) else dict(
            get_overload_constant_series(dtype))
        data_args = ', '.join(
            f'bodo.utils.conversion.fix_arr_dtype(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), {_get_dtype_str(qyo__tlxg[ghpd__ovmzd])}, copy, nan_to_str=_bodo_nan_to_str, from_series=True)'
             if ghpd__ovmzd in qyo__tlxg else
            f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})' for
            i, ghpd__ovmzd in enumerate(df.columns))
    else:
        data_args = ', '.join(
            f'bodo.utils.conversion.fix_arr_dtype(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), dtype, copy, nan_to_str=_bodo_nan_to_str, from_series=True)'
             for i in range(len(df.columns)))
    header = """def impl(df, dtype, copy=True, errors='raise', _bodo_nan_to_str=True, _bodo_object_typeref=None):
"""
    return _gen_init_df(header, df.columns, data_args, extra_globals=
        extra_globals)


@overload_method(DataFrameType, 'copy', inline='always', no_unliteral=True)
def overload_dataframe_copy(df, deep=True):
    check_runtime_cols_unsupported(df, 'DataFrame.copy()')
    dkqt__jrnc = []
    for i in range(len(df.columns)):
        arr = f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})'
        if is_overload_true(deep):
            dkqt__jrnc.append(arr + '.copy()')
        elif is_overload_false(deep):
            dkqt__jrnc.append(arr)
        else:
            dkqt__jrnc.append(f'{arr}.copy() if deep else {arr}')
    header = 'def impl(df, deep=True):\n'
    return _gen_init_df(header, df.columns, ', '.join(dkqt__jrnc))


@overload_method(DataFrameType, 'rename', inline='always', no_unliteral=True)
def overload_dataframe_rename(df, mapper=None, index=None, columns=None,
    axis=None, copy=True, inplace=False, level=None, errors='ignore',
    _bodo_transformed=False):
    check_runtime_cols_unsupported(df, 'DataFrame.rename()')
    handle_inplace_df_type_change(inplace, _bodo_transformed, 'rename')
    hwwj__nvjxf = {'index': index, 'level': level, 'errors': errors}
    hriw__fson = {'index': None, 'level': None, 'errors': 'ignore'}
    check_unsupported_args('DataFrame.rename', hwwj__nvjxf, hriw__fson,
        package_name='pandas', module_name='DataFrame')
    if not is_overload_constant_bool(inplace):
        raise BodoError(
            "DataFrame.rename(): 'inplace' keyword only supports boolean constant assignment"
            )
    if not is_overload_none(mapper):
        if not is_overload_none(columns):
            raise BodoError(
                "DataFrame.rename(): Cannot specify both 'mapper' and 'columns'"
                )
        if not (is_overload_constant_int(axis) and get_overload_const_int(
            axis) == 1):
            raise BodoError(
                "DataFrame.rename(): 'mapper' only supported with axis=1")
        if not is_overload_constant_dict(mapper):
            raise_bodo_error(
                "'mapper' argument to DataFrame.rename() should be a constant dictionary"
                )
        lyuys__zmjgk = get_overload_constant_dict(mapper)
    elif not is_overload_none(columns):
        if not is_overload_none(axis):
            raise BodoError(
                "DataFrame.rename(): Cannot specify both 'axis' and 'columns'")
        if not is_overload_constant_dict(columns):
            raise_bodo_error(
                "'columns' argument to DataFrame.rename() should be a constant dictionary"
                )
        lyuys__zmjgk = get_overload_constant_dict(columns)
    else:
        raise_bodo_error(
            "DataFrame.rename(): must pass columns either via 'mapper' and 'axis'=1 or 'columns'"
            )
    gippv__froes = [lyuys__zmjgk.get(df.columns[i], df.columns[i]) for i in
        range(len(df.columns))]
    dkqt__jrnc = []
    for i in range(len(df.columns)):
        arr = f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})'
        if is_overload_true(copy):
            dkqt__jrnc.append(arr + '.copy()')
        elif is_overload_false(copy):
            dkqt__jrnc.append(arr)
        else:
            dkqt__jrnc.append(f'{arr}.copy() if copy else {arr}')
    header = """def impl(df, mapper=None, index=None, columns=None, axis=None, copy=True, inplace=False, level=None, errors='ignore', _bodo_transformed=False):
"""
    return _gen_init_df(header, gippv__froes, ', '.join(dkqt__jrnc))


@overload_method(DataFrameType, 'filter', no_unliteral=True)
def overload_dataframe_filter(df, items=None, like=None, regex=None, axis=None
    ):
    check_runtime_cols_unsupported(df, 'DataFrame.filter()')
    fpwhc__mln = not is_overload_none(items)
    qtw__qya = not is_overload_none(like)
    zof__rjlj = not is_overload_none(regex)
    zfzq__yexj = fpwhc__mln ^ qtw__qya ^ zof__rjlj
    hyvb__nls = not (fpwhc__mln or qtw__qya or zof__rjlj)
    if hyvb__nls:
        raise BodoError(
            'DataFrame.filter(): one of keyword arguments `items`, `like`, and `regex` must be supplied'
            )
    if not zfzq__yexj:
        raise BodoError(
            'DataFrame.filter(): keyword arguments `items`, `like`, and `regex` are mutually exclusive'
            )
    if is_overload_none(axis):
        axis = 'columns'
    if is_overload_constant_str(axis):
        axis = get_overload_const_str(axis)
        if axis not in {'index', 'columns'}:
            raise_bodo_error(
                'DataFrame.filter(): keyword arguments `axis` must be either "index" or "columns" if string'
                )
        vyqt__lqmhw = 0 if axis == 'index' else 1
    elif is_overload_constant_int(axis):
        axis = get_overload_const_int(axis)
        if axis not in {0, 1}:
            raise_bodo_error(
                'DataFrame.filter(): keyword arguments `axis` must be either 0 or 1 if integer'
                )
        vyqt__lqmhw = axis
    else:
        raise_bodo_error(
            'DataFrame.filter(): keyword arguments `axis` must be constant string or integer'
            )
    assert vyqt__lqmhw in {0, 1}
    visq__yelo = (
        'def impl(df, items=None, like=None, regex=None, axis=None):\n')
    if vyqt__lqmhw == 0:
        raise BodoError(
            'DataFrame.filter(): filtering based on index is not supported.')
    if vyqt__lqmhw == 1:
        wzp__mpwzh = []
        xkip__sjh = []
        cyw__xthcp = []
        if fpwhc__mln:
            if is_overload_constant_list(items):
                yep__dbf = get_overload_const_list(items)
            else:
                raise BodoError(
                    "Dataframe.filter(): argument 'items' must be a list of constant strings."
                    )
        if qtw__qya:
            if is_overload_constant_str(like):
                vkyqo__whge = get_overload_const_str(like)
            else:
                raise BodoError(
                    "Dataframe.filter(): argument 'like' must be a constant string."
                    )
        if zof__rjlj:
            if is_overload_constant_str(regex):
                hgt__mcqg = get_overload_const_str(regex)
                snp__gfaxb = re.compile(hgt__mcqg)
            else:
                raise BodoError(
                    "Dataframe.filter(): argument 'regex' must be a constant string."
                    )
        for i, ghpd__ovmzd in enumerate(df.columns):
            if not is_overload_none(items
                ) and ghpd__ovmzd in yep__dbf or not is_overload_none(like
                ) and vkyqo__whge in str(ghpd__ovmzd) or not is_overload_none(
                regex) and snp__gfaxb.search(str(ghpd__ovmzd)):
                xkip__sjh.append(ghpd__ovmzd)
                cyw__xthcp.append(i)
        for i in cyw__xthcp:
            aawpg__yxbxg = f'data_{i}'
            wzp__mpwzh.append(aawpg__yxbxg)
            visq__yelo += f"""  {aawpg__yxbxg} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})
"""
        data_args = ', '.join(wzp__mpwzh)
        return _gen_init_df(visq__yelo, xkip__sjh, data_args)


@overload_method(DataFrameType, 'isna', inline='always', no_unliteral=True)
@overload_method(DataFrameType, 'isnull', inline='always', no_unliteral=True)
def overload_dataframe_isna(df):
    check_runtime_cols_unsupported(df, 'DataFrame.isna()')
    data_args = ', '.join(
        f'bodo.libs.array_ops.array_op_isna(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}))'
         for i in range(len(df.columns)))
    header = 'def impl(df):\n'
    return _gen_init_df(header, df.columns, data_args)


@overload_method(DataFrameType, 'select_dtypes', inline='always',
    no_unliteral=True)
def overload_dataframe_select_dtypes(df, include=None, exclude=None):
    check_runtime_cols_unsupported(df, 'DataFrame.select_dtypes')
    rouv__mio = is_overload_none(include)
    uzo__kzg = is_overload_none(exclude)
    ivpp__zvmv = 'DataFrame.select_dtypes'
    if rouv__mio and uzo__kzg:
        raise_bodo_error(
            'DataFrame.select_dtypes() At least one of include or exclude must not be none'
            )

    def is_legal_input(elem):
        return is_overload_constant_str(elem) or isinstance(elem, types.
            DTypeSpec) or isinstance(elem, types.Function)
    if not rouv__mio:
        if is_overload_constant_list(include):
            include = get_overload_const_list(include)
            nlmk__pfrvb = [dtype_to_array_type(parse_dtype(elem, ivpp__zvmv
                )) for elem in include]
        elif is_legal_input(include):
            nlmk__pfrvb = [dtype_to_array_type(parse_dtype(include,
                ivpp__zvmv))]
        else:
            raise_bodo_error(
                'DataFrame.select_dtypes() only supports constant strings or types as arguments'
                )
        nlmk__pfrvb = get_nullable_and_non_nullable_types(nlmk__pfrvb)
        gpec__bvz = tuple(ghpd__ovmzd for i, ghpd__ovmzd in enumerate(df.
            columns) if df.data[i] in nlmk__pfrvb)
    else:
        gpec__bvz = df.columns
    if not uzo__kzg:
        if is_overload_constant_list(exclude):
            exclude = get_overload_const_list(exclude)
            gfll__hvehb = [dtype_to_array_type(parse_dtype(elem, ivpp__zvmv
                )) for elem in exclude]
        elif is_legal_input(exclude):
            gfll__hvehb = [dtype_to_array_type(parse_dtype(exclude,
                ivpp__zvmv))]
        else:
            raise_bodo_error(
                'DataFrame.select_dtypes() only supports constant strings or types as arguments'
                )
        gfll__hvehb = get_nullable_and_non_nullable_types(gfll__hvehb)
        gpec__bvz = tuple(ghpd__ovmzd for ghpd__ovmzd in gpec__bvz if df.
            data[df.columns.index(ghpd__ovmzd)] not in gfll__hvehb)
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(ghpd__ovmzd)})'
         for ghpd__ovmzd in gpec__bvz)
    header = 'def impl(df, include=None, exclude=None):\n'
    return _gen_init_df(header, gpec__bvz, data_args)


@overload_method(DataFrameType, 'notna', inline='always', no_unliteral=True)
@overload_method(DataFrameType, 'notnull', inline='always', no_unliteral=True)
def overload_dataframe_notna(df):
    check_runtime_cols_unsupported(df, 'DataFrame.notna()')
    data_args = ', '.join(
        f'bodo.libs.array_ops.array_op_isna(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})) == False'
         for i in range(len(df.columns)))
    header = 'def impl(df):\n'
    return _gen_init_df(header, df.columns, data_args)


def overload_dataframe_head(df, n=5):
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})[:n]' for
        i in range(len(df.columns)))
    header = 'def impl(df, n=5):\n'
    index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[:n]'
    return _gen_init_df(header, df.columns, data_args, index)


@lower_builtin('df.head', DataFrameType, types.Integer)
@lower_builtin('df.head', DataFrameType, types.Omitted)
def dataframe_head_lower(context, builder, sig, args):
    impl = overload_dataframe_head(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


@overload_method(DataFrameType, 'tail', inline='always', no_unliteral=True)
def overload_dataframe_tail(df, n=5):
    check_runtime_cols_unsupported(df, 'DataFrame.tail()')
    if not is_overload_int(n):
        raise BodoError("Dataframe.tail(): 'n' must be an Integer")
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})[m:]' for
        i in range(len(df.columns)))
    header = 'def impl(df, n=5):\n'
    header += '  m = bodo.hiframes.series_impl.tail_slice(len(df), n)\n'
    index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[m:]'
    return _gen_init_df(header, df.columns, data_args, index)


@overload_method(DataFrameType, 'first', inline='always', no_unliteral=True)
def overload_dataframe_first(df, offset):
    check_runtime_cols_unsupported(df, 'DataFrame.first()')
    vqc__fmm = (types.unicode_type, bodo.month_begin_type, bodo.
        month_end_type, bodo.week_type, bodo.date_offset_type)
    if not isinstance(df.index, DatetimeIndexType):
        raise BodoError(
            'DataFrame.first(): only supports a DatetimeIndex index')
    if types.unliteral(offset) not in vqc__fmm:
        raise BodoError(
            "DataFrame.first(): 'offset' must be an string or DateOffset")
    index = (
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[:valid_entries]'
        )
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})[:valid_entries]'
         for i in range(len(df.columns)))
    header = 'def impl(df, offset):\n'
    header += (
        '  df_index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)\n'
        )
    header += '  if len(df_index):\n'
    header += '    start_date = df_index[0]\n'
    header += """    valid_entries = bodo.libs.array_kernels.get_valid_entries_from_date_offset(df_index, offset, start_date, False)
"""
    header += '  else:\n'
    header += '    valid_entries = 0\n'
    return _gen_init_df(header, df.columns, data_args, index)


@overload_method(DataFrameType, 'last', inline='always', no_unliteral=True)
def overload_dataframe_last(df, offset):
    check_runtime_cols_unsupported(df, 'DataFrame.last()')
    vqc__fmm = (types.unicode_type, bodo.month_begin_type, bodo.
        month_end_type, bodo.week_type, bodo.date_offset_type)
    if not isinstance(df.index, DatetimeIndexType):
        raise BodoError('DataFrame.last(): only supports a DatetimeIndex index'
            )
    if types.unliteral(offset) not in vqc__fmm:
        raise BodoError(
            "DataFrame.last(): 'offset' must be an string or DateOffset")
    index = (
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[len(df)-valid_entries:]'
        )
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})[len(df)-valid_entries:]'
         for i in range(len(df.columns)))
    header = 'def impl(df, offset):\n'
    header += (
        '  df_index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)\n'
        )
    header += '  if len(df_index):\n'
    header += '    final_date = df_index[-1]\n'
    header += """    valid_entries = bodo.libs.array_kernels.get_valid_entries_from_date_offset(df_index, offset, final_date, True)
"""
    header += '  else:\n'
    header += '    valid_entries = 0\n'
    return _gen_init_df(header, df.columns, data_args, index)


@overload_method(DataFrameType, 'to_string', no_unliteral=True)
def to_string_overload(df, buf=None, columns=None, col_space=None, header=
    True, index=True, na_rep='NaN', formatters=None, float_format=None,
    sparsify=None, index_names=True, justify=None, max_rows=None, min_rows=
    None, max_cols=None, show_dimensions=False, decimal='.', line_width=
    None, max_colwidth=None, encoding=None):
    check_runtime_cols_unsupported(df, 'DataFrame.to_string()')

    def impl(df, buf=None, columns=None, col_space=None, header=True, index
        =True, na_rep='NaN', formatters=None, float_format=None, sparsify=
        None, index_names=True, justify=None, max_rows=None, min_rows=None,
        max_cols=None, show_dimensions=False, decimal='.', line_width=None,
        max_colwidth=None, encoding=None):
        with numba.objmode(res='string'):
            res = df.to_string(buf=buf, columns=columns, col_space=
                col_space, header=header, index=index, na_rep=na_rep,
                formatters=formatters, float_format=float_format, sparsify=
                sparsify, index_names=index_names, justify=justify,
                max_rows=max_rows, min_rows=min_rows, max_cols=max_cols,
                show_dimensions=show_dimensions, decimal=decimal,
                line_width=line_width, max_colwidth=max_colwidth, encoding=
                encoding)
        return res
    return impl


@overload_method(DataFrameType, 'isin', inline='always', no_unliteral=True)
def overload_dataframe_isin(df, values):
    check_runtime_cols_unsupported(df, 'DataFrame.isin()')
    from bodo.utils.typing import is_iterable_type
    visq__yelo = 'def impl(df, values):\n'
    lqnbo__nml = {}
    msont__foc = False
    if isinstance(values, DataFrameType):
        msont__foc = True
        for i, ghpd__ovmzd in enumerate(df.columns):
            if ghpd__ovmzd in values.columns:
                sti__hiwe = 'val{}'.format(i)
                visq__yelo += (
                    """  {} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(values, {})
"""
                    .format(sti__hiwe, values.columns.index(ghpd__ovmzd)))
                lqnbo__nml[ghpd__ovmzd] = sti__hiwe
    elif is_iterable_type(values) and not isinstance(values, SeriesType):
        lqnbo__nml = {ghpd__ovmzd: 'values' for ghpd__ovmzd in df.columns}
    else:
        raise_bodo_error(f'pd.isin(): not supported for type {values}')
    data = []
    for i in range(len(df.columns)):
        sti__hiwe = 'data{}'.format(i)
        visq__yelo += (
            '  {} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})\n'
            .format(sti__hiwe, i))
        data.append(sti__hiwe)
    smrc__eyim = ['out{}'.format(i) for i in range(len(df.columns))]
    egm__lnau = """
  numba.parfors.parfor.init_prange()
  n = len({0})
  m = len({1})
  {2} = np.empty(n, np.bool_)
  for i in numba.parfors.parfor.internal_prange(n):
    {2}[i] = {0}[i] == {1}[i] if i < m else False
"""
    pzxl__aypbv = """
  numba.parfors.parfor.init_prange()
  n = len({0})
  {2} = np.empty(n, np.bool_)
  for i in numba.parfors.parfor.internal_prange(n):
    {2}[i] = {0}[i] in {1}
"""
    ejyo__nzrz = '  {} = np.zeros(len(df), np.bool_)\n'
    for i, (cname, afb__kmq) in enumerate(zip(df.columns, data)):
        if cname in lqnbo__nml:
            rdku__ink = lqnbo__nml[cname]
            if msont__foc:
                visq__yelo += egm__lnau.format(afb__kmq, rdku__ink,
                    smrc__eyim[i])
            else:
                visq__yelo += pzxl__aypbv.format(afb__kmq, rdku__ink,
                    smrc__eyim[i])
        else:
            visq__yelo += ejyo__nzrz.format(smrc__eyim[i])
    return _gen_init_df(visq__yelo, df.columns, ','.join(smrc__eyim))


@overload_method(DataFrameType, 'abs', inline='always', no_unliteral=True)
def overload_dataframe_abs(df):
    check_runtime_cols_unsupported(df, 'DataFrame.abs()')
    for arr_typ in df.data:
        if not (isinstance(arr_typ.dtype, types.Number) or arr_typ.dtype ==
            bodo.timedelta64ns):
            raise_bodo_error(
                f'DataFrame.abs(): Only supported for numeric and Timedelta. Encountered array with dtype {arr_typ.dtype}'
                )
    potqq__ublqz = len(df.columns)
    data_args = ', '.join(
        'np.abs(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}))'
        .format(i) for i in range(potqq__ublqz))
    header = 'def impl(df):\n'
    return _gen_init_df(header, df.columns, data_args)


def overload_dataframe_corr(df, method='pearson', min_periods=1):
    fhvp__ntwg = [ghpd__ovmzd for ghpd__ovmzd, numuq__qib in zip(df.columns,
        df.data) if bodo.utils.typing._is_pandas_numeric_dtype(numuq__qib.
        dtype)]
    assert len(fhvp__ntwg) != 0
    ezx__xfr = ''
    if not any(numuq__qib == types.float64 for numuq__qib in df.data):
        ezx__xfr = '.astype(np.float64)'
    ehe__ujg = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}){}'.
        format(df.columns.index(ghpd__ovmzd), '.astype(np.float64)' if 
        isinstance(df.data[df.columns.index(ghpd__ovmzd)], IntegerArrayType
        ) or df.data[df.columns.index(ghpd__ovmzd)] == boolean_array else
        '') for ghpd__ovmzd in fhvp__ntwg)
    lsmq__pik = 'np.stack(({},), 1){}'.format(ehe__ujg, ezx__xfr)
    data_args = ', '.join('res[:,{}]'.format(i) for i in range(len(fhvp__ntwg))
        )
    index = f'{generate_col_to_index_func_text(fhvp__ntwg)}\n'
    header = "def impl(df, method='pearson', min_periods=1):\n"
    header += '  mat = {}\n'.format(lsmq__pik)
    header += '  res = bodo.libs.array_kernels.nancorr(mat, 0, min_periods)\n'
    return _gen_init_df(header, fhvp__ntwg, data_args, index)


@lower_builtin('df.corr', DataFrameType, types.VarArg(types.Any))
def dataframe_corr_lower(context, builder, sig, args):
    impl = overload_dataframe_corr(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


@overload_method(DataFrameType, 'cov', inline='always', no_unliteral=True)
def overload_dataframe_cov(df, min_periods=None, ddof=1):
    check_runtime_cols_unsupported(df, 'DataFrame.cov()')
    qnjab__qwci = dict(ddof=ddof)
    klw__cab = dict(ddof=1)
    check_unsupported_args('DataFrame.cov', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    itifz__wgfe = '1' if is_overload_none(min_periods) else 'min_periods'
    fhvp__ntwg = [ghpd__ovmzd for ghpd__ovmzd, numuq__qib in zip(df.columns,
        df.data) if bodo.utils.typing._is_pandas_numeric_dtype(numuq__qib.
        dtype)]
    if len(fhvp__ntwg) == 0:
        raise_bodo_error('DataFrame.cov(): requires non-empty dataframe')
    ezx__xfr = ''
    if not any(numuq__qib == types.float64 for numuq__qib in df.data):
        ezx__xfr = '.astype(np.float64)'
    ehe__ujg = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}){}'.
        format(df.columns.index(ghpd__ovmzd), '.astype(np.float64)' if 
        isinstance(df.data[df.columns.index(ghpd__ovmzd)], IntegerArrayType
        ) or df.data[df.columns.index(ghpd__ovmzd)] == boolean_array else
        '') for ghpd__ovmzd in fhvp__ntwg)
    lsmq__pik = 'np.stack(({},), 1){}'.format(ehe__ujg, ezx__xfr)
    data_args = ', '.join('res[:,{}]'.format(i) for i in range(len(fhvp__ntwg))
        )
    index = f'pd.Index({fhvp__ntwg})\n'
    header = 'def impl(df, min_periods=None, ddof=1):\n'
    header += '  mat = {}\n'.format(lsmq__pik)
    header += '  res = bodo.libs.array_kernels.nancorr(mat, 1, {})\n'.format(
        itifz__wgfe)
    return _gen_init_df(header, fhvp__ntwg, data_args, index)


@overload_method(DataFrameType, 'count', inline='always', no_unliteral=True)
def overload_dataframe_count(df, axis=0, level=None, numeric_only=False):
    check_runtime_cols_unsupported(df, 'DataFrame.count()')
    qnjab__qwci = dict(axis=axis, level=level, numeric_only=numeric_only)
    klw__cab = dict(axis=0, level=None, numeric_only=False)
    check_unsupported_args('DataFrame.count', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    data_args = ', '.join(
        f'bodo.libs.array_ops.array_op_count(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}))'
         for i in range(len(df.columns)))
    visq__yelo = 'def impl(df, axis=0, level=None, numeric_only=False):\n'
    visq__yelo += '  data = np.array([{}])\n'.format(data_args)
    iylt__hhsz = bodo.hiframes.dataframe_impl.generate_col_to_index_func_text(
        df.columns)
    visq__yelo += (
        f'  return bodo.hiframes.pd_series_ext.init_series(data, {iylt__hhsz})\n'
        )
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'np': np}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@overload_method(DataFrameType, 'nunique', inline='always', no_unliteral=True)
def overload_dataframe_nunique(df, axis=0, dropna=True):
    check_runtime_cols_unsupported(df, 'DataFrame.unique()')
    qnjab__qwci = dict(axis=axis)
    klw__cab = dict(axis=0)
    if not is_overload_bool(dropna):
        raise BodoError('DataFrame.nunique: dropna must be a boolean value')
    check_unsupported_args('DataFrame.nunique', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    data_args = ', '.join(
        f'bodo.libs.array_kernels.nunique(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), dropna)'
         for i in range(len(df.columns)))
    visq__yelo = 'def impl(df, axis=0, dropna=True):\n'
    visq__yelo += '  data = np.asarray(({},))\n'.format(data_args)
    iylt__hhsz = bodo.hiframes.dataframe_impl.generate_col_to_index_func_text(
        df.columns)
    visq__yelo += (
        f'  return bodo.hiframes.pd_series_ext.init_series(data, {iylt__hhsz})\n'
        )
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'np': np}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@overload_method(DataFrameType, 'prod', inline='always', no_unliteral=True)
@overload_method(DataFrameType, 'product', inline='always', no_unliteral=True)
def overload_dataframe_prod(df, axis=None, skipna=None, level=None,
    numeric_only=None, min_count=0):
    check_runtime_cols_unsupported(df, 'DataFrame.prod()')
    qnjab__qwci = dict(skipna=skipna, level=level, numeric_only=
        numeric_only, min_count=min_count)
    klw__cab = dict(skipna=None, level=None, numeric_only=None, min_count=0)
    check_unsupported_args('DataFrame.prod', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'prod', axis=axis)


@overload_method(DataFrameType, 'sum', inline='always', no_unliteral=True)
def overload_dataframe_sum(df, axis=None, skipna=None, level=None,
    numeric_only=None, min_count=0):
    check_runtime_cols_unsupported(df, 'DataFrame.sum()')
    qnjab__qwci = dict(skipna=skipna, level=level, numeric_only=
        numeric_only, min_count=min_count)
    klw__cab = dict(skipna=None, level=None, numeric_only=None, min_count=0)
    check_unsupported_args('DataFrame.sum', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'sum', axis=axis)


@overload_method(DataFrameType, 'max', inline='always', no_unliteral=True)
def overload_dataframe_max(df, axis=None, skipna=None, level=None,
    numeric_only=None):
    check_runtime_cols_unsupported(df, 'DataFrame.max()')
    qnjab__qwci = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    klw__cab = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('DataFrame.max', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'max', axis=axis)


@overload_method(DataFrameType, 'min', inline='always', no_unliteral=True)
def overload_dataframe_min(df, axis=None, skipna=None, level=None,
    numeric_only=None):
    check_runtime_cols_unsupported(df, 'DataFrame.min()')
    qnjab__qwci = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    klw__cab = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('DataFrame.min', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'min', axis=axis)


@overload_method(DataFrameType, 'mean', inline='always', no_unliteral=True)
def overload_dataframe_mean(df, axis=None, skipna=None, level=None,
    numeric_only=None):
    check_runtime_cols_unsupported(df, 'DataFrame.mean()')
    qnjab__qwci = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    klw__cab = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('DataFrame.mean', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'mean', axis=axis)


@overload_method(DataFrameType, 'var', inline='always', no_unliteral=True)
def overload_dataframe_var(df, axis=None, skipna=None, level=None, ddof=1,
    numeric_only=None):
    check_runtime_cols_unsupported(df, 'DataFrame.var()')
    qnjab__qwci = dict(skipna=skipna, level=level, ddof=ddof, numeric_only=
        numeric_only)
    klw__cab = dict(skipna=None, level=None, ddof=1, numeric_only=None)
    check_unsupported_args('DataFrame.var', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'var', axis=axis)


@overload_method(DataFrameType, 'std', inline='always', no_unliteral=True)
def overload_dataframe_std(df, axis=None, skipna=None, level=None, ddof=1,
    numeric_only=None):
    check_runtime_cols_unsupported(df, 'DataFrame.std()')
    qnjab__qwci = dict(skipna=skipna, level=level, ddof=ddof, numeric_only=
        numeric_only)
    klw__cab = dict(skipna=None, level=None, ddof=1, numeric_only=None)
    check_unsupported_args('DataFrame.std', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'std', axis=axis)


@overload_method(DataFrameType, 'median', inline='always', no_unliteral=True)
def overload_dataframe_median(df, axis=None, skipna=None, level=None,
    numeric_only=None):
    check_runtime_cols_unsupported(df, 'DataFrame.median()')
    qnjab__qwci = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    klw__cab = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('DataFrame.median', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'median', axis=axis)


@overload_method(DataFrameType, 'quantile', inline='always', no_unliteral=True)
def overload_dataframe_quantile(df, q=0.5, axis=0, numeric_only=True,
    interpolation='linear'):
    check_runtime_cols_unsupported(df, 'DataFrame.quantile()')
    qnjab__qwci = dict(numeric_only=numeric_only, interpolation=interpolation)
    klw__cab = dict(numeric_only=True, interpolation='linear')
    check_unsupported_args('DataFrame.quantile', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    return _gen_reduce_impl(df, 'quantile', 'q', axis=axis)


@overload_method(DataFrameType, 'idxmax', inline='always', no_unliteral=True)
def overload_dataframe_idxmax(df, axis=0, skipna=True):
    check_runtime_cols_unsupported(df, 'DataFrame.idxmax()')
    qnjab__qwci = dict(axis=axis, skipna=skipna)
    klw__cab = dict(axis=0, skipna=True)
    check_unsupported_args('DataFrame.idxmax', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    for gltj__kumfb in df.data:
        if not (bodo.utils.utils.is_np_array_typ(gltj__kumfb) and (
            gltj__kumfb.dtype in [bodo.datetime64ns, bodo.timedelta64ns] or
            isinstance(gltj__kumfb.dtype, (types.Number, types.Boolean))) or
            isinstance(gltj__kumfb, (bodo.IntegerArrayType, bodo.
            CategoricalArrayType)) or gltj__kumfb in [bodo.boolean_array,
            bodo.datetime_date_array_type]):
            raise BodoError(
                f'DataFrame.idxmax() only supported for numeric column types. Column type: {gltj__kumfb} not supported.'
                )
        if isinstance(gltj__kumfb, bodo.CategoricalArrayType
            ) and not gltj__kumfb.dtype.ordered:
            raise BodoError(
                'DataFrame.idxmax(): categorical columns must be ordered')
    return _gen_reduce_impl(df, 'idxmax', axis=axis)


@overload_method(DataFrameType, 'idxmin', inline='always', no_unliteral=True)
def overload_dataframe_idxmin(df, axis=0, skipna=True):
    check_runtime_cols_unsupported(df, 'DataFrame.idxmin()')
    qnjab__qwci = dict(axis=axis, skipna=skipna)
    klw__cab = dict(axis=0, skipna=True)
    check_unsupported_args('DataFrame.idxmin', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    for gltj__kumfb in df.data:
        if not (bodo.utils.utils.is_np_array_typ(gltj__kumfb) and (
            gltj__kumfb.dtype in [bodo.datetime64ns, bodo.timedelta64ns] or
            isinstance(gltj__kumfb.dtype, (types.Number, types.Boolean))) or
            isinstance(gltj__kumfb, (bodo.IntegerArrayType, bodo.
            CategoricalArrayType)) or gltj__kumfb in [bodo.boolean_array,
            bodo.datetime_date_array_type]):
            raise BodoError(
                f'DataFrame.idxmin() only supported for numeric column types. Column type: {gltj__kumfb} not supported.'
                )
        if isinstance(gltj__kumfb, bodo.CategoricalArrayType
            ) and not gltj__kumfb.dtype.ordered:
            raise BodoError(
                'DataFrame.idxmin(): categorical columns must be ordered')
    return _gen_reduce_impl(df, 'idxmin', axis=axis)


@overload_method(DataFrameType, 'infer_objects', inline='always')
def overload_dataframe_infer_objects(df):
    check_runtime_cols_unsupported(df, 'DataFrame.infer_objects()')
    return lambda df: df.copy()


def _gen_reduce_impl(df, func_name, args=None, axis=None):
    args = '' if is_overload_none(args) else args
    if is_overload_none(axis):
        axis = 0
    elif is_overload_constant_int(axis):
        axis = get_overload_const_int(axis)
    else:
        raise_bodo_error(
            f'DataFrame.{func_name}: axis must be a constant Integer')
    assert axis in (0, 1), f'invalid axis argument for DataFrame.{func_name}'
    if func_name in ('idxmax', 'idxmin'):
        out_colnames = df.columns
    else:
        fhvp__ntwg = tuple(ghpd__ovmzd for ghpd__ovmzd, numuq__qib in zip(
            df.columns, df.data) if bodo.utils.typing.
            _is_pandas_numeric_dtype(numuq__qib.dtype))
        out_colnames = fhvp__ntwg
    assert len(out_colnames) != 0
    try:
        if func_name in ('idxmax', 'idxmin') and axis == 0:
            comm_dtype = None
        else:
            ybvpx__eazb = [numba.np.numpy_support.as_dtype(df.data[df.
                columns.index(ghpd__ovmzd)].dtype) for ghpd__ovmzd in
                out_colnames]
            comm_dtype = numba.np.numpy_support.from_dtype(np.
                find_common_type(ybvpx__eazb, []))
    except NotImplementedError as chldp__cjrq:
        raise BodoError(
            f'Dataframe.{func_name}() with column types: {df.data} could not be merged to a common type.'
            )
    scpxp__llkaa = ''
    if func_name in ('sum', 'prod'):
        scpxp__llkaa = ', min_count=0'
    ddof = ''
    if func_name in ('var', 'std'):
        ddof = 'ddof=1, '
    visq__yelo = (
        'def impl(df, axis=None, skipna=None, level=None,{} numeric_only=None{}):\n'
        .format(ddof, scpxp__llkaa))
    if func_name == 'quantile':
        visq__yelo = (
            "def impl(df, q=0.5, axis=0, numeric_only=True, interpolation='linear'):\n"
            )
    if func_name in ('idxmax', 'idxmin'):
        visq__yelo = 'def impl(df, axis=0, skipna=True):\n'
    if axis == 0:
        visq__yelo += _gen_reduce_impl_axis0(df, func_name, out_colnames,
            comm_dtype, args)
    else:
        visq__yelo += _gen_reduce_impl_axis1(func_name, out_colnames,
            comm_dtype, df)
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'np': np, 'pd': pd, 'numba': numba},
        zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


def _gen_reduce_impl_axis0(df, func_name, out_colnames, comm_dtype, args):
    arsig__sauyv = ''
    if func_name in ('min', 'max'):
        arsig__sauyv = ', dtype=np.{}'.format(comm_dtype)
    if comm_dtype == types.float32 and func_name in ('sum', 'prod', 'mean',
        'var', 'std', 'median'):
        arsig__sauyv = ', dtype=np.float32'
    dvu__ckwt = f'bodo.libs.array_ops.array_op_{func_name}'
    gnn__idkxw = ''
    if func_name in ['sum', 'prod']:
        gnn__idkxw = 'True, min_count'
    elif func_name in ['idxmax', 'idxmin']:
        gnn__idkxw = 'index'
    elif func_name == 'quantile':
        gnn__idkxw = 'q'
    elif func_name in ['std', 'var']:
        gnn__idkxw = 'True, ddof'
    elif func_name == 'median':
        gnn__idkxw = 'True'
    data_args = ', '.join(
        f'{dvu__ckwt}(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(ghpd__ovmzd)}), {gnn__idkxw})'
         for ghpd__ovmzd in out_colnames)
    visq__yelo = ''
    if func_name in ('idxmax', 'idxmin'):
        visq__yelo += (
            '  index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)\n'
            )
        visq__yelo += (
            '  data = bodo.utils.conversion.coerce_to_array(({},))\n'.
            format(data_args))
    else:
        visq__yelo += '  data = np.asarray(({},){})\n'.format(data_args,
            arsig__sauyv)
    visq__yelo += f"""  return bodo.hiframes.pd_series_ext.init_series(data, pd.Index({out_colnames}))
"""
    return visq__yelo


def _gen_reduce_impl_axis1(func_name, out_colnames, comm_dtype, df_type):
    zig__gsj = [df_type.columns.index(ghpd__ovmzd) for ghpd__ovmzd in
        out_colnames]
    index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
    data_args = '\n    '.join(
        'arr_{0} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0})'
        .format(i) for i in zig__gsj)
    bxawv__yuq = '\n        '.join(f'row[{i}] = arr_{zig__gsj[i]}[i]' for i in
        range(len(out_colnames)))
    assert len(data_args) > 0, f'empty dataframe in DataFrame.{func_name}()'
    sqci__jwled = f'len(arr_{zig__gsj[0]})'
    tkjl__rdnco = {'max': 'np.nanmax', 'min': 'np.nanmin', 'sum':
        'np.nansum', 'prod': 'np.nanprod', 'mean': 'np.nanmean', 'median':
        'np.nanmedian', 'var': 'bodo.utils.utils.nanvar_ddof1', 'std':
        'bodo.utils.utils.nanstd_ddof1'}
    if func_name in tkjl__rdnco:
        fdofq__wkr = tkjl__rdnco[func_name]
        ezhkt__nox = 'float64' if func_name in ['mean', 'median', 'std', 'var'
            ] else comm_dtype
        visq__yelo = f"""
    {data_args}
    numba.parfors.parfor.init_prange()
    n = {sqci__jwled}
    row = np.empty({len(out_colnames)}, np.{comm_dtype})
    A = np.empty(n, np.{ezhkt__nox})
    for i in numba.parfors.parfor.internal_prange(n):
        {bxawv__yuq}
        A[i] = {fdofq__wkr}(row)
    return bodo.hiframes.pd_series_ext.init_series(A, {index})
"""
        return visq__yelo
    else:
        raise BodoError(f'DataFrame.{func_name}(): Not supported for axis=1')


@overload_method(DataFrameType, 'pct_change', inline='always', no_unliteral
    =True)
def overload_dataframe_pct_change(df, periods=1, fill_method='pad', limit=
    None, freq=None):
    check_runtime_cols_unsupported(df, 'DataFrame.pct_change()')
    qnjab__qwci = dict(fill_method=fill_method, limit=limit, freq=freq)
    klw__cab = dict(fill_method='pad', limit=None, freq=None)
    check_unsupported_args('DataFrame.pct_change', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    data_args = ', '.join(
        f'bodo.hiframes.rolling.pct_change(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), periods, False)'
         for i in range(len(df.columns)))
    header = (
        "def impl(df, periods=1, fill_method='pad', limit=None, freq=None):\n")
    return _gen_init_df(header, df.columns, data_args)


@overload_method(DataFrameType, 'cumprod', inline='always', no_unliteral=True)
def overload_dataframe_cumprod(df, axis=None, skipna=True):
    check_runtime_cols_unsupported(df, 'DataFrame.cumprod()')
    qnjab__qwci = dict(axis=axis, skipna=skipna)
    klw__cab = dict(axis=None, skipna=True)
    check_unsupported_args('DataFrame.cumprod', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}).cumprod()'
         for i in range(len(df.columns)))
    header = 'def impl(df, axis=None, skipna=True):\n'
    return _gen_init_df(header, df.columns, data_args)


@overload_method(DataFrameType, 'cumsum', inline='always', no_unliteral=True)
def overload_dataframe_cumsum(df, axis=None, skipna=True):
    check_runtime_cols_unsupported(df, 'DataFrame.cumsum()')
    qnjab__qwci = dict(skipna=skipna)
    klw__cab = dict(skipna=True)
    check_unsupported_args('DataFrame.cumsum', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    data_args = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}).cumsum()'
         for i in range(len(df.columns)))
    header = 'def impl(df, axis=None, skipna=True):\n'
    return _gen_init_df(header, df.columns, data_args)


def _is_describe_type(data):
    return isinstance(data, IntegerArrayType) or isinstance(data, types.Array
        ) and isinstance(data.dtype, types.Number
        ) or data.dtype == bodo.datetime64ns


@overload_method(DataFrameType, 'describe', inline='always', no_unliteral=True)
def overload_dataframe_describe(df, percentiles=None, include=None, exclude
    =None, datetime_is_numeric=True):
    check_runtime_cols_unsupported(df, 'DataFrame.describe()')
    qnjab__qwci = dict(percentiles=percentiles, include=include, exclude=
        exclude, datetime_is_numeric=datetime_is_numeric)
    klw__cab = dict(percentiles=None, include=None, exclude=None,
        datetime_is_numeric=True)
    check_unsupported_args('DataFrame.describe', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    fhvp__ntwg = [ghpd__ovmzd for ghpd__ovmzd, numuq__qib in zip(df.columns,
        df.data) if _is_describe_type(numuq__qib)]
    if len(fhvp__ntwg) == 0:
        raise BodoError('df.describe() only supports numeric columns')
    bhm__erbl = sum(df.data[df.columns.index(ghpd__ovmzd)].dtype == bodo.
        datetime64ns for ghpd__ovmzd in fhvp__ntwg)

    def _get_describe(col_ind):
        swt__qrgv = df.data[col_ind].dtype == bodo.datetime64ns
        if bhm__erbl and bhm__erbl != len(fhvp__ntwg):
            if swt__qrgv:
                return f'des_{col_ind} + (np.nan,)'
            return (
                f'des_{col_ind}[:2] + des_{col_ind}[3:] + (des_{col_ind}[2],)')
        return f'des_{col_ind}'
    header = """def impl(df, percentiles=None, include=None, exclude=None, datetime_is_numeric=True):
"""
    for ghpd__ovmzd in fhvp__ntwg:
        col_ind = df.columns.index(ghpd__ovmzd)
        header += f"""  des_{col_ind} = bodo.libs.array_ops.array_op_describe(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {col_ind}))
"""
    data_args = ', '.join(_get_describe(df.columns.index(ghpd__ovmzd)) for
        ghpd__ovmzd in fhvp__ntwg)
    pmjl__rfscw = "['count', 'mean', 'std', 'min', '25%', '50%', '75%', 'max']"
    if bhm__erbl == len(fhvp__ntwg):
        pmjl__rfscw = "['count', 'mean', 'min', '25%', '50%', '75%', 'max']"
    elif bhm__erbl:
        pmjl__rfscw = (
            "['count', 'mean', 'min', '25%', '50%', '75%', 'max', 'std']")
    index = f'bodo.utils.conversion.convert_to_index({pmjl__rfscw})'
    return _gen_init_df(header, fhvp__ntwg, data_args, index)


@overload_method(DataFrameType, 'take', inline='always', no_unliteral=True)
def overload_dataframe_take(df, indices, axis=0, convert=None, is_copy=True):
    check_runtime_cols_unsupported(df, 'DataFrame.take()')
    qnjab__qwci = dict(axis=axis, convert=convert, is_copy=is_copy)
    klw__cab = dict(axis=0, convert=None, is_copy=True)
    check_unsupported_args('DataFrame.take', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    data_args = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})[indices_t]'
        .format(i) for i in range(len(df.columns)))
    header = 'def impl(df, indices, axis=0, convert=None, is_copy=True):\n'
    header += (
        '  indices_t = bodo.utils.conversion.coerce_to_ndarray(indices)\n')
    index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[indices_t]'
    return _gen_init_df(header, df.columns, data_args, index)


@overload_method(DataFrameType, 'shift', inline='always', no_unliteral=True)
def overload_dataframe_shift(df, periods=1, freq=None, axis=0, fill_value=None
    ):
    check_runtime_cols_unsupported(df, 'DataFrame.shift()')
    qnjab__qwci = dict(freq=freq, axis=axis, fill_value=fill_value)
    klw__cab = dict(freq=None, axis=0, fill_value=None)
    check_unsupported_args('DataFrame.shift', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    for gzcbd__pvl in df.data:
        if not is_supported_shift_array_type(gzcbd__pvl):
            raise BodoError(
                f'Dataframe.shift() column input type {gzcbd__pvl.dtype} not supported yet.'
                )
    if not is_overload_int(periods):
        raise BodoError(
            "DataFrame.shift(): 'periods' input must be an integer.")
    data_args = ', '.join(
        f'bodo.hiframes.rolling.shift(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), periods, False)'
         for i in range(len(df.columns)))
    header = 'def impl(df, periods=1, freq=None, axis=0, fill_value=None):\n'
    return _gen_init_df(header, df.columns, data_args)


@overload_method(DataFrameType, 'diff', inline='always', no_unliteral=True)
def overload_dataframe_diff(df, periods=1, axis=0):
    check_runtime_cols_unsupported(df, 'DataFrame.diff()')
    qnjab__qwci = dict(axis=axis)
    klw__cab = dict(axis=0)
    check_unsupported_args('DataFrame.diff', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    for gzcbd__pvl in df.data:
        if not (isinstance(gzcbd__pvl, types.Array) and (isinstance(
            gzcbd__pvl.dtype, types.Number) or gzcbd__pvl.dtype == bodo.
            datetime64ns)):
            raise BodoError(
                f'DataFrame.diff() column input type {gzcbd__pvl.dtype} not supported.'
                )
    if not is_overload_int(periods):
        raise BodoError("DataFrame.diff(): 'periods' input must be an integer."
            )
    header = 'def impl(df, periods=1, axis= 0):\n'
    for i in range(len(df.columns)):
        header += (
            f'  data_{i} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})\n'
            )
    data_args = ', '.join(
        f'bodo.hiframes.series_impl.dt64_arr_sub(data_{i}, bodo.hiframes.rolling.shift(data_{i}, periods, False))'
         if df.data[i] == types.Array(bodo.datetime64ns, 1, 'C') else
        f'data_{i} - bodo.hiframes.rolling.shift(data_{i}, periods, False)' for
        i in range(len(df.columns)))
    return _gen_init_df(header, df.columns, data_args)


@overload_method(DataFrameType, 'explode', inline='always', no_unliteral=True)
def overload_dataframe_explode(df, column, ignore_index=False):
    ubp__jzna = (
        "DataFrame.explode(): 'column' must a constant label or list of labels"
        )
    if not is_literal_type(column):
        raise_bodo_error(ubp__jzna)
    if is_overload_constant_list(column) or is_overload_constant_tuple(column):
        wdj__hoiu = get_overload_const_list(column)
    else:
        wdj__hoiu = [get_literal_value(column)]
    hyms__wdryu = {ghpd__ovmzd: i for i, ghpd__ovmzd in enumerate(df.columns)}
    abho__tjwdb = [hyms__wdryu[ghpd__ovmzd] for ghpd__ovmzd in wdj__hoiu]
    for i in abho__tjwdb:
        if not isinstance(df.data[i], ArrayItemArrayType) and df.data[i
            ].dtype != string_array_split_view_type:
            raise BodoError(
                f'DataFrame.explode(): columns must have array-like entries')
    n = len(df.columns)
    header = 'def impl(df, column, ignore_index=False):\n'
    header += (
        '  index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)\n')
    header += '  index_arr = bodo.utils.conversion.index_to_array(index)\n'
    for i in range(n):
        header += (
            f'  data{i} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})\n'
            )
    header += (
        f'  counts = bodo.libs.array_kernels.get_arr_lens(data{abho__tjwdb[0]})\n'
        )
    for i in range(n):
        if i in abho__tjwdb:
            header += (
                f'  out_data{i} = bodo.libs.array_kernels.explode_no_index(data{i}, counts)\n'
                )
        else:
            header += (
                f'  out_data{i} = bodo.libs.array_kernels.repeat_kernel(data{i}, counts)\n'
                )
    header += (
        '  new_index = bodo.libs.array_kernels.repeat_kernel(index_arr, counts)\n'
        )
    data_args = ', '.join(f'out_data{i}' for i in range(n))
    index = 'bodo.utils.conversion.convert_to_index(new_index)'
    return _gen_init_df(header, df.columns, data_args, index)


@overload_method(DataFrameType, 'set_index', inline='always', no_unliteral=True
    )
def overload_dataframe_set_index(df, keys, drop=True, append=False, inplace
    =False, verify_integrity=False):
    check_runtime_cols_unsupported(df, 'DataFrame.set_index()')
    hwwj__nvjxf = {'inplace': inplace, 'append': append, 'verify_integrity':
        verify_integrity}
    hriw__fson = {'inplace': False, 'append': False, 'verify_integrity': False}
    check_unsupported_args('DataFrame.set_index', hwwj__nvjxf, hriw__fson,
        package_name='pandas', module_name='DataFrame')
    if not is_overload_constant_str(keys):
        raise_bodo_error(
            "DataFrame.set_index(): 'keys' must be a constant string")
    col_name = get_overload_const_str(keys)
    col_ind = df.columns.index(col_name)
    if len(df.columns) == 1:
        raise BodoError(
            'DataFrame.set_index(): Not supported on single column DataFrames.'
            )
    data_args = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})'.format(
        i) for i in range(len(df.columns)) if i != col_ind)
    header = """def impl(df, keys, drop=True, append=False, inplace=False, verify_integrity=False):
"""
    columns = tuple(ghpd__ovmzd for ghpd__ovmzd in df.columns if 
        ghpd__ovmzd != col_name)
    index = (
        'bodo.utils.conversion.index_from_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}), {})'
        .format(col_ind, f"'{col_name}'" if isinstance(col_name, str) else
        col_name))
    return _gen_init_df(header, columns, data_args, index)


@overload_method(DataFrameType, 'query', no_unliteral=True)
def overload_dataframe_query(df, expr, inplace=False):
    check_runtime_cols_unsupported(df, 'DataFrame.query()')
    hwwj__nvjxf = {'inplace': inplace}
    hriw__fson = {'inplace': False}
    check_unsupported_args('query', hwwj__nvjxf, hriw__fson, package_name=
        'pandas', module_name='DataFrame')
    if not isinstance(expr, (types.StringLiteral, types.UnicodeType)):
        raise BodoError('query(): expr argument should be a string')

    def impl(df, expr, inplace=False):
        qiuv__sdc = bodo.hiframes.pd_dataframe_ext.query_dummy(df, expr)
        return df[qiuv__sdc]
    return impl


@overload_method(DataFrameType, 'duplicated', inline='always', no_unliteral
    =True)
def overload_dataframe_duplicated(df, subset=None, keep='first'):
    check_runtime_cols_unsupported(df, 'DataFrame.duplicated()')
    hwwj__nvjxf = {'subset': subset, 'keep': keep}
    hriw__fson = {'subset': None, 'keep': 'first'}
    check_unsupported_args('DataFrame.duplicated', hwwj__nvjxf, hriw__fson,
        package_name='pandas', module_name='DataFrame')
    potqq__ublqz = len(df.columns)
    visq__yelo = "def impl(df, subset=None, keep='first'):\n"
    for i in range(potqq__ublqz):
        visq__yelo += (
            '  data_{0} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0})\n'
            .format(i))
    index = (
        'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df))'
        )
    visq__yelo += (
        '  duplicated, index_arr = bodo.libs.array_kernels.duplicated(({},), {})\n'
        .format(', '.join('data_{}'.format(i) for i in range(potqq__ublqz)),
        index))
    visq__yelo += (
        '  index = bodo.utils.conversion.index_from_array(index_arr)\n')
    visq__yelo += (
        '  return bodo.hiframes.pd_series_ext.init_series(duplicated, index)\n'
        )
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@overload_method(DataFrameType, 'drop_duplicates', inline='always',
    no_unliteral=True)
def overload_dataframe_drop_duplicates(df, subset=None, keep='first',
    inplace=False, ignore_index=False):
    check_runtime_cols_unsupported(df, 'DataFrame.drop_duplicates()')
    hwwj__nvjxf = {'keep': keep, 'inplace': inplace, 'ignore_index':
        ignore_index}
    hriw__fson = {'keep': 'first', 'inplace': False, 'ignore_index': False}
    vklg__joade = []
    if is_overload_constant_list(subset):
        vklg__joade = get_overload_const_list(subset)
    elif is_overload_constant_str(subset):
        vklg__joade = [get_overload_const_str(subset)]
    elif is_overload_constant_int(subset):
        vklg__joade = [get_overload_const_int(subset)]
    elif not is_overload_none(subset):
        raise_bodo_error(
            'DataFrame.drop_duplicates(): subset must be a constant column name, constant list of column names or None'
            )
    uhp__sqxmi = []
    for col_name in vklg__joade:
        if col_name not in df.columns:
            raise BodoError(
                'DataFrame.drop_duplicates(): All subset columns must be found in the DataFrame.'
                 +
                f'Column {col_name} not found in DataFrame columns {df.columns}'
                )
        uhp__sqxmi.append(df.columns.index(col_name))
    check_unsupported_args('DataFrame.drop_duplicates', hwwj__nvjxf,
        hriw__fson, package_name='pandas', module_name='DataFrame')
    pmzq__lpmzz = []
    if uhp__sqxmi:
        for tyjha__vjj in uhp__sqxmi:
            if isinstance(df.data[tyjha__vjj], bodo.MapArrayType):
                pmzq__lpmzz.append(df.columns[tyjha__vjj])
    else:
        for i, col_name in enumerate(df.columns):
            if isinstance(df.data[i], bodo.MapArrayType):
                pmzq__lpmzz.append(col_name)
    if pmzq__lpmzz:
        raise BodoError(
            f'DataFrame.drop_duplicates(): Columns {pmzq__lpmzz} ' +
            f'have dictionary types which cannot be used to drop duplicates. '
             +
            "Please consider using the 'subset' argument to skip these columns."
            )
    potqq__ublqz = len(df.columns)
    gpcwn__nzxs = ['data_{}'.format(i) for i in uhp__sqxmi]
    jdd__jnk = ['data_{}'.format(i) for i in range(potqq__ublqz) if i not in
        uhp__sqxmi]
    if gpcwn__nzxs:
        smlx__lkf = len(gpcwn__nzxs)
    else:
        smlx__lkf = potqq__ublqz
    eta__dfhh = ', '.join(gpcwn__nzxs + jdd__jnk)
    data_args = ', '.join('data_{}'.format(i) for i in range(potqq__ublqz))
    visq__yelo = (
        "def impl(df, subset=None, keep='first', inplace=False, ignore_index=False):\n"
        )
    for i in range(potqq__ublqz):
        visq__yelo += (
            '  data_{0} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0})\n'
            .format(i))
    index = (
        'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df))'
        )
    visq__yelo += (
        """  ({0},), index_arr = bodo.libs.array_kernels.drop_duplicates(({0},), {1}, {2})
"""
        .format(eta__dfhh, index, smlx__lkf))
    visq__yelo += (
        '  index = bodo.utils.conversion.index_from_array(index_arr)\n')
    return _gen_init_df(visq__yelo, df.columns, data_args, 'index')


def create_dataframe_mask_where_overload(func_name):

    def overload_dataframe_mask_where(df, cond, other=np.nan, inplace=False,
        axis=None, level=None, errors='raise', try_cast=False):
        _validate_arguments_mask_where(f'DataFrame.{func_name}', df, cond,
            other, inplace, axis, level, errors, try_cast)
        header = """def impl(df, cond, other=np.nan, inplace=False, axis=None, level=None, errors='raise', try_cast=False):
"""
        if func_name == 'mask':
            header += '  cond = ~cond\n'
        gen_all_false = [False]
        if cond.ndim == 1:
            cond_str = lambda i, _: 'cond'
        elif cond.ndim == 2:
            if isinstance(cond, DataFrameType):
                vfu__ddexs = {ghpd__ovmzd: i for i, ghpd__ovmzd in
                    enumerate(cond.columns)}

                def cond_str(i, gen_all_false):
                    if df.columns[i] in vfu__ddexs:
                        return (
                            f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(cond, {vfu__ddexs[df.columns[i]]})'
                            )
                    else:
                        gen_all_false[0] = True
                        return 'all_false'
            elif isinstance(cond, types.Array):
                cond_str = lambda i, _: f'cond[:,{i}]'
        if not hasattr(other, 'ndim') or other.ndim == 1:
            rfo__jcb = lambda i: 'other'
        elif other.ndim == 2:
            if isinstance(other, DataFrameType):
                other_map = {ghpd__ovmzd: i for i, ghpd__ovmzd in enumerate
                    (other.columns)}
                rfo__jcb = (lambda i: 
                    f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(other, {other_map[df.columns[i]]})'
                     if df.columns[i] in other_map else 'None')
            elif isinstance(other, types.Array):
                rfo__jcb = lambda i: f'other[:,{i}]'
        potqq__ublqz = len(df.columns)
        data_args = ', '.join(
            f'bodo.hiframes.series_impl.where_impl({cond_str(i, gen_all_false)}, bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}), {rfo__jcb(i)})'
             for i in range(potqq__ublqz))
        if gen_all_false[0]:
            header += '  all_false = np.zeros(len(df), dtype=bool)\n'
        return _gen_init_df(header, df.columns, data_args)
    return overload_dataframe_mask_where


def _install_dataframe_mask_where_overload():
    for func_name in ('mask', 'where'):
        rgx__proe = create_dataframe_mask_where_overload(func_name)
        overload_method(DataFrameType, func_name, no_unliteral=True)(rgx__proe)


_install_dataframe_mask_where_overload()


def _validate_arguments_mask_where(func_name, df, cond, other, inplace,
    axis, level, errors, try_cast):
    qnjab__qwci = dict(inplace=inplace, level=level, errors=errors,
        try_cast=try_cast)
    klw__cab = dict(inplace=False, level=None, errors='raise', try_cast=False)
    check_unsupported_args(f'{func_name}', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error(f'{func_name}(): axis argument not supported')
    if not (isinstance(cond, (SeriesType, types.Array, BooleanArrayType)) and
        (cond.ndim == 1 or cond.ndim == 2) and cond.dtype == types.bool_
        ) and not (isinstance(cond, DataFrameType) and cond.ndim == 2 and
        all(cond.data[i].dtype == types.bool_ for i in range(len(df.columns)))
        ):
        raise BodoError(
            f"{func_name}(): 'cond' argument must be a DataFrame, Series, 1- or 2-dimensional array of booleans"
            )
    potqq__ublqz = len(df.columns)
    if hasattr(other, 'ndim') and (other.ndim != 1 or other.ndim != 2):
        if other.ndim == 2:
            if not isinstance(other, (DataFrameType, types.Array)):
                raise BodoError(
                    f"{func_name}(): 'other', if 2-dimensional, must be a DataFrame or array."
                    )
        elif other.ndim != 1:
            raise BodoError(
                f"{func_name}(): 'other' must be either 1 or 2-dimensional")
    if isinstance(other, DataFrameType):
        other_map = {ghpd__ovmzd: i for i, ghpd__ovmzd in enumerate(other.
            columns)}
        for i in range(potqq__ublqz):
            if df.columns[i] in other_map:
                bodo.hiframes.series_impl._validate_self_other_mask_where(
                    func_name, df.data[i], other.data[other_map[df.columns[i]]]
                    )
            else:
                bodo.hiframes.series_impl._validate_self_other_mask_where(
                    func_name, df.data[i], None, is_default=True)
    elif isinstance(other, SeriesType):
        for i in range(potqq__ublqz):
            bodo.hiframes.series_impl._validate_self_other_mask_where(func_name
                , df.data[i], other.data)
    else:
        for i in range(potqq__ublqz):
            bodo.hiframes.series_impl._validate_self_other_mask_where(func_name
                , df.data[i], other, max_ndim=2)


def _gen_init_df(header, columns, data_args, index=None, extra_globals=None,
    out_df_type=None):
    if index is None:
        index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
    if extra_globals is None:
        extra_globals = {}
    if out_df_type is not None:
        extra_globals['out_df_type'] = out_df_type
        vxntz__piuya = 'out_df_type'
    else:
        vxntz__piuya = gen_const_tup(columns)
    data_args = '({}{})'.format(data_args, ',' if data_args else '')
    visq__yelo = f"""{header}  return bodo.hiframes.pd_dataframe_ext.init_dataframe({data_args}, {index}, {vxntz__piuya})
"""
    zrt__acbb = {}
    bdbmb__vyoo = {'bodo': bodo, 'np': np, 'pd': pd, 'numba': numba}
    bdbmb__vyoo.update(extra_globals)
    exec(visq__yelo, bdbmb__vyoo, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


def _get_binop_columns(lhs, rhs, is_inplace=False):
    if lhs.columns != rhs.columns:
        wsb__jbfuy = pd.Index(lhs.columns)
        anuok__adl = pd.Index(rhs.columns)
        lgel__xbjf, mca__rpz, thtjh__aio = wsb__jbfuy.join(anuok__adl, how=
            'left' if is_inplace else 'outer', level=None, return_indexers=True
            )
        return tuple(lgel__xbjf), mca__rpz, thtjh__aio
    return lhs.columns, range(len(lhs.columns)), range(len(lhs.columns))


def create_binary_op_overload(op):

    def overload_dataframe_binary_op(lhs, rhs):
        szd__jmjo = numba.core.utils.OPERATORS_TO_BUILTINS[op]
        evc__gtm = operator.eq, operator.ne
        check_runtime_cols_unsupported(lhs, szd__jmjo)
        check_runtime_cols_unsupported(rhs, szd__jmjo)
        if isinstance(lhs, DataFrameType):
            if isinstance(rhs, DataFrameType):
                lgel__xbjf, mca__rpz, thtjh__aio = _get_binop_columns(lhs, rhs)
                data_args = ', '.join(
                    f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(lhs, {awyjj__xvvi}) {szd__jmjo}bodo.hiframes.pd_dataframe_ext.get_dataframe_data(rhs, {xah__xathi})'
                     if awyjj__xvvi != -1 and xah__xathi != -1 else
                    f'bodo.libs.array_kernels.gen_na_array(len(lhs), float64_arr_type)'
                     for awyjj__xvvi, xah__xathi in zip(mca__rpz, thtjh__aio))
                header = 'def impl(lhs, rhs):\n'
                index = (
                    'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(lhs)')
                return _gen_init_df(header, lgel__xbjf, data_args, index,
                    extra_globals={'float64_arr_type': types.Array(types.
                    float64, 1, 'C')})
            drx__pbjc = []
            ukxp__gcyzc = []
            if op in evc__gtm:
                for i, xeydf__mlim in enumerate(lhs.data):
                    if is_common_scalar_dtype([xeydf__mlim.dtype, rhs]):
                        drx__pbjc.append(
                            f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(lhs, {i}) {szd__jmjo} rhs'
                            )
                    else:
                        top__ovj = f'arr{i}'
                        ukxp__gcyzc.append(top__ovj)
                        drx__pbjc.append(top__ovj)
                data_args = ', '.join(drx__pbjc)
            else:
                data_args = ', '.join(
                    f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(lhs, {i}) {szd__jmjo} rhs'
                     for i in range(len(lhs.columns)))
            header = 'def impl(lhs, rhs):\n'
            if len(ukxp__gcyzc) > 0:
                header += '  numba.parfors.parfor.init_prange()\n'
                header += '  n = len(lhs)\n'
                header += ''.join(
                    f'  {top__ovj} = np.empty(n, dtype=np.bool_)\n' for
                    top__ovj in ukxp__gcyzc)
                header += (
                    '  for i in numba.parfors.parfor.internal_prange(n):\n')
                header += ''.join('    {0}[i] = {1}\n'.format(top__ovj, op ==
                    operator.ne) for top__ovj in ukxp__gcyzc)
            index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(lhs)'
            return _gen_init_df(header, lhs.columns, data_args, index)
        if isinstance(rhs, DataFrameType):
            drx__pbjc = []
            ukxp__gcyzc = []
            if op in evc__gtm:
                for i, xeydf__mlim in enumerate(rhs.data):
                    if is_common_scalar_dtype([lhs, xeydf__mlim.dtype]):
                        drx__pbjc.append(
                            f'lhs {szd__jmjo} bodo.hiframes.pd_dataframe_ext.get_dataframe_data(rhs, {i})'
                            )
                    else:
                        top__ovj = f'arr{i}'
                        ukxp__gcyzc.append(top__ovj)
                        drx__pbjc.append(top__ovj)
                data_args = ', '.join(drx__pbjc)
            else:
                data_args = ', '.join(
                    'lhs {1} bodo.hiframes.pd_dataframe_ext.get_dataframe_data(rhs, {0})'
                    .format(i, szd__jmjo) for i in range(len(rhs.columns)))
            header = 'def impl(lhs, rhs):\n'
            if len(ukxp__gcyzc) > 0:
                header += '  numba.parfors.parfor.init_prange()\n'
                header += '  n = len(rhs)\n'
                header += ''.join('  {0} = np.empty(n, dtype=np.bool_)\n'.
                    format(top__ovj) for top__ovj in ukxp__gcyzc)
                header += (
                    '  for i in numba.parfors.parfor.internal_prange(n):\n')
                header += ''.join('    {0}[i] = {1}\n'.format(top__ovj, op ==
                    operator.ne) for top__ovj in ukxp__gcyzc)
            index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(rhs)'
            return _gen_init_df(header, rhs.columns, data_args, index)
    return overload_dataframe_binary_op


skips = [operator.lt, operator.le, operator.eq, operator.ne, operator.gt,
    operator.ge, operator.add, operator.sub, operator.mul, operator.truediv,
    operator.floordiv, operator.pow, operator.mod]


def _install_binary_ops():
    for op in bodo.hiframes.pd_series_ext.series_binary_ops:
        if op in skips:
            continue
        rgx__proe = create_binary_op_overload(op)
        overload(op)(rgx__proe)


_install_binary_ops()


def create_inplace_binary_op_overload(op):

    def overload_dataframe_inplace_binary_op(left, right):
        szd__jmjo = numba.core.utils.OPERATORS_TO_BUILTINS[op]
        check_runtime_cols_unsupported(left, szd__jmjo)
        check_runtime_cols_unsupported(right, szd__jmjo)
        if isinstance(left, DataFrameType):
            if isinstance(right, DataFrameType):
                lgel__xbjf, _, thtjh__aio = _get_binop_columns(left, right,
                    True)
                visq__yelo = 'def impl(left, right):\n'
                for i, xah__xathi in enumerate(thtjh__aio):
                    if xah__xathi == -1:
                        visq__yelo += f"""  df_arr{i} = bodo.libs.array_kernels.gen_na_array(len(left), float64_arr_type)
"""
                        continue
                    visq__yelo += f"""  df_arr{i} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(left, {i})
"""
                    visq__yelo += f"""  df_arr{i} {szd__jmjo} bodo.hiframes.pd_dataframe_ext.get_dataframe_data(right, {xah__xathi})
"""
                data_args = ', '.join(f'df_arr{i}' for i in range(len(
                    lgel__xbjf)))
                index = (
                    'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(left)')
                return _gen_init_df(visq__yelo, lgel__xbjf, data_args,
                    index, extra_globals={'float64_arr_type': types.Array(
                    types.float64, 1, 'C')})
            visq__yelo = 'def impl(left, right):\n'
            for i in range(len(left.columns)):
                visq__yelo += (
                    """  df_arr{0} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(left, {0})
"""
                    .format(i))
                visq__yelo += '  df_arr{0} {1} right\n'.format(i, szd__jmjo)
            data_args = ', '.join('df_arr{}'.format(i) for i in range(len(
                left.columns)))
            index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(left)'
            return _gen_init_df(visq__yelo, left.columns, data_args, index)
    return overload_dataframe_inplace_binary_op


def _install_inplace_binary_ops():
    for op in bodo.hiframes.pd_series_ext.series_inplace_binary_ops:
        rgx__proe = create_inplace_binary_op_overload(op)
        overload(op, no_unliteral=True)(rgx__proe)


_install_inplace_binary_ops()


def create_unary_op_overload(op):

    def overload_dataframe_unary_op(df):
        if isinstance(df, DataFrameType):
            szd__jmjo = numba.core.utils.OPERATORS_TO_BUILTINS[op]
            check_runtime_cols_unsupported(df, szd__jmjo)
            data_args = ', '.join(
                '{1} bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0})'
                .format(i, szd__jmjo) for i in range(len(df.columns)))
            header = 'def impl(df):\n'
            return _gen_init_df(header, df.columns, data_args)
    return overload_dataframe_unary_op


def _install_unary_ops():
    for op in bodo.hiframes.pd_series_ext.series_unary_ops:
        rgx__proe = create_unary_op_overload(op)
        overload(op, no_unliteral=True)(rgx__proe)


_install_unary_ops()


def overload_isna(obj):
    check_runtime_cols_unsupported(obj, 'pd.isna()')
    if isinstance(obj, (DataFrameType, SeriesType)
        ) or bodo.hiframes.pd_index_ext.is_pd_index_type(obj):
        return lambda obj: obj.isna()
    if is_array_typ(obj):

        def impl(obj):
            numba.parfors.parfor.init_prange()
            n = len(obj)
            tkck__ivbw = np.empty(n, np.bool_)
            for i in numba.parfors.parfor.internal_prange(n):
                tkck__ivbw[i] = bodo.libs.array_kernels.isna(obj, i)
            return tkck__ivbw
        return impl


overload(pd.isna, inline='always')(overload_isna)
overload(pd.isnull, inline='always')(overload_isna)


@overload(pd.isna)
@overload(pd.isnull)
def overload_isna_scalar(obj):
    if isinstance(obj, (DataFrameType, SeriesType)
        ) or bodo.hiframes.pd_index_ext.is_pd_index_type(obj) or is_array_typ(
        obj):
        return
    if isinstance(obj, (types.List, types.UniTuple)):

        def impl(obj):
            n = len(obj)
            tkck__ivbw = np.empty(n, np.bool_)
            for i in range(n):
                tkck__ivbw[i] = pd.isna(obj[i])
            return tkck__ivbw
        return impl
    obj = types.unliteral(obj)
    if obj == bodo.string_type:
        return lambda obj: unliteral_val(False)
    if isinstance(obj, types.Integer):
        return lambda obj: unliteral_val(False)
    if isinstance(obj, types.Float):
        return lambda obj: np.isnan(obj)
    if isinstance(obj, (types.NPDatetime, types.NPTimedelta)):
        return lambda obj: np.isnat(obj)
    if obj == types.none:
        return lambda obj: unliteral_val(True)
    if obj == bodo.hiframes.pd_timestamp_ext.pd_timestamp_type:
        return lambda obj: np.isnat(bodo.hiframes.pd_timestamp_ext.
            integer_to_dt64(obj.value))
    if obj == bodo.hiframes.datetime_timedelta_ext.pd_timedelta_type:
        return lambda obj: np.isnat(bodo.hiframes.pd_timestamp_ext.
            integer_to_timedelta64(obj.value))
    if isinstance(obj, types.Optional):
        return lambda obj: obj is None
    return lambda obj: unliteral_val(False)


@overload(operator.setitem, no_unliteral=True)
def overload_setitem_arr_none(A, idx, val):
    if is_array_typ(A, False) and isinstance(idx, types.Integer
        ) and val == types.none:
        return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)


def overload_notna(obj):
    check_runtime_cols_unsupported(obj, 'pd.notna()')
    if isinstance(obj, DataFrameType):
        return lambda obj: obj.notna()
    if isinstance(obj, (SeriesType, types.Array, types.List, types.UniTuple)
        ) or bodo.hiframes.pd_index_ext.is_pd_index_type(obj
        ) or obj == bodo.string_array_type:
        return lambda obj: ~pd.isna(obj)
    return lambda obj: not pd.isna(obj)


overload(pd.notna, inline='always', no_unliteral=True)(overload_notna)
overload(pd.notnull, inline='always', no_unliteral=True)(overload_notna)


def _get_pd_dtype_str(t):
    if t.dtype == types.NPDatetime('ns'):
        return "'datetime64[ns]'"
    return bodo.ir.csv_ext._get_pd_dtype_str(t)


@overload_method(DataFrameType, 'replace', inline='always', no_unliteral=True)
def overload_dataframe_replace(df, to_replace=None, value=None, inplace=
    False, limit=None, regex=False, method='pad'):
    check_runtime_cols_unsupported(df, 'DataFrame.replace()')
    if is_overload_none(to_replace):
        raise BodoError('replace(): to_replace value of None is not supported')
    hwwj__nvjxf = {'inplace': inplace, 'limit': limit, 'regex': regex,
        'method': method}
    hriw__fson = {'inplace': False, 'limit': None, 'regex': False, 'method':
        'pad'}
    check_unsupported_args('replace', hwwj__nvjxf, hriw__fson, package_name
        ='pandas', module_name='DataFrame')
    data_args = ', '.join(
        f'df.iloc[:, {i}].replace(to_replace, value).values' for i in range
        (len(df.columns)))
    header = """def impl(df, to_replace=None, value=None, inplace=False, limit=None, regex=False, method='pad'):
"""
    return _gen_init_df(header, df.columns, data_args)


def _is_col_access(expr_node):
    diz__qvenb = str(expr_node)
    return diz__qvenb.startswith('left.') or diz__qvenb.startswith('right.')


def _insert_NA_cond(expr_node, left_columns, left_data, right_columns,
    right_data):
    sfv__ffjg = {'left': 0, 'right': 0, 'NOT_NA': 0}
    env = pd.core.computation.scope.ensure_scope(2, {}, {}, (sfv__ffjg,))
    djf__ncs = pd.core.computation.parsing.clean_column_name

    def append_null_checks(expr_node, null_set):
        if not null_set:
            return expr_node
        vdbba__koqvy = ' & '.join([('NOT_NA.`' + x + '`') for x in null_set])
        bjh__gsxei = {('NOT_NA', djf__ncs(xeydf__mlim)): xeydf__mlim for
            xeydf__mlim in null_set}
        dzx__yvxm, _, _ = _parse_query_expr(vdbba__koqvy, env, [], [], None,
            join_cleaned_cols=bjh__gsxei)
        brvc__lbzm = (pd.core.computation.ops.BinOp.
            _disallow_scalar_only_bool_ops)
        pd.core.computation.ops.BinOp._disallow_scalar_only_bool_ops = (lambda
            self: None)
        try:
            vam__aewyf = pd.core.computation.ops.BinOp('&', dzx__yvxm,
                expr_node)
        finally:
            (pd.core.computation.ops.BinOp._disallow_scalar_only_bool_ops
                ) = brvc__lbzm
        return vam__aewyf

    def _insert_NA_cond_body(expr_node, null_set):
        if isinstance(expr_node, pd.core.computation.ops.BinOp):
            if expr_node.op == '|':
                fkcm__gmkg = set()
                uzho__bnbhb = set()
                pmij__wfmpc = _insert_NA_cond_body(expr_node.lhs, fkcm__gmkg)
                ckzam__uzjv = _insert_NA_cond_body(expr_node.rhs, uzho__bnbhb)
                ocg__pnk = fkcm__gmkg.intersection(uzho__bnbhb)
                fkcm__gmkg.difference_update(ocg__pnk)
                uzho__bnbhb.difference_update(ocg__pnk)
                null_set.update(ocg__pnk)
                expr_node.lhs = append_null_checks(pmij__wfmpc, fkcm__gmkg)
                expr_node.rhs = append_null_checks(ckzam__uzjv, uzho__bnbhb)
                expr_node.operands = expr_node.lhs, expr_node.rhs
            else:
                expr_node.lhs = _insert_NA_cond_body(expr_node.lhs, null_set)
                expr_node.rhs = _insert_NA_cond_body(expr_node.rhs, null_set)
        elif _is_col_access(expr_node):
            npva__eipk = expr_node.name
            bhmqn__okxfn, col_name = npva__eipk.split('.')
            if bhmqn__okxfn == 'left':
                yei__yqe = left_columns
                data = left_data
            else:
                yei__yqe = right_columns
                data = right_data
            mto__fdrim = data[yei__yqe.index(col_name)]
            if bodo.utils.typing.is_nullable(mto__fdrim):
                null_set.add(expr_node.name)
        return expr_node
    null_set = set()
    aheke__iat = _insert_NA_cond_body(expr_node, null_set)
    return append_null_checks(expr_node, null_set)


def _extract_equal_conds(expr_node):
    if not hasattr(expr_node, 'op'):
        return [], [], expr_node
    if expr_node.op == '==' and _is_col_access(expr_node.lhs
        ) and _is_col_access(expr_node.rhs):
        mgns__stg = str(expr_node.lhs)
        tpj__xyqp = str(expr_node.rhs)
        if mgns__stg.startswith('left.') and tpj__xyqp.startswith('left.'
            ) or mgns__stg.startswith('right.') and tpj__xyqp.startswith(
            'right.'):
            return [], [], expr_node
        left_on = [mgns__stg.split('.')[1]]
        right_on = [tpj__xyqp.split('.')[1]]
        if mgns__stg.startswith('right.'):
            return right_on, left_on, None
        return left_on, right_on, None
    if expr_node.op == '&':
        vkknm__aaeol, ztt__buowv, ozq__hrvo = _extract_equal_conds(expr_node
            .lhs)
        yerit__atg, ujmy__mnqvo, zfk__sjzqg = _extract_equal_conds(expr_node
            .rhs)
        left_on = vkknm__aaeol + yerit__atg
        right_on = ztt__buowv + ujmy__mnqvo
        if ozq__hrvo is None:
            return left_on, right_on, zfk__sjzqg
        if zfk__sjzqg is None:
            return left_on, right_on, ozq__hrvo
        expr_node.lhs = ozq__hrvo
        expr_node.rhs = zfk__sjzqg
        expr_node.operands = expr_node.lhs, expr_node.rhs
        return left_on, right_on, expr_node
    return [], [], expr_node


def _parse_merge_cond(on_str, left_columns, left_data, right_columns,
    right_data):
    sfv__ffjg = {'left': 0, 'right': 0}
    env = pd.core.computation.scope.ensure_scope(2, {}, {}, (sfv__ffjg,))
    lyuys__zmjgk = dict()
    djf__ncs = pd.core.computation.parsing.clean_column_name
    for name, jea__yyy in (('left', left_columns), ('right', right_columns)):
        for xeydf__mlim in jea__yyy:
            ibch__iwwv = djf__ncs(xeydf__mlim)
            qzmv__nfxo = name, ibch__iwwv
            if qzmv__nfxo in lyuys__zmjgk:
                raise_bodo_error(
                    f"pd.merge(): {name} table contains two columns that are escaped to the same Python identifier '{xeydf__mlim}' and '{lyuys__zmjgk[ibch__iwwv]}' Please rename one of these columns. To avoid this issue, please use names that are valid Python identifiers."
                    )
            lyuys__zmjgk[qzmv__nfxo] = xeydf__mlim
    opopz__vhvh, _, _ = _parse_query_expr(on_str, env, [], [], None,
        join_cleaned_cols=lyuys__zmjgk)
    left_on, right_on, swuhm__mrge = _extract_equal_conds(opopz__vhvh.terms)
    return left_on, right_on, _insert_NA_cond(swuhm__mrge, left_columns,
        left_data, right_columns, right_data)


@overload_method(DataFrameType, 'merge', inline='always', no_unliteral=True)
@overload(pd.merge, inline='always', no_unliteral=True)
def overload_dataframe_merge(left, right, how='inner', on=None, left_on=
    None, right_on=None, left_index=False, right_index=False, sort=False,
    suffixes=('_x', '_y'), copy=True, indicator=False, validate=None,
    _bodo_na_equal=True):
    check_runtime_cols_unsupported(left, 'DataFrame.merge()')
    check_runtime_cols_unsupported(right, 'DataFrame.merge()')
    qnjab__qwci = dict(sort=sort, copy=copy, validate=validate)
    klw__cab = dict(sort=False, copy=True, validate=None)
    check_unsupported_args('DataFrame.merge', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    validate_merge_spec(left, right, how, on, left_on, right_on, left_index,
        right_index, sort, suffixes, copy, indicator, validate)
    how = get_overload_const_str(how)
    hskt__dnn = tuple(sorted(set(left.columns) & set(right.columns), key=lambda
        k: str(k)))
    fmp__shc = ''
    if not is_overload_none(on):
        left_on = right_on = on
        if is_overload_constant_str(on):
            on_str = get_overload_const_str(on)
            if on_str not in hskt__dnn and ('left.' in on_str or 'right.' in
                on_str):
                left_on, right_on, vkf__aqn = _parse_merge_cond(on_str,
                    left.columns, left.data, right.columns, right.data)
                if vkf__aqn is None:
                    fmp__shc = ''
                else:
                    fmp__shc = str(vkf__aqn)
    if is_overload_none(on) and is_overload_none(left_on) and is_overload_none(
        right_on) and is_overload_false(left_index) and is_overload_false(
        right_index):
        left_keys = hskt__dnn
        right_keys = hskt__dnn
    else:
        if is_overload_true(left_index):
            left_keys = ['$_bodo_index_']
        else:
            left_keys = get_overload_const_list(left_on)
            validate_keys(left_keys, left)
        if is_overload_true(right_index):
            right_keys = ['$_bodo_index_']
        else:
            right_keys = get_overload_const_list(right_on)
            validate_keys(right_keys, right)
    if (not left_on or not right_on) and not is_overload_none(on):
        raise BodoError(
            f"DataFrame.merge(): Merge condition '{get_overload_const_str(on)}' requires a cross join to implement, but cross join is not supported."
            )
    if not is_overload_bool(indicator):
        raise_bodo_error(
            'DataFrame.merge(): indicator must be a constant boolean')
    indicator_val = get_overload_const_bool(indicator)
    if not is_overload_bool(_bodo_na_equal):
        raise_bodo_error(
            'DataFrame.merge(): bodo extension _bodo_na_equal must be a constant boolean'
            )
    azz__gai = get_overload_const_bool(_bodo_na_equal)
    validate_keys_length(left_index, right_index, left_keys, right_keys)
    validate_keys_dtypes(left, right, left_index, right_index, left_keys,
        right_keys)
    if is_overload_constant_tuple(suffixes):
        jme__izm = get_overload_const_tuple(suffixes)
    if is_overload_constant_list(suffixes):
        jme__izm = list(get_overload_const_list(suffixes))
    suffix_x = jme__izm[0]
    suffix_y = jme__izm[1]
    validate_unicity_output_column_names(suffix_x, suffix_y, left_keys,
        right_keys, left.columns, right.columns, indicator_val)
    left_keys = gen_const_tup(left_keys)
    right_keys = gen_const_tup(right_keys)
    visq__yelo = "def _impl(left, right, how='inner', on=None, left_on=None,\n"
    visq__yelo += (
        '    right_on=None, left_index=False, right_index=False, sort=False,\n'
        )
    visq__yelo += """    suffixes=('_x', '_y'), copy=True, indicator=False, validate=None, _bodo_na_equal=True):
"""
    visq__yelo += (
        """  return bodo.hiframes.pd_dataframe_ext.join_dummy(left, right, {}, {}, '{}', '{}', '{}', False, {}, {}, '{}')
"""
        .format(left_keys, right_keys, how, suffix_x, suffix_y,
        indicator_val, azz__gai, fmp__shc))
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo}, zrt__acbb)
    _impl = zrt__acbb['_impl']
    return _impl


def common_validate_merge_merge_asof_spec(name_func, left, right, on,
    left_on, right_on, left_index, right_index, suffixes):
    if not isinstance(left, DataFrameType) or not isinstance(right,
        DataFrameType):
        raise BodoError(name_func + '() requires dataframe inputs')
    valid_dataframe_column_types = (ArrayItemArrayType, MapArrayType,
        StructArrayType, CategoricalArrayType, types.Array,
        IntegerArrayType, DecimalArrayType, IntervalArrayType)
    wst__ekg = {string_array_type, binary_array_type,
        datetime_date_array_type, datetime_timedelta_array_type, boolean_array}
    nsfn__sgh = {get_overload_const_str(jbjfj__dfee) for jbjfj__dfee in (
        left_on, right_on, on) if is_overload_constant_str(jbjfj__dfee)}
    for df in (left, right):
        for i, xeydf__mlim in enumerate(df.data):
            if not isinstance(xeydf__mlim, valid_dataframe_column_types
                ) and xeydf__mlim not in wst__ekg:
                raise BodoError(
                    f'{name_func}(): use of column with {type(xeydf__mlim)} in merge unsupported'
                    )
            if df.columns[i] in nsfn__sgh and isinstance(xeydf__mlim,
                MapArrayType):
                raise BodoError(
                    f'{name_func}(): merge on MapArrayType unsupported')
    ensure_constant_arg(name_func, 'left_index', left_index, bool)
    ensure_constant_arg(name_func, 'right_index', right_index, bool)
    if not is_overload_constant_tuple(suffixes
        ) and not is_overload_constant_list(suffixes):
        raise_const_error(name_func +
            "(): suffixes parameters should be ['_left', '_right']")
    if is_overload_constant_tuple(suffixes):
        jme__izm = get_overload_const_tuple(suffixes)
    if is_overload_constant_list(suffixes):
        jme__izm = list(get_overload_const_list(suffixes))
    if len(jme__izm) != 2:
        raise BodoError(name_func +
            '(): The number of suffixes should be exactly 2')
    hskt__dnn = tuple(set(left.columns) & set(right.columns))
    if not is_overload_none(on):
        dbgi__jemr = False
        if is_overload_constant_str(on):
            on_str = get_overload_const_str(on)
            dbgi__jemr = on_str not in hskt__dnn and ('left.' in on_str or 
                'right.' in on_str)
        if len(hskt__dnn) == 0 and not dbgi__jemr:
            raise_bodo_error(name_func +
                '(): No common columns to perform merge on. Merge options: left_on={lon}, right_on={ron}, left_index={lidx}, right_index={ridx}'
                .format(lon=is_overload_true(left_on), ron=is_overload_true
                (right_on), lidx=is_overload_true(left_index), ridx=
                is_overload_true(right_index)))
        if not is_overload_none(left_on) or not is_overload_none(right_on):
            raise BodoError(name_func +
                '(): Can only pass argument "on" OR "left_on" and "right_on", not a combination of both.'
                )
    if (is_overload_true(left_index) or not is_overload_none(left_on)
        ) and is_overload_none(right_on) and not is_overload_true(right_index):
        raise BodoError(name_func +
            '(): Must pass right_on or right_index=True')
    if (is_overload_true(right_index) or not is_overload_none(right_on)
        ) and is_overload_none(left_on) and not is_overload_true(left_index):
        raise BodoError(name_func + '(): Must pass left_on or left_index=True')


def validate_merge_spec(left, right, how, on, left_on, right_on, left_index,
    right_index, sort, suffixes, copy, indicator, validate):
    common_validate_merge_merge_asof_spec('merge', left, right, on, left_on,
        right_on, left_index, right_index, suffixes)
    ensure_constant_values('merge', 'how', how, ('left', 'right', 'outer',
        'inner'))


def validate_merge_asof_spec(left, right, on, left_on, right_on, left_index,
    right_index, by, left_by, right_by, suffixes, tolerance,
    allow_exact_matches, direction):
    common_validate_merge_merge_asof_spec('merge_asof', left, right, on,
        left_on, right_on, left_index, right_index, suffixes)
    if not is_overload_true(allow_exact_matches):
        raise BodoError(
            'merge_asof(): allow_exact_matches parameter only supports default value True'
            )
    if not is_overload_none(tolerance):
        raise BodoError(
            'merge_asof(): tolerance parameter only supports default value None'
            )
    if not is_overload_none(by):
        raise BodoError(
            'merge_asof(): by parameter only supports default value None')
    if not is_overload_none(left_by):
        raise BodoError(
            'merge_asof(): left_by parameter only supports default value None')
    if not is_overload_none(right_by):
        raise BodoError(
            'merge_asof(): right_by parameter only supports default value None'
            )
    if not is_overload_constant_str(direction):
        raise BodoError(
            'merge_asof(): direction parameter should be of type str')
    else:
        direction = get_overload_const_str(direction)
        if direction != 'backward':
            raise BodoError(
                "merge_asof(): direction parameter only supports default value 'backward'"
                )


def validate_merge_asof_keys_length(left_on, right_on, left_index,
    right_index, left_keys, right_keys):
    if not is_overload_true(left_index) and not is_overload_true(right_index):
        if len(right_keys) != len(left_keys):
            raise BodoError('merge(): len(right_on) must equal len(left_on)')
    if not is_overload_none(left_on) and is_overload_true(right_index):
        raise BodoError(
            'merge(): right_index = True and specifying left_on is not suppported yet.'
            )
    if not is_overload_none(right_on) and is_overload_true(left_index):
        raise BodoError(
            'merge(): left_index = True and specifying right_on is not suppported yet.'
            )


def validate_keys_length(left_index, right_index, left_keys, right_keys):
    if not is_overload_true(left_index) and not is_overload_true(right_index):
        if len(right_keys) != len(left_keys):
            raise BodoError('merge(): len(right_on) must equal len(left_on)')
    if is_overload_true(right_index):
        if len(left_keys) != 1:
            raise BodoError(
                'merge(): len(left_on) must equal the number of levels in the index of "right", which is 1'
                )
    if is_overload_true(left_index):
        if len(right_keys) != 1:
            raise BodoError(
                'merge(): len(right_on) must equal the number of levels in the index of "left", which is 1'
                )


def validate_keys_dtypes(left, right, left_index, right_index, left_keys,
    right_keys):
    njx__rdf = numba.core.registry.cpu_target.typing_context
    if is_overload_true(left_index) or is_overload_true(right_index):
        if is_overload_true(left_index) and is_overload_true(right_index):
            rewp__tcbod = left.index
            hhuam__ftyx = isinstance(rewp__tcbod, StringIndexType)
            xwlz__rnsh = right.index
            krm__yxe = isinstance(xwlz__rnsh, StringIndexType)
        elif is_overload_true(left_index):
            rewp__tcbod = left.index
            hhuam__ftyx = isinstance(rewp__tcbod, StringIndexType)
            xwlz__rnsh = right.data[right.columns.index(right_keys[0])]
            krm__yxe = xwlz__rnsh.dtype == string_type
        elif is_overload_true(right_index):
            rewp__tcbod = left.data[left.columns.index(left_keys[0])]
            hhuam__ftyx = rewp__tcbod.dtype == string_type
            xwlz__rnsh = right.index
            krm__yxe = isinstance(xwlz__rnsh, StringIndexType)
        if hhuam__ftyx and krm__yxe:
            return
        rewp__tcbod = rewp__tcbod.dtype
        xwlz__rnsh = xwlz__rnsh.dtype
        try:
            ftk__xnjg = njx__rdf.resolve_function_type(operator.eq, (
                rewp__tcbod, xwlz__rnsh), {})
        except:
            raise_bodo_error(
                'merge: You are trying to merge on {lk_dtype} and {rk_dtype} columns. If you wish to proceed you should use pd.concat'
                .format(lk_dtype=rewp__tcbod, rk_dtype=xwlz__rnsh))
    else:
        for wiy__wdnii, dcb__avgm in zip(left_keys, right_keys):
            rewp__tcbod = left.data[left.columns.index(wiy__wdnii)].dtype
            aux__vzitw = left.data[left.columns.index(wiy__wdnii)]
            xwlz__rnsh = right.data[right.columns.index(dcb__avgm)].dtype
            emdsw__fhew = right.data[right.columns.index(dcb__avgm)]
            if aux__vzitw == emdsw__fhew:
                continue
            evthv__oyfcd = (
                'merge: You are trying to merge on column {lk} of {lk_dtype} and column {rk} of {rk_dtype}. If you wish to proceed you should use pd.concat'
                .format(lk=wiy__wdnii, lk_dtype=rewp__tcbod, rk=dcb__avgm,
                rk_dtype=xwlz__rnsh))
            xccw__xxka = rewp__tcbod == string_type
            pqha__pjcq = xwlz__rnsh == string_type
            if xccw__xxka ^ pqha__pjcq:
                raise_bodo_error(evthv__oyfcd)
            try:
                ftk__xnjg = njx__rdf.resolve_function_type(operator.eq, (
                    rewp__tcbod, xwlz__rnsh), {})
            except:
                raise_bodo_error(evthv__oyfcd)


def validate_keys(keys, df):
    sla__ilc = set(keys).difference(set(df.columns))
    if len(sla__ilc) > 0:
        if is_overload_constant_str(df.index.name_typ
            ) and get_overload_const_str(df.index.name_typ) in sla__ilc:
            raise_bodo_error(
                f'merge(): use of index {df.index.name_typ} as key for on/left_on/right_on is unsupported'
                )
        raise_bodo_error(
            f"""merge(): invalid key {sla__ilc} for on/left_on/right_on
merge supports only valid column names {df.columns}"""
            )


@overload_method(DataFrameType, 'join', inline='always', no_unliteral=True)
def overload_dataframe_join(left, other, on=None, how='left', lsuffix='',
    rsuffix='', sort=False):
    check_runtime_cols_unsupported(left, 'DataFrame.join()')
    check_runtime_cols_unsupported(other, 'DataFrame.join()')
    qnjab__qwci = dict(lsuffix=lsuffix, rsuffix=rsuffix)
    klw__cab = dict(lsuffix='', rsuffix='')
    check_unsupported_args('DataFrame.join', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    validate_join_spec(left, other, on, how, lsuffix, rsuffix, sort)
    how = get_overload_const_str(how)
    if not is_overload_none(on):
        left_keys = get_overload_const_list(on)
    else:
        left_keys = ['$_bodo_index_']
    right_keys = ['$_bodo_index_']
    left_keys = gen_const_tup(left_keys)
    right_keys = gen_const_tup(right_keys)
    visq__yelo = "def _impl(left, other, on=None, how='left',\n"
    visq__yelo += "    lsuffix='', rsuffix='', sort=False):\n"
    visq__yelo += (
        """  return bodo.hiframes.pd_dataframe_ext.join_dummy(left, other, {}, {}, '{}', '{}', '{}', True, False, True, '')
"""
        .format(left_keys, right_keys, how, lsuffix, rsuffix))
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo}, zrt__acbb)
    _impl = zrt__acbb['_impl']
    return _impl


def validate_join_spec(left, other, on, how, lsuffix, rsuffix, sort):
    if not isinstance(other, DataFrameType):
        raise BodoError('join() requires dataframe inputs')
    ensure_constant_values('merge', 'how', how, ('left', 'right', 'outer',
        'inner'))
    if not is_overload_none(on) and len(get_overload_const_list(on)) != 1:
        raise BodoError('join(): len(on) must equals to 1 when specified.')
    if not is_overload_none(on):
        qubj__arngr = get_overload_const_list(on)
        validate_keys(qubj__arngr, left)
    if not is_overload_false(sort):
        raise BodoError(
            'join(): sort parameter only supports default value False')
    hskt__dnn = tuple(set(left.columns) & set(other.columns))
    if len(hskt__dnn) > 0:
        raise_bodo_error(
            'join(): not supporting joining on overlapping columns:{cols} Use DataFrame.merge() instead.'
            .format(cols=hskt__dnn))


def validate_unicity_output_column_names(suffix_x, suffix_y, left_keys,
    right_keys, left_columns, right_columns, indicator_val):
    skarb__uvsko = set(left_keys) & set(right_keys)
    llrsu__lbo = set(left_columns) & set(right_columns)
    ftg__hzkl = llrsu__lbo - skarb__uvsko
    pwmt__mops = set(left_columns) - llrsu__lbo
    sir__patlt = set(right_columns) - llrsu__lbo
    wvc__gtzh = {}

    def insertOutColumn(col_name):
        if col_name in wvc__gtzh:
            raise_bodo_error(
                'join(): two columns happen to have the same name : {}'.
                format(col_name))
        wvc__gtzh[col_name] = 0
    for hgylw__cnoy in skarb__uvsko:
        insertOutColumn(hgylw__cnoy)
    for hgylw__cnoy in ftg__hzkl:
        pcmpr__zwarg = str(hgylw__cnoy) + suffix_x
        lrbwc__sedtk = str(hgylw__cnoy) + suffix_y
        insertOutColumn(pcmpr__zwarg)
        insertOutColumn(lrbwc__sedtk)
    for hgylw__cnoy in pwmt__mops:
        insertOutColumn(hgylw__cnoy)
    for hgylw__cnoy in sir__patlt:
        insertOutColumn(hgylw__cnoy)
    if indicator_val:
        insertOutColumn('_merge')


@overload(pd.merge_asof, inline='always', no_unliteral=True)
def overload_dataframe_merge_asof(left, right, on=None, left_on=None,
    right_on=None, left_index=False, right_index=False, by=None, left_by=
    None, right_by=None, suffixes=('_x', '_y'), tolerance=None,
    allow_exact_matches=True, direction='backward'):
    validate_merge_asof_spec(left, right, on, left_on, right_on, left_index,
        right_index, by, left_by, right_by, suffixes, tolerance,
        allow_exact_matches, direction)
    if not isinstance(left, DataFrameType) or not isinstance(right,
        DataFrameType):
        raise BodoError('merge_asof() requires dataframe inputs')
    hskt__dnn = tuple(sorted(set(left.columns) & set(right.columns), key=lambda
        k: str(k)))
    if not is_overload_none(on):
        left_on = right_on = on
    if is_overload_none(on) and is_overload_none(left_on) and is_overload_none(
        right_on) and is_overload_false(left_index) and is_overload_false(
        right_index):
        left_keys = hskt__dnn
        right_keys = hskt__dnn
    else:
        if is_overload_true(left_index):
            left_keys = ['$_bodo_index_']
        else:
            left_keys = get_overload_const_list(left_on)
            validate_keys(left_keys, left)
        if is_overload_true(right_index):
            right_keys = ['$_bodo_index_']
        else:
            right_keys = get_overload_const_list(right_on)
            validate_keys(right_keys, right)
    validate_merge_asof_keys_length(left_on, right_on, left_index,
        right_index, left_keys, right_keys)
    validate_keys_dtypes(left, right, left_index, right_index, left_keys,
        right_keys)
    left_keys = gen_const_tup(left_keys)
    right_keys = gen_const_tup(right_keys)
    if isinstance(suffixes, tuple):
        jme__izm = suffixes
    if is_overload_constant_list(suffixes):
        jme__izm = list(get_overload_const_list(suffixes))
    if isinstance(suffixes, types.Omitted):
        jme__izm = suffixes.value
    suffix_x = jme__izm[0]
    suffix_y = jme__izm[1]
    visq__yelo = (
        'def _impl(left, right, on=None, left_on=None, right_on=None,\n')
    visq__yelo += (
        '    left_index=False, right_index=False, by=None, left_by=None,\n')
    visq__yelo += "    right_by=None, suffixes=('_x', '_y'), tolerance=None,\n"
    visq__yelo += "    allow_exact_matches=True, direction='backward'):\n"
    visq__yelo += '  suffix_x = suffixes[0]\n'
    visq__yelo += '  suffix_y = suffixes[1]\n'
    visq__yelo += (
        """  return bodo.hiframes.pd_dataframe_ext.join_dummy(left, right, {}, {}, 'asof', '{}', '{}', False, False, True, '')
"""
        .format(left_keys, right_keys, suffix_x, suffix_y))
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo}, zrt__acbb)
    _impl = zrt__acbb['_impl']
    return _impl


@overload_method(DataFrameType, 'groupby', inline='always', no_unliteral=True)
def overload_dataframe_groupby(df, by=None, axis=0, level=None, as_index=
    True, sort=False, group_keys=True, squeeze=False, observed=True, dropna
    =True):
    check_runtime_cols_unsupported(df, 'DataFrame.groupby()')
    validate_groupby_spec(df, by, axis, level, as_index, sort, group_keys,
        squeeze, observed, dropna)

    def _impl(df, by=None, axis=0, level=None, as_index=True, sort=False,
        group_keys=True, squeeze=False, observed=True, dropna=True):
        return bodo.hiframes.pd_groupby_ext.init_groupby(df, by, as_index,
            dropna)
    return _impl


def validate_groupby_spec(df, by, axis, level, as_index, sort, group_keys,
    squeeze, observed, dropna):
    if is_overload_none(by):
        raise BodoError("groupby(): 'by' must be supplied.")
    if not is_overload_zero(axis):
        raise BodoError(
            "groupby(): 'axis' parameter only supports integer value 0.")
    if not is_overload_none(level):
        raise BodoError(
            "groupby(): 'level' is not supported since MultiIndex is not supported."
            )
    if not is_literal_type(by) and not is_overload_constant_list(by):
        raise_const_error(
            f"groupby(): 'by' parameter only supports a constant column label or column labels, not {by}."
            )
    if len(set(get_overload_const_list(by)).difference(set(df.columns))) > 0:
        raise_const_error(
            "groupby(): invalid key {} for 'by' (not available in columns {})."
            .format(get_overload_const_list(by), df.columns))
    if not is_overload_constant_bool(as_index):
        raise_const_error(
            "groupby(): 'as_index' parameter must be a constant bool, not {}."
            .format(as_index))
    if not is_overload_constant_bool(dropna):
        raise_const_error(
            "groupby(): 'dropna' parameter must be a constant bool, not {}."
            .format(dropna))
    qnjab__qwci = dict(sort=sort, group_keys=group_keys, squeeze=squeeze,
        observed=observed)
    otb__lukh = dict(sort=False, group_keys=True, squeeze=False, observed=True)
    check_unsupported_args('Dataframe.groupby', qnjab__qwci, otb__lukh,
        package_name='pandas', module_name='GroupBy')


def pivot_error_checking(df, index, columns, values, func_name):
    mbwq__jna = func_name == 'DataFrame.pivot_table'
    if mbwq__jna:
        if is_overload_none(index) or not is_literal_type(index):
            raise BodoError(
                f"DataFrame.pivot_table(): 'index' argument is required and must be constant column labels"
                )
    elif not is_overload_none(index) and not is_literal_type(index):
        raise BodoError(
            f"{func_name}(): if 'index' argument is provided it must be constant column labels"
            )
    if is_overload_none(columns) or not is_literal_type(columns):
        raise BodoError(
            f"{func_name}(): 'columns' argument is required and must be a constant column label"
            )
    if not is_overload_none(values) and not is_literal_type(values):
        raise BodoError(
            f"{func_name}(): if 'values' argument is provided it must be constant column labels"
            )
    fxvc__arrkx = get_literal_value(columns)
    if isinstance(fxvc__arrkx, (list, tuple)):
        if len(fxvc__arrkx) > 1:
            raise BodoError(
                f"{func_name}(): 'columns' argument must be a constant column label not a {fxvc__arrkx}"
                )
        fxvc__arrkx = fxvc__arrkx[0]
    if fxvc__arrkx not in df.columns:
        raise BodoError(
            f"{func_name}(): 'columns' column {fxvc__arrkx} not found in DataFrame {df}."
            )
    rwpwh__vpm = {ghpd__ovmzd: i for i, ghpd__ovmzd in enumerate(df.columns)}
    eqbj__yubb = rwpwh__vpm[fxvc__arrkx]
    if is_overload_none(index):
        vgx__jhaj = []
        ukknq__kwn = []
    else:
        ukknq__kwn = get_literal_value(index)
        if not isinstance(ukknq__kwn, (list, tuple)):
            ukknq__kwn = [ukknq__kwn]
        vgx__jhaj = []
        for index in ukknq__kwn:
            if index not in rwpwh__vpm:
                raise BodoError(
                    f"{func_name}(): 'index' column {index} not found in DataFrame {df}."
                    )
            vgx__jhaj.append(rwpwh__vpm[index])
    if not (all(isinstance(ghpd__ovmzd, int) for ghpd__ovmzd in ukknq__kwn) or
        all(isinstance(ghpd__ovmzd, str) for ghpd__ovmzd in ukknq__kwn)):
        raise BodoError(
            f"{func_name}(): column names selected for 'index' must all share a common int or string type. Please convert your names to a common type using DataFrame.rename()"
            )
    if is_overload_none(values):
        awdbw__upo = []
        gzdfi__jrs = []
        rxyk__xcuqm = vgx__jhaj + [eqbj__yubb]
        for i, ghpd__ovmzd in enumerate(df.columns):
            if i not in rxyk__xcuqm:
                awdbw__upo.append(i)
                gzdfi__jrs.append(ghpd__ovmzd)
    else:
        gzdfi__jrs = get_literal_value(values)
        if not isinstance(gzdfi__jrs, (list, tuple)):
            gzdfi__jrs = [gzdfi__jrs]
        awdbw__upo = []
        for val in gzdfi__jrs:
            if val not in rwpwh__vpm:
                raise BodoError(
                    f"{func_name}(): 'values' column {val} not found in DataFrame {df}."
                    )
            awdbw__upo.append(rwpwh__vpm[val])
    if all(isinstance(ghpd__ovmzd, int) for ghpd__ovmzd in gzdfi__jrs):
        gzdfi__jrs = np.array(gzdfi__jrs, 'int64')
    elif all(isinstance(ghpd__ovmzd, str) for ghpd__ovmzd in gzdfi__jrs):
        gzdfi__jrs = pd.array(gzdfi__jrs, 'string')
    else:
        raise BodoError(
            f"{func_name}(): column names selected for 'values' must all share a common int or string type. Please convert your names to a common type using DataFrame.rename()"
            )
    xddcy__hhwk = set(awdbw__upo) | set(vgx__jhaj) | {eqbj__yubb}
    if len(xddcy__hhwk) != len(awdbw__upo) + len(vgx__jhaj) + 1:
        raise BodoError(
            f"{func_name}(): 'index', 'columns', and 'values' must all refer to different columns"
            )

    def check_valid_index_typ(index_column):
        if isinstance(index_column, (bodo.ArrayItemArrayType, bodo.
            MapArrayType, bodo.StructArrayType, bodo.TupleArrayType, bodo.
            IntervalArrayType)):
            raise BodoError(
                f"{func_name}(): 'index' DataFrame column must have scalar rows"
                )
        if isinstance(index_column, bodo.CategoricalArrayType):
            raise BodoError(
                f"{func_name}(): 'index' DataFrame column does not support categorical data"
                )
    if len(vgx__jhaj) == 0:
        index = df.index
        if isinstance(index, MultiIndexType):
            raise BodoError(
                f"{func_name}(): 'index' cannot be None with a DataFrame with a multi-index"
                )
        if not isinstance(index, RangeIndexType):
            check_valid_index_typ(index.data)
        if not is_literal_type(df.index.name_typ):
            raise BodoError(
                f"{func_name}(): If 'index' is None, the name of the DataFrame's Index must be constant at compile-time"
                )
    else:
        for liqa__loxp in vgx__jhaj:
            index_column = df.data[liqa__loxp]
            check_valid_index_typ(index_column)
    yzc__rpqo = df.data[eqbj__yubb]
    if isinstance(yzc__rpqo, (bodo.ArrayItemArrayType, bodo.MapArrayType,
        bodo.StructArrayType, bodo.TupleArrayType, bodo.IntervalArrayType)):
        raise BodoError(
            f"{func_name}(): 'columns' DataFrame column must have scalar rows")
    if isinstance(yzc__rpqo, bodo.CategoricalArrayType):
        raise BodoError(
            f"{func_name}(): 'columns' DataFrame column does not support categorical data"
            )
    for dynbw__hsq in awdbw__upo:
        bczab__djc = df.data[dynbw__hsq]
        if isinstance(bczab__djc, (bodo.ArrayItemArrayType, bodo.
            MapArrayType, bodo.StructArrayType, bodo.TupleArrayType)
            ) or bczab__djc == bodo.binary_array_type:
            raise BodoError(
                f"{func_name}(): 'values' DataFrame column must have scalar rows"
                )
    return (ukknq__kwn, fxvc__arrkx, gzdfi__jrs, vgx__jhaj, eqbj__yubb,
        awdbw__upo)


@overload(pd.pivot, inline='always', no_unliteral=True)
@overload_method(DataFrameType, 'pivot', inline='always', no_unliteral=True)
def overload_dataframe_pivot(data, index=None, columns=None, values=None):
    check_runtime_cols_unsupported(data, 'DataFrame.pivot()')
    if not isinstance(data, DataFrameType):
        raise BodoError("pandas.pivot(): 'data' argument must be a DataFrame")
    (ukknq__kwn, fxvc__arrkx, gzdfi__jrs, liqa__loxp, eqbj__yubb, ezeib__wlmhg
        ) = (pivot_error_checking(data, index, columns, values,
        'DataFrame.pivot'))
    if len(ukknq__kwn) == 0:
        if is_overload_none(data.index.name_typ):
            ukknq__kwn = [None]
        else:
            ukknq__kwn = [get_literal_value(data.index.name_typ)]
    if len(gzdfi__jrs) == 1:
        tri__alinv = None
    else:
        tri__alinv = gzdfi__jrs
    visq__yelo = 'def impl(data, index=None, columns=None, values=None):\n'
    visq__yelo += f'    pivot_values = data.iloc[:, {eqbj__yubb}].unique()\n'
    visq__yelo += '    return bodo.hiframes.pd_dataframe_ext.pivot_impl(\n'
    if len(liqa__loxp) == 0:
        visq__yelo += f"""        (bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(data)),),
"""
    else:
        visq__yelo += '        (\n'
        for oyw__hgrz in liqa__loxp:
            visq__yelo += f"""            bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {oyw__hgrz}),
"""
        visq__yelo += '        ),\n'
    visq__yelo += f"""        (bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {eqbj__yubb}),),
"""
    visq__yelo += '        (\n'
    for dynbw__hsq in ezeib__wlmhg:
        visq__yelo += f"""            bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {dynbw__hsq}),
"""
    visq__yelo += '        ),\n'
    visq__yelo += '        pivot_values,\n'
    visq__yelo += '        index_lit_tup,\n'
    visq__yelo += '        columns_lit,\n'
    visq__yelo += '        values_name_const,\n'
    visq__yelo += '    )\n'
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'index_lit_tup': tuple(ukknq__kwn),
        'columns_lit': fxvc__arrkx, 'values_name_const': tri__alinv}, zrt__acbb
        )
    impl = zrt__acbb['impl']
    return impl


@overload(pd.pivot_table, inline='always', no_unliteral=True)
@overload_method(DataFrameType, 'pivot_table', inline='always',
    no_unliteral=True)
def overload_dataframe_pivot_table(data, values=None, index=None, columns=
    None, aggfunc='mean', fill_value=None, margins=False, dropna=True,
    margins_name='All', observed=False, sort=True, _pivot_values=None):
    check_runtime_cols_unsupported(data, 'DataFrame.pivot_table()')
    qnjab__qwci = dict(fill_value=fill_value, margins=margins, dropna=
        dropna, margins_name=margins_name, observed=observed, sort=sort)
    klw__cab = dict(fill_value=None, margins=False, dropna=True,
        margins_name='All', observed=False, sort=True)
    check_unsupported_args('DataFrame.pivot_table', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    if not isinstance(data, DataFrameType):
        raise BodoError(
            "pandas.pivot_table(): 'data' argument must be a DataFrame")
    if _pivot_values is None:
        (ukknq__kwn, fxvc__arrkx, gzdfi__jrs, liqa__loxp, eqbj__yubb,
            ezeib__wlmhg) = (pivot_error_checking(data, index, columns,
            values, 'DataFrame.pivot_table'))
        if len(gzdfi__jrs) == 1:
            tri__alinv = None
        else:
            tri__alinv = gzdfi__jrs
        visq__yelo = 'def impl(\n'
        visq__yelo += '    data,\n'
        visq__yelo += '    values=None,\n'
        visq__yelo += '    index=None,\n'
        visq__yelo += '    columns=None,\n'
        visq__yelo += '    aggfunc="mean",\n'
        visq__yelo += '    fill_value=None,\n'
        visq__yelo += '    margins=False,\n'
        visq__yelo += '    dropna=True,\n'
        visq__yelo += '    margins_name="All",\n'
        visq__yelo += '    observed=False,\n'
        visq__yelo += '    sort=True,\n'
        visq__yelo += '    _pivot_values=None,\n'
        visq__yelo += '):\n'
        kvlxu__aku = liqa__loxp + [eqbj__yubb] + ezeib__wlmhg
        visq__yelo += f'    data = data.iloc[:, {kvlxu__aku}]\n'
        qfi__kvxad = ukknq__kwn + [fxvc__arrkx]
        visq__yelo += (
            f'    data = data.groupby({qfi__kvxad!r}, as_index=False).agg(aggfunc)\n'
            )
        visq__yelo += (
            f'    pivot_values = data.iloc[:, {len(liqa__loxp)}].unique()\n')
        visq__yelo += '    return bodo.hiframes.pd_dataframe_ext.pivot_impl(\n'
        visq__yelo += '        (\n'
        for i in range(0, len(liqa__loxp)):
            visq__yelo += f"""            bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {i}),
"""
        visq__yelo += '        ),\n'
        visq__yelo += f"""        (bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {len(liqa__loxp)}),),
"""
        visq__yelo += '        (\n'
        for i in range(len(liqa__loxp) + 1, len(ezeib__wlmhg) + len(
            liqa__loxp) + 1):
            visq__yelo += f"""            bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {i}),
"""
        visq__yelo += '        ),\n'
        visq__yelo += '        pivot_values,\n'
        visq__yelo += '        index_lit_tup,\n'
        visq__yelo += '        columns_lit,\n'
        visq__yelo += '        values_name_const,\n'
        visq__yelo += '        check_duplicates=False,\n'
        visq__yelo += '    )\n'
        zrt__acbb = {}
        exec(visq__yelo, {'bodo': bodo, 'numba': numba, 'index_lit_tup':
            tuple(ukknq__kwn), 'columns_lit': fxvc__arrkx,
            'values_name_const': tri__alinv}, zrt__acbb)
        impl = zrt__acbb['impl']
        return impl
    if aggfunc == 'mean':

        def _impl(data, values=None, index=None, columns=None, aggfunc=
            'mean', fill_value=None, margins=False, dropna=True,
            margins_name='All', observed=False, sort=True, _pivot_values=None):
            return bodo.hiframes.pd_groupby_ext.pivot_table_dummy(data,
                values, index, columns, 'mean', _pivot_values)
        return _impl

    def _impl(data, values=None, index=None, columns=None, aggfunc='mean',
        fill_value=None, margins=False, dropna=True, margins_name='All',
        observed=False, sort=True, _pivot_values=None):
        return bodo.hiframes.pd_groupby_ext.pivot_table_dummy(data, values,
            index, columns, aggfunc, _pivot_values)
    return _impl


@overload(pd.crosstab, inline='always', no_unliteral=True)
def crosstab_overload(index, columns, values=None, rownames=None, colnames=
    None, aggfunc=None, margins=False, margins_name='All', dropna=True,
    normalize=False, _pivot_values=None):
    qnjab__qwci = dict(values=values, rownames=rownames, colnames=colnames,
        aggfunc=aggfunc, margins=margins, margins_name=margins_name, dropna
        =dropna, normalize=normalize)
    klw__cab = dict(values=None, rownames=None, colnames=None, aggfunc=None,
        margins=False, margins_name='All', dropna=True, normalize=False)
    check_unsupported_args('pandas.crosstab', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    if not isinstance(index, SeriesType):
        raise BodoError(
            f"pandas.crosstab(): 'index' argument only supported for Series types, found {index}"
            )
    if not isinstance(columns, SeriesType):
        raise BodoError(
            f"pandas.crosstab(): 'columns' argument only supported for Series types, found {columns}"
            )

    def _impl(index, columns, values=None, rownames=None, colnames=None,
        aggfunc=None, margins=False, margins_name='All', dropna=True,
        normalize=False, _pivot_values=None):
        return bodo.hiframes.pd_groupby_ext.crosstab_dummy(index, columns,
            _pivot_values)
    return _impl


@overload_method(DataFrameType, 'sort_values', inline='always',
    no_unliteral=True)
def overload_dataframe_sort_values(df, by, axis=0, ascending=True, inplace=
    False, kind='quicksort', na_position='last', ignore_index=False, key=
    None, _bodo_transformed=False):
    check_runtime_cols_unsupported(df, 'DataFrame.sort_values()')
    qnjab__qwci = dict(ignore_index=ignore_index, key=key)
    klw__cab = dict(ignore_index=False, key=None)
    check_unsupported_args('DataFrame.sort_values', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    handle_inplace_df_type_change(inplace, _bodo_transformed, 'sort_values')
    validate_sort_values_spec(df, by, axis, ascending, inplace, kind,
        na_position)

    def _impl(df, by, axis=0, ascending=True, inplace=False, kind=
        'quicksort', na_position='last', ignore_index=False, key=None,
        _bodo_transformed=False):
        return bodo.hiframes.pd_dataframe_ext.sort_values_dummy(df, by,
            ascending, inplace, na_position)
    return _impl


def validate_sort_values_spec(df, by, axis, ascending, inplace, kind,
    na_position):
    if is_overload_none(by) or not is_literal_type(by
        ) and not is_overload_constant_list(by):
        raise_const_error(
            "sort_values(): 'by' parameter only supports a constant column label or column labels. by={}"
            .format(by))
    hzvb__ybt = set(df.columns)
    if is_overload_constant_str(df.index.name_typ):
        hzvb__ybt.add(get_overload_const_str(df.index.name_typ))
    if is_overload_constant_tuple(by):
        aplp__tmp = [get_overload_const_tuple(by)]
    else:
        aplp__tmp = get_overload_const_list(by)
    aplp__tmp = set((k, '') if (k, '') in hzvb__ybt else k for k in aplp__tmp)
    if len(aplp__tmp.difference(hzvb__ybt)) > 0:
        jhg__niu = list(set(get_overload_const_list(by)).difference(hzvb__ybt))
        raise_bodo_error(f'sort_values(): invalid keys {jhg__niu} for by.')
    if not is_overload_zero(axis):
        raise_bodo_error(
            "sort_values(): 'axis' parameter only supports integer value 0.")
    if not is_overload_bool(ascending) and not is_overload_bool_list(ascending
        ):
        raise_bodo_error(
            "sort_values(): 'ascending' parameter must be of type bool or list of bool, not {}."
            .format(ascending))
    if not is_overload_bool(inplace):
        raise_bodo_error(
            "sort_values(): 'inplace' parameter must be of type bool, not {}."
            .format(inplace))
    if kind != 'quicksort' and not isinstance(kind, types.Omitted):
        warnings.warn(BodoWarning(
            'sort_values(): specifying sorting algorithm is not supported in Bodo. Bodo uses stable sort.'
            ))
    if is_overload_constant_str(na_position):
        na_position = get_overload_const_str(na_position)
        if na_position not in ('first', 'last'):
            raise BodoError(
                "sort_values(): na_position should either be 'first' or 'last'"
                )
    elif is_overload_constant_list(na_position):
        gco__qpzby = get_overload_const_list(na_position)
        for na_position in gco__qpzby:
            if na_position not in ('first', 'last'):
                raise BodoError(
                    "sort_values(): Every value in na_position should either be 'first' or 'last'"
                    )
    else:
        raise_const_error(
            f'sort_values(): na_position parameter must be a literal constant of type str or a constant list of str with 1 entry per key column, not {na_position}'
            )
    na_position = get_overload_const_str(na_position)
    if na_position not in ['first', 'last']:
        raise BodoError(
            "sort_values(): na_position should either be 'first' or 'last'")


@overload_method(DataFrameType, 'sort_index', inline='always', no_unliteral
    =True)
def overload_dataframe_sort_index(df, axis=0, level=None, ascending=True,
    inplace=False, kind='quicksort', na_position='last', sort_remaining=
    True, ignore_index=False, key=None):
    check_runtime_cols_unsupported(df, 'DataFrame.sort_index()')
    qnjab__qwci = dict(axis=axis, level=level, kind=kind, sort_remaining=
        sort_remaining, ignore_index=ignore_index, key=key)
    klw__cab = dict(axis=0, level=None, kind='quicksort', sort_remaining=
        True, ignore_index=False, key=None)
    check_unsupported_args('DataFrame.sort_index', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    if not is_overload_bool(ascending):
        raise BodoError(
            "DataFrame.sort_index(): 'ascending' parameter must be of type bool"
            )
    if not is_overload_bool(inplace):
        raise BodoError(
            "DataFrame.sort_index(): 'inplace' parameter must be of type bool")
    if not is_overload_constant_str(na_position) or get_overload_const_str(
        na_position) not in ('first', 'last'):
        raise_bodo_error(
            "DataFrame.sort_index(): 'na_position' should either be 'first' or 'last'"
            )

    def _impl(df, axis=0, level=None, ascending=True, inplace=False, kind=
        'quicksort', na_position='last', sort_remaining=True, ignore_index=
        False, key=None):
        return bodo.hiframes.pd_dataframe_ext.sort_values_dummy(df,
            '$_bodo_index_', ascending, inplace, na_position)
    return _impl


@overload_method(DataFrameType, 'fillna', inline='always', no_unliteral=True)
def overload_dataframe_fillna(df, value=None, method=None, axis=None,
    inplace=False, limit=None, downcast=None):
    check_runtime_cols_unsupported(df, 'DataFrame.fillna()')
    qnjab__qwci = dict(limit=limit, downcast=downcast)
    klw__cab = dict(limit=None, downcast=None)
    check_unsupported_args('DataFrame.fillna', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise BodoError("DataFrame.fillna(): 'axis' argument not supported.")
    glgz__uwz = not is_overload_none(value)
    npbaj__wfauw = not is_overload_none(method)
    if glgz__uwz and npbaj__wfauw:
        raise BodoError(
            "DataFrame.fillna(): Cannot specify both 'value' and 'method'.")
    if not glgz__uwz and not npbaj__wfauw:
        raise BodoError(
            "DataFrame.fillna(): Must specify one of 'value' and 'method'.")
    if glgz__uwz:
        ytttk__fnk = 'value=value'
    else:
        ytttk__fnk = 'method=method'
    data_args = [(
        f"df['{ghpd__ovmzd}'].fillna({ytttk__fnk}, inplace=inplace)" if
        isinstance(ghpd__ovmzd, str) else
        f'df[{ghpd__ovmzd}].fillna({ytttk__fnk}, inplace=inplace)') for
        ghpd__ovmzd in df.columns]
    visq__yelo = """def impl(df, value=None, method=None, axis=None, inplace=False, limit=None, downcast=None):
"""
    if is_overload_true(inplace):
        visq__yelo += '  ' + '  \n'.join(data_args) + '\n'
        zrt__acbb = {}
        exec(visq__yelo, {}, zrt__acbb)
        impl = zrt__acbb['impl']
        return impl
    else:
        return _gen_init_df(visq__yelo, df.columns, ', '.join(numuq__qib +
            '.values' for numuq__qib in data_args))


@overload_method(DataFrameType, 'reset_index', inline='always',
    no_unliteral=True)
def overload_dataframe_reset_index(df, level=None, drop=False, inplace=
    False, col_level=0, col_fill='', _bodo_transformed=False):
    check_runtime_cols_unsupported(df, 'DataFrame.reset_index()')
    qnjab__qwci = dict(col_level=col_level, col_fill=col_fill)
    klw__cab = dict(col_level=0, col_fill='')
    check_unsupported_args('DataFrame.reset_index', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    handle_inplace_df_type_change(inplace, _bodo_transformed, 'reset_index')
    if not _is_all_levels(df, level):
        raise_bodo_error(
            'DataFrame.reset_index(): only dropping all index levels supported'
            )
    if not is_overload_constant_bool(drop):
        raise BodoError(
            "DataFrame.reset_index(): 'drop' parameter should be a constant boolean value"
            )
    if not is_overload_constant_bool(inplace):
        raise BodoError(
            "DataFrame.reset_index(): 'inplace' parameter should be a constant boolean value"
            )
    visq__yelo = """def impl(df, level=None, drop=False, inplace=False, col_level=0, col_fill='', _bodo_transformed=False,):
"""
    visq__yelo += (
        '  index = bodo.hiframes.pd_index_ext.init_range_index(0, len(df), 1, None)\n'
        )
    drop = is_overload_true(drop)
    inplace = is_overload_true(inplace)
    columns = df.columns
    data_args = [
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}){}\n'.
        format(i, '' if inplace else '.copy()') for i in range(len(df.columns))
        ]
    if not drop:
        uhs__ctcog = 'index' if 'index' not in columns else 'level_0'
        index_names = get_index_names(df.index, 'DataFrame.reset_index()',
            uhs__ctcog)
        columns = index_names + columns
        if isinstance(df.index, MultiIndexType):
            visq__yelo += (
                '  m_index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)\n'
                )
            fqsra__yyfl = ['m_index._data[{}]'.format(i) for i in range(df.
                index.nlevels)]
            data_args = fqsra__yyfl + data_args
        else:
            uyc__yrfx = (
                'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df))'
                )
            data_args = [uyc__yrfx] + data_args
    return _gen_init_df(visq__yelo, columns, ', '.join(data_args), 'index')


def _is_all_levels(df, level):
    yxlq__owm = len(get_index_data_arr_types(df.index))
    return is_overload_none(level) or is_overload_constant_int(level
        ) and get_overload_const_int(level
        ) == 0 and yxlq__owm == 1 or is_overload_constant_list(level) and list(
        get_overload_const_list(level)) == list(range(yxlq__owm))


@overload_method(DataFrameType, 'dropna', inline='always', no_unliteral=True)
def overload_dataframe_dropna(df, axis=0, how='any', thresh=None, subset=
    None, inplace=False):
    check_runtime_cols_unsupported(df, 'DataFrame.dropna()')
    if not is_overload_constant_bool(inplace) or is_overload_true(inplace):
        raise BodoError('DataFrame.dropna(): inplace=True is not supported')
    if not is_overload_zero(axis):
        raise_bodo_error(f'df.dropna(): only axis=0 supported')
    ensure_constant_values('dropna', 'how', how, ('any', 'all'))
    if is_overload_none(subset):
        zxroc__rtff = list(range(len(df.columns)))
    elif not is_overload_constant_list(subset):
        raise_bodo_error(
            f'df.dropna(): subset argument should a constant list, not {subset}'
            )
    else:
        asvqk__ocgt = get_overload_const_list(subset)
        zxroc__rtff = []
        for zez__xltop in asvqk__ocgt:
            if zez__xltop not in df.columns:
                raise_bodo_error(
                    f"df.dropna(): column '{zez__xltop}' not in data frame columns {df}"
                    )
            zxroc__rtff.append(df.columns.index(zez__xltop))
    potqq__ublqz = len(df.columns)
    data_args = ', '.join('data_{}'.format(i) for i in range(potqq__ublqz))
    visq__yelo = (
        "def impl(df, axis=0, how='any', thresh=None, subset=None, inplace=False):\n"
        )
    for i in range(potqq__ublqz):
        visq__yelo += (
            '  data_{0} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0})\n'
            .format(i))
    index = (
        'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df))'
        )
    visq__yelo += (
        """  ({0}, index_arr) = bodo.libs.array_kernels.dropna(({0}, {1}), how, thresh, ({2},))
"""
        .format(data_args, index, ', '.join(str(a) for a in zxroc__rtff)))
    visq__yelo += (
        '  index = bodo.utils.conversion.index_from_array(index_arr)\n')
    return _gen_init_df(visq__yelo, df.columns, data_args, 'index')


@overload_method(DataFrameType, 'drop', inline='always', no_unliteral=True)
def overload_dataframe_drop(df, labels=None, axis=0, index=None, columns=
    None, level=None, inplace=False, errors='raise', _bodo_transformed=False):
    check_runtime_cols_unsupported(df, 'DataFrame.drop()')
    qnjab__qwci = dict(index=index, level=level, errors=errors)
    klw__cab = dict(index=None, level=None, errors='raise')
    check_unsupported_args('DataFrame.drop', qnjab__qwci, klw__cab,
        package_name='pandas', module_name='DataFrame')
    handle_inplace_df_type_change(inplace, _bodo_transformed, 'drop')
    if not is_overload_constant_bool(inplace):
        raise_bodo_error(
            "DataFrame.drop(): 'inplace' parameter should be a constant bool")
    if not is_overload_none(labels):
        if not is_overload_none(columns):
            raise BodoError(
                "Dataframe.drop(): Cannot specify both 'labels' and 'columns'")
        if not is_overload_constant_int(axis) or get_overload_const_int(axis
            ) != 1:
            raise_bodo_error('DataFrame.drop(): only axis=1 supported')
        if is_overload_constant_str(labels):
            esoi__tdupl = get_overload_const_str(labels),
        elif is_overload_constant_list(labels):
            esoi__tdupl = get_overload_const_list(labels)
        else:
            raise_bodo_error(
                'constant list of columns expected for labels in DataFrame.drop()'
                )
    else:
        if is_overload_none(columns):
            raise BodoError(
                "DataFrame.drop(): Need to specify at least one of 'labels' or 'columns'"
                )
        if is_overload_constant_str(columns):
            esoi__tdupl = get_overload_const_str(columns),
        elif is_overload_constant_list(columns):
            esoi__tdupl = get_overload_const_list(columns)
        else:
            raise_bodo_error(
                'constant list of columns expected for labels in DataFrame.drop()'
                )
    for ghpd__ovmzd in esoi__tdupl:
        if ghpd__ovmzd not in df.columns:
            raise_bodo_error(
                'DataFrame.drop(): column {} not in DataFrame columns {}'.
                format(ghpd__ovmzd, df.columns))
    if len(set(esoi__tdupl)) == len(df.columns):
        raise BodoError('DataFrame.drop(): Dropping all columns not supported.'
            )
    inplace = is_overload_true(inplace)
    gippv__froes = tuple(ghpd__ovmzd for ghpd__ovmzd in df.columns if 
        ghpd__ovmzd not in esoi__tdupl)
    data_args = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}){}'.
        format(df.columns.index(ghpd__ovmzd), '.copy()' if not inplace else
        '') for ghpd__ovmzd in gippv__froes)
    visq__yelo = (
        'def impl(df, labels=None, axis=0, index=None, columns=None,\n')
    visq__yelo += (
        "     level=None, inplace=False, errors='raise', _bodo_transformed=False):\n"
        )
    index = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
    return _gen_init_df(visq__yelo, gippv__froes, data_args, index)


@overload_method(DataFrameType, 'append', inline='always', no_unliteral=True)
def overload_dataframe_append(df, other, ignore_index=False,
    verify_integrity=False, sort=None):
    check_runtime_cols_unsupported(df, 'DataFrame.append()')
    check_runtime_cols_unsupported(other, 'DataFrame.append()')
    if isinstance(other, DataFrameType):
        return (lambda df, other, ignore_index=False, verify_integrity=
            False, sort=None: pd.concat((df, other), ignore_index=
            ignore_index, verify_integrity=verify_integrity))
    if isinstance(other, types.BaseTuple):
        return (lambda df, other, ignore_index=False, verify_integrity=
            False, sort=None: pd.concat((df,) + other, ignore_index=
            ignore_index, verify_integrity=verify_integrity))
    if isinstance(other, types.List) and isinstance(other.dtype, DataFrameType
        ):
        return (lambda df, other, ignore_index=False, verify_integrity=
            False, sort=None: pd.concat([df] + other, ignore_index=
            ignore_index, verify_integrity=verify_integrity))
    raise BodoError(
        'invalid df.append() input. Only dataframe and list/tuple of dataframes supported'
        )


@overload_method(DataFrameType, 'sample', inline='always', no_unliteral=True)
def overload_dataframe_sample(df, n=None, frac=None, replace=False, weights
    =None, random_state=None, axis=None, ignore_index=False):
    check_runtime_cols_unsupported(df, 'DataFrame.sample()')
    qnjab__qwci = dict(random_state=random_state, weights=weights, axis=
        axis, ignore_index=ignore_index)
    vntj__zbmwm = dict(random_state=None, weights=None, axis=None,
        ignore_index=False)
    check_unsupported_args('DataFrame.sample', qnjab__qwci, vntj__zbmwm,
        package_name='pandas', module_name='DataFrame')
    if not is_overload_none(n) and not is_overload_none(frac):
        raise BodoError(
            'DataFrame.sample(): only one of n and frac option can be selected'
            )
    potqq__ublqz = len(df.columns)
    data_args = ', '.join('data_{}'.format(i) for i in range(potqq__ublqz))
    visq__yelo = """def impl(df, n=None, frac=None, replace=False, weights=None, random_state=None, axis=None, ignore_index=False):
"""
    for i in range(potqq__ublqz):
        visq__yelo += (
            '  data_{0} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0})\n'
            .format(i))
    visq__yelo += '  if frac is None:\n'
    visq__yelo += '    frac_d = -1.0\n'
    visq__yelo += '  else:\n'
    visq__yelo += '    frac_d = frac\n'
    visq__yelo += '  if n is None:\n'
    visq__yelo += '    n_i = 0\n'
    visq__yelo += '  else:\n'
    visq__yelo += '    n_i = n\n'
    index = (
        'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df))'
        )
    visq__yelo += (
        """  ({0},), index_arr = bodo.libs.array_kernels.sample_table_operation(({0},), {1}, n_i, frac_d, replace)
"""
        .format(data_args, index))
    visq__yelo += (
        '  index = bodo.utils.conversion.index_from_array(index_arr)\n')
    return bodo.hiframes.dataframe_impl._gen_init_df(visq__yelo, df.columns,
        data_args, 'index')


@numba.njit
def _sizeof_fmt(num, size_qualifier=''):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return f'{num:3.1f}{size_qualifier} {x}'
        num /= 1024.0
    return f'{num:3.1f}{size_qualifier} PB'


@overload_method(DataFrameType, 'info', no_unliteral=True)
def overload_dataframe_info(df, verbose=None, buf=None, max_cols=None,
    memory_usage=None, show_counts=None, null_counts=None):
    check_runtime_cols_unsupported(df, 'DataFrame.info()')
    hwwj__nvjxf = {'verbose': verbose, 'buf': buf, 'max_cols': max_cols,
        'memory_usage': memory_usage, 'show_counts': show_counts,
        'null_counts': null_counts}
    hriw__fson = {'verbose': None, 'buf': None, 'max_cols': None,
        'memory_usage': None, 'show_counts': None, 'null_counts': None}
    check_unsupported_args('DataFrame.info', hwwj__nvjxf, hriw__fson,
        package_name='pandas', module_name='DataFrame')
    cmu__vks = f"<class '{str(type(df)).split('.')[-1]}"
    if len(df.columns) == 0:

        def _info_impl(df, verbose=None, buf=None, max_cols=None,
            memory_usage=None, show_counts=None, null_counts=None):
            rblat__kxh = cmu__vks + '\n'
            rblat__kxh += 'Index: 0 entries\n'
            rblat__kxh += 'Empty DataFrame'
            print(rblat__kxh)
        return _info_impl
    else:
        visq__yelo = """def _info_impl(df, verbose=None, buf=None, max_cols=None, memory_usage=None, show_counts=None, null_counts=None): #pragma: no cover
"""
        visq__yelo += '    ncols = df.shape[1]\n'
        visq__yelo += f'    lines = "{cmu__vks}\\n"\n'
        visq__yelo += f'    lines += "{df.index}: "\n'
        visq__yelo += (
            '    index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)\n'
            )
        if isinstance(df.index, bodo.hiframes.pd_index_ext.RangeIndexType):
            visq__yelo += """    lines += f"{len(index)} entries, {index.start} to {index.stop-1}\\n\"
"""
        elif isinstance(df.index, bodo.hiframes.pd_index_ext.StringIndexType):
            visq__yelo += """    lines += f"{len(index)} entries, {index[0]} to {index[len(index)-1]}\\n\"
"""
        else:
            visq__yelo += (
                '    lines += f"{len(index)} entries, {index[0]} to {index[-1]}\\n"\n'
                )
        visq__yelo += (
            '    lines += f"Data columns (total {ncols} columns):\\n"\n')
        visq__yelo += (
            f'    space = {max(len(str(k)) for k in df.columns) + 1}\n')
        visq__yelo += '    column_width = max(space, 7)\n'
        visq__yelo += '    column= "Column"\n'
        visq__yelo += '    underl= "------"\n'
        visq__yelo += (
            '    lines += f"#   {column:<{column_width}} Non-Null Count  Dtype\\n"\n'
            )
        visq__yelo += (
            '    lines += f"--- {underl:<{column_width}} --------------  -----\\n"\n'
            )
        visq__yelo += '    mem_size = 0\n'
        visq__yelo += (
            '    col_name = bodo.libs.str_arr_ext.pre_alloc_string_array(ncols, -1)\n'
            )
        visq__yelo += """    non_null_count = bodo.libs.str_arr_ext.pre_alloc_string_array(ncols, -1)
"""
        visq__yelo += (
            '    col_dtype = bodo.libs.str_arr_ext.pre_alloc_string_array(ncols, -1)\n'
            )
        zeyf__lqay = dict()
        for i in range(len(df.columns)):
            visq__yelo += f"""    non_null_count[{i}] = str(bodo.libs.array_ops.array_op_count(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i})))
"""
            kxzil__fkd = f'{df.data[i].dtype}'
            if isinstance(df.data[i], bodo.CategoricalArrayType):
                kxzil__fkd = 'category'
            elif isinstance(df.data[i], bodo.IntegerArrayType):
                tupz__mvep = bodo.libs.int_arr_ext.IntDtype(df.data[i].dtype
                    ).name
                kxzil__fkd = f'{tupz__mvep[:-7]}'
            visq__yelo += f'    col_dtype[{i}] = "{kxzil__fkd}"\n'
            if kxzil__fkd in zeyf__lqay:
                zeyf__lqay[kxzil__fkd] += 1
            else:
                zeyf__lqay[kxzil__fkd] = 1
            visq__yelo += f'    col_name[{i}] = "{df.columns[i]}"\n'
            visq__yelo += f"""    mem_size += bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}).nbytes
"""
        visq__yelo += """    column_info = [f'{i:^3} {name:<{column_width}} {count} non-null      {dtype}' for i, (name, count, dtype) in enumerate(zip(col_name, non_null_count, col_dtype))]
"""
        visq__yelo += '    for i in column_info:\n'
        visq__yelo += "        lines += f'{i}\\n'\n"
        qfw__kvhl = ', '.join(f'{k}({zeyf__lqay[k]})' for k in sorted(
            zeyf__lqay))
        visq__yelo += f"    lines += 'dtypes: {qfw__kvhl}\\n'\n"
        visq__yelo += '    mem_size += df.index.nbytes\n'
        visq__yelo += '    total_size = _sizeof_fmt(mem_size)\n'
        visq__yelo += "    lines += f'memory usage: {total_size}'\n"
        visq__yelo += '    print(lines)\n'
        zrt__acbb = {}
        exec(visq__yelo, {'_sizeof_fmt': _sizeof_fmt, 'pd': pd, 'bodo':
            bodo, 'np': np}, zrt__acbb)
        _info_impl = zrt__acbb['_info_impl']
        return _info_impl


@overload_method(DataFrameType, 'memory_usage', inline='always',
    no_unliteral=True)
def overload_dataframe_memory_usage(df, index=True, deep=False):
    check_runtime_cols_unsupported(df, 'DataFrame.memory_usage()')
    visq__yelo = 'def impl(df, index=True, deep=False):\n'
    data = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}).nbytes'
         for i in range(len(df.columns)))
    if is_overload_true(index):
        jzb__okg = (
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df).nbytes\n,')
        dbp__ukc = ','.join(f"'{ghpd__ovmzd}'" for ghpd__ovmzd in df.columns)
        arr = f"bodo.utils.conversion.coerce_to_array(('Index',{dbp__ukc}))"
        index = f'bodo.hiframes.pd_index_ext.init_binary_str_index({arr})'
        visq__yelo += f"""  return bodo.hiframes.pd_series_ext.init_series(({jzb__okg}{data}), {index}, None)
"""
    else:
        tbot__julg = ',' if len(df.columns) == 1 else ''
        vxntz__piuya = gen_const_tup(df.columns)
        visq__yelo += f"""  return bodo.hiframes.pd_series_ext.init_series(({data}{tbot__julg}), pd.Index({vxntz__piuya}), None)
"""
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'pd': pd}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@overload(pd.read_excel, no_unliteral=True)
def overload_read_excel(io, sheet_name=0, header=0, names=None, index_col=
    None, usecols=None, squeeze=False, dtype=None, engine=None, converters=
    None, true_values=None, false_values=None, skiprows=None, nrows=None,
    na_values=None, keep_default_na=True, na_filter=True, verbose=False,
    parse_dates=False, date_parser=None, thousands=None, comment=None,
    skipfooter=0, convert_float=True, mangle_dupe_cols=True, _bodo_df_type=None
    ):
    df_type = _bodo_df_type.instance_type
    zjis__zgkbr = 'read_excel_df{}'.format(next_label())
    setattr(types, zjis__zgkbr, df_type)
    cejy__pfbg = False
    if is_overload_constant_list(parse_dates):
        cejy__pfbg = get_overload_const_list(parse_dates)
    xhli__qikj = ', '.join(["'{}':{}".format(cname, _get_pd_dtype_str(t)) for
        cname, t in zip(df_type.columns, df_type.data)])
    visq__yelo = (
        """
def impl(
    io,
    sheet_name=0,
    header=0,
    names=None,
    index_col=None,
    usecols=None,
    squeeze=False,
    dtype=None,
    engine=None,
    converters=None,
    true_values=None,
    false_values=None,
    skiprows=None,
    nrows=None,
    na_values=None,
    keep_default_na=True,
    na_filter=True,
    verbose=False,
    parse_dates=False,
    date_parser=None,
    thousands=None,
    comment=None,
    skipfooter=0,
    convert_float=True,
    mangle_dupe_cols=True,
    _bodo_df_type=None,
):
    with numba.objmode(df="{}"):
        df = pd.read_excel(
            io,
            sheet_name,
            header,
            {},
            index_col,
            usecols,
            squeeze,
            {{{}}},
            engine,
            converters,
            true_values,
            false_values,
            skiprows,
            nrows,
            na_values,
            keep_default_na,
            na_filter,
            verbose,
            {},
            date_parser,
            thousands,
            comment,
            skipfooter,
            convert_float,
            mangle_dupe_cols,
        )
    return df
    """
        .format(zjis__zgkbr, list(df_type.columns), xhli__qikj, cejy__pfbg))
    zrt__acbb = {}
    exec(visq__yelo, globals(), zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


def overload_dataframe_plot(df, x=None, y=None, kind='line', figsize=None,
    xlabel=None, ylabel=None, title=None, legend=True, fontsize=None,
    xticks=None, yticks=None, ax=None):
    try:
        import matplotlib.pyplot as plt
    except ImportError as chldp__cjrq:
        raise BodoError('df.plot needs matplotllib which is not installed.')
    visq__yelo = (
        "def impl(df, x=None, y=None, kind='line', figsize=None, xlabel=None, \n"
        )
    visq__yelo += '    ylabel=None, title=None, legend=True, fontsize=None, \n'
    visq__yelo += '    xticks=None, yticks=None, ax=None):\n'
    if is_overload_none(ax):
        visq__yelo += '   fig, ax = plt.subplots()\n'
    else:
        visq__yelo += '   fig = ax.get_figure()\n'
    if not is_overload_none(figsize):
        visq__yelo += '   fig.set_figwidth(figsize[0])\n'
        visq__yelo += '   fig.set_figheight(figsize[1])\n'
    if is_overload_none(xlabel):
        visq__yelo += '   xlabel = x\n'
    visq__yelo += '   ax.set_xlabel(xlabel)\n'
    if is_overload_none(ylabel):
        visq__yelo += '   ylabel = y\n'
    else:
        visq__yelo += '   ax.set_ylabel(ylabel)\n'
    if not is_overload_none(title):
        visq__yelo += '   ax.set_title(title)\n'
    if not is_overload_none(fontsize):
        visq__yelo += '   ax.tick_params(labelsize=fontsize)\n'
    kind = get_overload_const_str(kind)
    if kind == 'line':
        if is_overload_none(x) and is_overload_none(y):
            for i in range(len(df.columns)):
                if isinstance(df.data[i], (types.Array, IntegerArrayType)
                    ) and isinstance(df.data[i].dtype, (types.Integer,
                    types.Float)):
                    visq__yelo += (
                        f'   ax.plot(df.iloc[:, {i}], label=df.columns[{i}])\n'
                        )
        elif is_overload_none(x):
            visq__yelo += '   ax.plot(df[y], label=y)\n'
        elif is_overload_none(y):
            jmjem__axdk = get_overload_const_str(x)
            foiwz__qyyo = df.columns.index(jmjem__axdk)
            for i in range(len(df.columns)):
                if isinstance(df.data[i], (types.Array, IntegerArrayType)
                    ) and isinstance(df.data[i].dtype, (types.Integer,
                    types.Float)):
                    if foiwz__qyyo != i:
                        visq__yelo += (
                            f'   ax.plot(df[x], df.iloc[:, {i}], label=df.columns[{i}])\n'
                            )
        else:
            visq__yelo += '   ax.plot(df[x], df[y], label=y)\n'
    elif kind == 'scatter':
        legend = False
        visq__yelo += '   ax.scatter(df[x], df[y], s=20)\n'
        visq__yelo += '   ax.set_ylabel(ylabel)\n'
    if not is_overload_none(xticks):
        visq__yelo += '   ax.set_xticks(xticks)\n'
    if not is_overload_none(yticks):
        visq__yelo += '   ax.set_yticks(yticks)\n'
    if is_overload_true(legend):
        visq__yelo += '   ax.legend()\n'
    visq__yelo += '   return ax\n'
    zrt__acbb = {}
    exec(visq__yelo, {'bodo': bodo, 'plt': plt}, zrt__acbb)
    impl = zrt__acbb['impl']
    return impl


@lower_builtin('df.plot', DataFrameType, types.VarArg(types.Any))
def dataframe_plot_low(context, builder, sig, args):
    impl = overload_dataframe_plot(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def is_df_values_numpy_supported_dftyp(df_typ):
    for gurg__ajbbn in df_typ.data:
        if not (isinstance(gurg__ajbbn, IntegerArrayType) or isinstance(
            gurg__ajbbn.dtype, types.Number) or gurg__ajbbn.dtype in (bodo.
            datetime64ns, bodo.timedelta64ns)):
            return False
    return True


def typeref_to_type(v):
    if isinstance(v, types.BaseTuple):
        return types.BaseTuple.from_types(tuple(typeref_to_type(a) for a in v))
    return v.instance_type if isinstance(v, (types.TypeRef, types.NumberClass)
        ) else v


def _install_typer_for_type(type_name, typ):

    @type_callable(typ)
    def type_call_type(context):

        def typer(*args, **kws):
            args = tuple(typeref_to_type(v) for v in args)
            kws = {name: typeref_to_type(v) for name, v in kws.items()}
            return types.TypeRef(typ(*args, **kws))
        return typer
    no_side_effect_call_tuples.add((type_name, bodo))
    no_side_effect_call_tuples.add((typ,))


def _install_type_call_typers():
    for type_name in bodo_types_with_params:
        typ = getattr(bodo, type_name)
        _install_typer_for_type(type_name, typ)


_install_type_call_typers()


def set_df_col(df, cname, arr, inplace):
    df[cname] = arr


@infer_global(set_df_col)
class SetDfColInfer(AbstractTemplate):

    def generic(self, args, kws):
        from bodo.hiframes.pd_dataframe_ext import DataFrameType
        assert not kws
        assert len(args) == 4
        assert isinstance(args[1], types.Literal)
        yjgx__hmqv = args[0]
        fmh__ilim = args[1].literal_value
        val = args[2]
        assert val != types.unknown
        ykol__qcqt = yjgx__hmqv
        check_runtime_cols_unsupported(yjgx__hmqv, 'set_df_col()')
        if isinstance(yjgx__hmqv, DataFrameType):
            index = yjgx__hmqv.index
            if len(yjgx__hmqv.columns) == 0:
                index = bodo.hiframes.pd_index_ext.RangeIndexType(types.none)
            if isinstance(val, SeriesType):
                if len(yjgx__hmqv.columns) == 0:
                    index = val.index
                val = val.data
            if is_pd_index_type(val):
                val = bodo.utils.typing.get_index_data_arr_types(val)[0]
            if isinstance(val, types.List):
                val = dtype_to_array_type(val.dtype)
            if not is_array_typ(val):
                val = dtype_to_array_type(val)
            if fmh__ilim in yjgx__hmqv.columns:
                gippv__froes = yjgx__hmqv.columns
                qadp__bgvfj = yjgx__hmqv.columns.index(fmh__ilim)
                adpfg__couwj = list(yjgx__hmqv.data)
                adpfg__couwj[qadp__bgvfj] = val
                adpfg__couwj = tuple(adpfg__couwj)
            else:
                gippv__froes = yjgx__hmqv.columns + (fmh__ilim,)
                adpfg__couwj = yjgx__hmqv.data + (val,)
            ykol__qcqt = DataFrameType(adpfg__couwj, index, gippv__froes,
                yjgx__hmqv.dist, yjgx__hmqv.is_table_format)
        return ykol__qcqt(*args)


SetDfColInfer.prefer_literal = True


def _parse_query_expr(expr, env, columns, cleaned_columns, index_name=None,
    join_cleaned_cols=()):
    chh__neib = {}

    def _rewrite_membership_op(self, node, left, right):
        fqcl__sdfyq = node.op
        op = self.visit(fqcl__sdfyq)
        return op, fqcl__sdfyq, left, right

    def _maybe_evaluate_binop(self, op, op_class, lhs, rhs, eval_in_python=
        ('in', 'not in'), maybe_eval_in_python=('==', '!=', '<', '>', '<=',
        '>=')):
        res = op(lhs, rhs)
        return res
    jujqz__fnhc = []


    class NewFuncNode(pd.core.computation.ops.FuncNode):

        def __init__(self, name):
            if (name not in pd.core.computation.ops.MATHOPS or pd.core.
                computation.check._NUMEXPR_INSTALLED and pd.core.
                computation.check_NUMEXPR_VERSION < pd.core.computation.ops
                .LooseVersion('2.6.9') and name in ('floor', 'ceil')):
                if name not in jujqz__fnhc:
                    raise BodoError('"{0}" is not a supported function'.
                        format(name))
            self.name = name
            if name in jujqz__fnhc:
                self.func = name
            else:
                self.func = getattr(np, name)

        def __call__(self, *args):
            return pd.core.computation.ops.MathCall(self, args)

        def __repr__(self):
            return pd.io.formats.printing.pprint_thing(self.name)

    def visit_Attribute(self, node, **kwargs):
        ogwa__ivxe = node.attr
        value = node.value
        gejz__hmmrd = pd.core.computation.ops.LOCAL_TAG
        if ogwa__ivxe in ('str', 'dt'):
            try:
                ynhqn__irwwi = str(self.visit(value))
            except pd.core.computation.ops.UndefinedVariableError as bst__xlr:
                col_name = bst__xlr.args[0].split("'")[1]
                raise BodoError(
                    'df.query(): column {} is not found in dataframe columns {}'
                    .format(col_name, columns))
        else:
            ynhqn__irwwi = str(self.visit(value))
        qzmv__nfxo = ynhqn__irwwi, ogwa__ivxe
        if qzmv__nfxo in join_cleaned_cols:
            ogwa__ivxe = join_cleaned_cols[qzmv__nfxo]
        name = ynhqn__irwwi + '.' + ogwa__ivxe
        if name.startswith(gejz__hmmrd):
            name = name[len(gejz__hmmrd):]
        if ogwa__ivxe in ('str', 'dt'):
            zyjez__tzhth = columns[cleaned_columns.index(ynhqn__irwwi)]
            chh__neib[zyjez__tzhth] = ynhqn__irwwi
            self.env.scope[name] = 0
            return self.term_type(gejz__hmmrd + name, self.env)
        jujqz__fnhc.append(name)
        return NewFuncNode(name)

    def __str__(self):
        if isinstance(self.value, list):
            return '{}'.format(self.value)
        if isinstance(self.value, str):
            return "'{}'".format(self.value)
        return pd.io.formats.printing.pprint_thing(self.name)

    def math__str__(self):
        if self.op in jujqz__fnhc:
            return pd.io.formats.printing.pprint_thing('{0}({1})'.format(
                self.op, ','.join(map(str, self.operands))))
        vbnv__aqdtc = map(lambda a:
            'bodo.hiframes.pd_series_ext.get_series_data({})'.format(str(a)
            ), self.operands)
        op = 'np.{}'.format(self.op)
        fmh__ilim = (
            'bodo.hiframes.pd_index_ext.init_range_index(0, len({}), 1, None)'
            .format(str(self.operands[0])))
        return pd.io.formats.printing.pprint_thing(
            'bodo.hiframes.pd_series_ext.init_series({0}({1}), {2})'.format
            (op, ','.join(vbnv__aqdtc), fmh__ilim))

    def op__str__(self):
        lbnho__gvvy = ('({0})'.format(pd.io.formats.printing.pprint_thing(
            grzy__nmjw)) for grzy__nmjw in self.operands)
        if self.op == 'in':
            return pd.io.formats.printing.pprint_thing(
                'bodo.hiframes.pd_dataframe_ext.val_isin_dummy({})'.format(
                ', '.join(lbnho__gvvy)))
        if self.op == 'not in':
            return pd.io.formats.printing.pprint_thing(
                'bodo.hiframes.pd_dataframe_ext.val_notin_dummy({})'.format
                (', '.join(lbnho__gvvy)))
        return pd.io.formats.printing.pprint_thing(' {0} '.format(self.op).
            join(lbnho__gvvy))
    hywz__pzet = (pd.core.computation.expr.BaseExprVisitor.
        _rewrite_membership_op)
    hxz__lixs = pd.core.computation.expr.BaseExprVisitor._maybe_evaluate_binop
    xhnl__lsk = pd.core.computation.expr.BaseExprVisitor.visit_Attribute
    lho__tiuhg = (pd.core.computation.expr.BaseExprVisitor.
        _maybe_downcast_constants)
    pblz__mthhp = pd.core.computation.ops.Term.__str__
    ovu__ucn = pd.core.computation.ops.MathCall.__str__
    sckfg__bsn = pd.core.computation.ops.Op.__str__
    brvc__lbzm = pd.core.computation.ops.BinOp._disallow_scalar_only_bool_ops
    try:
        pd.core.computation.expr.BaseExprVisitor._rewrite_membership_op = (
            _rewrite_membership_op)
        pd.core.computation.expr.BaseExprVisitor._maybe_evaluate_binop = (
            _maybe_evaluate_binop)
        pd.core.computation.expr.BaseExprVisitor.visit_Attribute = (
            visit_Attribute)
        (pd.core.computation.expr.BaseExprVisitor._maybe_downcast_constants
            ) = lambda self, left, right: (left, right)
        pd.core.computation.ops.Term.__str__ = __str__
        pd.core.computation.ops.MathCall.__str__ = math__str__
        pd.core.computation.ops.Op.__str__ = op__str__
        pd.core.computation.ops.BinOp._disallow_scalar_only_bool_ops = (lambda
            self: None)
        opopz__vhvh = pd.core.computation.expr.Expr(expr, env=env)
        smpoj__iecbq = str(opopz__vhvh)
    except pd.core.computation.ops.UndefinedVariableError as bst__xlr:
        if not is_overload_none(index_name) and get_overload_const_str(
            index_name) == bst__xlr.args[0].split("'")[1]:
            raise BodoError(
                "df.query(): Refering to named index ('{}') by name is not supported"
                .format(get_overload_const_str(index_name)))
        else:
            raise BodoError(f'df.query(): undefined variable, {bst__xlr}')
    finally:
        pd.core.computation.expr.BaseExprVisitor._rewrite_membership_op = (
            hywz__pzet)
        pd.core.computation.expr.BaseExprVisitor._maybe_evaluate_binop = (
            hxz__lixs)
        pd.core.computation.expr.BaseExprVisitor.visit_Attribute = xhnl__lsk
        (pd.core.computation.expr.BaseExprVisitor._maybe_downcast_constants
            ) = lho__tiuhg
        pd.core.computation.ops.Term.__str__ = pblz__mthhp
        pd.core.computation.ops.MathCall.__str__ = ovu__ucn
        pd.core.computation.ops.Op.__str__ = sckfg__bsn
        pd.core.computation.ops.BinOp._disallow_scalar_only_bool_ops = (
            brvc__lbzm)
    zxytq__rzk = pd.core.computation.parsing.clean_column_name
    chh__neib.update({ghpd__ovmzd: zxytq__rzk(ghpd__ovmzd) for ghpd__ovmzd in
        columns if zxytq__rzk(ghpd__ovmzd) in opopz__vhvh.names})
    return opopz__vhvh, smpoj__iecbq, chh__neib


class DataFrameTupleIterator(types.SimpleIteratorType):

    def __init__(self, col_names, arr_typs):
        self.array_types = arr_typs
        self.col_names = col_names
        ozacw__axc = ['{}={}'.format(col_names[i], arr_typs[i]) for i in
            range(len(col_names))]
        name = 'itertuples({})'.format(','.join(ozacw__axc))
        oty__vzxx = namedtuple('Pandas', col_names)
        otzlf__lst = types.NamedTuple([_get_series_dtype(a) for a in
            arr_typs], oty__vzxx)
        super(DataFrameTupleIterator, self).__init__(name, otzlf__lst)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


def _get_series_dtype(arr_typ):
    if arr_typ == types.Array(types.NPDatetime('ns'), 1, 'C'):
        return pd_timestamp_type
    return arr_typ.dtype


def get_itertuples():
    pass


@infer_global(get_itertuples)
class TypeIterTuples(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) % 2 == 0, 'name and column pairs expected'
        col_names = [a.literal_value for a in args[:len(args) // 2]]
        kxro__ygxh = [if_series_to_array_type(a) for a in args[len(args) // 2:]
            ]
        assert 'Index' not in col_names[0]
        col_names = ['Index'] + col_names
        kxro__ygxh = [types.Array(types.int64, 1, 'C')] + kxro__ygxh
        yzsyb__tik = DataFrameTupleIterator(col_names, kxro__ygxh)
        return yzsyb__tik(*args)


TypeIterTuples.prefer_literal = True


@register_model(DataFrameTupleIterator)
class DataFrameTupleIteratorModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        hmwkl__vqt = [('index', types.EphemeralPointer(types.uintp))] + [(
            'array{}'.format(i), arr) for i, arr in enumerate(fe_type.
            array_types[1:])]
        super(DataFrameTupleIteratorModel, self).__init__(dmm, fe_type,
            hmwkl__vqt)

    def from_return(self, builder, value):
        return value


@lower_builtin(get_itertuples, types.VarArg(types.Any))
def get_itertuples_impl(context, builder, sig, args):
    lyu__bjs = args[len(args) // 2:]
    crx__oyx = sig.args[len(sig.args) // 2:]
    usnqs__mmr = context.make_helper(builder, sig.return_type)
    phgv__citr = context.get_constant(types.intp, 0)
    hapiz__bnebt = cgutils.alloca_once_value(builder, phgv__citr)
    usnqs__mmr.index = hapiz__bnebt
    for i, arr in enumerate(lyu__bjs):
        setattr(usnqs__mmr, 'array{}'.format(i), arr)
    for arr, arr_typ in zip(lyu__bjs, crx__oyx):
        context.nrt.incref(builder, arr_typ, arr)
    res = usnqs__mmr._getvalue()
    return impl_ret_new_ref(context, builder, sig.return_type, res)


@lower_builtin('getiter', DataFrameTupleIterator)
def getiter_itertuples(context, builder, sig, args):
    return impl_ret_borrowed(context, builder, sig.return_type, args[0])


@lower_builtin('iternext', DataFrameTupleIterator)
@iternext_impl(RefType.UNTRACKED)
def iternext_itertuples(context, builder, sig, args, result):
    phlsm__jnu, = sig.args
    npz__gbm, = args
    usnqs__mmr = context.make_helper(builder, phlsm__jnu, value=npz__gbm)
    vcpsj__ulmlv = signature(types.intp, phlsm__jnu.array_types[1])
    estci__xqi = context.compile_internal(builder, lambda a: len(a),
        vcpsj__ulmlv, [usnqs__mmr.array0])
    index = builder.load(usnqs__mmr.index)
    acu__umhvr = builder.icmp(lc.ICMP_SLT, index, estci__xqi)
    result.set_valid(acu__umhvr)
    with builder.if_then(acu__umhvr):
        values = [index]
        for i, arr_typ in enumerate(phlsm__jnu.array_types[1:]):
            jcsqg__zmxc = getattr(usnqs__mmr, 'array{}'.format(i))
            if arr_typ == types.Array(types.NPDatetime('ns'), 1, 'C'):
                rzh__ilqo = signature(pd_timestamp_type, arr_typ, types.intp)
                val = context.compile_internal(builder, lambda a, i: bodo.
                    hiframes.pd_timestamp_ext.
                    convert_datetime64_to_timestamp(np.int64(a[i])),
                    rzh__ilqo, [jcsqg__zmxc, index])
            else:
                rzh__ilqo = signature(arr_typ.dtype, arr_typ, types.intp)
                val = context.compile_internal(builder, lambda a, i: a[i],
                    rzh__ilqo, [jcsqg__zmxc, index])
            values.append(val)
        value = context.make_tuple(builder, phlsm__jnu.yield_type, values)
        result.yield_(value)
        ifj__fcpva = cgutils.increment_index(builder, index)
        builder.store(ifj__fcpva, usnqs__mmr.index)


def _analyze_op_pair_first(self, scope, equiv_set, expr, lhs):
    typ = self.typemap[expr.value.name].first_type
    if not isinstance(typ, types.NamedTuple):
        return None
    lhs = ir.Var(scope, mk_unique_var('tuple_var'), expr.loc)
    self.typemap[lhs.name] = typ
    rhs = ir.Expr.pair_first(expr.value, expr.loc)
    brnu__yqu = ir.Assign(rhs, lhs, expr.loc)
    mbdz__xpl = lhs
    vdtv__bcu = []
    twti__rgbjd = []
    brq__dqy = typ.count
    for i in range(brq__dqy):
        tcru__zihw = ir.Var(mbdz__xpl.scope, mk_unique_var('{}_size{}'.
            format(mbdz__xpl.name, i)), mbdz__xpl.loc)
        dzuhh__jsr = ir.Expr.static_getitem(lhs, i, None, mbdz__xpl.loc)
        self.calltypes[dzuhh__jsr] = None
        vdtv__bcu.append(ir.Assign(dzuhh__jsr, tcru__zihw, mbdz__xpl.loc))
        self._define(equiv_set, tcru__zihw, types.intp, dzuhh__jsr)
        twti__rgbjd.append(tcru__zihw)
    ptwas__cqbh = tuple(twti__rgbjd)
    return numba.parfors.array_analysis.ArrayAnalysis.AnalyzeResult(shape=
        ptwas__cqbh, pre=[brnu__yqu] + vdtv__bcu)


numba.parfors.array_analysis.ArrayAnalysis._analyze_op_pair_first = (
    _analyze_op_pair_first)
