"""
Indexing support for pd.DataFrame type.
"""
import operator
import numpy as np
import pandas as pd
from numba.core import cgutils, types
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import intrinsic, lower_builtin, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, register_model
import bodo
from bodo.hiframes.pd_dataframe_ext import DataFrameType, check_runtime_cols_unsupported
from bodo.utils.transform import gen_const_tup
from bodo.utils.typing import BodoError, get_overload_const_int, get_overload_const_list, get_overload_const_str, is_immutable_array, is_list_like_index_type, is_overload_constant_int, is_overload_constant_list, is_overload_constant_str, raise_bodo_error


@infer_global(operator.getitem)
class DataFrameGetItemTemplate(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 2
        check_runtime_cols_unsupported(args[0], 'DataFrame getitem (df[])')
        if isinstance(args[0], DataFrameType):
            return self.typecheck_df_getitem(args)
        elif isinstance(args[0], DataFrameLocType):
            return self.typecheck_loc_getitem(args)
        else:
            return

    def typecheck_loc_getitem(self, args):
        I = args[0]
        idx = args[1]
        df = I.df_type
        if isinstance(df.columns[0], tuple):
            raise_bodo_error(
                'DataFrame.loc[] getitem (location-based indexing) with multi-indexed columns not supported yet'
                )
        if is_list_like_index_type(idx) and idx.dtype == types.bool_:
            kuxb__pokv = idx
            nci__hryj = df.data
            kdrvr__gsrxu = df.columns
            fthq__dqfkm = self.replace_range_with_numeric_idx_if_needed(df,
                kuxb__pokv)
            jhbe__fyvz = DataFrameType(nci__hryj, fthq__dqfkm, kdrvr__gsrxu)
            return jhbe__fyvz(*args)
        if isinstance(idx, types.BaseTuple) and len(idx) == 2:
            nzbgq__cfxhq = idx.types[0]
            khxa__agjqw = idx.types[1]
            if isinstance(nzbgq__cfxhq, types.Integer):
                if not isinstance(df.index, bodo.hiframes.pd_index_ext.
                    RangeIndexType):
                    raise_bodo_error(
                        'Dataframe.loc[int, col_ind] getitem only supported for dataframes with RangeIndexes'
                        )
                if is_overload_constant_str(khxa__agjqw):
                    tqyb__exkm = get_overload_const_str(khxa__agjqw)
                    if tqyb__exkm not in df.columns:
                        raise_bodo_error(
                            'dataframe {} does not include column {}'.
                            format(df, tqyb__exkm))
                    vyeuz__ggp = df.columns.index(tqyb__exkm)
                    return df.data[vyeuz__ggp].dtype(*args)
                if isinstance(khxa__agjqw, types.UnicodeType):
                    raise_bodo_error(
                        f'DataFrame.loc[] getitem (location-based indexing) requires constant column names. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                        )
                else:
                    raise_bodo_error(
                        f'DataFrame.loc[] getitem (location-based indexing) using {idx} not supported yet.'
                        )
            if is_list_like_index_type(nzbgq__cfxhq
                ) and nzbgq__cfxhq.dtype == types.bool_ or isinstance(
                nzbgq__cfxhq, types.SliceType):
                fthq__dqfkm = self.replace_range_with_numeric_idx_if_needed(df,
                    nzbgq__cfxhq)
                if is_overload_constant_str(khxa__agjqw):
                    eyft__uwaz = get_overload_const_str(khxa__agjqw)
                    if eyft__uwaz not in df.columns:
                        raise_bodo_error(
                            f'dataframe {df} does not include column {eyft__uwaz}'
                            )
                    vyeuz__ggp = df.columns.index(eyft__uwaz)
                    btxx__uhg = df.data[vyeuz__ggp]
                    ufnwd__hpo = btxx__uhg.dtype
                    ste__uyl = types.literal(df.columns[vyeuz__ggp])
                    jhbe__fyvz = bodo.SeriesType(ufnwd__hpo, btxx__uhg,
                        fthq__dqfkm, ste__uyl)
                    return jhbe__fyvz(*args)
                if isinstance(khxa__agjqw, types.UnicodeType):
                    raise_bodo_error(
                        f'DataFrame.loc[] getitem (location-based indexing) requires constant column names. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                        )
                elif is_overload_constant_list(khxa__agjqw):
                    dujol__qzwsw = get_overload_const_list(khxa__agjqw)
                    useoe__xbz = types.unliteral(khxa__agjqw)
                    if useoe__xbz.dtype == types.bool_:
                        if len(df.columns) != len(dujol__qzwsw):
                            raise_bodo_error(
                                f'dataframe {df} has {len(df.columns)} columns, but boolean array used with DataFrame.loc[] {dujol__qzwsw} has {len(dujol__qzwsw)} values'
                                )
                        xrmpc__dte = []
                        qbqwc__fdgvv = []
                        for lwbak__wlrjs in range(len(dujol__qzwsw)):
                            if dujol__qzwsw[lwbak__wlrjs]:
                                xrmpc__dte.append(df.columns[lwbak__wlrjs])
                                qbqwc__fdgvv.append(df.data[lwbak__wlrjs])
                        fklbq__unoxw = tuple()
                        jhbe__fyvz = DataFrameType(tuple(qbqwc__fdgvv),
                            fthq__dqfkm, tuple(xrmpc__dte))
                        return jhbe__fyvz(*args)
                    elif useoe__xbz.dtype == bodo.string_type:
                        fklbq__unoxw, qbqwc__fdgvv = (self.
                            get_kept_cols_and_data(df, dujol__qzwsw))
                        jhbe__fyvz = DataFrameType(qbqwc__fdgvv,
                            fthq__dqfkm, fklbq__unoxw)
                        return jhbe__fyvz(*args)
        raise_bodo_error(
            f'DataFrame.loc[] getitem (location-based indexing) using {idx} not supported yet. If you are trying to select a subset of the columns by passing a list of column names, that list must be a compile time constant. See https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
            )

    def typecheck_df_getitem(self, args):
        df = args[0]
        ind = args[1]
        if is_overload_constant_str(ind) or is_overload_constant_int(ind):
            ind_val = get_overload_const_str(ind) if is_overload_constant_str(
                ind) else get_overload_const_int(ind)
            if isinstance(df.columns[0], tuple):
                xrmpc__dte = []
                qbqwc__fdgvv = []
                for lwbak__wlrjs, yqbhn__iok in enumerate(df.columns):
                    if yqbhn__iok[0] != ind_val:
                        continue
                    xrmpc__dte.append(yqbhn__iok[1] if len(yqbhn__iok) == 2
                         else yqbhn__iok[1:])
                    qbqwc__fdgvv.append(df.data[lwbak__wlrjs])
                btxx__uhg = tuple(qbqwc__fdgvv)
                qcux__arx = df.index
                uxd__stezx = tuple(xrmpc__dte)
                jhbe__fyvz = DataFrameType(btxx__uhg, qcux__arx, uxd__stezx)
                return jhbe__fyvz(*args)
            else:
                if ind_val not in df.columns:
                    raise_bodo_error('dataframe {} does not include column {}'
                        .format(df, ind_val))
                vyeuz__ggp = df.columns.index(ind_val)
                btxx__uhg = df.data[vyeuz__ggp]
                ufnwd__hpo = btxx__uhg.dtype
                qcux__arx = df.index
                ste__uyl = types.literal(df.columns[vyeuz__ggp])
                jhbe__fyvz = bodo.SeriesType(ufnwd__hpo, btxx__uhg,
                    qcux__arx, ste__uyl)
                return jhbe__fyvz(*args)
        if isinstance(ind, types.Integer) or isinstance(ind, types.UnicodeType
            ):
            raise_bodo_error(
                'df[] getitem selecting a subset of columns requires providing constant column names. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html'
                )
        if is_list_like_index_type(ind
            ) and ind.dtype == types.bool_ or isinstance(ind, types.SliceType):
            btxx__uhg = df.data
            qcux__arx = self.replace_range_with_numeric_idx_if_needed(df, ind)
            uxd__stezx = df.columns
            jhbe__fyvz = DataFrameType(btxx__uhg, qcux__arx, uxd__stezx,
                is_table_format=df.is_table_format)
            return jhbe__fyvz(*args)
        elif is_overload_constant_list(ind):
            tmeh__plu = get_overload_const_list(ind)
            uxd__stezx, btxx__uhg = self.get_kept_cols_and_data(df, tmeh__plu)
            qcux__arx = df.index
            jhbe__fyvz = DataFrameType(btxx__uhg, qcux__arx, uxd__stezx)
            return jhbe__fyvz(*args)
        raise_bodo_error(
            f'df[] getitem using {ind} not supported. If you are trying to select a subset of the columns, you must provide the column names you are selecting as a constant. See https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html'
            )

    def get_kept_cols_and_data(self, df, cols_to_keep_list):
        for yfull__jcmrb in cols_to_keep_list:
            if yfull__jcmrb not in df.columns:
                raise_bodo_error('Column {} not found in dataframe columns {}'
                    .format(yfull__jcmrb, df.columns))
        uxd__stezx = tuple(cols_to_keep_list)
        btxx__uhg = tuple(df.data[df.columns.index(lod__guo)] for lod__guo in
            uxd__stezx)
        return uxd__stezx, btxx__uhg

    def replace_range_with_numeric_idx_if_needed(self, df, ind):
        fthq__dqfkm = bodo.hiframes.pd_index_ext.NumericIndexType(types.
            int64, df.index.name_typ) if not isinstance(ind, types.SliceType
            ) and isinstance(df.index, bodo.hiframes.pd_index_ext.
            RangeIndexType) else df.index
        return fthq__dqfkm


DataFrameGetItemTemplate._no_unliteral = True


@lower_builtin(operator.getitem, DataFrameType, types.Any)
def getitem_df_lower(context, builder, sig, args):
    impl = df_getitem_overload(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def df_getitem_overload(df, ind):
    if not isinstance(df, DataFrameType):
        return
    if is_overload_constant_str(ind) or is_overload_constant_int(ind):
        ind_val = get_overload_const_str(ind) if is_overload_constant_str(ind
            ) else get_overload_const_int(ind)
        if isinstance(df.columns[0], tuple):
            xrmpc__dte = []
            qbqwc__fdgvv = []
            for lwbak__wlrjs, yqbhn__iok in enumerate(df.columns):
                if yqbhn__iok[0] != ind_val:
                    continue
                xrmpc__dte.append(yqbhn__iok[1] if len(yqbhn__iok) == 2 else
                    yqbhn__iok[1:])
                qbqwc__fdgvv.append(
                    'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})'
                    .format(lwbak__wlrjs))
            oqrh__vzhou = 'def impl(df, ind):\n'
            bhy__xiy = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
            return bodo.hiframes.dataframe_impl._gen_init_df(oqrh__vzhou,
                xrmpc__dte, ', '.join(qbqwc__fdgvv), bhy__xiy)
        if ind_val not in df.columns:
            raise_bodo_error('dataframe {} does not include column {}'.
                format(df, ind_val))
        col_no = df.columns.index(ind_val)
        return lambda df, ind: bodo.hiframes.pd_series_ext.init_series(bodo
            .hiframes.pd_dataframe_ext.get_dataframe_data(df, col_no), bodo
            .hiframes.pd_dataframe_ext.get_dataframe_index(df), ind_val)
    if is_overload_constant_list(ind):
        tmeh__plu = get_overload_const_list(ind)
        for yfull__jcmrb in tmeh__plu:
            if yfull__jcmrb not in df.columns:
                raise_bodo_error('Column {} not found in dataframe columns {}'
                    .format(yfull__jcmrb, df.columns))
        qbqwc__fdgvv = ', '.join(
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}).copy()'
            .format(df.columns.index(yfull__jcmrb)) for yfull__jcmrb in
            tmeh__plu)
        oqrh__vzhou = 'def impl(df, ind):\n'
        bhy__xiy = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)'
        return bodo.hiframes.dataframe_impl._gen_init_df(oqrh__vzhou,
            tmeh__plu, qbqwc__fdgvv, bhy__xiy)
    if is_list_like_index_type(ind) and ind.dtype == types.bool_ or isinstance(
        ind, types.SliceType):
        oqrh__vzhou = 'def impl(df, ind):\n'
        if not isinstance(ind, types.SliceType):
            oqrh__vzhou += (
                '  ind = bodo.utils.conversion.coerce_to_ndarray(ind)\n')
        bhy__xiy = (
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[ind]')
        if df.is_table_format:
            qbqwc__fdgvv = (
                f'bodo.hiframes.pd_dataframe_ext.get_dataframe_table(df)[ind]')
        else:
            qbqwc__fdgvv = ', '.join(
                f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(yfull__jcmrb)})[ind]'
                 for yfull__jcmrb in df.columns)
        return bodo.hiframes.dataframe_impl._gen_init_df(oqrh__vzhou, df.
            columns, qbqwc__fdgvv, bhy__xiy, out_df_type=df)
    raise_bodo_error('df[] getitem using {} not supported'.format(ind))


@overload(operator.setitem, no_unliteral=True)
def df_setitem_overload(df, idx, val):
    check_runtime_cols_unsupported(df, 'DataFrame setitem (df[])')
    if not isinstance(df, DataFrameType):
        return
    raise_bodo_error('DataFrame setitem: transform necessary')


class DataFrameILocType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        lod__guo = 'DataFrameILocType({})'.format(df_type)
        super(DataFrameILocType, self).__init__(lod__guo)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)
    ndim = 2


@register_model(DataFrameILocType)
class DataFrameILocModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        fhx__zgob = [('obj', fe_type.df_type)]
        super(DataFrameILocModel, self).__init__(dmm, fe_type, fhx__zgob)


make_attribute_wrapper(DataFrameILocType, 'obj', '_obj')


@intrinsic
def init_dataframe_iloc(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        czlu__juvg, = args
        gufva__xnee = signature.return_type
        tykp__pqi = cgutils.create_struct_proxy(gufva__xnee)(context, builder)
        tykp__pqi.obj = czlu__juvg
        context.nrt.incref(builder, signature.args[0], czlu__juvg)
        return tykp__pqi._getvalue()
    return DataFrameILocType(obj)(obj), codegen


@overload_attribute(DataFrameType, 'iloc')
def overload_dataframe_iloc(df):
    check_runtime_cols_unsupported(df, 'DataFrame.iloc')
    return lambda df: bodo.hiframes.dataframe_indexing.init_dataframe_iloc(df)


@overload(operator.getitem, no_unliteral=True)
def overload_iloc_getitem(I, idx):
    if not isinstance(I, DataFrameILocType):
        return
    df = I.df_type
    if isinstance(idx, types.Integer):
        return _gen_iloc_getitem_row_impl(df, df.columns, 'idx')
    if isinstance(idx, types.BaseTuple) and len(idx) == 2 and not isinstance(
        idx[1], types.SliceType):
        if not (is_overload_constant_list(idx.types[1]) or
            is_overload_constant_int(idx.types[1])):
            raise_bodo_error(
                'idx2 in df.iloc[idx1, idx2] should be a constant integer or constant list of integers. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                )
        ufqhw__cdx = len(df.data)
        if is_overload_constant_int(idx.types[1]):
            is_out_series = True
            nhjd__xhtix = get_overload_const_int(idx.types[1])
            if nhjd__xhtix < 0 or nhjd__xhtix >= ufqhw__cdx:
                raise BodoError(
                    'df.iloc: column integer must refer to a valid column number'
                    )
            jdmx__khkej = [nhjd__xhtix]
        else:
            is_out_series = False
            jdmx__khkej = get_overload_const_list(idx.types[1])
            if any(not isinstance(ind, int) or ind < 0 or ind >= ufqhw__cdx for
                ind in jdmx__khkej):
                raise BodoError(
                    'df.iloc: column list must be integers referring to a valid column number'
                    )
        col_names = tuple(pd.Series(df.columns, dtype=object)[jdmx__khkej])
        if isinstance(idx.types[0], types.Integer):
            if isinstance(idx.types[1], types.Integer):
                nhjd__xhtix = jdmx__khkej[0]

                def impl(I, idx):
                    df = I._obj
                    return bodo.utils.conversion.box_if_dt64(bodo.hiframes.
                        pd_dataframe_ext.get_dataframe_data(df, nhjd__xhtix
                        )[idx[0]])
                return impl
            return _gen_iloc_getitem_row_impl(df, col_names, 'idx[0]')
        if is_list_like_index_type(idx.types[0]) and isinstance(idx.types[0
            ].dtype, (types.Integer, types.Boolean)) or isinstance(idx.
            types[0], types.SliceType):
            return _gen_iloc_getitem_bool_slice_impl(df, col_names, idx.
                types[0], 'idx[0]', is_out_series)
    if is_list_like_index_type(idx) and isinstance(idx.dtype, (types.
        Integer, types.Boolean)) or isinstance(idx, types.SliceType):
        return _gen_iloc_getitem_bool_slice_impl(df, df.columns, idx, 'idx',
            False)
    if isinstance(idx, types.BaseTuple) and len(idx) == 2 and isinstance(idx
        [0], types.SliceType) and isinstance(idx[1], types.SliceType):
        raise_bodo_error(
            'slice2 in df.iloc[slice1,slice2] should be constant. For more information, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
            )
    raise_bodo_error(f'df.iloc[] getitem using {idx} not supported')


def _gen_iloc_getitem_bool_slice_impl(df, col_names, idx_typ, idx,
    is_out_series):
    oqrh__vzhou = 'def impl(I, idx):\n'
    oqrh__vzhou += '  df = I._obj\n'
    if isinstance(idx_typ, types.SliceType):
        oqrh__vzhou += f'  idx_t = {idx}\n'
    else:
        oqrh__vzhou += (
            f'  idx_t = bodo.utils.conversion.coerce_to_ndarray({idx})\n')
    bhy__xiy = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[idx_t]'
    qbqwc__fdgvv = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(yfull__jcmrb)})[idx_t]'
         for yfull__jcmrb in col_names)
    if is_out_series:
        mxpn__utcs = f"'{col_names[0]}'" if isinstance(col_names[0], str
            ) else f'{col_names[0]}'
        oqrh__vzhou += f"""  return bodo.hiframes.pd_series_ext.init_series({qbqwc__fdgvv}, {bhy__xiy}, {mxpn__utcs})
"""
        iygf__kyepq = {}
        exec(oqrh__vzhou, {'bodo': bodo}, iygf__kyepq)
        return iygf__kyepq['impl']
    return bodo.hiframes.dataframe_impl._gen_init_df(oqrh__vzhou, col_names,
        qbqwc__fdgvv, bhy__xiy)


def _gen_iloc_getitem_row_impl(df, col_names, idx):
    oqrh__vzhou = 'def impl(I, idx):\n'
    oqrh__vzhou += '  df = I._obj\n'
    jrpg__vryek = ', '.join(
        f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {df.columns.index(yfull__jcmrb)})[{idx}]'
         for yfull__jcmrb in col_names)
    oqrh__vzhou += f"""  row_idx = bodo.hiframes.pd_index_ext.init_heter_index({gen_const_tup(col_names)}, None)
"""
    oqrh__vzhou += f"""  return bodo.hiframes.pd_series_ext.init_series(({jrpg__vryek},), row_idx, None)
"""
    iygf__kyepq = {}
    exec(oqrh__vzhou, {'bodo': bodo}, iygf__kyepq)
    impl = iygf__kyepq['impl']
    return impl


@overload(operator.setitem, no_unliteral=True)
def df_iloc_setitem_overload(df, idx, val):
    if not isinstance(df, DataFrameILocType):
        return
    raise_bodo_error(
        f'DataFrame.iloc setitem unsupported for dataframe {df.df_type}, index {idx}, value {val}'
        )


class DataFrameLocType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        lod__guo = 'DataFrameLocType({})'.format(df_type)
        super(DataFrameLocType, self).__init__(lod__guo)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)
    ndim = 2


@register_model(DataFrameLocType)
class DataFrameLocModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        fhx__zgob = [('obj', fe_type.df_type)]
        super(DataFrameLocModel, self).__init__(dmm, fe_type, fhx__zgob)


make_attribute_wrapper(DataFrameLocType, 'obj', '_obj')


@intrinsic
def init_dataframe_loc(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        czlu__juvg, = args
        wap__qlzdn = signature.return_type
        wuoe__fdo = cgutils.create_struct_proxy(wap__qlzdn)(context, builder)
        wuoe__fdo.obj = czlu__juvg
        context.nrt.incref(builder, signature.args[0], czlu__juvg)
        return wuoe__fdo._getvalue()
    return DataFrameLocType(obj)(obj), codegen


@overload_attribute(DataFrameType, 'loc')
def overload_dataframe_loc(df):
    check_runtime_cols_unsupported(df, 'DataFrame.loc')
    return lambda df: bodo.hiframes.dataframe_indexing.init_dataframe_loc(df)


@lower_builtin(operator.getitem, DataFrameLocType, types.Any)
def loc_getitem_lower(context, builder, sig, args):
    impl = overload_loc_getitem(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def overload_loc_getitem(I, idx):
    if not isinstance(I, DataFrameLocType):
        return
    df = I.df_type
    if is_list_like_index_type(idx) and idx.dtype == types.bool_:
        oqrh__vzhou = 'def impl(I, idx):\n'
        oqrh__vzhou += '  df = I._obj\n'
        oqrh__vzhou += (
            '  idx_t = bodo.utils.conversion.coerce_to_ndarray(idx)\n')
        bhy__xiy = (
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[idx_t]')
        qbqwc__fdgvv = ', '.join(
            'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})[idx_t]'
            .format(df.columns.index(yfull__jcmrb)) for yfull__jcmrb in df.
            columns)
        return bodo.hiframes.dataframe_impl._gen_init_df(oqrh__vzhou, df.
            columns, qbqwc__fdgvv, bhy__xiy)
    if isinstance(idx, types.BaseTuple) and len(idx) == 2:
        mdgd__wjyp = idx.types[1]
        if is_overload_constant_str(mdgd__wjyp):
            saz__bcb = get_overload_const_str(mdgd__wjyp)
            nhjd__xhtix = df.columns.index(saz__bcb)

            def impl_col_name(I, idx):
                df = I._obj
                bhy__xiy = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(
                    df)
                lzf__mkm = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df
                    , nhjd__xhtix)
                return bodo.hiframes.pd_series_ext.init_series(lzf__mkm,
                    bhy__xiy, saz__bcb).loc[idx[0]]
            return impl_col_name
        if is_overload_constant_list(mdgd__wjyp):
            col_idx_list = get_overload_const_list(mdgd__wjyp)
            if len(col_idx_list) > 0 and not isinstance(col_idx_list[0], (
                bool, np.bool_)) and not all(yfull__jcmrb in df.columns for
                yfull__jcmrb in col_idx_list):
                raise_bodo_error(
                    f'DataFrame.loc[]: invalid column list {col_idx_list}; not all in dataframe columns {df.columns}'
                    )
            return gen_df_loc_col_select_impl(df, col_idx_list)
    raise_bodo_error(
        f'DataFrame.loc[] getitem (location-based indexing) using {idx} not supported yet.'
        )


def gen_df_loc_col_select_impl(df, col_idx_list):
    if len(col_idx_list) > 0 and isinstance(col_idx_list[0], (bool, np.bool_)):
        col_idx_list = list(pd.Series(df.columns, dtype=object)[col_idx_list])
    qbqwc__fdgvv = ', '.join(
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {})[idx[0]]'
        .format(df.columns.index(yfull__jcmrb)) for yfull__jcmrb in
        col_idx_list)
    bhy__xiy = 'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)[idx[0]]'
    oqrh__vzhou = 'def impl(I, idx):\n'
    oqrh__vzhou += '  df = I._obj\n'
    return bodo.hiframes.dataframe_impl._gen_init_df(oqrh__vzhou,
        col_idx_list, qbqwc__fdgvv, bhy__xiy)


@overload(operator.setitem, no_unliteral=True)
def df_loc_setitem_overload(df, idx, val):
    if not isinstance(df, DataFrameLocType):
        return
    raise_bodo_error(
        f'DataFrame.loc setitem unsupported for dataframe {df.df_type}, index {idx}, value {val}'
        )


class DataFrameIatType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        lod__guo = 'DataFrameIatType({})'.format(df_type)
        super(DataFrameIatType, self).__init__(lod__guo)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)
    ndim = 2


@register_model(DataFrameIatType)
class DataFrameIatModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        fhx__zgob = [('obj', fe_type.df_type)]
        super(DataFrameIatModel, self).__init__(dmm, fe_type, fhx__zgob)


make_attribute_wrapper(DataFrameIatType, 'obj', '_obj')


@intrinsic
def init_dataframe_iat(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        czlu__juvg, = args
        ahc__ushg = signature.return_type
        luam__tbwq = cgutils.create_struct_proxy(ahc__ushg)(context, builder)
        luam__tbwq.obj = czlu__juvg
        context.nrt.incref(builder, signature.args[0], czlu__juvg)
        return luam__tbwq._getvalue()
    return DataFrameIatType(obj)(obj), codegen


@overload_attribute(DataFrameType, 'iat')
def overload_dataframe_iat(df):
    check_runtime_cols_unsupported(df, 'DataFrame.iat')
    return lambda df: bodo.hiframes.dataframe_indexing.init_dataframe_iat(df)


@overload(operator.getitem, no_unliteral=True)
def overload_iat_getitem(I, idx):
    if not isinstance(I, DataFrameIatType):
        return
    if isinstance(idx, types.BaseTuple) and len(idx) == 2:
        if not isinstance(idx.types[0], types.Integer):
            raise BodoError(
                'DataFrame.iat: iAt based indexing can only have integer indexers'
                )
        if not is_overload_constant_int(idx.types[1]):
            raise_bodo_error(
                'DataFrame.iat getitem: column index must be a constant integer. For more informaton, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html.'
                )
        nhjd__xhtix = get_overload_const_int(idx.types[1])

        def impl_col_ind(I, idx):
            df = I._obj
            lzf__mkm = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df,
                nhjd__xhtix)
            return lzf__mkm[idx[0]]
        return impl_col_ind
    raise BodoError('df.iat[] getitem using {} not supported'.format(idx))


@overload(operator.setitem, no_unliteral=True)
def overload_iat_setitem(I, idx, val):
    if not isinstance(I, DataFrameIatType):
        return
    if isinstance(idx, types.BaseTuple) and len(idx) == 2:
        if not isinstance(idx.types[0], types.Integer):
            raise BodoError(
                'DataFrame.iat: iAt based indexing can only have integer indexers'
                )
        if not is_overload_constant_int(idx.types[1]):
            raise_bodo_error(
                'DataFrame.iat setitem: column index must be a constant integer. For more informaton, see https://docs.bodo.ai/latest/source/programming_with_bodo/require_constants.html'
                )
        nhjd__xhtix = get_overload_const_int(idx.types[1])
        if is_immutable_array(I.df_type.data[nhjd__xhtix]):
            raise BodoError(
                f'DataFrame setitem not supported for column with immutable array type {I.df_type.data}'
                )

        def impl_col_ind(I, idx, val):
            df = I._obj
            lzf__mkm = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df,
                nhjd__xhtix)
            lzf__mkm[idx[0]] = bodo.utils.conversion.unbox_if_timestamp(val)
        return impl_col_ind
    raise BodoError('df.iat[] setitem using {} not supported'.format(idx))


@lower_cast(DataFrameIatType, DataFrameIatType)
@lower_cast(DataFrameILocType, DataFrameILocType)
@lower_cast(DataFrameLocType, DataFrameLocType)
def cast_series_iat(context, builder, fromty, toty, val):
    luam__tbwq = cgutils.create_struct_proxy(fromty)(context, builder, val)
    ghbab__lgny = context.cast(builder, luam__tbwq.obj, fromty.df_type,
        toty.df_type)
    hemci__ige = cgutils.create_struct_proxy(toty)(context, builder)
    hemci__ige.obj = ghbab__lgny
    return hemci__ige._getvalue()
