"""
Implement pd.DataFrame typing and data model handling.
"""
import json
import operator
from urllib.parse import urlparse
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed, lower_constant
from numba.core.typing.templates import AbstractTemplate, bound_function, infer_global, signature
from numba.cpython.listobj import ListInstance
from numba.extending import infer_getattr, intrinsic, lower_builtin, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType
from bodo.hiframes.pd_index_ext import HeterogeneousIndexType, NumericIndexType, RangeIndexType, is_pd_index_type
from bodo.hiframes.pd_multi_index_ext import MultiIndexType
from bodo.hiframes.pd_series_ext import HeterogeneousSeriesType, SeriesType
from bodo.hiframes.series_indexing import SeriesIlocType
from bodo.hiframes.table import Table, TableType, get_table_data, set_table_data_codegen
from bodo.io import json_cpp
from bodo.libs.array import arr_info_list_to_table, array_to_info, delete_info_decref_array, delete_table, delete_table_decref_arrays, info_from_table, info_to_array, py_table_to_cpp_table, shuffle_table
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.distributed_api import bcast_scalar
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import str_arr_from_sequence, string_array_type
from bodo.libs.str_ext import string_type, unicode_to_utf8
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.utils.cg_helpers import is_ll_eq
from bodo.utils.conversion import index_to_array
from bodo.utils.templates import OverloadedKeyAttributeTemplate
from bodo.utils.transform import gen_const_tup, get_const_func_output_type, get_const_tup_vals
from bodo.utils.typing import BodoError, BodoWarning, check_unsupported_args, create_unsupported_overload, dtype_to_array_type, get_index_data_arr_types, get_literal_value, get_overload_const, get_overload_const_bool, get_overload_const_int, get_overload_const_list, get_overload_const_str, get_udf_error_msg, get_udf_out_arr_type, is_heterogeneous_tuple_type, is_iterable_type, is_literal_type, is_overload_bool, is_overload_constant_bool, is_overload_constant_int, is_overload_constant_str, is_overload_false, is_overload_int, is_overload_none, is_overload_true, is_tuple_like_type, raise_bodo_error, to_nullable_type
from bodo.utils.utils import is_null_pointer
_json_write = types.ExternalFunction('json_write', types.void(types.voidptr,
    types.voidptr, types.int64, types.int64, types.bool_, types.bool_,
    types.voidptr))
ll.add_symbol('json_write', json_cpp.json_write)


class DataFrameType(types.ArrayCompatible):
    ndim = 2

    def __init__(self, data=None, index=None, columns=None, dist=None,
        is_table_format=False):
        from bodo.transforms.distributed_analysis import Distribution
        self.data = data
        if index is None:
            index = RangeIndexType(types.none)
        self.index = index
        self.columns = columns
        dist = Distribution.OneD_Var if dist is None else dist
        self.dist = dist
        self.is_table_format = is_table_format
        if columns is None:
            assert is_table_format, 'Determining columns at runtime is only supported for DataFrame with table format'
            self.table_type = TableType(tuple(data[:-1]), True)
        else:
            self.table_type = TableType(data) if is_table_format else None
        super(DataFrameType, self).__init__(name=
            f'dataframe({data}, {index}, {columns}, {dist}, {is_table_format})'
            )

    def __str__(self):
        if not self.has_runtime_cols and len(self.columns) > 20:
            sit__ehsbl = f'{len(self.data)} columns of types {set(self.data)}'
            wadu__peeta = (
                f"('{self.columns[0]}', '{self.columns[1]}', ..., '{self.columns[-1]}')"
                )
            return (
                f'dataframe({sit__ehsbl}, {self.index}, {wadu__peeta}, {self.dist}, {self.is_table_format})'
                )
        return super().__str__()

    def copy(self, data=None, index=None, columns=None, dist=None,
        is_table_format=None):
        if data is None:
            data = self.data
        if columns is None:
            columns = self.columns
        if index is None:
            index = self.index
        if dist is None:
            dist = self.dist
        if is_table_format is None:
            is_table_format = self.is_table_format
        return DataFrameType(data, index, columns, dist, is_table_format)

    @property
    def has_runtime_cols(self):
        return self.columns is None

    @property
    def runtime_colname_typ(self):
        return self.data[-1] if self.has_runtime_cols else None

    @property
    def as_array(self):
        return types.Array(types.undefined, 2, 'C')

    @property
    def key(self):
        return (self.data, self.index, self.columns, self.dist, self.
            is_table_format)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)

    def unify(self, typingctx, other):
        from bodo.transforms.distributed_analysis import Distribution
        if (isinstance(other, DataFrameType) and len(other.data) == len(
            self.data) and other.columns == self.columns and other.
            has_runtime_cols == self.has_runtime_cols):
            wqu__bio = (self.index if self.index == other.index else self.
                index.unify(typingctx, other.index))
            data = tuple(cdlfw__yzr.unify(typingctx, dibo__xtm) if 
                cdlfw__yzr != dibo__xtm else cdlfw__yzr for cdlfw__yzr,
                dibo__xtm in zip(self.data, other.data))
            dist = Distribution(min(self.dist.value, other.dist.value))
            if wqu__bio is not None and None not in data:
                return DataFrameType(data, wqu__bio, self.columns, dist,
                    self.is_table_format)
        if isinstance(other, DataFrameType) and len(self.data
            ) == 0 and not self.has_runtime_cols:
            return other

    def can_convert_to(self, typingctx, other):
        from numba.core.typeconv import Conversion
        if (isinstance(other, DataFrameType) and self.data == other.data and
            self.index == other.index and self.columns == other.columns and
            self.dist != other.dist and self.has_runtime_cols == other.
            has_runtime_cols):
            return Conversion.safe

    def is_precise(self):
        return all(cdlfw__yzr.is_precise() for cdlfw__yzr in self.data
            ) and self.index.is_precise()

    def replace_col_type(self, col_name, new_type):
        if col_name not in self.columns:
            raise ValueError(
                f"DataFrameType.replace_col_type replaced column must be found in the DataFrameType. '{col_name}' not found in DataFrameType with columns {self.columns}"
                )
        wkolm__xelf = self.columns.index(col_name)
        digf__ctqv = tuple(list(self.data[:wkolm__xelf]) + [new_type] +
            list(self.data[wkolm__xelf + 1:]))
        return DataFrameType(digf__ctqv, self.index, self.columns, self.
            dist, self.is_table_format)


def check_runtime_cols_unsupported(df, func_name):
    if isinstance(df, DataFrameType) and df.has_runtime_cols:
        raise BodoError(
            f'{func_name} on DataFrames with columns determined at runtime is not yet supported. Please return the DataFrame to regular Python to update typing information.'
            )


class DataFramePayloadType(types.Type):

    def __init__(self, df_type):
        self.df_type = df_type
        super(DataFramePayloadType, self).__init__(name=
            f'DataFramePayloadType({df_type})')

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(DataFramePayloadType)
class DataFramePayloadModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        data_typ = types.Tuple(fe_type.df_type.data)
        if fe_type.df_type.is_table_format:
            data_typ = types.Tuple([fe_type.df_type.table_type])
        jhp__mua = [('data', data_typ), ('index', fe_type.df_type.index), (
            'parent', types.pyobject)]
        if fe_type.df_type.has_runtime_cols:
            jhp__mua.append(('columns', fe_type.df_type.runtime_colname_typ))
        super(DataFramePayloadModel, self).__init__(dmm, fe_type, jhp__mua)


@register_model(DataFrameType)
class DataFrameModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        payload_type = DataFramePayloadType(fe_type)
        jhp__mua = [('meminfo', types.MemInfoPointer(payload_type)), (
            'parent', types.pyobject)]
        super(DataFrameModel, self).__init__(dmm, fe_type, jhp__mua)


make_attribute_wrapper(DataFrameType, 'meminfo', '_meminfo')


@infer_getattr
class DataFrameAttribute(OverloadedKeyAttributeTemplate):
    key = DataFrameType

    def resolve_shape(self, df):
        return types.Tuple([types.int64, types.int64])

    @bound_function('df.head')
    def resolve_head(self, df, args, kws):
        func_name = 'DataFrame.head'
        check_runtime_cols_unsupported(df, f'{func_name}()')
        cjbx__gbht = 'n',
        qrl__wxsx = {'n': 5}
        vqbyt__jdi, gazxi__vwo = bodo.utils.typing.fold_typing_args(func_name,
            args, kws, cjbx__gbht, qrl__wxsx)
        xtyk__zpyb = gazxi__vwo[0]
        if not is_overload_int(xtyk__zpyb):
            raise BodoError(f"{func_name}(): 'n' must be an Integer")
        vslxi__flwc = df.copy(is_table_format=False)
        return vslxi__flwc(*gazxi__vwo).replace(pysig=vqbyt__jdi)

    @bound_function('df.corr')
    def resolve_corr(self, df, args, kws):
        func_name = 'DataFrame.corr'
        check_runtime_cols_unsupported(df, f'{func_name}()')
        hmq__njut = (df,) + args
        cjbx__gbht = 'df', 'method', 'min_periods'
        qrl__wxsx = {'method': 'pearson', 'min_periods': 1}
        rdir__ulbud = 'method',
        vqbyt__jdi, gazxi__vwo = bodo.utils.typing.fold_typing_args(func_name,
            hmq__njut, kws, cjbx__gbht, qrl__wxsx, rdir__ulbud)
        wnugs__nhot = gazxi__vwo[2]
        if not is_overload_int(wnugs__nhot):
            raise BodoError(f"{func_name}(): 'min_periods' must be an Integer")
        iubqc__tdr = []
        qql__ttmcm = []
        for ajkeq__tnmny, xupfv__pngz in zip(df.columns, df.data):
            if bodo.utils.typing._is_pandas_numeric_dtype(xupfv__pngz.dtype):
                iubqc__tdr.append(ajkeq__tnmny)
                qql__ttmcm.append(types.Array(types.float64, 1, 'A'))
        if len(iubqc__tdr) == 0:
            raise_bodo_error('DataFrame.corr(): requires non-empty dataframe')
        qql__ttmcm = tuple(qql__ttmcm)
        iubqc__tdr = tuple(iubqc__tdr)
        index_typ = bodo.utils.typing.type_col_to_index(iubqc__tdr)
        vslxi__flwc = DataFrameType(qql__ttmcm, index_typ, iubqc__tdr)
        return vslxi__flwc(*gazxi__vwo).replace(pysig=vqbyt__jdi)

    @bound_function('df.pipe', no_unliteral=True)
    def resolve_pipe(self, df, args, kws):
        check_runtime_cols_unsupported(df, 'DataFrame.pipe()')
        return bodo.hiframes.pd_groupby_ext.resolve_obj_pipe(self, df, args,
            kws, 'DataFrame')

    @bound_function('df.apply', no_unliteral=True)
    def resolve_apply(self, df, args, kws):
        check_runtime_cols_unsupported(df, 'DataFrame.apply()')
        kws = dict(kws)
        nnknp__xpva = args[0] if len(args) > 0 else kws.pop('func', None)
        axis = args[1] if len(args) > 1 else kws.pop('axis', types.literal(0))
        djf__wdyok = args[2] if len(args) > 2 else kws.pop('raw', types.
            literal(False))
        ekqf__lzd = args[3] if len(args) > 3 else kws.pop('result_type',
            types.none)
        hkiru__egmlu = args[4] if len(args) > 4 else kws.pop('args', types.
            Tuple([]))
        qjbin__nksm = dict(raw=djf__wdyok, result_type=ekqf__lzd)
        hcz__hffs = dict(raw=False, result_type=None)
        check_unsupported_args('Dataframe.apply', qjbin__nksm, hcz__hffs,
            package_name='pandas', module_name='DataFrame')
        rvae__timq = True
        if types.unliteral(nnknp__xpva) == types.unicode_type:
            if not is_overload_constant_str(nnknp__xpva):
                raise BodoError(
                    f'DataFrame.apply(): string argument (for builtins) must be a compile time constant'
                    )
            rvae__timq = False
        if not is_overload_constant_int(axis):
            raise BodoError(
                'Dataframe.apply(): axis argument must be a compile time constant.'
                )
        kxwe__yfjjs = get_overload_const_int(axis)
        if rvae__timq and kxwe__yfjjs != 1:
            raise BodoError(
                'Dataframe.apply(): only axis=1 supported for user-defined functions'
                )
        elif kxwe__yfjjs not in (0, 1):
            raise BodoError('Dataframe.apply(): axis must be either 0 or 1')
        hrl__wmqhd = []
        for arr_typ in df.data:
            zghf__mtvpo = SeriesType(arr_typ.dtype, arr_typ, df.index,
                string_type)
            tfs__ydulf = self.context.resolve_function_type(operator.
                getitem, (SeriesIlocType(zghf__mtvpo), types.int64), {}
                ).return_type
            hrl__wmqhd.append(tfs__ydulf)
        mzcv__skwx = types.none
        tekoc__nwqqd = HeterogeneousIndexType(types.BaseTuple.from_types(
            tuple(types.literal(ajkeq__tnmny) for ajkeq__tnmny in df.
            columns)), None)
        kxutf__xgmaw = types.BaseTuple.from_types(hrl__wmqhd)
        qrr__rjyxk = df.index.dtype
        if qrr__rjyxk == types.NPDatetime('ns'):
            qrr__rjyxk = bodo.pd_timestamp_type
        if qrr__rjyxk == types.NPTimedelta('ns'):
            qrr__rjyxk = bodo.pd_timedelta_type
        if is_heterogeneous_tuple_type(kxutf__xgmaw):
            aes__abvx = HeterogeneousSeriesType(kxutf__xgmaw, tekoc__nwqqd,
                qrr__rjyxk)
        else:
            aes__abvx = SeriesType(kxutf__xgmaw.dtype, kxutf__xgmaw,
                tekoc__nwqqd, qrr__rjyxk)
        ggy__nca = aes__abvx,
        if hkiru__egmlu is not None:
            ggy__nca += tuple(hkiru__egmlu.types)
        try:
            if not rvae__timq:
                one__ahrft = bodo.utils.transform.get_udf_str_return_type(df,
                    get_overload_const_str(nnknp__xpva), self.context,
                    'DataFrame.apply', axis if kxwe__yfjjs == 1 else None)
            else:
                one__ahrft = get_const_func_output_type(nnknp__xpva,
                    ggy__nca, kws, self.context, numba.core.registry.
                    cpu_target.target_context)
        except Exception as cwzji__vns:
            raise_bodo_error(get_udf_error_msg('DataFrame.apply()', cwzji__vns)
                )
        if rvae__timq:
            if not (is_overload_constant_int(axis) and 
                get_overload_const_int(axis) == 1):
                raise BodoError(
                    'Dataframe.apply(): only user-defined functions with axis=1 supported'
                    )
            if isinstance(one__ahrft, (SeriesType, HeterogeneousSeriesType)
                ) and one__ahrft.const_info is None:
                raise BodoError(
                    'Invalid Series output in UDF (Series with constant length and constant Index value expected)'
                    )
            if isinstance(one__ahrft, HeterogeneousSeriesType):
                cjxcc__zecxr, zmzu__dxqgu = one__ahrft.const_info
                cnh__cfcf = tuple(dtype_to_array_type(chpzo__khb) for
                    chpzo__khb in one__ahrft.data.types)
                lypij__hlr = DataFrameType(cnh__cfcf, df.index, zmzu__dxqgu)
            elif isinstance(one__ahrft, SeriesType):
                diccg__pebv, zmzu__dxqgu = one__ahrft.const_info
                cnh__cfcf = tuple(dtype_to_array_type(one__ahrft.dtype) for
                    cjxcc__zecxr in range(diccg__pebv))
                lypij__hlr = DataFrameType(cnh__cfcf, df.index, zmzu__dxqgu)
            else:
                dmn__xxwlj = get_udf_out_arr_type(one__ahrft)
                lypij__hlr = SeriesType(dmn__xxwlj.dtype, dmn__xxwlj, df.
                    index, None)
        else:
            lypij__hlr = one__ahrft
        mdu__jyouu = ', '.join("{} = ''".format(cdlfw__yzr) for cdlfw__yzr in
            kws.keys())
        typ__ilqs = f"""def apply_stub(func, axis=0, raw=False, result_type=None, args=(), {mdu__jyouu}):
"""
        typ__ilqs += '    pass\n'
        foa__gxbh = {}
        exec(typ__ilqs, {}, foa__gxbh)
        vylc__liat = foa__gxbh['apply_stub']
        vqbyt__jdi = numba.core.utils.pysignature(vylc__liat)
        ypccw__bcvw = (nnknp__xpva, axis, djf__wdyok, ekqf__lzd, hkiru__egmlu
            ) + tuple(kws.values())
        return signature(lypij__hlr, *ypccw__bcvw).replace(pysig=vqbyt__jdi)

    @bound_function('df.plot', no_unliteral=True)
    def resolve_plot(self, df, args, kws):
        func_name = 'DataFrame.plot'
        check_runtime_cols_unsupported(df, f'{func_name}()')
        cjbx__gbht = ('x', 'y', 'kind', 'figsize', 'ax', 'subplots',
            'sharex', 'sharey', 'layout', 'use_index', 'title', 'grid',
            'legend', 'style', 'logx', 'logy', 'loglog', 'xticks', 'yticks',
            'xlim', 'ylim', 'rot', 'fontsize', 'colormap', 'table', 'yerr',
            'xerr', 'secondary_y', 'sort_columns', 'xlabel', 'ylabel',
            'position', 'stacked', 'mark_right', 'include_bool', 'backend')
        qrl__wxsx = {'x': None, 'y': None, 'kind': 'line', 'figsize': None,
            'ax': None, 'subplots': False, 'sharex': None, 'sharey': False,
            'layout': None, 'use_index': True, 'title': None, 'grid': None,
            'legend': True, 'style': None, 'logx': False, 'logy': False,
            'loglog': False, 'xticks': None, 'yticks': None, 'xlim': None,
            'ylim': None, 'rot': None, 'fontsize': None, 'colormap': None,
            'table': False, 'yerr': None, 'xerr': None, 'secondary_y': 
            False, 'sort_columns': False, 'xlabel': None, 'ylabel': None,
            'position': 0.5, 'stacked': False, 'mark_right': True,
            'include_bool': False, 'backend': None}
        rdir__ulbud = ('subplots', 'sharex', 'sharey', 'layout',
            'use_index', 'grid', 'style', 'logx', 'logy', 'loglog', 'xlim',
            'ylim', 'rot', 'colormap', 'table', 'yerr', 'xerr',
            'sort_columns', 'secondary_y', 'colorbar', 'position',
            'stacked', 'mark_right', 'include_bool', 'backend')
        vqbyt__jdi, gazxi__vwo = bodo.utils.typing.fold_typing_args(func_name,
            args, kws, cjbx__gbht, qrl__wxsx, rdir__ulbud)
        qbe__amkjs = gazxi__vwo[2]
        if not is_overload_constant_str(qbe__amkjs):
            raise BodoError(
                f"{func_name}: kind must be a constant string and one of ('line', 'scatter')."
                )
        wkfov__xmrx = gazxi__vwo[0]
        if not is_overload_none(wkfov__xmrx) and not (is_overload_int(
            wkfov__xmrx) or is_overload_constant_str(wkfov__xmrx)):
            raise BodoError(
                f'{func_name}: x must be a constant column name, constant integer, or None.'
                )
        if is_overload_constant_str(wkfov__xmrx):
            kua__chzxt = get_overload_const_str(wkfov__xmrx)
            if kua__chzxt not in df.columns:
                raise BodoError(f'{func_name}: {kua__chzxt} column not found.')
        elif is_overload_int(wkfov__xmrx):
            rwo__cib = get_overload_const_int(wkfov__xmrx)
            if rwo__cib > len(df.columns):
                raise BodoError(
                    f'{func_name}: x: {rwo__cib} is out of bounds for axis 0 with size {len(df.columns)}'
                    )
            wkfov__xmrx = df.columns[wkfov__xmrx]
        xen__nrf = gazxi__vwo[1]
        if not is_overload_none(xen__nrf) and not (is_overload_int(xen__nrf
            ) or is_overload_constant_str(xen__nrf)):
            raise BodoError(
                'df.plot(): y must be a constant column name, constant integer, or None.'
                )
        if is_overload_constant_str(xen__nrf):
            gorh__qny = get_overload_const_str(xen__nrf)
            if gorh__qny not in df.columns:
                raise BodoError(f'{func_name}: {gorh__qny} column not found.')
        elif is_overload_int(xen__nrf):
            pde__macv = get_overload_const_int(xen__nrf)
            if pde__macv > len(df.columns):
                raise BodoError(
                    f'{func_name}: y: {pde__macv} is out of bounds for axis 0 with size {len(df.columns)}'
                    )
            xen__nrf = df.columns[xen__nrf]
        tqvhd__vezhq = gazxi__vwo[3]
        if not is_overload_none(tqvhd__vezhq) and not is_tuple_like_type(
            tqvhd__vezhq):
            raise BodoError(
                f'{func_name}: figsize must be a constant numeric tuple (width, height) or None.'
                )
        czndk__wiphz = gazxi__vwo[10]
        if not is_overload_none(czndk__wiphz) and not is_overload_constant_str(
            czndk__wiphz):
            raise BodoError(
                f'{func_name}: title must be a constant string or None.')
        ptr__ejvvq = gazxi__vwo[12]
        if not is_overload_bool(ptr__ejvvq):
            raise BodoError(f'{func_name}: legend must be a boolean type.')
        szh__ssx = gazxi__vwo[17]
        if not is_overload_none(szh__ssx) and not is_tuple_like_type(szh__ssx):
            raise BodoError(
                f'{func_name}: xticks must be a constant tuple or None.')
        hln__ugzk = gazxi__vwo[18]
        if not is_overload_none(hln__ugzk) and not is_tuple_like_type(hln__ugzk
            ):
            raise BodoError(
                f'{func_name}: yticks must be a constant tuple or None.')
        dymr__rvg = gazxi__vwo[22]
        if not is_overload_none(dymr__rvg) and not is_overload_int(dymr__rvg):
            raise BodoError(
                f'{func_name}: fontsize must be an integer or None.')
        kwjik__hky = gazxi__vwo[29]
        if not is_overload_none(kwjik__hky) and not is_overload_constant_str(
            kwjik__hky):
            raise BodoError(
                f'{func_name}: xlabel must be a constant string or None.')
        joiub__urwz = gazxi__vwo[30]
        if not is_overload_none(joiub__urwz) and not is_overload_constant_str(
            joiub__urwz):
            raise BodoError(
                f'{func_name}: ylabel must be a constant string or None.')
        vuks__vyb = types.List(types.mpl_line_2d_type)
        qbe__amkjs = get_overload_const_str(qbe__amkjs)
        if qbe__amkjs == 'scatter':
            if is_overload_none(wkfov__xmrx) and is_overload_none(xen__nrf):
                raise BodoError(
                    f'{func_name}: {qbe__amkjs} requires an x and y column.')
            elif is_overload_none(wkfov__xmrx):
                raise BodoError(
                    f'{func_name}: {qbe__amkjs} x column is missing.')
            elif is_overload_none(xen__nrf):
                raise BodoError(
                    f'{func_name}: {qbe__amkjs} y column is missing.')
            vuks__vyb = types.mpl_path_collection_type
        elif qbe__amkjs != 'line':
            raise BodoError(f'{func_name}: {qbe__amkjs} plot is not supported.'
                )
        return signature(vuks__vyb, *gazxi__vwo).replace(pysig=vqbyt__jdi)

    def generic_resolve(self, df, attr):
        if self._is_existing_attr(attr):
            return
        check_runtime_cols_unsupported(df,
            'Acessing DataFrame columns by attribute')
        if attr in df.columns:
            qiea__zvg = df.columns.index(attr)
            arr_typ = df.data[qiea__zvg]
            return SeriesType(arr_typ.dtype, arr_typ, df.index, types.
                StringLiteral(attr))
        if len(df.columns) > 0 and isinstance(df.columns[0], tuple):
            eacc__esgw = []
            digf__ctqv = []
            mwb__iicp = False
            for i, pbx__ybp in enumerate(df.columns):
                if pbx__ybp[0] != attr:
                    continue
                mwb__iicp = True
                eacc__esgw.append(pbx__ybp[1] if len(pbx__ybp) == 2 else
                    pbx__ybp[1:])
                digf__ctqv.append(df.data[i])
            if mwb__iicp:
                return DataFrameType(tuple(digf__ctqv), df.index, tuple(
                    eacc__esgw))


DataFrameAttribute._no_unliteral = True


@overload(operator.getitem, no_unliteral=True)
def namedtuple_getitem_overload(tup, idx):
    if isinstance(tup, types.BaseNamedTuple) and is_overload_constant_str(idx):
        rchx__ifa = get_overload_const_str(idx)
        val_ind = tup.instance_class._fields.index(rchx__ifa)
        return lambda tup, idx: tup[val_ind]


def decref_df_data(context, builder, payload, df_type):
    if df_type.is_table_format:
        context.nrt.decref(builder, df_type.table_type, builder.
            extract_value(payload.data, 0))
        context.nrt.decref(builder, df_type.index, payload.index)
        if df_type.has_runtime_cols:
            context.nrt.decref(builder, df_type.data[-1], payload.columns)
        return
    for i in range(len(df_type.data)):
        wit__gak = builder.extract_value(payload.data, i)
        context.nrt.decref(builder, df_type.data[i], wit__gak)
    context.nrt.decref(builder, df_type.index, payload.index)


def define_df_dtor(context, builder, df_type, payload_type):
    cjhgm__hbclm = builder.module
    qht__cvd = lir.FunctionType(lir.VoidType(), [cgutils.voidptr_t])
    qpbd__zvfpb = cgutils.get_or_insert_function(cjhgm__hbclm, qht__cvd,
        name='.dtor.df.{}'.format(df_type))
    if not qpbd__zvfpb.is_declaration:
        return qpbd__zvfpb
    qpbd__zvfpb.linkage = 'linkonce_odr'
    builder = lir.IRBuilder(qpbd__zvfpb.append_basic_block())
    tzsks__yhit = qpbd__zvfpb.args[0]
    hmirz__qtiem = context.get_value_type(payload_type).as_pointer()
    tlksg__stcjt = builder.bitcast(tzsks__yhit, hmirz__qtiem)
    payload = context.make_helper(builder, payload_type, ref=tlksg__stcjt)
    decref_df_data(context, builder, payload, df_type)
    has_parent = cgutils.is_not_null(builder, payload.parent)
    with builder.if_then(has_parent):
        bidh__ebc = context.get_python_api(builder)
        dwt__ewka = bidh__ebc.gil_ensure()
        bidh__ebc.decref(payload.parent)
        bidh__ebc.gil_release(dwt__ewka)
    builder.ret_void()
    return qpbd__zvfpb


def construct_dataframe(context, builder, df_type, data_tup, index_val,
    parent=None, colnames=None):
    payload_type = DataFramePayloadType(df_type)
    dnkdk__fpcek = cgutils.create_struct_proxy(payload_type)(context, builder)
    dnkdk__fpcek.data = data_tup
    dnkdk__fpcek.index = index_val
    if colnames is not None:
        assert df_type.has_runtime_cols, 'construct_dataframe can only provide colnames if columns are determined at runtime'
        dnkdk__fpcek.columns = colnames
    bjco__ykqi = context.get_value_type(payload_type)
    aji__atgb = context.get_abi_sizeof(bjco__ykqi)
    auz__lvix = define_df_dtor(context, builder, df_type, payload_type)
    jhpvj__rer = context.nrt.meminfo_alloc_dtor(builder, context.
        get_constant(types.uintp, aji__atgb), auz__lvix)
    xbdrp__aygn = context.nrt.meminfo_data(builder, jhpvj__rer)
    varki__rxah = builder.bitcast(xbdrp__aygn, bjco__ykqi.as_pointer())
    slsh__cki = cgutils.create_struct_proxy(df_type)(context, builder)
    slsh__cki.meminfo = jhpvj__rer
    if parent is None:
        slsh__cki.parent = cgutils.get_null_value(slsh__cki.parent.type)
    else:
        slsh__cki.parent = parent
        dnkdk__fpcek.parent = parent
        has_parent = cgutils.is_not_null(builder, parent)
        with builder.if_then(has_parent):
            bidh__ebc = context.get_python_api(builder)
            dwt__ewka = bidh__ebc.gil_ensure()
            bidh__ebc.incref(parent)
            bidh__ebc.gil_release(dwt__ewka)
    builder.store(dnkdk__fpcek._getvalue(), varki__rxah)
    return slsh__cki._getvalue()


@intrinsic
def init_runtime_cols_dataframe(typingctx, data_typ, index_typ,
    colnames_index_typ=None):
    assert isinstance(data_typ, types.BaseTuple) and isinstance(data_typ.
        dtype, TableType
        ) and data_typ.dtype.has_runtime_cols, 'init_runtime_cols_dataframe must be called with a table that determines columns at runtime.'
    assert bodo.hiframes.pd_index_ext.is_pd_index_type(colnames_index_typ
        ) or isinstance(colnames_index_typ, bodo.hiframes.
        pd_multi_index_ext.MultiIndexType), 'Column names must be an index'
    if isinstance(data_typ.dtype.arr_types, types.UniTuple):
        vfoar__ttk = [data_typ.dtype.arr_types.dtype] * len(data_typ.dtype.
            arr_types)
    else:
        vfoar__ttk = [chpzo__khb for chpzo__khb in data_typ.dtype.arr_types]
    kmfm__owyr = DataFrameType(tuple(vfoar__ttk + [colnames_index_typ]),
        index_typ, None, is_table_format=True)

    def codegen(context, builder, signature, args):
        df_type = signature.return_type
        data_tup, index, col_names = args
        parent = None
        ihz__jypih = construct_dataframe(context, builder, df_type,
            data_tup, index, parent, col_names)
        context.nrt.incref(builder, data_typ, data_tup)
        context.nrt.incref(builder, index_typ, index)
        context.nrt.incref(builder, colnames_index_typ, col_names)
        return ihz__jypih
    sig = signature(kmfm__owyr, data_typ, index_typ, colnames_index_typ)
    return sig, codegen


@intrinsic
def init_dataframe(typingctx, data_tup_typ, index_typ, col_names_typ=None):
    assert is_pd_index_type(index_typ) or isinstance(index_typ, MultiIndexType
        ), 'init_dataframe(): invalid index type'
    diccg__pebv = len(data_tup_typ.types)
    if diccg__pebv == 0:
        jvw__uhath = ()
    elif isinstance(col_names_typ, types.TypeRef):
        jvw__uhath = col_names_typ.instance_type.columns
    else:
        jvw__uhath = get_const_tup_vals(col_names_typ)
    if diccg__pebv == 1 and isinstance(data_tup_typ.types[0], TableType):
        diccg__pebv = len(data_tup_typ.types[0].arr_types)
    assert len(jvw__uhath
        ) == diccg__pebv, 'init_dataframe(): number of column names does not match number of columns'
    is_table_format = False
    joy__binpq = data_tup_typ.types
    if diccg__pebv != 0 and isinstance(data_tup_typ.types[0], TableType):
        joy__binpq = data_tup_typ.types[0].arr_types
        is_table_format = True
    kmfm__owyr = DataFrameType(joy__binpq, index_typ, jvw__uhath,
        is_table_format=is_table_format)

    def codegen(context, builder, signature, args):
        df_type = signature.return_type
        data_tup = args[0]
        index_val = args[1]
        parent = None
        if is_table_format:
            jzo__cbtv = cgutils.create_struct_proxy(kmfm__owyr.table_type)(
                context, builder, builder.extract_value(data_tup, 0))
            parent = jzo__cbtv.parent
        ihz__jypih = construct_dataframe(context, builder, df_type,
            data_tup, index_val, parent, None)
        context.nrt.incref(builder, data_tup_typ, data_tup)
        context.nrt.incref(builder, index_typ, index_val)
        return ihz__jypih
    sig = signature(kmfm__owyr, data_tup_typ, index_typ, col_names_typ)
    return sig, codegen


@intrinsic
def has_parent(typingctx, df=None):
    check_runtime_cols_unsupported(df, 'has_parent')

    def codegen(context, builder, sig, args):
        slsh__cki = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=args[0])
        return cgutils.is_not_null(builder, slsh__cki.parent)
    return signature(types.bool_, df), codegen


@intrinsic
def _column_needs_unboxing(typingctx, df_typ, i_typ=None):
    check_runtime_cols_unsupported(df_typ, '_column_needs_unboxing')
    assert isinstance(df_typ, DataFrameType) and is_overload_constant_int(i_typ
        )

    def codegen(context, builder, sig, args):
        dnkdk__fpcek = get_dataframe_payload(context, builder, df_typ, args[0])
        nhq__gwt = get_overload_const_int(i_typ)
        arr_typ = df_typ.data[nhq__gwt]
        if df_typ.is_table_format:
            jzo__cbtv = cgutils.create_struct_proxy(df_typ.table_type)(context,
                builder, builder.extract_value(dnkdk__fpcek.data, 0))
            zjfu__cacey = df_typ.table_type.type_to_blk[arr_typ]
            fyv__ezh = getattr(jzo__cbtv, f'block_{zjfu__cacey}')
            qke__elp = ListInstance(context, builder, types.List(arr_typ),
                fyv__ezh)
            qrp__toat = context.get_constant(types.int64, df_typ.table_type
                .block_offsets[nhq__gwt])
            wit__gak = qke__elp.getitem(qrp__toat)
        else:
            wit__gak = builder.extract_value(dnkdk__fpcek.data, nhq__gwt)
        wdx__jcvsl = cgutils.alloca_once_value(builder, wit__gak)
        hhn__zpx = cgutils.alloca_once_value(builder, context.
            get_constant_null(arr_typ))
        return is_ll_eq(builder, wdx__jcvsl, hhn__zpx)
    return signature(types.bool_, df_typ, i_typ), codegen


def get_dataframe_payload(context, builder, df_type, value):
    jhpvj__rer = cgutils.create_struct_proxy(df_type)(context, builder, value
        ).meminfo
    payload_type = DataFramePayloadType(df_type)
    payload = context.nrt.meminfo_data(builder, jhpvj__rer)
    hmirz__qtiem = context.get_value_type(payload_type).as_pointer()
    payload = builder.bitcast(payload, hmirz__qtiem)
    return context.make_helper(builder, payload_type, ref=payload)


@intrinsic
def _get_dataframe_data(typingctx, df_typ=None):
    check_runtime_cols_unsupported(df_typ, '_get_dataframe_data')
    kmfm__owyr = types.Tuple(df_typ.data)
    if df_typ.is_table_format:
        kmfm__owyr = types.Tuple([TableType(df_typ.data)])
    sig = signature(kmfm__owyr, df_typ)

    def codegen(context, builder, signature, args):
        dnkdk__fpcek = get_dataframe_payload(context, builder, signature.
            args[0], args[0])
        return impl_ret_borrowed(context, builder, signature.return_type,
            dnkdk__fpcek.data)
    return sig, codegen


@intrinsic
def get_dataframe_index(typingctx, df_typ=None):
    check_runtime_cols_unsupported(df_typ, 'get_dataframe_index')

    def codegen(context, builder, signature, args):
        dnkdk__fpcek = get_dataframe_payload(context, builder, signature.
            args[0], args[0])
        return impl_ret_borrowed(context, builder, df_typ.index,
            dnkdk__fpcek.index)
    kmfm__owyr = df_typ.index
    sig = signature(kmfm__owyr, df_typ)
    return sig, codegen


def get_dataframe_data(df, i):
    return df[i]


@infer_global(get_dataframe_data)
class GetDataFrameDataInfer(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 2
        if not is_overload_constant_int(args[1]):
            raise_bodo_error(
                'Selecting a DataFrame column requires a constant column label'
                )
        df = args[0]
        check_runtime_cols_unsupported(df, 'get_dataframe_data')
        i = get_overload_const_int(args[1])
        vslxi__flwc = df.data[i]
        return vslxi__flwc(*args)


GetDataFrameDataInfer.prefer_literal = True


def get_dataframe_data_impl(df, i):
    if df.is_table_format:

        def _impl(df, i):
            if has_parent(df) and _column_needs_unboxing(df, i):
                bodo.hiframes.boxing.unbox_dataframe_column(df, i)
            return get_table_data(_get_dataframe_data(df)[0], i)
        return _impl

    def _impl(df, i):
        if has_parent(df) and _column_needs_unboxing(df, i):
            bodo.hiframes.boxing.unbox_dataframe_column(df, i)
        return _get_dataframe_data(df)[i]
    return _impl


@intrinsic
def get_dataframe_table(typingctx, df_typ=None):
    assert df_typ.is_table_format, 'get_dataframe_table() expects table format'

    def codegen(context, builder, signature, args):
        dnkdk__fpcek = get_dataframe_payload(context, builder, signature.
            args[0], args[0])
        return impl_ret_borrowed(context, builder, df_typ.table_type,
            builder.extract_value(dnkdk__fpcek.data, 0))
    return df_typ.table_type(df_typ), codegen


@lower_builtin(get_dataframe_data, DataFrameType, types.IntegerLiteral)
def lower_get_dataframe_data(context, builder, sig, args):
    impl = get_dataframe_data_impl(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def alias_ext_dummy_func(lhs_name, args, alias_map, arg_aliases):
    assert len(args) >= 1
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['get_dataframe_data',
    'bodo.hiframes.pd_dataframe_ext'] = alias_ext_dummy_func
numba.core.ir_utils.alias_func_extensions['get_dataframe_index',
    'bodo.hiframes.pd_dataframe_ext'] = alias_ext_dummy_func
numba.core.ir_utils.alias_func_extensions['get_dataframe_table',
    'bodo.hiframes.pd_dataframe_ext'] = alias_ext_dummy_func


def alias_ext_init_dataframe(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 3
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)
    numba.core.ir_utils._add_alias(lhs_name, args[1].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_dataframe',
    'bodo.hiframes.pd_dataframe_ext'] = alias_ext_init_dataframe


def init_dataframe_equiv(self, scope, equiv_set, loc, args, kws):
    from bodo.hiframes.pd_index_ext import HeterogeneousIndexType
    assert len(args) == 3 and not kws
    data_tup = args[0]
    index = args[1]
    kxutf__xgmaw = self.typemap[data_tup.name]
    if any(is_tuple_like_type(chpzo__khb) for chpzo__khb in kxutf__xgmaw.types
        ):
        return None
    if equiv_set.has_shape(data_tup):
        fjg__myqab = equiv_set.get_shape(data_tup)
        if len(fjg__myqab) > 1:
            equiv_set.insert_equiv(*fjg__myqab)
        if len(fjg__myqab) > 0:
            tekoc__nwqqd = self.typemap[index.name]
            if not isinstance(tekoc__nwqqd, HeterogeneousIndexType
                ) and equiv_set.has_shape(index):
                equiv_set.insert_equiv(fjg__myqab[0], index)
            return ArrayAnalysis.AnalyzeResult(shape=(fjg__myqab[0], len(
                fjg__myqab)), pre=[])
    return None


(ArrayAnalysis._analyze_op_call_bodo_hiframes_pd_dataframe_ext_init_dataframe
    ) = init_dataframe_equiv


def get_dataframe_data_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    rxeie__nhwf = args[0]
    bvq__bzbh = self.typemap[rxeie__nhwf.name].data
    if any(is_tuple_like_type(chpzo__khb) for chpzo__khb in bvq__bzbh):
        return None
    if equiv_set.has_shape(rxeie__nhwf):
        return ArrayAnalysis.AnalyzeResult(shape=equiv_set.get_shape(
            rxeie__nhwf)[0], pre=[])
    return None


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_pd_dataframe_ext_get_dataframe_data
    ) = get_dataframe_data_equiv


def get_dataframe_index_equiv(self, scope, equiv_set, loc, args, kws):
    from bodo.hiframes.pd_index_ext import HeterogeneousIndexType
    assert len(args) == 1 and not kws
    rxeie__nhwf = args[0]
    tekoc__nwqqd = self.typemap[rxeie__nhwf.name].index
    if isinstance(tekoc__nwqqd, HeterogeneousIndexType):
        return None
    if equiv_set.has_shape(rxeie__nhwf):
        return ArrayAnalysis.AnalyzeResult(shape=equiv_set.get_shape(
            rxeie__nhwf)[0], pre=[])
    return None


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_pd_dataframe_ext_get_dataframe_index
    ) = get_dataframe_index_equiv


def get_dataframe_table_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 1 and not kws
    rxeie__nhwf = args[0]
    if equiv_set.has_shape(rxeie__nhwf):
        return ArrayAnalysis.AnalyzeResult(shape=equiv_set.get_shape(
            rxeie__nhwf), pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_pd_dataframe_ext_get_dataframe_table
    ) = get_dataframe_table_equiv


@intrinsic
def set_dataframe_data(typingctx, df_typ, c_ind_typ, arr_typ=None):
    check_runtime_cols_unsupported(df_typ, 'set_dataframe_data')
    assert is_overload_constant_int(c_ind_typ)
    nhq__gwt = get_overload_const_int(c_ind_typ)
    if df_typ.data[nhq__gwt] != arr_typ:
        raise BodoError(
            'Changing dataframe column data type inplace is not supported in conditionals/loops or for dataframe arguments'
            )

    def codegen(context, builder, signature, args):
        jdbfh__ghgc, cjxcc__zecxr, jhroo__pkkk = args
        dnkdk__fpcek = get_dataframe_payload(context, builder, df_typ,
            jdbfh__ghgc)
        if df_typ.is_table_format:
            jzo__cbtv = cgutils.create_struct_proxy(df_typ.table_type)(context,
                builder, builder.extract_value(dnkdk__fpcek.data, 0))
            zjfu__cacey = df_typ.table_type.type_to_blk[arr_typ]
            fyv__ezh = getattr(jzo__cbtv, f'block_{zjfu__cacey}')
            qke__elp = ListInstance(context, builder, types.List(arr_typ),
                fyv__ezh)
            qrp__toat = context.get_constant(types.int64, df_typ.table_type
                .block_offsets[nhq__gwt])
            qke__elp.setitem(qrp__toat, jhroo__pkkk, True)
        else:
            wit__gak = builder.extract_value(dnkdk__fpcek.data, nhq__gwt)
            context.nrt.decref(builder, df_typ.data[nhq__gwt], wit__gak)
            dnkdk__fpcek.data = builder.insert_value(dnkdk__fpcek.data,
                jhroo__pkkk, nhq__gwt)
            context.nrt.incref(builder, arr_typ, jhroo__pkkk)
        slsh__cki = cgutils.create_struct_proxy(df_typ)(context, builder,
            value=jdbfh__ghgc)
        payload_type = DataFramePayloadType(df_typ)
        tlksg__stcjt = context.nrt.meminfo_data(builder, slsh__cki.meminfo)
        hmirz__qtiem = context.get_value_type(payload_type).as_pointer()
        tlksg__stcjt = builder.bitcast(tlksg__stcjt, hmirz__qtiem)
        builder.store(dnkdk__fpcek._getvalue(), tlksg__stcjt)
        return impl_ret_borrowed(context, builder, df_typ, jdbfh__ghgc)
    sig = signature(df_typ, df_typ, c_ind_typ, arr_typ)
    return sig, codegen


@intrinsic
def set_df_index(typingctx, df_t, index_t=None):
    check_runtime_cols_unsupported(df_t, 'set_df_index')

    def codegen(context, builder, signature, args):
        vkjwl__zalm = args[0]
        index_val = args[1]
        df_typ = signature.args[0]
        ste__aes = cgutils.create_struct_proxy(df_typ)(context, builder,
            value=vkjwl__zalm)
        yctm__cjm = get_dataframe_payload(context, builder, df_typ, vkjwl__zalm
            )
        slsh__cki = construct_dataframe(context, builder, signature.
            return_type, yctm__cjm.data, index_val, ste__aes.parent, None)
        context.nrt.incref(builder, index_t, index_val)
        context.nrt.incref(builder, types.Tuple(df_t.data), yctm__cjm.data)
        return slsh__cki
    kmfm__owyr = DataFrameType(df_t.data, index_t, df_t.columns, df_t.dist,
        df_t.is_table_format)
    sig = signature(kmfm__owyr, df_t, index_t)
    return sig, codegen


@intrinsic
def set_df_column_with_reflect(typingctx, df_type, cname_type, arr_type=None):
    check_runtime_cols_unsupported(df_type, 'set_df_column_with_reflect')
    assert is_literal_type(cname_type), 'constant column name expected'
    col_name = get_literal_value(cname_type)
    diccg__pebv = len(df_type.columns)
    ztiej__cgk = diccg__pebv
    sjuh__msbk = df_type.data
    jvw__uhath = df_type.columns
    index_typ = df_type.index
    zreas__bbjvz = col_name not in df_type.columns
    nhq__gwt = diccg__pebv
    if zreas__bbjvz:
        sjuh__msbk += arr_type,
        jvw__uhath += col_name,
        ztiej__cgk += 1
    else:
        nhq__gwt = df_type.columns.index(col_name)
        sjuh__msbk = tuple(arr_type if i == nhq__gwt else sjuh__msbk[i] for
            i in range(diccg__pebv))

    def codegen(context, builder, signature, args):
        jdbfh__ghgc, cjxcc__zecxr, jhroo__pkkk = args
        in_dataframe_payload = get_dataframe_payload(context, builder,
            df_type, jdbfh__ghgc)
        fllw__obhi = cgutils.create_struct_proxy(df_type)(context, builder,
            value=jdbfh__ghgc)
        if df_type.is_table_format:
            zrb__stge = df_type.table_type
            mgu__nyk = builder.extract_value(in_dataframe_payload.data, 0)
            etk__via = TableType(sjuh__msbk)
            jdozt__mwca = set_table_data_codegen(context, builder,
                zrb__stge, mgu__nyk, etk__via, arr_type, jhroo__pkkk,
                nhq__gwt, zreas__bbjvz)
            data_tup = context.make_tuple(builder, types.Tuple([etk__via]),
                [jdozt__mwca])
        else:
            joy__binpq = [(builder.extract_value(in_dataframe_payload.data,
                i) if i != nhq__gwt else jhroo__pkkk) for i in range(
                diccg__pebv)]
            if zreas__bbjvz:
                joy__binpq.append(jhroo__pkkk)
            for rxeie__nhwf, ylw__rnkq in zip(joy__binpq, sjuh__msbk):
                context.nrt.incref(builder, ylw__rnkq, rxeie__nhwf)
            data_tup = context.make_tuple(builder, types.Tuple(sjuh__msbk),
                joy__binpq)
        index_val = in_dataframe_payload.index
        context.nrt.incref(builder, index_typ, index_val)
        dnyr__qtz = construct_dataframe(context, builder, signature.
            return_type, data_tup, index_val, fllw__obhi.parent, None)
        if not zreas__bbjvz and arr_type == df_type.data[nhq__gwt]:
            decref_df_data(context, builder, in_dataframe_payload, df_type)
            payload_type = DataFramePayloadType(df_type)
            tlksg__stcjt = context.nrt.meminfo_data(builder, fllw__obhi.meminfo
                )
            hmirz__qtiem = context.get_value_type(payload_type).as_pointer()
            tlksg__stcjt = builder.bitcast(tlksg__stcjt, hmirz__qtiem)
            vmda__xzf = get_dataframe_payload(context, builder, df_type,
                dnyr__qtz)
            builder.store(vmda__xzf._getvalue(), tlksg__stcjt)
            context.nrt.incref(builder, index_typ, index_val)
            if df_type.is_table_format:
                context.nrt.incref(builder, etk__via, builder.extract_value
                    (data_tup, 0))
            else:
                for rxeie__nhwf, ylw__rnkq in zip(joy__binpq, sjuh__msbk):
                    context.nrt.incref(builder, ylw__rnkq, rxeie__nhwf)
        has_parent = cgutils.is_not_null(builder, fllw__obhi.parent)
        with builder.if_then(has_parent):
            bidh__ebc = context.get_python_api(builder)
            dwt__ewka = bidh__ebc.gil_ensure()
            scrxm__hnr = context.get_env_manager(builder)
            context.nrt.incref(builder, arr_type, jhroo__pkkk)
            ajkeq__tnmny = numba.core.pythonapi._BoxContext(context,
                builder, bidh__ebc, scrxm__hnr)
            csyc__beyr = ajkeq__tnmny.pyapi.from_native_value(arr_type,
                jhroo__pkkk, ajkeq__tnmny.env_manager)
            if isinstance(col_name, str):
                uwkot__yrkt = context.insert_const_string(builder.module,
                    col_name)
                jves__jex = bidh__ebc.string_from_string(uwkot__yrkt)
            else:
                assert isinstance(col_name, int)
                jves__jex = bidh__ebc.long_from_longlong(context.
                    get_constant(types.intp, col_name))
            bidh__ebc.object_setitem(fllw__obhi.parent, jves__jex, csyc__beyr)
            bidh__ebc.decref(csyc__beyr)
            bidh__ebc.decref(jves__jex)
            bidh__ebc.gil_release(dwt__ewka)
        return dnyr__qtz
    kmfm__owyr = DataFrameType(sjuh__msbk, index_typ, jvw__uhath, df_type.
        dist, df_type.is_table_format)
    sig = signature(kmfm__owyr, df_type, cname_type, arr_type)
    return sig, codegen


@lower_constant(DataFrameType)
def lower_constant_dataframe(context, builder, df_type, pyval):
    check_runtime_cols_unsupported(df_type, 'lowering a constant DataFrame')
    diccg__pebv = len(pyval.columns)
    joy__binpq = tuple(pyval.iloc[:, i].values for i in range(diccg__pebv))
    if df_type.is_table_format:
        jzo__cbtv = context.get_constant_generic(builder, df_type.
            table_type, Table(joy__binpq))
        data_tup = lir.Constant.literal_struct([jzo__cbtv])
    else:
        data_tup = lir.Constant.literal_struct([context.
            get_constant_generic(builder, df_type.data[i], pbx__ybp) for i,
            pbx__ybp in enumerate(joy__binpq)])
    index_val = context.get_constant_generic(builder, df_type.index, pyval.
        index)
    vmz__ebfaj = context.get_constant_null(types.pyobject)
    payload = lir.Constant.literal_struct([data_tup, index_val, vmz__ebfaj])
    payload = cgutils.global_constant(builder, '.const.payload', payload
        ).bitcast(cgutils.voidptr_t)
    sfsn__dac = context.get_constant(types.int64, -1)
    jyp__xbf = context.get_constant_null(types.voidptr)
    jhpvj__rer = lir.Constant.literal_struct([sfsn__dac, jyp__xbf, jyp__xbf,
        payload, sfsn__dac])
    jhpvj__rer = cgutils.global_constant(builder, '.const.meminfo', jhpvj__rer
        ).bitcast(cgutils.voidptr_t)
    return lir.Constant.literal_struct([jhpvj__rer, vmz__ebfaj])


@lower_cast(DataFrameType, DataFrameType)
def cast_df_to_df(context, builder, fromty, toty, val):
    if (fromty.data == toty.data and fromty.index == toty.index and fromty.
        columns == toty.columns and fromty.is_table_format == toty.
        is_table_format and fromty.dist != toty.dist and fromty.
        has_runtime_cols == toty.has_runtime_cols):
        return val
    if not fromty.has_runtime_cols and not toty.has_runtime_cols and len(fromty
        .data) == 0 and len(toty.columns):
        return _cast_empty_df(context, builder, toty)
    if (fromty.data != toty.data or fromty.has_runtime_cols != toty.
        has_runtime_cols):
        raise BodoError(f'Invalid dataframe cast from {fromty} to {toty}')
    in_dataframe_payload = get_dataframe_payload(context, builder, fromty, val)
    if isinstance(fromty.index, RangeIndexType) and isinstance(toty.index,
        NumericIndexType):
        wqu__bio = context.cast(builder, in_dataframe_payload.index, fromty
            .index, toty.index)
    else:
        wqu__bio = in_dataframe_payload.index
        context.nrt.incref(builder, fromty.index, wqu__bio)
    if fromty.is_table_format == toty.is_table_format:
        digf__ctqv = in_dataframe_payload.data
        if fromty.is_table_format:
            context.nrt.incref(builder, types.Tuple([fromty.table_type]),
                digf__ctqv)
        else:
            context.nrt.incref(builder, types.BaseTuple.from_types(fromty.
                data), digf__ctqv)
    elif toty.is_table_format:
        digf__ctqv = _cast_df_data_to_table_format(context, builder, fromty,
            toty, in_dataframe_payload)
    else:
        digf__ctqv = _cast_df_data_to_tuple_format(context, builder, fromty,
            toty, in_dataframe_payload)
    return construct_dataframe(context, builder, toty, digf__ctqv, wqu__bio,
        in_dataframe_payload.parent, None)


def _cast_empty_df(context, builder, toty):
    hvb__aqt = {}
    if isinstance(toty.index, RangeIndexType):
        index = 'bodo.hiframes.pd_index_ext.init_range_index(0, 0, 1, None)'
    else:
        rfqqx__pzcu = get_index_data_arr_types(toty.index)[0]
        czczn__wtajn = bodo.utils.transform.get_type_alloc_counts(rfqqx__pzcu
            ) - 1
        ojcvk__yftgp = ', '.join('0' for cjxcc__zecxr in range(czczn__wtajn))
        index = (
            'bodo.utils.conversion.index_from_array(bodo.utils.utils.alloc_type(0, index_arr_type, ({}{})))'
            .format(ojcvk__yftgp, ', ' if czczn__wtajn == 1 else ''))
        hvb__aqt['index_arr_type'] = rfqqx__pzcu
    hdfmp__vvjul = []
    for i, arr_typ in enumerate(toty.data):
        czczn__wtajn = bodo.utils.transform.get_type_alloc_counts(arr_typ) - 1
        ojcvk__yftgp = ', '.join('0' for cjxcc__zecxr in range(czczn__wtajn))
        pprf__fqic = ('bodo.utils.utils.alloc_type(0, arr_type{}, ({}{}))'.
            format(i, ojcvk__yftgp, ', ' if czczn__wtajn == 1 else ''))
        hdfmp__vvjul.append(pprf__fqic)
        hvb__aqt[f'arr_type{i}'] = arr_typ
    hdfmp__vvjul = ', '.join(hdfmp__vvjul)
    typ__ilqs = 'def impl():\n'
    qaxs__ilt = bodo.hiframes.dataframe_impl._gen_init_df(typ__ilqs, toty.
        columns, hdfmp__vvjul, index, hvb__aqt)
    df = context.compile_internal(builder, qaxs__ilt, toty(), [])
    return df


def _cast_df_data_to_table_format(context, builder, fromty, toty,
    in_dataframe_payload):
    check_runtime_cols_unsupported(toty,
        'casting traditional DataFrame to table format')
    goxhb__gdv = toty.table_type
    jzo__cbtv = cgutils.create_struct_proxy(goxhb__gdv)(context, builder)
    jzo__cbtv.parent = in_dataframe_payload.parent
    for chpzo__khb, zjfu__cacey in goxhb__gdv.type_to_blk.items():
        pbl__szipg = context.get_constant(types.int64, len(goxhb__gdv.
            block_to_arr_ind[zjfu__cacey]))
        cjxcc__zecxr, ayfil__qvdvq = ListInstance.allocate_ex(context,
            builder, types.List(chpzo__khb), pbl__szipg)
        ayfil__qvdvq.size = pbl__szipg
        setattr(jzo__cbtv, f'block_{zjfu__cacey}', ayfil__qvdvq.value)
    for i, chpzo__khb in enumerate(fromty.data):
        wit__gak = builder.extract_value(in_dataframe_payload.data, i)
        zjfu__cacey = goxhb__gdv.type_to_blk[chpzo__khb]
        fyv__ezh = getattr(jzo__cbtv, f'block_{zjfu__cacey}')
        qke__elp = ListInstance(context, builder, types.List(chpzo__khb),
            fyv__ezh)
        qrp__toat = context.get_constant(types.int64, goxhb__gdv.
            block_offsets[i])
        qke__elp.setitem(qrp__toat, wit__gak, True)
    data_tup = context.make_tuple(builder, types.Tuple([goxhb__gdv]), [
        jzo__cbtv._getvalue()])
    return data_tup


def _cast_df_data_to_tuple_format(context, builder, fromty, toty,
    in_dataframe_payload):
    check_runtime_cols_unsupported(fromty,
        'casting table format to traditional DataFrame')
    goxhb__gdv = fromty.table_type
    jzo__cbtv = cgutils.create_struct_proxy(goxhb__gdv)(context, builder,
        builder.extract_value(in_dataframe_payload.data, 0))
    joy__binpq = []
    for i, chpzo__khb in enumerate(toty.data):
        zjfu__cacey = goxhb__gdv.type_to_blk[chpzo__khb]
        fyv__ezh = getattr(jzo__cbtv, f'block_{zjfu__cacey}')
        qke__elp = ListInstance(context, builder, types.List(chpzo__khb),
            fyv__ezh)
        qrp__toat = context.get_constant(types.int64, goxhb__gdv.
            block_offsets[i])
        wit__gak = qke__elp.getitem(qrp__toat)
        context.nrt.incref(builder, chpzo__khb, wit__gak)
        joy__binpq.append(wit__gak)
    data_tup = context.make_tuple(builder, types.Tuple(toty.data), joy__binpq)
    return data_tup


@overload(pd.DataFrame, inline='always', no_unliteral=True)
def pd_dataframe_overload(data=None, index=None, columns=None, dtype=None,
    copy=False):
    if not is_overload_constant_bool(copy):
        raise BodoError(
            "pd.DataFrame(): 'copy' argument should be a constant boolean")
    copy = get_overload_const(copy)
    qew__freyp, hdfmp__vvjul, index_arg = _get_df_args(data, index, columns,
        dtype, copy)
    skp__tfk = gen_const_tup(qew__freyp)
    typ__ilqs = (
        'def _init_df(data=None, index=None, columns=None, dtype=None, copy=False):\n'
        )
    typ__ilqs += (
        '  return bodo.hiframes.pd_dataframe_ext.init_dataframe({}, {}, {})\n'
        .format(hdfmp__vvjul, index_arg, skp__tfk))
    foa__gxbh = {}
    exec(typ__ilqs, {'bodo': bodo, 'np': np}, foa__gxbh)
    ayyi__zgfrv = foa__gxbh['_init_df']
    return ayyi__zgfrv


def _get_df_args(data, index, columns, dtype, copy):
    geap__ydkeh = ''
    if not is_overload_none(dtype):
        geap__ydkeh = '.astype(dtype)'
    index_is_none = is_overload_none(index)
    index_arg = 'bodo.utils.conversion.convert_to_index(index)'
    if isinstance(data, types.BaseTuple):
        if not data.types[0] == types.StringLiteral('__bodo_tup'):
            raise BodoError('pd.DataFrame tuple input data not supported yet')
        assert len(data.types) % 2 == 1, 'invalid const dict tuple structure'
        diccg__pebv = (len(data.types) - 1) // 2
        ina__uyf = [chpzo__khb.literal_value for chpzo__khb in data.types[1
            :diccg__pebv + 1]]
        data_val_types = dict(zip(ina__uyf, data.types[diccg__pebv + 1:]))
        joy__binpq = ['data[{}]'.format(i) for i in range(diccg__pebv + 1, 
            2 * diccg__pebv + 1)]
        data_dict = dict(zip(ina__uyf, joy__binpq))
        if is_overload_none(index):
            for i, chpzo__khb in enumerate(data.types[diccg__pebv + 1:]):
                if isinstance(chpzo__khb, SeriesType):
                    index_arg = (
                        'bodo.hiframes.pd_series_ext.get_series_index(data[{}])'
                        .format(diccg__pebv + 1 + i))
                    index_is_none = False
                    break
    elif is_overload_none(data):
        data_dict = {}
        data_val_types = {}
    else:
        if not (isinstance(data, types.Array) and data.ndim == 2):
            raise BodoError(
                'pd.DataFrame() only supports constant dictionary and array input'
                )
        if is_overload_none(columns):
            raise BodoError(
                "pd.DataFrame() 'columns' argument is required when an array is passed as data"
                )
        tupln__ifws = '.copy()' if copy else ''
        vnqds__sazct = get_overload_const_list(columns)
        diccg__pebv = len(vnqds__sazct)
        data_val_types = {ajkeq__tnmny: data.copy(ndim=1) for ajkeq__tnmny in
            vnqds__sazct}
        joy__binpq = ['data[:,{}]{}'.format(i, tupln__ifws) for i in range(
            diccg__pebv)]
        data_dict = dict(zip(vnqds__sazct, joy__binpq))
    if is_overload_none(columns):
        col_names = data_dict.keys()
    else:
        col_names = get_overload_const_list(columns)
    df_len = _get_df_len_from_info(data_dict, data_val_types, col_names,
        index_is_none, index_arg)
    _fill_null_arrays(data_dict, col_names, df_len, dtype)
    if index_is_none:
        if is_overload_none(data):
            index_arg = (
                'bodo.hiframes.pd_index_ext.init_binary_str_index(bodo.libs.str_arr_ext.pre_alloc_string_array(0, 0))'
                )
        else:
            index_arg = (
                'bodo.hiframes.pd_index_ext.init_range_index(0, {}, 1, None)'
                .format(df_len))
    hdfmp__vvjul = '({},)'.format(', '.join(
        'bodo.utils.conversion.coerce_to_array({}, True, scalar_to_arr_len={}){}'
        .format(data_dict[ajkeq__tnmny], df_len, geap__ydkeh) for
        ajkeq__tnmny in col_names))
    if len(col_names) == 0:
        hdfmp__vvjul = '()'
    return col_names, hdfmp__vvjul, index_arg


def _get_df_len_from_info(data_dict, data_val_types, col_names,
    index_is_none, index_arg):
    df_len = '0'
    for ajkeq__tnmny in col_names:
        if ajkeq__tnmny in data_dict and is_iterable_type(data_val_types[
            ajkeq__tnmny]):
            df_len = 'len({})'.format(data_dict[ajkeq__tnmny])
            break
    if df_len == '0' and not index_is_none:
        df_len = f'len({index_arg})'
    return df_len


def _fill_null_arrays(data_dict, col_names, df_len, dtype):
    if all(ajkeq__tnmny in data_dict for ajkeq__tnmny in col_names):
        return
    if is_overload_none(dtype):
        dtype = 'bodo.string_array_type'
    else:
        dtype = 'bodo.utils.conversion.array_type_from_dtype(dtype)'
    xaxwm__pled = 'bodo.libs.array_kernels.gen_na_array({}, {})'.format(df_len,
        dtype)
    for ajkeq__tnmny in col_names:
        if ajkeq__tnmny not in data_dict:
            data_dict[ajkeq__tnmny] = xaxwm__pled


@infer_global(len)
class LenTemplate(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 1
        if isinstance(args[0], (DataFrameType, bodo.TableType)):
            return types.int64(*args)


@lower_builtin(len, DataFrameType)
def table_len_lower(context, builder, sig, args):
    impl = df_len_overload(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def df_len_overload(df):
    if not isinstance(df, DataFrameType):
        return
    if df.has_runtime_cols:

        def impl(df):
            chpzo__khb = bodo.hiframes.pd_dataframe_ext.get_dataframe_table(df)
            return len(chpzo__khb)
        return impl
    if len(df.columns) == 0:
        return lambda df: 0

    def impl(df):
        if is_null_pointer(df._meminfo):
            return 0
        return len(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, 0))
    return impl


@infer_global(operator.getitem)
class GetItemTuple(AbstractTemplate):
    key = operator.getitem

    def generic(self, args, kws):
        tup, idx = args
        if not isinstance(tup, types.BaseTuple) or not isinstance(idx,
            types.IntegerLiteral):
            return
        ixh__jifx = idx.literal_value
        if isinstance(ixh__jifx, int):
            vslxi__flwc = tup.types[ixh__jifx]
        elif isinstance(ixh__jifx, slice):
            vslxi__flwc = types.BaseTuple.from_types(tup.types[ixh__jifx])
        return signature(vslxi__flwc, *args)


GetItemTuple.prefer_literal = True


@lower_builtin(operator.getitem, types.BaseTuple, types.IntegerLiteral)
@lower_builtin(operator.getitem, types.BaseTuple, types.SliceLiteral)
def getitem_tuple_lower(context, builder, sig, args):
    gekcg__ifu, idx = sig.args
    idx = idx.literal_value
    tup, cjxcc__zecxr = args
    if isinstance(idx, int):
        if idx < 0:
            idx += len(gekcg__ifu)
        if not 0 <= idx < len(gekcg__ifu):
            raise IndexError('cannot index at %d in %s' % (idx, gekcg__ifu))
        gmkkx__xzetz = builder.extract_value(tup, idx)
    elif isinstance(idx, slice):
        dugu__afdq = cgutils.unpack_tuple(builder, tup)[idx]
        gmkkx__xzetz = context.make_tuple(builder, sig.return_type, dugu__afdq)
    else:
        raise NotImplementedError('unexpected index %r for %s' % (idx, sig.
            args[0]))
    return impl_ret_borrowed(context, builder, sig.return_type, gmkkx__xzetz)


def join_dummy(left_df, right_df, left_on, right_on, how, suffix_x,
    suffix_y, is_join, indicator, _bodo_na_equal, gen_cond):
    return left_df


@infer_global(join_dummy)
class JoinTyper(AbstractTemplate):

    def generic(self, args, kws):
        from bodo.hiframes.pd_dataframe_ext import DataFrameType
        from bodo.utils.typing import is_overload_str
        assert not kws
        (left_df, right_df, left_on, right_on, yjxqk__fkie, suffix_x,
            suffix_y, is_join, indicator, _bodo_na_equal, wgqv__ixhwt) = args
        left_on = get_overload_const_list(left_on)
        right_on = get_overload_const_list(right_on)
        vjdh__hpbtt = set(left_on) & set(right_on)
        lyis__nttsw = set(left_df.columns) & set(right_df.columns)
        covbh__ttkz = lyis__nttsw - vjdh__hpbtt
        gbtq__tmc = '$_bodo_index_' in left_on
        lkep__mdh = '$_bodo_index_' in right_on
        how = get_overload_const_str(yjxqk__fkie)
        jnh__opf = how in {'left', 'outer'}
        mmuwi__msjgy = how in {'right', 'outer'}
        columns = []
        data = []
        if gbtq__tmc and not lkep__mdh and not is_join.literal_value:
            plmd__ksuk = right_on[0]
            if plmd__ksuk in left_df.columns:
                columns.append(plmd__ksuk)
                data.append(right_df.data[right_df.columns.index(plmd__ksuk)])
        if lkep__mdh and not gbtq__tmc and not is_join.literal_value:
            nsmiy__kuu = left_on[0]
            if nsmiy__kuu in right_df.columns:
                columns.append(nsmiy__kuu)
                data.append(left_df.data[left_df.columns.index(nsmiy__kuu)])
        for nnuca__dkjvu, uuvc__xnz in zip(left_df.data, left_df.columns):
            columns.append(str(uuvc__xnz) + suffix_x.literal_value if 
                uuvc__xnz in covbh__ttkz else uuvc__xnz)
            if uuvc__xnz in vjdh__hpbtt:
                data.append(nnuca__dkjvu)
            else:
                data.append(to_nullable_type(nnuca__dkjvu) if mmuwi__msjgy else
                    nnuca__dkjvu)
        for nnuca__dkjvu, uuvc__xnz in zip(right_df.data, right_df.columns):
            if uuvc__xnz not in vjdh__hpbtt:
                columns.append(str(uuvc__xnz) + suffix_y.literal_value if 
                    uuvc__xnz in covbh__ttkz else uuvc__xnz)
                data.append(to_nullable_type(nnuca__dkjvu) if jnh__opf else
                    nnuca__dkjvu)
        mkb__iuu = get_overload_const_bool(indicator)
        if mkb__iuu:
            columns.append('_merge')
            data.append(bodo.CategoricalArrayType(bodo.PDCategoricalDtype((
                'left_only', 'right_only', 'both'), bodo.string_type, False)))
        index_typ = RangeIndexType(types.none)
        if gbtq__tmc and lkep__mdh and not is_overload_str(how, 'asof'):
            index_typ = left_df.index
            if isinstance(index_typ, bodo.hiframes.pd_index_ext.RangeIndexType
                ):
                index_typ = bodo.hiframes.pd_index_ext.NumericIndexType(types
                    .int64)
        elif gbtq__tmc and not lkep__mdh:
            index_typ = right_df.index
            if isinstance(index_typ, bodo.hiframes.pd_index_ext.RangeIndexType
                ):
                index_typ = bodo.hiframes.pd_index_ext.NumericIndexType(types
                    .int64)
        elif lkep__mdh and not gbtq__tmc:
            index_typ = left_df.index
            if isinstance(index_typ, bodo.hiframes.pd_index_ext.RangeIndexType
                ):
                index_typ = bodo.hiframes.pd_index_ext.NumericIndexType(types
                    .int64)
        ubmd__kmtgz = DataFrameType(tuple(data), index_typ, tuple(columns))
        return signature(ubmd__kmtgz, *args)


JoinTyper._no_unliteral = True


@lower_builtin(join_dummy, types.VarArg(types.Any))
def lower_join_dummy(context, builder, sig, args):
    slsh__cki = cgutils.create_struct_proxy(sig.return_type)(context, builder)
    return slsh__cki._getvalue()


@overload(pd.concat, inline='always', no_unliteral=True)
def concat_overload(objs, axis=0, join='outer', join_axes=None,
    ignore_index=False, keys=None, levels=None, names=None,
    verify_integrity=False, sort=None, copy=True):
    if not is_overload_constant_int(axis):
        raise BodoError("pd.concat(): 'axis' should be a constant integer")
    if not is_overload_constant_bool(ignore_index):
        raise BodoError(
            "pd.concat(): 'ignore_index' should be a constant boolean")
    axis = get_overload_const_int(axis)
    ignore_index = is_overload_true(ignore_index)
    qjbin__nksm = dict(join=join, join_axes=join_axes, keys=keys, levels=
        levels, names=names, verify_integrity=verify_integrity, sort=sort,
        copy=copy)
    qrl__wxsx = dict(join='outer', join_axes=None, keys=None, levels=None,
        names=None, verify_integrity=False, sort=None, copy=True)
    check_unsupported_args('pandas.concat', qjbin__nksm, qrl__wxsx,
        package_name='pandas', module_name='General')
    typ__ilqs = """def impl(objs, axis=0, join='outer', join_axes=None, ignore_index=False, keys=None, levels=None, names=None, verify_integrity=False, sort=None, copy=True):
"""
    if axis == 1:
        if not isinstance(objs, types.BaseTuple):
            raise_bodo_error(
                'Only tuple argument for pd.concat(axis=1) expected')
        index = (
            'bodo.hiframes.pd_index_ext.init_range_index(0, len(objs[0]), 1, None)'
            )
        tgz__mueaf = 0
        hdfmp__vvjul = []
        names = []
        for i, yqjtd__faw in enumerate(objs.types):
            assert isinstance(yqjtd__faw, (SeriesType, DataFrameType))
            check_runtime_cols_unsupported(yqjtd__faw, 'pd.concat()')
            if isinstance(yqjtd__faw, SeriesType):
                names.append(str(tgz__mueaf))
                tgz__mueaf += 1
                hdfmp__vvjul.append(
                    'bodo.hiframes.pd_series_ext.get_series_data(objs[{}])'
                    .format(i))
            else:
                names.extend(yqjtd__faw.columns)
                for jhl__fymh in range(len(yqjtd__faw.data)):
                    hdfmp__vvjul.append(
                        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(objs[{}], {})'
                        .format(i, jhl__fymh))
        return bodo.hiframes.dataframe_impl._gen_init_df(typ__ilqs, names,
            ', '.join(hdfmp__vvjul), index)
    assert axis == 0
    if isinstance(objs, types.BaseTuple) and isinstance(objs.types[0],
        DataFrameType):
        assert all(isinstance(chpzo__khb, DataFrameType) for chpzo__khb in
            objs.types)
        exsay__jvw = []
        for df in objs.types:
            check_runtime_cols_unsupported(df, 'pd.concat()')
            exsay__jvw.extend(df.columns)
        exsay__jvw = list(dict.fromkeys(exsay__jvw).keys())
        vfoar__ttk = {}
        for tgz__mueaf, ajkeq__tnmny in enumerate(exsay__jvw):
            for df in objs.types:
                if ajkeq__tnmny in df.columns:
                    vfoar__ttk['arr_typ{}'.format(tgz__mueaf)] = df.data[df
                        .columns.index(ajkeq__tnmny)]
                    break
        assert len(vfoar__ttk) == len(exsay__jvw)
        wrx__krvpw = []
        for tgz__mueaf, ajkeq__tnmny in enumerate(exsay__jvw):
            args = []
            for i, df in enumerate(objs.types):
                if ajkeq__tnmny in df.columns:
                    nhq__gwt = df.columns.index(ajkeq__tnmny)
                    args.append(
                        'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(objs[{}], {})'
                        .format(i, nhq__gwt))
                else:
                    args.append(
                        'bodo.libs.array_kernels.gen_na_array(len(objs[{}]), arr_typ{})'
                        .format(i, tgz__mueaf))
            typ__ilqs += ('  A{} = bodo.libs.array_kernels.concat(({},))\n'
                .format(tgz__mueaf, ', '.join(args)))
        if ignore_index:
            index = (
                'bodo.hiframes.pd_index_ext.init_range_index(0, len(A0), 1, None)'
                )
        else:
            index = (
                """bodo.utils.conversion.index_from_array(bodo.libs.array_kernels.concat(({},)))
"""
                .format(', '.join(
                'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(objs[{}]))'
                .format(i) for i in range(len(objs.types)) if len(objs[i].
                columns) > 0)))
        return bodo.hiframes.dataframe_impl._gen_init_df(typ__ilqs,
            exsay__jvw, ', '.join('A{}'.format(i) for i in range(len(
            exsay__jvw))), index, vfoar__ttk)
    if isinstance(objs, types.BaseTuple) and isinstance(objs.types[0],
        SeriesType):
        assert all(isinstance(chpzo__khb, SeriesType) for chpzo__khb in
            objs.types)
        typ__ilqs += ('  out_arr = bodo.libs.array_kernels.concat(({},))\n'
            .format(', '.join(
            'bodo.hiframes.pd_series_ext.get_series_data(objs[{}])'.format(
            i) for i in range(len(objs.types)))))
        if ignore_index:
            typ__ilqs += """  index = bodo.hiframes.pd_index_ext.init_range_index(0, len(out_arr), 1, None)
"""
        else:
            typ__ilqs += (
                """  index = bodo.utils.conversion.index_from_array(bodo.libs.array_kernels.concat(({},)))
"""
                .format(', '.join(
                'bodo.utils.conversion.index_to_array(bodo.hiframes.pd_series_ext.get_series_index(objs[{}]))'
                .format(i) for i in range(len(objs.types)))))
        typ__ilqs += (
            '  return bodo.hiframes.pd_series_ext.init_series(out_arr, index)\n'
            )
        foa__gxbh = {}
        exec(typ__ilqs, {'bodo': bodo, 'np': np, 'numba': numba}, foa__gxbh)
        return foa__gxbh['impl']
    if isinstance(objs, types.List) and isinstance(objs.dtype, DataFrameType):
        check_runtime_cols_unsupported(objs.dtype, 'pd.concat()')
        df_type = objs.dtype
        for tgz__mueaf, ajkeq__tnmny in enumerate(df_type.columns):
            typ__ilqs += '  arrs{} = []\n'.format(tgz__mueaf)
            typ__ilqs += '  for i in range(len(objs)):\n'
            typ__ilqs += '    df = objs[i]\n'
            typ__ilqs += (
                """    arrs{0}.append(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {0}))
"""
                .format(tgz__mueaf))
            typ__ilqs += (
                '  out_arr{0} = bodo.libs.array_kernels.concat(arrs{0})\n'.
                format(tgz__mueaf))
        if ignore_index:
            index = (
                'bodo.hiframes.pd_index_ext.init_range_index(0, len(out_arr0), 1, None)'
                )
        else:
            typ__ilqs += '  arrs_index = []\n'
            typ__ilqs += '  for i in range(len(objs)):\n'
            typ__ilqs += '    df = objs[i]\n'
            typ__ilqs += """    arrs_index.append(bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)))
"""
            if objs.dtype.index.name_typ == types.none:
                name = None
            else:
                name = objs.dtype.index.name_typ.literal_value
            index = f"""bodo.utils.conversion.index_from_array(bodo.libs.array_kernels.concat(arrs_index), {name!r})
"""
        return bodo.hiframes.dataframe_impl._gen_init_df(typ__ilqs, df_type
            .columns, ', '.join('out_arr{}'.format(i) for i in range(len(
            df_type.columns))), index)
    if isinstance(objs, types.List) and isinstance(objs.dtype, SeriesType):
        typ__ilqs += '  arrs = []\n'
        typ__ilqs += '  for i in range(len(objs)):\n'
        typ__ilqs += (
            '    arrs.append(bodo.hiframes.pd_series_ext.get_series_data(objs[i]))\n'
            )
        typ__ilqs += '  out_arr = bodo.libs.array_kernels.concat(arrs)\n'
        if ignore_index:
            typ__ilqs += """  index = bodo.hiframes.pd_index_ext.init_range_index(0, len(out_arr), 1, None)
"""
        else:
            typ__ilqs += '  arrs_index = []\n'
            typ__ilqs += '  for i in range(len(objs)):\n'
            typ__ilqs += '    S = objs[i]\n'
            typ__ilqs += """    arrs_index.append(bodo.utils.conversion.index_to_array(bodo.hiframes.pd_series_ext.get_series_index(S)))
"""
            typ__ilqs += """  index = bodo.utils.conversion.index_from_array(bodo.libs.array_kernels.concat(arrs_index))
"""
        typ__ilqs += (
            '  return bodo.hiframes.pd_series_ext.init_series(out_arr, index)\n'
            )
        foa__gxbh = {}
        exec(typ__ilqs, {'bodo': bodo, 'np': np, 'numba': numba}, foa__gxbh)
        return foa__gxbh['impl']
    raise BodoError('pd.concat(): input type {} not supported yet'.format(objs)
        )


def sort_values_dummy(df, by, ascending, inplace, na_position):
    return df.sort_values(by, ascending=ascending, inplace=inplace,
        na_position=na_position)


@infer_global(sort_values_dummy)
class SortDummyTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        df, by, ascending, inplace, na_position = args
        index = df.index
        if isinstance(index, bodo.hiframes.pd_index_ext.RangeIndexType):
            index = bodo.hiframes.pd_index_ext.NumericIndexType(types.int64)
        kmfm__owyr = df.copy(index=index, is_table_format=False)
        return signature(kmfm__owyr, *args)


SortDummyTyper._no_unliteral = True


@lower_builtin(sort_values_dummy, types.VarArg(types.Any))
def lower_sort_values_dummy(context, builder, sig, args):
    if sig.return_type == types.none:
        return
    rdof__rhz = cgutils.create_struct_proxy(sig.return_type)(context, builder)
    return rdof__rhz._getvalue()


@overload_method(DataFrameType, 'itertuples', inline='always', no_unliteral
    =True)
def itertuples_overload(df, index=True, name='Pandas'):
    check_runtime_cols_unsupported(df, 'DataFrame.itertuples()')
    qjbin__nksm = dict(index=index, name=name)
    qrl__wxsx = dict(index=True, name='Pandas')
    check_unsupported_args('DataFrame.itertuples', qjbin__nksm, qrl__wxsx,
        package_name='pandas', module_name='DataFrame')

    def _impl(df, index=True, name='Pandas'):
        return bodo.hiframes.pd_dataframe_ext.itertuples_dummy(df)
    return _impl


def itertuples_dummy(df):
    return df


@infer_global(itertuples_dummy)
class ItertuplesDummyTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        df, = args
        assert 'Index' not in df.columns
        columns = ('Index',) + df.columns
        vfoar__ttk = (types.Array(types.int64, 1, 'C'),) + df.data
        kse__zhe = bodo.hiframes.dataframe_impl.DataFrameTupleIterator(columns,
            vfoar__ttk)
        return signature(kse__zhe, *args)


@lower_builtin(itertuples_dummy, types.VarArg(types.Any))
def lower_itertuples_dummy(context, builder, sig, args):
    rdof__rhz = cgutils.create_struct_proxy(sig.return_type)(context, builder)
    return rdof__rhz._getvalue()


def query_dummy(df, expr):
    return df.eval(expr)


@infer_global(query_dummy)
class QueryDummyTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(SeriesType(types.bool_, index=RangeIndexType(types
            .none)), *args)


@lower_builtin(query_dummy, types.VarArg(types.Any))
def lower_query_dummy(context, builder, sig, args):
    rdof__rhz = cgutils.create_struct_proxy(sig.return_type)(context, builder)
    return rdof__rhz._getvalue()


def val_isin_dummy(S, vals):
    return S in vals


def val_notin_dummy(S, vals):
    return S not in vals


@infer_global(val_isin_dummy)
@infer_global(val_notin_dummy)
class ValIsinTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(SeriesType(types.bool_, index=args[0].index), *args)


@lower_builtin(val_isin_dummy, types.VarArg(types.Any))
@lower_builtin(val_notin_dummy, types.VarArg(types.Any))
def lower_val_isin_dummy(context, builder, sig, args):
    rdof__rhz = cgutils.create_struct_proxy(sig.return_type)(context, builder)
    return rdof__rhz._getvalue()


@numba.generated_jit(nopython=True)
def pivot_impl(index_tup, columns_tup, values_tup, pivot_values,
    index_names, columns_name, value_names, check_duplicates=True, parallel
    =False):
    if not is_overload_constant_bool(check_duplicates):
        raise BodoError(
            'pivot_impl(): check_duplicates must be a constant boolean')
    wbln__bwy = get_overload_const_bool(check_duplicates)
    jrzfw__hqalw = not is_overload_none(value_names)
    hba__xwz = isinstance(values_tup, types.UniTuple)
    if hba__xwz:
        naqv__gdx = [to_nullable_type(values_tup.dtype)]
    else:
        naqv__gdx = [to_nullable_type(ylw__rnkq) for ylw__rnkq in values_tup]
    typ__ilqs = 'def impl(\n'
    typ__ilqs += """    index_tup, columns_tup, values_tup, pivot_values, index_names, columns_name, value_names, check_duplicates=True, parallel=False
"""
    typ__ilqs += '):\n'
    typ__ilqs += '    if parallel:\n'
    iqkq__kfqyc = ', '.join([f'array_to_info(index_tup[{i}])' for i in
        range(len(index_tup))] + [f'array_to_info(columns_tup[{i}])' for i in
        range(len(columns_tup))] + [f'array_to_info(values_tup[{i}])' for i in
        range(len(values_tup))])
    typ__ilqs += f'        info_list = [{iqkq__kfqyc}]\n'
    typ__ilqs += '        cpp_table = arr_info_list_to_table(info_list)\n'
    typ__ilqs += f"""        out_cpp_table = shuffle_table(cpp_table, {len(index_tup)}, parallel, 0)
"""
    ukch__boai = ', '.join([
        f'info_to_array(info_from_table(out_cpp_table, {i}), index_tup[{i}])'
         for i in range(len(index_tup))])
    snz__eda = ', '.join([
        f'info_to_array(info_from_table(out_cpp_table, {i + len(index_tup)}), columns_tup[{i}])'
         for i in range(len(columns_tup))])
    esm__rhgx = ', '.join([
        f'info_to_array(info_from_table(out_cpp_table, {i + len(index_tup) + len(columns_tup)}), values_tup[{i}])'
         for i in range(len(values_tup))])
    typ__ilqs += f'        index_tup = ({ukch__boai},)\n'
    typ__ilqs += f'        columns_tup = ({snz__eda},)\n'
    typ__ilqs += f'        values_tup = ({esm__rhgx},)\n'
    typ__ilqs += '        delete_table(cpp_table)\n'
    typ__ilqs += '        delete_table(out_cpp_table)\n'
    typ__ilqs += '    columns_arr = columns_tup[0]\n'
    if hba__xwz:
        typ__ilqs += '    values_arrs = [arr for arr in values_tup]\n'
    typ__ilqs += """    unique_index_arr_tup, row_vector = bodo.libs.array_ops.array_unique_vector_map(
"""
    typ__ilqs += '        index_tup\n'
    typ__ilqs += '    )\n'
    typ__ilqs += '    n_rows = len(unique_index_arr_tup[0])\n'
    typ__ilqs += '    num_values_arrays = len(values_tup)\n'
    typ__ilqs += '    n_unique_pivots = len(pivot_values)\n'
    if hba__xwz:
        typ__ilqs += '    n_cols = num_values_arrays * n_unique_pivots\n'
    else:
        typ__ilqs += '    n_cols = n_unique_pivots\n'
    typ__ilqs += '    col_map = {}\n'
    typ__ilqs += '    for i in range(n_unique_pivots):\n'
    typ__ilqs += '        if bodo.libs.array_kernels.isna(pivot_values, i):\n'
    typ__ilqs += '            raise ValueError(\n'
    typ__ilqs += """                "DataFrame.pivot(): NA values in 'columns' array not supported\"
"""
    typ__ilqs += '            )\n'
    typ__ilqs += '        col_map[pivot_values[i]] = i\n'
    yiwx__qwvw = False
    for i, exgkh__jfmc in enumerate(naqv__gdx):
        if exgkh__jfmc == bodo.string_array_type:
            yiwx__qwvw = True
            typ__ilqs += (
                f'    len_arrs_{i} = [np.zeros(n_rows, np.int64) for _ in range(n_cols)]\n'
                )
            typ__ilqs += f'    total_lens_{i} = np.zeros(n_cols, np.int64)\n'
    if yiwx__qwvw:
        if wbln__bwy:
            typ__ilqs += '    nbytes = (n_rows + 7) >> 3\n'
            typ__ilqs += """    seen_bitmaps = [np.zeros(nbytes, np.int8) for _ in range(n_unique_pivots)]
"""
        typ__ilqs += '    for i in range(len(columns_arr)):\n'
        typ__ilqs += '        col_name = columns_arr[i]\n'
        typ__ilqs += '        pivot_idx = col_map[col_name]\n'
        typ__ilqs += '        row_idx = row_vector[i]\n'
        if wbln__bwy:
            typ__ilqs += '        seen_bitmap = seen_bitmaps[pivot_idx]\n'
            typ__ilqs += """        if bodo.libs.int_arr_ext.get_bit_bitmap_arr(seen_bitmap, row_idx):
"""
            typ__ilqs += """            raise ValueError("DataFrame.pivot(): 'index' contains duplicate entries for the same output column")
"""
            typ__ilqs += '        else:\n'
            typ__ilqs += """            bodo.libs.int_arr_ext.set_bit_to_arr(seen_bitmap, row_idx, 1)
"""
        if hba__xwz:
            typ__ilqs += '        for j in range(num_values_arrays):\n'
            typ__ilqs += (
                '            col_idx = (j * len(pivot_values)) + pivot_idx\n')
            typ__ilqs += '            len_arr = len_arrs_0[col_idx]\n'
            typ__ilqs += '            values_arr = values_arrs[j]\n'
            typ__ilqs += (
                '            if not bodo.libs.array_kernels.isna(values_arr, i):\n'
                )
            typ__ilqs += (
                '                len_arr[row_idx] = len(values_arr[i])\n')
            typ__ilqs += (
                '                total_lens_0[col_idx] += len(values_arr[i])\n'
                )
        else:
            for i, exgkh__jfmc in enumerate(naqv__gdx):
                if exgkh__jfmc == bodo.string_array_type:
                    typ__ilqs += f"""        if not bodo.libs.array_kernels.isna(values_tup[{i}], i):
"""
                    typ__ilqs += f"""            len_arrs_{i}[pivot_idx][row_idx] = len(values_tup[{i}][i])
"""
                    typ__ilqs += f"""            total_lens_{i}[pivot_idx] += len(values_tup[{i}][i])
"""
    for i, exgkh__jfmc in enumerate(naqv__gdx):
        if exgkh__jfmc == bodo.string_array_type:
            typ__ilqs += f'    data_arrs_{i} = [\n'
            typ__ilqs += (
                '        bodo.libs.str_arr_ext.gen_na_str_array_lens(\n')
            typ__ilqs += (
                f'            n_rows, total_lens_{i}[i], len_arrs_{i}[i]\n')
            typ__ilqs += '        )\n'
            typ__ilqs += '        for i in range(n_cols)\n'
            typ__ilqs += '    ]\n'
        else:
            typ__ilqs += f'    data_arrs_{i} = [\n'
            typ__ilqs += (
                f'        bodo.libs.array_kernels.gen_na_array(n_rows, data_arr_typ_{i})\n'
                )
            typ__ilqs += '        for _ in range(n_cols)\n'
            typ__ilqs += '    ]\n'
    if not yiwx__qwvw and wbln__bwy:
        typ__ilqs += '    nbytes = (n_rows + 7) >> 3\n'
        typ__ilqs += """    seen_bitmaps = [np.zeros(nbytes, np.int8) for _ in range(n_unique_pivots)]
"""
    typ__ilqs += '    for i in range(len(columns_arr)):\n'
    typ__ilqs += '        col_name = columns_arr[i]\n'
    typ__ilqs += '        pivot_idx = col_map[col_name]\n'
    typ__ilqs += '        row_idx = row_vector[i]\n'
    if not yiwx__qwvw and wbln__bwy:
        typ__ilqs += '        seen_bitmap = seen_bitmaps[pivot_idx]\n'
        typ__ilqs += (
            '        if bodo.libs.int_arr_ext.get_bit_bitmap_arr(seen_bitmap, row_idx):\n'
            )
        typ__ilqs += """            raise ValueError("DataFrame.pivot(): 'index' contains duplicate entries for the same output column")
"""
        typ__ilqs += '        else:\n'
        typ__ilqs += (
            '            bodo.libs.int_arr_ext.set_bit_to_arr(seen_bitmap, row_idx, 1)\n'
            )
    if hba__xwz:
        typ__ilqs += '        for j in range(num_values_arrays):\n'
        typ__ilqs += (
            '            col_idx = (j * len(pivot_values)) + pivot_idx\n')
        typ__ilqs += '            col_arr = data_arrs_0[col_idx]\n'
        typ__ilqs += '            values_arr = values_arrs[j]\n'
        typ__ilqs += (
            '            if bodo.libs.array_kernels.isna(values_arr, i):\n')
        typ__ilqs += (
            '                bodo.libs.array_kernels.setna(col_arr, row_idx)\n'
            )
        typ__ilqs += '            else:\n'
        typ__ilqs += '                col_arr[row_idx] = values_arr[i]\n'
    else:
        for i, exgkh__jfmc in enumerate(naqv__gdx):
            typ__ilqs += f'        col_arr_{i} = data_arrs_{i}[pivot_idx]\n'
            typ__ilqs += (
                f'        if bodo.libs.array_kernels.isna(values_tup[{i}], i):\n'
                )
            typ__ilqs += (
                f'            bodo.libs.array_kernels.setna(col_arr_{i}, row_idx)\n'
                )
            typ__ilqs += f'        else:\n'
            typ__ilqs += (
                f'            col_arr_{i}[row_idx] = values_tup[{i}][i]\n')
    if len(index_tup) == 1:
        typ__ilqs += """    index = bodo.utils.conversion.index_from_array(unique_index_arr_tup[0], index_names[0])
"""
    else:
        typ__ilqs += """    index = bodo.hiframes.pd_multi_index_ext.init_multi_index(unique_index_arr_tup, index_names, None)
"""
    if jrzfw__hqalw:
        typ__ilqs += '    num_rows = len(value_names) * len(pivot_values)\n'
        if value_names == bodo.string_array_type:
            typ__ilqs += '    total_chars = 0\n'
            typ__ilqs += '    for i in range(len(value_names)):\n'
            typ__ilqs += '        total_chars += len(value_names[i])\n'
            typ__ilqs += """    new_value_names = bodo.libs.str_arr_ext.pre_alloc_string_array(num_rows, total_chars * len(pivot_values))
"""
        else:
            typ__ilqs += """    new_value_names = bodo.utils.utils.alloc_type(num_rows, value_names, (-1,))
"""
        if pivot_values == bodo.string_array_type:
            typ__ilqs += '    total_chars = 0\n'
            typ__ilqs += '    for i in range(len(pivot_values)):\n'
            typ__ilqs += '        total_chars += len(pivot_values[i])\n'
            typ__ilqs += """    new_pivot_values = bodo.libs.str_arr_ext.pre_alloc_string_array(num_rows, total_chars * len(value_names))
"""
        else:
            typ__ilqs += """    new_pivot_values = bodo.utils.utils.alloc_type(num_rows, pivot_values, (-1,))
"""
        typ__ilqs += '    for i in range(len(value_names)):\n'
        typ__ilqs += '        for j in range(len(pivot_values)):\n'
        typ__ilqs += (
            '            new_value_names[(i * len(pivot_values)) + j] = value_names[i]\n'
            )
        typ__ilqs += """            new_pivot_values[(i * len(pivot_values)) + j] = pivot_values[j]
"""
        typ__ilqs += """    column_index = bodo.hiframes.pd_multi_index_ext.init_multi_index((new_value_names, new_pivot_values), (None, columns_name), None)
"""
    else:
        typ__ilqs += """    column_index =  bodo.utils.conversion.index_from_array(pivot_values, columns_name)
"""
    ektpp__nklvf = ', '.join(f'data_arrs_{i}' for i in range(len(naqv__gdx)))
    typ__ilqs += f"""    table = bodo.hiframes.table.init_runtime_table_from_lists(({ektpp__nklvf},), n_rows)
"""
    typ__ilqs += (
        '    return bodo.hiframes.pd_dataframe_ext.init_runtime_cols_dataframe(\n'
        )
    typ__ilqs += '        (table,), index, column_index\n'
    typ__ilqs += '    )\n'
    foa__gxbh = {}
    fhiun__ghn = {f'data_arr_typ_{i}': exgkh__jfmc for i, exgkh__jfmc in
        enumerate(naqv__gdx)}
    prr__ygd = {'bodo': bodo, 'np': np, 'array_to_info': array_to_info,
        'arr_info_list_to_table': arr_info_list_to_table, 'shuffle_table':
        shuffle_table, 'info_to_array': info_to_array, 'delete_table':
        delete_table, 'info_from_table': info_from_table, **fhiun__ghn}
    exec(typ__ilqs, prr__ygd, foa__gxbh)
    impl = foa__gxbh['impl']
    return impl


def gen_pandas_parquet_metadata(df, write_non_range_index_to_metadata,
    write_rangeindex_to_metadata, partition_cols=None):
    nmapx__rlcm = {}
    nmapx__rlcm['columns'] = []
    if partition_cols is None:
        partition_cols = []
    for col_name, mncdu__ypbs in zip(df.columns, df.data):
        if col_name in partition_cols:
            continue
        if isinstance(mncdu__ypbs, types.Array
            ) or mncdu__ypbs == boolean_array:
            rdtsw__ckun = dbe__lbh = mncdu__ypbs.dtype.name
            if dbe__lbh.startswith('datetime'):
                rdtsw__ckun = 'datetime'
        elif mncdu__ypbs == string_array_type:
            rdtsw__ckun = 'unicode'
            dbe__lbh = 'object'
        elif mncdu__ypbs == binary_array_type:
            rdtsw__ckun = 'bytes'
            dbe__lbh = 'object'
        elif isinstance(mncdu__ypbs, DecimalArrayType):
            rdtsw__ckun = dbe__lbh = 'object'
        elif isinstance(mncdu__ypbs, IntegerArrayType):
            cglfy__axkjm = mncdu__ypbs.dtype.name
            if cglfy__axkjm.startswith('int'):
                rdtsw__ckun = 'Int' + cglfy__axkjm[3:]
            elif cglfy__axkjm.startswith('uint'):
                rdtsw__ckun = 'UInt' + cglfy__axkjm[4:]
            else:
                raise BodoError(
                    'to_parquet(): unknown dtype in nullable Integer column {} {}'
                    .format(col_name, mncdu__ypbs))
            dbe__lbh = mncdu__ypbs.dtype.name
        elif mncdu__ypbs == datetime_date_array_type:
            rdtsw__ckun = 'datetime'
            dbe__lbh = 'object'
        elif isinstance(mncdu__ypbs, (StructArrayType, ArrayItemArrayType)):
            rdtsw__ckun = 'object'
            dbe__lbh = 'object'
        else:
            raise BodoError(
                'to_parquet(): unsupported column type for metadata generation : {} {}'
                .format(col_name, mncdu__ypbs))
        pbp__wdc = {'name': col_name, 'field_name': col_name, 'pandas_type':
            rdtsw__ckun, 'numpy_type': dbe__lbh, 'metadata': None}
        nmapx__rlcm['columns'].append(pbp__wdc)
    if write_non_range_index_to_metadata:
        if isinstance(df.index, MultiIndexType):
            raise BodoError('to_parquet: MultiIndex not supported yet')
        if 'none' in df.index.name:
            cnnvh__eweba = '__index_level_0__'
            artc__eopu = None
        else:
            cnnvh__eweba = '%s'
            artc__eopu = '%s'
        nmapx__rlcm['index_columns'] = [cnnvh__eweba]
        nmapx__rlcm['columns'].append({'name': artc__eopu, 'field_name':
            cnnvh__eweba, 'pandas_type': df.index.pandas_type_name,
            'numpy_type': df.index.numpy_type_name, 'metadata': None})
    elif write_rangeindex_to_metadata:
        nmapx__rlcm['index_columns'] = [{'kind': 'range', 'name': '%s',
            'start': '%d', 'stop': '%d', 'step': '%d'}]
    else:
        nmapx__rlcm['index_columns'] = []
    nmapx__rlcm['pandas_version'] = pd.__version__
    return nmapx__rlcm


@overload_method(DataFrameType, 'to_parquet', no_unliteral=True)
def to_parquet_overload(df, fname, engine='auto', compression='snappy',
    index=None, partition_cols=None, storage_options=None, _is_parallel=False):
    check_runtime_cols_unsupported(df, 'DataFrame.to_parquet()')
    check_unsupported_args('DataFrame.to_parquet', {'storage_options':
        storage_options}, {'storage_options': None}, package_name='pandas',
        module_name='IO')
    if not is_overload_none(engine) and get_overload_const_str(engine) not in (
        'auto', 'pyarrow'):
        raise BodoError('DataFrame.to_parquet(): only pyarrow engine supported'
            )
    if not is_overload_none(compression) and get_overload_const_str(compression
        ) not in {'snappy', 'gzip', 'brotli'}:
        raise BodoError('to_parquet(): Unsupported compression: ' + str(
            get_overload_const_str(compression)))
    if not is_overload_none(partition_cols):
        partition_cols = get_overload_const_list(partition_cols)
        hxz__bcu = []
        for jrm__auhf in partition_cols:
            try:
                idx = df.columns.index(jrm__auhf)
            except ValueError as zihkb__fvmm:
                raise BodoError(
                    f'Partition column {jrm__auhf} is not in dataframe')
            hxz__bcu.append(idx)
    else:
        partition_cols = None
    if not is_overload_none(index) and not is_overload_constant_bool(index):
        raise BodoError('to_parquet(): index must be a constant bool or None')
    from bodo.io.parquet_pio import parquet_write_table_cpp, parquet_write_table_partitioned_cpp
    ueevw__opruu = isinstance(df.index, bodo.hiframes.pd_index_ext.
        RangeIndexType)
    naevl__oxb = df.index is not None and (is_overload_true(_is_parallel) or
        not is_overload_true(_is_parallel) and not ueevw__opruu)
    write_non_range_index_to_metadata = is_overload_true(index
        ) or is_overload_none(index) and (not ueevw__opruu or
        is_overload_true(_is_parallel))
    write_rangeindex_to_metadata = is_overload_none(index
        ) and ueevw__opruu and not is_overload_true(_is_parallel)
    kzmf__gqmq = json.dumps(gen_pandas_parquet_metadata(df,
        write_non_range_index_to_metadata, write_rangeindex_to_metadata,
        partition_cols=partition_cols))
    if not is_overload_true(_is_parallel) and ueevw__opruu:
        kzmf__gqmq = kzmf__gqmq.replace('"%d"', '%d')
        if df.index.name == 'RangeIndexType(none)':
            kzmf__gqmq = kzmf__gqmq.replace('"%s"', '%s')
    hdfmp__vvjul = ', '.join(
        'array_to_info(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {}))'
        .format(i) for i in range(len(df.columns)))
    typ__ilqs = """def df_to_parquet(df, fname, engine='auto', compression='snappy', index=None, partition_cols=None, storage_options=None, _is_parallel=False):
"""
    if df.is_table_format:
        typ__ilqs += (
            '    table = py_table_to_cpp_table(get_dataframe_table(df), py_table_typ)\n'
            )
    else:
        typ__ilqs += '    info_list = [{}]\n'.format(hdfmp__vvjul)
        typ__ilqs += '    table = arr_info_list_to_table(info_list)\n'
    typ__ilqs += '    col_names = array_to_info(col_names_arr)\n'
    if is_overload_true(index) or is_overload_none(index) and naevl__oxb:
        typ__ilqs += """    index_col = array_to_info(index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)))
"""
        jul__wgn = True
    else:
        typ__ilqs += '    index_col = array_to_info(np.empty(0))\n'
        jul__wgn = False
    typ__ilqs += '    metadata = """' + kzmf__gqmq + '"""\n'
    typ__ilqs += '    if compression is None:\n'
    typ__ilqs += "        compression = 'none'\n"
    typ__ilqs += '    if df.index.name is not None:\n'
    typ__ilqs += '        name_ptr = df.index.name\n'
    typ__ilqs += '    else:\n'
    typ__ilqs += "        name_ptr = 'null'\n"
    typ__ilqs += f"""    bucket_region = bodo.io.fs_io.get_s3_bucket_region_njit(fname, parallel=_is_parallel)
"""
    qcqc__hwxv = None
    if partition_cols:
        qcqc__hwxv = pd.array([col_name for col_name in df.columns if 
            col_name not in partition_cols])
        wgvlg__vft = ', '.join(
            f'array_to_info(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {i}).dtype.categories.values)'
             for i in range(len(df.columns)) if isinstance(df.data[i],
            CategoricalArrayType) and i in hxz__bcu)
        if wgvlg__vft:
            typ__ilqs += '    cat_info_list = [{}]\n'.format(wgvlg__vft)
            typ__ilqs += (
                '    cat_table = arr_info_list_to_table(cat_info_list)\n')
        else:
            typ__ilqs += '    cat_table = table\n'
        typ__ilqs += (
            '    col_names_no_partitions = array_to_info(col_names_no_parts_arr)\n'
            )
        typ__ilqs += (
            f'    part_cols_idxs = np.array({hxz__bcu}, dtype=np.int32)\n')
        typ__ilqs += (
            '    parquet_write_table_partitioned_cpp(unicode_to_utf8(fname),\n'
            )
        typ__ilqs += """                            table, col_names, col_names_no_partitions, cat_table,
"""
        typ__ilqs += (
            '                            part_cols_idxs.ctypes, len(part_cols_idxs),\n'
            )
        typ__ilqs += (
            '                            unicode_to_utf8(compression),\n')
        typ__ilqs += '                            _is_parallel,\n'
        typ__ilqs += (
            '                            unicode_to_utf8(bucket_region))\n')
        typ__ilqs += '    delete_table_decref_arrays(table)\n'
        typ__ilqs += '    delete_info_decref_array(index_col)\n'
        typ__ilqs += '    delete_info_decref_array(col_names_no_partitions)\n'
        typ__ilqs += '    delete_info_decref_array(col_names)\n'
        if wgvlg__vft:
            typ__ilqs += '    delete_table_decref_arrays(cat_table)\n'
    elif write_rangeindex_to_metadata:
        typ__ilqs += '    parquet_write_table_cpp(unicode_to_utf8(fname),\n'
        typ__ilqs += (
            '                            table, col_names, index_col,\n')
        typ__ilqs += '                            ' + str(jul__wgn) + ',\n'
        typ__ilqs += '                            unicode_to_utf8(metadata),\n'
        typ__ilqs += (
            '                            unicode_to_utf8(compression),\n')
        typ__ilqs += (
            '                            _is_parallel, 1, df.index.start,\n')
        typ__ilqs += (
            '                            df.index.stop, df.index.step,\n')
        typ__ilqs += '                            unicode_to_utf8(name_ptr),\n'
        typ__ilqs += (
            '                            unicode_to_utf8(bucket_region))\n')
        typ__ilqs += '    delete_table_decref_arrays(table)\n'
        typ__ilqs += '    delete_info_decref_array(index_col)\n'
        typ__ilqs += '    delete_info_decref_array(col_names)\n'
    else:
        typ__ilqs += '    parquet_write_table_cpp(unicode_to_utf8(fname),\n'
        typ__ilqs += (
            '                            table, col_names, index_col,\n')
        typ__ilqs += '                            ' + str(jul__wgn) + ',\n'
        typ__ilqs += '                            unicode_to_utf8(metadata),\n'
        typ__ilqs += (
            '                            unicode_to_utf8(compression),\n')
        typ__ilqs += '                            _is_parallel, 0, 0, 0, 0,\n'
        typ__ilqs += '                            unicode_to_utf8(name_ptr),\n'
        typ__ilqs += (
            '                            unicode_to_utf8(bucket_region))\n')
        typ__ilqs += '    delete_table_decref_arrays(table)\n'
        typ__ilqs += '    delete_info_decref_array(index_col)\n'
        typ__ilqs += '    delete_info_decref_array(col_names)\n'
    foa__gxbh = {}
    exec(typ__ilqs, {'np': np, 'bodo': bodo, 'unicode_to_utf8':
        unicode_to_utf8, 'array_to_info': array_to_info,
        'arr_info_list_to_table': arr_info_list_to_table,
        'str_arr_from_sequence': str_arr_from_sequence,
        'parquet_write_table_cpp': parquet_write_table_cpp,
        'parquet_write_table_partitioned_cpp':
        parquet_write_table_partitioned_cpp, 'index_to_array':
        index_to_array, 'delete_info_decref_array':
        delete_info_decref_array, 'delete_table_decref_arrays':
        delete_table_decref_arrays, 'col_names_arr': pd.array(df.columns),
        'py_table_to_cpp_table': py_table_to_cpp_table, 'py_table_typ': df.
        table_type, 'get_dataframe_table': get_dataframe_table,
        'col_names_no_parts_arr': qcqc__hwxv}, foa__gxbh)
    msqgq__ttcor = foa__gxbh['df_to_parquet']
    return msqgq__ttcor


def to_sql_exception_guard(df, name, con, schema=None, if_exists='fail',
    index=True, index_label=None, chunksize=None, dtype=None, method=None,
    _is_table_create=False, _is_parallel=False):
    zayku__yux = 'all_ok'
    ikun__qgbwv = urlparse(con).scheme
    if _is_parallel and bodo.get_rank() == 0:
        jsuru__dyj = 100
        if chunksize is None:
            wqs__zzoxc = jsuru__dyj
        else:
            wqs__zzoxc = min(chunksize, jsuru__dyj)
        if _is_table_create:
            df = df.iloc[:wqs__zzoxc, :]
        else:
            df = df.iloc[wqs__zzoxc:, :]
            if len(df) == 0:
                return zayku__yux
    if ikun__qgbwv == 'snowflake':
        try:
            from snowflake.connector.pandas_tools import pd_writer
            from bodo import snowflake_sqlalchemy_compat
            if method is not None and _is_table_create and bodo.get_rank(
                ) == 0:
                import warnings
                from bodo.utils.typing import BodoWarning
                warnings.warn(BodoWarning(
                    'DataFrame.to_sql(): method argument is not supported with Snowflake. Bodo always uses snowflake.connector.pandas_tools.pd_writer to write data.'
                    ))
            method = pd_writer
            df.columns = [(ajkeq__tnmny.upper() if ajkeq__tnmny.islower() else
                ajkeq__tnmny) for ajkeq__tnmny in df.columns]
        except ImportError as zihkb__fvmm:
            zayku__yux = (
                "Snowflake Python connector not found. It can be installed by calling 'conda install -c conda-forge snowflake-connector-python' or 'pip install snowflake-connector-python'."
                )
            return zayku__yux
    try:
        df.to_sql(name, con, schema, if_exists, index, index_label,
            chunksize, dtype, method)
    except Exception as cwzji__vns:
        zayku__yux = cwzji__vns.args[0]
    return zayku__yux


@numba.njit
def to_sql_exception_guard_encaps(df, name, con, schema=None, if_exists=
    'fail', index=True, index_label=None, chunksize=None, dtype=None,
    method=None, _is_table_create=False, _is_parallel=False):
    with numba.objmode(out='unicode_type'):
        out = to_sql_exception_guard(df, name, con, schema, if_exists,
            index, index_label, chunksize, dtype, method, _is_table_create,
            _is_parallel)
    return out


@overload_method(DataFrameType, 'to_sql')
def to_sql_overload(df, name, con, schema=None, if_exists='fail', index=
    True, index_label=None, chunksize=None, dtype=None, method=None,
    _is_parallel=False):
    check_runtime_cols_unsupported(df, 'DataFrame.to_sql()')
    qjbin__nksm = dict(chunksize=chunksize)
    qrl__wxsx = dict(chunksize=None)
    check_unsupported_args('to_sql', qjbin__nksm, qrl__wxsx, package_name=
        'pandas', module_name='IO')
    if is_overload_none(schema):
        if bodo.get_rank() == 0:
            import warnings
            warnings.warn(BodoWarning(
                f'DataFrame.to_sql(): schema argument is recommended to avoid permission issues for writing to table {name}.'
                ))

    def _impl(df, name, con, schema=None, if_exists='fail', index=True,
        index_label=None, chunksize=None, dtype=None, method=None,
        _is_parallel=False):
        hca__mmfeo = bodo.libs.distributed_api.get_rank()
        zayku__yux = 'unset'
        if hca__mmfeo != 0:
            zayku__yux = bcast_scalar(zayku__yux)
        elif hca__mmfeo == 0:
            zayku__yux = to_sql_exception_guard_encaps(df, name, con,
                schema, if_exists, index, index_label, chunksize, dtype,
                method, True, _is_parallel)
            zayku__yux = bcast_scalar(zayku__yux)
        if_exists = 'append'
        if _is_parallel and zayku__yux == 'all_ok':
            zayku__yux = to_sql_exception_guard_encaps(df, name, con,
                schema, if_exists, index, index_label, chunksize, dtype,
                method, False, _is_parallel)
        if zayku__yux != 'all_ok':
            print('err_msg=', zayku__yux)
            raise ValueError('error in to_sql() operation')
    return _impl


@overload_method(DataFrameType, 'to_csv', no_unliteral=True)
def to_csv_overload(df, path_or_buf=None, sep=',', na_rep='', float_format=
    None, columns=None, header=True, index=True, index_label=None, mode='w',
    encoding=None, compression=None, quoting=None, quotechar='"',
    line_terminator=None, chunksize=None, date_format=None, doublequote=
    True, escapechar=None, decimal='.', errors='strict', storage_options=None):
    check_runtime_cols_unsupported(df, 'DataFrame.to_csv()')
    check_unsupported_args('DataFrame.to_csv', {'encoding': encoding,
        'mode': mode, 'errors': errors, 'storage_options': storage_options},
        {'encoding': None, 'mode': 'w', 'errors': 'strict',
        'storage_options': None}, package_name='pandas', module_name='IO')
    if not (is_overload_none(path_or_buf) or is_overload_constant_str(
        path_or_buf) or path_or_buf == string_type):
        raise BodoError(
            "DataFrame.to_csv(): 'path_or_buf' argument should be None or string"
            )
    if not is_overload_none(compression):
        raise BodoError(
            "DataFrame.to_csv(): 'compression' argument supports only None, which is the default in JIT code."
            )
    if is_overload_constant_str(path_or_buf):
        yxkt__ryb = get_overload_const_str(path_or_buf)
        if yxkt__ryb.endswith(('.gz', '.bz2', '.zip', '.xz')):
            import warnings
            from bodo.utils.typing import BodoWarning
            warnings.warn(BodoWarning(
                "DataFrame.to_csv(): 'compression' argument defaults to None in JIT code, which is the only supported value."
                ))
    if isinstance(columns, types.List):
        raise BodoError(
            "DataFrame.to_csv(): 'columns' argument must not be list type. Please convert to tuple type."
            )
    if is_overload_none(path_or_buf):

        def _impl(df, path_or_buf=None, sep=',', na_rep='', float_format=
            None, columns=None, header=True, index=True, index_label=None,
            mode='w', encoding=None, compression=None, quoting=None,
            quotechar='"', line_terminator=None, chunksize=None,
            date_format=None, doublequote=True, escapechar=None, decimal=
            '.', errors='strict', storage_options=None):
            with numba.objmode(D='unicode_type'):
                D = df.to_csv(path_or_buf, sep, na_rep, float_format,
                    columns, header, index, index_label, mode, encoding,
                    compression, quoting, quotechar, line_terminator,
                    chunksize, date_format, doublequote, escapechar,
                    decimal, errors, storage_options)
            return D
        return _impl

    def _impl(df, path_or_buf=None, sep=',', na_rep='', float_format=None,
        columns=None, header=True, index=True, index_label=None, mode='w',
        encoding=None, compression=None, quoting=None, quotechar='"',
        line_terminator=None, chunksize=None, date_format=None, doublequote
        =True, escapechar=None, decimal='.', errors='strict',
        storage_options=None):
        with numba.objmode(D='unicode_type'):
            D = df.to_csv(None, sep, na_rep, float_format, columns, header,
                index, index_label, mode, encoding, compression, quoting,
                quotechar, line_terminator, chunksize, date_format,
                doublequote, escapechar, decimal, errors, storage_options)
        bodo.io.fs_io.csv_write(path_or_buf, D)
    return _impl


@overload_method(DataFrameType, 'to_json', no_unliteral=True)
def to_json_overload(df, path_or_buf=None, orient='columns', date_format=
    None, double_precision=10, force_ascii=True, date_unit='ms',
    default_handler=None, lines=False, compression='infer', index=True,
    indent=None, storage_options=None):
    check_runtime_cols_unsupported(df, 'DataFrame.to_json()')
    check_unsupported_args('DataFrame.to_json', {'storage_options':
        storage_options}, {'storage_options': None}, package_name='pandas',
        module_name='IO')
    if path_or_buf is None or path_or_buf == types.none:

        def _impl(df, path_or_buf=None, orient='columns', date_format=None,
            double_precision=10, force_ascii=True, date_unit='ms',
            default_handler=None, lines=False, compression='infer', index=
            True, indent=None, storage_options=None):
            with numba.objmode(D='unicode_type'):
                D = df.to_json(path_or_buf, orient, date_format,
                    double_precision, force_ascii, date_unit,
                    default_handler, lines, compression, index, indent,
                    storage_options)
            return D
        return _impl

    def _impl(df, path_or_buf=None, orient='columns', date_format=None,
        double_precision=10, force_ascii=True, date_unit='ms',
        default_handler=None, lines=False, compression='infer', index=True,
        indent=None, storage_options=None):
        with numba.objmode(D='unicode_type'):
            D = df.to_json(None, orient, date_format, double_precision,
                force_ascii, date_unit, default_handler, lines, compression,
                index, indent, storage_options)
        iooip__htas = bodo.io.fs_io.get_s3_bucket_region_njit(path_or_buf,
            parallel=False)
        if lines and orient == 'records':
            bodo.hiframes.pd_dataframe_ext._json_write(unicode_to_utf8(
                path_or_buf), unicode_to_utf8(D), 0, len(D), False, True,
                unicode_to_utf8(iooip__htas))
            bodo.utils.utils.check_and_propagate_cpp_exception()
        else:
            bodo.hiframes.pd_dataframe_ext._json_write(unicode_to_utf8(
                path_or_buf), unicode_to_utf8(D), 0, len(D), False, False,
                unicode_to_utf8(iooip__htas))
            bodo.utils.utils.check_and_propagate_cpp_exception()
    return _impl


@overload(pd.get_dummies, inline='always', no_unliteral=True)
def get_dummies(data, prefix=None, prefix_sep='_', dummy_na=False, columns=
    None, sparse=False, drop_first=False, dtype=None):
    bcy__afbb = {'prefix': prefix, 'prefix_sep': prefix_sep, 'dummy_na':
        dummy_na, 'columns': columns, 'sparse': sparse, 'drop_first':
        drop_first, 'dtype': dtype}
    mtbf__teom = {'prefix': None, 'prefix_sep': '_', 'dummy_na': False,
        'columns': None, 'sparse': False, 'drop_first': False, 'dtype': None}
    check_unsupported_args('pandas.get_dummies', bcy__afbb, mtbf__teom,
        package_name='pandas', module_name='General')
    if not categorical_can_construct_dataframe(data):
        raise BodoError(
            'pandas.get_dummies() only support categorical data types with explicitly known categories'
            )
    typ__ilqs = """def impl(data, prefix=None, prefix_sep='_', dummy_na=False, columns=None, sparse=False, drop_first=False, dtype=None,):
"""
    if isinstance(data, SeriesType):
        gibb__egk = data.data.dtype.categories
        typ__ilqs += (
            '  data_values = bodo.hiframes.pd_series_ext.get_series_data(data)\n'
            )
    else:
        gibb__egk = data.dtype.categories
        typ__ilqs += '  data_values = data\n'
    diccg__pebv = len(gibb__egk)
    typ__ilqs += """  codes = bodo.hiframes.pd_categorical_ext.get_categorical_arr_codes(data_values)
"""
    typ__ilqs += '  numba.parfors.parfor.init_prange()\n'
    typ__ilqs += '  n = len(data_values)\n'
    for i in range(diccg__pebv):
        typ__ilqs += '  data_arr_{} = np.empty(n, np.uint8)\n'.format(i)
    typ__ilqs += '  for i in numba.parfors.parfor.internal_prange(n):\n'
    typ__ilqs += '      if bodo.libs.array_kernels.isna(data_values, i):\n'
    for jhl__fymh in range(diccg__pebv):
        typ__ilqs += '          data_arr_{}[i] = 0\n'.format(jhl__fymh)
    typ__ilqs += '      else:\n'
    for omfgl__wpz in range(diccg__pebv):
        typ__ilqs += '          data_arr_{0}[i] = codes[i] == {0}\n'.format(
            omfgl__wpz)
    hdfmp__vvjul = ', '.join(f'data_arr_{i}' for i in range(diccg__pebv))
    index = 'bodo.hiframes.pd_index_ext.init_range_index(0, n, 1, None)'
    if isinstance(gibb__egk[0], np.datetime64):
        gibb__egk = tuple(pd.Timestamp(ajkeq__tnmny) for ajkeq__tnmny in
            gibb__egk)
    elif isinstance(gibb__egk[0], np.timedelta64):
        gibb__egk = tuple(pd.Timedelta(ajkeq__tnmny) for ajkeq__tnmny in
            gibb__egk)
    return bodo.hiframes.dataframe_impl._gen_init_df(typ__ilqs, gibb__egk,
        hdfmp__vvjul, index)


def categorical_can_construct_dataframe(val):
    if isinstance(val, CategoricalArrayType):
        return val.dtype.categories is not None
    elif isinstance(val, SeriesType) and isinstance(val.data,
        CategoricalArrayType):
        return val.data.dtype.categories is not None
    return False


def handle_inplace_df_type_change(inplace, _bodo_transformed, func_name):
    if is_overload_false(_bodo_transformed
        ) and bodo.transforms.typing_pass.in_partial_typing and (
        is_overload_true(inplace) or not is_overload_constant_bool(inplace)):
        bodo.transforms.typing_pass.typing_transform_required = True
        raise Exception('DataFrame.{}(): transform necessary for inplace'.
            format(func_name))


pd_unsupported = (pd.read_pickle, pd.read_table, pd.read_fwf, pd.
    read_clipboard, pd.ExcelFile, pd.read_html, pd.read_xml, pd.read_hdf,
    pd.read_feather, pd.read_orc, pd.read_sas, pd.read_spss, pd.
    read_sql_table, pd.read_sql_query, pd.read_gbq, pd.read_stata, pd.
    ExcelWriter, pd.json_normalize, pd.melt, pd.merge_ordered, pd.factorize,
    pd.unique, pd.wide_to_long, pd.bdate_range, pd.period_range, pd.
    infer_freq, pd.interval_range, pd.eval, pd.test, pd.Grouper)
pd_util_unsupported = pd.util.hash_array, pd.util.hash_pandas_object
dataframe_unsupported = ['set_flags', 'convert_dtypes', 'bool', '__iter__',
    'items', 'iteritems', 'keys', 'iterrows', 'lookup', 'pop', 'xs', 'get',
    'add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'dot',
    'radd', 'rsub', 'rmul', 'rdiv', 'rtruediv', 'rfloordiv', 'rmod', 'rpow',
    'lt', 'gt', 'le', 'ge', 'ne', 'eq', 'combine', 'combine_first',
    'subtract', 'divide', 'multiply', 'applymap', 'agg', 'aggregate',
    'transform', 'expanding', 'ewm', 'all', 'any', 'clip', 'corrwith',
    'cummax', 'cummin', 'eval', 'kurt', 'kurtosis', 'mad', 'mode', 'rank',
    'round', 'sem', 'skew', 'value_counts', 'add_prefix', 'add_suffix',
    'align', 'at_time', 'between_time', 'equals', 'reindex', 'reindex_like',
    'rename_axis', 'set_axis', 'truncate', 'backfill', 'bfill', 'ffill',
    'interpolate', 'pad', 'droplevel', 'reorder_levels', 'nlargest',
    'nsmallest', 'swaplevel', 'stack', 'unstack', 'swapaxes', 'melt',
    'squeeze', 'to_xarray', 'T', 'transpose', 'compare', 'update', 'asfreq',
    'asof', 'slice_shift', 'tshift', 'first_valid_index',
    'last_valid_index', 'resample', 'to_period', 'to_timestamp',
    'tz_convert', 'tz_localize', 'boxplot', 'hist', 'from_dict',
    'from_records', 'to_pickle', 'to_hdf', 'to_dict', 'to_excel', 'to_html',
    'to_feather', 'to_latex', 'to_stata', 'to_gbq', 'to_records',
    'to_clipboard', 'to_markdown', 'to_xml']
dataframe_unsupported_attrs = ['at', 'attrs', 'axes', 'flags', 'style',
    'sparse']


def _install_pd_unsupported(mod_name, pd_unsupported):
    for rch__djyij in pd_unsupported:
        fname = mod_name + '.' + rch__djyij.__name__
        overload(rch__djyij, no_unliteral=True)(create_unsupported_overload
            (fname))


def _install_dataframe_unsupported():
    for kfjc__esb in dataframe_unsupported_attrs:
        mjttt__ojupi = 'DataFrame.' + kfjc__esb
        overload_attribute(DataFrameType, kfjc__esb)(
            create_unsupported_overload(mjttt__ojupi))
    for fname in dataframe_unsupported:
        mjttt__ojupi = 'DataFrame.' + fname + '()'
        overload_method(DataFrameType, fname)(create_unsupported_overload(
            mjttt__ojupi))


_install_pd_unsupported('pandas', pd_unsupported)
_install_pd_unsupported('pandas.util', pd_util_unsupported)
_install_dataframe_unsupported()
