"""Support for Pandas Groupby operations
"""
import operator
from enum import Enum
import numba
import numpy as np
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed
from numba.core.registry import CPUDispatcher
from numba.core.typing.templates import AbstractTemplate, bound_function, infer_global, signature
from numba.extending import infer, infer_getattr, intrinsic, lower_builtin, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model
import bodo
from bodo.hiframes.pd_dataframe_ext import DataFrameType
from bodo.hiframes.pd_index_ext import NumericIndexType, RangeIndexType
from bodo.hiframes.pd_multi_index_ext import MultiIndexType
from bodo.hiframes.pd_series_ext import HeterogeneousSeriesType, SeriesType
from bodo.libs.array import arr_info_list_to_table, array_to_info, delete_table, delete_table_decref_arrays, get_groupby_labels, get_null_shuffle_info, get_shuffle_info, info_from_table, info_to_array, reverse_shuffle_table, shuffle_table
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.decimal_arr_ext import Decimal128Type
from bodo.libs.int_arr_ext import IntDtype, IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.tuple_arr_ext import TupleArrayType
from bodo.utils.templates import OverloadedKeyAttributeTemplate
from bodo.utils.transform import gen_const_tup, get_call_expr_arg, get_const_func_output_type
from bodo.utils.typing import BodoError, check_unsupported_args, create_unsupported_overload, dtype_to_array_type, get_index_data_arr_types, get_index_name_types, get_literal_value, get_overload_const_bool, get_overload_const_func, get_overload_const_list, get_overload_const_str, get_overload_constant_dict, get_udf_error_msg, get_udf_out_arr_type, is_dtype_nullable, is_literal_type, is_overload_constant_bool, is_overload_constant_dict, is_overload_constant_list, is_overload_constant_str, is_overload_false, is_overload_none, is_overload_true, list_cumulative, raise_bodo_error, raise_const_error
from bodo.utils.utils import dt_err, is_expr


class DataFrameGroupByType(types.Type):

    def __init__(self, df_type, keys, selection, as_index, dropna=True,
        explicit_select=False, series_select=False):
        self.df_type = df_type
        self.keys = keys
        self.selection = selection
        self.as_index = as_index
        self.dropna = dropna
        self.explicit_select = explicit_select
        self.series_select = series_select
        super(DataFrameGroupByType, self).__init__(name=
            f'DataFrameGroupBy({df_type}, {keys}, {selection}, {as_index}, {dropna}, {explicit_select}, {series_select})'
            )

    def copy(self):
        return DataFrameGroupByType(self.df_type, self.keys, self.selection,
            self.as_index, self.dropna, self.explicit_select, self.
            series_select)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(DataFrameGroupByType)
class GroupbyModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        awui__nlxpw = [('obj', fe_type.df_type)]
        super(GroupbyModel, self).__init__(dmm, fe_type, awui__nlxpw)


make_attribute_wrapper(DataFrameGroupByType, 'obj', 'obj')


def validate_udf(func_name, func):
    if not isinstance(func, (types.functions.MakeFunctionLiteral, bodo.
        utils.typing.FunctionLiteral, types.Dispatcher, CPUDispatcher)):
        raise_const_error(
            f"Groupby.{func_name}: 'func' must be user defined function")


@intrinsic
def init_groupby(typingctx, obj_type, by_type, as_index_type=None,
    dropna_type=None):

    def codegen(context, builder, signature, args):
        nswc__cxce = args[0]
        ijnr__zrhs = signature.return_type
        mhb__oxlrs = cgutils.create_struct_proxy(ijnr__zrhs)(context, builder)
        mhb__oxlrs.obj = nswc__cxce
        context.nrt.incref(builder, signature.args[0], nswc__cxce)
        return mhb__oxlrs._getvalue()
    if is_overload_constant_list(by_type):
        keys = tuple(get_overload_const_list(by_type))
    elif is_literal_type(by_type):
        keys = get_literal_value(by_type),
    else:
        assert False, 'Reached unreachable code in init_groupby; there is an validate_groupby_spec'
    selection = list(obj_type.columns)
    for oflt__cee in keys:
        selection.remove(oflt__cee)
    if is_overload_constant_bool(as_index_type):
        as_index = is_overload_true(as_index_type)
    else:
        as_index = True
    if is_overload_constant_bool(dropna_type):
        dropna = is_overload_true(dropna_type)
    else:
        dropna = True
    ijnr__zrhs = DataFrameGroupByType(obj_type, keys, tuple(selection),
        as_index, dropna, False)
    return ijnr__zrhs(obj_type, by_type, as_index_type, dropna_type), codegen


@lower_builtin('groupby.count', types.VarArg(types.Any))
@lower_builtin('groupby.size', types.VarArg(types.Any))
@lower_builtin('groupby.apply', types.VarArg(types.Any))
@lower_builtin('groupby.agg', types.VarArg(types.Any))
def lower_groupby_count_dummy(context, builder, sig, args):
    return context.get_constant_null(sig.return_type)


@infer
class StaticGetItemDataFrameGroupBy(AbstractTemplate):
    key = 'static_getitem'

    def generic(self, args, kws):
        grpby, tdxqw__lytb = args
        if isinstance(grpby, DataFrameGroupByType):
            series_select = False
            if isinstance(tdxqw__lytb, (tuple, list)):
                if len(set(tdxqw__lytb).difference(set(grpby.df_type.columns))
                    ) > 0:
                    raise_const_error(
                        'groupby: selected column {} not found in dataframe'
                        .format(set(tdxqw__lytb).difference(set(grpby.
                        df_type.columns))))
                selection = tdxqw__lytb
            else:
                if tdxqw__lytb not in grpby.df_type.columns:
                    raise_const_error(
                        'groupby: selected column {} not found in dataframe'
                        .format(tdxqw__lytb))
                selection = tdxqw__lytb,
                series_select = True
            sena__eivb = DataFrameGroupByType(grpby.df_type, grpby.keys,
                selection, grpby.as_index, grpby.dropna, True, series_select)
            return signature(sena__eivb, *args)


@infer_global(operator.getitem)
class GetItemDataFrameGroupBy(AbstractTemplate):

    def generic(self, args, kws):
        grpby, tdxqw__lytb = args
        if isinstance(grpby, DataFrameGroupByType) and is_literal_type(
            tdxqw__lytb):
            sena__eivb = StaticGetItemDataFrameGroupBy.generic(self, (grpby,
                get_literal_value(tdxqw__lytb)), {}).return_type
            return signature(sena__eivb, *args)


GetItemDataFrameGroupBy.prefer_literal = True


@lower_builtin('static_getitem', DataFrameGroupByType, types.Any)
@lower_builtin(operator.getitem, DataFrameGroupByType, types.Any)
def static_getitem_df_groupby(context, builder, sig, args):
    return impl_ret_borrowed(context, builder, sig.return_type, args[0])


def get_groupby_output_dtype(arr_type, func_name, index_type=None):
    ujzn__kvoni = arr_type == ArrayItemArrayType(string_array_type)
    rcg__xsydn = arr_type.dtype
    if isinstance(rcg__xsydn, bodo.hiframes.datetime_timedelta_ext.
        DatetimeTimeDeltaType):
        raise BodoError(
            f"""column type of {rcg__xsydn} is not supported in groupby built-in function {func_name}.
{dt_err}"""
            )
    if func_name == 'median' and not isinstance(rcg__xsydn, (Decimal128Type,
        types.Float, types.Integer)):
        return (None,
            'For median, only column of integer, float or Decimal type are allowed'
            )
    if func_name in ('first', 'last', 'sum', 'prod', 'min', 'max', 'count',
        'nunique', 'head') and isinstance(arr_type, (TupleArrayType,
        ArrayItemArrayType)):
        return (None,
            f'column type of list/tuple of {rcg__xsydn} is not supported in groupby built-in function {func_name}'
            )
    if func_name in {'median', 'mean', 'var', 'std'} and isinstance(rcg__xsydn,
        (Decimal128Type, types.Integer, types.Float)):
        return dtype_to_array_type(types.float64), 'ok'
    if not isinstance(rcg__xsydn, (types.Integer, types.Float, types.Boolean)):
        if ujzn__kvoni or rcg__xsydn == types.unicode_type:
            if func_name not in {'count', 'nunique', 'min', 'max', 'sum',
                'first', 'last', 'head'}:
                return (None,
                    f'column type of strings or list of strings is not supported in groupby built-in function {func_name}'
                    )
        else:
            if isinstance(rcg__xsydn, bodo.PDCategoricalDtype):
                if func_name in ('min', 'max') and not rcg__xsydn.ordered:
                    return (None,
                        f'categorical column must be ordered in groupby built-in function {func_name}'
                        )
            if func_name not in {'count', 'nunique', 'min', 'max', 'first',
                'last', 'head'}:
                return (None,
                    f'column type of {rcg__xsydn} is not supported in groupby built-in function {func_name}'
                    )
    if isinstance(rcg__xsydn, types.Boolean) and func_name in {'cumsum',
        'sum', 'mean', 'std', 'var'}:
        return (None,
            f'groupby built-in functions {func_name} does not support boolean column'
            )
    if func_name in {'idxmin', 'idxmax'}:
        return dtype_to_array_type(get_index_data_arr_types(index_type)[0].
            dtype), 'ok'
    if func_name in {'count', 'nunique'}:
        return dtype_to_array_type(types.int64), 'ok'
    else:
        return arr_type, 'ok'


def get_pivot_output_dtype(arr_type, func_name, index_type=None):
    rcg__xsydn = arr_type.dtype
    if func_name in {'count'}:
        return IntDtype(types.int64)
    if func_name in {'sum', 'prod', 'min', 'max'}:
        if func_name in {'sum', 'prod'} and not isinstance(rcg__xsydn, (
            types.Integer, types.Float)):
            raise BodoError(
                'pivot_table(): sum and prod operations require integer or float input'
                )
        if isinstance(rcg__xsydn, types.Integer):
            return IntDtype(rcg__xsydn)
        return rcg__xsydn
    if func_name in {'mean', 'var', 'std'}:
        return types.float64
    raise BodoError('invalid pivot operation')


def check_args_kwargs(func_name, len_args, args, kws):
    if len(kws) > 0:
        xjbi__pyc = list(kws.keys())[0]
        raise BodoError(
            f"Groupby.{func_name}() got an unexpected keyword argument '{xjbi__pyc}'."
            )
    elif len(args) > len_args:
        raise BodoError(
            f'Groupby.{func_name}() takes {len_args + 1} positional argument but {len(args)} were given.'
            )


class ColumnType(Enum):
    KeyColumn = 0
    NumericalColumn = 1
    NonNumericalColumn = 2


def get_keys_not_as_index(grp, out_columns, out_data, out_column_type,
    multi_level_names=False):
    for oflt__cee in grp.keys:
        if multi_level_names:
            nte__sng = oflt__cee, ''
        else:
            nte__sng = oflt__cee
        mhh__vcz = grp.df_type.columns.index(oflt__cee)
        data = grp.df_type.data[mhh__vcz]
        out_columns.append(nte__sng)
        out_data.append(data)
        out_column_type.append(ColumnType.KeyColumn.value)


def get_agg_typ(grp, args, func_name, typing_context, target_context, func=
    None, kws=None):
    index = RangeIndexType(types.none)
    out_data = []
    out_columns = []
    out_column_type = []
    if func_name == 'head':
        grp.dropna = False
        grp.as_index = True
    if not grp.as_index:
        get_keys_not_as_index(grp, out_columns, out_data, out_column_type)
    elif func_name == 'head':
        if grp.df_type.index == index:
            index = NumericIndexType(types.int64, types.none)
        else:
            index = grp.df_type.index
    elif len(grp.keys) > 1:
        tnv__qeece = tuple(grp.df_type.columns.index(grp.keys[lko__kwjd]) for
            lko__kwjd in range(len(grp.keys)))
        ikwml__qgd = tuple(grp.df_type.data[mhh__vcz] for mhh__vcz in
            tnv__qeece)
        index = MultiIndexType(ikwml__qgd, tuple(types.StringLiteral(
            oflt__cee) for oflt__cee in grp.keys))
    else:
        mhh__vcz = grp.df_type.columns.index(grp.keys[0])
        index = bodo.hiframes.pd_index_ext.array_type_to_index(grp.df_type.
            data[mhh__vcz], types.StringLiteral(grp.keys[0]))
    lky__zieuz = {}
    ktqph__ydgu = []
    if func_name in ('size', 'count'):
        kws = dict(kws) if kws else {}
        check_args_kwargs(func_name, 0, args, kws)
    if func_name == 'size':
        out_data.append(types.Array(types.int64, 1, 'C'))
        out_columns.append('size')
        lky__zieuz[None, 'size'] = 'size'
    else:
        columns = (grp.selection if func_name != 'head' or grp.
            explicit_select else grp.df_type.columns)
        for psgb__vbgre in columns:
            mhh__vcz = grp.df_type.columns.index(psgb__vbgre)
            data = grp.df_type.data[mhh__vcz]
            tslw__ybi = ColumnType.NonNumericalColumn.value
            if isinstance(data, (types.Array, IntegerArrayType)
                ) and isinstance(data.dtype, (types.Integer, types.Float)):
                tslw__ybi = ColumnType.NumericalColumn.value
            if func_name == 'agg':
                try:
                    dvck__ykqip = SeriesType(data.dtype, data, None,
                        string_type)
                    tpp__rou = get_const_func_output_type(func, (
                        dvck__ykqip,), {}, typing_context, target_context)
                    if tpp__rou != ArrayItemArrayType(string_array_type):
                        tpp__rou = dtype_to_array_type(tpp__rou)
                    err_msg = 'ok'
                except:
                    raise_bodo_error(
                        'Groupy.agg()/Groupy.aggregate(): column {col} of type {type} is unsupported/not a valid input type for user defined function'
                        .format(col=psgb__vbgre, type=data.dtype))
            else:
                if func_name in ('first', 'last', 'min', 'max'):
                    kws = dict(kws) if kws else {}
                    xjdk__mffp = args[0] if len(args) > 0 else kws.pop(
                        'numeric_only', False)
                    svkt__evvcf = args[1] if len(args) > 1 else kws.pop(
                        'min_count', -1)
                    goz__hzh = dict(numeric_only=xjdk__mffp, min_count=
                        svkt__evvcf)
                    bbq__epg = dict(numeric_only=False, min_count=-1)
                    check_unsupported_args(f'Groupby.{func_name}', goz__hzh,
                        bbq__epg, package_name='pandas', module_name='GroupBy')
                elif func_name in ('sum', 'prod'):
                    kws = dict(kws) if kws else {}
                    xjdk__mffp = args[0] if len(args) > 0 else kws.pop(
                        'numeric_only', True)
                    svkt__evvcf = args[1] if len(args) > 1 else kws.pop(
                        'min_count', 0)
                    goz__hzh = dict(numeric_only=xjdk__mffp, min_count=
                        svkt__evvcf)
                    bbq__epg = dict(numeric_only=True, min_count=0)
                    check_unsupported_args(f'Groupby.{func_name}', goz__hzh,
                        bbq__epg, package_name='pandas', module_name='GroupBy')
                elif func_name in ('mean', 'median'):
                    kws = dict(kws) if kws else {}
                    xjdk__mffp = args[0] if len(args) > 0 else kws.pop(
                        'numeric_only', True)
                    goz__hzh = dict(numeric_only=xjdk__mffp)
                    bbq__epg = dict(numeric_only=True)
                    check_unsupported_args(f'Groupby.{func_name}', goz__hzh,
                        bbq__epg, package_name='pandas', module_name='GroupBy')
                elif func_name in ('idxmin', 'idxmax'):
                    kws = dict(kws) if kws else {}
                    zlnmo__jmdom = args[0] if len(args) > 0 else kws.pop('axis'
                        , 0)
                    avgcz__uiete = args[1] if len(args) > 1 else kws.pop(
                        'skipna', True)
                    goz__hzh = dict(axis=zlnmo__jmdom, skipna=avgcz__uiete)
                    bbq__epg = dict(axis=0, skipna=True)
                    check_unsupported_args(f'Groupby.{func_name}', goz__hzh,
                        bbq__epg, package_name='pandas', module_name='GroupBy')
                elif func_name in ('var', 'std'):
                    kws = dict(kws) if kws else {}
                    lijd__vjar = args[0] if len(args) > 0 else kws.pop('ddof',
                        1)
                    goz__hzh = dict(ddof=lijd__vjar)
                    bbq__epg = dict(ddof=1)
                    check_unsupported_args(f'Groupby.{func_name}', goz__hzh,
                        bbq__epg, package_name='pandas', module_name='GroupBy')
                elif func_name == 'nunique':
                    kws = dict(kws) if kws else {}
                    dropna = args[0] if len(args) > 0 else kws.pop('dropna', 1)
                    check_args_kwargs(func_name, 1, args, kws)
                elif func_name == 'head':
                    if len(args) == 0:
                        kws.pop('n', None)
                tpp__rou, err_msg = get_groupby_output_dtype(data,
                    func_name, grp.df_type.index)
            if err_msg == 'ok':
                qgul__jfrb = tpp__rou
                out_data.append(qgul__jfrb)
                out_columns.append(psgb__vbgre)
                if func_name == 'agg':
                    cjzxw__guf = bodo.ir.aggregate._get_udf_name(bodo.ir.
                        aggregate._get_const_agg_func(func, None))
                    lky__zieuz[psgb__vbgre, cjzxw__guf] = psgb__vbgre
                else:
                    lky__zieuz[psgb__vbgre, func_name] = psgb__vbgre
                out_column_type.append(tslw__ybi)
            else:
                ktqph__ydgu.append(err_msg)
    if func_name == 'sum':
        vxyd__zgra = any([(aqpu__erui == ColumnType.NumericalColumn.value) for
            aqpu__erui in out_column_type])
        if vxyd__zgra:
            out_data = [aqpu__erui for aqpu__erui, ygoac__gjzb in zip(
                out_data, out_column_type) if ygoac__gjzb != ColumnType.
                NonNumericalColumn.value]
            out_columns = [aqpu__erui for aqpu__erui, ygoac__gjzb in zip(
                out_columns, out_column_type) if ygoac__gjzb != ColumnType.
                NonNumericalColumn.value]
            lky__zieuz = {}
            for psgb__vbgre in out_columns:
                if grp.as_index is False and psgb__vbgre in grp.keys:
                    continue
                lky__zieuz[psgb__vbgre, func_name] = psgb__vbgre
    ceha__qkblr = len(ktqph__ydgu)
    if len(out_data) == 0:
        if ceha__qkblr == 0:
            raise BodoError('No columns in output.')
        else:
            raise BodoError(
                'No columns in output. {} column{} dropped for following reasons: {}'
                .format(ceha__qkblr, ' was' if ceha__qkblr == 1 else
                's were', ','.join(ktqph__ydgu)))
    dts__kvzrv = DataFrameType(tuple(out_data), index, tuple(out_columns))
    if (len(grp.selection) == 1 and grp.series_select and grp.as_index or 
        func_name == 'size' and grp.as_index):
        if isinstance(out_data[0], IntegerArrayType):
            cqduk__ulmf = IntDtype(out_data[0].dtype)
        else:
            cqduk__ulmf = out_data[0].dtype
        igugr__ghj = (types.none if func_name == 'size' else types.
            StringLiteral(grp.selection[0]))
        dts__kvzrv = SeriesType(cqduk__ulmf, index=index, name_typ=igugr__ghj)
    return signature(dts__kvzrv, *args), lky__zieuz


def get_agg_funcname_and_outtyp(grp, col, f_val, typing_context, target_context
    ):
    aij__iqupj = True
    if isinstance(f_val, str):
        aij__iqupj = False
        ppzx__vfof = f_val
    elif is_overload_constant_str(f_val):
        aij__iqupj = False
        ppzx__vfof = get_overload_const_str(f_val)
    elif bodo.utils.typing.is_builtin_function(f_val):
        aij__iqupj = False
        ppzx__vfof = bodo.utils.typing.get_builtin_function_name(f_val)
    if not aij__iqupj:
        if ppzx__vfof not in bodo.ir.aggregate.supported_agg_funcs[:-1]:
            raise BodoError(f'unsupported aggregate function {ppzx__vfof}')
        sena__eivb = DataFrameGroupByType(grp.df_type, grp.keys, (col,),
            grp.as_index, grp.dropna, True, True)
        out_tp = get_agg_typ(sena__eivb, (), ppzx__vfof, typing_context,
            target_context)[0].return_type
    else:
        if is_expr(f_val, 'make_function'):
            kqg__yii = types.functions.MakeFunctionLiteral(f_val)
        else:
            kqg__yii = f_val
        validate_udf('agg', kqg__yii)
        func = get_overload_const_func(kqg__yii, None)
        fmoa__gff = func.code if hasattr(func, 'code') else func.__code__
        ppzx__vfof = fmoa__gff.co_name
        sena__eivb = DataFrameGroupByType(grp.df_type, grp.keys, (col,),
            grp.as_index, grp.dropna, True, True)
        out_tp = get_agg_typ(sena__eivb, (), 'agg', typing_context,
            target_context, kqg__yii)[0].return_type
    return ppzx__vfof, out_tp


def resolve_agg(grp, args, kws, typing_context, target_context):
    func = get_call_expr_arg('agg', args, dict(kws), 0, 'func', default=
        types.none)
    qll__vdp = kws and all(isinstance(nqml__qnsmu, types.Tuple) and len(
        nqml__qnsmu) == 2 for nqml__qnsmu in kws.values())
    if is_overload_none(func) and not qll__vdp:
        raise_bodo_error("Groupby.agg()/aggregate(): Must provide 'func'")
    if len(args) > 1 or kws and not qll__vdp:
        raise_bodo_error(
            'Groupby.agg()/aggregate(): passing extra arguments to functions not supported yet.'
            )
    ztnwf__bnxdl = False

    def _append_out_type(grp, out_data, out_tp):
        if grp.as_index is False:
            out_data.append(out_tp.data[len(grp.keys)])
        else:
            out_data.append(out_tp.data)
    if qll__vdp or is_overload_constant_dict(func):
        if qll__vdp:
            etq__xqe = [get_literal_value(qjain__hqz) for qjain__hqz,
                jpnlj__vwp in kws.values()]
            ziu__cpp = [get_literal_value(cnm__ewxs) for jpnlj__vwp,
                cnm__ewxs in kws.values()]
        else:
            hwvle__ttmvn = get_overload_constant_dict(func)
            etq__xqe = tuple(hwvle__ttmvn.keys())
            ziu__cpp = tuple(hwvle__ttmvn.values())
        if 'head' in ziu__cpp:
            raise BodoError(
                'Groupby.agg()/aggregate(): head cannot be mixed with other groupby operations.'
                )
        if any(psgb__vbgre not in grp.selection and psgb__vbgre not in grp.
            keys for psgb__vbgre in etq__xqe):
            raise_const_error(
                f'Selected column names {etq__xqe} not all available in dataframe column names {grp.selection}'
                )
        multi_level_names = any(isinstance(f_val, (tuple, list)) for f_val in
            ziu__cpp)
        if qll__vdp and multi_level_names:
            raise_bodo_error(
                'Groupby.agg()/aggregate(): cannot pass multiple functions in a single pd.NamedAgg()'
                )
        lky__zieuz = {}
        out_columns = []
        out_data = []
        out_column_type = []
        ihpw__ijm = []
        if not grp.as_index:
            get_keys_not_as_index(grp, out_columns, out_data,
                out_column_type, multi_level_names=multi_level_names)
        for xwh__chla, f_val in zip(etq__xqe, ziu__cpp):
            if isinstance(f_val, (tuple, list)):
                hnwsk__ymfnc = 0
                for kqg__yii in f_val:
                    ppzx__vfof, out_tp = get_agg_funcname_and_outtyp(grp,
                        xwh__chla, kqg__yii, typing_context, target_context)
                    ztnwf__bnxdl = ppzx__vfof in list_cumulative
                    if ppzx__vfof == '<lambda>' and len(f_val) > 1:
                        ppzx__vfof = '<lambda_' + str(hnwsk__ymfnc) + '>'
                        hnwsk__ymfnc += 1
                    out_columns.append((xwh__chla, ppzx__vfof))
                    lky__zieuz[xwh__chla, ppzx__vfof] = xwh__chla, ppzx__vfof
                    _append_out_type(grp, out_data, out_tp)
            else:
                ppzx__vfof, out_tp = get_agg_funcname_and_outtyp(grp,
                    xwh__chla, f_val, typing_context, target_context)
                ztnwf__bnxdl = ppzx__vfof in list_cumulative
                if multi_level_names:
                    out_columns.append((xwh__chla, ppzx__vfof))
                    lky__zieuz[xwh__chla, ppzx__vfof] = xwh__chla, ppzx__vfof
                elif not qll__vdp:
                    out_columns.append(xwh__chla)
                    lky__zieuz[xwh__chla, ppzx__vfof] = xwh__chla
                elif qll__vdp:
                    ihpw__ijm.append(ppzx__vfof)
                _append_out_type(grp, out_data, out_tp)
        if qll__vdp:
            for lko__kwjd, ngd__epkpd in enumerate(kws.keys()):
                out_columns.append(ngd__epkpd)
                lky__zieuz[etq__xqe[lko__kwjd], ihpw__ijm[lko__kwjd]
                    ] = ngd__epkpd
        if ztnwf__bnxdl:
            index = grp.df_type.index
        else:
            index = out_tp.index
        dts__kvzrv = DataFrameType(tuple(out_data), index, tuple(out_columns))
        return signature(dts__kvzrv, *args), lky__zieuz
    if isinstance(func, types.BaseTuple) and not isinstance(func, types.
        LiteralStrKeyDict):
        if not (len(grp.selection) == 1 and grp.explicit_select):
            raise_bodo_error(
                'Groupby.agg()/aggregate(): must select exactly one column when more than one functions supplied'
                )
        assert len(func) > 0
        out_data = []
        out_columns = []
        out_column_type = []
        hnwsk__ymfnc = 0
        if not grp.as_index:
            get_keys_not_as_index(grp, out_columns, out_data, out_column_type)
        lky__zieuz = {}
        zxnn__dds = grp.selection[0]
        for f_val in func.types:
            ppzx__vfof, out_tp = get_agg_funcname_and_outtyp(grp, zxnn__dds,
                f_val, typing_context, target_context)
            ztnwf__bnxdl = ppzx__vfof in list_cumulative
            if ppzx__vfof == '<lambda>':
                ppzx__vfof = '<lambda_' + str(hnwsk__ymfnc) + '>'
                hnwsk__ymfnc += 1
            out_columns.append(ppzx__vfof)
            lky__zieuz[zxnn__dds, ppzx__vfof] = ppzx__vfof
            _append_out_type(grp, out_data, out_tp)
        if ztnwf__bnxdl:
            index = grp.df_type.index
        else:
            index = out_tp.index
        dts__kvzrv = DataFrameType(tuple(out_data), index, tuple(out_columns))
        return signature(dts__kvzrv, *args), lky__zieuz
    ppzx__vfof = ''
    if types.unliteral(func) == types.unicode_type:
        ppzx__vfof = get_overload_const_str(func)
    if bodo.utils.typing.is_builtin_function(func):
        ppzx__vfof = bodo.utils.typing.get_builtin_function_name(func)
    if ppzx__vfof:
        args = args[1:]
        kws.pop('func', None)
        return get_agg_typ(grp, args, ppzx__vfof, typing_context, kws)
    validate_udf('agg', func)
    return get_agg_typ(grp, args, 'agg', typing_context, target_context, func)


def resolve_transformative(grp, args, kws, msg, name_operation):
    index = grp.df_type.index
    out_columns = []
    out_data = []
    if name_operation in list_cumulative:
        kws = dict(kws) if kws else {}
        zlnmo__jmdom = args[0] if len(args) > 0 else kws.pop('axis', 0)
        xjdk__mffp = args[1] if len(args) > 1 else kws.pop('numeric_only', 
            False)
        avgcz__uiete = args[2] if len(args) > 2 else kws.pop('skipna', 1)
        goz__hzh = dict(axis=zlnmo__jmdom, numeric_only=xjdk__mffp)
        bbq__epg = dict(axis=0, numeric_only=False)
        check_unsupported_args(f'Groupby.{name_operation}', goz__hzh,
            bbq__epg, package_name='pandas', module_name='GroupBy')
        check_args_kwargs(name_operation, 3, args, kws)
    elif name_operation == 'shift':
        glw__gativ = args[0] if len(args) > 0 else kws.pop('periods', 1)
        ycb__oaqg = args[1] if len(args) > 1 else kws.pop('freq', None)
        zlnmo__jmdom = args[2] if len(args) > 2 else kws.pop('axis', 0)
        ljqh__eddq = args[3] if len(args) > 3 else kws.pop('fill_value', None)
        goz__hzh = dict(freq=ycb__oaqg, axis=zlnmo__jmdom, fill_value=
            ljqh__eddq)
        bbq__epg = dict(freq=None, axis=0, fill_value=None)
        check_unsupported_args(f'Groupby.{name_operation}', goz__hzh,
            bbq__epg, package_name='pandas', module_name='GroupBy')
        check_args_kwargs(name_operation, 4, args, kws)
    elif name_operation == 'transform':
        kws = dict(kws)
        nqnka__kynio = args[0] if len(args) > 0 else kws.pop('func', None)
        iel__hod = kws.pop('engine', None)
        qgjj__jqugc = kws.pop('engine_kwargs', None)
        goz__hzh = dict(engine=iel__hod, engine_kwargs=qgjj__jqugc)
        bbq__epg = dict(engine=None, engine_kwargs=None)
        check_unsupported_args(f'Groupby.transform', goz__hzh, bbq__epg,
            package_name='pandas', module_name='GroupBy')
    lky__zieuz = {}
    for psgb__vbgre in grp.selection:
        out_columns.append(psgb__vbgre)
        lky__zieuz[psgb__vbgre, name_operation] = psgb__vbgre
        mhh__vcz = grp.df_type.columns.index(psgb__vbgre)
        data = grp.df_type.data[mhh__vcz]
        if name_operation == 'cumprod':
            if not isinstance(data.dtype, (types.Integer, types.Float)):
                raise BodoError(msg)
        if name_operation == 'cumsum':
            if data.dtype != types.unicode_type and data != ArrayItemArrayType(
                string_array_type) and not isinstance(data.dtype, (types.
                Integer, types.Float)):
                raise BodoError(msg)
        if name_operation in ('cummin', 'cummax'):
            if not isinstance(data.dtype, types.Integer
                ) and not is_dtype_nullable(data.dtype):
                raise BodoError(msg)
        if name_operation == 'shift':
            if isinstance(data, (TupleArrayType, ArrayItemArrayType)):
                raise BodoError(msg)
            if isinstance(data.dtype, bodo.hiframes.datetime_timedelta_ext.
                DatetimeTimeDeltaType):
                raise BodoError(
                    f"""column type of {data.dtype} is not supported in groupby built-in function shift.
{dt_err}"""
                    )
        if name_operation == 'transform':
            tpp__rou, err_msg = get_groupby_output_dtype(data,
                get_literal_value(nqnka__kynio), grp.df_type.index)
            if err_msg == 'ok':
                data = tpp__rou
            else:
                raise BodoError(
                    f'column type of {data.dtype} is not supported by {args[0]} yet.\n'
                    )
        out_data.append(data)
    if len(out_data) == 0:
        raise BodoError('No columns in output.')
    dts__kvzrv = DataFrameType(tuple(out_data), index, tuple(out_columns))
    if len(grp.selection) == 1 and grp.series_select and grp.as_index:
        dts__kvzrv = SeriesType(out_data[0].dtype, data=out_data[0], index=
            index, name_typ=types.StringLiteral(grp.selection[0]))
    return signature(dts__kvzrv, *args), lky__zieuz


def resolve_gb(grp, args, kws, func_name, typing_context, target_context,
    err_msg=''):
    if func_name in set(list_cumulative) | {'shift', 'transform'}:
        return resolve_transformative(grp, args, kws, err_msg, func_name)
    elif func_name in {'agg', 'aggregate'}:
        return resolve_agg(grp, args, kws, typing_context, target_context)
    else:
        return get_agg_typ(grp, args, func_name, typing_context,
            target_context, kws=kws)


@infer_getattr
class DataframeGroupByAttribute(OverloadedKeyAttributeTemplate):
    key = DataFrameGroupByType
    _attr_set = None

    @bound_function('groupby.agg', no_unliteral=True)
    def resolve_agg(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'agg', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.aggregate', no_unliteral=True)
    def resolve_aggregate(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'agg', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.sum', no_unliteral=True)
    def resolve_sum(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'sum', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.count', no_unliteral=True)
    def resolve_count(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'count', self.context, numba.core
            .registry.cpu_target.target_context)[0]

    @bound_function('groupby.nunique', no_unliteral=True)
    def resolve_nunique(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'nunique', self.context, numba.
            core.registry.cpu_target.target_context)[0]

    @bound_function('groupby.median', no_unliteral=True)
    def resolve_median(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'median', self.context, numba.
            core.registry.cpu_target.target_context)[0]

    @bound_function('groupby.mean', no_unliteral=True)
    def resolve_mean(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'mean', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.min', no_unliteral=True)
    def resolve_min(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'min', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.max', no_unliteral=True)
    def resolve_max(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'max', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.prod', no_unliteral=True)
    def resolve_prod(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'prod', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.var', no_unliteral=True)
    def resolve_var(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'var', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.std', no_unliteral=True)
    def resolve_std(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'std', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.first', no_unliteral=True)
    def resolve_first(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'first', self.context, numba.core
            .registry.cpu_target.target_context)[0]

    @bound_function('groupby.last', no_unliteral=True)
    def resolve_last(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'last', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.idxmin', no_unliteral=True)
    def resolve_idxmin(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'idxmin', self.context, numba.
            core.registry.cpu_target.target_context)[0]

    @bound_function('groupby.idxmax', no_unliteral=True)
    def resolve_idxmax(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'idxmax', self.context, numba.
            core.registry.cpu_target.target_context)[0]

    @bound_function('groupby.size', no_unliteral=True)
    def resolve_size(self, grp, args, kws):
        return resolve_gb(grp, args, kws, 'size', self.context, numba.core.
            registry.cpu_target.target_context)[0]

    @bound_function('groupby.cumsum', no_unliteral=True)
    def resolve_cumsum(self, grp, args, kws):
        msg = (
            'Groupby.cumsum() only supports columns of types integer, float, string or liststring'
            )
        return resolve_gb(grp, args, kws, 'cumsum', self.context, numba.
            core.registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.cumprod', no_unliteral=True)
    def resolve_cumprod(self, grp, args, kws):
        msg = (
            'Groupby.cumprod() only supports columns of types integer and float'
            )
        return resolve_gb(grp, args, kws, 'cumprod', self.context, numba.
            core.registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.cummin', no_unliteral=True)
    def resolve_cummin(self, grp, args, kws):
        msg = (
            'Groupby.cummin() only supports columns of types integer, float, string, liststring, date, datetime or timedelta'
            )
        return resolve_gb(grp, args, kws, 'cummin', self.context, numba.
            core.registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.cummax', no_unliteral=True)
    def resolve_cummax(self, grp, args, kws):
        msg = (
            'Groupby.cummax() only supports columns of types integer, float, string, liststring, date, datetime or timedelta'
            )
        return resolve_gb(grp, args, kws, 'cummax', self.context, numba.
            core.registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.shift', no_unliteral=True)
    def resolve_shift(self, grp, args, kws):
        msg = (
            'Column type of list/tuple is not supported in groupby built-in function shift'
            )
        return resolve_gb(grp, args, kws, 'shift', self.context, numba.core
            .registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.pipe', no_unliteral=True)
    def resolve_pipe(self, grp, args, kws):
        return resolve_obj_pipe(self, grp, args, kws, 'GroupBy')

    @bound_function('groupby.transform', no_unliteral=True)
    def resolve_transform(self, grp, args, kws):
        msg = (
            'Groupby.transform() only supports sum, count, min, max, mean, and std operations'
            )
        return resolve_gb(grp, args, kws, 'transform', self.context, numba.
            core.registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.head', no_unliteral=True)
    def resolve_head(self, grp, args, kws):
        msg = 'Unsupported Gropupby head operation.\n'
        return resolve_gb(grp, args, kws, 'head', self.context, numba.core.
            registry.cpu_target.target_context, err_msg=msg)[0]

    @bound_function('groupby.apply', no_unliteral=True)
    def resolve_apply(self, grp, args, kws):
        kws = dict(kws)
        func = args[0] if len(args) > 0 else kws.pop('func', None)
        f_args = tuple(args[1:]) if len(args) > 0 else ()
        xrbl__vib = _get_groupby_apply_udf_out_type(func, grp, f_args, kws,
            self.context, numba.core.registry.cpu_target.target_context)
        cgowf__qftr = isinstance(xrbl__vib, (SeriesType,
            HeterogeneousSeriesType)
            ) and xrbl__vib.const_info is not None or not isinstance(xrbl__vib,
            (SeriesType, DataFrameType))
        if cgowf__qftr:
            out_data = []
            out_columns = []
            out_column_type = []
            if not grp.as_index:
                get_keys_not_as_index(grp, out_columns, out_data,
                    out_column_type)
                lcr__omms = NumericIndexType(types.int64, types.none)
            elif len(grp.keys) > 1:
                tnv__qeece = tuple(grp.df_type.columns.index(grp.keys[
                    lko__kwjd]) for lko__kwjd in range(len(grp.keys)))
                ikwml__qgd = tuple(grp.df_type.data[mhh__vcz] for mhh__vcz in
                    tnv__qeece)
                lcr__omms = MultiIndexType(ikwml__qgd, tuple(types.literal(
                    oflt__cee) for oflt__cee in grp.keys))
            else:
                mhh__vcz = grp.df_type.columns.index(grp.keys[0])
                lcr__omms = bodo.hiframes.pd_index_ext.array_type_to_index(grp
                    .df_type.data[mhh__vcz], types.literal(grp.keys[0]))
            out_data = tuple(out_data)
            out_columns = tuple(out_columns)
        else:
            cdwei__eao = tuple(grp.df_type.data[grp.df_type.columns.index(
                psgb__vbgre)] for psgb__vbgre in grp.keys)
            jeqa__bmw = tuple(types.literal(nqml__qnsmu) for nqml__qnsmu in
                grp.keys) + get_index_name_types(xrbl__vib.index)
            if not grp.as_index:
                cdwei__eao = types.Array(types.int64, 1, 'C'),
                jeqa__bmw = (types.none,) + get_index_name_types(xrbl__vib.
                    index)
            lcr__omms = MultiIndexType(cdwei__eao +
                get_index_data_arr_types(xrbl__vib.index), jeqa__bmw)
        if cgowf__qftr:
            if isinstance(xrbl__vib, HeterogeneousSeriesType):
                jpnlj__vwp, bgj__vrkc = xrbl__vib.const_info
                hlrdc__njfj = tuple(dtype_to_array_type(ebne__kmqg) for
                    ebne__kmqg in xrbl__vib.data.types)
                dxkm__nykcc = DataFrameType(out_data + hlrdc__njfj,
                    lcr__omms, out_columns + bgj__vrkc)
            elif isinstance(xrbl__vib, SeriesType):
                enpx__upat, bgj__vrkc = xrbl__vib.const_info
                hlrdc__njfj = tuple(dtype_to_array_type(xrbl__vib.dtype) for
                    jpnlj__vwp in range(enpx__upat))
                dxkm__nykcc = DataFrameType(out_data + hlrdc__njfj,
                    lcr__omms, out_columns + bgj__vrkc)
            else:
                gqocq__iuakh = get_udf_out_arr_type(xrbl__vib)
                if not grp.as_index:
                    dxkm__nykcc = DataFrameType(out_data + (gqocq__iuakh,),
                        lcr__omms, out_columns + ('',))
                else:
                    dxkm__nykcc = SeriesType(gqocq__iuakh.dtype,
                        gqocq__iuakh, lcr__omms, None)
        elif isinstance(xrbl__vib, SeriesType):
            dxkm__nykcc = SeriesType(xrbl__vib.dtype, xrbl__vib.data,
                lcr__omms, xrbl__vib.name_typ)
        else:
            dxkm__nykcc = DataFrameType(xrbl__vib.data, lcr__omms,
                xrbl__vib.columns)
        wwmk__dcbn = gen_apply_pysig(len(f_args), kws.keys())
        tyg__yxg = (func, *f_args) + tuple(kws.values())
        return signature(dxkm__nykcc, *tyg__yxg).replace(pysig=wwmk__dcbn)

    def generic_resolve(self, grpby, attr):
        if self._is_existing_attr(attr):
            return
        if attr not in grpby.df_type.columns:
            raise_const_error(
                f'groupby: invalid attribute {attr} (column not found in dataframe or unsupported function)'
                )
        return DataFrameGroupByType(grpby.df_type, grpby.keys, (attr,),
            grpby.as_index, grpby.dropna, True, True)


def _get_groupby_apply_udf_out_type(func, grp, f_args, kws, typing_context,
    target_context):
    kbdo__nrv = grp.df_type
    if grp.explicit_select:
        if len(grp.selection) == 1:
            xwh__chla = grp.selection[0]
            gqocq__iuakh = kbdo__nrv.data[kbdo__nrv.columns.index(xwh__chla)]
            fife__auih = SeriesType(gqocq__iuakh.dtype, gqocq__iuakh,
                kbdo__nrv.index, types.literal(xwh__chla))
        else:
            uuo__yhfo = tuple(kbdo__nrv.data[kbdo__nrv.columns.index(
                psgb__vbgre)] for psgb__vbgre in grp.selection)
            fife__auih = DataFrameType(uuo__yhfo, kbdo__nrv.index, tuple(
                grp.selection))
    else:
        fife__auih = kbdo__nrv
    yzxjk__ibo = fife__auih,
    yzxjk__ibo += tuple(f_args)
    try:
        xrbl__vib = get_const_func_output_type(func, yzxjk__ibo, kws,
            typing_context, target_context)
    except Exception as giqk__rdq:
        raise_bodo_error(get_udf_error_msg('GroupBy.apply()', giqk__rdq),
            getattr(giqk__rdq, 'loc', None))
    return xrbl__vib


def resolve_obj_pipe(self, grp, args, kws, obj_name):
    kws = dict(kws)
    func = args[0] if len(args) > 0 else kws.pop('func', None)
    f_args = tuple(args[1:]) if len(args) > 0 else ()
    yzxjk__ibo = (grp,) + f_args
    try:
        xrbl__vib = get_const_func_output_type(func, yzxjk__ibo, kws, self.
            context, numba.core.registry.cpu_target.target_context, False)
    except Exception as giqk__rdq:
        raise_bodo_error(get_udf_error_msg(f'{obj_name}.pipe()', giqk__rdq),
            getattr(giqk__rdq, 'loc', None))
    wwmk__dcbn = gen_apply_pysig(len(f_args), kws.keys())
    tyg__yxg = (func, *f_args) + tuple(kws.values())
    return signature(xrbl__vib, *tyg__yxg).replace(pysig=wwmk__dcbn)


def gen_apply_pysig(n_args, kws):
    kllug__zlro = ', '.join(f'arg{lko__kwjd}' for lko__kwjd in range(n_args))
    kllug__zlro = kllug__zlro + ', ' if kllug__zlro else ''
    vwik__weonu = ', '.join(f"{tkjx__eqoo} = ''" for tkjx__eqoo in kws)
    jqpg__sits = f'def apply_stub(func, {kllug__zlro}{vwik__weonu}):\n'
    jqpg__sits += '    pass\n'
    cpu__yyei = {}
    exec(jqpg__sits, {}, cpu__yyei)
    guqu__ezq = cpu__yyei['apply_stub']
    return numba.core.utils.pysignature(guqu__ezq)


def pivot_table_dummy(df, values, index, columns, aggfunc, _pivot_values):
    return 0


@infer_global(pivot_table_dummy)
class PivotTableTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        df, values, index, columns, aggfunc, _pivot_values = args
        if not (is_overload_constant_str(values) and
            is_overload_constant_str(index) and is_overload_constant_str(
            columns)):
            raise BodoError(
                "pivot_table() only support string constants for 'values', 'index' and 'columns' arguments"
                )
        values = values.literal_value
        index = index.literal_value
        columns = columns.literal_value
        data = df.data[df.columns.index(values)]
        tpp__rou = get_pivot_output_dtype(data, aggfunc.literal_value)
        uiuis__vca = dtype_to_array_type(tpp__rou)
        if is_overload_none(_pivot_values):
            raise_bodo_error(
                'Dataframe.pivot_table() requires explicit annotation to determine output columns. For more information, see: https://docs.bodo.ai/latest/source/programming_with_bodo/pandas.html'
                )
        nlxyu__tqxcc = _pivot_values.meta
        lnqop__ypl = len(nlxyu__tqxcc)
        mhh__vcz = df.columns.index(index)
        gofcf__drib = bodo.hiframes.pd_index_ext.array_type_to_index(df.
            data[mhh__vcz], types.StringLiteral(index))
        dhop__vfh = DataFrameType((uiuis__vca,) * lnqop__ypl, gofcf__drib,
            tuple(nlxyu__tqxcc))
        return signature(dhop__vfh, *args)


PivotTableTyper._no_unliteral = True


@lower_builtin(pivot_table_dummy, types.VarArg(types.Any))
def lower_pivot_table_dummy(context, builder, sig, args):
    return context.get_constant_null(sig.return_type)


def crosstab_dummy(index, columns, _pivot_values):
    return 0


@infer_global(crosstab_dummy)
class CrossTabTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        index, columns, _pivot_values = args
        uiuis__vca = types.Array(types.int64, 1, 'C')
        nlxyu__tqxcc = _pivot_values.meta
        lnqop__ypl = len(nlxyu__tqxcc)
        gofcf__drib = bodo.hiframes.pd_index_ext.array_type_to_index(index.
            data, types.StringLiteral('index'))
        dhop__vfh = DataFrameType((uiuis__vca,) * lnqop__ypl, gofcf__drib,
            tuple(nlxyu__tqxcc))
        return signature(dhop__vfh, *args)


CrossTabTyper._no_unliteral = True


@lower_builtin(crosstab_dummy, types.VarArg(types.Any))
def lower_crosstab_dummy(context, builder, sig, args):
    return context.get_constant_null(sig.return_type)


def get_group_indices(keys, dropna, _is_parallel):
    return np.arange(len(keys))


@overload(get_group_indices)
def get_group_indices_overload(keys, dropna, _is_parallel):
    jqpg__sits = 'def impl(keys, dropna, _is_parallel):\n'
    jqpg__sits += (
        "    ev = bodo.utils.tracing.Event('get_group_indices', _is_parallel)\n"
        )
    jqpg__sits += '    info_list = [{}]\n'.format(', '.join(
        f'array_to_info(keys[{lko__kwjd}])' for lko__kwjd in range(len(keys
        .types))))
    jqpg__sits += '    table = arr_info_list_to_table(info_list)\n'
    jqpg__sits += '    group_labels = np.empty(len(keys[0]), np.int64)\n'
    jqpg__sits += '    sort_idx = np.empty(len(keys[0]), np.int64)\n'
    jqpg__sits += """    ngroups = get_groupby_labels(table, group_labels.ctypes, sort_idx.ctypes, dropna, _is_parallel)
"""
    jqpg__sits += '    delete_table_decref_arrays(table)\n'
    jqpg__sits += '    ev.finalize()\n'
    jqpg__sits += '    return sort_idx, group_labels, ngroups\n'
    cpu__yyei = {}
    exec(jqpg__sits, {'bodo': bodo, 'np': np, 'get_groupby_labels':
        get_groupby_labels, 'array_to_info': array_to_info,
        'arr_info_list_to_table': arr_info_list_to_table,
        'delete_table_decref_arrays': delete_table_decref_arrays}, cpu__yyei)
    ghjs__yep = cpu__yyei['impl']
    return ghjs__yep


@numba.njit(no_cpython_wrapper=True)
def generate_slices(labels, ngroups):
    yqm__jlg = len(labels)
    edzoz__lvjl = np.zeros(ngroups, dtype=np.int64)
    qey__svf = np.zeros(ngroups, dtype=np.int64)
    vhysg__llm = 0
    sfzar__crngo = 0
    for lko__kwjd in range(yqm__jlg):
        xal__wlxqr = labels[lko__kwjd]
        if xal__wlxqr < 0:
            vhysg__llm += 1
        else:
            sfzar__crngo += 1
            if lko__kwjd == yqm__jlg - 1 or xal__wlxqr != labels[lko__kwjd + 1
                ]:
                edzoz__lvjl[xal__wlxqr] = vhysg__llm
                qey__svf[xal__wlxqr] = vhysg__llm + sfzar__crngo
                vhysg__llm += sfzar__crngo
                sfzar__crngo = 0
    return edzoz__lvjl, qey__svf


def shuffle_dataframe(df, keys, _is_parallel):
    return df, keys, _is_parallel


@overload(shuffle_dataframe, prefer_literal=True)
def overload_shuffle_dataframe(df, keys, _is_parallel):
    enpx__upat = len(df.columns)
    oxmh__inw = len(keys.types)
    assert is_overload_constant_bool(_is_parallel
        ), 'shuffle_dataframe: _is_parallel is not a constant'
    jqpg__sits = 'def impl(df, keys, _is_parallel):\n'
    if is_overload_false(_is_parallel):
        jqpg__sits += '  return df, keys, get_null_shuffle_info()\n'
        cpu__yyei = {}
        exec(jqpg__sits, {'get_null_shuffle_info': get_null_shuffle_info},
            cpu__yyei)
        ghjs__yep = cpu__yyei['impl']
        return ghjs__yep
    for lko__kwjd in range(enpx__upat):
        jqpg__sits += f"""  in_arr{lko__kwjd} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {lko__kwjd})
"""
    jqpg__sits += f"""  in_index_arr = bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df))
"""
    jqpg__sits += '  info_list = [{}, {}, {}]\n'.format(', '.join(
        f'array_to_info(keys[{lko__kwjd}])' for lko__kwjd in range(
        oxmh__inw)), ', '.join(f'array_to_info(in_arr{lko__kwjd})' for
        lko__kwjd in range(enpx__upat)), 'array_to_info(in_index_arr)')
    jqpg__sits += '  table = arr_info_list_to_table(info_list)\n'
    jqpg__sits += (
        f'  out_table = shuffle_table(table, {oxmh__inw}, _is_parallel, 1)\n')
    for lko__kwjd in range(oxmh__inw):
        jqpg__sits += f"""  out_key{lko__kwjd} = info_to_array(info_from_table(out_table, {lko__kwjd}), keys[{lko__kwjd}])
"""
    for lko__kwjd in range(enpx__upat):
        jqpg__sits += f"""  out_arr{lko__kwjd} = info_to_array(info_from_table(out_table, {lko__kwjd + oxmh__inw}), in_arr{lko__kwjd})
"""
    jqpg__sits += f"""  out_arr_index = info_to_array(info_from_table(out_table, {oxmh__inw + enpx__upat}), in_index_arr)
"""
    jqpg__sits += '  shuffle_info = get_shuffle_info(out_table)\n'
    jqpg__sits += '  delete_table(out_table)\n'
    jqpg__sits += '  delete_table(table)\n'
    out_data = ', '.join(f'out_arr{lko__kwjd}' for lko__kwjd in range(
        enpx__upat))
    jqpg__sits += (
        '  out_index = bodo.utils.conversion.index_from_array(out_arr_index)\n'
        )
    jqpg__sits += f"""  out_df = bodo.hiframes.pd_dataframe_ext.init_dataframe(({out_data},), out_index, {gen_const_tup(df.columns)})
"""
    jqpg__sits += '  return out_df, ({},), shuffle_info\n'.format(', '.join
        (f'out_key{lko__kwjd}' for lko__kwjd in range(oxmh__inw)))
    cpu__yyei = {}
    exec(jqpg__sits, {'bodo': bodo, 'array_to_info': array_to_info,
        'arr_info_list_to_table': arr_info_list_to_table, 'shuffle_table':
        shuffle_table, 'info_from_table': info_from_table, 'info_to_array':
        info_to_array, 'delete_table': delete_table, 'get_shuffle_info':
        get_shuffle_info}, cpu__yyei)
    ghjs__yep = cpu__yyei['impl']
    return ghjs__yep


def reverse_shuffle(data, shuffle_info):
    return data


@overload(reverse_shuffle)
def overload_reverse_shuffle(data, shuffle_info):
    if isinstance(data, bodo.hiframes.pd_multi_index_ext.MultiIndexType):
        afbkq__sgt = len(data.array_types)
        jqpg__sits = 'def impl(data, shuffle_info):\n'
        jqpg__sits += '  info_list = [{}]\n'.format(', '.join(
            f'array_to_info(data._data[{lko__kwjd}])' for lko__kwjd in
            range(afbkq__sgt)))
        jqpg__sits += '  table = arr_info_list_to_table(info_list)\n'
        jqpg__sits += (
            '  out_table = reverse_shuffle_table(table, shuffle_info)\n')
        for lko__kwjd in range(afbkq__sgt):
            jqpg__sits += f"""  out_arr{lko__kwjd} = info_to_array(info_from_table(out_table, {lko__kwjd}), data._data[{lko__kwjd}])
"""
        jqpg__sits += '  delete_table(out_table)\n'
        jqpg__sits += '  delete_table(table)\n'
        jqpg__sits += (
            '  return init_multi_index(({},), data._names, data._name)\n'.
            format(', '.join(f'out_arr{lko__kwjd}' for lko__kwjd in range(
            afbkq__sgt))))
        cpu__yyei = {}
        exec(jqpg__sits, {'bodo': bodo, 'array_to_info': array_to_info,
            'arr_info_list_to_table': arr_info_list_to_table,
            'reverse_shuffle_table': reverse_shuffle_table,
            'info_from_table': info_from_table, 'info_to_array':
            info_to_array, 'delete_table': delete_table, 'init_multi_index':
            bodo.hiframes.pd_multi_index_ext.init_multi_index}, cpu__yyei)
        ghjs__yep = cpu__yyei['impl']
        return ghjs__yep
    if bodo.hiframes.pd_index_ext.is_index_type(data):

        def impl_index(data, shuffle_info):
            qefal__pgfmf = bodo.utils.conversion.index_to_array(data)
            qgul__jfrb = reverse_shuffle(qefal__pgfmf, shuffle_info)
            return bodo.utils.conversion.index_from_array(qgul__jfrb)
        return impl_index

    def impl_arr(data, shuffle_info):
        yyu__tidg = [array_to_info(data)]
        fawac__mdeq = arr_info_list_to_table(yyu__tidg)
        oxmj__icos = reverse_shuffle_table(fawac__mdeq, shuffle_info)
        qgul__jfrb = info_to_array(info_from_table(oxmj__icos, 0), data)
        delete_table(oxmj__icos)
        delete_table(fawac__mdeq)
        return qgul__jfrb
    return impl_arr


@overload_method(DataFrameGroupByType, 'value_counts', inline='always',
    no_unliteral=True)
def groupby_value_counts(grp, normalize=False, sort=True, ascending=False,
    bins=None, dropna=True):
    goz__hzh = dict(normalize=normalize, sort=sort, bins=bins, dropna=dropna)
    bbq__epg = dict(normalize=False, sort=True, bins=None, dropna=True)
    check_unsupported_args('Groupby.value_counts', goz__hzh, bbq__epg,
        package_name='pandas', module_name='GroupBy')
    if len(grp.selection) > 1 or not grp.as_index:
        raise BodoError(
            "'DataFrameGroupBy' object has no attribute 'value_counts'")
    if not is_overload_constant_bool(ascending):
        raise BodoError(
            'Groupby.value_counts() ascending must be a constant boolean')
    rxhz__nzbdb = get_overload_const_bool(ascending)
    frzik__oshk = grp.selection[0]
    jqpg__sits = f"""def impl(grp, normalize=False, sort=True, ascending=False, bins=None, dropna=True):
"""
    hjd__wbuk = (
        f"lambda S: S.value_counts(ascending={rxhz__nzbdb}, _index_name='{frzik__oshk}')"
        )
    jqpg__sits += f'    return grp.apply({hjd__wbuk})\n'
    cpu__yyei = {}
    exec(jqpg__sits, {'bodo': bodo}, cpu__yyei)
    ghjs__yep = cpu__yyei['impl']
    return ghjs__yep


groupby_unsupported_attr = {'groups', 'indices'}
groupby_unsupported = {'__iter__', 'get_group', 'all', 'any', 'bfill',
    'backfill', 'cumcount', 'cummax', 'cummin', 'cumprod', 'ffill',
    'ngroup', 'nth', 'ohlc', 'pad', 'rank', 'pct_change', 'sem', 'tail',
    'corr', 'cov', 'describe', 'diff', 'fillna', 'filter', 'hist', 'mad',
    'plot', 'quantile', 'resample', 'sample', 'skew', 'take', 'tshift'}
series_only_unsupported_attrs = {'is_monotonic_increasing',
    'is_monotonic_decreasing'}
series_only_unsupported = {'nlargest', 'nsmallest', 'unique'}
dataframe_only_unsupported = {'corrwith', 'boxplot'}


def _install_groupy_unsupported():
    for yvn__agqxl in groupby_unsupported_attr:
        overload_attribute(DataFrameGroupByType, yvn__agqxl, no_unliteral=True
            )(create_unsupported_overload(f'DataFrameGroupBy.{yvn__agqxl}'))
    for yvn__agqxl in groupby_unsupported:
        overload_method(DataFrameGroupByType, yvn__agqxl, no_unliteral=True)(
            create_unsupported_overload(f'DataFrameGroupBy.{yvn__agqxl}'))
    for yvn__agqxl in series_only_unsupported_attrs:
        overload_attribute(DataFrameGroupByType, yvn__agqxl, no_unliteral=True
            )(create_unsupported_overload(f'SeriesGroupBy.{yvn__agqxl}'))
    for yvn__agqxl in series_only_unsupported:
        overload_method(DataFrameGroupByType, yvn__agqxl, no_unliteral=True)(
            create_unsupported_overload(f'SeriesGroupBy.{yvn__agqxl}'))
    for yvn__agqxl in dataframe_only_unsupported:
        overload_method(DataFrameGroupByType, yvn__agqxl, no_unliteral=True)(
            create_unsupported_overload(f'DataFrameGroupBy.{yvn__agqxl}'))


_install_groupy_unsupported()
