"""Support for MultiIndex type of Pandas
"""
import operator
import numba
import pandas as pd
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, lower_builtin, make_attribute_wrapper, models, overload, register_model, typeof_impl, unbox
from bodo.utils.typing import BodoError, check_unsupported_args, dtype_to_array_type, get_val_type_maybe_str_literal, is_overload_none


class MultiIndexType(types.Type):

    def __init__(self, array_types, names_typ=None, name_typ=None):
        names_typ = (types.none,) * len(array_types
            ) if names_typ is None else names_typ
        name_typ = types.none if name_typ is None else name_typ
        self.array_types = array_types
        self.names_typ = names_typ
        self.name_typ = name_typ
        super(MultiIndexType, self).__init__(name=
            'MultiIndexType({}, {}, {})'.format(array_types, names_typ,
            name_typ))
    ndim = 1

    def copy(self):
        return MultiIndexType(self.array_types, self.names_typ, self.name_typ)

    @property
    def nlevels(self):
        return len(self.array_types)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(MultiIndexType)
class MultiIndexModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        qfifp__qarxa = [('data', types.Tuple(fe_type.array_types)), (
            'names', types.Tuple(fe_type.names_typ)), ('name', fe_type.
            name_typ)]
        super(MultiIndexModel, self).__init__(dmm, fe_type, qfifp__qarxa)


make_attribute_wrapper(MultiIndexType, 'data', '_data')
make_attribute_wrapper(MultiIndexType, 'names', '_names')
make_attribute_wrapper(MultiIndexType, 'name', '_name')


@typeof_impl.register(pd.MultiIndex)
def typeof_multi_index(val, c):
    array_types = tuple(numba.typeof(val.levels[vwz__pcter].values) for
        vwz__pcter in range(val.nlevels))
    return MultiIndexType(array_types, tuple(get_val_type_maybe_str_literal
        (plarm__aowh) for plarm__aowh in val.names), numba.typeof(val.name))


@box(MultiIndexType)
def box_multi_index(typ, val, c):
    yobqe__zvfd = c.context.insert_const_string(c.builder.module, 'pandas')
    fqcva__igrym = c.pyapi.import_module_noblock(yobqe__zvfd)
    dmhl__sqj = c.pyapi.object_getattr_string(fqcva__igrym, 'MultiIndex')
    doh__bjyd = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, types.Tuple(typ.array_types), doh__bjyd
        .data)
    bgh__ixcs = c.pyapi.from_native_value(types.Tuple(typ.array_types),
        doh__bjyd.data, c.env_manager)
    c.context.nrt.incref(c.builder, types.Tuple(typ.names_typ), doh__bjyd.names
        )
    wmdai__eiht = c.pyapi.from_native_value(types.Tuple(typ.names_typ),
        doh__bjyd.names, c.env_manager)
    c.context.nrt.incref(c.builder, typ.name_typ, doh__bjyd.name)
    jsvdn__nqnvc = c.pyapi.from_native_value(typ.name_typ, doh__bjyd.name,
        c.env_manager)
    pudyw__gluys = c.pyapi.borrow_none()
    xgpr__jai = c.pyapi.call_method(dmhl__sqj, 'from_arrays', (bgh__ixcs,
        pudyw__gluys, wmdai__eiht))
    c.pyapi.object_setattr_string(xgpr__jai, 'name', jsvdn__nqnvc)
    c.pyapi.decref(bgh__ixcs)
    c.pyapi.decref(wmdai__eiht)
    c.pyapi.decref(jsvdn__nqnvc)
    c.pyapi.decref(fqcva__igrym)
    c.pyapi.decref(dmhl__sqj)
    c.context.nrt.decref(c.builder, typ, val)
    return xgpr__jai


@unbox(MultiIndexType)
def unbox_multi_index(typ, val, c):
    nzt__qvjj = []
    objr__kawio = []
    for vwz__pcter in range(typ.nlevels):
        psu__qeobg = c.pyapi.unserialize(c.pyapi.serialize_object(vwz__pcter))
        kaw__utr = c.pyapi.call_method(val, 'get_level_values', (psu__qeobg,))
        lmjkf__nobx = c.pyapi.object_getattr_string(kaw__utr, 'values')
        c.pyapi.decref(kaw__utr)
        c.pyapi.decref(psu__qeobg)
        hlavc__ybd = c.pyapi.to_native_value(typ.array_types[vwz__pcter],
            lmjkf__nobx).value
        nzt__qvjj.append(hlavc__ybd)
        objr__kawio.append(lmjkf__nobx)
    if isinstance(types.Tuple(typ.array_types), types.UniTuple):
        data = cgutils.pack_array(c.builder, nzt__qvjj)
    else:
        data = cgutils.pack_struct(c.builder, nzt__qvjj)
    wmdai__eiht = c.pyapi.object_getattr_string(val, 'names')
    quetf__ewgpp = c.pyapi.unserialize(c.pyapi.serialize_object(tuple))
    giyd__htx = c.pyapi.call_function_objargs(quetf__ewgpp, (wmdai__eiht,))
    names = c.pyapi.to_native_value(types.Tuple(typ.names_typ), giyd__htx
        ).value
    jsvdn__nqnvc = c.pyapi.object_getattr_string(val, 'name')
    name = c.pyapi.to_native_value(typ.name_typ, jsvdn__nqnvc).value
    doh__bjyd = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    doh__bjyd.data = data
    doh__bjyd.names = names
    doh__bjyd.name = name
    for lmjkf__nobx in objr__kawio:
        c.pyapi.decref(lmjkf__nobx)
    c.pyapi.decref(wmdai__eiht)
    c.pyapi.decref(quetf__ewgpp)
    c.pyapi.decref(giyd__htx)
    c.pyapi.decref(jsvdn__nqnvc)
    return NativeValue(doh__bjyd._getvalue())


def from_product_error_checking(iterables, sortorder, names):
    hvs__jut = 'pandas.MultiIndex.from_product'
    qexa__uzll = dict(sortorder=sortorder)
    xovih__mkbjp = dict(sortorder=None)
    check_unsupported_args(hvs__jut, qexa__uzll, xovih__mkbjp, package_name
        ='pandas', module_name='Index')
    if not (is_overload_none(names) or isinstance(names, types.BaseTuple)):
        raise BodoError(f'{hvs__jut}: names must be None or a tuple.')
    elif not isinstance(iterables, types.BaseTuple):
        raise BodoError(f'{hvs__jut}: iterables must be a tuple.')
    elif not is_overload_none(names) and len(iterables) != len(names):
        raise BodoError(
            f'{hvs__jut}: iterables and names must be of the same length.')


def from_product(iterable, sortorder=None, names=None):
    pass


@overload(from_product)
def from_product_overload(iterables, sortorder=None, names=None):
    from_product_error_checking(iterables, sortorder, names)
    array_types = tuple(dtype_to_array_type(iterable.dtype) for iterable in
        iterables)
    if is_overload_none(names):
        names_typ = tuple([types.none] * len(iterables))
    else:
        names_typ = names.types
    fnukf__rmlyh = MultiIndexType(array_types, names_typ)
    cjs__ctgs = f'from_product_multiindex{numba.core.ir_utils.next_label()}'
    setattr(types, cjs__ctgs, fnukf__rmlyh)
    necrm__fcug = f"""
def impl(iterables, sortorder=None, names=None):
    with numba.objmode(mi='{cjs__ctgs}'):
        mi = pd.MultiIndex.from_product(iterables, names=names)
    return mi
"""
    khiw__cjf = {}
    exec(necrm__fcug, globals(), khiw__cjf)
    jml__dxjfd = khiw__cjf['impl']
    return jml__dxjfd


@intrinsic
def init_multi_index(typingctx, data, names, name=None):
    name = types.none if name is None else name
    names = types.Tuple(names.types)

    def codegen(context, builder, signature, args):
        xlupg__ukhr, dmld__jol, mmfks__awp = args
        tvwr__byum = cgutils.create_struct_proxy(signature.return_type)(context
            , builder)
        tvwr__byum.data = xlupg__ukhr
        tvwr__byum.names = dmld__jol
        tvwr__byum.name = mmfks__awp
        context.nrt.incref(builder, signature.args[0], xlupg__ukhr)
        context.nrt.incref(builder, signature.args[1], dmld__jol)
        context.nrt.incref(builder, signature.args[2], mmfks__awp)
        return tvwr__byum._getvalue()
    qwsw__neev = MultiIndexType(data.types, names.types, name)
    return qwsw__neev(data, names, name), codegen


@overload(len, no_unliteral=True)
def overload_len_pd_multiindex(A):
    if isinstance(A, MultiIndexType):
        return lambda A: len(A._data[0])


@overload(operator.getitem, no_unliteral=True)
def overload_multi_index_getitem(I, ind):
    if not isinstance(I, MultiIndexType):
        return
    if not isinstance(ind, types.Integer):
        fpo__rpv = len(I.array_types)
        necrm__fcug = 'def impl(I, ind):\n'
        necrm__fcug += '  data = I._data\n'
        necrm__fcug += ('  return init_multi_index(({},), I._names, I._name)\n'
            .format(', '.join(f'data[{vwz__pcter}][ind]' for vwz__pcter in
            range(fpo__rpv))))
        khiw__cjf = {}
        exec(necrm__fcug, {'init_multi_index': init_multi_index}, khiw__cjf)
        jml__dxjfd = khiw__cjf['impl']
        return jml__dxjfd


@lower_builtin(operator.is_, MultiIndexType, MultiIndexType)
def multi_index_is(context, builder, sig, args):
    usyfg__gcixc, ctisx__ydj = sig.args
    if usyfg__gcixc != ctisx__ydj:
        return cgutils.false_bit

    def index_is_impl(a, b):
        return (a._data is b._data and a._names is b._names and a._name is
            b._name)
    return context.compile_internal(builder, index_is_impl, sig, args)
