"""typing for rolling window functions
"""
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed
from numba.core.typing.templates import AbstractTemplate, AttributeTemplate, signature
from numba.extending import infer, infer_getattr, intrinsic, lower_builtin, make_attribute_wrapper, models, overload, overload_method, register_model
import bodo
from bodo.hiframes.datetime_timedelta_ext import datetime_timedelta_type, pd_timedelta_type
from bodo.hiframes.pd_dataframe_ext import DataFrameType, check_runtime_cols_unsupported
from bodo.hiframes.pd_groupby_ext import DataFrameGroupByType
from bodo.hiframes.pd_series_ext import SeriesType
from bodo.hiframes.rolling import supported_rolling_funcs, unsupported_rolling_methods
from bodo.utils.typing import BodoError, check_unsupported_args, create_unsupported_overload, get_literal_value, is_const_func_type, is_literal_type, is_overload_bool, is_overload_constant_str, is_overload_int, is_overload_none, raise_bodo_error, raise_const_error


class RollingType(types.Type):

    def __init__(self, obj_type, window_type, on, selection,
        explicit_select=False, series_select=False):
        self.obj_type = obj_type
        self.window_type = window_type
        self.on = on
        self.selection = selection
        self.explicit_select = explicit_select
        self.series_select = series_select
        super(RollingType, self).__init__(name=
            f'RollingType({obj_type}, {window_type}, {on}, {selection}, {explicit_select}, {series_select})'
            )

    def copy(self):
        return RollingType(self.obj_type, self.window_type, self.on, self.
            selection, self.explicit_select, self.series_select)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(RollingType)
class RollingModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        vzshi__abto = [('obj', fe_type.obj_type), ('window', fe_type.
            window_type), ('min_periods', types.int64), ('center', types.bool_)
            ]
        super(RollingModel, self).__init__(dmm, fe_type, vzshi__abto)


make_attribute_wrapper(RollingType, 'obj', 'obj')
make_attribute_wrapper(RollingType, 'window', 'window')
make_attribute_wrapper(RollingType, 'center', 'center')
make_attribute_wrapper(RollingType, 'min_periods', 'min_periods')


@overload_method(DataFrameType, 'rolling', inline='always', no_unliteral=True)
def df_rolling_overload(df, window, min_periods=None, center=False,
    win_type=None, on=None, axis=0, closed=None):
    check_runtime_cols_unsupported(df, 'DataFrame.rolling()')
    iue__cxc = dict(win_type=win_type, axis=axis, closed=closed)
    dsb__qvkhq = dict(win_type=None, axis=0, closed=None)
    check_unsupported_args('DataFrame.rolling', iue__cxc, dsb__qvkhq,
        package_name='pandas', module_name='Window')
    _validate_rolling_args(df, window, min_periods, center, on)

    def impl(df, window, min_periods=None, center=False, win_type=None, on=
        None, axis=0, closed=None):
        min_periods = _handle_default_min_periods(min_periods, window)
        return bodo.hiframes.pd_rolling_ext.init_rolling(df, window,
            min_periods, center, on)
    return impl


@overload_method(SeriesType, 'rolling', inline='always', no_unliteral=True)
def overload_series_rolling(S, window, min_periods=None, center=False,
    win_type=None, on=None, axis=0, closed=None):
    iue__cxc = dict(win_type=win_type, axis=axis, closed=closed)
    dsb__qvkhq = dict(win_type=None, axis=0, closed=None)
    check_unsupported_args('Series.rolling', iue__cxc, dsb__qvkhq,
        package_name='pandas', module_name='Window')
    _validate_rolling_args(S, window, min_periods, center, on)

    def impl(S, window, min_periods=None, center=False, win_type=None, on=
        None, axis=0, closed=None):
        min_periods = _handle_default_min_periods(min_periods, window)
        return bodo.hiframes.pd_rolling_ext.init_rolling(S, window,
            min_periods, center, on)
    return impl


@intrinsic
def init_rolling(typingctx, obj_type, window_type, min_periods_type,
    center_type, on_type=None):

    def codegen(context, builder, signature, args):
        nsni__hkg, tryg__inog, xaord__npn, unhxk__zmu, fcbr__xpfk = args
        spma__pism = signature.return_type
        tpx__qllpx = cgutils.create_struct_proxy(spma__pism)(context, builder)
        tpx__qllpx.obj = nsni__hkg
        tpx__qllpx.window = tryg__inog
        tpx__qllpx.min_periods = xaord__npn
        tpx__qllpx.center = unhxk__zmu
        context.nrt.incref(builder, signature.args[0], nsni__hkg)
        context.nrt.incref(builder, signature.args[1], tryg__inog)
        context.nrt.incref(builder, signature.args[2], xaord__npn)
        context.nrt.incref(builder, signature.args[3], unhxk__zmu)
        return tpx__qllpx._getvalue()
    on = get_literal_value(on_type)
    if isinstance(obj_type, SeriesType):
        selection = None
    elif isinstance(obj_type, DataFrameType):
        selection = obj_type.columns
    else:
        assert isinstance(obj_type, DataFrameGroupByType
            ), f'invalid obj type for rolling: {obj_type}'
        selection = obj_type.selection
    spma__pism = RollingType(obj_type, window_type, on, selection, False)
    return spma__pism(obj_type, window_type, min_periods_type, center_type,
        on_type), codegen


def _handle_default_min_periods(min_periods, window):
    return min_periods


@overload(_handle_default_min_periods)
def overload_handle_default_min_periods(min_periods, window):
    if is_overload_none(min_periods):
        if isinstance(window, types.Integer):
            return lambda min_periods, window: window
        else:
            return lambda min_periods, window: 1
    else:
        return lambda min_periods, window: min_periods


def _gen_df_rolling_out_data(rolling):
    twk__gffvn = not isinstance(rolling.window_type, types.Integer)
    ztoa__rrk = 'variable' if twk__gffvn else 'fixed'
    dbdcy__cpadj = 'None'
    if twk__gffvn:
        dbdcy__cpadj = ('bodo.utils.conversion.index_to_array(index)' if 
            rolling.on is None else
            f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {rolling.obj_type.columns.index(rolling.on)})'
            )
    yoij__niwbk = []
    esyh__gwofq = 'on_arr, ' if twk__gffvn else ''
    if isinstance(rolling.obj_type, SeriesType):
        return (
            f'bodo.hiframes.rolling.rolling_{ztoa__rrk}(bodo.hiframes.pd_series_ext.get_series_data(df), {esyh__gwofq}index_arr, window, minp, center, func, raw)'
            , dbdcy__cpadj, rolling.selection)
    assert isinstance(rolling.obj_type, DataFrameType
        ), 'expected df in rolling obj'
    phzkb__affmf = rolling.obj_type.data
    out_cols = []
    for zescr__lnlck in rolling.selection:
        wbdix__dvexv = rolling.obj_type.columns.index(zescr__lnlck)
        if zescr__lnlck == rolling.on:
            if len(rolling.selection) == 2 and rolling.series_select:
                continue
            wbwlo__pkkng = (
                f'bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {wbdix__dvexv})'
                )
            out_cols.append(zescr__lnlck)
        else:
            if not isinstance(phzkb__affmf[wbdix__dvexv].dtype, (types.
                Boolean, types.Number)):
                continue
            wbwlo__pkkng = (
                f'bodo.hiframes.rolling.rolling_{ztoa__rrk}(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {wbdix__dvexv}), {esyh__gwofq}index_arr, window, minp, center, func, raw)'
                )
            out_cols.append(zescr__lnlck)
        yoij__niwbk.append(wbwlo__pkkng)
    return ', '.join(yoij__niwbk), dbdcy__cpadj, tuple(out_cols)


@overload_method(RollingType, 'apply', inline='always', no_unliteral=True)
def overload_rolling_apply(rolling, func, raw=False, engine=None,
    engine_kwargs=None, args=None, kwargs=None):
    iue__cxc = dict(engine=engine, engine_kwargs=engine_kwargs, args=args,
        kwargs=kwargs)
    dsb__qvkhq = dict(engine=None, engine_kwargs=None, args=None, kwargs=None)
    check_unsupported_args('Rolling.apply', iue__cxc, dsb__qvkhq,
        package_name='pandas', module_name='Window')
    if not is_const_func_type(func):
        raise BodoError(
            f"Rolling.apply(): 'func' parameter must be a function, not {func} (builtin functions not supported yet)."
            )
    if not is_overload_bool(raw):
        raise BodoError(
            f"Rolling.apply(): 'raw' parameter must be bool, not {raw}.")
    return _gen_rolling_impl(rolling, 'apply')


@overload_method(DataFrameGroupByType, 'rolling', inline='always',
    no_unliteral=True)
def groupby_rolling_overload(grp, window, min_periods=None, center=False,
    win_type=None, on=None, axis=0, closed=None, method='single'):
    iue__cxc = dict(win_type=win_type, axis=axis, closed=closed, method=method)
    dsb__qvkhq = dict(win_type=None, axis=0, closed=None, method='single')
    check_unsupported_args('GroupBy.rolling', iue__cxc, dsb__qvkhq,
        package_name='pandas', module_name='Window')
    _validate_rolling_args(grp, window, min_periods, center, on)

    def _impl(grp, window, min_periods=None, center=False, win_type=None,
        on=None, axis=0, closed=None, method='single'):
        min_periods = _handle_default_min_periods(min_periods, window)
        return bodo.hiframes.pd_rolling_ext.init_rolling(grp, window,
            min_periods, center, on)
    return _impl


def _gen_rolling_impl(rolling, fname, other=None):
    if isinstance(rolling.obj_type, DataFrameGroupByType):
        vngs__naflr = f'def impl(rolling, {_get_rolling_func_args(fname)}):\n'
        ubm__qaka = f"'{rolling.on}'" if isinstance(rolling.on, str
            ) else f'{rolling.on}'
        selection = ''
        if rolling.explicit_select:
            selection = '[{}]'.format(', '.join(f"'{vrpv__fxal}'" if
                isinstance(vrpv__fxal, str) else f'{vrpv__fxal}' for
                vrpv__fxal in rolling.selection if vrpv__fxal != rolling.on))
        bhx__sdo = inu__zppl = ''
        if fname == 'apply':
            bhx__sdo = 'func, raw, args, kwargs'
            inu__zppl = 'func, raw, None, None, args, kwargs'
        if fname == 'corr':
            bhx__sdo = inu__zppl = 'other, pairwise'
        if fname == 'cov':
            bhx__sdo = inu__zppl = 'other, pairwise, ddof'
        gmxn__ytg = (
            f'lambda df, window, minp, center, {bhx__sdo}: bodo.hiframes.pd_rolling_ext.init_rolling(df, window, minp, center, {ubm__qaka}){selection}.{fname}({inu__zppl})'
            )
        vngs__naflr += f"""  return rolling.obj.apply({gmxn__ytg}, rolling.window, rolling.min_periods, rolling.center, {bhx__sdo})
"""
        tmdbj__rlql = {}
        exec(vngs__naflr, {'bodo': bodo}, tmdbj__rlql)
        impl = tmdbj__rlql['impl']
        return impl
    judw__muksr = isinstance(rolling.obj_type, SeriesType)
    if fname in ('corr', 'cov'):
        out_cols = None if judw__muksr else _get_corr_cov_out_cols(rolling,
            other, fname)
        df_cols = None if judw__muksr else rolling.obj_type.columns
        other_cols = None if judw__muksr else other.columns
        yoij__niwbk, dbdcy__cpadj = _gen_corr_cov_out_data(out_cols,
            df_cols, other_cols, rolling.window_type, fname)
    else:
        yoij__niwbk, dbdcy__cpadj, out_cols = _gen_df_rolling_out_data(rolling)
    lgbs__mkzyl = judw__muksr or len(rolling.selection) == (1 if rolling.on is
        None else 2) and rolling.series_select
    nuapy__ekr = f'def impl(rolling, {_get_rolling_func_args(fname)}):\n'
    nuapy__ekr += '  df = rolling.obj\n'
    nuapy__ekr += '  index = {}\n'.format(
        'bodo.hiframes.pd_series_ext.get_series_index(df)' if judw__muksr else
        'bodo.hiframes.pd_dataframe_ext.get_dataframe_index(df)')
    bgb__asmc = 'None'
    if judw__muksr:
        bgb__asmc = 'bodo.hiframes.pd_series_ext.get_series_name(df)'
    elif lgbs__mkzyl:
        zescr__lnlck = (set(out_cols) - set([rolling.on])).pop()
        bgb__asmc = f"'{zescr__lnlck}'" if isinstance(zescr__lnlck, str
            ) else str(zescr__lnlck)
    nuapy__ekr += f'  name = {bgb__asmc}\n'
    nuapy__ekr += '  window = rolling.window\n'
    nuapy__ekr += '  center = rolling.center\n'
    nuapy__ekr += '  minp = rolling.min_periods\n'
    nuapy__ekr += f'  on_arr = {dbdcy__cpadj}\n'
    if fname == 'apply':
        nuapy__ekr += (
            f'  index_arr = bodo.utils.conversion.index_to_array(index)\n')
    else:
        nuapy__ekr += f"  func = '{fname}'\n"
        nuapy__ekr += f'  index_arr = None\n'
        nuapy__ekr += f'  raw = False\n'
    if lgbs__mkzyl:
        nuapy__ekr += (
            f'  return bodo.hiframes.pd_series_ext.init_series({yoij__niwbk}, index, name)'
            )
        tmdbj__rlql = {}
        tlzxm__mfy = {'bodo': bodo}
        exec(nuapy__ekr, tlzxm__mfy, tmdbj__rlql)
        impl = tmdbj__rlql['impl']
        return impl
    return bodo.hiframes.dataframe_impl._gen_init_df(nuapy__ekr, out_cols,
        yoij__niwbk)


def _get_rolling_func_args(fname):
    if fname == 'apply':
        return (
            'func, raw=False, engine=None, engine_kwargs=None, args=None, kwargs=None\n'
            )
    elif fname == 'corr':
        return 'other=None, pairwise=None, ddof=1\n'
    elif fname == 'cov':
        return 'other=None, pairwise=None, ddof=1\n'
    return ''


def create_rolling_overload(fname):

    def overload_rolling_func(rolling):
        return _gen_rolling_impl(rolling, fname)
    return overload_rolling_func


def _install_rolling_methods():
    for fname in supported_rolling_funcs:
        if fname in ('apply', 'corr', 'cov'):
            continue
        evs__umby = create_rolling_overload(fname)
        overload_method(RollingType, fname, inline='always', no_unliteral=True
            )(evs__umby)


def _install_rolling_unsupported_methods():
    for fname in unsupported_rolling_methods:
        overload_method(RollingType, fname, no_unliteral=True)(
            create_unsupported_overload(
            f'pandas.core.window.rolling.Rolling.{fname}()'))


_install_rolling_methods()
_install_rolling_unsupported_methods()


def _get_corr_cov_out_cols(rolling, other, func_name):
    if not isinstance(other, DataFrameType):
        raise_bodo_error(
            f"DataFrame.rolling.{func_name}(): requires providing a DataFrame for 'other'"
            )
    cfifx__rsw = rolling.selection
    if rolling.on is not None:
        raise BodoError(
            f'variable window rolling {func_name} not supported yet.')
    out_cols = tuple(sorted(set(cfifx__rsw) | set(other.columns), key=lambda
        k: str(k)))
    return out_cols


def _gen_corr_cov_out_data(out_cols, df_cols, other_cols, window_type,
    func_name):
    twk__gffvn = not isinstance(window_type, types.Integer)
    dbdcy__cpadj = 'None'
    if twk__gffvn:
        dbdcy__cpadj = 'bodo.utils.conversion.index_to_array(index)'
    esyh__gwofq = 'on_arr, ' if twk__gffvn else ''
    yoij__niwbk = []
    if out_cols is None:
        return (
            f'bodo.hiframes.rolling.rolling_{func_name}(bodo.hiframes.pd_series_ext.get_series_data(df), bodo.hiframes.pd_series_ext.get_series_data(other), {esyh__gwofq}window, minp, center)'
            , dbdcy__cpadj)
    for zescr__lnlck in out_cols:
        if zescr__lnlck in df_cols and zescr__lnlck in other_cols:
            ixr__ggnhc = df_cols.index(zescr__lnlck)
            ktt__hsjcx = other_cols.index(zescr__lnlck)
            wbwlo__pkkng = (
                f'bodo.hiframes.rolling.rolling_{func_name}(bodo.hiframes.pd_dataframe_ext.get_dataframe_data(df, {ixr__ggnhc}), bodo.hiframes.pd_dataframe_ext.get_dataframe_data(other, {ktt__hsjcx}), {esyh__gwofq}window, minp, center)'
                )
        else:
            wbwlo__pkkng = 'np.full(len(df), np.nan)'
        yoij__niwbk.append(wbwlo__pkkng)
    return ', '.join(yoij__niwbk), dbdcy__cpadj


@overload_method(RollingType, 'corr', inline='always', no_unliteral=True)
def overload_rolling_corr(rolling, other=None, pairwise=None, ddof=1):
    ffxaf__dkbsr = {'pairwise': pairwise, 'ddof': ddof}
    izo__zyng = {'pairwise': None, 'ddof': 1}
    check_unsupported_args('pandas.core.window.rolling.Rolling.corr',
        ffxaf__dkbsr, izo__zyng, package_name='pandas', module_name='Window')
    return _gen_rolling_impl(rolling, 'corr', other)


@overload_method(RollingType, 'cov', inline='always', no_unliteral=True)
def overload_rolling_cov(rolling, other=None, pairwise=None, ddof=1):
    ffxaf__dkbsr = {'ddof': ddof, 'pairwise': pairwise}
    izo__zyng = {'ddof': 1, 'pairwise': None}
    check_unsupported_args('pandas.core.window.rolling.Rolling.cov',
        ffxaf__dkbsr, izo__zyng, package_name='pandas', module_name='Window')
    return _gen_rolling_impl(rolling, 'cov', other)


@infer
class GetItemDataFrameRolling2(AbstractTemplate):
    key = 'static_getitem'

    def generic(self, args, kws):
        rolling, fwdbc__rahm = args
        if isinstance(rolling, RollingType):
            cfifx__rsw = rolling.obj_type.selection if isinstance(rolling.
                obj_type, DataFrameGroupByType) else rolling.obj_type.columns
            series_select = False
            if isinstance(fwdbc__rahm, (tuple, list)):
                if len(set(fwdbc__rahm).difference(set(cfifx__rsw))) > 0:
                    raise_const_error(
                        'rolling: selected column {} not found in dataframe'
                        .format(set(fwdbc__rahm).difference(set(cfifx__rsw))))
                selection = list(fwdbc__rahm)
            else:
                if fwdbc__rahm not in cfifx__rsw:
                    raise_const_error(
                        'rolling: selected column {} not found in dataframe'
                        .format(fwdbc__rahm))
                selection = [fwdbc__rahm]
                series_select = True
            if rolling.on is not None:
                selection.append(rolling.on)
            csn__rrhtf = RollingType(rolling.obj_type, rolling.window_type,
                rolling.on, tuple(selection), True, series_select)
            return signature(csn__rrhtf, *args)


@lower_builtin('static_getitem', RollingType, types.Any)
def static_getitem_df_groupby(context, builder, sig, args):
    return impl_ret_borrowed(context, builder, sig.return_type, args[0])


@infer_getattr
class RollingAttribute(AttributeTemplate):
    key = RollingType

    def generic_resolve(self, rolling, attr):
        cfifx__rsw = ()
        if isinstance(rolling.obj_type, DataFrameGroupByType):
            cfifx__rsw = rolling.obj_type.selection
        if isinstance(rolling.obj_type, DataFrameType):
            cfifx__rsw = rolling.obj_type.columns
        if attr in cfifx__rsw:
            return RollingType(rolling.obj_type, rolling.window_type,
                rolling.on, (attr,) if rolling.on is None else (attr,
                rolling.on), True, True)


def _validate_rolling_args(obj, window, min_periods, center, on):
    assert isinstance(obj, (SeriesType, DataFrameType, DataFrameGroupByType)
        ), 'invalid rolling obj'
    func_name = 'Series' if isinstance(obj, SeriesType
        ) else 'DataFrame' if isinstance(obj, DataFrameType
        ) else 'DataFrameGroupBy'
    if not (is_overload_int(window) or is_overload_constant_str(window) or 
        window == bodo.string_type or window in (pd_timedelta_type,
        datetime_timedelta_type)):
        raise BodoError(
            f"{func_name}.rolling(): 'window' should be int or time offset (str, pd.Timedelta, datetime.timedelta), not {window}"
            )
    if not is_overload_bool(center):
        raise BodoError(
            f'{func_name}.rolling(): center must be a boolean, not {center}')
    if not (is_overload_none(min_periods) or isinstance(min_periods, types.
        Integer)):
        raise BodoError(
            f'{func_name}.rolling(): min_periods must be an integer, not {min_periods}'
            )
    if isinstance(obj, SeriesType) and not is_overload_none(on):
        raise BodoError(
            f"{func_name}.rolling(): 'on' not supported for Series yet (can use a DataFrame instead)."
            )
    ryi__jafjj = obj.columns if isinstance(obj, DataFrameType
        ) else obj.df_type.columns if isinstance(obj, DataFrameGroupByType
        ) else []
    phzkb__affmf = [obj.data] if isinstance(obj, SeriesType
        ) else obj.data if isinstance(obj, DataFrameType) else obj.df_type.data
    if not is_overload_none(on) and (not is_literal_type(on) or 
        get_literal_value(on) not in ryi__jafjj):
        raise BodoError(
            f"{func_name}.rolling(): 'on' should be a constant column name.")
    if not is_overload_none(on):
        pjhfn__wly = phzkb__affmf[ryi__jafjj.index(get_literal_value(on))]
        if not isinstance(pjhfn__wly, types.Array
            ) or pjhfn__wly.dtype != bodo.datetime64ns:
            raise BodoError(
                f"{func_name}.rolling(): 'on' column should have datetime64 data."
                )
    if not any(isinstance(prb__uyw.dtype, (types.Boolean, types.Number)) for
        prb__uyw in phzkb__affmf):
        raise BodoError(f'{func_name}.rolling(): No numeric types to aggregate'
            )
