"""
Implementation of Series attributes and methods using overload.
"""
import operator
import numba
import numpy as np
import pandas as pd
from numba.core import types
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import lower_builtin, overload, overload_attribute, overload_method, register_jitable
import bodo
from bodo.hiframes.datetime_datetime_ext import datetime_datetime_type
from bodo.hiframes.datetime_timedelta_ext import PDTimeDeltaType, datetime_timedelta_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.hiframes.pd_offsets_ext import is_offsets_type
from bodo.hiframes.pd_series_ext import HeterogeneousSeriesType, SeriesType, if_series_to_array_type, is_series_type
from bodo.hiframes.pd_timestamp_ext import PandasTimestampType, pd_timestamp_type
from bodo.hiframes.rolling import is_supported_shift_array_type
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import BinaryArrayType, binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import BooleanArrayType, boolean_array
from bodo.libs.decimal_arr_ext import Decimal128Type, DecimalArrayType
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import StringArrayType, string_array_type
from bodo.libs.str_ext import string_type
from bodo.utils.transform import gen_const_tup, is_var_size_item_array_type
from bodo.utils.typing import BodoError, can_replace, check_unsupported_args, dtype_to_array_type, element_type, get_common_scalar_dtype, get_literal_value, get_overload_const_bytes, get_overload_const_int, get_overload_const_str, is_common_scalar_dtype, is_iterable_type, is_literal_type, is_nullable_type, is_overload_bool, is_overload_constant_bool, is_overload_constant_bytes, is_overload_constant_int, is_overload_constant_nan, is_overload_constant_str, is_overload_false, is_overload_int, is_overload_none, is_overload_true, is_overload_zero, is_scalar_type, raise_bodo_error, to_nullable_type


@overload_attribute(HeterogeneousSeriesType, 'index', inline='always')
@overload_attribute(SeriesType, 'index', inline='always')
def overload_series_index(s):
    return lambda s: bodo.hiframes.pd_series_ext.get_series_index(s)


@overload_attribute(HeterogeneousSeriesType, 'values', inline='always')
@overload_attribute(SeriesType, 'values', inline='always')
def overload_series_values(s):
    return lambda s: bodo.hiframes.pd_series_ext.get_series_data(s)


@overload_attribute(SeriesType, 'dtype', inline='always')
def overload_series_dtype(s):
    if s.dtype == bodo.string_type:
        raise BodoError('Series.dtype not supported for string Series yet')
    return lambda s: bodo.hiframes.pd_series_ext.get_series_data(s).dtype


@overload_attribute(HeterogeneousSeriesType, 'shape')
@overload_attribute(SeriesType, 'shape')
def overload_series_shape(s):
    return lambda s: (len(bodo.hiframes.pd_series_ext.get_series_data(s)),)


@overload_attribute(HeterogeneousSeriesType, 'ndim', inline='always')
@overload_attribute(SeriesType, 'ndim', inline='always')
def overload_series_ndim(s):
    return lambda s: 1


@overload_attribute(HeterogeneousSeriesType, 'size')
@overload_attribute(SeriesType, 'size')
def overload_series_size(s):
    return lambda s: len(bodo.hiframes.pd_series_ext.get_series_data(s))


@overload_attribute(HeterogeneousSeriesType, 'T', inline='always')
@overload_attribute(SeriesType, 'T', inline='always')
def overload_series_T(s):
    return lambda s: s


@overload_attribute(SeriesType, 'hasnans', inline='always')
def overload_series_hasnans(s):
    return lambda s: s.isna().sum() != 0


@overload_attribute(HeterogeneousSeriesType, 'empty')
@overload_attribute(SeriesType, 'empty')
def overload_series_empty(s):
    return lambda s: len(bodo.hiframes.pd_series_ext.get_series_data(s)) == 0


@overload_attribute(SeriesType, 'dtypes', inline='always')
def overload_series_dtypes(s):
    return lambda s: s.dtype


@overload_attribute(HeterogeneousSeriesType, 'name', inline='always')
@overload_attribute(SeriesType, 'name', inline='always')
def overload_series_name(s):
    return lambda s: bodo.hiframes.pd_series_ext.get_series_name(s)


@overload(len, no_unliteral=True)
def overload_series_len(S):
    if isinstance(S, (SeriesType, HeterogeneousSeriesType)):
        return lambda S: len(bodo.hiframes.pd_series_ext.get_series_data(S))


@overload_method(SeriesType, 'copy', inline='always', no_unliteral=True)
def overload_series_copy(S, deep=True):
    if is_overload_true(deep):

        def impl1(S, deep=True):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            return bodo.hiframes.pd_series_ext.init_series(arr.copy(),
                index, name)
        return impl1
    if is_overload_false(deep):

        def impl2(S, deep=True):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            return bodo.hiframes.pd_series_ext.init_series(arr, index, name)
        return impl2

    def impl(S, deep=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        if deep:
            arr = arr.copy()
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(arr, index, name)
    return impl


@overload_method(SeriesType, 'to_list', no_unliteral=True)
@overload_method(SeriesType, 'tolist', no_unliteral=True)
def overload_series_to_list(S):
    if isinstance(S.dtype, types.Float):

        def impl_float(S):
            tjrjs__xty = list()
            for dot__kzj in range(len(S)):
                tjrjs__xty.append(S.iat[dot__kzj])
            return tjrjs__xty
        return impl_float

    def impl(S):
        tjrjs__xty = list()
        for dot__kzj in range(len(S)):
            if bodo.libs.array_kernels.isna(S.values, dot__kzj):
                raise ValueError(
                    'Series.to_list(): Not supported for NA values with non-float dtypes'
                    )
            tjrjs__xty.append(S.iat[dot__kzj])
        return tjrjs__xty
    return impl


@overload_method(SeriesType, 'to_numpy', inline='always', no_unliteral=True)
def overload_series_to_numpy(S, dtype=None, copy=False, na_value=None):
    vxivd__glivu = dict(dtype=dtype, copy=copy, na_value=na_value)
    prr__rqtl = dict(dtype=None, copy=False, na_value=None)
    check_unsupported_args('Series.to_numpy', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, dtype=None, copy=False, na_value=None):
        return S.values
    return impl


@overload_method(SeriesType, 'reset_index', inline='always', no_unliteral=True)
def overload_series_reset_index(S, level=None, drop=False, name=None,
    inplace=False):
    vxivd__glivu = dict(name=name, inplace=inplace)
    prr__rqtl = dict(name=None, inplace=False)
    check_unsupported_args('Series.reset_index', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not bodo.hiframes.dataframe_impl._is_all_levels(S, level):
        raise_bodo_error(
            'Series.reset_index(): only dropping all index levels supported')
    if not is_overload_constant_bool(drop):
        raise_bodo_error(
            "Series.reset_index(): 'drop' parameter should be a constant boolean value"
            )
    if is_overload_true(drop):

        def impl_drop(S, level=None, drop=False, name=None, inplace=False):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_index_ext.init_range_index(0, len(arr),
                1, None)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            return bodo.hiframes.pd_series_ext.init_series(arr, index, name)
        return impl_drop

    def get_name_literal(name_typ, is_index=False, series_name=None):
        if is_overload_none(name_typ):
            if is_index:
                return 'index' if series_name != 'index' else 'level_0'
            return 0
        if is_literal_type(name_typ):
            return get_literal_value(name_typ)
        else:
            raise BodoError(
                'Series.reset_index() not supported for non-literal series names'
                )
    series_name = get_name_literal(S.name_typ)
    xojo__cvvhb = get_name_literal(S.index.name_typ, True, series_name)
    columns = [xojo__cvvhb, series_name]
    utmrs__nduqj = (
        'def _impl(S, level=None, drop=False, name=None, inplace=False):\n')
    utmrs__nduqj += (
        '    arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    utmrs__nduqj += """    index = bodo.utils.conversion.index_to_array(bodo.hiframes.pd_series_ext.get_series_index(S))
"""
    utmrs__nduqj += """    df_index = bodo.hiframes.pd_index_ext.init_range_index(0, len(S), 1, None)
"""
    utmrs__nduqj += '    col_var = {}\n'.format(gen_const_tup(columns))
    utmrs__nduqj += """    return bodo.hiframes.pd_dataframe_ext.init_dataframe((index, arr), df_index, col_var)
"""
    snqc__makr = {}
    exec(utmrs__nduqj, {'bodo': bodo}, snqc__makr)
    ieyvt__msc = snqc__makr['_impl']
    return ieyvt__msc


@overload_method(SeriesType, 'isna', inline='always', no_unliteral=True)
@overload_method(SeriesType, 'isnull', inline='always', no_unliteral=True)
def overload_series_isna(S):

    def impl(S):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo = bodo.libs.array_ops.array_op_isna(arr)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'round', inline='always', no_unliteral=True)
def overload_series_round(S, decimals=0):

    def impl(S, decimals=0):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        numba.parfors.parfor.init_prange()
        n = len(arr)
        ivga__befgo = bodo.utils.utils.alloc_type(n, arr, (-1,))
        for dot__kzj in numba.parfors.parfor.internal_prange(n):
            if pd.isna(arr[dot__kzj]):
                bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
            else:
                ivga__befgo[dot__kzj] = np.round(arr[dot__kzj], decimals)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'sum', inline='always', no_unliteral=True)
def overload_series_sum(S, axis=None, skipna=True, level=None, numeric_only
    =None, min_count=0):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.sum', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.sum(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.sum(): skipna argument must be a boolean')
    if not is_overload_int(min_count):
        raise BodoError('Series.sum(): min_count argument must be an integer')

    def impl(S, axis=None, skipna=True, level=None, numeric_only=None,
        min_count=0):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_sum(arr, skipna, min_count)
    return impl


@overload_method(SeriesType, 'prod', inline='always', no_unliteral=True)
@overload_method(SeriesType, 'product', inline='always', no_unliteral=True)
def overload_series_prod(S, axis=None, skipna=True, level=None,
    numeric_only=None, min_count=0):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.product', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.product(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.product(): skipna argument must be a boolean')

    def impl(S, axis=None, skipna=True, level=None, numeric_only=None,
        min_count=0):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_prod(arr, skipna, min_count)
    return impl


@overload_method(SeriesType, 'any', inline='always', no_unliteral=True)
def overload_series_any(S, axis=0, bool_only=None, skipna=True, level=None):
    vxivd__glivu = dict(axis=axis, bool_only=bool_only, skipna=skipna,
        level=level)
    prr__rqtl = dict(axis=0, bool_only=None, skipna=True, level=None)
    check_unsupported_args('Series.any', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, axis=0, bool_only=None, skipna=True, level=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        numba.parfors.parfor.init_prange()
        plvb__sxq = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = 0
            if not bodo.libs.array_kernels.isna(A, dot__kzj):
                jrrz__tkvr = int(A[dot__kzj])
            plvb__sxq += jrrz__tkvr
        return plvb__sxq != 0
    return impl


@overload_method(SeriesType, 'equals', inline='always', no_unliteral=True)
def overload_series_equals(S, other):
    if not isinstance(other, SeriesType):
        raise BodoError("Series.equals() 'other' must be a Series")
    if isinstance(S.data, bodo.ArrayItemArrayType):
        raise BodoError(
            'Series.equals() not supported for Series where each element is an array or list'
            )
    if S.data != other.data:
        return lambda S, other: False

    def impl(S, other):
        qvcs__lldbm = bodo.hiframes.pd_series_ext.get_series_data(S)
        gqdmu__ywt = bodo.hiframes.pd_series_ext.get_series_data(other)
        numba.parfors.parfor.init_prange()
        plvb__sxq = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(qvcs__lldbm)):
            jrrz__tkvr = 0
            uqe__wgevi = bodo.libs.array_kernels.isna(qvcs__lldbm, dot__kzj)
            mtxf__fbult = bodo.libs.array_kernels.isna(gqdmu__ywt, dot__kzj)
            if (uqe__wgevi and not mtxf__fbult or not uqe__wgevi and
                mtxf__fbult):
                jrrz__tkvr = 1
            elif not uqe__wgevi:
                if qvcs__lldbm[dot__kzj] != gqdmu__ywt[dot__kzj]:
                    jrrz__tkvr = 1
            plvb__sxq += jrrz__tkvr
        return plvb__sxq == 0
    return impl


@overload_method(SeriesType, 'all', inline='always', no_unliteral=True)
def overload_series_all(S, axis=0, bool_only=None, skipna=True, level=None):
    vxivd__glivu = dict(axis=axis, bool_only=bool_only, skipna=skipna,
        level=level)
    prr__rqtl = dict(axis=0, bool_only=None, skipna=True, level=None)
    check_unsupported_args('Series.all', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, axis=0, bool_only=None, skipna=True, level=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        numba.parfors.parfor.init_prange()
        plvb__sxq = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = 0
            if not bodo.libs.array_kernels.isna(A, dot__kzj):
                jrrz__tkvr = int(not A[dot__kzj])
            plvb__sxq += jrrz__tkvr
        return plvb__sxq == 0
    return impl


@overload_method(SeriesType, 'mad', inline='always', no_unliteral=True)
def overload_series_mad(S, axis=None, skipna=True, level=None):
    vxivd__glivu = dict(level=level)
    prr__rqtl = dict(level=None)
    check_unsupported_args('Series.mad', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_bool(skipna):
        raise BodoError("Series.mad(): 'skipna' argument must be a boolean")
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.mad(): axis argument not supported')
    hiijx__tdkv = types.float64
    htbxp__jtk = types.float64
    if S.dtype == types.float32:
        hiijx__tdkv = types.float32
        htbxp__jtk = types.float32
    yoi__iqofg = hiijx__tdkv(0)
    smym__uex = htbxp__jtk(0)
    razk__apfoe = htbxp__jtk(1)

    def impl(S, axis=None, skipna=True, level=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        numba.parfors.parfor.init_prange()
        hrslf__bbh = yoi__iqofg
        plvb__sxq = smym__uex
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = yoi__iqofg
            ftfmt__bua = smym__uex
            if not bodo.libs.array_kernels.isna(A, dot__kzj) or not skipna:
                jrrz__tkvr = A[dot__kzj]
                ftfmt__bua = razk__apfoe
            hrslf__bbh += jrrz__tkvr
            plvb__sxq += ftfmt__bua
        ajx__fbu = bodo.hiframes.series_kernels._mean_handle_nan(hrslf__bbh,
            plvb__sxq)
        wpmpu__zsz = yoi__iqofg
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = yoi__iqofg
            if not bodo.libs.array_kernels.isna(A, dot__kzj) or not skipna:
                jrrz__tkvr = abs(A[dot__kzj] - ajx__fbu)
            wpmpu__zsz += jrrz__tkvr
        iktgm__mnll = bodo.hiframes.series_kernels._mean_handle_nan(wpmpu__zsz,
            plvb__sxq)
        return iktgm__mnll
    return impl


@overload_method(SeriesType, 'mean', inline='always', no_unliteral=True)
def overload_series_mean(S, axis=None, skipna=None, level=None,
    numeric_only=None):
    if not isinstance(S.dtype, types.Number) and S.dtype not in [bodo.
        datetime64ns, types.bool_]:
        raise BodoError(f"Series.mean(): Series with type '{S}' not supported")
    vxivd__glivu = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    prr__rqtl = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('Series.mean', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.mean(): axis argument not supported')

    def impl(S, axis=None, skipna=None, level=None, numeric_only=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_mean(arr)
    return impl


@overload_method(SeriesType, 'sem', inline='always', no_unliteral=True)
def overload_series_sem(S, axis=None, skipna=True, level=None, ddof=1,
    numeric_only=None):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.sem', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.sem(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.sem(): skipna argument must be a boolean')
    if not is_overload_int(ddof):
        raise BodoError('Series.sem(): ddof argument must be an integer')

    def impl(S, axis=None, skipna=True, level=None, ddof=1, numeric_only=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        numba.parfors.parfor.init_prange()
        peazu__iwgxq = 0
        gleio__qxle = 0
        plvb__sxq = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = 0
            ftfmt__bua = 0
            if not bodo.libs.array_kernels.isna(A, dot__kzj) or not skipna:
                jrrz__tkvr = A[dot__kzj]
                ftfmt__bua = 1
            peazu__iwgxq += jrrz__tkvr
            gleio__qxle += jrrz__tkvr * jrrz__tkvr
            plvb__sxq += ftfmt__bua
        xzb__vwpz = bodo.hiframes.series_kernels._compute_var_nan_count_ddof(
            peazu__iwgxq, gleio__qxle, plvb__sxq, ddof)
        emrja__npne = bodo.hiframes.series_kernels._sem_handle_nan(xzb__vwpz,
            plvb__sxq)
        return emrja__npne
    return impl


@overload_method(SeriesType, 'kurt', inline='always', no_unliteral=True)
@overload_method(SeriesType, 'kurtosis', inline='always', no_unliteral=True)
def overload_series_kurt(S, axis=None, skipna=True, level=None,
    numeric_only=None):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.kurtosis', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.kurtosis(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError(
            "Series.kurtosis(): 'skipna' argument must be a boolean")

    def impl(S, axis=None, skipna=True, level=None, numeric_only=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        numba.parfors.parfor.init_prange()
        peazu__iwgxq = 0.0
        gleio__qxle = 0.0
        tlnx__pqmb = 0.0
        sjgdt__xkp = 0.0
        plvb__sxq = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = 0.0
            ftfmt__bua = 0
            if not bodo.libs.array_kernels.isna(A, dot__kzj) or not skipna:
                jrrz__tkvr = np.float64(A[dot__kzj])
                ftfmt__bua = 1
            peazu__iwgxq += jrrz__tkvr
            gleio__qxle += jrrz__tkvr ** 2
            tlnx__pqmb += jrrz__tkvr ** 3
            sjgdt__xkp += jrrz__tkvr ** 4
            plvb__sxq += ftfmt__bua
        xzb__vwpz = bodo.hiframes.series_kernels.compute_kurt(peazu__iwgxq,
            gleio__qxle, tlnx__pqmb, sjgdt__xkp, plvb__sxq)
        return xzb__vwpz
    return impl


@overload_method(SeriesType, 'skew', inline='always', no_unliteral=True)
def overload_series_skew(S, axis=None, skipna=True, level=None,
    numeric_only=None):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.skew', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.skew(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.skew(): skipna argument must be a boolean')

    def impl(S, axis=None, skipna=True, level=None, numeric_only=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        numba.parfors.parfor.init_prange()
        peazu__iwgxq = 0.0
        gleio__qxle = 0.0
        tlnx__pqmb = 0.0
        plvb__sxq = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(A)):
            jrrz__tkvr = 0.0
            ftfmt__bua = 0
            if not bodo.libs.array_kernels.isna(A, dot__kzj) or not skipna:
                jrrz__tkvr = np.float64(A[dot__kzj])
                ftfmt__bua = 1
            peazu__iwgxq += jrrz__tkvr
            gleio__qxle += jrrz__tkvr ** 2
            tlnx__pqmb += jrrz__tkvr ** 3
            plvb__sxq += ftfmt__bua
        xzb__vwpz = bodo.hiframes.series_kernels.compute_skew(peazu__iwgxq,
            gleio__qxle, tlnx__pqmb, plvb__sxq)
        return xzb__vwpz
    return impl


@overload_method(SeriesType, 'var', inline='always', no_unliteral=True)
def overload_series_var(S, axis=None, skipna=True, level=None, ddof=1,
    numeric_only=None):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.var', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.var(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.var(): skipna argument must be a boolean')
    if not is_overload_int(ddof):
        raise BodoError('Series.var(): ddof argument must be an integer')

    def impl(S, axis=None, skipna=True, level=None, ddof=1, numeric_only=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_var(arr, skipna, ddof)
    return impl


@overload_method(SeriesType, 'std', inline='always', no_unliteral=True)
def overload_series_std(S, axis=None, skipna=True, level=None, ddof=1,
    numeric_only=None):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.std', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.std(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.std(): skipna argument must be a boolean')
    if not is_overload_int(ddof):
        raise BodoError('Series.std(): ddof argument must be an integer')

    def impl(S, axis=None, skipna=True, level=None, ddof=1, numeric_only=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_std(arr, skipna, ddof)
    return impl


@overload_method(SeriesType, 'dot', inline='always', no_unliteral=True)
def overload_series_dot(S, other):

    def impl(S, other):
        qvcs__lldbm = bodo.hiframes.pd_series_ext.get_series_data(S)
        gqdmu__ywt = bodo.hiframes.pd_series_ext.get_series_data(other)
        numba.parfors.parfor.init_prange()
        poj__orx = 0
        for dot__kzj in numba.parfors.parfor.internal_prange(len(qvcs__lldbm)):
            exda__dfhmt = qvcs__lldbm[dot__kzj]
            soy__grq = gqdmu__ywt[dot__kzj]
            poj__orx += exda__dfhmt * soy__grq
        return poj__orx
    return impl


@overload_method(SeriesType, 'cumsum', inline='always', no_unliteral=True)
def overload_series_cumsum(S, axis=None, skipna=True):
    vxivd__glivu = dict(skipna=skipna)
    prr__rqtl = dict(skipna=True)
    check_unsupported_args('Series.cumsum', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.cumsum(): axis argument not supported')

    def impl(S, axis=None, skipna=True):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(A.cumsum(), index, name)
    return impl


@overload_method(SeriesType, 'cumprod', inline='always', no_unliteral=True)
def overload_series_cumprod(S, axis=None, skipna=True):
    vxivd__glivu = dict(skipna=skipna)
    prr__rqtl = dict(skipna=True)
    check_unsupported_args('Series.cumprod', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.cumprod(): axis argument not supported')

    def impl(S, axis=None, skipna=True):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(A.cumprod(), index, name
            )
    return impl


@overload_method(SeriesType, 'cummin', inline='always', no_unliteral=True)
def overload_series_cummin(S, axis=None, skipna=True):
    vxivd__glivu = dict(skipna=skipna)
    prr__rqtl = dict(skipna=True)
    check_unsupported_args('Series.cummin', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.cummin(): axis argument not supported')

    def impl(S, axis=None, skipna=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(bodo.libs.
            array_kernels.cummin(arr), index, name)
    return impl


@overload_method(SeriesType, 'cummax', inline='always', no_unliteral=True)
def overload_series_cummax(S, axis=None, skipna=True):
    vxivd__glivu = dict(skipna=skipna)
    prr__rqtl = dict(skipna=True)
    check_unsupported_args('Series.cummax', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.cummax(): axis argument not supported')

    def impl(S, axis=None, skipna=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(bodo.libs.
            array_kernels.cummax(arr), index, name)
    return impl


@overload_method(SeriesType, 'rename', inline='always', no_unliteral=True)
def overload_series_rename(S, index=None, axis=None, copy=True, inplace=
    False, level=None, errors='ignore'):
    if not (index == bodo.string_type or isinstance(index, types.StringLiteral)
        ):
        raise BodoError("Series.rename() 'index' can only be a string")
    vxivd__glivu = dict(copy=copy, inplace=inplace, level=level, errors=errors)
    prr__rqtl = dict(copy=True, inplace=False, level=None, errors='ignore')
    check_unsupported_args('Series.rename', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, index=None, axis=None, copy=True, inplace=False, level=None,
        errors='ignore'):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        ijh__aul = bodo.hiframes.pd_series_ext.get_series_index(S)
        return bodo.hiframes.pd_series_ext.init_series(A, ijh__aul, index)
    return impl


@overload_method(SeriesType, 'rename_axis', inline='always', no_unliteral=True)
def overload_series_rename_axis(S, mapper=None, index=None, columns=None,
    axis=None, copy=True, inplace=False):
    vxivd__glivu = dict(index=index, columns=columns, axis=axis, copy=copy,
        inplace=inplace)
    prr__rqtl = dict(index=None, columns=None, axis=None, copy=True,
        inplace=False)
    check_unsupported_args('Series.rename_axis', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if is_overload_none(mapper) or not is_scalar_type(mapper):
        raise BodoError(
            "Series.rename_axis(): 'mapper' is required and must be a scalar type."
            )

    def impl(S, mapper=None, index=None, columns=None, axis=None, copy=True,
        inplace=False):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        index = index.rename(mapper)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(arr, index, name)
    return impl


@overload_method(SeriesType, 'abs', inline='always', no_unliteral=True)
def overload_series_abs(S):

    def impl(S):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(np.abs(A), index, name)
    return impl


@overload_method(SeriesType, 'count', no_unliteral=True)
def overload_series_count(S, level=None):
    vxivd__glivu = dict(level=level)
    prr__rqtl = dict(level=None)
    check_unsupported_args('Series.count', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, level=None):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_count(A)
    return impl


@overload_method(SeriesType, 'corr', inline='always', no_unliteral=True)
def overload_series_corr(S, other, method='pearson', min_periods=None):
    vxivd__glivu = dict(method=method, min_periods=min_periods)
    prr__rqtl = dict(method='pearson', min_periods=None)
    check_unsupported_args('Series.corr', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, other, method='pearson', min_periods=None):
        n = S.count()
        kxnc__qcz = S.sum()
        wbpqr__tqnwu = other.sum()
        a = n * (S * other).sum() - kxnc__qcz * wbpqr__tqnwu
        rsr__xpjkz = n * (S ** 2).sum() - kxnc__qcz ** 2
        khum__mbu = n * (other ** 2).sum() - wbpqr__tqnwu ** 2
        return a / np.sqrt(rsr__xpjkz * khum__mbu)
    return impl


@overload_method(SeriesType, 'cov', inline='always', no_unliteral=True)
def overload_series_cov(S, other, min_periods=None, ddof=1):
    vxivd__glivu = dict(min_periods=min_periods)
    prr__rqtl = dict(min_periods=None)
    check_unsupported_args('Series.cov', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, other, min_periods=None, ddof=1):
        kxnc__qcz = S.mean()
        wbpqr__tqnwu = other.mean()
        smeom__zjsg = ((S - kxnc__qcz) * (other - wbpqr__tqnwu)).sum()
        N = np.float64(S.count() - ddof)
        nonzero_len = S.count() * other.count()
        return _series_cov_helper(smeom__zjsg, N, nonzero_len)
    return impl


def _series_cov_helper(sum_val, N, nonzero_len):
    return


@overload(_series_cov_helper, no_unliteral=True)
def _overload_series_cov_helper(sum_val, N, nonzero_len):

    def impl(sum_val, N, nonzero_len):
        if not nonzero_len:
            return np.nan
        if N <= 0.0:
            cqir__tpa = np.sign(sum_val)
            return np.inf * cqir__tpa
        return sum_val / N
    return impl


@overload_method(SeriesType, 'min', inline='always', no_unliteral=True)
def overload_series_min(S, axis=None, skipna=None, level=None, numeric_only
    =None):
    vxivd__glivu = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    prr__rqtl = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('Series.min', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.min(): axis argument not supported')
    if isinstance(S.dtype, PDCategoricalDtype):
        if not S.dtype.ordered:
            raise BodoError(
                'Series.min(): only ordered categoricals are possible')

    def impl(S, axis=None, skipna=None, level=None, numeric_only=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_min(arr)
    return impl


@overload(max, no_unliteral=True)
def overload_series_builtins_max(S):
    if isinstance(S, SeriesType):

        def impl(S):
            return S.max()
        return impl


@overload(min, no_unliteral=True)
def overload_series_builtins_min(S):
    if isinstance(S, SeriesType):

        def impl(S):
            return S.min()
        return impl


@overload(sum, no_unliteral=True)
def overload_series_builtins_sum(S):
    if isinstance(S, SeriesType):

        def impl(S):
            return S.sum()
        return impl


@overload(np.prod, inline='always', no_unliteral=True)
def overload_series_np_prod(S):
    if isinstance(S, SeriesType):

        def impl(S):
            return S.prod()
        return impl


@overload_method(SeriesType, 'max', inline='always', no_unliteral=True)
def overload_series_max(S, axis=None, skipna=None, level=None, numeric_only
    =None):
    vxivd__glivu = dict(skipna=skipna, level=level, numeric_only=numeric_only)
    prr__rqtl = dict(skipna=None, level=None, numeric_only=None)
    check_unsupported_args('Series.max', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.max(): axis argument not supported')
    if isinstance(S.dtype, PDCategoricalDtype):
        if not S.dtype.ordered:
            raise BodoError(
                'Series.max(): only ordered categoricals are possible')

    def impl(S, axis=None, skipna=None, level=None, numeric_only=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_ops.array_op_max(arr)
    return impl


@overload_method(SeriesType, 'idxmin', inline='always', no_unliteral=True)
def overload_series_idxmin(S, axis=0, skipna=True):
    vxivd__glivu = dict(axis=axis, skipna=skipna)
    prr__rqtl = dict(axis=0, skipna=True)
    check_unsupported_args('Series.idxmin', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (S.dtype == types.none or bodo.utils.utils.is_np_array_typ(S.
        data) and (S.dtype in [bodo.datetime64ns, bodo.timedelta64ns] or
        isinstance(S.dtype, (types.Number, types.Boolean))) or isinstance(S
        .data, (bodo.IntegerArrayType, bodo.CategoricalArrayType)) or S.
        data in [bodo.boolean_array, bodo.datetime_date_array_type]):
        raise BodoError(
            f'Series.idxmin() only supported for numeric array types. Array type: {S.data} not supported.'
            )
    if isinstance(S.data, bodo.CategoricalArrayType) and not S.dtype.ordered:
        raise BodoError(
            'Series.idxmin(): only ordered categoricals are possible')

    def impl(S, axis=0, skipna=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        return bodo.libs.array_ops.array_op_idxmin(arr, index)
    return impl


@overload_method(SeriesType, 'idxmax', inline='always', no_unliteral=True)
def overload_series_idxmax(S, axis=0, skipna=True):
    vxivd__glivu = dict(axis=axis, skipna=skipna)
    prr__rqtl = dict(axis=0, skipna=True)
    check_unsupported_args('Series.idxmax', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (S.dtype == types.none or bodo.utils.utils.is_np_array_typ(S.
        data) and (S.dtype in [bodo.datetime64ns, bodo.timedelta64ns] or
        isinstance(S.dtype, (types.Number, types.Boolean))) or isinstance(S
        .data, (bodo.IntegerArrayType, bodo.CategoricalArrayType)) or S.
        data in [bodo.boolean_array, bodo.datetime_date_array_type]):
        raise BodoError(
            f'Series.idxmax() only supported for numeric array types. Array type: {S.data} not supported.'
            )
    if isinstance(S.data, bodo.CategoricalArrayType) and not S.dtype.ordered:
        raise BodoError(
            'Series.idxmax(): only ordered categoricals are possible')

    def impl(S, axis=0, skipna=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        return bodo.libs.array_ops.array_op_idxmax(arr, index)
    return impl


@overload_method(SeriesType, 'infer_objects', inline='always')
def overload_series_infer_objects(S):
    return lambda S: S.copy()


@overload_attribute(SeriesType, 'is_monotonic', inline='always')
@overload_attribute(SeriesType, 'is_monotonic_increasing', inline='always')
def overload_series_is_monotonic_increasing(S):
    return lambda S: bodo.libs.array_kernels.series_monotonicity(bodo.
        hiframes.pd_series_ext.get_series_data(S), 1)


@overload_attribute(SeriesType, 'is_monotonic_decreasing', inline='always')
def overload_series_is_monotonic_decreasing(S):
    return lambda S: bodo.libs.array_kernels.series_monotonicity(bodo.
        hiframes.pd_series_ext.get_series_data(S), 2)


@overload_attribute(SeriesType, 'nbytes', inline='always')
def overload_series_nbytes(S):
    return lambda S: bodo.hiframes.pd_series_ext.get_series_data(S).nbytes


@overload_method(SeriesType, 'autocorr', inline='always', no_unliteral=True)
def overload_series_autocorr(S, lag=1):
    return lambda S, lag=1: bodo.libs.array_kernels.autocorr(bodo.hiframes.
        pd_series_ext.get_series_data(S), lag)


@overload_method(SeriesType, 'median', inline='always', no_unliteral=True)
def overload_series_median(S, axis=None, skipna=True, level=None,
    numeric_only=None):
    vxivd__glivu = dict(level=level, numeric_only=numeric_only)
    prr__rqtl = dict(level=None, numeric_only=None)
    check_unsupported_args('Series.median', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.median(): axis argument not supported')
    if not is_overload_bool(skipna):
        raise BodoError('Series.median(): skipna argument must be a boolean')
    return (lambda S, axis=None, skipna=True, level=None, numeric_only=None:
        bodo.libs.array_ops.array_op_median(bodo.hiframes.pd_series_ext.
        get_series_data(S), skipna))


def overload_series_head(S, n=5):

    def impl(S, n=5):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        yke__xvxor = arr[:n]
        rxf__zgot = index[:n]
        return bodo.hiframes.pd_series_ext.init_series(yke__xvxor,
            rxf__zgot, name)
    return impl


@lower_builtin('series.head', SeriesType, types.Integer)
@lower_builtin('series.head', SeriesType, types.Omitted)
def series_head_lower(context, builder, sig, args):
    impl = overload_series_head(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


@numba.extending.register_jitable
def tail_slice(k, n):
    if n == 0:
        return k
    return -n


@overload_method(SeriesType, 'tail', inline='always', no_unliteral=True)
def overload_series_tail(S, n=5):
    if not is_overload_int(n):
        raise BodoError("Series.tail(): 'n' must be an Integer")

    def impl(S, n=5):
        kcx__pqvx = tail_slice(len(S), n)
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        yke__xvxor = arr[kcx__pqvx:]
        rxf__zgot = index[kcx__pqvx:]
        return bodo.hiframes.pd_series_ext.init_series(yke__xvxor,
            rxf__zgot, name)
    return impl


@overload_method(SeriesType, 'first', inline='always', no_unliteral=True)
def overload_series_first(S, offset):
    lzfx__gjak = (types.unicode_type, bodo.month_begin_type, bodo.
        month_end_type, bodo.week_type, bodo.date_offset_type)
    if types.unliteral(offset) not in lzfx__gjak:
        raise BodoError(
            "Series.first(): 'offset' must be a string or a DateOffset")

    def impl(S, offset):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        if len(index):
            pzfas__zilx = index[0]
            xqh__dux = (bodo.libs.array_kernels.
                get_valid_entries_from_date_offset(index, offset,
                pzfas__zilx, False))
        else:
            xqh__dux = 0
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        yke__xvxor = arr[:xqh__dux]
        rxf__zgot = index[:xqh__dux]
        return bodo.hiframes.pd_series_ext.init_series(yke__xvxor,
            rxf__zgot, name)
    return impl


@overload_method(SeriesType, 'last', inline='always', no_unliteral=True)
def overload_series_last(S, offset):
    lzfx__gjak = (types.unicode_type, bodo.month_begin_type, bodo.
        month_end_type, bodo.week_type, bodo.date_offset_type)
    if types.unliteral(offset) not in lzfx__gjak:
        raise BodoError(
            "Series.last(): 'offset' must be a string or a DateOffset")

    def impl(S, offset):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        if len(index):
            dsaw__dtxsm = index[-1]
            xqh__dux = (bodo.libs.array_kernels.
                get_valid_entries_from_date_offset(index, offset,
                dsaw__dtxsm, True))
        else:
            xqh__dux = 0
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        yke__xvxor = arr[len(arr) - xqh__dux:]
        rxf__zgot = index[len(arr) - xqh__dux:]
        return bodo.hiframes.pd_series_ext.init_series(yke__xvxor,
            rxf__zgot, name)
    return impl


@overload_method(SeriesType, 'nlargest', inline='always', no_unliteral=True)
def overload_series_nlargest(S, n=5, keep='first'):
    vxivd__glivu = dict(keep=keep)
    prr__rqtl = dict(keep='first')
    check_unsupported_args('Series.nlargest', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_int(n):
        raise BodoError('Series.nlargest(): n argument must be an integer')

    def impl(S, n=5, keep='first'):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        hls__iymkx = bodo.utils.conversion.coerce_to_ndarray(index)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo, vpk__rcjxf = bodo.libs.array_kernels.nlargest(arr,
            hls__iymkx, n, True, bodo.hiframes.series_kernels.gt_f)
        juqvh__wrsfg = bodo.utils.conversion.convert_to_index(vpk__rcjxf)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl


@overload_method(SeriesType, 'nsmallest', inline='always', no_unliteral=True)
def overload_series_nsmallest(S, n=5, keep='first'):
    vxivd__glivu = dict(keep=keep)
    prr__rqtl = dict(keep='first')
    check_unsupported_args('Series.nsmallest', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_int(n):
        raise BodoError('Series.nsmallest(): n argument must be an integer')

    def impl(S, n=5, keep='first'):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        hls__iymkx = bodo.utils.conversion.coerce_to_ndarray(index)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo, vpk__rcjxf = bodo.libs.array_kernels.nlargest(arr,
            hls__iymkx, n, False, bodo.hiframes.series_kernels.lt_f)
        juqvh__wrsfg = bodo.utils.conversion.convert_to_index(vpk__rcjxf)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl


@overload_method(SeriesType, 'notnull', inline='always', no_unliteral=True)
@overload_method(SeriesType, 'notna', inline='always', no_unliteral=True)
def overload_series_notna(S):
    return lambda S: S.isna() == False


@overload_method(SeriesType, 'astype', inline='always', no_unliteral=True)
def overload_series_astype(S, dtype, copy=True, errors='raise',
    _bodo_nan_to_str=True):
    vxivd__glivu = dict(errors=errors)
    prr__rqtl = dict(errors='raise')
    check_unsupported_args('Series.astype', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if dtype == types.unicode_type:
        raise_bodo_error(
            "Series.astype(): 'dtype' when passed as string must be a constant value"
            )

    def impl(S, dtype, copy=True, errors='raise', _bodo_nan_to_str=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo = bodo.utils.conversion.fix_arr_dtype(arr, dtype, copy,
            nan_to_str=_bodo_nan_to_str, from_series=True)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'take', inline='always', no_unliteral=True)
def overload_series_take(S, indices, axis=0, is_copy=True):
    vxivd__glivu = dict(axis=axis, is_copy=is_copy)
    prr__rqtl = dict(axis=0, is_copy=True)
    check_unsupported_args('Series.take', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_iterable_type(indices) and isinstance(indices.dtype, types.
        Integer)):
        raise BodoError(
            f"Series.take() 'indices' must be an array-like and contain integers. Found type {indices}."
            )

    def impl(S, indices, axis=0, is_copy=True):
        jwrek__qordb = bodo.utils.conversion.coerce_to_ndarray(indices)
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(arr[jwrek__qordb],
            index[jwrek__qordb], name)
    return impl


@overload_method(SeriesType, 'argsort', inline='always', no_unliteral=True)
def overload_series_argsort(S, axis=0, kind='quicksort', order=None):
    vxivd__glivu = dict(axis=axis, kind=kind, order=order)
    prr__rqtl = dict(axis=0, kind='quicksort', order=None)
    check_unsupported_args('Series.argsort', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')

    def impl(S, axis=0, kind='quicksort', order=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        n = len(arr)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        jwhz__rxze = S.notna().values
        if not jwhz__rxze.all():
            ivga__befgo = np.full(n, -1, np.int64)
            ivga__befgo[jwhz__rxze] = argsort(arr[jwhz__rxze])
        else:
            ivga__befgo = argsort(arr)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'sort_index', inline='always', no_unliteral=True)
def overload_series_sort_index(S, axis=0, level=None, ascending=True,
    inplace=False, kind='quicksort', na_position='last', sort_remaining=
    True, ignore_index=False, key=None):
    vxivd__glivu = dict(axis=axis, level=level, inplace=inplace, kind=kind,
        sort_remaining=sort_remaining, ignore_index=ignore_index, key=key)
    prr__rqtl = dict(axis=0, level=None, inplace=False, kind='quicksort',
        sort_remaining=True, ignore_index=False, key=None)
    check_unsupported_args('Series.sort_index', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_bool(ascending):
        raise BodoError(
            "Series.sort_index(): 'ascending' parameter must be of type bool")
    if not is_overload_constant_str(na_position) or get_overload_const_str(
        na_position) not in ('first', 'last'):
        raise_bodo_error(
            "Series.sort_index(): 'na_position' should either be 'first' or 'last'"
            )

    def impl(S, axis=0, level=None, ascending=True, inplace=False, kind=
        'quicksort', na_position='last', sort_remaining=True, ignore_index=
        False, key=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        pyxt__vqtlq = bodo.hiframes.pd_dataframe_ext.init_dataframe((arr,),
            index, ('$_bodo_col3_',))
        rvp__ofabl = pyxt__vqtlq.sort_index(ascending=ascending, inplace=
            inplace, na_position=na_position)
        ivga__befgo = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(
            rvp__ofabl, 0)
        juqvh__wrsfg = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(
            rvp__ofabl)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl


@overload_method(SeriesType, 'sort_values', inline='always', no_unliteral=True)
def overload_series_sort_values(S, axis=0, ascending=True, inplace=False,
    kind='quicksort', na_position='last', ignore_index=False, key=None):
    vxivd__glivu = dict(axis=axis, inplace=inplace, kind=kind, ignore_index
        =ignore_index, key=key)
    prr__rqtl = dict(axis=0, inplace=False, kind='quicksort', ignore_index=
        False, key=None)
    check_unsupported_args('Series.sort_values', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_bool(ascending):
        raise BodoError(
            "Series.sort_values(): 'ascending' parameter must be of type bool")
    if not is_overload_constant_str(na_position) or get_overload_const_str(
        na_position) not in ('first', 'last'):
        raise_bodo_error(
            "Series.sort_values(): 'na_position' should either be 'first' or 'last'"
            )

    def impl(S, axis=0, ascending=True, inplace=False, kind='quicksort',
        na_position='last', ignore_index=False, key=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        pyxt__vqtlq = bodo.hiframes.pd_dataframe_ext.init_dataframe((arr,),
            index, ('$_bodo_col_',))
        rvp__ofabl = pyxt__vqtlq.sort_values(['$_bodo_col_'], ascending=
            ascending, inplace=inplace, na_position=na_position)
        ivga__befgo = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(
            rvp__ofabl, 0)
        juqvh__wrsfg = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(
            rvp__ofabl)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl


def get_bin_inds(bins, arr):
    return arr


@overload(get_bin_inds, inline='always', no_unliteral=True)
def overload_get_bin_inds(bins, arr, is_nullable=True, include_lowest=True):
    assert is_overload_constant_bool(is_nullable)
    pvnu__dsp = is_overload_true(is_nullable)
    utmrs__nduqj = (
        'def impl(bins, arr, is_nullable=True, include_lowest=True):\n')
    utmrs__nduqj += '  numba.parfors.parfor.init_prange()\n'
    utmrs__nduqj += '  n = len(arr)\n'
    if pvnu__dsp:
        utmrs__nduqj += (
            '  out_arr = bodo.libs.int_arr_ext.alloc_int_array(n, np.int64)\n')
    else:
        utmrs__nduqj += '  out_arr = np.empty(n, np.int64)\n'
    utmrs__nduqj += '  for i in numba.parfors.parfor.internal_prange(n):\n'
    utmrs__nduqj += '    if bodo.libs.array_kernels.isna(arr, i):\n'
    if pvnu__dsp:
        utmrs__nduqj += '      bodo.libs.array_kernels.setna(out_arr, i)\n'
    else:
        utmrs__nduqj += '      out_arr[i] = -1\n'
    utmrs__nduqj += '      continue\n'
    utmrs__nduqj += '    val = arr[i]\n'
    utmrs__nduqj += '    if include_lowest and val == bins[0]:\n'
    utmrs__nduqj += '      ind = 1\n'
    utmrs__nduqj += '    else:\n'
    utmrs__nduqj += '      ind = np.searchsorted(bins, val)\n'
    utmrs__nduqj += '    if ind == 0 or ind == len(bins):\n'
    if pvnu__dsp:
        utmrs__nduqj += '      bodo.libs.array_kernels.setna(out_arr, i)\n'
    else:
        utmrs__nduqj += '      out_arr[i] = -1\n'
    utmrs__nduqj += '    else:\n'
    utmrs__nduqj += '      out_arr[i] = ind - 1\n'
    utmrs__nduqj += '  return out_arr\n'
    snqc__makr = {}
    exec(utmrs__nduqj, {'bodo': bodo, 'np': np, 'numba': numba}, snqc__makr)
    impl = snqc__makr['impl']
    return impl


@register_jitable
def _round_frac(x, precision: int):
    if not np.isfinite(x) or x == 0:
        return x
    else:
        mjcad__fyis, xlsbt__ynuka = np.divmod(x, 1)
        if mjcad__fyis == 0:
            ovdt__wfjos = -int(np.floor(np.log10(abs(xlsbt__ynuka)))
                ) - 1 + precision
        else:
            ovdt__wfjos = precision
        return np.around(x, ovdt__wfjos)


@register_jitable
def _infer_precision(base_precision: int, bins) ->int:
    for precision in range(base_precision, 20):
        hfliw__kbtjx = np.array([_round_frac(b, precision) for b in bins])
        if len(np.unique(hfliw__kbtjx)) == len(bins):
            return precision
    return base_precision


def get_bin_labels(bins):
    pass


@overload(get_bin_labels, no_unliteral=True)
def overload_get_bin_labels(bins, right=True, include_lowest=True):
    dtype = np.float64 if isinstance(bins.dtype, types.Integer) else bins.dtype
    if dtype == bodo.datetime64ns:
        dzfri__cqzp = bodo.timedelta64ns(1)

        def impl_dt64(bins, right=True, include_lowest=True):
            kyhuq__tnc = bins.copy()
            if right and include_lowest:
                kyhuq__tnc[0] = kyhuq__tnc[0] - dzfri__cqzp
            jlm__yda = bodo.libs.interval_arr_ext.init_interval_array(
                kyhuq__tnc[:-1], kyhuq__tnc[1:])
            return bodo.hiframes.pd_index_ext.init_interval_index(jlm__yda,
                None)
        return impl_dt64

    def impl(bins, right=True, include_lowest=True):
        base_precision = 3
        precision = _infer_precision(base_precision, bins)
        kyhuq__tnc = np.array([_round_frac(b, precision) for b in bins],
            dtype=dtype)
        if right and include_lowest:
            kyhuq__tnc[0] = kyhuq__tnc[0] - 10.0 ** -precision
        jlm__yda = bodo.libs.interval_arr_ext.init_interval_array(kyhuq__tnc
            [:-1], kyhuq__tnc[1:])
        return bodo.hiframes.pd_index_ext.init_interval_index(jlm__yda, None)
    return impl


def get_output_bin_counts(count_series, nbins):
    pass


@overload(get_output_bin_counts, no_unliteral=True)
def overload_get_output_bin_counts(count_series, nbins):

    def impl(count_series, nbins):
        ift__lbeze = bodo.hiframes.pd_series_ext.get_series_data(count_series)
        yipvg__kxor = bodo.utils.conversion.index_to_array(bodo.hiframes.
            pd_series_ext.get_series_index(count_series))
        ivga__befgo = np.zeros(nbins, np.int64)
        for dot__kzj in range(len(ift__lbeze)):
            ivga__befgo[yipvg__kxor[dot__kzj]] = ift__lbeze[dot__kzj]
        return ivga__befgo
    return impl


def compute_bins(nbins, min_val, max_val):
    pass


@overload(compute_bins, no_unliteral=True)
def overload_compute_bins(nbins, min_val, max_val, right=True):

    def impl(nbins, min_val, max_val, right=True):
        if nbins < 1:
            raise ValueError('`bins` should be a positive integer.')
        min_val = min_val + 0.0
        max_val = max_val + 0.0
        if np.isinf(min_val) or np.isinf(max_val):
            raise ValueError(
                'cannot specify integer `bins` when input data contains infinity'
                )
        elif min_val == max_val:
            min_val -= 0.001 * abs(min_val) if min_val != 0 else 0.001
            max_val += 0.001 * abs(max_val) if max_val != 0 else 0.001
            bins = np.linspace(min_val, max_val, nbins + 1, endpoint=True)
        else:
            bins = np.linspace(min_val, max_val, nbins + 1, endpoint=True)
            etqxi__mrovf = (max_val - min_val) * 0.001
            if right:
                bins[0] -= etqxi__mrovf
            else:
                bins[-1] += etqxi__mrovf
        return bins
    return impl


@overload_method(SeriesType, 'value_counts', inline='always', no_unliteral=True
    )
def overload_series_value_counts(S, normalize=False, sort=True, ascending=
    False, bins=None, dropna=True, _index_name=None):
    vxivd__glivu = dict(dropna=dropna)
    prr__rqtl = dict(dropna=True)
    check_unsupported_args('Series.value_counts', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_constant_bool(normalize):
        raise_bodo_error(
            'Series.value_counts(): normalize argument must be a constant boolean'
            )
    if not is_overload_constant_bool(sort):
        raise_bodo_error(
            'Series.value_counts(): sort argument must be a constant boolean')
    if not is_overload_bool(ascending):
        raise_bodo_error(
            'Series.value_counts(): ascending argument must be a constant boolean'
            )
    znyb__uesms = not is_overload_none(bins)
    utmrs__nduqj = 'def impl(\n'
    utmrs__nduqj += '    S,\n'
    utmrs__nduqj += '    normalize=False,\n'
    utmrs__nduqj += '    sort=True,\n'
    utmrs__nduqj += '    ascending=False,\n'
    utmrs__nduqj += '    bins=None,\n'
    utmrs__nduqj += '    dropna=True,\n'
    utmrs__nduqj += (
        '    _index_name=None,  # bodo argument. See groupby.value_counts\n')
    utmrs__nduqj += '):\n'
    utmrs__nduqj += (
        '    arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    utmrs__nduqj += (
        '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    utmrs__nduqj += (
        '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
    if znyb__uesms:
        utmrs__nduqj += '    right = True\n'
        utmrs__nduqj += _gen_bins_handling(bins, S.dtype)
        utmrs__nduqj += '    arr = get_bin_inds(bins, arr)\n'
    utmrs__nduqj += (
        '    in_df = bodo.hiframes.pd_dataframe_ext.init_dataframe(\n')
    utmrs__nduqj += "        (arr,), index, ('$_bodo_col2_',)\n"
    utmrs__nduqj += '    )\n'
    utmrs__nduqj += "    count_series = in_df.groupby('$_bodo_col2_').size()\n"
    if znyb__uesms:
        utmrs__nduqj += """    count_series = bodo.gatherv(count_series, allgather=True, warn_if_rep=False)
"""
        utmrs__nduqj += (
            '    count_arr = get_output_bin_counts(count_series, len(bins) - 1)\n'
            )
        utmrs__nduqj += '    index = get_bin_labels(bins)\n'
    else:
        utmrs__nduqj += """    count_arr = bodo.hiframes.pd_series_ext.get_series_data(count_series)
"""
        utmrs__nduqj += (
            '    ind_arr = bodo.utils.conversion.coerce_to_array(\n')
        utmrs__nduqj += (
            '        bodo.hiframes.pd_series_ext.get_series_index(count_series)\n'
            )
        utmrs__nduqj += '    )\n'
        utmrs__nduqj += """    index = bodo.utils.conversion.index_from_array(ind_arr, name=_index_name)
"""
    utmrs__nduqj += (
        '    res = bodo.hiframes.pd_series_ext.init_series(count_arr, index, name)\n'
        )
    if is_overload_true(sort):
        utmrs__nduqj += '    res = res.sort_values(ascending=ascending)\n'
    if is_overload_true(normalize):
        uznp__jliz = 'len(S)' if znyb__uesms else 'count_arr.sum()'
        utmrs__nduqj += f'    res = res / float({uznp__jliz})\n'
    utmrs__nduqj += '    return res\n'
    snqc__makr = {}
    exec(utmrs__nduqj, {'bodo': bodo, 'pd': pd, 'np': np, 'get_bin_inds':
        get_bin_inds, 'get_bin_labels': get_bin_labels,
        'get_output_bin_counts': get_output_bin_counts, 'compute_bins':
        compute_bins}, snqc__makr)
    impl = snqc__makr['impl']
    return impl


def _gen_bins_handling(bins, dtype):
    utmrs__nduqj = ''
    if isinstance(bins, types.Integer):
        utmrs__nduqj += '    min_val = bodo.libs.array_ops.array_op_min(arr)\n'
        utmrs__nduqj += '    max_val = bodo.libs.array_ops.array_op_max(arr)\n'
        if dtype == bodo.datetime64ns:
            utmrs__nduqj += '    min_val = min_val.value\n'
            utmrs__nduqj += '    max_val = max_val.value\n'
        utmrs__nduqj += (
            '    bins = compute_bins(bins, min_val, max_val, right)\n')
        if dtype == bodo.datetime64ns:
            utmrs__nduqj += (
                "    bins = bins.astype(np.int64).view(np.dtype('datetime64[ns]'))\n"
                )
    else:
        utmrs__nduqj += (
            '    bins = bodo.utils.conversion.coerce_to_ndarray(bins)\n')
    return utmrs__nduqj


@overload(pd.cut, inline='always', no_unliteral=True)
def overload_cut(x, bins, right=True, labels=None, retbins=False, precision
    =3, include_lowest=False, duplicates='raise', ordered=True):
    vxivd__glivu = dict(right=right, labels=labels, retbins=retbins,
        precision=precision, duplicates=duplicates, ordered=ordered)
    prr__rqtl = dict(right=True, labels=None, retbins=False, precision=3,
        duplicates='raise', ordered=True)
    check_unsupported_args('pandas.cut', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='General')
    utmrs__nduqj = 'def impl(\n'
    utmrs__nduqj += '    x,\n'
    utmrs__nduqj += '    bins,\n'
    utmrs__nduqj += '    right=True,\n'
    utmrs__nduqj += '    labels=None,\n'
    utmrs__nduqj += '    retbins=False,\n'
    utmrs__nduqj += '    precision=3,\n'
    utmrs__nduqj += '    include_lowest=False,\n'
    utmrs__nduqj += "    duplicates='raise',\n"
    utmrs__nduqj += '    ordered=True\n'
    utmrs__nduqj += '):\n'
    if isinstance(x, SeriesType):
        utmrs__nduqj += (
            '    arr = bodo.hiframes.pd_series_ext.get_series_data(x)\n')
        utmrs__nduqj += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(x)\n')
        utmrs__nduqj += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(x)\n')
    else:
        utmrs__nduqj += '    arr = bodo.utils.conversion.coerce_to_array(x)\n'
    utmrs__nduqj += _gen_bins_handling(bins, x.dtype)
    utmrs__nduqj += (
        '    arr = get_bin_inds(bins, arr, False, include_lowest)\n')
    utmrs__nduqj += (
        '    label_index = get_bin_labels(bins, right, include_lowest)\n')
    utmrs__nduqj += """    cat_dtype = bodo.hiframes.pd_categorical_ext.init_cat_dtype(label_index, ordered, None, None)
"""
    utmrs__nduqj += """    out_arr = bodo.hiframes.pd_categorical_ext.init_categorical_array(arr, cat_dtype)
"""
    if isinstance(x, SeriesType):
        utmrs__nduqj += (
            '    res = bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
            )
        utmrs__nduqj += '    return res\n'
    else:
        utmrs__nduqj += '    return out_arr\n'
    snqc__makr = {}
    exec(utmrs__nduqj, {'bodo': bodo, 'pd': pd, 'np': np, 'get_bin_inds':
        get_bin_inds, 'get_bin_labels': get_bin_labels,
        'get_output_bin_counts': get_output_bin_counts, 'compute_bins':
        compute_bins}, snqc__makr)
    impl = snqc__makr['impl']
    return impl


def _get_q_list(q):
    return q


@overload(_get_q_list, no_unliteral=True)
def get_q_list_overload(q):
    if is_overload_int(q):
        return lambda q: np.linspace(0, 1, q + 1)
    return lambda q: q


@overload(pd.qcut, inline='always', no_unliteral=True)
def overload_qcut(x, q, labels=None, retbins=False, precision=3, duplicates
    ='raise'):
    vxivd__glivu = dict(labels=labels, retbins=retbins, precision=precision,
        duplicates=duplicates)
    prr__rqtl = dict(labels=None, retbins=False, precision=3, duplicates=
        'raise')
    check_unsupported_args('pandas.qcut', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='General')
    if not (is_overload_int(q) or is_iterable_type(q)):
        raise BodoError(
            "pd.qcut(): 'q' should be an integer or a list of quantiles")

    def impl(x, q, labels=None, retbins=False, precision=3, duplicates='raise'
        ):
        mkrqt__yoo = _get_q_list(q)
        arr = bodo.utils.conversion.coerce_to_array(x)
        bins = bodo.libs.array_ops.array_op_quantile(arr, mkrqt__yoo)
        return pd.cut(x, bins, include_lowest=True)
    return impl


@overload_method(SeriesType, 'groupby', inline='always', no_unliteral=True)
def overload_series_groupby(S, by=None, axis=0, level=None, as_index=True,
    sort=True, group_keys=True, squeeze=False, observed=True, dropna=True):
    vxivd__glivu = dict(axis=axis, sort=sort, group_keys=group_keys,
        squeeze=squeeze, observed=observed, dropna=dropna)
    prr__rqtl = dict(axis=0, sort=True, group_keys=True, squeeze=False,
        observed=True, dropna=True)
    check_unsupported_args('Series.groupby', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='GroupBy')
    if not is_overload_true(as_index):
        raise BodoError('as_index=False only valid with DataFrame')
    if is_overload_none(by) and is_overload_none(level):
        raise BodoError("You have to supply one of 'by' and 'level'")
    if not is_overload_none(by) and not is_overload_none(level):
        raise BodoError(
            "Series.groupby(): 'level' argument should be None if 'by' is not None"
            )
    if not is_overload_none(level):
        if not (is_overload_constant_int(level) and get_overload_const_int(
            level) == 0) or isinstance(S.index, bodo.hiframes.
            pd_multi_index_ext.MultiIndexType):
            raise BodoError(
                "Series.groupby(): MultiIndex case or 'level' other than 0 not supported yet"
                )

        def impl_index(S, by=None, axis=0, level=None, as_index=True, sort=
            True, group_keys=True, squeeze=False, observed=True, dropna=True):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            kqwm__tbr = bodo.utils.conversion.coerce_to_array(index)
            pyxt__vqtlq = bodo.hiframes.pd_dataframe_ext.init_dataframe((
                kqwm__tbr, arr), index, (' ', ''))
            return pyxt__vqtlq.groupby(' ')['']
        return impl_index
    kuzvk__pbdm = by
    if isinstance(by, SeriesType):
        kuzvk__pbdm = by.data
    if isinstance(kuzvk__pbdm, DecimalArrayType):
        raise BodoError(
            'Series.groupby(): by argument with decimal type is not supported yet.'
            )
    if isinstance(by, bodo.hiframes.pd_categorical_ext.CategoricalArrayType):
        raise BodoError(
            'Series.groupby(): by argument with categorical type is not supported yet.'
            )

    def impl(S, by=None, axis=0, level=None, as_index=True, sort=True,
        group_keys=True, squeeze=False, observed=True, dropna=True):
        kqwm__tbr = bodo.utils.conversion.coerce_to_array(by)
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        pyxt__vqtlq = bodo.hiframes.pd_dataframe_ext.init_dataframe((
            kqwm__tbr, arr), index, (' ', ''))
        return pyxt__vqtlq.groupby(' ')['']
    return impl


@overload_method(SeriesType, 'append', inline='always', no_unliteral=True)
def overload_series_append(S, to_append, ignore_index=False,
    verify_integrity=False):
    vxivd__glivu = dict(verify_integrity=verify_integrity)
    prr__rqtl = dict(verify_integrity=False)
    check_unsupported_args('Series.append', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if isinstance(to_append, SeriesType):
        return (lambda S, to_append, ignore_index=False, verify_integrity=
            False: pd.concat((S, to_append), ignore_index=ignore_index,
            verify_integrity=verify_integrity))
    if isinstance(to_append, types.BaseTuple):
        return (lambda S, to_append, ignore_index=False, verify_integrity=
            False: pd.concat((S,) + to_append, ignore_index=ignore_index,
            verify_integrity=verify_integrity))
    return (lambda S, to_append, ignore_index=False, verify_integrity=False:
        pd.concat([S] + to_append, ignore_index=ignore_index,
        verify_integrity=verify_integrity))


@overload_method(SeriesType, 'isin', inline='always', no_unliteral=True)
def overload_series_isin(S, values):
    if bodo.utils.utils.is_array_typ(values):

        def impl_arr(S, values):
            uyeea__fwaq = bodo.utils.conversion.coerce_to_array(values)
            A = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            n = len(A)
            ivga__befgo = np.empty(n, np.bool_)
            bodo.libs.array.array_isin(ivga__befgo, A, uyeea__fwaq, False)
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl_arr
    if not isinstance(values, (types.Set, types.List)):
        raise BodoError(
            "Series.isin(): 'values' parameter should be a set or a list")

    def impl(S, values):
        A = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo = bodo.libs.array_ops.array_op_isin(A, values)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'quantile', inline='always', no_unliteral=True)
def overload_series_quantile(S, q=0.5, interpolation='linear'):
    vxivd__glivu = dict(interpolation=interpolation)
    prr__rqtl = dict(interpolation='linear')
    check_unsupported_args('Series.quantile', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if is_iterable_type(q) and isinstance(q.dtype, types.Number):

        def impl_list(S, q=0.5, interpolation='linear'):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            ivga__befgo = bodo.libs.array_ops.array_op_quantile(arr, q)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            index = bodo.hiframes.pd_index_ext.init_numeric_index(bodo.
                utils.conversion.coerce_to_array(q), None)
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl_list
    elif isinstance(q, (float, types.Number)) or is_overload_constant_int(q):

        def impl(S, q=0.5, interpolation='linear'):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            return bodo.libs.array_ops.array_op_quantile(arr, q)
        return impl
    else:
        raise BodoError(
            f'Series.quantile() q type must be float or iterable of floats only.'
            )


@overload_method(SeriesType, 'nunique', inline='always', no_unliteral=True)
def overload_series_nunique(S, dropna=True):
    if not is_overload_bool(dropna):
        raise BodoError('Series.nunique: dropna must be a boolean value')

    def impl(S, dropna=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        return bodo.libs.array_kernels.nunique(arr, dropna)
    return impl


@overload_method(SeriesType, 'unique', inline='always', no_unliteral=True)
def overload_series_unique(S):

    def impl(S):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        qrli__yawkl = bodo.libs.array_kernels.unique(arr)
        return bodo.allgatherv(qrli__yawkl, False)
    return impl


@overload_method(SeriesType, 'describe', inline='always', no_unliteral=True)
def overload_series_describe(S, percentiles=None, include=None, exclude=
    None, datetime_is_numeric=True):
    vxivd__glivu = dict(percentiles=percentiles, include=include, exclude=
        exclude, datetime_is_numeric=datetime_is_numeric)
    prr__rqtl = dict(percentiles=None, include=None, exclude=None,
        datetime_is_numeric=True)
    check_unsupported_args('Series.describe', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (isinstance(S.data, types.Array) and (isinstance(S.data.dtype,
        types.Number) or S.data.dtype == bodo.datetime64ns)
        ) and not isinstance(S.data, IntegerArrayType):
        raise BodoError(f'describe() column input type {S.data} not supported.'
            )
    if S.data.dtype == bodo.datetime64ns:

        def impl_dt(S, percentiles=None, include=None, exclude=None,
            datetime_is_numeric=True):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            return bodo.hiframes.pd_series_ext.init_series(bodo.libs.
                array_ops.array_op_describe(arr), bodo.utils.conversion.
                convert_to_index(['count', 'mean', 'min', '25%', '50%',
                '75%', 'max']), name)
        return impl_dt

    def impl(S, percentiles=None, include=None, exclude=None,
        datetime_is_numeric=True):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        return bodo.hiframes.pd_series_ext.init_series(bodo.libs.array_ops.
            array_op_describe(arr), bodo.utils.conversion.convert_to_index(
            ['count', 'mean', 'std', 'min', '25%', '50%', '75%', 'max']), name)
    return impl


@overload_method(SeriesType, 'memory_usage', inline='always', no_unliteral=True
    )
def overload_series_memory_usage(S, index=True, deep=False):
    if is_overload_true(index):

        def impl(S, index=True, deep=False):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            return arr.nbytes + index.nbytes
        return impl
    else:

        def impl(S, index=True, deep=False):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            return arr.nbytes
        return impl


def binary_str_fillna_inplace_series_impl(is_binary=False):
    if is_binary:
        amndl__sihes = 'bodo.libs.binary_arr_ext.pre_alloc_binary_array'
    else:
        amndl__sihes = 'bodo.libs.str_arr_ext.pre_alloc_string_array'
    utmrs__nduqj = '\n'.join(('def impl(', '    S,', '    value=None,',
        '    method=None,', '    axis=None,', '    inplace=False,',
        '    limit=None,', '    downcast=None,', '):',
        '    in_arr = bodo.hiframes.pd_series_ext.get_series_data(S)',
        '    fill_arr = bodo.hiframes.pd_series_ext.get_series_data(value)',
        '    n = len(in_arr)', '    nf = len(fill_arr)',
        "    assert n == nf, 'fillna() requires same length arrays'",
        f'    out_arr = {amndl__sihes}(n, -1)',
        '    for j in numba.parfors.parfor.internal_prange(n):',
        '        s = in_arr[j]',
        '        if bodo.libs.array_kernels.isna(in_arr, j) and not bodo.libs.array_kernels.isna('
        , '            fill_arr, j', '        ):',
        '            s = fill_arr[j]', '        out_arr[j] = s',
        '    bodo.libs.str_arr_ext.move_str_binary_arr_payload(in_arr, out_arr)'
        ))
    axv__stlbs = dict()
    exec(utmrs__nduqj, {'bodo': bodo, 'numba': numba}, axv__stlbs)
    ybg__gfruf = axv__stlbs['impl']
    return ybg__gfruf


def binary_str_fillna_inplace_impl(is_binary=False):
    if is_binary:
        amndl__sihes = 'bodo.libs.binary_arr_ext.pre_alloc_binary_array'
    else:
        amndl__sihes = 'bodo.libs.str_arr_ext.pre_alloc_string_array'
    utmrs__nduqj = 'def impl(S,\n'
    utmrs__nduqj += '     value=None,\n'
    utmrs__nduqj += '    method=None,\n'
    utmrs__nduqj += '    axis=None,\n'
    utmrs__nduqj += '    inplace=False,\n'
    utmrs__nduqj += '    limit=None,\n'
    utmrs__nduqj += '   downcast=None,\n'
    utmrs__nduqj += '):\n'
    utmrs__nduqj += (
        '    in_arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    utmrs__nduqj += '    n = len(in_arr)\n'
    utmrs__nduqj += f'    out_arr = {amndl__sihes}(n, -1)\n'
    utmrs__nduqj += '    for j in numba.parfors.parfor.internal_prange(n):\n'
    utmrs__nduqj += '        s = in_arr[j]\n'
    utmrs__nduqj += '        if bodo.libs.array_kernels.isna(in_arr, j):\n'
    utmrs__nduqj += '            s = value\n'
    utmrs__nduqj += '        out_arr[j] = s\n'
    utmrs__nduqj += (
        '    bodo.libs.str_arr_ext.move_str_binary_arr_payload(in_arr, out_arr)\n'
        )
    axv__stlbs = dict()
    exec(utmrs__nduqj, {'bodo': bodo, 'numba': numba}, axv__stlbs)
    ybg__gfruf = axv__stlbs['impl']
    return ybg__gfruf


def fillna_inplace_series_impl(S, value=None, method=None, axis=None,
    inplace=False, limit=None, downcast=None):
    glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
    xvx__fypxl = bodo.hiframes.pd_series_ext.get_series_data(value)
    for dot__kzj in numba.parfors.parfor.internal_prange(len(glfic__kmbxv)):
        s = glfic__kmbxv[dot__kzj]
        if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj
            ) and not bodo.libs.array_kernels.isna(xvx__fypxl, dot__kzj):
            s = xvx__fypxl[dot__kzj]
        glfic__kmbxv[dot__kzj] = s


def fillna_inplace_impl(S, value=None, method=None, axis=None, inplace=
    False, limit=None, downcast=None):
    glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
    for dot__kzj in numba.parfors.parfor.internal_prange(len(glfic__kmbxv)):
        s = glfic__kmbxv[dot__kzj]
        if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj):
            s = value
        glfic__kmbxv[dot__kzj] = s


def str_fillna_alloc_series_impl(S, value=None, method=None, axis=None,
    inplace=False, limit=None, downcast=None):
    glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
    index = bodo.hiframes.pd_series_ext.get_series_index(S)
    name = bodo.hiframes.pd_series_ext.get_series_name(S)
    xvx__fypxl = bodo.hiframes.pd_series_ext.get_series_data(value)
    n = len(glfic__kmbxv)
    ivga__befgo = bodo.libs.str_arr_ext.pre_alloc_string_array(n, -1)
    for enk__kweh in numba.parfors.parfor.internal_prange(n):
        s = glfic__kmbxv[enk__kweh]
        if bodo.libs.array_kernels.isna(glfic__kmbxv, enk__kweh
            ) and not bodo.libs.array_kernels.isna(xvx__fypxl, enk__kweh):
            s = xvx__fypxl[enk__kweh]
        ivga__befgo[enk__kweh] = s
        if bodo.libs.array_kernels.isna(glfic__kmbxv, enk__kweh
            ) and bodo.libs.array_kernels.isna(xvx__fypxl, enk__kweh):
            bodo.libs.array_kernels.setna(ivga__befgo, enk__kweh)
    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name)


def fillna_series_impl(S, value=None, method=None, axis=None, inplace=False,
    limit=None, downcast=None):
    glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
    index = bodo.hiframes.pd_series_ext.get_series_index(S)
    name = bodo.hiframes.pd_series_ext.get_series_name(S)
    xvx__fypxl = bodo.hiframes.pd_series_ext.get_series_data(value)
    n = len(glfic__kmbxv)
    ivga__befgo = bodo.utils.utils.alloc_type(n, glfic__kmbxv.dtype, (-1,))
    for dot__kzj in numba.parfors.parfor.internal_prange(n):
        s = glfic__kmbxv[dot__kzj]
        if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj
            ) and not bodo.libs.array_kernels.isna(xvx__fypxl, dot__kzj):
            s = xvx__fypxl[dot__kzj]
        ivga__befgo[dot__kzj] = s
    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name)


@overload_method(SeriesType, 'fillna', no_unliteral=True)
def overload_series_fillna(S, value=None, method=None, axis=None, inplace=
    False, limit=None, downcast=None):
    vxivd__glivu = dict(limit=limit, downcast=downcast)
    prr__rqtl = dict(limit=None, downcast=None)
    check_unsupported_args('Series.fillna', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    dssz__nefug = not is_overload_none(value)
    wqz__ckmtm = not is_overload_none(method)
    if dssz__nefug and wqz__ckmtm:
        raise BodoError(
            "Series.fillna(): Cannot specify both 'value' and 'method'.")
    if not dssz__nefug and not wqz__ckmtm:
        raise BodoError(
            "Series.fillna(): Must specify one of 'value' and 'method'.")
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error('Series.fillna(): axis argument not supported')
    elif is_iterable_type(value) and not isinstance(value, SeriesType):
        raise BodoError('Series.fillna(): "value" parameter cannot be a list')
    elif is_var_size_item_array_type(S.data
        ) and not S.dtype == bodo.string_type:
        raise BodoError(
            f'Series.fillna() with inplace=True not supported for {S.dtype} values yet.'
            )
    if not is_overload_constant_bool(inplace):
        raise_bodo_error(
            "Series.fillna(): 'inplace' argument must be a constant boolean")
    if wqz__ckmtm:
        if is_overload_true(inplace):
            raise BodoError(
                "Series.fillna() with inplace=True not supported with 'method' argument yet."
                )
        qxpn__ngil = (
            "Series.fillna(): 'method' argument if provided must be a constant string and one of ('backfill', 'bfill', 'pad' 'ffill')."
            )
        if not is_overload_constant_str(method):
            raise_bodo_error(qxpn__ngil)
        elif get_overload_const_str(method) not in ('backfill', 'bfill',
            'pad', 'ffill'):
            raise BodoError(qxpn__ngil)
    uxa__feij = element_type(S.data)
    tytbb__sofd = None
    if dssz__nefug:
        tytbb__sofd = element_type(types.unliteral(value))
    if tytbb__sofd and not can_replace(uxa__feij, tytbb__sofd):
        raise BodoError(
            f'Series.fillna(): Cannot use value type {tytbb__sofd} with series type {uxa__feij}'
            )
    if is_overload_true(inplace):
        if S.dtype == bodo.string_type:
            if is_overload_constant_str(value) and get_overload_const_str(value
                ) == '':
                return (lambda S, value=None, method=None, axis=None,
                    inplace=False, limit=None, downcast=None: bodo.libs.
                    str_arr_ext.set_null_bits_to_value(bodo.hiframes.
                    pd_series_ext.get_series_data(S), -1))
            if isinstance(value, SeriesType):
                return binary_str_fillna_inplace_series_impl(is_binary=False)
            return binary_str_fillna_inplace_impl(is_binary=False)
        if S.dtype == bodo.bytes_type:
            if is_overload_constant_bytes(value) and get_overload_const_bytes(
                value) == b'':
                return (lambda S, value=None, method=None, axis=None,
                    inplace=False, limit=None, downcast=None: bodo.libs.
                    str_arr_ext.set_null_bits_to_value(bodo.hiframes.
                    pd_series_ext.get_series_data(S), -1))
            if isinstance(value, SeriesType):
                return binary_str_fillna_inplace_series_impl(is_binary=True)
            return binary_str_fillna_inplace_impl(is_binary=True)
        else:
            if isinstance(value, SeriesType):
                return fillna_inplace_series_impl
            return fillna_inplace_impl
    else:
        guc__wgs = S.data
        if isinstance(value, SeriesType):

            def fillna_series_impl(S, value=None, method=None, axis=None,
                inplace=False, limit=None, downcast=None):
                glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
                index = bodo.hiframes.pd_series_ext.get_series_index(S)
                name = bodo.hiframes.pd_series_ext.get_series_name(S)
                xvx__fypxl = bodo.hiframes.pd_series_ext.get_series_data(value)
                n = len(glfic__kmbxv)
                ivga__befgo = bodo.utils.utils.alloc_type(n, guc__wgs, (-1,))
                for dot__kzj in numba.parfors.parfor.internal_prange(n):
                    if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj
                        ) and bodo.libs.array_kernels.isna(xvx__fypxl, dot__kzj
                        ):
                        bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                        continue
                    if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj):
                        ivga__befgo[dot__kzj
                            ] = bodo.utils.conversion.unbox_if_timestamp(
                            xvx__fypxl[dot__kzj])
                        continue
                    ivga__befgo[dot__kzj
                        ] = bodo.utils.conversion.unbox_if_timestamp(
                        glfic__kmbxv[dot__kzj])
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return fillna_series_impl
        if wqz__ckmtm:
            yvx__dzf = (types.unicode_type, types.bool_, bodo.datetime64ns,
                bodo.timedelta64ns)
            if not isinstance(uxa__feij, (types.Integer, types.Float)
                ) and uxa__feij not in yvx__dzf:
                raise BodoError(
                    f"Series.fillna(): series of type {uxa__feij} are not supported with 'method' argument."
                    )

            def fillna_method_impl(S, value=None, method=None, axis=None,
                inplace=False, limit=None, downcast=None):
                glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
                index = bodo.hiframes.pd_series_ext.get_series_index(S)
                name = bodo.hiframes.pd_series_ext.get_series_name(S)
                ivga__befgo = bodo.libs.array_kernels.ffill_bfill_arr(
                    glfic__kmbxv, method)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return fillna_method_impl

        def fillna_impl(S, value=None, method=None, axis=None, inplace=
            False, limit=None, downcast=None):
            value = bodo.utils.conversion.unbox_if_timestamp(value)
            glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            n = len(glfic__kmbxv)
            ivga__befgo = bodo.utils.utils.alloc_type(n, guc__wgs, (-1,))
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                s = bodo.utils.conversion.unbox_if_timestamp(glfic__kmbxv[
                    dot__kzj])
                if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj):
                    s = value
                ivga__befgo[dot__kzj] = s
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return fillna_impl


def create_fillna_specific_method_overload(overload_name):

    def overload_series_fillna_specific_method(S, axis=None, inplace=False,
        limit=None, downcast=None):
        nxb__wfd = {'ffill': 'ffill', 'bfill': 'bfill', 'pad': 'ffill',
            'backfill': 'bfill'}[overload_name]
        vxivd__glivu = dict(limit=limit, downcast=downcast)
        prr__rqtl = dict(limit=None, downcast=None)
        check_unsupported_args(f'Series.{overload_name}', vxivd__glivu,
            prr__rqtl, package_name='pandas', module_name='Series')
        if not (is_overload_none(axis) or is_overload_zero(axis)):
            raise BodoError(
                f'Series.{overload_name}(): axis argument not supported')
        uxa__feij = element_type(S.data)
        yvx__dzf = (types.unicode_type, types.bool_, bodo.datetime64ns,
            bodo.timedelta64ns)
        if not isinstance(uxa__feij, (types.Integer, types.Float)
            ) and uxa__feij not in yvx__dzf:
            raise BodoError(
                f'Series.{overload_name}(): series of type {uxa__feij} are not supported.'
                )

        def impl(S, axis=None, inplace=False, limit=None, downcast=None):
            glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            ivga__befgo = bodo.libs.array_kernels.ffill_bfill_arr(glfic__kmbxv,
                nxb__wfd)
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl
    return overload_series_fillna_specific_method


fillna_specific_methods = 'ffill', 'bfill', 'pad', 'backfill'


def _install_fillna_specific_methods():
    for overload_name in fillna_specific_methods:
        rgkm__wnfc = create_fillna_specific_method_overload(overload_name)
        overload_method(SeriesType, overload_name, no_unliteral=True)(
            rgkm__wnfc)


_install_fillna_specific_methods()


def check_unsupported_types(S, to_replace, value):
    if any(bodo.utils.utils.is_array_typ(x, True) for x in [S.dtype,
        to_replace, value]):
        ynwi__fmq = (
            'Series.replace(): only support with Scalar, List, or Dictionary')
        raise BodoError(ynwi__fmq)
    elif isinstance(to_replace, types.DictType) and not is_overload_none(value
        ):
        ynwi__fmq = (
            "Series.replace(): 'value' must be None when 'to_replace' is a dictionary"
            )
        raise BodoError(ynwi__fmq)
    elif any(isinstance(x, (PandasTimestampType, PDTimeDeltaType)) for x in
        [to_replace, value]):
        ynwi__fmq = (
            f'Series.replace(): Not supported for types {to_replace} and {value}'
            )
        raise BodoError(ynwi__fmq)


def series_replace_error_checking(S, to_replace, value, inplace, limit,
    regex, method):
    vxivd__glivu = dict(inplace=inplace, limit=limit, regex=regex, method=
        method)
    fvjk__bpo = dict(inplace=False, limit=None, regex=False, method='pad')
    check_unsupported_args('Series.replace', vxivd__glivu, fvjk__bpo,
        package_name='pandas', module_name='Series')
    check_unsupported_types(S, to_replace, value)


@overload_method(SeriesType, 'replace', inline='always', no_unliteral=True)
def overload_series_replace(S, to_replace=None, value=None, inplace=False,
    limit=None, regex=False, method='pad'):
    series_replace_error_checking(S, to_replace, value, inplace, limit,
        regex, method)
    uxa__feij = element_type(S.data)
    if isinstance(to_replace, types.DictType):
        drz__wnus = element_type(to_replace.key_type)
        tytbb__sofd = element_type(to_replace.value_type)
    else:
        drz__wnus = element_type(to_replace)
        tytbb__sofd = element_type(value)
    emjy__diwq = None
    if uxa__feij != types.unliteral(drz__wnus):
        if bodo.utils.typing.equality_always_false(uxa__feij, types.
            unliteral(drz__wnus)
            ) or not bodo.utils.typing.types_equality_exists(uxa__feij,
            drz__wnus):

            def impl(S, to_replace=None, value=None, inplace=False, limit=
                None, regex=False, method='pad'):
                return S.copy()
            return impl
        if isinstance(uxa__feij, (types.Float, types.Integer)
            ) or uxa__feij == np.bool_:
            emjy__diwq = uxa__feij
    if not can_replace(uxa__feij, types.unliteral(tytbb__sofd)):

        def impl(S, to_replace=None, value=None, inplace=False, limit=None,
            regex=False, method='pad'):
            return S.copy()
        return impl
    txlh__vallb = S.data
    if isinstance(txlh__vallb, CategoricalArrayType):

        def cat_impl(S, to_replace=None, value=None, inplace=False, limit=
            None, regex=False, method='pad'):
            glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            return bodo.hiframes.pd_series_ext.init_series(glfic__kmbxv.
                replace(to_replace, value), index, name)
        return cat_impl

    def impl(S, to_replace=None, value=None, inplace=False, limit=None,
        regex=False, method='pad'):
        glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        n = len(glfic__kmbxv)
        ivga__befgo = bodo.utils.utils.alloc_type(n, txlh__vallb, (-1,))
        tocyg__yjg = build_replace_dict(to_replace, value, emjy__diwq)
        for dot__kzj in numba.parfors.parfor.internal_prange(n):
            if bodo.libs.array_kernels.isna(glfic__kmbxv, dot__kzj):
                bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                continue
            s = glfic__kmbxv[dot__kzj]
            if s in tocyg__yjg:
                s = tocyg__yjg[s]
            ivga__befgo[dot__kzj] = s
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


def build_replace_dict(to_replace, value, key_dtype_conv):
    pass


@overload(build_replace_dict)
def _build_replace_dict(to_replace, value, key_dtype_conv):
    pgg__hvgn = isinstance(to_replace, (types.Number, Decimal128Type)
        ) or to_replace in [bodo.string_type, types.boolean, bodo.bytes_type]
    mlxbi__dvwb = is_iterable_type(to_replace)
    ffm__grx = isinstance(value, (types.Number, Decimal128Type)) or value in [
        bodo.string_type, bodo.bytes_type, types.boolean]
    fsd__qzqzx = is_iterable_type(value)
    if pgg__hvgn and ffm__grx:
        if not is_overload_none(key_dtype_conv):

            def impl_cast(to_replace, value, key_dtype_conv):
                tocyg__yjg = {}
                tocyg__yjg[key_dtype_conv(to_replace)] = value
                return tocyg__yjg
            return impl_cast

        def impl(to_replace, value, key_dtype_conv):
            tocyg__yjg = {}
            tocyg__yjg[to_replace] = value
            return tocyg__yjg
        return impl
    if mlxbi__dvwb and ffm__grx:
        if not is_overload_none(key_dtype_conv):

            def impl_cast(to_replace, value, key_dtype_conv):
                tocyg__yjg = {}
                for ovyb__lfp in to_replace:
                    tocyg__yjg[key_dtype_conv(ovyb__lfp)] = value
                return tocyg__yjg
            return impl_cast

        def impl(to_replace, value, key_dtype_conv):
            tocyg__yjg = {}
            for ovyb__lfp in to_replace:
                tocyg__yjg[ovyb__lfp] = value
            return tocyg__yjg
        return impl
    if mlxbi__dvwb and fsd__qzqzx:
        if not is_overload_none(key_dtype_conv):

            def impl_cast(to_replace, value, key_dtype_conv):
                tocyg__yjg = {}
                assert len(to_replace) == len(value
                    ), 'To_replace and value lengths must be the same'
                for dot__kzj in range(len(to_replace)):
                    tocyg__yjg[key_dtype_conv(to_replace[dot__kzj])] = value[
                        dot__kzj]
                return tocyg__yjg
            return impl_cast

        def impl(to_replace, value, key_dtype_conv):
            tocyg__yjg = {}
            assert len(to_replace) == len(value
                ), 'To_replace and value lengths must be the same'
            for dot__kzj in range(len(to_replace)):
                tocyg__yjg[to_replace[dot__kzj]] = value[dot__kzj]
            return tocyg__yjg
        return impl
    if isinstance(to_replace, numba.types.DictType) and is_overload_none(value
        ):
        return lambda to_replace, value, key_dtype_conv: to_replace
    raise BodoError(
        'Series.replace(): Not supported for types to_replace={} and value={}'
        .format(to_replace, value))


@overload_method(SeriesType, 'diff', inline='always', no_unliteral=True)
def overload_series_diff(S, periods=1):
    if not (isinstance(S.data, types.Array) and (isinstance(S.data.dtype,
        types.Number) or S.data.dtype == bodo.datetime64ns)):
        raise BodoError(
            f'Series.diff() column input type {S.data} not supported.')
    if not is_overload_int(periods):
        raise BodoError("Series.diff(): 'periods' input must be an integer.")
    if S.data == types.Array(bodo.datetime64ns, 1, 'C'):

        def impl_datetime(S, periods=1):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            ivga__befgo = bodo.hiframes.series_impl.dt64_arr_sub(arr, bodo.
                hiframes.rolling.shift(arr, periods, False))
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl_datetime

    def impl(S, periods=1):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo = arr - bodo.hiframes.rolling.shift(arr, periods, False)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'explode', inline='always', no_unliteral=True)
def overload_series_explode(S, ignore_index=False):
    from bodo.hiframes.split_impl import string_array_split_view_type
    vxivd__glivu = dict(ignore_index=ignore_index)
    vnze__yjh = dict(ignore_index=False)
    check_unsupported_args('Series.explode', vxivd__glivu, vnze__yjh,
        package_name='pandas', module_name='Series')
    if not (isinstance(S.data, ArrayItemArrayType) or S.data ==
        string_array_split_view_type):
        return lambda S, ignore_index=False: S.copy()

    def impl(S, ignore_index=False):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        hls__iymkx = bodo.utils.conversion.index_to_array(index)
        ivga__befgo, vurp__mxlc = bodo.libs.array_kernels.explode(arr,
            hls__iymkx)
        juqvh__wrsfg = bodo.utils.conversion.index_from_array(vurp__mxlc)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl


@overload(np.digitize, inline='always', no_unliteral=True)
def overload_series_np_digitize(x, bins, right=False):
    if isinstance(x, SeriesType):

        def impl(x, bins, right=False):
            arr = bodo.hiframes.pd_series_ext.get_series_data(x)
            return np.digitize(arr, bins, right)
        return impl


@overload(np.argmax, inline='always', no_unliteral=True)
def argmax_overload(a, axis=None, out=None):
    if isinstance(a, types.Array) and is_overload_constant_int(axis
        ) and get_overload_const_int(axis) == 1:

        def impl(a, axis=None, out=None):
            oxb__azgkw = np.empty(len(a), a.dtype)
            numba.parfors.parfor.init_prange()
            n = len(a)
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                oxb__azgkw[dot__kzj] = np.argmax(a[dot__kzj])
            return oxb__azgkw
        return impl


@overload(np.argmin, inline='always', no_unliteral=True)
def argmin_overload(a, axis=None, out=None):
    if isinstance(a, types.Array) and is_overload_constant_int(axis
        ) and get_overload_const_int(axis) == 1:

        def impl(a, axis=None, out=None):
            bqycg__cpoi = np.empty(len(a), a.dtype)
            numba.parfors.parfor.init_prange()
            n = len(a)
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                bqycg__cpoi[dot__kzj] = np.argmin(a[dot__kzj])
            return bqycg__cpoi
        return impl


def overload_series_np_dot(a, b, out=None):
    if (isinstance(a, SeriesType) or isinstance(b, SeriesType)
        ) and not is_overload_none(out):
        raise BodoError("np.dot(): 'out' parameter not supported yet")
    if isinstance(a, SeriesType):

        def impl(a, b, out=None):
            arr = bodo.hiframes.pd_series_ext.get_series_data(a)
            return np.dot(arr, b)
        return impl
    if isinstance(b, SeriesType):

        def impl(a, b, out=None):
            arr = bodo.hiframes.pd_series_ext.get_series_data(b)
            return np.dot(a, arr)
        return impl


overload(np.dot, inline='always', no_unliteral=True)(overload_series_np_dot)
overload(operator.matmul, inline='always', no_unliteral=True)(
    overload_series_np_dot)


@overload_method(SeriesType, 'dropna', inline='always', no_unliteral=True)
def overload_series_dropna(S, axis=0, inplace=False, how=None):
    vxivd__glivu = dict(axis=axis, inplace=inplace, how=how)
    gvkgr__luxrl = dict(axis=0, inplace=False, how=None)
    check_unsupported_args('Series.dropna', vxivd__glivu, gvkgr__luxrl,
        package_name='pandas', module_name='Series')
    if S.dtype == bodo.string_type:

        def dropna_str_impl(S, axis=0, inplace=False, how=None):
            glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            jwhz__rxze = S.notna().values
            hls__iymkx = bodo.utils.conversion.extract_index_array(S)
            juqvh__wrsfg = bodo.utils.conversion.convert_to_index(hls__iymkx
                [jwhz__rxze])
            ivga__befgo = (bodo.hiframes.series_kernels.
                _series_dropna_str_alloc_impl_inner(glfic__kmbxv))
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                juqvh__wrsfg, name)
        return dropna_str_impl
    else:

        def dropna_impl(S, axis=0, inplace=False, how=None):
            glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            hls__iymkx = bodo.utils.conversion.extract_index_array(S)
            jwhz__rxze = S.notna().values
            juqvh__wrsfg = bodo.utils.conversion.convert_to_index(hls__iymkx
                [jwhz__rxze])
            ivga__befgo = glfic__kmbxv[jwhz__rxze]
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                juqvh__wrsfg, name)
        return dropna_impl


@overload_method(SeriesType, 'shift', inline='always', no_unliteral=True)
def overload_series_shift(S, periods=1, freq=None, axis=0, fill_value=None):
    vxivd__glivu = dict(freq=freq, axis=axis, fill_value=fill_value)
    prr__rqtl = dict(freq=None, axis=0, fill_value=None)
    check_unsupported_args('Series.shift', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_supported_shift_array_type(S.data):
        raise BodoError(
            f"Series.shift(): Series input type '{S.data.dtype}' not supported yet."
            )
    if not is_overload_int(periods):
        raise BodoError("Series.shift(): 'periods' input must be an integer.")

    def impl(S, periods=1, freq=None, axis=0, fill_value=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo = bodo.hiframes.rolling.shift(arr, periods, False)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'pct_change', inline='always', no_unliteral=True)
def overload_series_pct_change(S, periods=1, fill_method='pad', limit=None,
    freq=None):
    vxivd__glivu = dict(fill_method=fill_method, limit=limit, freq=freq)
    prr__rqtl = dict(fill_method='pad', limit=None, freq=None)
    check_unsupported_args('Series.pct_change', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not is_overload_int(periods):
        raise BodoError(
            'Series.pct_change(): periods argument must be an Integer')

    def impl(S, periods=1, fill_method='pad', limit=None, freq=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        ivga__befgo = bodo.hiframes.rolling.pct_change(arr, periods, False)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


def create_series_mask_where_overload(func_name):

    def overload_series_mask_where(S, cond, other=np.nan, inplace=False,
        axis=None, level=None, errors='raise', try_cast=False):
        _validate_arguments_mask_where(f'Series.{func_name}', S, cond,
            other, inplace, axis, level, errors, try_cast)
        if is_overload_constant_nan(other):
            buhy__dofte = 'None'
        else:
            buhy__dofte = 'other'
        utmrs__nduqj = """def impl(S, cond, other=np.nan, inplace=False, axis=None, level=None, errors='raise',try_cast=False):
"""
        if func_name == 'mask':
            utmrs__nduqj += '  cond = ~cond\n'
        utmrs__nduqj += (
            '  arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        utmrs__nduqj += (
            '  index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        utmrs__nduqj += (
            '  name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
        utmrs__nduqj += f"""  out_arr = bodo.hiframes.series_impl.where_impl(cond, arr, {buhy__dofte})
"""
        utmrs__nduqj += (
            '  return bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
            )
        snqc__makr = {}
        exec(utmrs__nduqj, {'bodo': bodo, 'np': np}, snqc__makr)
        impl = snqc__makr['impl']
        return impl
    return overload_series_mask_where


def _install_series_mask_where_overload():
    for func_name in ('mask', 'where'):
        rgkm__wnfc = create_series_mask_where_overload(func_name)
        overload_method(SeriesType, func_name, no_unliteral=True)(rgkm__wnfc)


_install_series_mask_where_overload()


def _validate_arguments_mask_where(func_name, S, cond, other, inplace, axis,
    level, errors, try_cast):
    vxivd__glivu = dict(inplace=inplace, level=level, errors=errors,
        try_cast=try_cast)
    prr__rqtl = dict(inplace=False, level=None, errors='raise', try_cast=False)
    check_unsupported_args(f'{func_name}', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (is_overload_none(axis) or is_overload_zero(axis)):
        raise_bodo_error(f'{func_name}(): axis argument not supported')
    if isinstance(other, SeriesType):
        _validate_self_other_mask_where(func_name, S.data, other.data)
    else:
        _validate_self_other_mask_where(func_name, S.data, other)
    if not (isinstance(cond, (SeriesType, types.Array, BooleanArrayType)) and
        cond.ndim == 1 and cond.dtype == types.bool_):
        raise BodoError(
            f"{func_name}() 'cond' argument must be a Series or 1-dim array of booleans"
            )


def _validate_self_other_mask_where(func_name, arr, other, max_ndim=1,
    is_default=False):
    if not (isinstance(arr, types.Array) or isinstance(arr,
        BooleanArrayType) or isinstance(arr, IntegerArrayType) or bodo.
        utils.utils.is_array_typ(arr, False) and arr.dtype in [bodo.
        string_type, bodo.bytes_type] or isinstance(arr, bodo.
        CategoricalArrayType) and arr.dtype.elem_type not in [bodo.
        datetime64ns, bodo.timedelta64ns, bodo.pd_timestamp_type, bodo.
        pd_timedelta_type]):
        raise BodoError(
            f'{func_name}() Series data with type {arr} not yet supported')
    xex__llmp = is_overload_constant_nan(other)
    if not (is_default or xex__llmp or is_scalar_type(other) or isinstance(
        other, types.Array) and other.ndim >= 1 and other.ndim <= max_ndim or
        isinstance(other, SeriesType) and (isinstance(arr, types.Array) or 
        arr.dtype in [bodo.string_type, bodo.bytes_type]) or isinstance(
        other, StringArrayType) and (arr.dtype == bodo.string_type or 
        isinstance(arr, bodo.CategoricalArrayType) and arr.dtype.elem_type ==
        bodo.string_type) or isinstance(other, BinaryArrayType) and (arr.
        dtype == bodo.bytes_type or isinstance(arr, bodo.
        CategoricalArrayType) and arr.dtype.elem_type == bodo.bytes_type) or
        (not isinstance(other, (StringArrayType, BinaryArrayType)) and (
        isinstance(arr.dtype, types.Integer) and (bodo.utils.utils.
        is_array_typ(other) and isinstance(other.dtype, types.Integer) or 
        is_series_type(other) and isinstance(other.dtype, types.Integer))) or
        (bodo.utils.utils.is_array_typ(other) and arr.dtype == other.dtype or
        is_series_type(other) and arr.dtype == other.dtype)) and (
        isinstance(arr, BooleanArrayType) or isinstance(arr, IntegerArrayType))
        ):
        raise BodoError(
            f"{func_name}() 'other' must be a scalar, non-categorical series, 1-dim numpy array or StringArray with a matching type for Series."
            )
    if not is_default:
        if isinstance(arr.dtype, bodo.PDCategoricalDtype):
            pqt__nxya = arr.dtype.elem_type
        else:
            pqt__nxya = arr.dtype
        if is_iterable_type(other):
            jad__odmvg = other.dtype
        elif xex__llmp:
            jad__odmvg = types.float64
        else:
            jad__odmvg = types.unliteral(other)
        if not xex__llmp and not is_common_scalar_dtype([pqt__nxya, jad__odmvg]
            ):
            raise BodoError(
                f"{func_name}() series and 'other' must share a common type.")


def create_explicit_binary_op_overload(op):

    def overload_series_explicit_binary_op(S, other, level=None, fill_value
        =None, axis=0):
        vxivd__glivu = dict(level=level, axis=axis)
        prr__rqtl = dict(level=None, axis=0)
        check_unsupported_args('series.{}'.format(op.__name__),
            vxivd__glivu, prr__rqtl, package_name='pandas', module_name=
            'Series')
        mqe__xlyos = other == string_type or is_overload_constant_str(other)
        wwmf__dimio = is_iterable_type(other) and other.dtype == string_type
        rckqf__bzttc = S.dtype == string_type and (op == operator.add and (
            mqe__xlyos or wwmf__dimio) or op == operator.mul and isinstance
            (other, types.Integer))
        ybypj__zowl = S.dtype == bodo.timedelta64ns
        pux__dbjr = S.dtype == bodo.datetime64ns
        yez__seufd = is_iterable_type(other) and (other.dtype ==
            datetime_timedelta_type or other.dtype == bodo.timedelta64ns)
        zcwix__dvgz = is_iterable_type(other) and (other.dtype ==
            datetime_datetime_type or other.dtype == pd_timestamp_type or 
            other.dtype == bodo.datetime64ns)
        jkz__regs = ybypj__zowl and (yez__seufd or zcwix__dvgz
            ) or pux__dbjr and yez__seufd
        jkz__regs = jkz__regs and op == operator.add
        if not (isinstance(S.dtype, types.Number) or rckqf__bzttc or jkz__regs
            ):
            raise BodoError(f'Unsupported types for Series.{op.__name__}')
        odj__gbsl = numba.core.registry.cpu_target.typing_context
        if is_scalar_type(other):
            args = S.data, other
            txlh__vallb = odj__gbsl.resolve_function_type(op, args, {}
                ).return_type
            if isinstance(S.data, IntegerArrayType
                ) and txlh__vallb == types.Array(types.bool_, 1, 'C'):
                txlh__vallb = boolean_array

            def impl_scalar(S, other, level=None, fill_value=None, axis=0):
                arr = bodo.hiframes.pd_series_ext.get_series_data(S)
                index = bodo.hiframes.pd_series_ext.get_series_index(S)
                name = bodo.hiframes.pd_series_ext.get_series_name(S)
                numba.parfors.parfor.init_prange()
                other = bodo.utils.conversion.unbox_if_timestamp(other)
                n = len(arr)
                ivga__befgo = bodo.utils.utils.alloc_type(n, txlh__vallb, (-1,)
                    )
                for dot__kzj in numba.parfors.parfor.internal_prange(n):
                    xxfwb__vuiov = bodo.libs.array_kernels.isna(arr, dot__kzj)
                    if xxfwb__vuiov:
                        if fill_value is None:
                            bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj
                                )
                        else:
                            ivga__befgo[dot__kzj] = op(fill_value, other)
                    else:
                        ivga__befgo[dot__kzj] = op(arr[dot__kzj], other)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl_scalar
        args = S.data, types.Array(other.dtype, 1, 'C')
        txlh__vallb = odj__gbsl.resolve_function_type(op, args, {}).return_type
        if isinstance(S.data, IntegerArrayType) and txlh__vallb == types.Array(
            types.bool_, 1, 'C'):
            txlh__vallb = boolean_array

        def impl(S, other, level=None, fill_value=None, axis=0):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            cuw__okn = bodo.utils.conversion.coerce_to_array(other)
            numba.parfors.parfor.init_prange()
            n = len(arr)
            ivga__befgo = bodo.utils.utils.alloc_type(n, txlh__vallb, (-1,))
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                xxfwb__vuiov = bodo.libs.array_kernels.isna(arr, dot__kzj)
                pswcn__qmvb = bodo.libs.array_kernels.isna(cuw__okn, dot__kzj)
                if xxfwb__vuiov and pswcn__qmvb:
                    bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                elif xxfwb__vuiov:
                    if fill_value is None:
                        bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                    else:
                        ivga__befgo[dot__kzj] = op(fill_value, cuw__okn[
                            dot__kzj])
                elif pswcn__qmvb:
                    if fill_value is None:
                        bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                    else:
                        ivga__befgo[dot__kzj] = op(arr[dot__kzj], fill_value)
                else:
                    ivga__befgo[dot__kzj] = op(arr[dot__kzj], cuw__okn[
                        dot__kzj])
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl
    return overload_series_explicit_binary_op


def create_explicit_binary_reverse_op_overload(op):

    def overload_series_explicit_binary_reverse_op(S, other, level=None,
        fill_value=None, axis=0):
        if not is_overload_none(level):
            raise BodoError('level argument not supported')
        if not is_overload_zero(axis):
            raise BodoError('axis argument not supported')
        if not isinstance(S.dtype, types.Number):
            raise BodoError('only numeric values supported')
        odj__gbsl = numba.core.registry.cpu_target.typing_context
        if isinstance(other, types.Number):
            args = other, S.data
            txlh__vallb = odj__gbsl.resolve_function_type(op, args, {}
                ).return_type
            if isinstance(S.data, IntegerArrayType
                ) and txlh__vallb == types.Array(types.bool_, 1, 'C'):
                txlh__vallb = boolean_array

            def impl_scalar(S, other, level=None, fill_value=None, axis=0):
                arr = bodo.hiframes.pd_series_ext.get_series_data(S)
                index = bodo.hiframes.pd_series_ext.get_series_index(S)
                name = bodo.hiframes.pd_series_ext.get_series_name(S)
                numba.parfors.parfor.init_prange()
                n = len(arr)
                ivga__befgo = bodo.utils.utils.alloc_type(n, txlh__vallb, None)
                for dot__kzj in numba.parfors.parfor.internal_prange(n):
                    xxfwb__vuiov = bodo.libs.array_kernels.isna(arr, dot__kzj)
                    if xxfwb__vuiov:
                        if fill_value is None:
                            bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj
                                )
                        else:
                            ivga__befgo[dot__kzj] = op(other, fill_value)
                    else:
                        ivga__befgo[dot__kzj] = op(other, arr[dot__kzj])
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl_scalar
        args = types.Array(other.dtype, 1, 'C'), S.data
        txlh__vallb = odj__gbsl.resolve_function_type(op, args, {}).return_type
        if isinstance(S.data, IntegerArrayType) and txlh__vallb == types.Array(
            types.bool_, 1, 'C'):
            txlh__vallb = boolean_array

        def impl(S, other, level=None, fill_value=None, axis=0):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            cuw__okn = bodo.hiframes.pd_series_ext.get_series_data(other)
            numba.parfors.parfor.init_prange()
            n = len(arr)
            ivga__befgo = bodo.utils.utils.alloc_type(n, txlh__vallb, None)
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                xxfwb__vuiov = bodo.libs.array_kernels.isna(arr, dot__kzj)
                pswcn__qmvb = bodo.libs.array_kernels.isna(cuw__okn, dot__kzj)
                ivga__befgo[dot__kzj] = op(cuw__okn[dot__kzj], arr[dot__kzj])
                if xxfwb__vuiov and pswcn__qmvb:
                    bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                elif xxfwb__vuiov:
                    if fill_value is None:
                        bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                    else:
                        ivga__befgo[dot__kzj] = op(cuw__okn[dot__kzj],
                            fill_value)
                elif pswcn__qmvb:
                    if fill_value is None:
                        bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                    else:
                        ivga__befgo[dot__kzj] = op(fill_value, arr[dot__kzj])
                else:
                    ivga__befgo[dot__kzj] = op(cuw__okn[dot__kzj], arr[
                        dot__kzj])
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl
    return overload_series_explicit_binary_reverse_op


explicit_binop_funcs_two_ways = {operator.add: {'add'}, operator.sub: {
    'sub'}, operator.mul: {'mul'}, operator.truediv: {'div', 'truediv'},
    operator.floordiv: {'floordiv'}, operator.mod: {'mod'}, operator.pow: {
    'pow'}}
explicit_binop_funcs_single = {operator.lt: 'lt', operator.gt: 'gt',
    operator.le: 'le', operator.ge: 'ge', operator.ne: 'ne', operator.eq: 'eq'}
explicit_binop_funcs = set()
split_logical_binops_funcs = [operator.or_, operator.and_]


def _install_explicit_binary_ops():
    for op, yqz__jcgno in explicit_binop_funcs_two_ways.items():
        for name in yqz__jcgno:
            rgkm__wnfc = create_explicit_binary_op_overload(op)
            nrzqw__ogtbn = create_explicit_binary_reverse_op_overload(op)
            qxo__jbem = 'r' + name
            overload_method(SeriesType, name, no_unliteral=True)(rgkm__wnfc)
            overload_method(SeriesType, qxo__jbem, no_unliteral=True)(
                nrzqw__ogtbn)
            explicit_binop_funcs.add(name)
    for op, name in explicit_binop_funcs_single.items():
        rgkm__wnfc = create_explicit_binary_op_overload(op)
        overload_method(SeriesType, name, no_unliteral=True)(rgkm__wnfc)
        explicit_binop_funcs.add(name)


_install_explicit_binary_ops()


def create_binary_op_overload(op):

    def overload_series_binary_op(lhs, rhs):
        if (isinstance(lhs, SeriesType) and isinstance(rhs, SeriesType) and
            lhs.dtype == bodo.datetime64ns and rhs.dtype == bodo.
            datetime64ns and op == operator.sub):

            def impl_dt64(lhs, rhs):
                arr = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                index = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                name = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                kmsuz__dqn = (bodo.utils.conversion.
                    get_array_if_series_or_index(rhs))
                ivga__befgo = dt64_arr_sub(arr, kmsuz__dqn)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl_dt64
        if op in [operator.add, operator.sub] and isinstance(lhs, SeriesType
            ) and lhs.dtype == bodo.datetime64ns and is_offsets_type(rhs):

            def impl_offsets(lhs, rhs):
                arr = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                index = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                name = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                ivga__befgo = np.empty(n, np.dtype('datetime64[ns]'))
                for dot__kzj in numba.parfors.parfor.internal_prange(n):
                    if bodo.libs.array_kernels.isna(arr, dot__kzj):
                        bodo.libs.array_kernels.setna(ivga__befgo, dot__kzj)
                        continue
                    sbm__mjm = (bodo.hiframes.pd_timestamp_ext.
                        convert_datetime64_to_timestamp(arr[dot__kzj]))
                    dew__qieb = op(sbm__mjm, rhs)
                    ivga__befgo[dot__kzj
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_dt64(
                        dew__qieb.value)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl_offsets
        if op == operator.add and is_offsets_type(lhs) and isinstance(rhs,
            SeriesType) and rhs.dtype == bodo.datetime64ns:

            def impl(lhs, rhs):
                return op(rhs, lhs)
            return impl
        if isinstance(lhs, SeriesType):
            if lhs.dtype in [bodo.datetime64ns, bodo.timedelta64ns]:

                def impl(lhs, rhs):
                    arr = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                    index = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                    name = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                    kmsuz__dqn = (bodo.utils.conversion.
                        get_array_if_series_or_index(rhs))
                    ivga__befgo = op(arr, bodo.utils.conversion.
                        unbox_if_timestamp(kmsuz__dqn))
                    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                        index, name)
                return impl

            def impl(lhs, rhs):
                arr = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                index = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                name = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                kmsuz__dqn = (bodo.utils.conversion.
                    get_array_if_series_or_index(rhs))
                ivga__befgo = op(arr, kmsuz__dqn)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl
        if isinstance(rhs, SeriesType):
            if rhs.dtype in [bodo.datetime64ns, bodo.timedelta64ns]:

                def impl(lhs, rhs):
                    arr = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                    index = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                    name = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                    iiz__sfaa = (bodo.utils.conversion.
                        get_array_if_series_or_index(lhs))
                    ivga__befgo = op(bodo.utils.conversion.
                        unbox_if_timestamp(iiz__sfaa), arr)
                    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                        index, name)
                return impl

            def impl(lhs, rhs):
                arr = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                index = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                name = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                iiz__sfaa = bodo.utils.conversion.get_array_if_series_or_index(
                    lhs)
                ivga__befgo = op(iiz__sfaa, arr)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl
    return overload_series_binary_op


skips = list(explicit_binop_funcs_two_ways.keys()) + list(
    explicit_binop_funcs_single.keys()) + split_logical_binops_funcs


def _install_binary_ops():
    for op in bodo.hiframes.pd_series_ext.series_binary_ops:
        if op in skips:
            continue
        rgkm__wnfc = create_binary_op_overload(op)
        overload(op)(rgkm__wnfc)


_install_binary_ops()


def dt64_arr_sub(arg1, arg2):
    return arg1 - arg2


@overload(dt64_arr_sub, no_unliteral=True)
def overload_dt64_arr_sub(arg1, arg2):
    assert arg1 == types.Array(bodo.datetime64ns, 1, 'C'
        ) and arg2 == types.Array(bodo.datetime64ns, 1, 'C')
    ofe__kfizq = np.dtype('timedelta64[ns]')

    def impl(arg1, arg2):
        numba.parfors.parfor.init_prange()
        n = len(arg1)
        S = np.empty(n, ofe__kfizq)
        for dot__kzj in numba.parfors.parfor.internal_prange(n):
            if bodo.libs.array_kernels.isna(arg1, dot__kzj
                ) or bodo.libs.array_kernels.isna(arg2, dot__kzj):
                bodo.libs.array_kernels.setna(S, dot__kzj)
                continue
            S[dot__kzj
                ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                bodo.hiframes.pd_timestamp_ext.dt64_to_integer(arg1[
                dot__kzj]) - bodo.hiframes.pd_timestamp_ext.dt64_to_integer
                (arg2[dot__kzj]))
        return S
    return impl


def create_inplace_binary_op_overload(op):

    def overload_series_inplace_binary_op(S, other):
        if isinstance(S, SeriesType) or isinstance(other, SeriesType):

            def impl(S, other):
                arr = bodo.utils.conversion.get_array_if_series_or_index(S)
                cuw__okn = bodo.utils.conversion.get_array_if_series_or_index(
                    other)
                op(arr, cuw__okn)
                return S
            return impl
    return overload_series_inplace_binary_op


def _install_inplace_binary_ops():
    for op in bodo.hiframes.pd_series_ext.series_inplace_binary_ops:
        rgkm__wnfc = create_inplace_binary_op_overload(op)
        overload(op, no_unliteral=True)(rgkm__wnfc)


_install_inplace_binary_ops()


def create_unary_op_overload(op):

    def overload_series_unary_op(S):
        if isinstance(S, SeriesType):

            def impl(S):
                arr = bodo.hiframes.pd_series_ext.get_series_data(S)
                index = bodo.hiframes.pd_series_ext.get_series_index(S)
                name = bodo.hiframes.pd_series_ext.get_series_name(S)
                ivga__befgo = op(arr)
                return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                    index, name)
            return impl
    return overload_series_unary_op


def _install_unary_ops():
    for op in bodo.hiframes.pd_series_ext.series_unary_ops:
        rgkm__wnfc = create_unary_op_overload(op)
        overload(op, no_unliteral=True)(rgkm__wnfc)


_install_unary_ops()


def create_ufunc_overload(ufunc):
    if ufunc.nin == 1:

        def overload_series_ufunc_nin_1(S):
            if isinstance(S, SeriesType):

                def impl(S):
                    arr = bodo.hiframes.pd_series_ext.get_series_data(S)
                    index = bodo.hiframes.pd_series_ext.get_series_index(S)
                    name = bodo.hiframes.pd_series_ext.get_series_name(S)
                    ivga__befgo = ufunc(arr)
                    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                        index, name)
                return impl
        return overload_series_ufunc_nin_1
    elif ufunc.nin == 2:

        def overload_series_ufunc_nin_2(S1, S2):
            if isinstance(S1, SeriesType):

                def impl(S1, S2):
                    arr = bodo.hiframes.pd_series_ext.get_series_data(S1)
                    index = bodo.hiframes.pd_series_ext.get_series_index(S1)
                    name = bodo.hiframes.pd_series_ext.get_series_name(S1)
                    cuw__okn = (bodo.utils.conversion.
                        get_array_if_series_or_index(S2))
                    ivga__befgo = ufunc(arr, cuw__okn)
                    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                        index, name)
                return impl
            elif isinstance(S2, SeriesType):

                def impl(S1, S2):
                    arr = bodo.utils.conversion.get_array_if_series_or_index(S1
                        )
                    cuw__okn = bodo.hiframes.pd_series_ext.get_series_data(S2)
                    index = bodo.hiframes.pd_series_ext.get_series_index(S2)
                    name = bodo.hiframes.pd_series_ext.get_series_name(S2)
                    ivga__befgo = ufunc(arr, cuw__okn)
                    return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                        index, name)
                return impl
        return overload_series_ufunc_nin_2
    else:
        raise RuntimeError(
            "Don't know how to register ufuncs from ufunc_db with arity > 2")


def _install_np_ufuncs():
    import numba.np.ufunc_db
    for ufunc in numba.np.ufunc_db.get_ufuncs():
        rgkm__wnfc = create_ufunc_overload(ufunc)
        overload(ufunc, no_unliteral=True)(rgkm__wnfc)


_install_np_ufuncs()


def argsort(A):
    return np.argsort(A)


@overload(argsort, no_unliteral=True)
def overload_argsort(A):

    def impl(A):
        n = len(A)
        jxs__gsrtl = bodo.libs.str_arr_ext.to_list_if_immutable_arr((A.copy(),)
            )
        djq__aqsup = np.arange(n),
        bodo.libs.timsort.sort(jxs__gsrtl, 0, n, djq__aqsup)
        return djq__aqsup[0]
    return impl


@overload(pd.to_numeric, inline='always', no_unliteral=True)
def overload_to_numeric(arg_a, errors='raise', downcast=None):
    if not is_overload_none(downcast) and not (is_overload_constant_str(
        downcast) and get_overload_const_str(downcast) in ('integer',
        'signed', 'unsigned', 'float')):
        raise BodoError(
            'pd.to_numeric(): invalid downcasting method provided {}'.
            format(downcast))
    out_dtype = types.float64
    if not is_overload_none(downcast):
        croz__avs = get_overload_const_str(downcast)
        if croz__avs in ('integer', 'signed'):
            out_dtype = types.int64
        elif croz__avs == 'unsigned':
            out_dtype = types.uint64
        else:
            assert croz__avs == 'float'
    if isinstance(arg_a, (types.Array, IntegerArrayType)):
        return lambda arg_a, errors='raise', downcast=None: arg_a.astype(
            out_dtype)
    if isinstance(arg_a, SeriesType):

        def impl_series(arg_a, errors='raise', downcast=None):
            glfic__kmbxv = bodo.hiframes.pd_series_ext.get_series_data(arg_a)
            index = bodo.hiframes.pd_series_ext.get_series_index(arg_a)
            name = bodo.hiframes.pd_series_ext.get_series_name(arg_a)
            ivga__befgo = pd.to_numeric(glfic__kmbxv, errors, downcast)
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                index, name)
        return impl_series
    if arg_a != string_array_type:
        raise BodoError('pd.to_numeric(): invalid argument type {}'.format(
            arg_a))
    if out_dtype == types.float64:

        def to_numeric_float_impl(arg_a, errors='raise', downcast=None):
            numba.parfors.parfor.init_prange()
            n = len(arg_a)
            rmli__zihf = np.empty(n, np.float64)
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                if bodo.libs.array_kernels.isna(arg_a, dot__kzj):
                    bodo.libs.array_kernels.setna(rmli__zihf, dot__kzj)
                else:
                    bodo.libs.str_arr_ext.str_arr_item_to_numeric(rmli__zihf,
                        dot__kzj, arg_a, dot__kzj)
            return rmli__zihf
        return to_numeric_float_impl
    else:

        def to_numeric_int_impl(arg_a, errors='raise', downcast=None):
            numba.parfors.parfor.init_prange()
            n = len(arg_a)
            rmli__zihf = bodo.libs.int_arr_ext.alloc_int_array(n, np.int64)
            for dot__kzj in numba.parfors.parfor.internal_prange(n):
                if bodo.libs.array_kernels.isna(arg_a, dot__kzj):
                    bodo.libs.array_kernels.setna(rmli__zihf, dot__kzj)
                else:
                    bodo.libs.str_arr_ext.str_arr_item_to_numeric(rmli__zihf,
                        dot__kzj, arg_a, dot__kzj)
            return rmli__zihf
        return to_numeric_int_impl


def series_filter_bool(arr, bool_arr):
    return arr[bool_arr]


@infer_global(series_filter_bool)
class SeriesFilterBoolInfer(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 2
        vob__wnyf = if_series_to_array_type(args[0])
        if isinstance(vob__wnyf, types.Array) and isinstance(vob__wnyf.
            dtype, types.Integer):
            vob__wnyf = types.Array(types.float64, 1, 'C')
        return vob__wnyf(*args)


def where_impl_one_arg(c):
    return np.where(c)


@overload(where_impl_one_arg, no_unliteral=True)
def overload_where_unsupported_one_arg(condition):
    if isinstance(condition, SeriesType) or bodo.utils.utils.is_array_typ(
        condition, False):
        return lambda condition: np.where(condition)


def overload_np_where_one_arg(condition):
    if isinstance(condition, SeriesType):

        def impl_series(condition):
            condition = bodo.hiframes.pd_series_ext.get_series_data(condition)
            return bodo.libs.array_kernels.nonzero(condition)
        return impl_series
    elif bodo.utils.utils.is_array_typ(condition, False):

        def impl(condition):
            return bodo.libs.array_kernels.nonzero(condition)
        return impl


overload(np.where, inline='always', no_unliteral=True)(
    overload_np_where_one_arg)
overload(where_impl_one_arg, inline='always', no_unliteral=True)(
    overload_np_where_one_arg)


def where_impl(c, x, y):
    return np.where(c, x, y)


@overload(where_impl, no_unliteral=True)
def overload_where_unsupported(condition, x, y):
    if not isinstance(condition, (SeriesType, types.Array, BooleanArrayType)
        ) or condition.ndim != 1:
        return lambda condition, x, y: np.where(condition, x, y)


@overload(where_impl, no_unliteral=True)
@overload(np.where, no_unliteral=True)
def overload_np_where(condition, x, y):
    if not isinstance(condition, (SeriesType, types.Array, BooleanArrayType)
        ) or condition.ndim != 1:
        return
    assert condition.dtype == types.bool_, 'invalid condition dtype'
    qchg__sjuxc = bodo.utils.utils.is_array_typ(x, True)
    cif__esg = bodo.utils.utils.is_array_typ(y, True)
    utmrs__nduqj = 'def _impl(condition, x, y):\n'
    if isinstance(condition, SeriesType):
        utmrs__nduqj += (
            '  condition = bodo.hiframes.pd_series_ext.get_series_data(condition)\n'
            )
    if qchg__sjuxc and not bodo.utils.utils.is_array_typ(x, False):
        utmrs__nduqj += '  x = bodo.utils.conversion.coerce_to_array(x)\n'
    if cif__esg and not bodo.utils.utils.is_array_typ(y, False):
        utmrs__nduqj += '  y = bodo.utils.conversion.coerce_to_array(y)\n'
    utmrs__nduqj += '  n = len(condition)\n'
    jdja__ybppq = x.dtype if qchg__sjuxc else types.unliteral(x)
    iirxc__tnv = y.dtype if cif__esg else types.unliteral(y)
    if not isinstance(x, CategoricalArrayType):
        jdja__ybppq = element_type(x)
    if not isinstance(y, CategoricalArrayType):
        iirxc__tnv = element_type(y)

    def get_data(x):
        if isinstance(x, SeriesType):
            return x.data
        elif isinstance(x, types.Array):
            return x
        return types.unliteral(x)
    jlmbo__yik = get_data(x)
    dbyw__uqica = get_data(y)
    is_nullable = any(bodo.utils.typing.is_nullable(djq__aqsup) for
        djq__aqsup in [jlmbo__yik, dbyw__uqica])
    if dbyw__uqica == types.none:
        if isinstance(jdja__ybppq, types.Number):
            out_dtype = types.Array(types.float64, 1, 'C')
        else:
            out_dtype = to_nullable_type(x)
    elif jlmbo__yik == dbyw__uqica and not is_nullable:
        out_dtype = dtype_to_array_type(jdja__ybppq)
    elif jdja__ybppq == string_type or iirxc__tnv == string_type:
        out_dtype = bodo.string_array_type
    elif jlmbo__yik == bytes_type or (qchg__sjuxc and jdja__ybppq == bytes_type
        ) and (dbyw__uqica == bytes_type or cif__esg and iirxc__tnv ==
        bytes_type):
        out_dtype = binary_array_type
    elif isinstance(jdja__ybppq, bodo.PDCategoricalDtype):
        out_dtype = None
    elif jdja__ybppq in [bodo.timedelta64ns, bodo.datetime64ns]:
        out_dtype = types.Array(jdja__ybppq, 1, 'C')
    elif iirxc__tnv in [bodo.timedelta64ns, bodo.datetime64ns]:
        out_dtype = types.Array(iirxc__tnv, 1, 'C')
    else:
        out_dtype = numba.from_dtype(np.promote_types(numba.np.
            numpy_support.as_dtype(jdja__ybppq), numba.np.numpy_support.
            as_dtype(iirxc__tnv)))
        out_dtype = types.Array(out_dtype, 1, 'C')
        if is_nullable:
            out_dtype = bodo.utils.typing.to_nullable_type(out_dtype)
    if isinstance(jdja__ybppq, bodo.PDCategoricalDtype):
        ymj__oxh = 'x'
    else:
        ymj__oxh = 'out_dtype'
    utmrs__nduqj += (
        f'  out_arr = bodo.utils.utils.alloc_type(n, {ymj__oxh}, (-1,))\n')
    if isinstance(jdja__ybppq, bodo.PDCategoricalDtype):
        utmrs__nduqj += """  out_codes = bodo.hiframes.pd_categorical_ext.get_categorical_arr_codes(out_arr)
"""
        utmrs__nduqj += """  x_codes = bodo.hiframes.pd_categorical_ext.get_categorical_arr_codes(x)
"""
    utmrs__nduqj += '  for j in numba.parfors.parfor.internal_prange(n):\n'
    utmrs__nduqj += (
        '    if not bodo.libs.array_kernels.isna(condition, j) and condition[j]:\n'
        )
    if qchg__sjuxc:
        utmrs__nduqj += '      if bodo.libs.array_kernels.isna(x, j):\n'
        utmrs__nduqj += '        setna(out_arr, j)\n'
        utmrs__nduqj += '        continue\n'
    if isinstance(jdja__ybppq, bodo.PDCategoricalDtype):
        utmrs__nduqj += '      out_codes[j] = x_codes[j]\n'
    else:
        utmrs__nduqj += (
            '      out_arr[j] = bodo.utils.conversion.unbox_if_timestamp({})\n'
            .format('x[j]' if qchg__sjuxc else 'x'))
    utmrs__nduqj += '    else:\n'
    if cif__esg:
        utmrs__nduqj += '      if bodo.libs.array_kernels.isna(y, j):\n'
        utmrs__nduqj += '        setna(out_arr, j)\n'
        utmrs__nduqj += '        continue\n'
    if dbyw__uqica == types.none:
        if isinstance(jdja__ybppq, bodo.PDCategoricalDtype):
            utmrs__nduqj += '      out_codes[j] = -1\n'
        else:
            utmrs__nduqj += '      setna(out_arr, j)\n'
    else:
        utmrs__nduqj += (
            '      out_arr[j] = bodo.utils.conversion.unbox_if_timestamp({})\n'
            .format('y[j]' if cif__esg else 'y'))
    utmrs__nduqj += '  return out_arr\n'
    snqc__makr = {}
    exec(utmrs__nduqj, {'bodo': bodo, 'numba': numba, 'setna': bodo.libs.
        array_kernels.setna, 'np': np, 'out_dtype': out_dtype}, snqc__makr)
    ieyvt__msc = snqc__makr['_impl']
    return ieyvt__msc


def _verify_np_select_arg_typs(condlist, choicelist, default):
    if isinstance(condlist, (types.List, types.UniTuple)):
        if not (bodo.utils.utils.is_np_array_typ(condlist.dtype) and 
            condlist.dtype.dtype == types.bool_):
            raise BodoError(
                "np.select(): 'condlist' argument must be list or tuple of boolean ndarrays. If passing a Series, please convert with pd.Series.to_numpy()."
                )
    else:
        raise BodoError(
            "np.select(): 'condlist' argument must be list or tuple of boolean ndarrays. If passing a Series, please convert with pd.Series.to_numpy()."
            )
    if not isinstance(choicelist, (types.List, types.UniTuple, types.BaseTuple)
        ):
        raise BodoError(
            "np.select(): 'choicelist' argument must be list or tuple type")
    if isinstance(choicelist, (types.List, types.UniTuple)):
        uuwq__oizk = choicelist.dtype
        if not bodo.utils.utils.is_array_typ(uuwq__oizk, True):
            raise BodoError(
                "np.select(): 'choicelist' argument must be list or tuple of series/arrays types"
                )
        if is_series_type(uuwq__oizk):
            zpaq__vnn = uuwq__oizk.data.dtype
        else:
            zpaq__vnn = uuwq__oizk.dtype
        if isinstance(zpaq__vnn, bodo.PDCategoricalDtype):
            raise BodoError(
                'np.select(): data with choicelist of type Categorical not yet supported'
                )
        lvr__mfong = uuwq__oizk
    else:
        vzxp__lrzlc = []
        for uuwq__oizk in choicelist:
            if not bodo.utils.utils.is_array_typ(uuwq__oizk, True):
                raise BodoError(
                    "np.select(): 'choicelist' argument must be list or tuple of series/arrays types"
                    )
            if is_series_type(uuwq__oizk):
                zpaq__vnn = uuwq__oizk.data.dtype
            else:
                zpaq__vnn = uuwq__oizk.dtype
            if isinstance(zpaq__vnn, bodo.PDCategoricalDtype):
                raise BodoError(
                    'np.select(): data with choicelist of type Categorical not yet supported'
                    )
            vzxp__lrzlc.append(zpaq__vnn)
        if not is_common_scalar_dtype(vzxp__lrzlc):
            raise BodoError(
                f"np.select(): 'choicelist' items must be arrays with a commmon data type. Found a tuple with the following data types {choicelist}."
                )
        lvr__mfong = choicelist[0]
    if is_series_type(lvr__mfong):
        lvr__mfong = lvr__mfong.data
    if is_overload_constant_int(default) and get_overload_const_int(default
        ) == 0:
        pass
    else:
        if not is_scalar_type(default):
            raise BodoError(
                "np.select(): 'default' argument must be scalar type")
        if not (is_common_scalar_dtype([default, lvr__mfong.dtype]) or 
            default == types.none or is_overload_constant_nan(default)):
            raise BodoError(
                f"np.select(): 'default' is not type compatible with the array types in choicelist. Choicelist type: {choicelist}, Default type: {default}"
                )
    if not (isinstance(lvr__mfong, types.Array) or isinstance(lvr__mfong,
        BooleanArrayType) or isinstance(lvr__mfong, IntegerArrayType) or 
        bodo.utils.utils.is_array_typ(lvr__mfong, False) and lvr__mfong.
        dtype in [bodo.string_type, bodo.bytes_type]):
        raise BodoError(
            f'np.select(): data with choicelist of type {lvr__mfong} not yet supported'
            )


@overload(np.select)
def overload_np_select(condlist, choicelist, default=0):
    _verify_np_select_arg_typs(condlist, choicelist, default)
    uziix__idp = isinstance(choicelist, (types.List, types.UniTuple)
        ) and isinstance(condlist, (types.List, types.UniTuple))
    if isinstance(choicelist, (types.List, types.UniTuple)):
        entks__fwrzs = choicelist.dtype
    else:
        jdep__xnsfh = False
        vzxp__lrzlc = []
        for uuwq__oizk in choicelist:
            if is_nullable_type(uuwq__oizk):
                jdep__xnsfh = True
            if is_series_type(uuwq__oizk):
                zpaq__vnn = uuwq__oizk.data.dtype
            else:
                zpaq__vnn = uuwq__oizk.dtype
            if isinstance(zpaq__vnn, bodo.PDCategoricalDtype):
                raise BodoError(
                    'np.select(): data with choicelist of type Categorical not yet supported'
                    )
            vzxp__lrzlc.append(zpaq__vnn)
        fhi__ovunp, ocpyd__ihh = get_common_scalar_dtype(vzxp__lrzlc)
        if not ocpyd__ihh:
            raise BodoError('Internal error in overload_np_select')
        wyejz__azdy = dtype_to_array_type(fhi__ovunp)
        if jdep__xnsfh:
            wyejz__azdy = to_nullable_type(wyejz__azdy)
        entks__fwrzs = wyejz__azdy
    if isinstance(entks__fwrzs, SeriesType):
        entks__fwrzs = entks__fwrzs.data
    if is_overload_constant_int(default) and get_overload_const_int(default
        ) == 0:
        uutw__jzml = True
    else:
        uutw__jzml = False
    bfsrl__rhc = False
    pbqv__fpxd = False
    if uutw__jzml:
        if isinstance(entks__fwrzs.dtype, types.Number):
            pass
        elif entks__fwrzs.dtype == types.bool_:
            pbqv__fpxd = True
        else:
            bfsrl__rhc = True
            entks__fwrzs = to_nullable_type(entks__fwrzs)
    elif default == types.none or is_overload_constant_nan(default):
        bfsrl__rhc = True
        entks__fwrzs = to_nullable_type(entks__fwrzs)
    utmrs__nduqj = 'def np_select_impl(condlist, choicelist, default=0):\n'
    utmrs__nduqj += '  if len(condlist) != len(choicelist):\n'
    utmrs__nduqj += """    raise ValueError('list of cases must be same length as list of conditions')
"""
    utmrs__nduqj += '  output_len = len(choicelist[0])\n'
    utmrs__nduqj += (
        '  out = bodo.utils.utils.alloc_type(output_len, alloc_typ, (-1,))\n')
    utmrs__nduqj += '  for i in range(output_len):\n'
    if bfsrl__rhc:
        utmrs__nduqj += '    bodo.libs.array_kernels.setna(out, i)\n'
    elif pbqv__fpxd:
        utmrs__nduqj += '    out[i] = False\n'
    else:
        utmrs__nduqj += '    out[i] = default\n'
    if uziix__idp:
        utmrs__nduqj += '  for i in range(len(condlist) - 1, -1, -1):\n'
        utmrs__nduqj += '    cond = condlist[i]\n'
        utmrs__nduqj += '    choice = choicelist[i]\n'
        utmrs__nduqj += '    out = np.where(cond, choice, out)\n'
    else:
        for dot__kzj in range(len(choicelist) - 1, -1, -1):
            utmrs__nduqj += f'  cond = condlist[{dot__kzj}]\n'
            utmrs__nduqj += f'  choice = choicelist[{dot__kzj}]\n'
            utmrs__nduqj += f'  out = np.where(cond, choice, out)\n'
    utmrs__nduqj += '  return out'
    snqc__makr = dict()
    exec(utmrs__nduqj, {'bodo': bodo, 'numba': numba, 'setna': bodo.libs.
        array_kernels.setna, 'np': np, 'alloc_typ': entks__fwrzs}, snqc__makr)
    impl = snqc__makr['np_select_impl']
    return impl


@overload_method(SeriesType, 'duplicated', inline='always', no_unliteral=True)
def overload_series_duplicated(S, keep='first'):

    def impl(S, keep='first'):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        hls__iymkx = bodo.utils.conversion.index_to_array(index)
        ivga__befgo, vurp__mxlc = bodo.libs.array_kernels.duplicated((arr,),
            hls__iymkx)
        juqvh__wrsfg = bodo.utils.conversion.index_from_array(vurp__mxlc)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl


@overload_method(SeriesType, 'drop_duplicates', inline='always',
    no_unliteral=True)
def overload_series_drop_duplicates(S, subset=None, keep='first', inplace=False
    ):
    vxivd__glivu = dict(subset=subset, keep=keep, inplace=inplace)
    prr__rqtl = dict(subset=None, keep='first', inplace=False)
    check_unsupported_args('Series.drop_duplicates', vxivd__glivu,
        prr__rqtl, package_name='pandas', module_name='Series')

    def impl(S, subset=None, keep='first', inplace=False):
        kzipb__omk = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.utils.conversion.index_to_array(bodo.hiframes.
            pd_series_ext.get_series_index(S))
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        (kzipb__omk,), hls__iymkx = bodo.libs.array_kernels.drop_duplicates((
            kzipb__omk,), index, 1)
        index = bodo.utils.conversion.index_from_array(hls__iymkx)
        return bodo.hiframes.pd_series_ext.init_series(kzipb__omk, index, name)
    return impl


@overload_method(SeriesType, 'between', inline='always', no_unliteral=True)
def overload_series_between(S, left, right, inclusive='both'):
    vaqv__gsugn = element_type(S.data)
    if not is_common_scalar_dtype([vaqv__gsugn, left]):
        raise_bodo_error(
            "Series.between(): 'left' must be compariable with the Series data"
            )
    if not is_common_scalar_dtype([vaqv__gsugn, right]):
        raise_bodo_error(
            "Series.between(): 'right' must be compariable with the Series data"
            )
    if not is_overload_constant_str(inclusive) or get_overload_const_str(
        inclusive) not in ('both', 'neither'):
        raise_bodo_error(
            "Series.between(): 'inclusive' must be a constant string and one of ('both', 'neither')"
            )

    def impl(S, left, right, inclusive='both'):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        numba.parfors.parfor.init_prange()
        n = len(arr)
        ivga__befgo = np.empty(n, np.bool_)
        for dot__kzj in numba.parfors.parfor.internal_prange(n):
            jrrz__tkvr = bodo.utils.conversion.box_if_dt64(arr[dot__kzj])
            if inclusive == 'both':
                ivga__befgo[dot__kzj
                    ] = jrrz__tkvr <= right and jrrz__tkvr >= left
            else:
                ivga__befgo[dot__kzj
                    ] = jrrz__tkvr < right and jrrz__tkvr > left
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo, index, name
            )
    return impl


@overload_method(SeriesType, 'repeat', inline='always', no_unliteral=True)
def overload_series_repeat(S, repeats, axis=None):
    vxivd__glivu = dict(axis=axis)
    prr__rqtl = dict(axis=None)
    check_unsupported_args('Series.repeat', vxivd__glivu, prr__rqtl,
        package_name='pandas', module_name='Series')
    if not (isinstance(repeats, types.Integer) or is_iterable_type(repeats) and
        isinstance(repeats.dtype, types.Integer)):
        raise BodoError(
            "Series.repeat(): 'repeats' should be an integer or array of integers"
            )
    if isinstance(repeats, types.Integer):

        def impl_int(S, repeats, axis=None):
            arr = bodo.hiframes.pd_series_ext.get_series_data(S)
            index = bodo.hiframes.pd_series_ext.get_series_index(S)
            name = bodo.hiframes.pd_series_ext.get_series_name(S)
            hls__iymkx = bodo.utils.conversion.index_to_array(index)
            ivga__befgo = bodo.libs.array_kernels.repeat_kernel(arr, repeats)
            vurp__mxlc = bodo.libs.array_kernels.repeat_kernel(hls__iymkx,
                repeats)
            juqvh__wrsfg = bodo.utils.conversion.index_from_array(vurp__mxlc)
            return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
                juqvh__wrsfg, name)
        return impl_int

    def impl_arr(S, repeats, axis=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        name = bodo.hiframes.pd_series_ext.get_series_name(S)
        hls__iymkx = bodo.utils.conversion.index_to_array(index)
        repeats = bodo.utils.conversion.coerce_to_array(repeats)
        ivga__befgo = bodo.libs.array_kernels.repeat_kernel(arr, repeats)
        vurp__mxlc = bodo.libs.array_kernels.repeat_kernel(hls__iymkx, repeats)
        juqvh__wrsfg = bodo.utils.conversion.index_from_array(vurp__mxlc)
        return bodo.hiframes.pd_series_ext.init_series(ivga__befgo,
            juqvh__wrsfg, name)
    return impl_arr


@overload_method(SeriesType, 'to_dict', no_unliteral=True)
def overload_to_dict(S, into=None):

    def impl(S, into=None):
        djq__aqsup = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.utils.conversion.index_to_array(bodo.hiframes.
            pd_series_ext.get_series_index(S))
        n = len(djq__aqsup)
        hvgb__ore = {}
        for dot__kzj in range(n):
            jrrz__tkvr = bodo.utils.conversion.box_if_dt64(djq__aqsup[dot__kzj]
                )
            hvgb__ore[index[dot__kzj]] = jrrz__tkvr
        return hvgb__ore
    return impl


@overload_method(SeriesType, 'to_frame', inline='always', no_unliteral=True)
def overload_series_to_frame(S, name=None):
    qxpn__ngil = (
        "Series.to_frame(): output column name should be known at compile time. Set 'name' to a constant value."
        )
    if is_overload_none(name):
        if is_literal_type(S.name_typ):
            zytlh__ysca = get_literal_value(S.name_typ)
        else:
            raise_bodo_error(qxpn__ngil)
    elif is_literal_type(name):
        zytlh__ysca = get_literal_value(name)
    else:
        raise_bodo_error(qxpn__ngil)
    zytlh__ysca = 0 if zytlh__ysca is None else zytlh__ysca

    def impl(S, name=None):
        arr = bodo.hiframes.pd_series_ext.get_series_data(S)
        index = bodo.hiframes.pd_series_ext.get_series_index(S)
        return bodo.hiframes.pd_dataframe_ext.init_dataframe((arr,), index,
            (zytlh__ysca,))
    return impl


@overload_method(SeriesType, 'keys', inline='always', no_unliteral=True)
def overload_series_keys(S):

    def impl(S):
        return bodo.hiframes.pd_series_ext.get_series_index(S)
    return impl
