"""
Class information for DataFrame iterators returned by pd.read_csv. This is used
to handle situations in which pd.read_csv is used to return chunks with separate
read calls instead of just a single read.
"""
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, ir_utils, types
from numba.core.imputils import RefType, impl_ret_borrowed, iternext_impl
from numba.core.typing.templates import signature
from numba.extending import intrinsic, lower_builtin, models, register_model
import bodo
import bodo.ir.connector
import bodo.ir.csv_ext
from bodo import objmode
from bodo.hiframes.pd_dataframe_ext import DataFrameType
from bodo.hiframes.table import Table, TableType
from bodo.io import csv_cpp
from bodo.ir.csv_ext import _gen_read_csv_objmode, astype
from bodo.utils.utils import check_java_installation
from bodo.utils.utils import sanitize_varname
ll.add_symbol('update_csv_reader', csv_cpp.update_csv_reader)
ll.add_symbol('initialize_csv_reader', csv_cpp.initialize_csv_reader)


class CSVIteratorType(types.SimpleIteratorType):

    def __init__(self, df_type, out_colnames, out_types, usecols, sep,
        index_ind, index_arr_typ, index_name, escapechar):
        assert isinstance(df_type, DataFrameType
            ), 'CSVIterator must return a DataFrame'
        wsb__ofbqb = (
            f'CSVIteratorType({df_type}, {out_colnames}, {out_types}, {usecols}, {sep}, {index_ind}, {index_arr_typ}, {index_name}, {escapechar})'
            )
        super(types.SimpleIteratorType, self).__init__(wsb__ofbqb)
        self._yield_type = df_type
        self._out_colnames = out_colnames
        self._out_types = out_types
        self._usecols = usecols
        self._sep = sep
        self._index_ind = index_ind
        self._index_arr_typ = index_arr_typ
        self._index_name = index_name
        self._escapechar = escapechar

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(CSVIteratorType)
class CSVIteratorModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        wsma__lou = [('csv_reader', bodo.ir.connector.stream_reader_type),
            ('index', types.EphemeralPointer(types.uintp))]
        super(CSVIteratorModel, self).__init__(dmm, fe_type, wsma__lou)


@lower_builtin('getiter', CSVIteratorType)
def getiter_csv_iterator(context, builder, sig, args):
    xndq__gwjfx = cgutils.create_struct_proxy(sig.args[0])(context, builder,
        value=args[0])
    ojh__yspur = lir.FunctionType(lir.VoidType(), [lir.IntType(8).as_pointer()]
        )
    burig__aft = cgutils.get_or_insert_function(builder.module, ojh__yspur,
        name='initialize_csv_reader')
    builder.call(burig__aft, [xndq__gwjfx.csv_reader])
    builder.store(context.get_constant(types.uint64, 0), xndq__gwjfx.index)
    return impl_ret_borrowed(context, builder, sig.return_type, args[0])


@lower_builtin('iternext', CSVIteratorType)
@iternext_impl(RefType.NEW)
def iternext_csv_iterator(context, builder, sig, args, result):
    [zml__pygc] = sig.args
    [pvm__onh] = args
    xndq__gwjfx = cgutils.create_struct_proxy(zml__pygc)(context, builder,
        value=pvm__onh)
    ojh__yspur = lir.FunctionType(lir.IntType(1), [lir.IntType(8).as_pointer()]
        )
    burig__aft = cgutils.get_or_insert_function(builder.module, ojh__yspur,
        name='update_csv_reader')
    kyu__xyam = builder.call(burig__aft, [xndq__gwjfx.csv_reader])
    result.set_valid(kyu__xyam)
    with builder.if_then(kyu__xyam):
        xsh__hpv = builder.load(xndq__gwjfx.index)
        raof__tmtnw = types.Tuple([sig.return_type.first_type, types.int64])
        vxsm__uwj = gen_read_csv_objmode(sig.args[0])
        nxr__nkyme = signature(raof__tmtnw, bodo.ir.connector.
            stream_reader_type, types.int64)
        sdase__sitak = context.compile_internal(builder, vxsm__uwj,
            nxr__nkyme, [xndq__gwjfx.csv_reader, xsh__hpv])
        myxto__sesqk, srwd__hef = cgutils.unpack_tuple(builder, sdase__sitak)
        piupa__dyhyo = builder.add(xsh__hpv, srwd__hef, flags=['nsw'])
        builder.store(piupa__dyhyo, xndq__gwjfx.index)
        result.yield_(myxto__sesqk)


@intrinsic
def init_csv_iterator(typingctx, csv_reader, csv_iterator_typeref):

    def codegen(context, builder, signature, args):
        hzde__ywecd = cgutils.create_struct_proxy(signature.return_type)(
            context, builder)
        hzde__ywecd.csv_reader = args[0]
        yctks__uwly = context.get_constant(types.uintp, 0)
        hzde__ywecd.index = cgutils.alloca_once_value(builder, yctks__uwly)
        return hzde__ywecd._getvalue()
    assert isinstance(csv_iterator_typeref, types.TypeRef
        ), 'Initializing a csv iterator requires a typeref'
    jkrk__gcdsr = csv_iterator_typeref.instance_type
    sig = signature(jkrk__gcdsr, csv_reader, csv_iterator_typeref)
    return sig, codegen


def gen_read_csv_objmode(csv_iterator_type):
    eifc__pxep = 'def read_csv_objmode(f_reader):\n'
    czt__yagsm = [sanitize_varname(esrqx__nzcd) for esrqx__nzcd in
        csv_iterator_type._out_colnames]
    aar__hoftf = ir_utils.next_label()
    ziflt__nzl = globals()
    out_types = csv_iterator_type._out_types
    ziflt__nzl[f'table_type_{aar__hoftf}'] = TableType(tuple(out_types))
    ziflt__nzl[f'idx_array_typ'] = csv_iterator_type._index_arr_typ
    nqjwd__qlna = list(range(len(csv_iterator_type._usecols)))
    eifc__pxep += _gen_read_csv_objmode(csv_iterator_type._out_colnames,
        czt__yagsm, out_types, csv_iterator_type._usecols, nqjwd__qlna,
        csv_iterator_type._sep, csv_iterator_type._escapechar, aar__hoftf,
        ziflt__nzl, parallel=False, check_parallel_runtime=True,
        idx_col_index=csv_iterator_type._index_ind, idx_col_typ=
        csv_iterator_type._index_arr_typ)
    pey__uyc = bodo.ir.csv_ext._gen_parallel_flag_name(czt__yagsm)
    vscjk__epl = ['T'] + (['idx_arr'] if csv_iterator_type._index_ind is not
        None else []) + [pey__uyc]
    eifc__pxep += f"  return {', '.join(vscjk__epl)}"
    ziflt__nzl = globals()
    rap__aui = {}
    exec(eifc__pxep, ziflt__nzl, rap__aui)
    ivl__rbn = rap__aui['read_csv_objmode']
    aolol__orlt = numba.njit(ivl__rbn)
    bodo.ir.csv_ext.compiled_funcs.append(aolol__orlt)
    tube__hzw = 'def read_func(reader, local_start):\n'
    tube__hzw += f"  {', '.join(vscjk__epl)} = objmode_func(reader)\n"
    index_ind = csv_iterator_type._index_ind
    if index_ind is None:
        tube__hzw += f'  local_len = len(T)\n'
        tube__hzw += '  total_size = local_len\n'
        tube__hzw += f'  if ({pey__uyc}):\n'
        tube__hzw += """    local_start = local_start + bodo.libs.distributed_api.dist_exscan(local_len, _op)
"""
        tube__hzw += (
            '    total_size = bodo.libs.distributed_api.dist_reduce(local_len, _op)\n'
            )
        bdis__olz = (
            f'bodo.hiframes.pd_index_ext.init_range_index(local_start, local_start + local_len, 1, None)'
            )
    else:
        tube__hzw += '  total_size = 0\n'
        bdis__olz = (
            f'bodo.utils.conversion.convert_to_index({vscjk__epl[1]}, {csv_iterator_type._index_name!r})'
            )
    tube__hzw += f"""  return (bodo.hiframes.pd_dataframe_ext.init_dataframe(({vscjk__epl[0]},), {bdis__olz}, out_df_typ), total_size)
"""
    exec(tube__hzw, {'bodo': bodo, 'objmode_func': aolol__orlt, '_op': np.
        int32(bodo.libs.distributed_api.Reduce_Type.Sum.value),
        'out_df_typ': csv_iterator_type.yield_type}, rap__aui)
    return rap__aui['read_func']
