"""
S3 & Hadoop file system supports, and file system dependent calls
"""
import glob
import os
import warnings
from urllib.parse import urlparse
import llvmlite.binding as ll
import numba
import numpy as np
from numba.core import types
from numba.extending import overload
import bodo
from bodo.io import csv_cpp
from bodo.libs.distributed_api import Reduce_Type
from bodo.libs.str_ext import unicode_to_utf8, unicode_to_utf8_and_len
from bodo.utils.typing import BodoError, BodoWarning
from bodo.utils.utils import check_java_installation
from fsspec.implementations.arrow import ArrowFSWrapper, ArrowFile, wrap_exceptions


def fsspec_arrowfswrapper__open(self, path, mode='rb', block_size=None, **
    kwargs):
    if mode == 'rb':
        try:
            trjn__ejlw = self.fs.open_input_file(path)
        except:
            trjn__ejlw = self.fs.open_input_stream(path)
    elif mode == 'wb':
        trjn__ejlw = self.fs.open_output_stream(path)
    else:
        raise ValueError(f'unsupported mode for Arrow filesystem: {mode!r}')
    return ArrowFile(self, trjn__ejlw, path, mode, block_size, **kwargs)


ArrowFSWrapper._open = wrap_exceptions(fsspec_arrowfswrapper__open)
_csv_write = types.ExternalFunction('csv_write', types.void(types.voidptr,
    types.voidptr, types.int64, types.int64, types.bool_, types.voidptr))
ll.add_symbol('csv_write', csv_cpp.csv_write)
bodo_error_msg = """
    Some possible causes:
        (1) Incorrect path: Specified file/directory doesn't exist or is unreachable.
        (2) Missing credentials: You haven't provided S3 credentials, neither through 
            environment variables, nor through a local AWS setup 
            that makes the credentials available at ~/.aws/credentials.
        (3) Incorrect credentials: Your S3 credentials are incorrect or do not have
            the correct permissions.
    """


def get_proxy_uri_from_env_vars():
    return os.environ.get('http_proxy', None) or os.environ.get('https_proxy',
        None) or os.environ.get('HTTP_PROXY', None) or os.environ.get(
        'HTTPS_PROXY', None)


def get_s3_fs(region=None, storage_options=None):
    from pyarrow.fs import S3FileSystem
    jbb__qpeou = os.environ.get('AWS_S3_ENDPOINT', None)
    if not region:
        region = os.environ.get('AWS_DEFAULT_REGION', None)
    nyyo__styx = False
    nvvtc__cpoe = get_proxy_uri_from_env_vars()
    if storage_options:
        nyyo__styx = storage_options.get('anon', False)
    return S3FileSystem(anonymous=nyyo__styx, region=region,
        endpoint_override=jbb__qpeou, proxy_options=nvvtc__cpoe)


def get_s3_subtree_fs(bucket_name, region=None, storage_options=None):
    from pyarrow._fs import SubTreeFileSystem
    from pyarrow._s3fs import S3FileSystem
    jbb__qpeou = os.environ.get('AWS_S3_ENDPOINT', None)
    if not region:
        region = os.environ.get('AWS_DEFAULT_REGION', None)
    nyyo__styx = False
    nvvtc__cpoe = get_proxy_uri_from_env_vars()
    if storage_options:
        nyyo__styx = storage_options.get('anon', False)
    fs = S3FileSystem(region=region, endpoint_override=jbb__qpeou,
        anonymous=nyyo__styx, proxy_options=nvvtc__cpoe)
    return SubTreeFileSystem(bucket_name, fs)


def get_s3_fs_from_path(path, parallel=False, storage_options=None):
    region = get_s3_bucket_region_njit(path, parallel=parallel)
    if region == '':
        region = None
    return get_s3_fs(region, storage_options)


def get_hdfs_fs(path):
    from pyarrow.fs import HadoopFileSystem as HdFS
    bzwo__hhxz = urlparse(path)
    if bzwo__hhxz.scheme in ('abfs', 'abfss'):
        uyzb__splh = path
        if bzwo__hhxz.port is None:
            wjvcm__umo = 0
        else:
            wjvcm__umo = bzwo__hhxz.port
        tjk__efpe = None
    else:
        uyzb__splh = bzwo__hhxz.hostname
        wjvcm__umo = bzwo__hhxz.port
        tjk__efpe = bzwo__hhxz.username
    try:
        fs = HdFS(host=uyzb__splh, port=wjvcm__umo, user=tjk__efpe)
    except Exception as birc__snta:
        raise BodoError('Hadoop file system cannot be created: {}'.format(
            birc__snta))
    return fs


def gcs_is_directory(path):
    import gcsfs
    fs = gcsfs.GCSFileSystem(token=None)
    try:
        iqeyl__mkdm = fs.isdir(path)
    except gcsfs.utils.HttpError as birc__snta:
        raise BodoError(
            f'{birc__snta}. Make sure your google cloud credentials are set!')
    return iqeyl__mkdm


def gcs_list_dir_fnames(path):
    import gcsfs
    fs = gcsfs.GCSFileSystem(token=None)
    return [owq__dox.split('/')[-1] for owq__dox in fs.ls(path)]


def s3_is_directory(fs, path):
    from pyarrow import fs as pa_fs
    try:
        bzwo__hhxz = urlparse(path)
        ubd__ridv = (bzwo__hhxz.netloc + bzwo__hhxz.path).rstrip('/')
        hrmfd__fsgu = fs.get_file_info(ubd__ridv)
        if hrmfd__fsgu.type in (pa_fs.FileType.NotFound, pa_fs.FileType.Unknown
            ):
            raise FileNotFoundError('{} is a non-existing or unreachable file'
                .format(path))
        if (not hrmfd__fsgu.size and hrmfd__fsgu.type == pa_fs.FileType.
            Directory):
            return True
        return False
    except (FileNotFoundError, OSError) as birc__snta:
        raise
    except BodoError as ltgk__ayv:
        raise
    except Exception as birc__snta:
        raise BodoError(
            f"""error from pyarrow S3FileSystem: {type(birc__snta).__name__}: {str(birc__snta)}
{bodo_error_msg}"""
            )


def s3_list_dir_fnames(fs, path):
    from pyarrow import fs as pa_fs
    zgr__uutg = None
    try:
        if s3_is_directory(fs, path):
            bzwo__hhxz = urlparse(path)
            ubd__ridv = (bzwo__hhxz.netloc + bzwo__hhxz.path).rstrip('/')
            twli__bpksy = pa_fs.FileSelector(ubd__ridv, recursive=False)
            zkgx__mavgu = fs.get_file_info(twli__bpksy)
            if zkgx__mavgu and zkgx__mavgu[0].path in [ubd__ridv,
                f'{ubd__ridv}/'] and int(zkgx__mavgu[0].size or 0) == 0:
                zkgx__mavgu = zkgx__mavgu[1:]
            zgr__uutg = [tbayq__akce.base_name for tbayq__akce in zkgx__mavgu]
    except BodoError as ltgk__ayv:
        raise
    except Exception as birc__snta:
        raise BodoError(
            f"""error from pyarrow S3FileSystem: {type(birc__snta).__name__}: {str(birc__snta)}
{bodo_error_msg}"""
            )
    return zgr__uutg


def hdfs_is_directory(path):
    from pyarrow.fs import FileType, HadoopFileSystem
    check_java_installation(path)
    bzwo__hhxz = urlparse(path)
    fevc__quutu = bzwo__hhxz.path
    try:
        pfqs__flxof = HadoopFileSystem.from_uri(path)
    except Exception as birc__snta:
        raise BodoError(' Hadoop file system cannot be created: {}'.format(
            birc__snta))
    bghhc__iaim = pfqs__flxof.get_file_info([fevc__quutu])
    if bghhc__iaim[0].type in (FileType.NotFound, FileType.Unknown):
        raise BodoError('{} is a non-existing or unreachable file'.format(path)
            )
    if not bghhc__iaim[0].size and bghhc__iaim[0].type == FileType.Directory:
        return pfqs__flxof, True
    return pfqs__flxof, False


def hdfs_list_dir_fnames(path):
    from pyarrow.fs import FileSelector
    zgr__uutg = None
    pfqs__flxof, iqeyl__mkdm = hdfs_is_directory(path)
    if iqeyl__mkdm:
        bzwo__hhxz = urlparse(path)
        fevc__quutu = bzwo__hhxz.path
        twli__bpksy = FileSelector(fevc__quutu, recursive=True)
        try:
            zkgx__mavgu = pfqs__flxof.get_file_info(twli__bpksy)
        except Exception as birc__snta:
            raise BodoError('Exception on getting directory info of {}: {}'
                .format(fevc__quutu, birc__snta))
        zgr__uutg = [tbayq__akce.base_name for tbayq__akce in zkgx__mavgu]
    return pfqs__flxof, zgr__uutg


def abfs_is_directory(path):
    pfqs__flxof = get_hdfs_fs(path)
    try:
        bghhc__iaim = pfqs__flxof.info(path)
    except OSError as ltgk__ayv:
        raise BodoError('{} is a non-existing or unreachable file'.format(path)
            )
    if bghhc__iaim['size'] == 0 and bghhc__iaim['kind'].lower() == 'directory':
        return pfqs__flxof, True
    return pfqs__flxof, False


def abfs_list_dir_fnames(path):
    zgr__uutg = None
    pfqs__flxof, iqeyl__mkdm = abfs_is_directory(path)
    if iqeyl__mkdm:
        bzwo__hhxz = urlparse(path)
        fevc__quutu = bzwo__hhxz.path
        try:
            zlzh__hai = pfqs__flxof.ls(fevc__quutu)
        except Exception as birc__snta:
            raise BodoError('Exception on getting directory info of {}: {}'
                .format(fevc__quutu, birc__snta))
        zgr__uutg = [fname[fname.rindex('/') + 1:] for fname in zlzh__hai]
    return pfqs__flxof, zgr__uutg


def directory_of_files_common_filter(fname):
    return not (fname.endswith('.crc') or fname.endswith('_$folder$') or
        fname.startswith('.') or fname.startswith('_') and fname !=
        '_delta_log')


def find_file_name_or_handler(path, ftype):
    from urllib.parse import urlparse
    odvn__etjq = urlparse(path)
    fname = path
    fs = None
    icyap__egi = 'read_json' if ftype == 'json' else 'read_csv'
    wfvj__xnle = (
        f'pd.{icyap__egi}(): there is no {ftype} file in directory: {fname}')
    ytobz__vmoqp = directory_of_files_common_filter
    if odvn__etjq.scheme == 's3':
        wml__gyz = True
        fs = get_s3_fs_from_path(path)
        bmhtq__qetu = s3_list_dir_fnames(fs, path)
        ubd__ridv = (odvn__etjq.netloc + odvn__etjq.path).rstrip('/')
        fname = ubd__ridv
        if bmhtq__qetu:
            bmhtq__qetu = [(ubd__ridv + '/' + owq__dox) for owq__dox in
                sorted(filter(ytobz__vmoqp, bmhtq__qetu))]
            lblp__rccg = [owq__dox for owq__dox in bmhtq__qetu if int(fs.
                get_file_info(owq__dox).size or 0) > 0]
            if len(lblp__rccg) == 0:
                raise BodoError(wfvj__xnle)
            fname = lblp__rccg[0]
        dyj__why = int(fs.get_file_info(fname).size or 0)
        aylnz__bet = fs.open_input_file(fname)
    elif odvn__etjq.scheme == 'hdfs':
        wml__gyz = True
        fs, bmhtq__qetu = hdfs_list_dir_fnames(path)
        dyj__why = fs.get_file_info([odvn__etjq.path])[0].size
        if bmhtq__qetu:
            path = path.rstrip('/')
            bmhtq__qetu = [(path + '/' + owq__dox) for owq__dox in sorted(
                filter(ytobz__vmoqp, bmhtq__qetu))]
            lblp__rccg = [owq__dox for owq__dox in bmhtq__qetu if fs.
                get_file_info([urlparse(owq__dox).path])[0].size > 0]
            if len(lblp__rccg) == 0:
                raise BodoError(wfvj__xnle)
            fname = lblp__rccg[0]
            fname = urlparse(fname).path
            dyj__why = fs.get_file_info([fname])[0].size
        aylnz__bet = fs.open_input_file(fname)
    elif odvn__etjq.scheme in ('abfs', 'abfss'):
        wml__gyz = True
        fs, bmhtq__qetu = abfs_list_dir_fnames(path)
        dyj__why = fs.info(fname)['size']
        if bmhtq__qetu:
            path = path.rstrip('/')
            bmhtq__qetu = [(path + '/' + owq__dox) for owq__dox in sorted(
                filter(ytobz__vmoqp, bmhtq__qetu))]
            lblp__rccg = [owq__dox for owq__dox in bmhtq__qetu if fs.info(
                owq__dox)['size'] > 0]
            if len(lblp__rccg) == 0:
                raise BodoError(wfvj__xnle)
            fname = lblp__rccg[0]
            dyj__why = fs.info(fname)['size']
            fname = urlparse(fname).path
        aylnz__bet = fs.open(fname, 'rb')
    else:
        if odvn__etjq.scheme != '':
            raise BodoError(
                f'Unrecognized scheme {odvn__etjq.scheme}. Please refer to https://docs.bodo.ai/latest/source/file_io.html'
                )
        wml__gyz = False
        if os.path.isdir(path):
            zlzh__hai = filter(ytobz__vmoqp, glob.glob(os.path.join(os.path
                .abspath(path), '*')))
            lblp__rccg = [owq__dox for owq__dox in sorted(zlzh__hai) if os.
                path.getsize(owq__dox) > 0]
            if len(lblp__rccg) == 0:
                raise BodoError(wfvj__xnle)
            fname = lblp__rccg[0]
        dyj__why = os.path.getsize(fname)
        aylnz__bet = fname
    return wml__gyz, aylnz__bet, dyj__why, fs


def get_s3_bucket_region(s3_filepath, parallel):
    try:
        from pyarrow import fs as pa_fs
    except:
        raise BodoError('Reading from s3 requires pyarrow currently.')
    from mpi4py import MPI
    giub__ywv = MPI.COMM_WORLD
    bucket_loc = None
    if parallel and bodo.get_rank() == 0 or not parallel:
        try:
            ayp__xdg, kxp__wvc = pa_fs.S3FileSystem.from_uri(s3_filepath)
            bucket_loc = ayp__xdg.region
        except Exception as birc__snta:
            if os.environ.get('AWS_DEFAULT_REGION', '') == '':
                warnings.warn(BodoWarning(
                    f"""Unable to get S3 Bucket Region.
{birc__snta}.
Value not defined in the AWS_DEFAULT_REGION environment variable either. Region defaults to us-east-1 currently."""
                    ))
            bucket_loc = ''
    if parallel:
        bucket_loc = giub__ywv.bcast(bucket_loc)
    return bucket_loc


@numba.njit()
def get_s3_bucket_region_njit(s3_filepath, parallel):
    with numba.objmode(bucket_loc='unicode_type'):
        bucket_loc = ''
        if isinstance(s3_filepath, list):
            s3_filepath = s3_filepath[0]
        if s3_filepath.startswith('s3://'):
            bucket_loc = get_s3_bucket_region(s3_filepath, parallel)
    return bucket_loc


def csv_write(path_or_buf, D, is_parallel=False):
    return None


@overload(csv_write, no_unliteral=True)
def csv_write_overload(path_or_buf, D, is_parallel=False):

    def impl(path_or_buf, D, is_parallel=False):
        gmk__dtig = get_s3_bucket_region_njit(path_or_buf, parallel=is_parallel
            )
        qwn__btan, yfky__wat = unicode_to_utf8_and_len(D)
        ddpky__bfq = 0
        if is_parallel:
            ddpky__bfq = bodo.libs.distributed_api.dist_exscan(yfky__wat,
                np.int32(Reduce_Type.Sum.value))
        _csv_write(unicode_to_utf8(path_or_buf), qwn__btan, ddpky__bfq,
            yfky__wat, is_parallel, unicode_to_utf8(gmk__dtig))
        bodo.utils.utils.check_and_propagate_cpp_exception()
    return impl
