"""IR node for the groupby, pivot and cross_tabulation"""
import ctypes
import operator
import types as pytypes
from collections import defaultdict, namedtuple
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, compiler, ir, ir_utils, types
from numba.core.analysis import compute_use_defs
from numba.core.ir_utils import build_definitions, compile_to_numba_ir, find_callname, find_const, find_topo_order, get_definition, get_ir_of_code, get_name_var_table, guard, is_getitem, mk_unique_var, next_label, remove_dels, replace_arg_nodes, replace_var_names, replace_vars_inner, visit_vars_inner
from numba.core.typing import signature
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import intrinsic, lower_builtin, overload
from numba.parfors.parfor import Parfor, unwrap_parfor_blocks, wrap_parfor_blocks
import bodo
from bodo.hiframes.datetime_date_ext import DatetimeDateArrayType
from bodo.hiframes.pd_series_ext import SeriesType
from bodo.libs.array import arr_info_list_to_table, array_to_info, compute_node_partition_by_hash, delete_info_decref_array, delete_table, delete_table_decref_arrays, groupby_and_aggregate, info_from_table, info_to_array, pivot_groupby_and_aggregate
from bodo.libs.array_item_arr_ext import ArrayItemArrayType, pre_alloc_array_item_array
from bodo.libs.binary_arr_ext import BinaryArrayType, pre_alloc_binary_array
from bodo.libs.bool_arr_ext import BooleanArrayType
from bodo.libs.decimal_arr_ext import DecimalArrayType, alloc_decimal_array
from bodo.libs.int_arr_ext import IntDtype, IntegerArrayType
from bodo.libs.str_arr_ext import StringArrayType, pre_alloc_string_array, string_array_type
from bodo.libs.str_ext import string_type
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.transforms.distributed_analysis import Distribution
from bodo.utils.transform import get_call_expr_arg
from bodo.utils.typing import BodoError, get_literal_value, get_overload_const_func, get_overload_const_str, get_overload_constant_dict, is_overload_constant_dict, is_overload_constant_str, list_cumulative
from bodo.utils.utils import debug_prints, incref, is_assign, is_call_assign, is_expr, is_null_pointer, is_var_assign, sanitize_varname, unliteral_all
gb_agg_cfunc = {}
gb_agg_cfunc_addr = {}


@intrinsic
def add_agg_cfunc_sym(typingctx, func, sym):

    def codegen(context, builder, signature, args):
        sig = func.signature
        if sig == types.none(types.voidptr):
            tlin__ffvul = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer()])
            wtjf__jhkn = cgutils.get_or_insert_function(builder.module,
                tlin__ffvul, sym._literal_value)
            builder.call(wtjf__jhkn, [context.get_constant_null(sig.args[0])])
        elif sig == types.none(types.int64, types.voidptr, types.voidptr):
            tlin__ffvul = lir.FunctionType(lir.VoidType(), [lir.IntType(64),
                lir.IntType(8).as_pointer(), lir.IntType(8).as_pointer()])
            wtjf__jhkn = cgutils.get_or_insert_function(builder.module,
                tlin__ffvul, sym._literal_value)
            builder.call(wtjf__jhkn, [context.get_constant(types.int64, 0),
                context.get_constant_null(sig.args[1]), context.
                get_constant_null(sig.args[2])])
        else:
            tlin__ffvul = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(64).
                as_pointer()])
            wtjf__jhkn = cgutils.get_or_insert_function(builder.module,
                tlin__ffvul, sym._literal_value)
            builder.call(wtjf__jhkn, [context.get_constant_null(sig.args[0]
                ), context.get_constant_null(sig.args[1]), context.
                get_constant_null(sig.args[2])])
        context.add_linking_libs([gb_agg_cfunc[sym._literal_value]._library])
        return
    return types.none(func, sym), codegen


@numba.jit
def get_agg_udf_addr(name):
    with numba.objmode(addr='int64'):
        addr = gb_agg_cfunc_addr[name]
    return addr


class AggUDFStruct(object):

    def __init__(self, regular_udf_funcs=None, general_udf_funcs=None):
        assert regular_udf_funcs is not None or general_udf_funcs is not None
        self.regular_udfs = False
        self.general_udfs = False
        self.regular_udf_cfuncs = None
        self.general_udf_cfunc = None
        if regular_udf_funcs is not None:
            (self.var_typs, self.init_func, self.update_all_func, self.
                combine_all_func, self.eval_all_func) = regular_udf_funcs
            self.regular_udfs = True
        if general_udf_funcs is not None:
            self.general_udf_funcs = general_udf_funcs
            self.general_udfs = True

    def set_regular_cfuncs(self, update_cb, combine_cb, eval_cb):
        assert self.regular_udfs and self.regular_udf_cfuncs is None
        self.regular_udf_cfuncs = [update_cb, combine_cb, eval_cb]

    def set_general_cfunc(self, general_udf_cb):
        assert self.general_udfs and self.general_udf_cfunc is None
        self.general_udf_cfunc = general_udf_cb


AggFuncStruct = namedtuple('AggFuncStruct', ['func', 'ftype'])
supported_agg_funcs = ['no_op', 'head', 'transform', 'size', 'shift', 'sum',
    'count', 'nunique', 'median', 'cumsum', 'cumprod', 'cummin', 'cummax',
    'mean', 'min', 'max', 'prod', 'first', 'last', 'idxmin', 'idxmax',
    'var', 'std', 'udf', 'gen_udf']
supported_transform_funcs = ['no_op', 'sum', 'count', 'nunique', 'median',
    'mean', 'min', 'max', 'prod', 'first', 'last', 'var', 'std']


def get_agg_func(func_ir, func_name, rhs, series_type=None, typemap=None):
    if func_name == 'no_op':
        raise BodoError('Unknown aggregation function used in groupby.')
    if series_type is None:
        series_type = SeriesType(types.float64)
    if func_name in {'var', 'std'}:
        func = pytypes.SimpleNamespace()
        func.ftype = func_name
        func.fname = func_name
        func.ncols_pre_shuffle = 3
        func.ncols_post_shuffle = 4
        return func
    if func_name in {'first', 'last'}:
        func = pytypes.SimpleNamespace()
        func.ftype = func_name
        func.fname = func_name
        func.ncols_pre_shuffle = 1
        func.ncols_post_shuffle = 1
        return func
    if func_name in {'idxmin', 'idxmax'}:
        func = pytypes.SimpleNamespace()
        func.ftype = func_name
        func.fname = func_name
        func.ncols_pre_shuffle = 2
        func.ncols_post_shuffle = 2
        return func
    if func_name in supported_agg_funcs[:-8]:
        func = pytypes.SimpleNamespace()
        func.ftype = func_name
        func.fname = func_name
        func.ncols_pre_shuffle = 1
        func.ncols_post_shuffle = 1
        wofs__wlnft = True
        ipzf__tekke = 1
        lxthx__yht = -1
        if isinstance(rhs, ir.Expr):
            for wjuvx__cbz in rhs.kws:
                if func_name in list_cumulative:
                    if wjuvx__cbz[0] == 'skipna':
                        wofs__wlnft = guard(find_const, func_ir, wjuvx__cbz[1])
                        if not isinstance(wofs__wlnft, bool):
                            raise BodoError(
                                'For {} argument of skipna should be a boolean'
                                .format(func_name))
                if func_name == 'nunique':
                    if wjuvx__cbz[0] == 'dropna':
                        wofs__wlnft = guard(find_const, func_ir, wjuvx__cbz[1])
                        if not isinstance(wofs__wlnft, bool):
                            raise BodoError(
                                'argument of dropna to nunique should be a boolean'
                                )
        if func_name == 'shift' and (len(rhs.args) > 0 or len(rhs.kws) > 0):
            ipzf__tekke = get_call_expr_arg('shift', rhs.args, dict(rhs.kws
                ), 0, 'periods', ipzf__tekke)
            ipzf__tekke = guard(find_const, func_ir, ipzf__tekke)
        if func_name == 'head':
            lxthx__yht = get_call_expr_arg('head', rhs.args, dict(rhs.kws),
                0, 'n', 5)
            if not isinstance(lxthx__yht, int):
                lxthx__yht = guard(find_const, func_ir, lxthx__yht)
            if lxthx__yht < 0:
                raise BodoError(
                    f'groupby.{func_name} does not work with negative values.')
        func.skipdropna = wofs__wlnft
        func.periods = ipzf__tekke
        func.head_n = lxthx__yht
        if func_name == 'transform':
            kws = dict(rhs.kws)
            bswvt__vjy = get_call_expr_arg(func_name, rhs.args, kws, 0,
                'func', '')
            zwhwd__zpsx = typemap[bswvt__vjy.name]
            itybq__zpj = None
            if isinstance(zwhwd__zpsx, str):
                itybq__zpj = zwhwd__zpsx
            elif is_overload_constant_str(zwhwd__zpsx):
                itybq__zpj = get_overload_const_str(zwhwd__zpsx)
            elif bodo.utils.typing.is_builtin_function(zwhwd__zpsx):
                itybq__zpj = bodo.utils.typing.get_builtin_function_name(
                    zwhwd__zpsx)
            if itybq__zpj not in bodo.ir.aggregate.supported_transform_funcs[:
                ]:
                raise BodoError(f'unsupported transform function {itybq__zpj}')
            func.transform_func = supported_agg_funcs.index(itybq__zpj)
        else:
            func.transform_func = supported_agg_funcs.index('no_op')
        return func
    assert func_name in ['agg', 'aggregate']
    assert typemap is not None
    kws = dict(rhs.kws)
    bswvt__vjy = get_call_expr_arg(func_name, rhs.args, kws, 0, 'func', '')
    if bswvt__vjy == '':
        zwhwd__zpsx = types.none
    else:
        zwhwd__zpsx = typemap[bswvt__vjy.name]
    if is_overload_constant_dict(zwhwd__zpsx):
        ruzpz__bou = get_overload_constant_dict(zwhwd__zpsx)
        fxql__fdc = [get_agg_func_udf(func_ir, f_val, rhs, series_type,
            typemap) for f_val in ruzpz__bou.values()]
        return fxql__fdc
    if zwhwd__zpsx == types.none:
        return [get_agg_func_udf(func_ir, get_literal_value(typemap[f_val.
            name])[1], rhs, series_type, typemap) for f_val in kws.values()]
    if isinstance(zwhwd__zpsx, types.BaseTuple):
        fxql__fdc = []
        hjjk__chh = 0
        for t in zwhwd__zpsx.types:
            if is_overload_constant_str(t):
                func_name = get_overload_const_str(t)
                fxql__fdc.append(get_agg_func(func_ir, func_name, rhs,
                    series_type, typemap))
            else:
                assert typemap is not None, 'typemap is required for agg UDF handling'
                func = _get_const_agg_func(t, func_ir)
                func.ftype = 'udf'
                func.fname = _get_udf_name(func)
                if func.fname == '<lambda>':
                    func.fname = '<lambda_' + str(hjjk__chh) + '>'
                    hjjk__chh += 1
                fxql__fdc.append(func)
        return [fxql__fdc]
    if is_overload_constant_str(zwhwd__zpsx):
        func_name = get_overload_const_str(zwhwd__zpsx)
        return get_agg_func(func_ir, func_name, rhs, series_type, typemap)
    if bodo.utils.typing.is_builtin_function(zwhwd__zpsx):
        func_name = bodo.utils.typing.get_builtin_function_name(zwhwd__zpsx)
        return get_agg_func(func_ir, func_name, rhs, series_type, typemap)
    assert typemap is not None, 'typemap is required for agg UDF handling'
    func = _get_const_agg_func(typemap[rhs.args[0].name], func_ir)
    func.ftype = 'udf'
    func.fname = _get_udf_name(func)
    return func


def get_agg_func_udf(func_ir, f_val, rhs, series_type, typemap):
    if isinstance(f_val, str):
        return get_agg_func(func_ir, f_val, rhs, series_type, typemap)
    if bodo.utils.typing.is_builtin_function(f_val):
        func_name = bodo.utils.typing.get_builtin_function_name(f_val)
        return get_agg_func(func_ir, func_name, rhs, series_type, typemap)
    if isinstance(f_val, (tuple, list)):
        hjjk__chh = 0
        tmk__czqp = []
        for aaq__qpzs in f_val:
            func = get_agg_func_udf(func_ir, aaq__qpzs, rhs, series_type,
                typemap)
            if func.fname == '<lambda>' and len(f_val) > 1:
                func.fname = f'<lambda_{hjjk__chh}>'
                hjjk__chh += 1
            tmk__czqp.append(func)
        return tmk__czqp
    else:
        assert is_expr(f_val, 'make_function') or isinstance(f_val, (numba.
            core.registry.CPUDispatcher, types.Dispatcher))
        assert typemap is not None, 'typemap is required for agg UDF handling'
        func = _get_const_agg_func(f_val, func_ir)
        func.ftype = 'udf'
        func.fname = _get_udf_name(func)
        return func


def _get_udf_name(func):
    code = func.code if hasattr(func, 'code') else func.__code__
    itybq__zpj = code.co_name
    return itybq__zpj


def _get_const_agg_func(func_typ, func_ir):
    agg_func = get_overload_const_func(func_typ, func_ir)
    if is_expr(agg_func, 'make_function'):

        def agg_func_wrapper(A):
            return A
        agg_func_wrapper.__code__ = agg_func.code
        agg_func = agg_func_wrapper
        return agg_func
    return agg_func


@infer_global(type)
class TypeDt64(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        if len(args) == 1 and isinstance(args[0], (types.NPDatetime, types.
            NPTimedelta)):
            woult__xdwm = types.DType(args[0])
            return signature(woult__xdwm, *args)


@numba.njit(no_cpython_wrapper=True)
def _var_combine(ssqdm_a, mean_a, nobs_a, ssqdm_b, mean_b, nobs_b):
    rlinl__wpsf = nobs_a + nobs_b
    tixw__ktbzi = (nobs_a * mean_a + nobs_b * mean_b) / rlinl__wpsf
    xhhyc__dei = mean_b - mean_a
    tiy__sqh = (ssqdm_a + ssqdm_b + xhhyc__dei * xhhyc__dei * nobs_a *
        nobs_b / rlinl__wpsf)
    return tiy__sqh, tixw__ktbzi, rlinl__wpsf


def __special_combine(*args):
    return


@infer_global(__special_combine)
class SpecialCombineTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(types.void, *unliteral_all(args))


@lower_builtin(__special_combine, types.VarArg(types.Any))
def lower_special_combine(context, builder, sig, args):
    return context.get_dummy_value()


class Aggregate(ir.Stmt):

    def __init__(self, df_out, df_in, key_names, gb_info_in, gb_info_out,
        out_key_vars, df_out_vars, df_in_vars, key_arrs, input_has_index,
        same_index, return_key, loc, func_name, dropna=True, pivot_arr=None,
        pivot_values=None, is_crosstab=False):
        self.df_out = df_out
        self.df_in = df_in
        self.key_names = key_names
        self.gb_info_in = gb_info_in
        self.gb_info_out = gb_info_out
        self.out_key_vars = out_key_vars
        self.df_out_vars = df_out_vars
        self.df_in_vars = df_in_vars
        self.key_arrs = key_arrs
        self.input_has_index = input_has_index
        self.same_index = same_index
        self.return_key = return_key
        self.loc = loc
        self.func_name = func_name
        self.dropna = dropna
        self.pivot_arr = pivot_arr
        self.pivot_values = pivot_values
        self.is_crosstab = is_crosstab

    def __repr__(self):
        bzfyl__vbii = ''
        for vat__plxf, v in self.df_out_vars.items():
            bzfyl__vbii += "'{}':{}, ".format(vat__plxf, v.name)
        rvqy__jcpqx = '{}{{{}}}'.format(self.df_out, bzfyl__vbii)
        nypp__tayc = ''
        for vat__plxf, v in self.df_in_vars.items():
            nypp__tayc += "'{}':{}, ".format(vat__plxf, v.name)
        gzqpe__vey = '{}{{{}}}'.format(self.df_in, nypp__tayc)
        zbq__gexon = 'pivot {}:{}'.format(self.pivot_arr.name, self.
            pivot_values) if self.pivot_arr is not None else ''
        key_names = ','.join(self.key_names)
        wisny__nrziq = ','.join([v.name for v in self.key_arrs])
        return 'aggregate: {} = {} [key: {}:{}] {}'.format(rvqy__jcpqx,
            gzqpe__vey, key_names, wisny__nrziq, zbq__gexon)

    def remove_out_col(self, out_col_name):
        self.df_out_vars.pop(out_col_name)
        ijwn__qvl, getir__teh = self.gb_info_out.pop(out_col_name)
        if ijwn__qvl is None and not self.is_crosstab:
            return
        dpsxj__drz = self.gb_info_in[ijwn__qvl]
        if self.pivot_arr is not None:
            self.pivot_values.remove(out_col_name)
            for i, (func, bzfyl__vbii) in enumerate(dpsxj__drz):
                try:
                    bzfyl__vbii.remove(out_col_name)
                    if len(bzfyl__vbii) == 0:
                        dpsxj__drz.pop(i)
                        break
                except ValueError as nmt__polh:
                    continue
        else:
            for i, (func, yku__icj) in enumerate(dpsxj__drz):
                if yku__icj == out_col_name:
                    dpsxj__drz.pop(i)
                    break
        if len(dpsxj__drz) == 0:
            self.gb_info_in.pop(ijwn__qvl)
            self.df_in_vars.pop(ijwn__qvl)


def aggregate_usedefs(aggregate_node, use_set=None, def_set=None):
    if use_set is None:
        use_set = set()
    if def_set is None:
        def_set = set()
    use_set.update({v.name for v in aggregate_node.key_arrs})
    use_set.update({v.name for v in aggregate_node.df_in_vars.values()})
    if aggregate_node.pivot_arr is not None:
        use_set.add(aggregate_node.pivot_arr.name)
    def_set.update({v.name for v in aggregate_node.df_out_vars.values()})
    if aggregate_node.out_key_vars is not None:
        def_set.update({v.name for v in aggregate_node.out_key_vars})
    return numba.core.analysis._use_defs_result(usemap=use_set, defmap=def_set)


numba.core.analysis.ir_extension_usedefs[Aggregate] = aggregate_usedefs


def remove_dead_aggregate(aggregate_node, lives_no_aliases, lives,
    arg_aliases, alias_map, func_ir, typemap):
    plep__ars = [qjv__xoaj for qjv__xoaj, zrssf__enqsd in aggregate_node.
        df_out_vars.items() if zrssf__enqsd.name not in lives]
    for zxpdz__xzr in plep__ars:
        aggregate_node.remove_out_col(zxpdz__xzr)
    out_key_vars = aggregate_node.out_key_vars
    if out_key_vars is not None and all(v.name not in lives for v in
        out_key_vars):
        aggregate_node.out_key_vars = None
    if len(aggregate_node.df_out_vars
        ) == 0 and aggregate_node.out_key_vars is None:
        return None
    return aggregate_node


ir_utils.remove_dead_extensions[Aggregate] = remove_dead_aggregate


def get_copies_aggregate(aggregate_node, typemap):
    upmk__phivr = set(v.name for v in aggregate_node.df_out_vars.values())
    if aggregate_node.out_key_vars is not None:
        upmk__phivr.update({v.name for v in aggregate_node.out_key_vars})
    return set(), upmk__phivr


ir_utils.copy_propagate_extensions[Aggregate] = get_copies_aggregate


def apply_copies_aggregate(aggregate_node, var_dict, name_var_table,
    typemap, calltypes, save_copies):
    for i in range(len(aggregate_node.key_arrs)):
        aggregate_node.key_arrs[i] = replace_vars_inner(aggregate_node.
            key_arrs[i], var_dict)
    for qjv__xoaj in list(aggregate_node.df_in_vars.keys()):
        aggregate_node.df_in_vars[qjv__xoaj] = replace_vars_inner(
            aggregate_node.df_in_vars[qjv__xoaj], var_dict)
    for qjv__xoaj in list(aggregate_node.df_out_vars.keys()):
        aggregate_node.df_out_vars[qjv__xoaj] = replace_vars_inner(
            aggregate_node.df_out_vars[qjv__xoaj], var_dict)
    if aggregate_node.out_key_vars is not None:
        for i in range(len(aggregate_node.out_key_vars)):
            aggregate_node.out_key_vars[i] = replace_vars_inner(aggregate_node
                .out_key_vars[i], var_dict)
    if aggregate_node.pivot_arr is not None:
        aggregate_node.pivot_arr = replace_vars_inner(aggregate_node.
            pivot_arr, var_dict)


ir_utils.apply_copy_propagate_extensions[Aggregate] = apply_copies_aggregate


def visit_vars_aggregate(aggregate_node, callback, cbdata):
    if debug_prints():
        print('visiting aggregate vars for:', aggregate_node)
        print('cbdata: ', sorted(cbdata.items()))
    for i in range(len(aggregate_node.key_arrs)):
        aggregate_node.key_arrs[i] = visit_vars_inner(aggregate_node.
            key_arrs[i], callback, cbdata)
    for qjv__xoaj in list(aggregate_node.df_in_vars.keys()):
        aggregate_node.df_in_vars[qjv__xoaj] = visit_vars_inner(aggregate_node
            .df_in_vars[qjv__xoaj], callback, cbdata)
    for qjv__xoaj in list(aggregate_node.df_out_vars.keys()):
        aggregate_node.df_out_vars[qjv__xoaj] = visit_vars_inner(aggregate_node
            .df_out_vars[qjv__xoaj], callback, cbdata)
    if aggregate_node.out_key_vars is not None:
        for i in range(len(aggregate_node.out_key_vars)):
            aggregate_node.out_key_vars[i] = visit_vars_inner(aggregate_node
                .out_key_vars[i], callback, cbdata)
    if aggregate_node.pivot_arr is not None:
        aggregate_node.pivot_arr = visit_vars_inner(aggregate_node.
            pivot_arr, callback, cbdata)


ir_utils.visit_vars_extensions[Aggregate] = visit_vars_aggregate


def aggregate_array_analysis(aggregate_node, equiv_set, typemap, array_analysis
    ):
    assert len(aggregate_node.df_out_vars
        ) > 0 or aggregate_node.out_key_vars is not None or aggregate_node.is_crosstab, 'empty aggregate in array analysis'
    wid__dpin = []
    for uuuod__lsaph in aggregate_node.key_arrs:
        rzpbl__jektp = equiv_set.get_shape(uuuod__lsaph)
        if rzpbl__jektp:
            wid__dpin.append(rzpbl__jektp[0])
    if aggregate_node.pivot_arr is not None:
        rzpbl__jektp = equiv_set.get_shape(aggregate_node.pivot_arr)
        if rzpbl__jektp:
            wid__dpin.append(rzpbl__jektp[0])
    for zrssf__enqsd in aggregate_node.df_in_vars.values():
        rzpbl__jektp = equiv_set.get_shape(zrssf__enqsd)
        if rzpbl__jektp:
            wid__dpin.append(rzpbl__jektp[0])
    if len(wid__dpin) > 1:
        equiv_set.insert_equiv(*wid__dpin)
    wpsfi__uusxq = []
    wid__dpin = []
    uyrs__dvel = list(aggregate_node.df_out_vars.values())
    if aggregate_node.out_key_vars is not None:
        uyrs__dvel.extend(aggregate_node.out_key_vars)
    for zrssf__enqsd in uyrs__dvel:
        qfcg__xhnv = typemap[zrssf__enqsd.name]
        knhbi__msnk = array_analysis._gen_shape_call(equiv_set,
            zrssf__enqsd, qfcg__xhnv.ndim, None, wpsfi__uusxq)
        equiv_set.insert_equiv(zrssf__enqsd, knhbi__msnk)
        wid__dpin.append(knhbi__msnk[0])
        equiv_set.define(zrssf__enqsd, set())
    if len(wid__dpin) > 1:
        equiv_set.insert_equiv(*wid__dpin)
    return [], wpsfi__uusxq


numba.parfors.array_analysis.array_analysis_extensions[Aggregate
    ] = aggregate_array_analysis


def aggregate_distributed_analysis(aggregate_node, array_dists):
    zekwy__ghlr = Distribution.OneD
    for zrssf__enqsd in aggregate_node.df_in_vars.values():
        zekwy__ghlr = Distribution(min(zekwy__ghlr.value, array_dists[
            zrssf__enqsd.name].value))
    for uuuod__lsaph in aggregate_node.key_arrs:
        zekwy__ghlr = Distribution(min(zekwy__ghlr.value, array_dists[
            uuuod__lsaph.name].value))
    if aggregate_node.pivot_arr is not None:
        zekwy__ghlr = Distribution(min(zekwy__ghlr.value, array_dists[
            aggregate_node.pivot_arr.name].value))
        array_dists[aggregate_node.pivot_arr.name] = zekwy__ghlr
    for zrssf__enqsd in aggregate_node.df_in_vars.values():
        array_dists[zrssf__enqsd.name] = zekwy__ghlr
    for uuuod__lsaph in aggregate_node.key_arrs:
        array_dists[uuuod__lsaph.name] = zekwy__ghlr
    plt__top = Distribution.OneD_Var
    for zrssf__enqsd in aggregate_node.df_out_vars.values():
        if zrssf__enqsd.name in array_dists:
            plt__top = Distribution(min(plt__top.value, array_dists[
                zrssf__enqsd.name].value))
    if aggregate_node.out_key_vars is not None:
        for zrssf__enqsd in aggregate_node.out_key_vars:
            if zrssf__enqsd.name in array_dists:
                plt__top = Distribution(min(plt__top.value, array_dists[
                    zrssf__enqsd.name].value))
    plt__top = Distribution(min(plt__top.value, zekwy__ghlr.value))
    for zrssf__enqsd in aggregate_node.df_out_vars.values():
        array_dists[zrssf__enqsd.name] = plt__top
    if aggregate_node.out_key_vars is not None:
        for zxeth__mco in aggregate_node.out_key_vars:
            array_dists[zxeth__mco.name] = plt__top
    if plt__top != Distribution.OneD_Var:
        for uuuod__lsaph in aggregate_node.key_arrs:
            array_dists[uuuod__lsaph.name] = plt__top
        if aggregate_node.pivot_arr is not None:
            array_dists[aggregate_node.pivot_arr.name] = plt__top
        for zrssf__enqsd in aggregate_node.df_in_vars.values():
            array_dists[zrssf__enqsd.name] = plt__top


distributed_analysis.distributed_analysis_extensions[Aggregate
    ] = aggregate_distributed_analysis


def build_agg_definitions(agg_node, definitions=None):
    if definitions is None:
        definitions = defaultdict(list)
    for zrssf__enqsd in agg_node.df_out_vars.values():
        definitions[zrssf__enqsd.name].append(agg_node)
    if agg_node.out_key_vars is not None:
        for zxeth__mco in agg_node.out_key_vars:
            definitions[zxeth__mco.name].append(agg_node)
    return definitions


ir_utils.build_defs_extensions[Aggregate] = build_agg_definitions


def __update_redvars():
    pass


@infer_global(__update_redvars)
class UpdateDummyTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(types.void, *args)


def __combine_redvars():
    pass


@infer_global(__combine_redvars)
class CombineDummyTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(types.void, *args)


def __eval_res():
    pass


@infer_global(__eval_res)
class EvalDummyTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(args[0].dtype, *args)


def agg_distributed_run(agg_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    parallel = False
    if array_dists is not None:
        parallel = True
        for v in (list(agg_node.df_in_vars.values()) + list(agg_node.
            df_out_vars.values()) + agg_node.key_arrs):
            if array_dists[v.name
                ] != distributed_pass.Distribution.OneD and array_dists[v.name
                ] != distributed_pass.Distribution.OneD_Var:
                parallel = False
    mbzr__ozp = tuple(typemap[v.name] for v in agg_node.key_arrs)
    bwvtv__xzfr = [v for hppbg__rvrn, v in agg_node.df_in_vars.items()]
    zak__jtmw = [v for hppbg__rvrn, v in agg_node.df_out_vars.items()]
    in_col_typs = []
    fxql__fdc = []
    if agg_node.pivot_arr is not None:
        for ijwn__qvl, dpsxj__drz in agg_node.gb_info_in.items():
            for func, getir__teh in dpsxj__drz:
                if ijwn__qvl is not None:
                    in_col_typs.append(typemap[agg_node.df_in_vars[
                        ijwn__qvl].name])
                fxql__fdc.append(func)
    else:
        for ijwn__qvl, func in agg_node.gb_info_out.values():
            if ijwn__qvl is not None:
                in_col_typs.append(typemap[agg_node.df_in_vars[ijwn__qvl].name]
                    )
            fxql__fdc.append(func)
    out_col_typs = tuple(typemap[v.name] for v in zak__jtmw)
    pivot_typ = types.none if agg_node.pivot_arr is None else typemap[agg_node
        .pivot_arr.name]
    arg_typs = tuple(mbzr__ozp + tuple(typemap[v.name] for v in bwvtv__xzfr
        ) + (pivot_typ,))
    gtnop__fsn = {'bodo': bodo, 'np': np, 'dt64_dtype': np.dtype(
        'datetime64[ns]'), 'td64_dtype': np.dtype('timedelta64[ns]')}
    for i, in_col_typ in enumerate(in_col_typs):
        if isinstance(in_col_typ, bodo.CategoricalArrayType):
            gtnop__fsn.update({f'in_cat_dtype_{i}': in_col_typ})
    for i, zfau__kow in enumerate(out_col_typs):
        if isinstance(zfau__kow, bodo.CategoricalArrayType):
            gtnop__fsn.update({f'out_cat_dtype_{i}': zfau__kow})
    udf_func_struct = get_udf_func_struct(fxql__fdc, agg_node.
        input_has_index, in_col_typs, out_col_typs, typingctx, targetctx,
        pivot_typ, agg_node.pivot_values, agg_node.is_crosstab)
    yoz__ktr = gen_top_level_agg_func(agg_node, in_col_typs, out_col_typs,
        parallel, udf_func_struct)
    gtnop__fsn.update({'pd': pd, 'pre_alloc_string_array':
        pre_alloc_string_array, 'pre_alloc_binary_array':
        pre_alloc_binary_array, 'pre_alloc_array_item_array':
        pre_alloc_array_item_array, 'string_array_type': string_array_type,
        'alloc_decimal_array': alloc_decimal_array, 'array_to_info':
        array_to_info, 'arr_info_list_to_table': arr_info_list_to_table,
        'coerce_to_array': bodo.utils.conversion.coerce_to_array,
        'groupby_and_aggregate': groupby_and_aggregate,
        'pivot_groupby_and_aggregate': pivot_groupby_and_aggregate,
        'compute_node_partition_by_hash': compute_node_partition_by_hash,
        'info_from_table': info_from_table, 'info_to_array': info_to_array,
        'delete_info_decref_array': delete_info_decref_array,
        'delete_table': delete_table, 'add_agg_cfunc_sym':
        add_agg_cfunc_sym, 'get_agg_udf_addr': get_agg_udf_addr,
        'delete_table_decref_arrays': delete_table_decref_arrays})
    if udf_func_struct is not None:
        if udf_func_struct.regular_udfs:
            gtnop__fsn.update({'__update_redvars': udf_func_struct.
                update_all_func, '__init_func': udf_func_struct.init_func,
                '__combine_redvars': udf_func_struct.combine_all_func,
                '__eval_res': udf_func_struct.eval_all_func,
                'cpp_cb_update': udf_func_struct.regular_udf_cfuncs[0],
                'cpp_cb_combine': udf_func_struct.regular_udf_cfuncs[1],
                'cpp_cb_eval': udf_func_struct.regular_udf_cfuncs[2]})
        if udf_func_struct.general_udfs:
            gtnop__fsn.update({'cpp_cb_general': udf_func_struct.
                general_udf_cfunc})
    vuhmr__utkk = compile_to_numba_ir(yoz__ktr, gtnop__fsn, typingctx=
        typingctx, targetctx=targetctx, arg_typs=arg_typs, typemap=typemap,
        calltypes=calltypes).blocks.popitem()[1]
    efnni__lqbsn = []
    if agg_node.pivot_arr is None:
        mtkp__rwpi = agg_node.key_arrs[0].scope
        loc = agg_node.loc
        gbm__aehqt = ir.Var(mtkp__rwpi, mk_unique_var('dummy_none'), loc)
        typemap[gbm__aehqt.name] = types.none
        efnni__lqbsn.append(ir.Assign(ir.Const(None, loc), gbm__aehqt, loc))
        bwvtv__xzfr.append(gbm__aehqt)
    else:
        bwvtv__xzfr.append(agg_node.pivot_arr)
    replace_arg_nodes(vuhmr__utkk, agg_node.key_arrs + bwvtv__xzfr)
    ucqy__mcdb = vuhmr__utkk.body[-3]
    assert is_assign(ucqy__mcdb) and isinstance(ucqy__mcdb.value, ir.Expr
        ) and ucqy__mcdb.value.op == 'build_tuple'
    efnni__lqbsn += vuhmr__utkk.body[:-3]
    uyrs__dvel = list(agg_node.df_out_vars.values())
    if agg_node.out_key_vars is not None:
        uyrs__dvel += agg_node.out_key_vars
    for i, tot__hjro in enumerate(uyrs__dvel):
        peaux__yxob = ucqy__mcdb.value.items[i]
        efnni__lqbsn.append(ir.Assign(peaux__yxob, tot__hjro, tot__hjro.loc))
    return efnni__lqbsn


distributed_pass.distributed_run_extensions[Aggregate] = agg_distributed_run


def get_numba_set(dtype):
    pass


@infer_global(get_numba_set)
class GetNumbaSetTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 1
        arr = args[0]
        dtype = types.Tuple([t.dtype for t in arr.types]) if isinstance(arr,
            types.BaseTuple) else arr.dtype
        if isinstance(arr, types.BaseTuple) and len(arr.types) == 1:
            dtype = arr.types[0].dtype
        return signature(types.Set(dtype), *args)


@lower_builtin(get_numba_set, types.Any)
def lower_get_numba_set(context, builder, sig, args):
    return numba.cpython.setobj.set_empty_constructor(context, builder, sig,
        args)


@infer_global(bool)
class BoolNoneTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 1
        ozaf__szqqm = args[0]
        if ozaf__szqqm == types.none:
            return signature(types.boolean, *args)


@lower_builtin(bool, types.none)
def lower_column_mean_impl(context, builder, sig, args):
    gcwd__fbd = context.compile_internal(builder, lambda a: False, sig, args)
    return gcwd__fbd


def setitem_array_with_str(arr, i, v):
    return


@overload(setitem_array_with_str)
def setitem_array_with_str_overload(arr, i, val):
    if arr == string_array_type:

        def setitem_str_arr(arr, i, val):
            arr[i] = val
        return setitem_str_arr
    if val == string_type:
        return lambda arr, i, val: None

    def setitem_impl(arr, i, val):
        arr[i] = val
    return setitem_impl


def _gen_dummy_alloc(t, colnum=0, is_input=False):
    if isinstance(t, IntegerArrayType):
        wio__pmwm = IntDtype(t.dtype).name
        assert wio__pmwm.endswith('Dtype()')
        wio__pmwm = wio__pmwm[:-7]
        return (
            f"bodo.hiframes.pd_series_ext.get_series_data(pd.Series([1], dtype='{wio__pmwm}'))"
            )
    elif isinstance(t, BooleanArrayType):
        return (
            'bodo.libs.bool_arr_ext.init_bool_array(np.empty(0, np.bool_), np.empty(0, np.uint8))'
            )
    elif isinstance(t, StringArrayType):
        return 'pre_alloc_string_array(1, 1)'
    elif isinstance(t, BinaryArrayType):
        return 'pre_alloc_binary_array(1, 1)'
    elif t == ArrayItemArrayType(string_array_type):
        return 'pre_alloc_array_item_array(1, (1, 1), string_array_type)'
    elif isinstance(t, DecimalArrayType):
        return 'alloc_decimal_array(1, {}, {})'.format(t.precision, t.scale)
    elif isinstance(t, DatetimeDateArrayType):
        return (
            'bodo.hiframes.datetime_date_ext.init_datetime_date_array(np.empty(1, np.int64), np.empty(1, np.uint8))'
            )
    elif isinstance(t, bodo.CategoricalArrayType):
        if t.dtype.categories is None:
            raise BodoError(
                'Groupby agg operations on Categorical types require constant categories'
                )
        yriyl__wzc = 'in' if is_input else 'out'
        return (
            f'bodo.utils.utils.alloc_type(1, {yriyl__wzc}_cat_dtype_{colnum})')
    else:
        return 'np.empty(1, {})'.format(_get_np_dtype(t.dtype))


def _get_np_dtype(t):
    if t == types.bool_:
        return 'np.bool_'
    if t == types.NPDatetime('ns'):
        return 'dt64_dtype'
    if t == types.NPTimedelta('ns'):
        return 'td64_dtype'
    return 'np.{}'.format(t)


def gen_update_cb(udf_func_struct, allfuncs, n_keys, data_in_typs_,
    out_data_typs, do_combine, func_idx_to_in_col, label_suffix):
    ajanq__hvh = udf_func_struct.var_typs
    eek__zwbod = len(ajanq__hvh)
    uxeq__cuh = (
        'def bodo_gb_udf_update_local{}(in_table, out_table, row_to_group):\n'
        .format(label_suffix))
    uxeq__cuh += '    if is_null_pointer(in_table):\n'
    uxeq__cuh += '        return\n'
    uxeq__cuh += '    data_redvar_dummy = ({}{})\n'.format(','.join([
        'np.empty(1, {})'.format(_get_np_dtype(t)) for t in ajanq__hvh]), 
        ',' if len(ajanq__hvh) == 1 else '')
    zujp__mmuk = n_keys
    hbp__oyifl = []
    redvar_offsets = []
    zwqx__ieyu = []
    if do_combine:
        for i, aaq__qpzs in enumerate(allfuncs):
            if aaq__qpzs.ftype != 'udf':
                zujp__mmuk += aaq__qpzs.ncols_pre_shuffle
            else:
                redvar_offsets += list(range(zujp__mmuk, zujp__mmuk +
                    aaq__qpzs.n_redvars))
                zujp__mmuk += aaq__qpzs.n_redvars
                zwqx__ieyu.append(data_in_typs_[func_idx_to_in_col[i]])
                hbp__oyifl.append(func_idx_to_in_col[i] + n_keys)
    else:
        for i, aaq__qpzs in enumerate(allfuncs):
            if aaq__qpzs.ftype != 'udf':
                zujp__mmuk += aaq__qpzs.ncols_post_shuffle
            else:
                redvar_offsets += list(range(zujp__mmuk + 1, zujp__mmuk + 1 +
                    aaq__qpzs.n_redvars))
                zujp__mmuk += aaq__qpzs.n_redvars + 1
                zwqx__ieyu.append(data_in_typs_[func_idx_to_in_col[i]])
                hbp__oyifl.append(func_idx_to_in_col[i] + n_keys)
    assert len(redvar_offsets) == eek__zwbod
    cpwjd__aec = len(zwqx__ieyu)
    xxnq__wqfu = []
    for i, t in enumerate(zwqx__ieyu):
        xxnq__wqfu.append(_gen_dummy_alloc(t, i, True))
    uxeq__cuh += '    data_in_dummy = ({}{})\n'.format(','.join(xxnq__wqfu),
        ',' if len(zwqx__ieyu) == 1 else '')
    uxeq__cuh += """
    # initialize redvar cols
"""
    uxeq__cuh += '    init_vals = __init_func()\n'
    for i in range(eek__zwbod):
        uxeq__cuh += (
            """    redvar_arr_{} = info_to_array(info_from_table(out_table, {}), data_redvar_dummy[{}])
"""
            .format(i, redvar_offsets[i], i))
        uxeq__cuh += '    incref(redvar_arr_{})\n'.format(i)
        uxeq__cuh += '    redvar_arr_{}.fill(init_vals[{}])\n'.format(i, i)
    uxeq__cuh += '    redvars = ({}{})\n'.format(','.join(['redvar_arr_{}'.
        format(i) for i in range(eek__zwbod)]), ',' if eek__zwbod == 1 else '')
    uxeq__cuh += '\n'
    for i in range(cpwjd__aec):
        uxeq__cuh += (
            """    data_in_{} = info_to_array(info_from_table(in_table, {}), data_in_dummy[{}])
"""
            .format(i, hbp__oyifl[i], i))
        uxeq__cuh += '    incref(data_in_{})\n'.format(i)
    uxeq__cuh += '    data_in = ({}{})\n'.format(','.join(['data_in_{}'.
        format(i) for i in range(cpwjd__aec)]), ',' if cpwjd__aec == 1 else '')
    uxeq__cuh += '\n'
    uxeq__cuh += '    for i in range(len(data_in_0)):\n'
    uxeq__cuh += '        w_ind = row_to_group[i]\n'
    uxeq__cuh += '        if w_ind != -1:\n'
    uxeq__cuh += (
        '            __update_redvars(redvars, data_in, w_ind, i, pivot_arr=None)\n'
        )
    oooax__bmx = {}
    exec(uxeq__cuh, {'bodo': bodo, 'np': np, 'pd': pd, 'info_to_array':
        info_to_array, 'info_from_table': info_from_table, 'incref': incref,
        'pre_alloc_string_array': pre_alloc_string_array, '__init_func':
        udf_func_struct.init_func, '__update_redvars': udf_func_struct.
        update_all_func, 'is_null_pointer': is_null_pointer, 'dt64_dtype':
        np.dtype('datetime64[ns]'), 'td64_dtype': np.dtype(
        'timedelta64[ns]')}, oooax__bmx)
    return oooax__bmx['bodo_gb_udf_update_local{}'.format(label_suffix)]


def gen_combine_cb(udf_func_struct, allfuncs, n_keys, out_data_typs,
    label_suffix):
    ajanq__hvh = udf_func_struct.var_typs
    eek__zwbod = len(ajanq__hvh)
    uxeq__cuh = (
        'def bodo_gb_udf_combine{}(in_table, out_table, row_to_group):\n'.
        format(label_suffix))
    uxeq__cuh += '    if is_null_pointer(in_table):\n'
    uxeq__cuh += '        return\n'
    uxeq__cuh += '    data_redvar_dummy = ({}{})\n'.format(','.join([
        'np.empty(1, {})'.format(_get_np_dtype(t)) for t in ajanq__hvh]), 
        ',' if len(ajanq__hvh) == 1 else '')
    kgs__psw = n_keys
    mlr__tgjg = n_keys
    iiv__votk = []
    anzki__fgpxt = []
    for aaq__qpzs in allfuncs:
        if aaq__qpzs.ftype != 'udf':
            kgs__psw += aaq__qpzs.ncols_pre_shuffle
            mlr__tgjg += aaq__qpzs.ncols_post_shuffle
        else:
            iiv__votk += list(range(kgs__psw, kgs__psw + aaq__qpzs.n_redvars))
            anzki__fgpxt += list(range(mlr__tgjg + 1, mlr__tgjg + 1 +
                aaq__qpzs.n_redvars))
            kgs__psw += aaq__qpzs.n_redvars
            mlr__tgjg += 1 + aaq__qpzs.n_redvars
    assert len(iiv__votk) == eek__zwbod
    uxeq__cuh += """
    # initialize redvar cols
"""
    uxeq__cuh += '    init_vals = __init_func()\n'
    for i in range(eek__zwbod):
        uxeq__cuh += (
            """    redvar_arr_{} = info_to_array(info_from_table(out_table, {}), data_redvar_dummy[{}])
"""
            .format(i, anzki__fgpxt[i], i))
        uxeq__cuh += '    incref(redvar_arr_{})\n'.format(i)
        uxeq__cuh += '    redvar_arr_{}.fill(init_vals[{}])\n'.format(i, i)
    uxeq__cuh += '    redvars = ({}{})\n'.format(','.join(['redvar_arr_{}'.
        format(i) for i in range(eek__zwbod)]), ',' if eek__zwbod == 1 else '')
    uxeq__cuh += '\n'
    for i in range(eek__zwbod):
        uxeq__cuh += (
            """    recv_redvar_arr_{} = info_to_array(info_from_table(in_table, {}), data_redvar_dummy[{}])
"""
            .format(i, iiv__votk[i], i))
        uxeq__cuh += '    incref(recv_redvar_arr_{})\n'.format(i)
    uxeq__cuh += '    recv_redvars = ({}{})\n'.format(','.join([
        'recv_redvar_arr_{}'.format(i) for i in range(eek__zwbod)]), ',' if
        eek__zwbod == 1 else '')
    uxeq__cuh += '\n'
    if eek__zwbod:
        uxeq__cuh += '    for i in range(len(recv_redvar_arr_0)):\n'
        uxeq__cuh += '        w_ind = row_to_group[i]\n'
        uxeq__cuh += (
            '        __combine_redvars(redvars, recv_redvars, w_ind, i, pivot_arr=None)\n'
            )
    oooax__bmx = {}
    exec(uxeq__cuh, {'np': np, 'info_to_array': info_to_array,
        'info_from_table': info_from_table, 'incref': incref, '__init_func':
        udf_func_struct.init_func, '__combine_redvars': udf_func_struct.
        combine_all_func, 'is_null_pointer': is_null_pointer, 'dt64_dtype':
        np.dtype('datetime64[ns]'), 'td64_dtype': np.dtype(
        'timedelta64[ns]')}, oooax__bmx)
    return oooax__bmx['bodo_gb_udf_combine{}'.format(label_suffix)]


def gen_eval_cb(udf_func_struct, allfuncs, n_keys, out_data_typs_, label_suffix
    ):
    ajanq__hvh = udf_func_struct.var_typs
    eek__zwbod = len(ajanq__hvh)
    zujp__mmuk = n_keys
    redvar_offsets = []
    lsab__tcmh = []
    out_data_typs = []
    for i, aaq__qpzs in enumerate(allfuncs):
        if aaq__qpzs.ftype != 'udf':
            zujp__mmuk += aaq__qpzs.ncols_post_shuffle
        else:
            lsab__tcmh.append(zujp__mmuk)
            redvar_offsets += list(range(zujp__mmuk + 1, zujp__mmuk + 1 +
                aaq__qpzs.n_redvars))
            zujp__mmuk += 1 + aaq__qpzs.n_redvars
            out_data_typs.append(out_data_typs_[i])
    assert len(redvar_offsets) == eek__zwbod
    cpwjd__aec = len(out_data_typs)
    uxeq__cuh = 'def bodo_gb_udf_eval{}(table):\n'.format(label_suffix)
    uxeq__cuh += '    if is_null_pointer(table):\n'
    uxeq__cuh += '        return\n'
    uxeq__cuh += '    data_redvar_dummy = ({}{})\n'.format(','.join([
        'np.empty(1, {})'.format(_get_np_dtype(t)) for t in ajanq__hvh]), 
        ',' if len(ajanq__hvh) == 1 else '')
    uxeq__cuh += '    out_data_dummy = ({}{})\n'.format(','.join([
        'np.empty(1, {})'.format(_get_np_dtype(t.dtype)) for t in
        out_data_typs]), ',' if len(out_data_typs) == 1 else '')
    for i in range(eek__zwbod):
        uxeq__cuh += (
            """    redvar_arr_{} = info_to_array(info_from_table(table, {}), data_redvar_dummy[{}])
"""
            .format(i, redvar_offsets[i], i))
        uxeq__cuh += '    incref(redvar_arr_{})\n'.format(i)
    uxeq__cuh += '    redvars = ({}{})\n'.format(','.join(['redvar_arr_{}'.
        format(i) for i in range(eek__zwbod)]), ',' if eek__zwbod == 1 else '')
    uxeq__cuh += '\n'
    for i in range(cpwjd__aec):
        uxeq__cuh += (
            """    data_out_{} = info_to_array(info_from_table(table, {}), out_data_dummy[{}])
"""
            .format(i, lsab__tcmh[i], i))
        uxeq__cuh += '    incref(data_out_{})\n'.format(i)
    uxeq__cuh += '    data_out = ({}{})\n'.format(','.join(['data_out_{}'.
        format(i) for i in range(cpwjd__aec)]), ',' if cpwjd__aec == 1 else '')
    uxeq__cuh += '\n'
    uxeq__cuh += '    for i in range(len(data_out_0)):\n'
    uxeq__cuh += '        __eval_res(redvars, data_out, i)\n'
    oooax__bmx = {}
    exec(uxeq__cuh, {'np': np, 'info_to_array': info_to_array,
        'info_from_table': info_from_table, 'incref': incref, '__eval_res':
        udf_func_struct.eval_all_func, 'is_null_pointer': is_null_pointer,
        'dt64_dtype': np.dtype('datetime64[ns]'), 'td64_dtype': np.dtype(
        'timedelta64[ns]')}, oooax__bmx)
    return oooax__bmx['bodo_gb_udf_eval{}'.format(label_suffix)]


def gen_general_udf_cb(udf_func_struct, allfuncs, n_keys, in_col_typs,
    out_col_typs, func_idx_to_in_col, label_suffix):
    zujp__mmuk = n_keys
    jerad__gkhzc = []
    for i, aaq__qpzs in enumerate(allfuncs):
        if aaq__qpzs.ftype == 'gen_udf':
            jerad__gkhzc.append(zujp__mmuk)
            zujp__mmuk += 1
        elif aaq__qpzs.ftype != 'udf':
            zujp__mmuk += aaq__qpzs.ncols_post_shuffle
        else:
            zujp__mmuk += aaq__qpzs.n_redvars + 1
    uxeq__cuh = (
        'def bodo_gb_apply_general_udfs{}(num_groups, in_table, out_table):\n'
        .format(label_suffix))
    uxeq__cuh += '    if num_groups == 0:\n'
    uxeq__cuh += '        return\n'
    for i, func in enumerate(udf_func_struct.general_udf_funcs):
        uxeq__cuh += '    # col {}\n'.format(i)
        uxeq__cuh += (
            """    out_col = info_to_array(info_from_table(out_table, {}), out_col_{}_typ)
"""
            .format(jerad__gkhzc[i], i))
        uxeq__cuh += '    incref(out_col)\n'
        uxeq__cuh += '    for j in range(num_groups):\n'
        uxeq__cuh += (
            """        in_col = info_to_array(info_from_table(in_table, {}*num_groups + j), in_col_{}_typ)
"""
            .format(i, i))
        uxeq__cuh += '        incref(in_col)\n'
        uxeq__cuh += (
            '        out_col[j] = func_{}(pd.Series(in_col))  # func returns scalar\n'
            .format(i))
    gtnop__fsn = {'pd': pd, 'info_to_array': info_to_array,
        'info_from_table': info_from_table, 'incref': incref}
    lxql__wxp = 0
    for i, func in enumerate(allfuncs):
        if func.ftype != 'gen_udf':
            continue
        func = udf_func_struct.general_udf_funcs[lxql__wxp]
        gtnop__fsn['func_{}'.format(lxql__wxp)] = func
        gtnop__fsn['in_col_{}_typ'.format(lxql__wxp)] = in_col_typs[
            func_idx_to_in_col[i]]
        gtnop__fsn['out_col_{}_typ'.format(lxql__wxp)] = out_col_typs[i]
        lxql__wxp += 1
    oooax__bmx = {}
    exec(uxeq__cuh, gtnop__fsn, oooax__bmx)
    aaq__qpzs = oooax__bmx['bodo_gb_apply_general_udfs{}'.format(label_suffix)]
    gyvm__ldq = types.void(types.int64, types.voidptr, types.voidptr)
    return numba.cfunc(gyvm__ldq, nopython=True)(aaq__qpzs)


def gen_top_level_agg_func(agg_node, in_col_typs, out_col_typs, parallel,
    udf_func_struct):
    jbf__kthoj = agg_node.pivot_arr is not None
    if agg_node.same_index:
        assert agg_node.input_has_index
    if agg_node.pivot_values is None:
        vew__vcn = 1
    else:
        vew__vcn = len(agg_node.pivot_values)
    cmv__hyeu = tuple('key_' + sanitize_varname(vat__plxf) for vat__plxf in
        agg_node.key_names)
    jijdq__eoaj = {vat__plxf: 'in_{}'.format(sanitize_varname(vat__plxf)) for
        vat__plxf in agg_node.gb_info_in.keys() if vat__plxf is not None}
    uiswl__ehu = {vat__plxf: ('out_' + sanitize_varname(vat__plxf)) for
        vat__plxf in agg_node.gb_info_out.keys()}
    n_keys = len(agg_node.key_names)
    qswth__zylu = ', '.join(cmv__hyeu)
    zpru__qwfod = ', '.join(jijdq__eoaj.values())
    if zpru__qwfod != '':
        zpru__qwfod = ', ' + zpru__qwfod
    uxeq__cuh = 'def agg_top({}{}{}, pivot_arr):\n'.format(qswth__zylu,
        zpru__qwfod, ', index_arg' if agg_node.input_has_index else '')
    if jbf__kthoj:
        qerxt__lamke = []
        for ijwn__qvl, dpsxj__drz in agg_node.gb_info_in.items():
            if ijwn__qvl is not None:
                for func, getir__teh in dpsxj__drz:
                    qerxt__lamke.append(jijdq__eoaj[ijwn__qvl])
    else:
        qerxt__lamke = tuple(jijdq__eoaj[ijwn__qvl] for ijwn__qvl,
            getir__teh in agg_node.gb_info_out.values() if ijwn__qvl is not
            None)
    arenf__zsm = cmv__hyeu + tuple(qerxt__lamke)
    uxeq__cuh += '    info_list = [{}{}{}]\n'.format(', '.join(
        'array_to_info({})'.format(a) for a in arenf__zsm), 
        ', array_to_info(index_arg)' if agg_node.input_has_index else '', 
        ', array_to_info(pivot_arr)' if agg_node.is_crosstab else '')
    uxeq__cuh += '    table = arr_info_list_to_table(info_list)\n'
    for i, vat__plxf in enumerate(agg_node.gb_info_out.keys()):
        yik__qao = uiswl__ehu[vat__plxf] + '_dummy'
        zfau__kow = out_col_typs[i]
        ijwn__qvl, func = agg_node.gb_info_out[vat__plxf]
        if isinstance(func, pytypes.SimpleNamespace) and func.fname in ['min',
            'max', 'shift'] and isinstance(zfau__kow, bodo.CategoricalArrayType
            ):
            uxeq__cuh += '    {} = {}\n'.format(yik__qao, jijdq__eoaj[
                ijwn__qvl])
        else:
            uxeq__cuh += '    {} = {}\n'.format(yik__qao, _gen_dummy_alloc(
                zfau__kow, i, False))
    do_combine = parallel
    allfuncs = []
    sblig__akwb = []
    func_idx_to_in_col = []
    tczj__rqbt = []
    wofs__wlnft = False
    heolz__foqe = 1
    lxthx__yht = -1
    umn__lkm = 0
    paaz__shh = 0
    if not jbf__kthoj:
        fxql__fdc = [func for getir__teh, func in agg_node.gb_info_out.values()
            ]
    else:
        fxql__fdc = [func for func, getir__teh in dpsxj__drz for dpsxj__drz in
            agg_node.gb_info_in.values()]
    for dag__rdpw, func in enumerate(fxql__fdc):
        sblig__akwb.append(len(allfuncs))
        if func.ftype in {'median', 'nunique'}:
            do_combine = False
        if func.ftype in list_cumulative:
            umn__lkm += 1
        if hasattr(func, 'skipdropna'):
            wofs__wlnft = func.skipdropna
        if func.ftype == 'shift':
            heolz__foqe = func.periods
            do_combine = False
        if func.ftype in {'transform'}:
            paaz__shh = func.transform_func
            do_combine = False
        if func.ftype == 'head':
            lxthx__yht = func.head_n
            do_combine = False
        allfuncs.append(func)
        func_idx_to_in_col.append(dag__rdpw)
        if func.ftype == 'udf':
            tczj__rqbt.append(func.n_redvars)
        elif func.ftype == 'gen_udf':
            tczj__rqbt.append(0)
            do_combine = False
    sblig__akwb.append(len(allfuncs))
    if agg_node.is_crosstab:
        assert len(agg_node.gb_info_out
            ) == vew__vcn, 'invalid number of groupby outputs for pivot'
    else:
        assert len(agg_node.gb_info_out) == len(allfuncs
            ) * vew__vcn, 'invalid number of groupby outputs'
    if umn__lkm > 0:
        if umn__lkm != len(allfuncs):
            raise BodoError(
                f'{agg_node.func_name}(): Cannot mix cumulative operations with other aggregation functions'
                , loc=agg_node.loc)
        do_combine = False
    if udf_func_struct is not None:
        bji__gnv = next_label()
        if udf_func_struct.regular_udfs:
            gyvm__ldq = types.void(types.voidptr, types.voidptr, types.
                CPointer(types.int64))
            udcq__cwc = numba.cfunc(gyvm__ldq, nopython=True)(gen_update_cb
                (udf_func_struct, allfuncs, n_keys, in_col_typs,
                out_col_typs, do_combine, func_idx_to_in_col, bji__gnv))
            ymekw__znkp = numba.cfunc(gyvm__ldq, nopython=True)(gen_combine_cb
                (udf_func_struct, allfuncs, n_keys, out_col_typs, bji__gnv))
            phqk__wyyag = numba.cfunc('void(voidptr)', nopython=True)(
                gen_eval_cb(udf_func_struct, allfuncs, n_keys, out_col_typs,
                bji__gnv))
            udf_func_struct.set_regular_cfuncs(udcq__cwc, ymekw__znkp,
                phqk__wyyag)
            for osxak__oczov in udf_func_struct.regular_udf_cfuncs:
                gb_agg_cfunc[osxak__oczov.native_name] = osxak__oczov
                gb_agg_cfunc_addr[osxak__oczov.native_name
                    ] = osxak__oczov.address
        if udf_func_struct.general_udfs:
            uwfx__sef = gen_general_udf_cb(udf_func_struct, allfuncs,
                n_keys, in_col_typs, out_col_typs, func_idx_to_in_col, bji__gnv
                )
            udf_func_struct.set_general_cfunc(uwfx__sef)
        pjtsj__cputt = []
        nqig__lxk = 0
        i = 0
        for yik__qao, aaq__qpzs in zip(uiswl__ehu.values(), allfuncs):
            if aaq__qpzs.ftype in ('udf', 'gen_udf'):
                pjtsj__cputt.append(yik__qao + '_dummy')
                for vtfn__hokn in range(nqig__lxk, nqig__lxk + tczj__rqbt[i]):
                    pjtsj__cputt.append('data_redvar_dummy_' + str(vtfn__hokn))
                nqig__lxk += tczj__rqbt[i]
                i += 1
        if udf_func_struct.regular_udfs:
            ajanq__hvh = udf_func_struct.var_typs
            for i, t in enumerate(ajanq__hvh):
                uxeq__cuh += ('    data_redvar_dummy_{} = np.empty(1, {})\n'
                    .format(i, _get_np_dtype(t)))
        uxeq__cuh += '    out_info_list_dummy = [{}]\n'.format(', '.join(
            'array_to_info({})'.format(a) for a in pjtsj__cputt))
        uxeq__cuh += (
            '    udf_table_dummy = arr_info_list_to_table(out_info_list_dummy)\n'
            )
        if udf_func_struct.regular_udfs:
            uxeq__cuh += "    add_agg_cfunc_sym(cpp_cb_update, '{}')\n".format(
                udcq__cwc.native_name)
            uxeq__cuh += ("    add_agg_cfunc_sym(cpp_cb_combine, '{}')\n".
                format(ymekw__znkp.native_name))
            uxeq__cuh += "    add_agg_cfunc_sym(cpp_cb_eval, '{}')\n".format(
                phqk__wyyag.native_name)
            uxeq__cuh += ("    cpp_cb_update_addr = get_agg_udf_addr('{}')\n"
                .format(udcq__cwc.native_name))
            uxeq__cuh += ("    cpp_cb_combine_addr = get_agg_udf_addr('{}')\n"
                .format(ymekw__znkp.native_name))
            uxeq__cuh += ("    cpp_cb_eval_addr = get_agg_udf_addr('{}')\n"
                .format(phqk__wyyag.native_name))
        else:
            uxeq__cuh += '    cpp_cb_update_addr = 0\n'
            uxeq__cuh += '    cpp_cb_combine_addr = 0\n'
            uxeq__cuh += '    cpp_cb_eval_addr = 0\n'
        if udf_func_struct.general_udfs:
            osxak__oczov = udf_func_struct.general_udf_cfunc
            gb_agg_cfunc[osxak__oczov.native_name] = osxak__oczov
            gb_agg_cfunc_addr[osxak__oczov.native_name] = osxak__oczov.address
            uxeq__cuh += ("    add_agg_cfunc_sym(cpp_cb_general, '{}')\n".
                format(osxak__oczov.native_name))
            uxeq__cuh += ("    cpp_cb_general_addr = get_agg_udf_addr('{}')\n"
                .format(osxak__oczov.native_name))
        else:
            uxeq__cuh += '    cpp_cb_general_addr = 0\n'
    else:
        uxeq__cuh += (
            '    udf_table_dummy = arr_info_list_to_table([array_to_info(np.empty(1))])\n'
            )
        uxeq__cuh += '    cpp_cb_update_addr = 0\n'
        uxeq__cuh += '    cpp_cb_combine_addr = 0\n'
        uxeq__cuh += '    cpp_cb_eval_addr = 0\n'
        uxeq__cuh += '    cpp_cb_general_addr = 0\n'
    uxeq__cuh += '    ftypes = np.array([{}, 0], dtype=np.int32)\n'.format(', '
        .join([str(supported_agg_funcs.index(aaq__qpzs.ftype)) for
        aaq__qpzs in allfuncs] + ['0']))
    uxeq__cuh += '    func_offsets = np.array({}, dtype=np.int32)\n'.format(str
        (sblig__akwb))
    if len(tczj__rqbt) > 0:
        uxeq__cuh += '    udf_ncols = np.array({}, dtype=np.int32)\n'.format(
            str(tczj__rqbt))
    else:
        uxeq__cuh += '    udf_ncols = np.array([0], np.int32)\n'
    if jbf__kthoj:
        uxeq__cuh += '    arr_type = coerce_to_array({})\n'.format(agg_node
            .pivot_values)
        uxeq__cuh += '    arr_info = array_to_info(arr_type)\n'
        uxeq__cuh += (
            '    dispatch_table = arr_info_list_to_table([arr_info])\n')
        uxeq__cuh += '    pivot_info = array_to_info(pivot_arr)\n'
        uxeq__cuh += (
            '    dispatch_info = arr_info_list_to_table([pivot_info])\n')
        uxeq__cuh += (
            """    out_table = pivot_groupby_and_aggregate(table, {}, dispatch_table, dispatch_info, {}, ftypes.ctypes, func_offsets.ctypes, udf_ncols.ctypes, {}, {}, {}, {}, {}, cpp_cb_update_addr, cpp_cb_combine_addr, cpp_cb_eval_addr, udf_table_dummy)
"""
            .format(n_keys, agg_node.input_has_index, parallel, agg_node.
            is_crosstab, wofs__wlnft, agg_node.return_key, agg_node.same_index)
            )
        uxeq__cuh += '    delete_info_decref_array(pivot_info)\n'
        uxeq__cuh += '    delete_info_decref_array(arr_info)\n'
    else:
        uxeq__cuh += (
            """    out_table = groupby_and_aggregate(table, {}, {}, ftypes.ctypes, func_offsets.ctypes, udf_ncols.ctypes, {}, {}, {}, {}, {}, {}, {}, {}, cpp_cb_update_addr, cpp_cb_combine_addr, cpp_cb_eval_addr, cpp_cb_general_addr, udf_table_dummy)
"""
            .format(n_keys, agg_node.input_has_index, parallel, wofs__wlnft,
            heolz__foqe, paaz__shh, lxthx__yht, agg_node.return_key,
            agg_node.same_index, agg_node.dropna))
    uoyvj__dfci = 0
    if agg_node.return_key:
        for i, umc__mlb in enumerate(cmv__hyeu):
            uxeq__cuh += (
                '    {} = info_to_array(info_from_table(out_table, {}), {})\n'
                .format(umc__mlb, uoyvj__dfci, umc__mlb))
            uoyvj__dfci += 1
    for yik__qao in uiswl__ehu.values():
        uxeq__cuh += (
            '    {} = info_to_array(info_from_table(out_table, {}), {})\n'.
            format(yik__qao, uoyvj__dfci, yik__qao + '_dummy'))
        uoyvj__dfci += 1
    if agg_node.same_index:
        uxeq__cuh += (
            """    out_index_arg = info_to_array(info_from_table(out_table, {}), index_arg)
"""
            .format(uoyvj__dfci))
        uoyvj__dfci += 1
    uxeq__cuh += (
        f"    ev_clean = bodo.utils.tracing.Event('tables_clean_up', {parallel})\n"
        )
    uxeq__cuh += '    delete_table_decref_arrays(table)\n'
    uxeq__cuh += '    delete_table_decref_arrays(udf_table_dummy)\n'
    uxeq__cuh += '    delete_table(out_table)\n'
    uxeq__cuh += f'    ev_clean.finalize()\n'
    fvov__ast = tuple(uiswl__ehu.values())
    if agg_node.return_key:
        fvov__ast += tuple(cmv__hyeu)
    uxeq__cuh += '    return ({},{})\n'.format(', '.join(fvov__ast), 
        ' out_index_arg,' if agg_node.same_index else '')
    oooax__bmx = {}
    exec(uxeq__cuh, {}, oooax__bmx)
    ycmbt__kmh = oooax__bmx['agg_top']
    return ycmbt__kmh


def compile_to_optimized_ir(func, arg_typs, typingctx, targetctx):
    code = func.code if hasattr(func, 'code') else func.__code__
    closure = func.closure if hasattr(func, 'closure') else func.__closure__
    f_ir = get_ir_of_code(func.__globals__, code)
    replace_closures(f_ir, closure, code)
    for block in f_ir.blocks.values():
        for vbmp__juw in block.body:
            if is_call_assign(vbmp__juw) and find_callname(f_ir, vbmp__juw.
                value) == ('len', 'builtins') and vbmp__juw.value.args[0
                ].name == f_ir.arg_names[0]:
                gqb__flxb = get_definition(f_ir, vbmp__juw.value.func)
                gqb__flxb.name = 'dummy_agg_count'
                gqb__flxb.value = dummy_agg_count
    oijcs__dbzq = get_name_var_table(f_ir.blocks)
    mwsqc__ahkn = {}
    for name, getir__teh in oijcs__dbzq.items():
        mwsqc__ahkn[name] = mk_unique_var(name)
    replace_var_names(f_ir.blocks, mwsqc__ahkn)
    f_ir._definitions = build_definitions(f_ir.blocks)
    assert f_ir.arg_count == 1, 'agg function should have one input'
    iaw__jbeqf = numba.core.compiler.Flags()
    iaw__jbeqf.nrt = True
    ijaj__vgauj = bodo.transforms.untyped_pass.UntypedPass(f_ir, typingctx,
        arg_typs, {}, {}, iaw__jbeqf)
    ijaj__vgauj.run()
    f_ir._definitions = build_definitions(f_ir.blocks)
    typemap, qwnts__vsoy, calltypes, getir__teh = (numba.core.typed_passes.
        type_inference_stage(typingctx, targetctx, f_ir, arg_typs, None))
    chi__udo = numba.core.cpu.ParallelOptions(True)
    targetctx = numba.core.cpu.CPUContext(typingctx)
    bvypg__onsqj = namedtuple('DummyPipeline', ['typingctx', 'targetctx',
        'args', 'func_ir', 'typemap', 'return_type', 'calltypes',
        'type_annotation', 'locals', 'flags', 'pipeline'])
    ybh__gtek = namedtuple('TypeAnnotation', ['typemap', 'calltypes'])
    mxe__hqpnb = ybh__gtek(typemap, calltypes)
    pm = bvypg__onsqj(typingctx, targetctx, None, f_ir, typemap,
        qwnts__vsoy, calltypes, mxe__hqpnb, {}, iaw__jbeqf, None)
    reio__xbp = numba.core.compiler.DefaultPassBuilder.define_untyped_pipeline(
        pm)
    pm = bvypg__onsqj(typingctx, targetctx, None, f_ir, typemap,
        qwnts__vsoy, calltypes, mxe__hqpnb, {}, iaw__jbeqf, reio__xbp)
    ucypf__xtnqd = numba.core.typed_passes.InlineOverloads()
    ucypf__xtnqd.run_pass(pm)
    laxu__puxrs = bodo.transforms.series_pass.SeriesPass(f_ir, typingctx,
        targetctx, typemap, calltypes, {}, False)
    laxu__puxrs.run()
    for block in f_ir.blocks.values():
        for vbmp__juw in block.body:
            if is_assign(vbmp__juw) and isinstance(vbmp__juw.value, (ir.Arg,
                ir.Var)) and isinstance(typemap[vbmp__juw.target.name],
                SeriesType):
                qfcg__xhnv = typemap.pop(vbmp__juw.target.name)
                typemap[vbmp__juw.target.name] = qfcg__xhnv.data
            if is_call_assign(vbmp__juw) and find_callname(f_ir, vbmp__juw.
                value) == ('get_series_data', 'bodo.hiframes.pd_series_ext'):
                f_ir._definitions[vbmp__juw.target.name].remove(vbmp__juw.value
                    )
                vbmp__juw.value = vbmp__juw.value.args[0]
                f_ir._definitions[vbmp__juw.target.name].append(vbmp__juw.value
                    )
            if is_call_assign(vbmp__juw) and find_callname(f_ir, vbmp__juw.
                value) == ('isna', 'bodo.libs.array_kernels'):
                f_ir._definitions[vbmp__juw.target.name].remove(vbmp__juw.value
                    )
                vbmp__juw.value = ir.Const(False, vbmp__juw.loc)
                f_ir._definitions[vbmp__juw.target.name].append(vbmp__juw.value
                    )
            if is_call_assign(vbmp__juw) and find_callname(f_ir, vbmp__juw.
                value) == ('setna', 'bodo.libs.array_kernels'):
                f_ir._definitions[vbmp__juw.target.name].remove(vbmp__juw.value
                    )
                vbmp__juw.value = ir.Const(False, vbmp__juw.loc)
                f_ir._definitions[vbmp__juw.target.name].append(vbmp__juw.value
                    )
    bodo.transforms.untyped_pass.remove_dead_branches(f_ir)
    eyinb__xjy = numba.parfors.parfor.PreParforPass(f_ir, typemap,
        calltypes, typingctx, targetctx, chi__udo)
    eyinb__xjy.run()
    f_ir._definitions = build_definitions(f_ir.blocks)
    miauh__bjb = numba.core.compiler.StateDict()
    miauh__bjb.func_ir = f_ir
    miauh__bjb.typemap = typemap
    miauh__bjb.calltypes = calltypes
    miauh__bjb.typingctx = typingctx
    miauh__bjb.targetctx = targetctx
    miauh__bjb.return_type = qwnts__vsoy
    numba.core.rewrites.rewrite_registry.apply('after-inference', miauh__bjb)
    zrf__cfjj = numba.parfors.parfor.ParforPass(f_ir, typemap, calltypes,
        qwnts__vsoy, typingctx, targetctx, chi__udo, iaw__jbeqf, {})
    zrf__cfjj.run()
    remove_dels(f_ir.blocks)
    numba.parfors.parfor.maximize_fusion(f_ir, f_ir.blocks, typemap, False)
    return f_ir, pm


def replace_closures(f_ir, closure, code):
    if closure:
        closure = f_ir.get_definition(closure)
        if isinstance(closure, tuple):
            qrp__xhp = ctypes.pythonapi.PyCell_Get
            qrp__xhp.restype = ctypes.py_object
            qrp__xhp.argtypes = ctypes.py_object,
            ruzpz__bou = tuple(qrp__xhp(lpvh__see) for lpvh__see in closure)
        else:
            assert isinstance(closure, ir.Expr) and closure.op == 'build_tuple'
            ruzpz__bou = closure.items
        assert len(code.co_freevars) == len(ruzpz__bou)
        numba.core.inline_closurecall._replace_freevars(f_ir.blocks, ruzpz__bou
            )


class RegularUDFGenerator(object):

    def __init__(self, in_col_types, out_col_types, pivot_typ, pivot_values,
        is_crosstab, typingctx, targetctx):
        self.in_col_types = in_col_types
        self.out_col_types = out_col_types
        self.pivot_typ = pivot_typ
        self.pivot_values = pivot_values
        self.is_crosstab = is_crosstab
        self.typingctx = typingctx
        self.targetctx = targetctx
        self.all_reduce_vars = []
        self.all_vartypes = []
        self.all_init_nodes = []
        self.all_eval_funcs = []
        self.all_update_funcs = []
        self.all_combine_funcs = []
        self.curr_offset = 0
        self.redvar_offsets = [0]

    def add_udf(self, in_col_typ, func):
        bpnly__qozwf = SeriesType(in_col_typ.dtype, in_col_typ, None,
            string_type)
        f_ir, pm = compile_to_optimized_ir(func, (bpnly__qozwf,), self.
            typingctx, self.targetctx)
        f_ir._definitions = build_definitions(f_ir.blocks)
        assert len(f_ir.blocks
            ) == 1 and 0 in f_ir.blocks, 'only simple functions with one block supported for aggregation'
        block = f_ir.blocks[0]
        qlq__cait, arr_var = _rm_arg_agg_block(block, pm.typemap)
        indzj__cvjw = -1
        for i, vbmp__juw in enumerate(qlq__cait):
            if isinstance(vbmp__juw, numba.parfors.parfor.Parfor):
                assert indzj__cvjw == -1, 'only one parfor for aggregation function'
                indzj__cvjw = i
        parfor = None
        if indzj__cvjw != -1:
            parfor = qlq__cait[indzj__cvjw]
            remove_dels(parfor.loop_body)
            remove_dels({(0): parfor.init_block})
        init_nodes = []
        if parfor:
            init_nodes = qlq__cait[:indzj__cvjw] + parfor.init_block.body
        eval_nodes = qlq__cait[indzj__cvjw + 1:]
        redvars = []
        var_to_redvar = {}
        if parfor:
            redvars, var_to_redvar = get_parfor_reductions(parfor, parfor.
                params, pm.calltypes)
        func.ncols_pre_shuffle = len(redvars)
        func.ncols_post_shuffle = len(redvars) + 1
        func.n_redvars = len(redvars)
        reduce_vars = [0] * len(redvars)
        for vbmp__juw in init_nodes:
            if is_assign(vbmp__juw) and vbmp__juw.target.name in redvars:
                ind = redvars.index(vbmp__juw.target.name)
                reduce_vars[ind] = vbmp__juw.target
        var_types = [pm.typemap[v] for v in redvars]
        vxn__bkt = gen_combine_func(f_ir, parfor, redvars, var_to_redvar,
            var_types, arr_var, pm, self.typingctx, self.targetctx)
        init_nodes = _mv_read_only_init_vars(init_nodes, parfor, eval_nodes)
        oqho__xujfl = gen_update_func(parfor, redvars, var_to_redvar,
            var_types, arr_var, in_col_typ, pm, self.typingctx, self.targetctx)
        fpet__izjl = gen_eval_func(f_ir, eval_nodes, reduce_vars, var_types,
            pm, self.typingctx, self.targetctx)
        self.all_reduce_vars += reduce_vars
        self.all_vartypes += var_types
        self.all_init_nodes += init_nodes
        self.all_eval_funcs.append(fpet__izjl)
        self.all_update_funcs.append(oqho__xujfl)
        self.all_combine_funcs.append(vxn__bkt)
        self.curr_offset += len(redvars)
        self.redvar_offsets.append(self.curr_offset)

    def gen_all_func(self):
        if len(self.all_update_funcs) == 0:
            return None
        self.all_vartypes = self.all_vartypes * len(self.pivot_values
            ) if self.pivot_values is not None else self.all_vartypes
        self.all_reduce_vars = self.all_reduce_vars * len(self.pivot_values
            ) if self.pivot_values is not None else self.all_reduce_vars
        cobq__sgxc = gen_init_func(self.all_init_nodes, self.
            all_reduce_vars, self.all_vartypes, self.typingctx, self.targetctx)
        moda__qkabi = gen_all_update_func(self.all_update_funcs, self.
            all_vartypes, self.in_col_types, self.redvar_offsets, self.
            typingctx, self.targetctx, self.pivot_typ, self.pivot_values,
            self.is_crosstab)
        yeghv__txy = gen_all_combine_func(self.all_combine_funcs, self.
            all_vartypes, self.redvar_offsets, self.typingctx, self.
            targetctx, self.pivot_typ, self.pivot_values)
        qcp__eif = gen_all_eval_func(self.all_eval_funcs, self.all_vartypes,
            self.redvar_offsets, self.out_col_types, self.typingctx, self.
            targetctx, self.pivot_values)
        return self.all_vartypes, cobq__sgxc, moda__qkabi, yeghv__txy, qcp__eif


class GeneralUDFGenerator(object):

    def __init__(self):
        self.funcs = []

    def add_udf(self, func):
        self.funcs.append(bodo.jit(distributed=False)(func))
        func.ncols_pre_shuffle = 1
        func.ncols_post_shuffle = 1
        func.n_redvars = 0

    def gen_all_func(self):
        if len(self.funcs) > 0:
            return self.funcs
        else:
            return None


def get_udf_func_struct(agg_func, input_has_index, in_col_types,
    out_col_types, typingctx, targetctx, pivot_typ, pivot_values, is_crosstab):
    if is_crosstab and len(in_col_types) == 0:
        in_col_types = [types.Array(types.intp, 1, 'C')]
    mtir__wrk = []
    for t, aaq__qpzs in zip(in_col_types, agg_func):
        mtir__wrk.append((t, aaq__qpzs))
    rlab__bnit = RegularUDFGenerator(in_col_types, out_col_types, pivot_typ,
        pivot_values, is_crosstab, typingctx, targetctx)
    kkgb__ndnn = GeneralUDFGenerator()
    for in_col_typ, func in mtir__wrk:
        if func.ftype not in ('udf', 'gen_udf'):
            continue
        try:
            rlab__bnit.add_udf(in_col_typ, func)
        except:
            kkgb__ndnn.add_udf(func)
            func.ftype = 'gen_udf'
    regular_udf_funcs = rlab__bnit.gen_all_func()
    general_udf_funcs = kkgb__ndnn.gen_all_func()
    if regular_udf_funcs is not None or general_udf_funcs is not None:
        return AggUDFStruct(regular_udf_funcs, general_udf_funcs)
    else:
        return None


def _mv_read_only_init_vars(init_nodes, parfor, eval_nodes):
    if not parfor:
        return init_nodes
    tboww__drqet = compute_use_defs(parfor.loop_body)
    qcz__lhzxg = set()
    for vmrwn__cmu in tboww__drqet.usemap.values():
        qcz__lhzxg |= vmrwn__cmu
    jejq__ghkg = set()
    for vmrwn__cmu in tboww__drqet.defmap.values():
        jejq__ghkg |= vmrwn__cmu
    spcwd__ryvy = ir.Block(ir.Scope(None, parfor.loc), parfor.loc)
    spcwd__ryvy.body = eval_nodes
    lwq__klk = compute_use_defs({(0): spcwd__ryvy})
    zkxmg__lxb = lwq__klk.usemap[0]
    arjr__sfnj = set()
    ioxup__odcd = []
    hmkan__oow = []
    for vbmp__juw in reversed(init_nodes):
        vpyn__perd = {v.name for v in vbmp__juw.list_vars()}
        if is_assign(vbmp__juw):
            v = vbmp__juw.target.name
            vpyn__perd.remove(v)
            if (v in qcz__lhzxg and v not in arjr__sfnj and v not in
                zkxmg__lxb and v not in jejq__ghkg):
                hmkan__oow.append(vbmp__juw)
                qcz__lhzxg |= vpyn__perd
                jejq__ghkg.add(v)
                continue
        arjr__sfnj |= vpyn__perd
        ioxup__odcd.append(vbmp__juw)
    hmkan__oow.reverse()
    ioxup__odcd.reverse()
    zzj__yittx = min(parfor.loop_body.keys())
    hod__ahqov = parfor.loop_body[zzj__yittx]
    hod__ahqov.body = hmkan__oow + hod__ahqov.body
    return ioxup__odcd


def gen_init_func(init_nodes, reduce_vars, var_types, typingctx, targetctx):
    ahm__jxhb = (numba.parfors.parfor.max_checker, numba.parfors.parfor.
        min_checker, numba.parfors.parfor.argmax_checker, numba.parfors.
        parfor.argmin_checker)
    lcwel__sze = set()
    qwkrw__mqd = []
    for vbmp__juw in init_nodes:
        if is_assign(vbmp__juw) and isinstance(vbmp__juw.value, ir.Global
            ) and isinstance(vbmp__juw.value.value, pytypes.FunctionType
            ) and vbmp__juw.value.value in ahm__jxhb:
            lcwel__sze.add(vbmp__juw.target.name)
        elif is_call_assign(vbmp__juw
            ) and vbmp__juw.value.func.name in lcwel__sze:
            pass
        else:
            qwkrw__mqd.append(vbmp__juw)
    init_nodes = qwkrw__mqd
    jetp__orcx = types.Tuple(var_types)
    kydlj__xoyg = lambda : None
    f_ir = compile_to_numba_ir(kydlj__xoyg, {})
    block = list(f_ir.blocks.values())[0]
    loc = block.loc
    iohxk__wxv = ir.Var(block.scope, mk_unique_var('init_tup'), loc)
    kmyj__kscr = ir.Assign(ir.Expr.build_tuple(reduce_vars, loc),
        iohxk__wxv, loc)
    block.body = block.body[-2:]
    block.body = init_nodes + [kmyj__kscr] + block.body
    block.body[-2].value.value = iohxk__wxv
    mfgr__siayv = compiler.compile_ir(typingctx, targetctx, f_ir, (),
        jetp__orcx, compiler.DEFAULT_FLAGS, {})
    from numba.core.target_extension import cpu_target
    zyob__kmgt = numba.core.target_extension.dispatcher_registry[cpu_target](
        kydlj__xoyg)
    zyob__kmgt.add_overload(mfgr__siayv)
    return zyob__kmgt


def gen_all_update_func(update_funcs, reduce_var_types, in_col_types,
    redvar_offsets, typingctx, targetctx, pivot_typ, pivot_values, is_crosstab
    ):
    mnxxs__yqzv = len(update_funcs)
    htd__uis = len(in_col_types)
    if pivot_values is not None:
        assert htd__uis == 1
    uxeq__cuh = (
        'def update_all_f(redvar_arrs, data_in, w_ind, i, pivot_arr):\n')
    if pivot_values is not None:
        yfr__ikm = redvar_offsets[htd__uis]
        uxeq__cuh += '  pv = pivot_arr[i]\n'
        for vtfn__hokn, wvro__ultrx in enumerate(pivot_values):
            qyfpe__iado = 'el' if vtfn__hokn != 0 else ''
            uxeq__cuh += "  {}if pv == '{}':\n".format(qyfpe__iado, wvro__ultrx
                )
            ydwxy__glxyt = yfr__ikm * vtfn__hokn
            mpus__izw = ', '.join(['redvar_arrs[{}][w_ind]'.format(i) for i in
                range(ydwxy__glxyt + redvar_offsets[0], ydwxy__glxyt +
                redvar_offsets[1])])
            knxi__vvozi = 'data_in[0][i]'
            if is_crosstab:
                knxi__vvozi = '0'
            uxeq__cuh += '    {} = update_vars_0({}, {})\n'.format(mpus__izw,
                mpus__izw, knxi__vvozi)
    else:
        for vtfn__hokn in range(mnxxs__yqzv):
            mpus__izw = ', '.join(['redvar_arrs[{}][w_ind]'.format(i) for i in
                range(redvar_offsets[vtfn__hokn], redvar_offsets[vtfn__hokn +
                1])])
            if mpus__izw:
                uxeq__cuh += ('  {} = update_vars_{}({},  data_in[{}][i])\n'
                    .format(mpus__izw, vtfn__hokn, mpus__izw, 0 if htd__uis ==
                    1 else vtfn__hokn))
    uxeq__cuh += '  return\n'
    gtnop__fsn = {}
    for i, aaq__qpzs in enumerate(update_funcs):
        gtnop__fsn['update_vars_{}'.format(i)] = aaq__qpzs
    oooax__bmx = {}
    exec(uxeq__cuh, gtnop__fsn, oooax__bmx)
    zfk__sjkxa = oooax__bmx['update_all_f']
    return numba.njit(no_cpython_wrapper=True)(zfk__sjkxa)


def gen_all_combine_func(combine_funcs, reduce_var_types, redvar_offsets,
    typingctx, targetctx, pivot_typ, pivot_values):
    wcsvr__ixwzw = types.Tuple([types.Array(t, 1, 'C') for t in
        reduce_var_types])
    arg_typs = wcsvr__ixwzw, wcsvr__ixwzw, types.intp, types.intp, pivot_typ
    iuy__wyfq = len(redvar_offsets) - 1
    yfr__ikm = redvar_offsets[iuy__wyfq]
    uxeq__cuh = (
        'def combine_all_f(redvar_arrs, recv_arrs, w_ind, i, pivot_arr):\n')
    if pivot_values is not None:
        assert iuy__wyfq == 1
        for ivpec__ryi in range(len(pivot_values)):
            ydwxy__glxyt = yfr__ikm * ivpec__ryi
            mpus__izw = ', '.join(['redvar_arrs[{}][w_ind]'.format(i) for i in
                range(ydwxy__glxyt + redvar_offsets[0], ydwxy__glxyt +
                redvar_offsets[1])])
            lir__nhvy = ', '.join(['recv_arrs[{}][i]'.format(i) for i in
                range(ydwxy__glxyt + redvar_offsets[0], ydwxy__glxyt +
                redvar_offsets[1])])
            uxeq__cuh += '  {} = combine_vars_0({}, {})\n'.format(mpus__izw,
                mpus__izw, lir__nhvy)
    else:
        for vtfn__hokn in range(iuy__wyfq):
            mpus__izw = ', '.join(['redvar_arrs[{}][w_ind]'.format(i) for i in
                range(redvar_offsets[vtfn__hokn], redvar_offsets[vtfn__hokn +
                1])])
            lir__nhvy = ', '.join(['recv_arrs[{}][i]'.format(i) for i in
                range(redvar_offsets[vtfn__hokn], redvar_offsets[vtfn__hokn +
                1])])
            if lir__nhvy:
                uxeq__cuh += '  {} = combine_vars_{}({}, {})\n'.format(
                    mpus__izw, vtfn__hokn, mpus__izw, lir__nhvy)
    uxeq__cuh += '  return\n'
    gtnop__fsn = {}
    for i, aaq__qpzs in enumerate(combine_funcs):
        gtnop__fsn['combine_vars_{}'.format(i)] = aaq__qpzs
    oooax__bmx = {}
    exec(uxeq__cuh, gtnop__fsn, oooax__bmx)
    ntqm__wyc = oooax__bmx['combine_all_f']
    f_ir = compile_to_numba_ir(ntqm__wyc, gtnop__fsn)
    yeghv__txy = compiler.compile_ir(typingctx, targetctx, f_ir, arg_typs,
        types.none, compiler.DEFAULT_FLAGS, {})
    from numba.core.target_extension import cpu_target
    zyob__kmgt = numba.core.target_extension.dispatcher_registry[cpu_target](
        ntqm__wyc)
    zyob__kmgt.add_overload(yeghv__txy)
    return zyob__kmgt


def gen_all_eval_func(eval_funcs, reduce_var_types, redvar_offsets,
    out_col_typs, typingctx, targetctx, pivot_values):
    wcsvr__ixwzw = types.Tuple([types.Array(t, 1, 'C') for t in
        reduce_var_types])
    out_col_typs = types.Tuple(out_col_typs)
    iuy__wyfq = len(redvar_offsets) - 1
    yfr__ikm = redvar_offsets[iuy__wyfq]
    uxeq__cuh = 'def eval_all_f(redvar_arrs, out_arrs, j):\n'
    if pivot_values is not None:
        assert iuy__wyfq == 1
        for vtfn__hokn in range(len(pivot_values)):
            ydwxy__glxyt = yfr__ikm * vtfn__hokn
            mpus__izw = ', '.join(['redvar_arrs[{}][j]'.format(i) for i in
                range(ydwxy__glxyt + redvar_offsets[0], ydwxy__glxyt +
                redvar_offsets[1])])
            uxeq__cuh += '  out_arrs[{}][j] = eval_vars_0({})\n'.format(
                vtfn__hokn, mpus__izw)
    else:
        for vtfn__hokn in range(iuy__wyfq):
            mpus__izw = ', '.join(['redvar_arrs[{}][j]'.format(i) for i in
                range(redvar_offsets[vtfn__hokn], redvar_offsets[vtfn__hokn +
                1])])
            uxeq__cuh += '  out_arrs[{}][j] = eval_vars_{}({})\n'.format(
                vtfn__hokn, vtfn__hokn, mpus__izw)
    uxeq__cuh += '  return\n'
    gtnop__fsn = {}
    for i, aaq__qpzs in enumerate(eval_funcs):
        gtnop__fsn['eval_vars_{}'.format(i)] = aaq__qpzs
    oooax__bmx = {}
    exec(uxeq__cuh, gtnop__fsn, oooax__bmx)
    xqa__psw = oooax__bmx['eval_all_f']
    return numba.njit(no_cpython_wrapper=True)(xqa__psw)


def gen_eval_func(f_ir, eval_nodes, reduce_vars, var_types, pm, typingctx,
    targetctx):
    vlqbj__mkqs = len(var_types)
    dqqw__xrw = [f'in{i}' for i in range(vlqbj__mkqs)]
    jetp__orcx = types.unliteral(pm.typemap[eval_nodes[-1].value.name])
    qrae__mbzol = jetp__orcx(0)
    uxeq__cuh = 'def agg_eval({}):\n return _zero\n'.format(', '.join(
        dqqw__xrw))
    oooax__bmx = {}
    exec(uxeq__cuh, {'_zero': qrae__mbzol}, oooax__bmx)
    yxt__pxm = oooax__bmx['agg_eval']
    arg_typs = tuple(var_types)
    f_ir = compile_to_numba_ir(yxt__pxm, {'numba': numba, 'bodo': bodo,
        'np': np, '_zero': qrae__mbzol}, typingctx=typingctx, targetctx=
        targetctx, arg_typs=arg_typs, typemap=pm.typemap, calltypes=pm.
        calltypes)
    block = list(f_ir.blocks.values())[0]
    zkk__jab = []
    for i, v in enumerate(reduce_vars):
        zkk__jab.append(ir.Assign(block.body[i].target, v, v.loc))
        for vehwv__lnb in v.versioned_names:
            zkk__jab.append(ir.Assign(v, ir.Var(v.scope, vehwv__lnb, v.loc),
                v.loc))
    block.body = block.body[:vlqbj__mkqs] + zkk__jab + eval_nodes
    fpet__izjl = compiler.compile_ir(typingctx, targetctx, f_ir, arg_typs,
        jetp__orcx, compiler.DEFAULT_FLAGS, {})
    from numba.core.target_extension import cpu_target
    zyob__kmgt = numba.core.target_extension.dispatcher_registry[cpu_target](
        yxt__pxm)
    zyob__kmgt.add_overload(fpet__izjl)
    return zyob__kmgt


def gen_combine_func(f_ir, parfor, redvars, var_to_redvar, var_types,
    arr_var, pm, typingctx, targetctx):
    if not parfor:
        return numba.njit(lambda : ())
    vlqbj__mkqs = len(redvars)
    vosd__yqoah = [f'v{i}' for i in range(vlqbj__mkqs)]
    dqqw__xrw = [f'in{i}' for i in range(vlqbj__mkqs)]
    uxeq__cuh = 'def agg_combine({}):\n'.format(', '.join(vosd__yqoah +
        dqqw__xrw))
    wou__ipwu = wrap_parfor_blocks(parfor)
    kxals__zsl = find_topo_order(wou__ipwu)
    kxals__zsl = kxals__zsl[1:]
    unwrap_parfor_blocks(parfor)
    fqjeo__fbfx = {}
    wwrn__djvw = []
    for vwnge__wnhwr in kxals__zsl:
        gtcw__wpjxg = parfor.loop_body[vwnge__wnhwr]
        for vbmp__juw in gtcw__wpjxg.body:
            if is_call_assign(vbmp__juw) and guard(find_callname, f_ir,
                vbmp__juw.value) == ('__special_combine', 'bodo.ir.aggregate'):
                args = vbmp__juw.value.args
                mdd__kkxw = []
                rblj__qrnrq = []
                for v in args[:-1]:
                    ind = redvars.index(v.name)
                    wwrn__djvw.append(ind)
                    mdd__kkxw.append('v{}'.format(ind))
                    rblj__qrnrq.append('in{}'.format(ind))
                lik__ujfsu = '__special_combine__{}'.format(len(fqjeo__fbfx))
                uxeq__cuh += '    ({},) = {}({})\n'.format(', '.join(
                    mdd__kkxw), lik__ujfsu, ', '.join(mdd__kkxw + rblj__qrnrq))
                fmewg__lts = ir.Expr.call(args[-1], [], (), gtcw__wpjxg.loc)
                tripy__slou = guard(find_callname, f_ir, fmewg__lts)
                assert tripy__slou == ('_var_combine', 'bodo.ir.aggregate')
                tripy__slou = bodo.ir.aggregate._var_combine
                fqjeo__fbfx[lik__ujfsu] = tripy__slou
            if is_assign(vbmp__juw) and vbmp__juw.target.name in redvars:
                ttd__oqyi = vbmp__juw.target.name
                ind = redvars.index(ttd__oqyi)
                if ind in wwrn__djvw:
                    continue
                if len(f_ir._definitions[ttd__oqyi]) == 2:
                    var_def = f_ir._definitions[ttd__oqyi][0]
                    uxeq__cuh += _match_reduce_def(var_def, f_ir, ind)
                    var_def = f_ir._definitions[ttd__oqyi][1]
                    uxeq__cuh += _match_reduce_def(var_def, f_ir, ind)
    uxeq__cuh += '    return {}'.format(', '.join(['v{}'.format(i) for i in
        range(vlqbj__mkqs)]))
    oooax__bmx = {}
    exec(uxeq__cuh, {}, oooax__bmx)
    xotdh__jwl = oooax__bmx['agg_combine']
    arg_typs = tuple(2 * var_types)
    gtnop__fsn = {'numba': numba, 'bodo': bodo, 'np': np}
    gtnop__fsn.update(fqjeo__fbfx)
    f_ir = compile_to_numba_ir(xotdh__jwl, gtnop__fsn, typingctx=typingctx,
        targetctx=targetctx, arg_typs=arg_typs, typemap=pm.typemap,
        calltypes=pm.calltypes)
    block = list(f_ir.blocks.values())[0]
    jetp__orcx = pm.typemap[block.body[-1].value.name]
    vxn__bkt = compiler.compile_ir(typingctx, targetctx, f_ir, arg_typs,
        jetp__orcx, compiler.DEFAULT_FLAGS, {})
    from numba.core.target_extension import cpu_target
    zyob__kmgt = numba.core.target_extension.dispatcher_registry[cpu_target](
        xotdh__jwl)
    zyob__kmgt.add_overload(vxn__bkt)
    return zyob__kmgt


def _match_reduce_def(var_def, f_ir, ind):
    uxeq__cuh = ''
    while isinstance(var_def, ir.Var):
        var_def = guard(get_definition, f_ir, var_def)
    if isinstance(var_def, ir.Expr
        ) and var_def.op == 'inplace_binop' and var_def.fn in ('+=',
        operator.iadd):
        uxeq__cuh = '    v{} += in{}\n'.format(ind, ind)
    if isinstance(var_def, ir.Expr) and var_def.op == 'call':
        cyett__hrn = guard(find_callname, f_ir, var_def)
        if cyett__hrn == ('min', 'builtins'):
            uxeq__cuh = '    v{} = min(v{}, in{})\n'.format(ind, ind, ind)
        if cyett__hrn == ('max', 'builtins'):
            uxeq__cuh = '    v{} = max(v{}, in{})\n'.format(ind, ind, ind)
    return uxeq__cuh


def gen_update_func(parfor, redvars, var_to_redvar, var_types, arr_var,
    in_col_typ, pm, typingctx, targetctx):
    if not parfor:
        return numba.njit(lambda A: ())
    vlqbj__mkqs = len(redvars)
    izrob__fnga = 1
    ckrm__ipz = []
    for i in range(izrob__fnga):
        ouvmk__lzpd = ir.Var(arr_var.scope, f'$input{i}', arr_var.loc)
        ckrm__ipz.append(ouvmk__lzpd)
    kkv__flk = parfor.loop_nests[0].index_variable
    ewcla__mzzpz = [0] * vlqbj__mkqs
    for gtcw__wpjxg in parfor.loop_body.values():
        aeub__mqr = []
        for vbmp__juw in gtcw__wpjxg.body:
            if is_var_assign(vbmp__juw
                ) and vbmp__juw.value.name == kkv__flk.name:
                continue
            if is_getitem(vbmp__juw
                ) and vbmp__juw.value.value.name == arr_var.name:
                vbmp__juw.value = ckrm__ipz[0]
            if is_call_assign(vbmp__juw) and guard(find_callname, pm.
                func_ir, vbmp__juw.value) == ('isna', 'bodo.libs.array_kernels'
                ) and vbmp__juw.value.args[0].name == arr_var.name:
                vbmp__juw.value = ir.Const(False, vbmp__juw.target.loc)
            if is_assign(vbmp__juw) and vbmp__juw.target.name in redvars:
                ind = redvars.index(vbmp__juw.target.name)
                ewcla__mzzpz[ind] = vbmp__juw.target
            aeub__mqr.append(vbmp__juw)
        gtcw__wpjxg.body = aeub__mqr
    vosd__yqoah = ['v{}'.format(i) for i in range(vlqbj__mkqs)]
    dqqw__xrw = ['in{}'.format(i) for i in range(izrob__fnga)]
    uxeq__cuh = 'def agg_update({}):\n'.format(', '.join(vosd__yqoah +
        dqqw__xrw))
    uxeq__cuh += '    __update_redvars()\n'
    uxeq__cuh += '    return {}'.format(', '.join(['v{}'.format(i) for i in
        range(vlqbj__mkqs)]))
    oooax__bmx = {}
    exec(uxeq__cuh, {}, oooax__bmx)
    xnd__bmgy = oooax__bmx['agg_update']
    arg_typs = tuple(var_types + [in_col_typ.dtype] * izrob__fnga)
    f_ir = compile_to_numba_ir(xnd__bmgy, {'__update_redvars':
        __update_redvars}, typingctx=typingctx, targetctx=targetctx,
        arg_typs=arg_typs, typemap=pm.typemap, calltypes=pm.calltypes)
    f_ir._definitions = build_definitions(f_ir.blocks)
    gdwc__jgwag = f_ir.blocks.popitem()[1].body
    jetp__orcx = pm.typemap[gdwc__jgwag[-1].value.name]
    wou__ipwu = wrap_parfor_blocks(parfor)
    kxals__zsl = find_topo_order(wou__ipwu)
    kxals__zsl = kxals__zsl[1:]
    unwrap_parfor_blocks(parfor)
    f_ir.blocks = parfor.loop_body
    hod__ahqov = f_ir.blocks[kxals__zsl[0]]
    bgfg__khd = f_ir.blocks[kxals__zsl[-1]]
    jobf__dsbu = gdwc__jgwag[:vlqbj__mkqs + izrob__fnga]
    if vlqbj__mkqs > 1:
        kct__zooql = gdwc__jgwag[-3:]
        assert is_assign(kct__zooql[0]) and isinstance(kct__zooql[0].value,
            ir.Expr) and kct__zooql[0].value.op == 'build_tuple'
    else:
        kct__zooql = gdwc__jgwag[-2:]
    for i in range(vlqbj__mkqs):
        nzob__plnxw = gdwc__jgwag[i].target
        vrx__zwd = ir.Assign(nzob__plnxw, ewcla__mzzpz[i], nzob__plnxw.loc)
        jobf__dsbu.append(vrx__zwd)
    for i in range(vlqbj__mkqs, vlqbj__mkqs + izrob__fnga):
        nzob__plnxw = gdwc__jgwag[i].target
        vrx__zwd = ir.Assign(nzob__plnxw, ckrm__ipz[i - vlqbj__mkqs],
            nzob__plnxw.loc)
        jobf__dsbu.append(vrx__zwd)
    hod__ahqov.body = jobf__dsbu + hod__ahqov.body
    wftuh__mxqtb = []
    for i in range(vlqbj__mkqs):
        nzob__plnxw = gdwc__jgwag[i].target
        vrx__zwd = ir.Assign(ewcla__mzzpz[i], nzob__plnxw, nzob__plnxw.loc)
        wftuh__mxqtb.append(vrx__zwd)
    bgfg__khd.body += wftuh__mxqtb + kct__zooql
    fhnn__hang = compiler.compile_ir(typingctx, targetctx, f_ir, arg_typs,
        jetp__orcx, compiler.DEFAULT_FLAGS, {})
    from numba.core.target_extension import cpu_target
    zyob__kmgt = numba.core.target_extension.dispatcher_registry[cpu_target](
        xnd__bmgy)
    zyob__kmgt.add_overload(fhnn__hang)
    return zyob__kmgt


def _rm_arg_agg_block(block, typemap):
    qlq__cait = []
    arr_var = None
    for i, vbmp__juw in enumerate(block.body):
        if is_assign(vbmp__juw) and isinstance(vbmp__juw.value, ir.Arg):
            arr_var = vbmp__juw.target
            edx__wmdv = typemap[arr_var.name]
            if not isinstance(edx__wmdv, types.ArrayCompatible):
                qlq__cait += block.body[i + 1:]
                break
            dfvl__lgxju = block.body[i + 1]
            assert is_assign(dfvl__lgxju) and isinstance(dfvl__lgxju.value,
                ir.Expr
                ) and dfvl__lgxju.value.op == 'getattr' and dfvl__lgxju.value.attr == 'shape' and dfvl__lgxju.value.value.name == arr_var.name
            vuman__ozof = dfvl__lgxju.target
            xinwc__mihez = block.body[i + 2]
            assert is_assign(xinwc__mihez) and isinstance(xinwc__mihez.
                value, ir.Expr
                ) and xinwc__mihez.value.op == 'static_getitem' and xinwc__mihez.value.value.name == vuman__ozof.name
            qlq__cait += block.body[i + 3:]
            break
        qlq__cait.append(vbmp__juw)
    return qlq__cait, arr_var


def get_parfor_reductions(parfor, parfor_params, calltypes, reduce_varnames
    =None, param_uses=None, var_to_param=None):
    if reduce_varnames is None:
        reduce_varnames = []
    if param_uses is None:
        param_uses = defaultdict(list)
    if var_to_param is None:
        var_to_param = {}
    wou__ipwu = wrap_parfor_blocks(parfor)
    kxals__zsl = find_topo_order(wou__ipwu)
    kxals__zsl = kxals__zsl[1:]
    unwrap_parfor_blocks(parfor)
    for vwnge__wnhwr in reversed(kxals__zsl):
        for vbmp__juw in reversed(parfor.loop_body[vwnge__wnhwr].body):
            if isinstance(vbmp__juw, ir.Assign) and (vbmp__juw.target.name in
                parfor_params or vbmp__juw.target.name in var_to_param):
                wwptl__jncu = vbmp__juw.target.name
                rhs = vbmp__juw.value
                osc__adxr = (wwptl__jncu if wwptl__jncu in parfor_params else
                    var_to_param[wwptl__jncu])
                hydv__jjsl = []
                if isinstance(rhs, ir.Var):
                    hydv__jjsl = [rhs.name]
                elif isinstance(rhs, ir.Expr):
                    hydv__jjsl = [v.name for v in vbmp__juw.value.list_vars()]
                param_uses[osc__adxr].extend(hydv__jjsl)
                for v in hydv__jjsl:
                    var_to_param[v] = osc__adxr
            if isinstance(vbmp__juw, Parfor):
                get_parfor_reductions(vbmp__juw, parfor_params, calltypes,
                    reduce_varnames, param_uses, var_to_param)
    for qrk__eoutw, hydv__jjsl in param_uses.items():
        if qrk__eoutw in hydv__jjsl and qrk__eoutw not in reduce_varnames:
            reduce_varnames.append(qrk__eoutw)
    return reduce_varnames, var_to_param


@numba.extending.register_jitable
def dummy_agg_count(A):
    return len(A)
