"""
Common IR extension functions for connectors such as CSV, Parquet and JSON readers.
"""
from collections import defaultdict
import numba
from numba.core import ir, types
from numba.core.ir_utils import replace_vars_inner, visit_vars_inner
from numba.extending import box, models, register_model
from bodo.hiframes.table import TableType
from bodo.transforms.distributed_analysis import Distribution
from bodo.utils.utils import debug_prints


def connector_array_analysis(node, equiv_set, typemap, array_analysis):
    xuy__kqkmf = []
    assert len(node.out_vars) > 0, 'empty {} in array analysis'.format(node
        .connector_typ)
    if node.connector_typ == 'csv' and node.chunksize is not None:
        return [], []
    khdq__ydxdr = []
    for irwp__uxv in node.out_vars:
        typ = typemap[irwp__uxv.name]
        if typ == types.none:
            continue
        ndcw__pkd = array_analysis._gen_shape_call(equiv_set, irwp__uxv,
            typ.ndim, None, xuy__kqkmf)
        equiv_set.insert_equiv(irwp__uxv, ndcw__pkd)
        khdq__ydxdr.append(ndcw__pkd[0])
        equiv_set.define(irwp__uxv, set())
    if len(khdq__ydxdr) > 1:
        equiv_set.insert_equiv(*khdq__ydxdr)
    return [], xuy__kqkmf


def connector_distributed_analysis(node, array_dists):
    from bodo.ir.sql_ext import SqlReader
    if isinstance(node, SqlReader) and node.limit is not None:
        tvrtf__lhyg = Distribution.OneD_Var
    else:
        tvrtf__lhyg = Distribution.OneD
    for czzr__lyk in node.out_vars:
        if czzr__lyk.name in array_dists:
            tvrtf__lhyg = Distribution(min(tvrtf__lhyg.value, array_dists[
                czzr__lyk.name].value))
    for czzr__lyk in node.out_vars:
        array_dists[czzr__lyk.name] = tvrtf__lhyg


def connector_typeinfer(node, typeinferer):
    if node.connector_typ == 'csv':
        if node.chunksize is not None:
            typeinferer.lock_type(node.out_vars[0].name, node.out_types[0],
                loc=node.loc)
        else:
            typeinferer.lock_type(node.out_vars[0].name, TableType(tuple(
                node.out_types)), loc=node.loc)
            typeinferer.lock_type(node.out_vars[1].name, node.
                index_column_typ, loc=node.loc)
        return
    if node.connector_typ == 'parquet':
        typeinferer.lock_type(node.out_vars[0].name, TableType(tuple(node.
            out_types)), loc=node.loc)
        typeinferer.lock_type(node.out_vars[1].name, node.index_column_type,
            loc=node.loc)
        return
    for irwp__uxv, typ in zip(node.out_vars, node.out_types):
        typeinferer.lock_type(irwp__uxv.name, typ, loc=node.loc)


def visit_vars_connector(node, callback, cbdata):
    if debug_prints():
        print('visiting {} vars for:'.format(node.connector_typ), node)
        print('cbdata: ', sorted(cbdata.items()))
    glxo__kysg = []
    for irwp__uxv in node.out_vars:
        unnn__hjyou = visit_vars_inner(irwp__uxv, callback, cbdata)
        glxo__kysg.append(unnn__hjyou)
    node.out_vars = glxo__kysg
    if node.connector_typ in ('csv', 'parquet', 'json'):
        node.file_name = visit_vars_inner(node.file_name, callback, cbdata)
    if node.connector_typ == 'csv':
        node.nrows = visit_vars_inner(node.nrows, callback, cbdata)
        node.skiprows = visit_vars_inner(node.skiprows, callback, cbdata)
    if node.connector_typ in ('parquet', 'sql') and node.filters:
        for ybb__cvhhy in node.filters:
            for liiy__cyl in range(len(ybb__cvhhy)):
                val = ybb__cvhhy[liiy__cyl]
                ybb__cvhhy[liiy__cyl] = val[0], val[1], visit_vars_inner(val
                    [2], callback, cbdata)


def connector_usedefs(node, use_set=None, def_set=None):
    if use_set is None:
        use_set = set()
    if def_set is None:
        def_set = set()
    def_set.update({czzr__lyk.name for czzr__lyk in node.out_vars})
    if node.connector_typ in ('csv', 'parquet', 'json'):
        use_set.add(node.file_name.name)
    if node.connector_typ == 'csv':
        if isinstance(node.nrows, numba.core.ir.Var):
            use_set.add(node.nrows.name)
        if isinstance(node.skiprows, numba.core.ir.Var):
            use_set.add(node.skiprows.name)
    if node.connector_typ in ('parquet', 'sql') and node.filters:
        for hcl__hpkj in node.filters:
            for czzr__lyk in hcl__hpkj:
                if isinstance(czzr__lyk[2], ir.Var):
                    use_set.add(czzr__lyk[2].name)
    return numba.core.analysis._use_defs_result(usemap=use_set, defmap=def_set)


def get_copies_connector(node, typemap):
    yyifj__gcjb = set(czzr__lyk.name for czzr__lyk in node.out_vars)
    return set(), yyifj__gcjb


def apply_copies_connector(node, var_dict, name_var_table, typemap,
    calltypes, save_copies):
    glxo__kysg = []
    for irwp__uxv in node.out_vars:
        unnn__hjyou = replace_vars_inner(irwp__uxv, var_dict)
        glxo__kysg.append(unnn__hjyou)
    node.out_vars = glxo__kysg
    if node.connector_typ in ('csv', 'parquet', 'json'):
        node.file_name = replace_vars_inner(node.file_name, var_dict)
    if node.connector_typ in ('parquet', 'sql') and node.filters:
        for ybb__cvhhy in node.filters:
            for liiy__cyl in range(len(ybb__cvhhy)):
                val = ybb__cvhhy[liiy__cyl]
                ybb__cvhhy[liiy__cyl] = val[0], val[1], replace_vars_inner(val
                    [2], var_dict)
    if node.connector_typ == 'csv':
        node.nrows = replace_vars_inner(node.nrows, var_dict)
        node.skiprows = replace_vars_inner(node.skiprows, var_dict)


def build_connector_definitions(node, definitions=None):
    if definitions is None:
        definitions = defaultdict(list)
    for irwp__uxv in node.out_vars:
        lxj__tmtch = definitions[irwp__uxv.name]
        if node not in lxj__tmtch:
            lxj__tmtch.append(node)
    return definitions


def generate_filter_map(filters):
    if filters:
        ztyy__kcm = []
        gag__buxi = [czzr__lyk[2] for hcl__hpkj in filters for czzr__lyk in
            hcl__hpkj]
        uhdge__dchz = set()
        for fluo__ujtox in gag__buxi:
            if isinstance(fluo__ujtox, ir.Var):
                if fluo__ujtox.name not in uhdge__dchz:
                    ztyy__kcm.append(fluo__ujtox)
                uhdge__dchz.add(fluo__ujtox.name)
        return {czzr__lyk.name: f'f{liiy__cyl}' for liiy__cyl, czzr__lyk in
            enumerate(ztyy__kcm)}, ztyy__kcm
    else:
        return {}, []


class StreamReaderType(types.Opaque):

    def __init__(self):
        super(StreamReaderType, self).__init__(name='StreamReaderType')


stream_reader_type = StreamReaderType()
register_model(StreamReaderType)(models.OpaqueModel)


@box(StreamReaderType)
def box_stream_reader(typ, val, c):
    c.pyapi.incref(val)
    return val


def trim_extra_used_columns(used_columns, num_columns):
    tea__mtr = len(used_columns)
    for liiy__cyl in range(len(used_columns) - 1, -1, -1):
        if used_columns[liiy__cyl] < num_columns:
            break
        tea__mtr = liiy__cyl
    return used_columns[:tea__mtr]
