from collections import defaultdict
import numba
import numpy as np
import pandas as pd
from mpi4py import MPI
from numba.core import ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, replace_arg_nodes
import bodo
import bodo.ir.connector
from bodo import objmode
from bodo.hiframes.datetime_date_ext import datetime_date_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.hiframes.table import Table, TableType
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import StringArrayType, string_array_type
from bodo.libs.str_ext import string_type
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.transforms.table_column_del_pass import get_live_column_nums_block, ir_extension_table_column_use, remove_dead_column_extensions
from bodo.utils.typing import BodoError
from bodo.utils.utils import check_java_installation
from bodo.utils.utils import sanitize_varname


class CsvReader(ir.Stmt):

    def __init__(self, file_name, df_out, sep, df_colnames, out_vars,
        out_types, usecols, loc, header, compression, nrows, skiprows,
        chunksize, is_skiprows_list, low_memory, escapechar,
        index_column_index=None, index_column_typ=types.none):
        self.connector_typ = 'csv'
        self.file_name = file_name
        self.df_out = df_out
        self.sep = sep
        self.df_colnames = df_colnames
        self.out_vars = out_vars
        self.out_types = out_types
        self.usecols = usecols
        self.loc = loc
        self.skiprows = skiprows
        self.nrows = nrows
        self.header = header
        self.compression = compression
        self.chunksize = chunksize
        self.is_skiprows_list = is_skiprows_list
        self.pd_low_memory = low_memory
        self.escapechar = escapechar
        self.index_column_index = index_column_index
        self.index_column_typ = index_column_typ
        self.type_usecol_offset = list(range(len(usecols)))

    def __repr__(self):
        return (
            '{} = ReadCsv(file={}, col_names={}, types={}, vars={}, nrows={}, skiprows={}, chunksize={}, is_skiprows_list={}, pd_low_memory={}, index_column_index={}, index_colum_typ = {}, type_usecol_offsets={})'
            .format(self.df_out, self.file_name, self.df_colnames, self.
            out_types, self.out_vars, self.nrows, self.skiprows, self.
            chunksize, self.is_skiprows_list, self.pd_low_memory, self.
            index_column_index, self.index_column_typ, self.type_usecol_offset)
            )


def check_node_typing(node, typemap):
    yqp__xlk = typemap[node.file_name.name]
    if types.unliteral(yqp__xlk) != types.unicode_type:
        raise BodoError(
            f"pd.read_csv(): 'filepath_or_buffer' must be a string. Found type: {yqp__xlk}."
            , node.file_name.loc)
    if not isinstance(node.skiprows, ir.Const):
        qxkea__spju = typemap[node.skiprows.name]
        if isinstance(qxkea__spju, types.Dispatcher):
            raise BodoError(
                f"pd.read_csv(): 'skiprows' callable not supported yet.",
                node.file_name.loc)
        elif not isinstance(qxkea__spju, types.Integer) and not (isinstance
            (qxkea__spju, (types.List, types.Tuple)) and isinstance(
            qxkea__spju.dtype, types.Integer)) and not isinstance(qxkea__spju,
            (types.LiteralList, bodo.utils.typing.ListLiteral)):
            raise BodoError(
                f"pd.read_csv(): 'skiprows' must be an integer or list of integers. Found type {qxkea__spju}."
                , loc=node.skiprows.loc)
        elif isinstance(qxkea__spju, (types.List, types.Tuple)):
            node.is_skiprows_list = True
    if not isinstance(node.nrows, ir.Const):
        dha__uqcn = typemap[node.nrows.name]
        if not isinstance(dha__uqcn, types.Integer):
            raise BodoError(
                f"pd.read_csv(): 'nrows' must be an integer. Found type {dha__uqcn}."
                , loc=node.nrows.loc)


import llvmlite.binding as ll
from bodo.io import csv_cpp
ll.add_symbol('csv_file_chunk_reader', csv_cpp.csv_file_chunk_reader)
csv_file_chunk_reader = types.ExternalFunction('csv_file_chunk_reader',
    bodo.ir.connector.stream_reader_type(types.voidptr, types.bool_, types.
    voidptr, types.int64, types.bool_, types.voidptr, types.voidptr, types.
    int64, types.bool_, types.int64, types.bool_))


def remove_dead_csv(csv_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    if csv_node.chunksize is not None:
        jfl__uydb = csv_node.out_vars[0]
        if jfl__uydb.name not in lives:
            return None
    else:
        lbw__use = csv_node.out_vars[0]
        mpd__exy = csv_node.out_vars[1]
        if lbw__use.name not in lives and mpd__exy.name not in lives:
            return None
        elif mpd__exy.name not in lives:
            csv_node.index_column_index = None
            csv_node.index_column_typ = types.none
        elif lbw__use.name not in lives:
            csv_node.usecols = []
            csv_node.out_types = []
            csv_node.type_usecol_offset = []
    return csv_node


def csv_distributed_run(csv_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    parallel = False
    qxkea__spju = types.int64 if isinstance(csv_node.skiprows, ir.Const
        ) else types.unliteral(typemap[csv_node.skiprows.name])
    if csv_node.chunksize is not None:
        if array_dists is not None:
            vmhv__hjl = csv_node.out_vars[0].name
            parallel = array_dists[vmhv__hjl] in (distributed_pass.
                Distribution.OneD, distributed_pass.Distribution.OneD_Var)
        rtzti__qkym = 'def csv_iterator_impl(fname, nrows, skiprows):\n'
        rtzti__qkym += (
            f'    reader = _csv_reader_init(fname, nrows, skiprows)\n')
        rtzti__qkym += (
            f'    iterator = init_csv_iterator(reader, csv_iterator_type)\n')
        efcbm__ilb = {}
        from bodo.io.csv_iterator_ext import init_csv_iterator
        exec(rtzti__qkym, {}, efcbm__ilb)
        uvzba__lff = efcbm__ilb['csv_iterator_impl']
        kil__qny = 'def csv_reader_init(fname, nrows, skiprows):\n'
        kil__qny += _gen_csv_file_reader_init(parallel, csv_node.header,
            csv_node.compression, csv_node.chunksize, csv_node.
            is_skiprows_list, csv_node.pd_low_memory)
        kil__qny += '  return f_reader\n'
        exec(kil__qny, globals(), efcbm__ilb)
        lljh__vrl = efcbm__ilb['csv_reader_init']
        dnqr__wuesl = numba.njit(lljh__vrl)
        compiled_funcs.append(dnqr__wuesl)
        jhyt__hdj = compile_to_numba_ir(uvzba__lff, {'_csv_reader_init':
            dnqr__wuesl, 'init_csv_iterator': init_csv_iterator,
            'csv_iterator_type': typemap[csv_node.out_vars[0].name]},
            typingctx=typingctx, targetctx=targetctx, arg_typs=(string_type,
            types.int64, qxkea__spju), typemap=typemap, calltypes=calltypes
            ).blocks.popitem()[1]
        replace_arg_nodes(jhyt__hdj, [csv_node.file_name, csv_node.nrows,
            csv_node.skiprows])
        jspv__ajnw = jhyt__hdj.body[:-3]
        jspv__ajnw[-1].target = csv_node.out_vars[0]
        return jspv__ajnw
    if array_dists is not None:
        zgazg__yvgtt = csv_node.out_vars[0].name
        parallel = array_dists[zgazg__yvgtt] in (distributed_pass.
            Distribution.OneD, distributed_pass.Distribution.OneD_Var)
        zkt__scij = csv_node.out_vars[1].name
        assert typemap[zkt__scij] == types.none or not parallel or array_dists[
            zkt__scij] in (distributed_pass.Distribution.OneD,
            distributed_pass.Distribution.OneD_Var
            ), 'pq data/index parallelization does not match'
    rtzti__qkym = 'def csv_impl(fname, nrows, skiprows):\n'
    rtzti__qkym += (
        f'    (table_val, idx_col) = _csv_reader_py(fname, nrows, skiprows)\n')
    efcbm__ilb = {}
    exec(rtzti__qkym, {}, efcbm__ilb)
    qpim__caf = efcbm__ilb['csv_impl']
    vcti__mmh = csv_node.usecols
    if vcti__mmh:
        vcti__mmh = [csv_node.usecols[int__vpm] for int__vpm in csv_node.
            type_usecol_offset]
    xio__lrjdx = _gen_csv_reader_py(csv_node.df_colnames, csv_node.
        out_types, vcti__mmh, csv_node.type_usecol_offset, csv_node.sep,
        parallel, csv_node.header, csv_node.compression, csv_node.
        is_skiprows_list, csv_node.pd_low_memory, csv_node.escapechar,
        idx_col_index=csv_node.index_column_index, idx_col_typ=csv_node.
        index_column_typ)
    jhyt__hdj = compile_to_numba_ir(qpim__caf, {'_csv_reader_py':
        xio__lrjdx}, typingctx=typingctx, targetctx=targetctx, arg_typs=(
        string_type, types.int64, qxkea__spju), typemap=typemap, calltypes=
        calltypes).blocks.popitem()[1]
    replace_arg_nodes(jhyt__hdj, [csv_node.file_name, csv_node.nrows,
        csv_node.skiprows, csv_node.is_skiprows_list])
    jspv__ajnw = jhyt__hdj.body[:-3]
    jspv__ajnw[-1].target = csv_node.out_vars[1]
    jspv__ajnw[-2].target = csv_node.out_vars[0]
    if csv_node.index_column_index is None:
        jspv__ajnw.pop(-1)
    elif not vcti__mmh:
        jspv__ajnw.pop(-2)
    return jspv__ajnw


def csv_remove_dead_column(csv_node, column_live_map, equiv_vars, typemap):
    if csv_node.chunksize is not None:
        return False
    assert len(csv_node.out_vars) == 2, 'invalid CsvReader node'
    rxt__pst = csv_node.out_vars[0].name
    if isinstance(typemap[rxt__pst], TableType) and csv_node.usecols:
        rdqn__xzwun, wzvou__kxf = get_live_column_nums_block(column_live_map,
            equiv_vars, rxt__pst)
        rdqn__xzwun = bodo.ir.connector.trim_extra_used_columns(rdqn__xzwun,
            len(csv_node.usecols))
        if not wzvou__kxf and not rdqn__xzwun:
            rdqn__xzwun = [0]
        if not wzvou__kxf and len(rdqn__xzwun) != len(csv_node.
            type_usecol_offset):
            csv_node.type_usecol_offset = rdqn__xzwun
            return True
    return False


def csv_table_column_use(csv_node, block_use_map, equiv_vars, typemap):
    return


numba.parfors.array_analysis.array_analysis_extensions[CsvReader
    ] = bodo.ir.connector.connector_array_analysis
distributed_analysis.distributed_analysis_extensions[CsvReader
    ] = bodo.ir.connector.connector_distributed_analysis
typeinfer.typeinfer_extensions[CsvReader
    ] = bodo.ir.connector.connector_typeinfer
ir_utils.visit_vars_extensions[CsvReader
    ] = bodo.ir.connector.visit_vars_connector
ir_utils.remove_dead_extensions[CsvReader] = remove_dead_csv
numba.core.analysis.ir_extension_usedefs[CsvReader
    ] = bodo.ir.connector.connector_usedefs
ir_utils.copy_propagate_extensions[CsvReader
    ] = bodo.ir.connector.get_copies_connector
ir_utils.apply_copy_propagate_extensions[CsvReader
    ] = bodo.ir.connector.apply_copies_connector
ir_utils.build_defs_extensions[CsvReader
    ] = bodo.ir.connector.build_connector_definitions
distributed_pass.distributed_run_extensions[CsvReader] = csv_distributed_run
remove_dead_column_extensions[CsvReader] = csv_remove_dead_column
ir_extension_table_column_use[CsvReader] = csv_table_column_use


def _get_dtype_str(t):
    xhnbm__wtbge = t.dtype
    if isinstance(xhnbm__wtbge, PDCategoricalDtype):
        nphm__agty = CategoricalArrayType(xhnbm__wtbge)
        rvg__uqy = 'CategoricalArrayType' + str(ir_utils.next_label())
        setattr(types, rvg__uqy, nphm__agty)
        return rvg__uqy
    if xhnbm__wtbge == types.NPDatetime('ns'):
        xhnbm__wtbge = 'NPDatetime("ns")'
    if t == string_array_type:
        types.string_array_type = string_array_type
        return 'string_array_type'
    if isinstance(t, IntegerArrayType):
        lxco__vyx = 'int_arr_{}'.format(xhnbm__wtbge)
        setattr(types, lxco__vyx, t)
        return lxco__vyx
    if t == boolean_array:
        types.boolean_array = boolean_array
        return 'boolean_array'
    if xhnbm__wtbge == types.bool_:
        xhnbm__wtbge = 'bool_'
    if xhnbm__wtbge == datetime_date_type:
        return 'datetime_date_array_type'
    if isinstance(t, ArrayItemArrayType) and isinstance(xhnbm__wtbge, (
        StringArrayType, ArrayItemArrayType)):
        vsjem__vfch = f'ArrayItemArrayType{str(ir_utils.next_label())}'
        setattr(types, vsjem__vfch, t)
        return vsjem__vfch
    return '{}[::1]'.format(xhnbm__wtbge)


def _get_pd_dtype_str(t):
    xhnbm__wtbge = t.dtype
    if isinstance(xhnbm__wtbge, PDCategoricalDtype):
        return 'pd.CategoricalDtype({})'.format(xhnbm__wtbge.categories)
    if xhnbm__wtbge == types.NPDatetime('ns'):
        return 'str'
    if t == string_array_type:
        return 'str'
    if isinstance(t, IntegerArrayType):
        return '"{}Int{}"'.format('' if xhnbm__wtbge.signed else 'U',
            xhnbm__wtbge.bitwidth)
    if t == boolean_array:
        return 'np.bool_'
    if isinstance(t, ArrayItemArrayType) and isinstance(xhnbm__wtbge, (
        StringArrayType, ArrayItemArrayType)):
        return 'object'
    return 'np.{}'.format(xhnbm__wtbge)


compiled_funcs = []


@numba.njit
def check_nrows_skiprows_value(nrows, skiprows):
    if nrows < -1:
        raise ValueError('pd.read_csv: nrows must be integer >= 0.')
    if skiprows[0] < 0:
        raise ValueError('pd.read_csv: skiprows must be integer >= 0.')


def astype(df, typemap, parallel):
    qsyqg__rtwz = ''
    from collections import defaultdict
    iwzy__ikjbt = defaultdict(list)
    for vjjy__zye, gvn__soe in typemap.items():
        iwzy__ikjbt[gvn__soe].append(vjjy__zye)
    tcbh__atnu = df.columns.to_list()
    hss__xlh = []
    for gvn__soe, gjc__ztthx in iwzy__ikjbt.items():
        try:
            hss__xlh.append(df.loc[:, gjc__ztthx].astype(gvn__soe, copy=False))
            df = df.drop(gjc__ztthx, axis=1)
        except (ValueError, TypeError) as lvfg__zfelk:
            qsyqg__rtwz = (
                f"Caught the runtime error '{lvfg__zfelk}' on columns {gjc__ztthx}. Consider setting the 'dtype' argument in 'read_csv' or investigate if the data is corrupted."
                )
            break
    hcx__ukxbh = bool(qsyqg__rtwz)
    if parallel:
        biwj__wxnyb = MPI.COMM_WORLD
        hcx__ukxbh = biwj__wxnyb.allreduce(hcx__ukxbh, op=MPI.LOR)
    if hcx__ukxbh:
        twjhz__qdpa = 'pd.read_csv(): Bodo could not infer dtypes correctly.'
        if qsyqg__rtwz:
            raise TypeError(f'{twjhz__qdpa}\n{qsyqg__rtwz}')
        else:
            raise TypeError(
                f'{twjhz__qdpa}\nPlease refer to errors on other ranks.')
    df = pd.concat(hss__xlh + [df], axis=1)
    zqayw__ked = df.loc[:, tcbh__atnu]
    return zqayw__ked


def _gen_csv_file_reader_init(parallel, header, compression, chunksize,
    is_skiprows_list, pd_low_memory):
    kche__vqlv = header == 0
    if compression is None:
        compression = 'uncompressed'
    if is_skiprows_list:
        rtzti__qkym = '  skiprows = sorted(set(skiprows))\n'
    else:
        rtzti__qkym = '  skiprows = [skiprows]\n'
    rtzti__qkym += '  skiprows_list_len = len(skiprows)\n'
    rtzti__qkym += '  check_nrows_skiprows_value(nrows, skiprows)\n'
    rtzti__qkym += '  check_java_installation(fname)\n'
    rtzti__qkym += f"""  bucket_region = bodo.io.fs_io.get_s3_bucket_region_njit(fname, parallel={parallel})
"""
    rtzti__qkym += (
        '  f_reader = bodo.ir.csv_ext.csv_file_chunk_reader(bodo.libs.str_ext.unicode_to_utf8(fname), '
        )
    rtzti__qkym += (
        """    {}, bodo.utils.conversion.coerce_to_ndarray(skiprows, scalar_to_arr_len=1).ctypes, nrows, {}, bodo.libs.str_ext.unicode_to_utf8('{}'), bodo.libs.str_ext.unicode_to_utf8(bucket_region), {}, {}, skiprows_list_len, {})
"""
        .format(parallel, kche__vqlv, compression, chunksize,
        is_skiprows_list, pd_low_memory))
    rtzti__qkym += '  bodo.utils.utils.check_and_propagate_cpp_exception()\n'
    rtzti__qkym += '  if bodo.utils.utils.is_null_pointer(f_reader):\n'
    rtzti__qkym += "      raise FileNotFoundError('File does not exist')\n"
    return rtzti__qkym


def _gen_read_csv_objmode(col_names, sanitized_cnames, col_typs, usecols,
    type_usecol_offset, sep, escapechar, call_id, glbs, parallel,
    check_parallel_runtime, idx_col_index, idx_col_typ):
    wmcv__vhs = [str(grnbg__krl) for int__vpm, grnbg__krl in enumerate(
        usecols) if col_typs[type_usecol_offset[int__vpm]].dtype == types.
        NPDatetime('ns')]
    if idx_col_typ == types.NPDatetime('ns'):
        assert not idx_col_index is None
        wmcv__vhs.append(str(idx_col_index))
    hbdap__wdr = ', '.join(wmcv__vhs)
    wxli__twruj = _gen_parallel_flag_name(sanitized_cnames)
    nyouz__dkb = f"{wxli__twruj}='bool_'" if check_parallel_runtime else ''
    vfkkr__cfuf = [_get_pd_dtype_str(col_typs[type_usecol_offset[int__vpm]]
        ) for int__vpm in range(len(usecols))]
    syy__vxhs = None if idx_col_index is None else _get_pd_dtype_str(
        idx_col_typ)
    pjm__rnye = [grnbg__krl for int__vpm, grnbg__krl in enumerate(usecols) if
        vfkkr__cfuf[int__vpm] == 'str']
    if idx_col_index is not None and syy__vxhs == 'str':
        pjm__rnye.append(idx_col_index)
    adi__xja = np.array(pjm__rnye, dtype=np.int64)
    glbs[f'str_col_nums_{call_id}'] = adi__xja
    rtzti__qkym = f'  str_col_nums_{call_id}_2 = str_col_nums_{call_id}\n'
    qdsk__hyqy = np.array(usecols + ([idx_col_index] if idx_col_index is not
        None else []))
    glbs[f'usecols_arr_{call_id}'] = qdsk__hyqy
    rtzti__qkym += f'  usecols_arr_{call_id}_2 = usecols_arr_{call_id}\n'
    hxta__rxemi = np.array(type_usecol_offset, dtype=np.int64)
    if usecols:
        glbs[f'type_usecols_offsets_arr_{call_id}'] = hxta__rxemi
        rtzti__qkym += f"""  type_usecols_offsets_arr_{call_id}_2 = type_usecols_offsets_arr_{call_id}
"""
    hjrd__nqtp = defaultdict(list)
    for int__vpm, grnbg__krl in enumerate(usecols):
        if vfkkr__cfuf[int__vpm] == 'str':
            continue
        hjrd__nqtp[vfkkr__cfuf[int__vpm]].append(grnbg__krl)
    if idx_col_index is not None and syy__vxhs != 'str':
        hjrd__nqtp[syy__vxhs].append(idx_col_index)
    for int__vpm, mekn__sfl in enumerate(hjrd__nqtp.values()):
        glbs[f't_arr_{int__vpm}_{call_id}'] = np.asarray(mekn__sfl)
        rtzti__qkym += (
            f'  t_arr_{int__vpm}_{call_id}_2 = t_arr_{int__vpm}_{call_id}\n')
    if idx_col_index != None:
        rtzti__qkym += f"""  with objmode(T=table_type_{call_id}, idx_arr=idx_array_typ, {nyouz__dkb}):
"""
    else:
        rtzti__qkym += (
            f'  with objmode(T=table_type_{call_id}, {nyouz__dkb}):\n')
    rtzti__qkym += f'    typemap = {{}}\n'
    for int__vpm, oycw__gkfa in enumerate(hjrd__nqtp.keys()):
        rtzti__qkym += f"""    typemap.update({{i:{oycw__gkfa} for i in t_arr_{int__vpm}_{call_id}_2}})
"""
    rtzti__qkym += '    if f_reader.get_chunk_size() == 0:\n'
    rtzti__qkym += (
        f'      df = pd.DataFrame(columns=usecols_arr_{call_id}_2, dtype=str)\n'
        )
    rtzti__qkym += '    else:\n'
    rtzti__qkym += '      df = pd.read_csv(f_reader,\n'
    rtzti__qkym += '        header=None,\n'
    rtzti__qkym += '        parse_dates=[{}],\n'.format(hbdap__wdr)
    rtzti__qkym += (
        f'        dtype={{i:str for i in str_col_nums_{call_id}_2}},\n')
    rtzti__qkym += f"""        usecols=usecols_arr_{call_id}_2, sep={sep!r}, low_memory=False, escapechar={escapechar!r})
"""
    if check_parallel_runtime:
        rtzti__qkym += f'    {wxli__twruj} = f_reader.is_parallel()\n'
    else:
        rtzti__qkym += f'    {wxli__twruj} = {parallel}\n'
    rtzti__qkym += f'    df = astype(df, typemap, {wxli__twruj})\n'
    if idx_col_index != None:
        ffqvx__byz = sorted(qdsk__hyqy).index(idx_col_index)
        rtzti__qkym += f'    idx_arr = df.iloc[:, {ffqvx__byz}].values\n'
        rtzti__qkym += (
            f'    df.drop(columns=df.columns[{ffqvx__byz}], inplace=True)\n')
    if len(usecols) == 0:
        rtzti__qkym += f'    T = None\n'
    else:
        rtzti__qkym += f'    arrs = []\n'
        rtzti__qkym += f'    for i in range(df.shape[1]):\n'
        rtzti__qkym += f'      arrs.append(df.iloc[:, i].values)\n'
        rtzti__qkym += f"""    T = Table(arrs, type_usecols_offsets_arr_{call_id}_2, {len(col_names)})
"""
    return rtzti__qkym


def _gen_parallel_flag_name(sanitized_cnames):
    wxli__twruj = '_parallel_value'
    while wxli__twruj in sanitized_cnames:
        wxli__twruj = '_' + wxli__twruj
    return wxli__twruj


def _gen_csv_reader_py(col_names, col_typs, usecols, type_usecol_offset,
    sep, parallel, header, compression, is_skiprows_list, pd_low_memory,
    escapechar, idx_col_index=None, idx_col_typ=types.none):
    sanitized_cnames = [sanitize_varname(cacfa__nwhi) for cacfa__nwhi in
        col_names]
    rtzti__qkym = 'def csv_reader_py(fname, nrows, skiprows):\n'
    rtzti__qkym += _gen_csv_file_reader_init(parallel, header, compression,
        -1, is_skiprows_list, pd_low_memory)
    call_id = ir_utils.next_label()
    wvv__ndlv = globals()
    if idx_col_typ != types.none:
        wvv__ndlv[f'idx_array_typ'] = idx_col_typ
    if len(usecols) == 0:
        wvv__ndlv[f'table_type_{call_id}'] = types.none
    else:
        wvv__ndlv[f'table_type_{call_id}'] = TableType(tuple(col_typs))
    rtzti__qkym += _gen_read_csv_objmode(col_names, sanitized_cnames,
        col_typs, usecols, type_usecol_offset, sep, escapechar, call_id,
        wvv__ndlv, parallel=parallel, check_parallel_runtime=False,
        idx_col_index=idx_col_index, idx_col_typ=idx_col_typ)
    if idx_col_index != None:
        rtzti__qkym += '  return (T, idx_arr)\n'
    else:
        rtzti__qkym += '  return (T, None)\n'
    efcbm__ilb = {}
    exec(rtzti__qkym, wvv__ndlv, efcbm__ilb)
    xio__lrjdx = efcbm__ilb['csv_reader_py']
    dnqr__wuesl = numba.njit(xio__lrjdx)
    compiled_funcs.append(dnqr__wuesl)
    return dnqr__wuesl
