"""IR node for the join and merge"""
from collections import defaultdict
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba import generated_jit
from numba.core import cgutils, ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, next_label, replace_arg_nodes, replace_vars_inner, visit_vars_inner
from numba.extending import intrinsic, overload
import bodo
from bodo.libs.array import arr_info_list_to_table, array_to_info, compute_node_partition_by_hash, delete_table, delete_table_decref_arrays, hash_join_table, info_from_table, info_to_array
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.int_arr_ext import IntDtype, IntegerArrayType
from bodo.libs.str_arr_ext import copy_str_arr_slice, cp_str_list_to_array, get_bit_bitmap, get_null_bitmap_ptr, get_str_arr_item_length, get_str_arr_item_ptr, get_utf8_size, getitem_str_offset, num_total_chars, pre_alloc_string_array, set_bit_to, str_copy_ptr, string_array_type, to_list_if_immutable_arr
from bodo.libs.str_ext import string_type
from bodo.libs.timsort import getitem_arr_tup, setitem_arr_tup
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.transforms.distributed_analysis import Distribution
from bodo.utils.shuffle import _get_data_tup, _get_keys_tup, alloc_pre_shuffle_metadata, alltoallv_tup, finalize_shuffle_meta, getitem_arr_tup_single, update_shuffle_meta
from bodo.utils.typing import BodoError, dtype_to_array_type, find_common_np_dtype, is_dtype_nullable, is_nullable_type, to_nullable_type
from bodo.utils.utils import alloc_arr_tup, debug_prints, is_null_pointer
join_gen_cond_cfunc = {}
join_gen_cond_cfunc_addr = {}


@intrinsic
def add_join_gen_cond_cfunc_sym(typingctx, func, sym):

    def codegen(context, builder, signature, args):
        yjc__wlet = func.signature
        wyil__mhrep = lir.FunctionType(lir.IntType(1), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(64), lir
            .IntType(64)])
        enpp__pewlc = cgutils.get_or_insert_function(builder.module,
            wyil__mhrep, sym._literal_value)
        builder.call(enpp__pewlc, [context.get_constant_null(yjc__wlet.args
            [0]), context.get_constant_null(yjc__wlet.args[1]), context.
            get_constant_null(yjc__wlet.args[2]), context.get_constant_null
            (yjc__wlet.args[3]), context.get_constant_null(yjc__wlet.args[4
            ]), context.get_constant_null(yjc__wlet.args[5]), context.
            get_constant(types.int64, 0), context.get_constant(types.int64, 0)]
            )
        context.add_linking_libs([join_gen_cond_cfunc[sym._literal_value].
            _library])
        return
    return types.none(func, sym), codegen


@numba.jit
def get_join_cond_addr(name):
    with numba.objmode(addr='int64'):
        addr = join_gen_cond_cfunc_addr[name]
    return addr


class Join(ir.Stmt):

    def __init__(self, df_out, left_df, right_df, left_keys, right_keys,
        out_data_vars, left_vars, right_vars, how, suffix_x, suffix_y, loc,
        is_left, is_right, is_join, left_index, right_index, indicator,
        is_na_equal, gen_cond_expr):
        self.df_out = df_out
        self.left_df = left_df
        self.right_df = right_df
        self.left_keys = left_keys
        self.right_keys = right_keys
        self.out_data_vars = out_data_vars
        self.left_vars = left_vars
        self.right_vars = right_vars
        self.how = how
        self.suffix_x = suffix_x
        self.suffix_y = suffix_y
        self.loc = loc
        self.is_left = is_left
        self.is_right = is_right
        self.is_join = is_join
        self.left_index = left_index
        self.right_index = right_index
        self.indicator = indicator
        self.is_na_equal = is_na_equal
        self.gen_cond_expr = gen_cond_expr
        self.left_cond_cols = set(amu__bzd for amu__bzd in left_vars.keys() if
            f'(left.{amu__bzd})' in gen_cond_expr)
        self.right_cond_cols = set(amu__bzd for amu__bzd in right_vars.keys
            () if f'(right.{amu__bzd})' in gen_cond_expr)
        eyfp__vozw = set(left_keys) & set(right_keys)
        vcgk__zqo = set(left_vars.keys()) & set(right_vars.keys())
        llyu__mocn = vcgk__zqo - eyfp__vozw
        vect_same_key = []
        n_keys = len(left_keys)
        for bvxgl__pro in range(n_keys):
            sggq__rbg = left_keys[bvxgl__pro]
            tqvm__phf = right_keys[bvxgl__pro]
            vect_same_key.append(sggq__rbg == tqvm__phf)
        self.vect_same_key = vect_same_key
        self.column_origins = {(str(amu__bzd) + suffix_x if amu__bzd in
            llyu__mocn else amu__bzd): ('left', amu__bzd) for amu__bzd in
            left_vars.keys()}
        self.column_origins.update({(str(amu__bzd) + suffix_y if amu__bzd in
            llyu__mocn else amu__bzd): ('right', amu__bzd) for amu__bzd in
            right_vars.keys()})
        if '$_bodo_index_' in llyu__mocn:
            llyu__mocn.remove('$_bodo_index_')
        self.add_suffix = llyu__mocn

    def __repr__(self):
        ttw__edg = ''
        for amu__bzd, ttf__mmsjq in self.out_data_vars.items():
            ttw__edg += "'{}':{}, ".format(amu__bzd, ttf__mmsjq.name)
        waof__rarnm = '{}{{{}}}'.format(self.df_out, ttw__edg)
        iwul__bew = ''
        for amu__bzd, ttf__mmsjq in self.left_vars.items():
            iwul__bew += "'{}':{}, ".format(amu__bzd, ttf__mmsjq.name)
        lii__haoum = '{}{{{}}}'.format(self.left_df, iwul__bew)
        iwul__bew = ''
        for amu__bzd, ttf__mmsjq in self.right_vars.items():
            iwul__bew += "'{}':{}, ".format(amu__bzd, ttf__mmsjq.name)
        wad__kahp = '{}{{{}}}'.format(self.right_df, iwul__bew)
        return 'join [{}={}]: {} , {}, {}'.format(self.left_keys, self.
            right_keys, waof__rarnm, lii__haoum, wad__kahp)


def join_array_analysis(join_node, equiv_set, typemap, array_analysis):
    hpnx__jnto = []
    assert len(join_node.out_data_vars) > 0, 'empty join in array analysis'
    xvoga__ogw = []
    wbr__pxk = list(join_node.left_vars.values())
    for amgwf__nigvl in wbr__pxk:
        euu__rpkou = typemap[amgwf__nigvl.name]
        pad__xxyxy = equiv_set.get_shape(amgwf__nigvl)
        if pad__xxyxy:
            xvoga__ogw.append(pad__xxyxy[0])
    if len(xvoga__ogw) > 1:
        equiv_set.insert_equiv(*xvoga__ogw)
    xvoga__ogw = []
    wbr__pxk = list(join_node.right_vars.values())
    for amgwf__nigvl in wbr__pxk:
        euu__rpkou = typemap[amgwf__nigvl.name]
        pad__xxyxy = equiv_set.get_shape(amgwf__nigvl)
        if pad__xxyxy:
            xvoga__ogw.append(pad__xxyxy[0])
    if len(xvoga__ogw) > 1:
        equiv_set.insert_equiv(*xvoga__ogw)
    xvoga__ogw = []
    for amgwf__nigvl in join_node.out_data_vars.values():
        euu__rpkou = typemap[amgwf__nigvl.name]
        zbj__fpow = array_analysis._gen_shape_call(equiv_set, amgwf__nigvl,
            euu__rpkou.ndim, None, hpnx__jnto)
        equiv_set.insert_equiv(amgwf__nigvl, zbj__fpow)
        xvoga__ogw.append(zbj__fpow[0])
        equiv_set.define(amgwf__nigvl, set())
    if len(xvoga__ogw) > 1:
        equiv_set.insert_equiv(*xvoga__ogw)
    return [], hpnx__jnto


numba.parfors.array_analysis.array_analysis_extensions[Join
    ] = join_array_analysis


def join_distributed_analysis(join_node, array_dists):
    hfu__wjfw = Distribution.OneD
    hif__xoav = Distribution.OneD
    for amgwf__nigvl in join_node.left_vars.values():
        hfu__wjfw = Distribution(min(hfu__wjfw.value, array_dists[
            amgwf__nigvl.name].value))
    for amgwf__nigvl in join_node.right_vars.values():
        hif__xoav = Distribution(min(hif__xoav.value, array_dists[
            amgwf__nigvl.name].value))
    tdrwz__qwrxd = Distribution.OneD_Var
    for amgwf__nigvl in join_node.out_data_vars.values():
        if amgwf__nigvl.name in array_dists:
            tdrwz__qwrxd = Distribution(min(tdrwz__qwrxd.value, array_dists
                [amgwf__nigvl.name].value))
    kxhqn__cai = Distribution(min(tdrwz__qwrxd.value, hfu__wjfw.value))
    wst__oufli = Distribution(min(tdrwz__qwrxd.value, hif__xoav.value))
    tdrwz__qwrxd = Distribution(max(kxhqn__cai.value, wst__oufli.value))
    for amgwf__nigvl in join_node.out_data_vars.values():
        array_dists[amgwf__nigvl.name] = tdrwz__qwrxd
    if tdrwz__qwrxd != Distribution.OneD_Var:
        hfu__wjfw = tdrwz__qwrxd
        hif__xoav = tdrwz__qwrxd
    for amgwf__nigvl in join_node.left_vars.values():
        array_dists[amgwf__nigvl.name] = hfu__wjfw
    for amgwf__nigvl in join_node.right_vars.values():
        array_dists[amgwf__nigvl.name] = hif__xoav
    return


distributed_analysis.distributed_analysis_extensions[Join
    ] = join_distributed_analysis


def join_typeinfer(join_node, typeinferer):
    eyfp__vozw = set(join_node.left_keys) & set(join_node.right_keys)
    vcgk__zqo = set(join_node.left_vars.keys()) & set(join_node.right_vars.
        keys())
    llyu__mocn = vcgk__zqo - eyfp__vozw
    for mag__evmdq, qibo__smmu in join_node.out_data_vars.items():
        if join_node.indicator and mag__evmdq == '_merge':
            continue
        if not mag__evmdq in join_node.column_origins:
            raise BodoError('join(): The variable ' + mag__evmdq +
                ' is absent from the output')
        lxvo__qry = join_node.column_origins[mag__evmdq]
        if lxvo__qry[0] == 'left':
            amgwf__nigvl = join_node.left_vars[lxvo__qry[1]]
        else:
            amgwf__nigvl = join_node.right_vars[lxvo__qry[1]]
        typeinferer.constraints.append(typeinfer.Propagate(dst=qibo__smmu.
            name, src=amgwf__nigvl.name, loc=join_node.loc))
    return


typeinfer.typeinfer_extensions[Join] = join_typeinfer


def visit_vars_join(join_node, callback, cbdata):
    if debug_prints():
        print('visiting join vars for:', join_node)
        print('cbdata: ', sorted(cbdata.items()))
    for cqfa__ktj in list(join_node.left_vars.keys()):
        join_node.left_vars[cqfa__ktj] = visit_vars_inner(join_node.
            left_vars[cqfa__ktj], callback, cbdata)
    for cqfa__ktj in list(join_node.right_vars.keys()):
        join_node.right_vars[cqfa__ktj] = visit_vars_inner(join_node.
            right_vars[cqfa__ktj], callback, cbdata)
    for cqfa__ktj in list(join_node.out_data_vars.keys()):
        join_node.out_data_vars[cqfa__ktj] = visit_vars_inner(join_node.
            out_data_vars[cqfa__ktj], callback, cbdata)


ir_utils.visit_vars_extensions[Join] = visit_vars_join


def remove_dead_join(join_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    vgv__fkykp = []
    oiq__hoaqo = True
    for cqfa__ktj, amgwf__nigvl in join_node.out_data_vars.items():
        if amgwf__nigvl.name in lives:
            oiq__hoaqo = False
            continue
        if cqfa__ktj == '$_bodo_index_':
            continue
        if join_node.indicator and cqfa__ktj == '_merge':
            vgv__fkykp.append('_merge')
            join_node.indicator = False
            continue
        kfhg__oqc, ymlyw__dgsx = join_node.column_origins[cqfa__ktj]
        if (kfhg__oqc == 'left' and ymlyw__dgsx not in join_node.left_keys and
            ymlyw__dgsx not in join_node.left_cond_cols):
            join_node.left_vars.pop(ymlyw__dgsx)
            vgv__fkykp.append(cqfa__ktj)
        if (kfhg__oqc == 'right' and ymlyw__dgsx not in join_node.
            right_keys and ymlyw__dgsx not in join_node.right_cond_cols):
            join_node.right_vars.pop(ymlyw__dgsx)
            vgv__fkykp.append(cqfa__ktj)
    for cname in vgv__fkykp:
        join_node.out_data_vars.pop(cname)
    if oiq__hoaqo:
        return None
    return join_node


ir_utils.remove_dead_extensions[Join] = remove_dead_join


def join_usedefs(join_node, use_set=None, def_set=None):
    if use_set is None:
        use_set = set()
    if def_set is None:
        def_set = set()
    use_set.update({ttf__mmsjq.name for ttf__mmsjq in join_node.left_vars.
        values()})
    use_set.update({ttf__mmsjq.name for ttf__mmsjq in join_node.right_vars.
        values()})
    def_set.update({ttf__mmsjq.name for ttf__mmsjq in join_node.
        out_data_vars.values()})
    return numba.core.analysis._use_defs_result(usemap=use_set, defmap=def_set)


numba.core.analysis.ir_extension_usedefs[Join] = join_usedefs


def get_copies_join(join_node, typemap):
    tab__ypmy = set(ttf__mmsjq.name for ttf__mmsjq in join_node.
        out_data_vars.values())
    return set(), tab__ypmy


ir_utils.copy_propagate_extensions[Join] = get_copies_join


def apply_copies_join(join_node, var_dict, name_var_table, typemap,
    calltypes, save_copies):
    for cqfa__ktj in list(join_node.left_vars.keys()):
        join_node.left_vars[cqfa__ktj] = replace_vars_inner(join_node.
            left_vars[cqfa__ktj], var_dict)
    for cqfa__ktj in list(join_node.right_vars.keys()):
        join_node.right_vars[cqfa__ktj] = replace_vars_inner(join_node.
            right_vars[cqfa__ktj], var_dict)
    for cqfa__ktj in list(join_node.out_data_vars.keys()):
        join_node.out_data_vars[cqfa__ktj] = replace_vars_inner(join_node.
            out_data_vars[cqfa__ktj], var_dict)
    return


ir_utils.apply_copy_propagate_extensions[Join] = apply_copies_join


def build_join_definitions(join_node, definitions=None):
    if definitions is None:
        definitions = defaultdict(list)
    for amgwf__nigvl in join_node.out_data_vars.values():
        definitions[amgwf__nigvl.name].append(join_node)
    return definitions


ir_utils.build_defs_extensions[Join] = build_join_definitions


def join_distributed_run(join_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    left_parallel, right_parallel = False, False
    if array_dists is not None:
        left_parallel, right_parallel = _get_table_parallel_flags(join_node,
            array_dists)
    n_keys = len(join_node.left_keys)
    lvsbc__rbjo = tuple(join_node.left_vars[amu__bzd] for amu__bzd in
        join_node.left_keys)
    arm__dlc = tuple(join_node.right_vars[amu__bzd] for amu__bzd in
        join_node.right_keys)
    trjt__gmlr = tuple(join_node.left_vars.keys())
    qec__uty = tuple(join_node.right_vars.keys())
    gbdn__say = ()
    zdx__rkou = ()
    optional_column = False
    if (join_node.left_index and not join_node.right_index and not
        join_node.is_join):
        lsz__ishn = join_node.right_keys[0]
        if lsz__ishn in trjt__gmlr:
            zdx__rkou = lsz__ishn,
            gbdn__say = join_node.right_vars[lsz__ishn],
            optional_column = True
    if (join_node.right_index and not join_node.left_index and not
        join_node.is_join):
        lsz__ishn = join_node.left_keys[0]
        if lsz__ishn in qec__uty:
            zdx__rkou = lsz__ishn,
            gbdn__say = join_node.left_vars[lsz__ishn],
            optional_column = True
    mnjij__elqt = tuple(join_node.out_data_vars[cname] for cname in zdx__rkou)
    mbry__lmt = tuple(ttf__mmsjq for tks__ojq, ttf__mmsjq in sorted(
        join_node.left_vars.items(), key=lambda a: str(a[0])) if tks__ojq
         not in join_node.left_keys)
    ggny__irw = tuple(ttf__mmsjq for tks__ojq, ttf__mmsjq in sorted(
        join_node.right_vars.items(), key=lambda a: str(a[0])) if tks__ojq
         not in join_node.right_keys)
    ryyc__benp = gbdn__say + lvsbc__rbjo + arm__dlc + mbry__lmt + ggny__irw
    ama__chek = tuple(typemap[ttf__mmsjq.name] for ttf__mmsjq in ryyc__benp)
    ojrtv__yepi = tuple('opti_c' + str(i) for i in range(len(gbdn__say)))
    left_other_names = tuple('t1_c' + str(i) for i in range(len(mbry__lmt)))
    right_other_names = tuple('t2_c' + str(i) for i in range(len(ggny__irw)))
    left_other_types = tuple([typemap[amu__bzd.name] for amu__bzd in mbry__lmt]
        )
    right_other_types = tuple([typemap[amu__bzd.name] for amu__bzd in
        ggny__irw])
    left_key_names = tuple('t1_key' + str(i) for i in range(n_keys))
    right_key_names = tuple('t2_key' + str(i) for i in range(n_keys))
    glbs = {}
    loc = join_node.loc
    func_text = 'def f({}{}, {},{}{}{}):\n'.format('{},'.format(ojrtv__yepi
        [0]) if len(ojrtv__yepi) == 1 else '', ','.join(left_key_names),
        ','.join(right_key_names), ','.join(left_other_names), ',' if len(
        left_other_names) != 0 else '', ','.join(right_other_names))
    left_key_types = tuple(typemap[ttf__mmsjq.name] for ttf__mmsjq in
        lvsbc__rbjo)
    right_key_types = tuple(typemap[ttf__mmsjq.name] for ttf__mmsjq in arm__dlc
        )
    for i in range(n_keys):
        glbs[f'key_type_{i}'] = _match_join_key_types(left_key_types[i],
            right_key_types[i], loc)
    func_text += '    t1_keys = ({},)\n'.format(', '.join(
        f'bodo.utils.utils.astype({left_key_names[i]}, key_type_{i})' for i in
        range(n_keys)))
    func_text += '    t2_keys = ({},)\n'.format(', '.join(
        f'bodo.utils.utils.astype({right_key_names[i]}, key_type_{i})' for
        i in range(n_keys)))
    func_text += '    data_left = ({}{})\n'.format(','.join(
        left_other_names), ',' if len(left_other_names) != 0 else '')
    func_text += '    data_right = ({}{})\n'.format(','.join(
        right_other_names), ',' if len(right_other_names) != 0 else '')
    cfh__vsco = []
    for cname in join_node.left_keys:
        if cname in join_node.add_suffix:
            gcz__zaglp = str(cname) + join_node.suffix_x
        else:
            gcz__zaglp = cname
        assert gcz__zaglp in join_node.out_data_vars
        cfh__vsco.append(join_node.out_data_vars[gcz__zaglp])
    for i, cname in enumerate(join_node.right_keys):
        if not join_node.vect_same_key[i] and not join_node.is_join:
            if cname in join_node.add_suffix:
                gcz__zaglp = str(cname) + join_node.suffix_y
            else:
                gcz__zaglp = cname
            assert gcz__zaglp in join_node.out_data_vars
            cfh__vsco.append(join_node.out_data_vars[gcz__zaglp])

    def _get_out_col_var(cname, is_left):
        if cname in join_node.add_suffix:
            if is_left:
                gcz__zaglp = str(cname) + join_node.suffix_x
            else:
                gcz__zaglp = str(cname) + join_node.suffix_y
        else:
            gcz__zaglp = cname
        return join_node.out_data_vars[gcz__zaglp]
    yum__dfy = mnjij__elqt + tuple(cfh__vsco)
    yum__dfy += tuple(_get_out_col_var(tks__ojq, True) for tks__ojq,
        ttf__mmsjq in sorted(join_node.left_vars.items(), key=lambda a: str
        (a[0])) if tks__ojq not in join_node.left_keys)
    yum__dfy += tuple(_get_out_col_var(tks__ojq, False) for tks__ojq,
        ttf__mmsjq in sorted(join_node.right_vars.items(), key=lambda a:
        str(a[0])) if tks__ojq not in join_node.right_keys)
    if join_node.indicator:
        yum__dfy += _get_out_col_var('_merge', False),
    wbk__hmdag = [('t3_c' + str(i)) for i in range(len(yum__dfy))]
    general_cond_cfunc, left_col_nums, right_col_nums = (
        _gen_general_cond_cfunc(join_node, typemap))
    if join_node.how == 'asof':
        if left_parallel or right_parallel:
            assert left_parallel and right_parallel
            func_text += """    t2_keys, data_right = parallel_asof_comm(t1_keys, t2_keys, data_right)
"""
        func_text += """    out_t1_keys, out_t2_keys, out_data_left, out_data_right = bodo.ir.join.local_merge_asof(t1_keys, t2_keys, data_left, data_right)
"""
    else:
        func_text += _gen_local_hash_join(optional_column, left_key_names,
            right_key_names, left_key_types, right_key_types,
            left_other_names, right_other_names, left_other_types,
            right_other_types, join_node.vect_same_key, join_node.is_left,
            join_node.is_right, join_node.is_join, left_parallel,
            right_parallel, glbs, [typemap[ttf__mmsjq.name] for ttf__mmsjq in
            yum__dfy], join_node.loc, join_node.indicator, join_node.
            is_na_equal, general_cond_cfunc, left_col_nums, right_col_nums)
    if join_node.how == 'asof':
        for i in range(len(left_other_names)):
            func_text += '    left_{} = out_data_left[{}]\n'.format(i, i)
        for i in range(len(right_other_names)):
            func_text += '    right_{} = out_data_right[{}]\n'.format(i, i)
        for i in range(n_keys):
            func_text += f'    t1_keys_{i} = out_t1_keys[{i}]\n'
        for i in range(n_keys):
            func_text += f'    t2_keys_{i} = out_t2_keys[{i}]\n'
    idx = 0
    if optional_column:
        func_text += f'    {wbk__hmdag[idx]} = opti_0\n'
        idx += 1
    for i in range(n_keys):
        func_text += f'    {wbk__hmdag[idx]} = t1_keys_{i}\n'
        idx += 1
    for i in range(n_keys):
        if not join_node.vect_same_key[i] and not join_node.is_join:
            func_text += f'    {wbk__hmdag[idx]} = t2_keys_{i}\n'
            idx += 1
    for i in range(len(left_other_names)):
        func_text += f'    {wbk__hmdag[idx]} = left_{i}\n'
        idx += 1
    for i in range(len(right_other_names)):
        func_text += f'    {wbk__hmdag[idx]} = right_{i}\n'
        idx += 1
    if join_node.indicator:
        func_text += f'    {wbk__hmdag[idx]} = indicator_col\n'
        idx += 1
    wqhck__zscak = {}
    exec(func_text, {}, wqhck__zscak)
    nkgu__gkmi = wqhck__zscak['f']
    glbs.update({'bodo': bodo, 'np': np, 'pd': pd,
        'to_list_if_immutable_arr': to_list_if_immutable_arr,
        'cp_str_list_to_array': cp_str_list_to_array, 'parallel_asof_comm':
        parallel_asof_comm, 'array_to_info': array_to_info,
        'arr_info_list_to_table': arr_info_list_to_table, 'hash_join_table':
        hash_join_table, 'info_from_table': info_from_table,
        'info_to_array': info_to_array, 'delete_table': delete_table,
        'delete_table_decref_arrays': delete_table_decref_arrays,
        'add_join_gen_cond_cfunc_sym': add_join_gen_cond_cfunc_sym,
        'get_join_cond_addr': get_join_cond_addr})
    if general_cond_cfunc:
        glbs.update({'general_cond_cfunc': general_cond_cfunc})
    qpqhq__fqeyc = compile_to_numba_ir(nkgu__gkmi, glbs, typingctx=
        typingctx, targetctx=targetctx, arg_typs=ama__chek, typemap=typemap,
        calltypes=calltypes).blocks.popitem()[1]
    replace_arg_nodes(qpqhq__fqeyc, ryyc__benp)
    prib__jsodj = qpqhq__fqeyc.body[:-3]
    for i in range(len(yum__dfy)):
        prib__jsodj[-len(yum__dfy) + i].target = yum__dfy[i]
    return prib__jsodj


distributed_pass.distributed_run_extensions[Join] = join_distributed_run


def _gen_general_cond_cfunc(join_node, typemap):
    expr = join_node.gen_cond_expr
    if not expr:
        return None, [], []
    cdn__bdoxl = next_label()
    zfpnm__mjej = _get_col_to_ind(join_node.left_keys, join_node.left_vars)
    qtkxp__bzqrj = _get_col_to_ind(join_node.right_keys, join_node.right_vars)
    table_getitem_funcs = {'bodo': bodo, 'numba': numba, 'is_null_pointer':
        is_null_pointer}
    na_check_name = 'NOT_NA'
    func_text = f"""def bodo_join_gen_cond{cdn__bdoxl}(left_table, right_table, left_data1, right_data1, left_null_bitmap, right_null_bitmap, left_ind, right_ind):
"""
    func_text += '  if is_null_pointer(left_table):\n'
    func_text += '    return False\n'
    expr, func_text, left_col_nums = _replace_column_accesses(expr,
        zfpnm__mjej, typemap, join_node.left_vars, table_getitem_funcs,
        func_text, 'left', len(join_node.left_keys), na_check_name)
    expr, func_text, right_col_nums = _replace_column_accesses(expr,
        qtkxp__bzqrj, typemap, join_node.right_vars, table_getitem_funcs,
        func_text, 'right', len(join_node.right_keys), na_check_name)
    func_text += f'  return {expr}'
    wqhck__zscak = {}
    exec(func_text, table_getitem_funcs, wqhck__zscak)
    izfgw__upmij = wqhck__zscak[f'bodo_join_gen_cond{cdn__bdoxl}']
    huciv__jwex = types.bool_(types.voidptr, types.voidptr, types.voidptr,
        types.voidptr, types.voidptr, types.voidptr, types.int64, types.int64)
    bti__icv = numba.cfunc(huciv__jwex, nopython=True)(izfgw__upmij)
    join_gen_cond_cfunc[bti__icv.native_name] = bti__icv
    join_gen_cond_cfunc_addr[bti__icv.native_name] = bti__icv.address
    return bti__icv, left_col_nums, right_col_nums


def _replace_column_accesses(expr, col_to_ind, typemap, col_vars,
    table_getitem_funcs, func_text, table_name, n_keys, na_check_name):
    amf__tlb = []
    for amu__bzd, rod__ivs in col_to_ind.items():
        cname = f'({table_name}.{amu__bzd})'
        if cname not in expr:
            continue
        npjwa__nsa = f'getitem_{table_name}_val_{rod__ivs}'
        uhefb__quziq = f'_bodo_{table_name}_val_{rod__ivs}'
        urb__dlpjy = typemap[col_vars[amu__bzd].name].dtype
        if urb__dlpjy == types.unicode_type:
            func_text += f"""  {uhefb__quziq}, {uhefb__quziq}_size = {npjwa__nsa}({table_name}_table, {table_name}_ind)
"""
            func_text += f"""  {uhefb__quziq} = bodo.libs.str_arr_ext.decode_utf8({uhefb__quziq}, {uhefb__quziq}_size)
"""
        else:
            func_text += (
                f'  {uhefb__quziq} = {npjwa__nsa}({table_name}_data1, {table_name}_ind)\n'
                )
        table_getitem_funcs[npjwa__nsa
            ] = bodo.libs.array._gen_row_access_intrinsic(urb__dlpjy, rod__ivs)
        expr = expr.replace(cname, uhefb__quziq)
        rxmhv__gme = f'({na_check_name}.{table_name}.{amu__bzd})'
        if rxmhv__gme in expr:
            ntdq__ldkna = typemap[col_vars[amu__bzd].name]
            hjc__yco = f'nacheck_{table_name}_val_{rod__ivs}'
            zxij__kecr = f'_bodo_isna_{table_name}_val_{rod__ivs}'
            if isinstance(ntdq__ldkna, bodo.libs.int_arr_ext.IntegerArrayType
                ) or ntdq__ldkna in [bodo.libs.bool_arr_ext.boolean_array,
                bodo.libs.str_arr_ext.string_array_type]:
                func_text += f"""  {zxij__kecr} = {hjc__yco}({table_name}_null_bitmap, {table_name}_ind)
"""
            else:
                func_text += (
                    f'  {zxij__kecr} = {hjc__yco}({table_name}_data1, {table_name}_ind)\n'
                    )
            table_getitem_funcs[hjc__yco
                ] = bodo.libs.array._gen_row_na_check_intrinsic(ntdq__ldkna,
                rod__ivs)
            expr = expr.replace(rxmhv__gme, zxij__kecr)
        if rod__ivs >= n_keys:
            amf__tlb.append(rod__ivs)
    return expr, func_text, amf__tlb


def _get_col_to_ind(key_names, col_vars):
    n_keys = len(key_names)
    col_to_ind = {amu__bzd: i for i, amu__bzd in enumerate(key_names)}
    i = n_keys
    for amu__bzd in sorted(col_vars, key=lambda a: str(a)):
        if amu__bzd in key_names:
            continue
        col_to_ind[amu__bzd] = i
        i += 1
    return col_to_ind


def _match_join_key_types(t1, t2, loc):
    if t1 == t2:
        return t1
    try:
        arr = dtype_to_array_type(find_common_np_dtype([t1, t2]))
        return to_nullable_type(arr) if is_nullable_type(t1
            ) or is_nullable_type(t2) else arr
    except:
        raise BodoError(f'Join key types {t1} and {t2} do not match', loc=loc)


def _get_table_parallel_flags(join_node, array_dists):
    lkzq__kxzzy = (distributed_pass.Distribution.OneD, distributed_pass.
        Distribution.OneD_Var)
    left_parallel = all(array_dists[ttf__mmsjq.name] in lkzq__kxzzy for
        ttf__mmsjq in join_node.left_vars.values())
    right_parallel = all(array_dists[ttf__mmsjq.name] in lkzq__kxzzy for
        ttf__mmsjq in join_node.right_vars.values())
    if not left_parallel:
        assert not any(array_dists[ttf__mmsjq.name] in lkzq__kxzzy for
            ttf__mmsjq in join_node.left_vars.values())
    if not right_parallel:
        assert not any(array_dists[ttf__mmsjq.name] in lkzq__kxzzy for
            ttf__mmsjq in join_node.right_vars.values())
    if left_parallel or right_parallel:
        assert all(array_dists[ttf__mmsjq.name] in lkzq__kxzzy for
            ttf__mmsjq in join_node.out_data_vars.values())
    return left_parallel, right_parallel


def _gen_local_hash_join(optional_column, left_key_names, right_key_names,
    left_key_types, right_key_types, left_other_names, right_other_names,
    left_other_types, right_other_types, vect_same_key, is_left, is_right,
    is_join, left_parallel, right_parallel, glbs, out_types, loc, indicator,
    is_na_equal, general_cond_cfunc, left_col_nums, right_col_nums):

    def needs_typechange(in_type, need_nullable, is_same_key):
        return isinstance(in_type, types.Array) and not is_dtype_nullable(
            in_type.dtype) and need_nullable and not is_same_key
    npse__opp = []
    for i in range(len(left_key_names)):
        bjyj__mgjx = _match_join_key_types(left_key_types[i],
            right_key_types[i], loc)
        npse__opp.append(needs_typechange(bjyj__mgjx, is_right,
            vect_same_key[i]))
    for i in range(len(left_other_names)):
        npse__opp.append(needs_typechange(left_other_types[i], is_right, False)
            )
    for i in range(len(right_key_names)):
        if not vect_same_key[i] and not is_join:
            bjyj__mgjx = _match_join_key_types(left_key_types[i],
                right_key_types[i], loc)
            npse__opp.append(needs_typechange(bjyj__mgjx, is_left, False))
    for i in range(len(right_other_names)):
        npse__opp.append(needs_typechange(right_other_types[i], is_left, False)
            )

    def get_out_type(idx, in_type, in_name, need_nullable, is_same_key):
        if isinstance(in_type, types.Array) and not is_dtype_nullable(in_type
            .dtype) and need_nullable and not is_same_key:
            if isinstance(in_type.dtype, types.Integer):
                jgx__jwtyc = IntDtype(in_type.dtype).name
                assert jgx__jwtyc.endswith('Dtype()')
                jgx__jwtyc = jgx__jwtyc[:-7]
                ixqsn__byep = f"""    typ_{idx} = bodo.hiframes.pd_series_ext.get_series_data(pd.Series([1], dtype="{jgx__jwtyc}"))
"""
                qesqy__zye = f'typ_{idx}'
            else:
                assert in_type.dtype == types.bool_, 'unexpected non-nullable type in join'
                ixqsn__byep = (
                    f'    typ_{idx} = bodo.libs.bool_arr_ext.alloc_bool_array(1)\n'
                    )
                qesqy__zye = f'typ_{idx}'
        else:
            ixqsn__byep = ''
            qesqy__zye = in_name
        return ixqsn__byep, qesqy__zye
    n_keys = len(left_key_names)
    func_text = '    # beginning of _gen_local_hash_join\n'
    xsl__oyq = []
    for i in range(n_keys):
        xsl__oyq.append('t1_keys[{}]'.format(i))
    for i in range(len(left_other_names)):
        xsl__oyq.append('data_left[{}]'.format(i))
    func_text += '    info_list_total_l = [{}]\n'.format(','.join(
        'array_to_info({})'.format(a) for a in xsl__oyq))
    func_text += '    table_left = arr_info_list_to_table(info_list_total_l)\n'
    ytigp__xnax = []
    for i in range(n_keys):
        ytigp__xnax.append('t2_keys[{}]'.format(i))
    for i in range(len(right_other_names)):
        ytigp__xnax.append('data_right[{}]'.format(i))
    func_text += '    info_list_total_r = [{}]\n'.format(','.join(
        'array_to_info({})'.format(a) for a in ytigp__xnax))
    func_text += (
        '    table_right = arr_info_list_to_table(info_list_total_r)\n')
    func_text += '    vect_same_key = np.array([{}])\n'.format(','.join('1' if
        ykn__vzdb else '0' for ykn__vzdb in vect_same_key))
    func_text += '    vect_need_typechange = np.array([{}])\n'.format(','.
        join('1' if ykn__vzdb else '0' for ykn__vzdb in npse__opp))
    func_text += f"""    left_table_cond_columns = np.array({left_col_nums if len(left_col_nums) > 0 else [-1]}, dtype=np.int64)
"""
    func_text += f"""    right_table_cond_columns = np.array({right_col_nums if len(right_col_nums) > 0 else [-1]}, dtype=np.int64)
"""
    if general_cond_cfunc:
        func_text += f"""    cfunc_cond = add_join_gen_cond_cfunc_sym(general_cond_cfunc, '{general_cond_cfunc.native_name}')
"""
        func_text += (
            f"    cfunc_cond = get_join_cond_addr('{general_cond_cfunc.native_name}')\n"
            )
    else:
        func_text += '    cfunc_cond = 0\n'
    func_text += (
        """    out_table = hash_join_table(table_left, table_right, {}, {}, {}, {}, {}, vect_same_key.ctypes, vect_need_typechange.ctypes, {}, {}, {}, {}, {}, {}, cfunc_cond, left_table_cond_columns.ctypes, {}, right_table_cond_columns.ctypes, {})
"""
        .format(left_parallel, right_parallel, n_keys, len(left_other_names
        ), len(right_other_names), is_left, is_right, is_join,
        optional_column, indicator, is_na_equal, len(left_col_nums), len(
        right_col_nums)))
    func_text += '    delete_table(table_left)\n'
    func_text += '    delete_table(table_right)\n'
    idx = 0
    if optional_column:
        func_text += (
            f'    opti_0 = info_to_array(info_from_table(out_table, {idx}), opti_c0)\n'
            )
        idx += 1
    for i, gce__rcp in enumerate(left_key_names):
        bjyj__mgjx = _match_join_key_types(left_key_types[i],
            right_key_types[i], loc)
        wete__jwj = get_out_type(idx, bjyj__mgjx, f't1_keys[{i}]', is_right,
            vect_same_key[i])
        func_text += wete__jwj[0]
        glbs[f'out_type_{idx}'] = out_types[idx]
        if bjyj__mgjx != left_key_types[i]:
            func_text += f"""    t1_keys_{i} = bodo.utils.utils.astype(info_to_array(info_from_table(out_table, {idx}), {wete__jwj[1]}), out_type_{idx})
"""
        else:
            func_text += f"""    t1_keys_{i} = info_to_array(info_from_table(out_table, {idx}), {wete__jwj[1]})
"""
        idx += 1
    for i, gce__rcp in enumerate(left_other_names):
        wete__jwj = get_out_type(idx, left_other_types[i], gce__rcp,
            is_right, False)
        func_text += wete__jwj[0]
        func_text += (
            '    left_{} = info_to_array(info_from_table(out_table, {}), {})\n'
            .format(i, idx, wete__jwj[1]))
        idx += 1
    for i, gce__rcp in enumerate(right_key_names):
        if not vect_same_key[i] and not is_join:
            bjyj__mgjx = _match_join_key_types(left_key_types[i],
                right_key_types[i], loc)
            wete__jwj = get_out_type(idx, bjyj__mgjx, f't2_keys[{i}]',
                is_left, False)
            func_text += wete__jwj[0]
            glbs[f'out_type_{idx}'] = out_types[idx - len(left_other_names)]
            if bjyj__mgjx != right_key_types[i]:
                func_text += f"""    t2_keys_{i} = bodo.utils.utils.astype(info_to_array(info_from_table(out_table, {idx}), {wete__jwj[1]}), out_type_{idx})
"""
            else:
                func_text += f"""    t2_keys_{i} = info_to_array(info_from_table(out_table, {idx}), {wete__jwj[1]})
"""
            idx += 1
    for i, gce__rcp in enumerate(right_other_names):
        wete__jwj = get_out_type(idx, right_other_types[i], gce__rcp,
            is_left, False)
        func_text += wete__jwj[0]
        func_text += (
            '    right_{} = info_to_array(info_from_table(out_table, {}), {})\n'
            .format(i, idx, wete__jwj[1]))
        idx += 1
    if indicator:
        func_text += f"""    typ_{idx} = pd.Categorical(values=['both'], categories=('left_only', 'right_only', 'both'))
"""
        func_text += f"""    indicator_col = info_to_array(info_from_table(out_table, {idx}), typ_{idx})
"""
        idx += 1
    func_text += '    delete_table(out_table)\n'
    return func_text


def parallel_join_impl(key_arrs, data):
    gcl__qvr = bodo.libs.distributed_api.get_size()
    zrgu__lml = alloc_pre_shuffle_metadata(key_arrs, data, gcl__qvr, False)
    tks__ojq = len(key_arrs[0])
    egmtt__mns = np.empty(tks__ojq, np.int32)
    ice__zll = arr_info_list_to_table([array_to_info(key_arrs[0])])
    lxmg__frpqt = 1
    zapml__lvp = compute_node_partition_by_hash(ice__zll, lxmg__frpqt, gcl__qvr
        )
    namx__ezfa = np.empty(1, np.int32)
    orryp__zawh = info_to_array(info_from_table(zapml__lvp, 0), namx__ezfa)
    delete_table(zapml__lvp)
    delete_table(ice__zll)
    for i in range(tks__ojq):
        val = getitem_arr_tup_single(key_arrs, i)
        node_id = orryp__zawh[i]
        egmtt__mns[i] = node_id
        update_shuffle_meta(zrgu__lml, node_id, i, key_arrs, data, False)
    shuffle_meta = finalize_shuffle_meta(key_arrs, data, zrgu__lml,
        gcl__qvr, False)
    for i in range(tks__ojq):
        node_id = egmtt__mns[i]
        write_send_buff(shuffle_meta, node_id, i, key_arrs, data)
        shuffle_meta.tmp_offset[node_id] += 1
    gqz__ycu = alltoallv_tup(key_arrs + data, shuffle_meta, key_arrs)
    cfh__vsco = _get_keys_tup(gqz__ycu, key_arrs)
    hnej__ybvl = _get_data_tup(gqz__ycu, key_arrs)
    return cfh__vsco, hnej__ybvl


@generated_jit(nopython=True, cache=True)
def parallel_shuffle(key_arrs, data):
    return parallel_join_impl


@numba.njit
def parallel_asof_comm(left_key_arrs, right_key_arrs, right_data):
    gcl__qvr = bodo.libs.distributed_api.get_size()
    fjesb__dzwnj = np.empty(gcl__qvr, left_key_arrs[0].dtype)
    otvwo__lopb = np.empty(gcl__qvr, left_key_arrs[0].dtype)
    bodo.libs.distributed_api.allgather(fjesb__dzwnj, left_key_arrs[0][0])
    bodo.libs.distributed_api.allgather(otvwo__lopb, left_key_arrs[0][-1])
    pkyyt__vthn = np.zeros(gcl__qvr, np.int32)
    owh__lfq = np.zeros(gcl__qvr, np.int32)
    rplr__lgak = np.zeros(gcl__qvr, np.int32)
    tith__mcpd = right_key_arrs[0][0]
    ocmmt__gekv = right_key_arrs[0][-1]
    saxw__qdkw = -1
    i = 0
    while i < gcl__qvr - 1 and otvwo__lopb[i] < tith__mcpd:
        i += 1
    while i < gcl__qvr and fjesb__dzwnj[i] <= ocmmt__gekv:
        saxw__qdkw, sczei__hrhz = _count_overlap(right_key_arrs[0],
            fjesb__dzwnj[i], otvwo__lopb[i])
        if saxw__qdkw != 0:
            saxw__qdkw -= 1
            sczei__hrhz += 1
        pkyyt__vthn[i] = sczei__hrhz
        owh__lfq[i] = saxw__qdkw
        i += 1
    while i < gcl__qvr:
        pkyyt__vthn[i] = 1
        owh__lfq[i] = len(right_key_arrs[0]) - 1
        i += 1
    bodo.libs.distributed_api.alltoall(pkyyt__vthn, rplr__lgak, 1)
    ixtpu__zxsd = rplr__lgak.sum()
    bbu__agq = np.empty(ixtpu__zxsd, right_key_arrs[0].dtype)
    plu__grva = alloc_arr_tup(ixtpu__zxsd, right_data)
    ewmd__dkaga = bodo.ir.join.calc_disp(rplr__lgak)
    bodo.libs.distributed_api.alltoallv(right_key_arrs[0], bbu__agq,
        pkyyt__vthn, rplr__lgak, owh__lfq, ewmd__dkaga)
    bodo.libs.distributed_api.alltoallv_tup(right_data, plu__grva,
        pkyyt__vthn, rplr__lgak, owh__lfq, ewmd__dkaga)
    return (bbu__agq,), plu__grva


@numba.njit
def _count_overlap(r_key_arr, start, end):
    sczei__hrhz = 0
    saxw__qdkw = 0
    utcs__gtat = 0
    while utcs__gtat < len(r_key_arr) and r_key_arr[utcs__gtat] < start:
        saxw__qdkw += 1
        utcs__gtat += 1
    while utcs__gtat < len(r_key_arr) and start <= r_key_arr[utcs__gtat
        ] <= end:
        utcs__gtat += 1
        sczei__hrhz += 1
    return saxw__qdkw, sczei__hrhz


def write_send_buff(shuffle_meta, node_id, i, key_arrs, data):
    return i


@overload(write_send_buff, no_unliteral=True)
def write_data_buff_overload(meta, node_id, i, key_arrs, data):
    func_text = 'def f(meta, node_id, i, key_arrs, data):\n'
    func_text += (
        '  w_ind = meta.send_disp[node_id] + meta.tmp_offset[node_id]\n')
    n_keys = len(key_arrs.types)
    for i, euu__rpkou in enumerate(key_arrs.types + data.types):
        arr = 'key_arrs[{}]'.format(i) if i < n_keys else 'data[{}]'.format(
            i - n_keys)
        if not euu__rpkou in (string_type, string_array_type,
            binary_array_type, bytes_type):
            func_text += '  meta.send_buff_tup[{}][w_ind] = {}[i]\n'.format(i,
                arr)
        else:
            func_text += ('  n_chars_{} = get_str_arr_item_length({}, i)\n'
                .format(i, arr))
            func_text += ('  meta.send_arr_lens_tup[{}][w_ind] = n_chars_{}\n'
                .format(i, i))
            if i >= n_keys:
                func_text += (
                    """  out_bitmap = meta.send_arr_nulls_tup[{}][meta.send_disp_nulls[node_id]:].ctypes
"""
                    .format(i))
                func_text += (
                    '  bit_val = get_bit_bitmap(get_null_bitmap_ptr(data[{}]), i)\n'
                    .format(i - n_keys))
                func_text += (
                    '  set_bit_to(out_bitmap, meta.tmp_offset[node_id], bit_val)\n'
                    )
            func_text += (
                """  indc_{} = meta.send_disp_char_tup[{}][node_id] + meta.tmp_offset_char_tup[{}][node_id]
"""
                .format(i, i, i))
            func_text += ('  item_ptr_{} = get_str_arr_item_ptr({}, i)\n'.
                format(i, arr))
            func_text += (
                """  str_copy_ptr(meta.send_arr_chars_tup[{}], indc_{}, item_ptr_{}, n_chars_{})
"""
                .format(i, i, i, i))
            func_text += (
                '  meta.tmp_offset_char_tup[{}][node_id] += n_chars_{}\n'.
                format(i, i))
    func_text += '  return w_ind\n'
    wqhck__zscak = {}
    exec(func_text, {'str_copy_ptr': str_copy_ptr, 'get_null_bitmap_ptr':
        get_null_bitmap_ptr, 'get_bit_bitmap': get_bit_bitmap, 'set_bit_to':
        set_bit_to, 'get_str_arr_item_length': get_str_arr_item_length,
        'get_str_arr_item_ptr': get_str_arr_item_ptr}, wqhck__zscak)
    tmv__ifxbj = wqhck__zscak['f']
    return tmv__ifxbj


import llvmlite.binding as ll
from bodo.libs import hdist
ll.add_symbol('c_alltoallv', hdist.c_alltoallv)


@numba.njit
def calc_disp(arr):
    ratb__vipd = np.empty_like(arr)
    ratb__vipd[0] = 0
    for i in range(1, len(arr)):
        ratb__vipd[i] = ratb__vipd[i - 1] + arr[i - 1]
    return ratb__vipd


def ensure_capacity(arr, new_size):
    oryfs__siu = arr
    hjz__zhxvq = len(arr)
    if hjz__zhxvq < new_size:
        fjned__mwt = 2 * hjz__zhxvq
        oryfs__siu = bodo.utils.utils.alloc_type(fjned__mwt, arr)
        oryfs__siu[:hjz__zhxvq] = arr
    return oryfs__siu


@overload(ensure_capacity, no_unliteral=True)
def ensure_capacity_overload(arr, new_size):
    if isinstance(arr, types.Array) or arr == boolean_array:
        return ensure_capacity
    assert isinstance(arr, types.BaseTuple)
    sczei__hrhz = arr.count
    func_text = 'def f(arr, new_size):\n'
    func_text += '  return ({}{})\n'.format(','.join([
        'ensure_capacity(arr[{}], new_size)'.format(i) for i in range(
        sczei__hrhz)]), ',' if sczei__hrhz == 1 else '')
    wqhck__zscak = {}
    exec(func_text, {'ensure_capacity': ensure_capacity}, wqhck__zscak)
    bmmq__akz = wqhck__zscak['f']
    return bmmq__akz


@numba.njit
def ensure_capacity_str(arr, new_size, n_chars):
    oryfs__siu = arr
    hjz__zhxvq = len(arr)
    gzjxh__huj = num_total_chars(arr)
    flk__psrk = getitem_str_offset(arr, new_size - 1) + n_chars
    if hjz__zhxvq < new_size or flk__psrk > gzjxh__huj:
        fjned__mwt = int(2 * hjz__zhxvq if hjz__zhxvq < new_size else
            hjz__zhxvq)
        maw__fwqcv = int(2 * gzjxh__huj + n_chars if flk__psrk > gzjxh__huj
             else gzjxh__huj)
        oryfs__siu = pre_alloc_string_array(fjned__mwt, maw__fwqcv)
        copy_str_arr_slice(oryfs__siu, arr, new_size - 1)
    return oryfs__siu


def trim_arr_tup(data, new_size):
    return data


@overload(trim_arr_tup, no_unliteral=True)
def trim_arr_tup_overload(data, new_size):
    assert isinstance(data, types.BaseTuple)
    sczei__hrhz = data.count
    func_text = 'def f(data, new_size):\n'
    func_text += '  return ({}{})\n'.format(','.join([
        'trim_arr(data[{}], new_size)'.format(i) for i in range(sczei__hrhz
        )]), ',' if sczei__hrhz == 1 else '')
    wqhck__zscak = {}
    exec(func_text, {'trim_arr': trim_arr}, wqhck__zscak)
    bmmq__akz = wqhck__zscak['f']
    return bmmq__akz


def copy_elem_buff(arr, ind, val):
    oryfs__siu = ensure_capacity(arr, ind + 1)
    oryfs__siu[ind] = val
    return oryfs__siu


@overload(copy_elem_buff, no_unliteral=True)
def copy_elem_buff_overload(arr, ind, val):
    if isinstance(arr, types.Array) or arr == boolean_array:
        return copy_elem_buff
    assert arr == string_array_type

    def copy_elem_buff_str(arr, ind, val):
        oryfs__siu = ensure_capacity_str(arr, ind + 1, get_utf8_size(val))
        oryfs__siu[ind] = val
        return oryfs__siu
    return copy_elem_buff_str


def copy_elem_buff_tup(arr, ind, val):
    return arr


@overload(copy_elem_buff_tup, no_unliteral=True)
def copy_elem_buff_tup_overload(data, ind, val):
    assert isinstance(data, types.BaseTuple)
    sczei__hrhz = data.count
    func_text = 'def f(data, ind, val):\n'
    for i in range(sczei__hrhz):
        func_text += ('  arr_{} = copy_elem_buff(data[{}], ind, val[{}])\n'
            .format(i, i, i))
    func_text += '  return ({}{})\n'.format(','.join(['arr_{}'.format(i) for
        i in range(sczei__hrhz)]), ',' if sczei__hrhz == 1 else '')
    wqhck__zscak = {}
    exec(func_text, {'copy_elem_buff': copy_elem_buff}, wqhck__zscak)
    lfo__zjf = wqhck__zscak['f']
    return lfo__zjf


def trim_arr(arr, size):
    return arr[:size]


@overload(trim_arr, no_unliteral=True)
def trim_arr_overload(arr, size):
    if isinstance(arr, types.Array) or arr == boolean_array:
        return trim_arr
    assert arr == string_array_type

    def trim_arr_str(arr, size):
        oryfs__siu = pre_alloc_string_array(size, np.int64(
            getitem_str_offset(arr, size)))
        copy_str_arr_slice(oryfs__siu, arr, size)
        return oryfs__siu
    return trim_arr_str


def setnan_elem_buff(arr, ind):
    oryfs__siu = ensure_capacity(arr, ind + 1)
    bodo.libs.array_kernels.setna(oryfs__siu, ind)
    return oryfs__siu


@overload(setnan_elem_buff, no_unliteral=True)
def setnan_elem_buff_overload(arr, ind):
    if isinstance(arr, types.Array) or arr == boolean_array:
        return setnan_elem_buff
    assert arr == string_array_type

    def setnan_elem_buff_str(arr, ind):
        oryfs__siu = ensure_capacity_str(arr, ind + 1, 0)
        oryfs__siu[ind] = ''
        bodo.libs.array_kernels.setna(oryfs__siu, ind)
        return oryfs__siu
    return setnan_elem_buff_str


def setnan_elem_buff_tup(arr, ind):
    return arr


@overload(setnan_elem_buff_tup, no_unliteral=True)
def setnan_elem_buff_tup_overload(data, ind):
    assert isinstance(data, types.BaseTuple)
    sczei__hrhz = data.count
    func_text = 'def f(data, ind):\n'
    for i in range(sczei__hrhz):
        func_text += '  arr_{} = setnan_elem_buff(data[{}], ind)\n'.format(i, i
            )
    func_text += '  return ({}{})\n'.format(','.join(['arr_{}'.format(i) for
        i in range(sczei__hrhz)]), ',' if sczei__hrhz == 1 else '')
    wqhck__zscak = {}
    exec(func_text, {'setnan_elem_buff': setnan_elem_buff}, wqhck__zscak)
    lfo__zjf = wqhck__zscak['f']
    return lfo__zjf


@generated_jit(nopython=True, cache=True)
def _check_ind_if_hashed(right_keys, r_ind, l_key):
    if right_keys == types.Tuple((types.intp[::1],)):
        return lambda right_keys, r_ind, l_key: r_ind

    def _impl(right_keys, r_ind, l_key):
        oqnyw__xfdxu = getitem_arr_tup(right_keys, r_ind)
        if oqnyw__xfdxu != l_key:
            return -1
        return r_ind
    return _impl


@numba.njit
def local_merge_asof(left_keys, right_keys, data_left, data_right):
    rbdpf__muf = len(left_keys[0])
    fnnyh__reje = len(right_keys[0])
    dhd__lhio = alloc_arr_tup(rbdpf__muf, left_keys)
    mzyvl__itahd = alloc_arr_tup(rbdpf__muf, right_keys)
    rdo__mbb = alloc_arr_tup(rbdpf__muf, data_left)
    nwwe__hyap = alloc_arr_tup(rbdpf__muf, data_right)
    trbco__rsy = 0
    rboa__ggjh = 0
    for trbco__rsy in range(rbdpf__muf):
        if rboa__ggjh < 0:
            rboa__ggjh = 0
        while rboa__ggjh < fnnyh__reje and getitem_arr_tup(right_keys,
            rboa__ggjh) <= getitem_arr_tup(left_keys, trbco__rsy):
            rboa__ggjh += 1
        rboa__ggjh -= 1
        setitem_arr_tup(dhd__lhio, trbco__rsy, getitem_arr_tup(left_keys,
            trbco__rsy))
        setitem_arr_tup(rdo__mbb, trbco__rsy, getitem_arr_tup(data_left,
            trbco__rsy))
        if rboa__ggjh >= 0:
            setitem_arr_tup(mzyvl__itahd, trbco__rsy, getitem_arr_tup(
                right_keys, rboa__ggjh))
            setitem_arr_tup(nwwe__hyap, trbco__rsy, getitem_arr_tup(
                data_right, rboa__ggjh))
        else:
            bodo.libs.array_kernels.setna_tup(mzyvl__itahd, trbco__rsy)
            bodo.libs.array_kernels.setna_tup(nwwe__hyap, trbco__rsy)
    return dhd__lhio, mzyvl__itahd, rdo__mbb, nwwe__hyap


def copy_arr_tup(arrs):
    return tuple(a.copy() for a in arrs)


@overload(copy_arr_tup, no_unliteral=True)
def copy_arr_tup_overload(arrs):
    sczei__hrhz = arrs.count
    func_text = 'def f(arrs):\n'
    func_text += '  return ({},)\n'.format(','.join('arrs[{}].copy()'.
        format(i) for i in range(sczei__hrhz)))
    wqhck__zscak = {}
    exec(func_text, {}, wqhck__zscak)
    impl = wqhck__zscak['f']
    return impl


def get_nan_bits(arr, ind):
    return 0


@overload(get_nan_bits, no_unliteral=True)
def overload_get_nan_bits(arr, ind):
    if arr == string_array_type:

        def impl_str(arr, ind):
            ohkxl__solfg = get_null_bitmap_ptr(arr)
            return get_bit_bitmap(ohkxl__solfg, ind)
        return impl_str
    if isinstance(arr, IntegerArrayType) or arr == boolean_array:

        def impl(arr, ind):
            return bodo.libs.int_arr_ext.get_bit_bitmap_arr(arr.
                _null_bitmap, ind)
        return impl
    return lambda arr, ind: False


def get_nan_bits_tup(arr_tup, ind):
    return tuple(get_nan_bits(arr, ind) for arr in arr_tup)


@overload(get_nan_bits_tup, no_unliteral=True)
def overload_get_nan_bits_tup(arr_tup, ind):
    sczei__hrhz = arr_tup.count
    func_text = 'def f(arr_tup, ind):\n'
    func_text += '  return ({}{})\n'.format(','.join([
        'get_nan_bits(arr_tup[{}], ind)'.format(i) for i in range(
        sczei__hrhz)]), ',' if sczei__hrhz == 1 else '')
    wqhck__zscak = {}
    exec(func_text, {'get_nan_bits': get_nan_bits}, wqhck__zscak)
    impl = wqhck__zscak['f']
    return impl


def set_nan_bits(arr, ind, na_val):
    return 0


@overload(set_nan_bits, no_unliteral=True)
def overload_set_nan_bits(arr, ind, na_val):
    if arr == string_array_type:

        def impl_str(arr, ind, na_val):
            ohkxl__solfg = get_null_bitmap_ptr(arr)
            set_bit_to(ohkxl__solfg, ind, na_val)
        return impl_str
    if isinstance(arr, IntegerArrayType) or arr == boolean_array:

        def impl(arr, ind, na_val):
            bodo.libs.int_arr_ext.set_bit_to_arr(arr._null_bitmap, ind, na_val)
        return impl
    return lambda arr, ind, na_val: None


def set_nan_bits_tup(arr_tup, ind, na_val):
    return tuple(set_nan_bits(arr, ind, na_val) for arr in arr_tup)


@overload(set_nan_bits_tup, no_unliteral=True)
def overload_set_nan_bits_tup(arr_tup, ind, na_val):
    sczei__hrhz = arr_tup.count
    func_text = 'def f(arr_tup, ind, na_val):\n'
    for i in range(sczei__hrhz):
        func_text += '  set_nan_bits(arr_tup[{}], ind, na_val[{}])\n'.format(i,
            i)
    func_text += '  return\n'
    wqhck__zscak = {}
    exec(func_text, {'set_nan_bits': set_nan_bits}, wqhck__zscak)
    impl = wqhck__zscak['f']
    return impl
