import numba
import numpy as np
import pandas as pd
from numba.core import ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, replace_arg_nodes
import bodo
import bodo.ir.connector
from bodo import objmode
from bodo.libs.str_ext import string_type
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.utils.utils import check_java_installation
from bodo.utils.utils import sanitize_varname


class JsonReader(ir.Stmt):

    def __init__(self, df_out, loc, out_vars, out_types, file_name,
        df_colnames, orient, convert_dates, precise_float, lines, compression):
        self.connector_typ = 'json'
        self.df_out = df_out
        self.loc = loc
        self.out_vars = out_vars
        self.out_types = out_types
        self.file_name = file_name
        self.df_colnames = df_colnames
        self.orient = orient
        self.convert_dates = convert_dates
        self.precise_float = precise_float
        self.lines = lines
        self.compression = compression

    def __repr__(self):
        return ('{} = ReadJson(file={}, col_names={}, types={}, vars={})'.
            format(self.df_out, self.file_name, self.df_colnames, self.
            out_types, self.out_vars))


import llvmlite.binding as ll
from bodo.io import json_cpp
ll.add_symbol('json_file_chunk_reader', json_cpp.json_file_chunk_reader)
json_file_chunk_reader = types.ExternalFunction('json_file_chunk_reader',
    bodo.ir.connector.stream_reader_type(types.voidptr, types.bool_, types.
    bool_, types.int64, types.voidptr, types.voidptr))


def remove_dead_json(json_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    lyrit__pbcxi = []
    wbxh__obtjb = []
    mbecx__zwrhy = []
    for cxt__cxg, yvzja__mru in enumerate(json_node.out_vars):
        if yvzja__mru.name in lives:
            lyrit__pbcxi.append(json_node.df_colnames[cxt__cxg])
            wbxh__obtjb.append(json_node.out_vars[cxt__cxg])
            mbecx__zwrhy.append(json_node.out_types[cxt__cxg])
    json_node.df_colnames = lyrit__pbcxi
    json_node.out_vars = wbxh__obtjb
    json_node.out_types = mbecx__zwrhy
    if len(json_node.out_vars) == 0:
        return None
    return json_node


def json_distributed_run(json_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    parallel = False
    if array_dists is not None:
        parallel = True
        for cid__acg in json_node.out_vars:
            if array_dists[cid__acg.name
                ] != distributed_pass.Distribution.OneD and array_dists[
                cid__acg.name] != distributed_pass.Distribution.OneD_Var:
                parallel = False
    atdx__wyv = len(json_node.out_vars)
    qsj__coto = ', '.join('arr' + str(cxt__cxg) for cxt__cxg in range(
        atdx__wyv))
    piwvi__ybeut = 'def json_impl(fname):\n'
    piwvi__ybeut += '    ({},) = _json_reader_py(fname)\n'.format(qsj__coto)
    bnkb__btwpu = {}
    exec(piwvi__ybeut, {}, bnkb__btwpu)
    use__ednfg = bnkb__btwpu['json_impl']
    xsrod__jvhq = _gen_json_reader_py(json_node.df_colnames, json_node.
        out_types, typingctx, targetctx, parallel, json_node.orient,
        json_node.convert_dates, json_node.precise_float, json_node.lines,
        json_node.compression)
    wru__trm = compile_to_numba_ir(use__ednfg, {'_json_reader_py':
        xsrod__jvhq}, typingctx=typingctx, targetctx=targetctx, arg_typs=(
        string_type,), typemap=typemap, calltypes=calltypes).blocks.popitem()[1
        ]
    replace_arg_nodes(wru__trm, [json_node.file_name])
    gcy__cby = wru__trm.body[:-3]
    for cxt__cxg in range(len(json_node.out_vars)):
        gcy__cby[-len(json_node.out_vars) + cxt__cxg
            ].target = json_node.out_vars[cxt__cxg]
    return gcy__cby


numba.parfors.array_analysis.array_analysis_extensions[JsonReader
    ] = bodo.ir.connector.connector_array_analysis
distributed_analysis.distributed_analysis_extensions[JsonReader
    ] = bodo.ir.connector.connector_distributed_analysis
typeinfer.typeinfer_extensions[JsonReader
    ] = bodo.ir.connector.connector_typeinfer
ir_utils.visit_vars_extensions[JsonReader
    ] = bodo.ir.connector.visit_vars_connector
ir_utils.remove_dead_extensions[JsonReader] = remove_dead_json
numba.core.analysis.ir_extension_usedefs[JsonReader
    ] = bodo.ir.connector.connector_usedefs
ir_utils.copy_propagate_extensions[JsonReader
    ] = bodo.ir.connector.get_copies_connector
ir_utils.apply_copy_propagate_extensions[JsonReader
    ] = bodo.ir.connector.apply_copies_connector
ir_utils.build_defs_extensions[JsonReader
    ] = bodo.ir.connector.build_connector_definitions
distributed_pass.distributed_run_extensions[JsonReader] = json_distributed_run
compiled_funcs = []


def _gen_json_reader_py(col_names, col_typs, typingctx, targetctx, parallel,
    orient, convert_dates, precise_float, lines, compression):
    mwbf__hml = [sanitize_varname(gotbe__eois) for gotbe__eois in col_names]
    pkfxv__pcyf = ', '.join(str(cxt__cxg) for cxt__cxg, evvue__pdusu in
        enumerate(col_typs) if evvue__pdusu.dtype == types.NPDatetime('ns'))
    bevk__nfda = ', '.join(["{}='{}'".format(wlyxl__ifwbi, bodo.ir.csv_ext.
        _get_dtype_str(evvue__pdusu)) for wlyxl__ifwbi, evvue__pdusu in zip
        (mwbf__hml, col_typs)])
    xqa__itq = ', '.join(["'{}':{}".format(iyw__bgpt, bodo.ir.csv_ext.
        _get_pd_dtype_str(evvue__pdusu)) for iyw__bgpt, evvue__pdusu in zip
        (col_names, col_typs)])
    if compression is None:
        compression = 'uncompressed'
    piwvi__ybeut = 'def json_reader_py(fname):\n'
    piwvi__ybeut += '  check_java_installation(fname)\n'
    piwvi__ybeut += f"""  bucket_region = bodo.io.fs_io.get_s3_bucket_region_njit(fname, parallel={parallel})
"""
    piwvi__ybeut += (
        '  f_reader = bodo.ir.json_ext.json_file_chunk_reader(bodo.libs.str_ext.unicode_to_utf8(fname), '
        )
    piwvi__ybeut += (
        """    {}, {}, -1, bodo.libs.str_ext.unicode_to_utf8('{}'), bodo.libs.str_ext.unicode_to_utf8(bucket_region) )
"""
        .format(lines, parallel, compression))
    piwvi__ybeut += '  bodo.utils.utils.check_and_propagate_cpp_exception()\n'
    piwvi__ybeut += '  if bodo.utils.utils.is_null_pointer(f_reader):\n'
    piwvi__ybeut += "      raise FileNotFoundError('File does not exist')\n"
    piwvi__ybeut += '  with objmode({}):\n'.format(bevk__nfda)
    piwvi__ybeut += "    df = pd.read_json(f_reader, orient='{}',\n".format(
        orient)
    piwvi__ybeut += '       convert_dates = {}, \n'.format(convert_dates)
    piwvi__ybeut += '       precise_float={}, \n'.format(precise_float)
    piwvi__ybeut += '       lines={}, \n'.format(lines)
    piwvi__ybeut += '       dtype={{{}}},\n'.format(xqa__itq)
    piwvi__ybeut += '       )\n'
    for wlyxl__ifwbi, iyw__bgpt in zip(mwbf__hml, col_names):
        piwvi__ybeut += '    if len(df) > 0:\n'
        piwvi__ybeut += "        {} = df['{}'].values\n".format(wlyxl__ifwbi,
            iyw__bgpt)
        piwvi__ybeut += '    else:\n'
        piwvi__ybeut += '        {} = np.array([])\n'.format(wlyxl__ifwbi)
    piwvi__ybeut += '  return ({},)\n'.format(', '.join(lxret__qpzmc for
        lxret__qpzmc in mwbf__hml))
    ggrve__kgw = globals()
    bnkb__btwpu = {}
    exec(piwvi__ybeut, ggrve__kgw, bnkb__btwpu)
    xsrod__jvhq = bnkb__btwpu['json_reader_py']
    zkbw__glun = numba.njit(xsrod__jvhq)
    compiled_funcs.append(zkbw__glun)
    return zkbw__glun
