"""IR node for the data sorting"""
from collections import defaultdict
import numba
import numpy as np
from numba.core import ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, mk_unique_var, replace_arg_nodes, replace_vars_inner, visit_vars_inner
import bodo
import bodo.libs.timsort
from bodo.libs.array import arr_info_list_to_table, array_to_info, delete_table, delete_table_decref_arrays, info_from_table, info_to_array, sort_values_table
from bodo.libs.str_arr_ext import cp_str_list_to_array, to_list_if_immutable_arr
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.transforms.distributed_analysis import Distribution
from bodo.utils.utils import debug_prints, gen_getitem
MIN_SAMPLES = 1000000
samplePointsPerPartitionHint = 20
MPI_ROOT = 0


class Sort(ir.Stmt):

    def __init__(self, df_in, df_out, key_arrs, out_key_arrs, df_in_vars,
        df_out_vars, inplace, loc, ascending_list=True, na_position='last'):
        self.df_in = df_in
        self.df_out = df_out
        self.key_arrs = key_arrs
        self.out_key_arrs = out_key_arrs
        self.df_in_vars = df_in_vars
        self.df_out_vars = df_out_vars
        self.inplace = inplace
        if isinstance(na_position, str):
            if na_position == 'last':
                self.na_position_b = (True,) * len(key_arrs)
            else:
                self.na_position_b = (False,) * len(key_arrs)
        else:
            self.na_position_b = tuple([(True if etowl__glyph == 'last' else
                False) for etowl__glyph in na_position])
        if isinstance(ascending_list, bool):
            ascending_list = (ascending_list,) * len(key_arrs)
        self.ascending_list = ascending_list
        self.loc = loc

    def __repr__(self):
        uis__zotr = ''
        for phbjr__eisn, pmq__woqs in self.df_in_vars.items():
            uis__zotr += "'{}':{}, ".format(phbjr__eisn, pmq__woqs.name)
        dhunf__sjiph = '{}{{{}}}'.format(self.df_in, uis__zotr)
        sxx__gda = ''
        for phbjr__eisn, pmq__woqs in self.df_out_vars.items():
            sxx__gda += "'{}':{}, ".format(phbjr__eisn, pmq__woqs.name)
        eet__wia = '{}{{{}}}'.format(self.df_out, sxx__gda)
        return 'sort: [key: {}] {} [key: {}] {}'.format(', '.join(pmq__woqs
            .name for pmq__woqs in self.key_arrs), dhunf__sjiph, ', '.join(
            pmq__woqs.name for pmq__woqs in self.out_key_arrs), eet__wia)


def sort_array_analysis(sort_node, equiv_set, typemap, array_analysis):
    fcjmg__kxq = []
    sfv__ezbng = sort_node.key_arrs + list(sort_node.df_in_vars.values())
    for fxz__zhit in sfv__ezbng:
        wmct__zuq = equiv_set.get_shape(fxz__zhit)
        if wmct__zuq is not None:
            fcjmg__kxq.append(wmct__zuq[0])
    if len(fcjmg__kxq) > 1:
        equiv_set.insert_equiv(*fcjmg__kxq)
    abcw__erlg = []
    fcjmg__kxq = []
    jtdx__rawdt = sort_node.out_key_arrs + list(sort_node.df_out_vars.values())
    for fxz__zhit in jtdx__rawdt:
        faylm__slh = typemap[fxz__zhit.name]
        pridl__uxn = array_analysis._gen_shape_call(equiv_set, fxz__zhit,
            faylm__slh.ndim, None, abcw__erlg)
        equiv_set.insert_equiv(fxz__zhit, pridl__uxn)
        fcjmg__kxq.append(pridl__uxn[0])
        equiv_set.define(fxz__zhit, set())
    if len(fcjmg__kxq) > 1:
        equiv_set.insert_equiv(*fcjmg__kxq)
    return [], abcw__erlg


numba.parfors.array_analysis.array_analysis_extensions[Sort
    ] = sort_array_analysis


def sort_distributed_analysis(sort_node, array_dists):
    sfv__ezbng = sort_node.key_arrs + list(sort_node.df_in_vars.values())
    etj__hrzfa = sort_node.out_key_arrs + list(sort_node.df_out_vars.values())
    xlwz__jajce = Distribution.OneD
    for fxz__zhit in sfv__ezbng:
        xlwz__jajce = Distribution(min(xlwz__jajce.value, array_dists[
            fxz__zhit.name].value))
    aqom__mekjo = Distribution(min(xlwz__jajce.value, Distribution.OneD_Var
        .value))
    for fxz__zhit in etj__hrzfa:
        if fxz__zhit.name in array_dists:
            aqom__mekjo = Distribution(min(aqom__mekjo.value, array_dists[
                fxz__zhit.name].value))
    if aqom__mekjo != Distribution.OneD_Var:
        xlwz__jajce = aqom__mekjo
    for fxz__zhit in sfv__ezbng:
        array_dists[fxz__zhit.name] = xlwz__jajce
    for fxz__zhit in etj__hrzfa:
        array_dists[fxz__zhit.name] = aqom__mekjo
    return


distributed_analysis.distributed_analysis_extensions[Sort
    ] = sort_distributed_analysis


def sort_typeinfer(sort_node, typeinferer):
    for wih__qhy, nqkup__luv in zip(sort_node.key_arrs, sort_node.out_key_arrs
        ):
        typeinferer.constraints.append(typeinfer.Propagate(dst=nqkup__luv.
            name, src=wih__qhy.name, loc=sort_node.loc))
    for ptuw__wtmcb, fxz__zhit in sort_node.df_in_vars.items():
        slk__ahoto = sort_node.df_out_vars[ptuw__wtmcb]
        typeinferer.constraints.append(typeinfer.Propagate(dst=slk__ahoto.
            name, src=fxz__zhit.name, loc=sort_node.loc))
    return


typeinfer.typeinfer_extensions[Sort] = sort_typeinfer


def build_sort_definitions(sort_node, definitions=None):
    if definitions is None:
        definitions = defaultdict(list)
    if not sort_node.inplace:
        for fxz__zhit in (sort_node.out_key_arrs + list(sort_node.
            df_out_vars.values())):
            definitions[fxz__zhit.name].append(sort_node)
    return definitions


ir_utils.build_defs_extensions[Sort] = build_sort_definitions


def visit_vars_sort(sort_node, callback, cbdata):
    if debug_prints():
        print('visiting sort vars for:', sort_node)
        print('cbdata: ', sorted(cbdata.items()))
    for kny__uygui in range(len(sort_node.key_arrs)):
        sort_node.key_arrs[kny__uygui] = visit_vars_inner(sort_node.
            key_arrs[kny__uygui], callback, cbdata)
        sort_node.out_key_arrs[kny__uygui] = visit_vars_inner(sort_node.
            out_key_arrs[kny__uygui], callback, cbdata)
    for ptuw__wtmcb in list(sort_node.df_in_vars.keys()):
        sort_node.df_in_vars[ptuw__wtmcb] = visit_vars_inner(sort_node.
            df_in_vars[ptuw__wtmcb], callback, cbdata)
    for ptuw__wtmcb in list(sort_node.df_out_vars.keys()):
        sort_node.df_out_vars[ptuw__wtmcb] = visit_vars_inner(sort_node.
            df_out_vars[ptuw__wtmcb], callback, cbdata)


ir_utils.visit_vars_extensions[Sort] = visit_vars_sort


def remove_dead_sort(sort_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    uot__dac = []
    for ptuw__wtmcb, fxz__zhit in sort_node.df_out_vars.items():
        if fxz__zhit.name not in lives:
            uot__dac.append(ptuw__wtmcb)
    for vhh__eehj in uot__dac:
        sort_node.df_in_vars.pop(vhh__eehj)
        sort_node.df_out_vars.pop(vhh__eehj)
    if len(sort_node.df_out_vars) == 0 and all(pmq__woqs.name not in lives for
        pmq__woqs in sort_node.out_key_arrs):
        return None
    return sort_node


ir_utils.remove_dead_extensions[Sort] = remove_dead_sort


def sort_usedefs(sort_node, use_set=None, def_set=None):
    if use_set is None:
        use_set = set()
    if def_set is None:
        def_set = set()
    use_set.update({pmq__woqs.name for pmq__woqs in sort_node.key_arrs})
    use_set.update({pmq__woqs.name for pmq__woqs in sort_node.df_in_vars.
        values()})
    if not sort_node.inplace:
        def_set.update({pmq__woqs.name for pmq__woqs in sort_node.out_key_arrs}
            )
        def_set.update({pmq__woqs.name for pmq__woqs in sort_node.
            df_out_vars.values()})
    return numba.core.analysis._use_defs_result(usemap=use_set, defmap=def_set)


numba.core.analysis.ir_extension_usedefs[Sort] = sort_usedefs


def get_copies_sort(sort_node, typemap):
    cqa__htpxc = set()
    if not sort_node.inplace:
        cqa__htpxc = set(pmq__woqs.name for pmq__woqs in sort_node.
            df_out_vars.values())
        cqa__htpxc.update({pmq__woqs.name for pmq__woqs in sort_node.
            out_key_arrs})
    return set(), cqa__htpxc


ir_utils.copy_propagate_extensions[Sort] = get_copies_sort


def apply_copies_sort(sort_node, var_dict, name_var_table, typemap,
    calltypes, save_copies):
    for kny__uygui in range(len(sort_node.key_arrs)):
        sort_node.key_arrs[kny__uygui] = replace_vars_inner(sort_node.
            key_arrs[kny__uygui], var_dict)
        sort_node.out_key_arrs[kny__uygui] = replace_vars_inner(sort_node.
            out_key_arrs[kny__uygui], var_dict)
    for ptuw__wtmcb in list(sort_node.df_in_vars.keys()):
        sort_node.df_in_vars[ptuw__wtmcb] = replace_vars_inner(sort_node.
            df_in_vars[ptuw__wtmcb], var_dict)
    for ptuw__wtmcb in list(sort_node.df_out_vars.keys()):
        sort_node.df_out_vars[ptuw__wtmcb] = replace_vars_inner(sort_node.
            df_out_vars[ptuw__wtmcb], var_dict)
    return


ir_utils.apply_copy_propagate_extensions[Sort] = apply_copies_sort


def sort_distributed_run(sort_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    jili__togwp = False
    mbjl__uwp = list(sort_node.df_in_vars.values())
    jtdx__rawdt = list(sort_node.df_out_vars.values())
    if array_dists is not None:
        jili__togwp = True
        for pmq__woqs in (sort_node.key_arrs + sort_node.out_key_arrs +
            mbjl__uwp + jtdx__rawdt):
            if array_dists[pmq__woqs.name
                ] != distributed_pass.Distribution.OneD and array_dists[
                pmq__woqs.name] != distributed_pass.Distribution.OneD_Var:
                jili__togwp = False
    loc = sort_node.loc
    qisk__xcln = sort_node.key_arrs[0].scope
    nodes = []
    key_arrs = sort_node.key_arrs
    if not sort_node.inplace:
        nmrkv__tvc = []
        for pmq__woqs in key_arrs:
            vxpa__gfcdd = _copy_array_nodes(pmq__woqs, nodes, typingctx,
                targetctx, typemap, calltypes)
            nmrkv__tvc.append(vxpa__gfcdd)
        key_arrs = nmrkv__tvc
        qmrh__beyen = []
        for pmq__woqs in mbjl__uwp:
            nfuh__eyoe = _copy_array_nodes(pmq__woqs, nodes, typingctx,
                targetctx, typemap, calltypes)
            qmrh__beyen.append(nfuh__eyoe)
        mbjl__uwp = qmrh__beyen
    key_name_args = [('key' + str(kny__uygui)) for kny__uygui in range(len(
        key_arrs))]
    zqol__jewo = ', '.join(key_name_args)
    col_name_args = [('c' + str(kny__uygui)) for kny__uygui in range(len(
        mbjl__uwp))]
    aih__jlyd = ', '.join(col_name_args)
    qdvsn__yqa = 'def f({}, {}):\n'.format(zqol__jewo, aih__jlyd)
    qdvsn__yqa += get_sort_cpp_section(key_name_args, col_name_args,
        sort_node.ascending_list, sort_node.na_position_b, jili__togwp)
    qdvsn__yqa += '  return key_arrs, data\n'
    rdpks__axqbd = {}
    exec(qdvsn__yqa, {}, rdpks__axqbd)
    tekz__nhvyi = rdpks__axqbd['f']
    frdgm__lsem = types.Tuple([typemap[pmq__woqs.name] for pmq__woqs in
        key_arrs])
    gptbv__ncv = types.Tuple([typemap[pmq__woqs.name] for pmq__woqs in
        mbjl__uwp])
    yrn__ckzh = compile_to_numba_ir(tekz__nhvyi, {'bodo': bodo, 'np': np,
        'to_list_if_immutable_arr': to_list_if_immutable_arr,
        'cp_str_list_to_array': cp_str_list_to_array, 'delete_table':
        delete_table, 'delete_table_decref_arrays':
        delete_table_decref_arrays, 'info_to_array': info_to_array,
        'info_from_table': info_from_table, 'sort_values_table':
        sort_values_table, 'arr_info_list_to_table': arr_info_list_to_table,
        'array_to_info': array_to_info}, typingctx=typingctx, targetctx=
        targetctx, arg_typs=tuple(list(frdgm__lsem.types) + list(gptbv__ncv
        .types)), typemap=typemap, calltypes=calltypes).blocks.popitem()[1]
    replace_arg_nodes(yrn__ckzh, key_arrs + mbjl__uwp)
    nodes += yrn__ckzh.body[:-2]
    hbp__nuowm = nodes[-1].target
    ssl__rtwkh = ir.Var(qisk__xcln, mk_unique_var('key_data'), loc)
    typemap[ssl__rtwkh.name] = frdgm__lsem
    gen_getitem(ssl__rtwkh, hbp__nuowm, 0, calltypes, nodes)
    icmtx__kcy = ir.Var(qisk__xcln, mk_unique_var('sort_data'), loc)
    typemap[icmtx__kcy.name] = gptbv__ncv
    gen_getitem(icmtx__kcy, hbp__nuowm, 1, calltypes, nodes)
    for kny__uygui, var in enumerate(sort_node.out_key_arrs):
        gen_getitem(var, ssl__rtwkh, kny__uygui, calltypes, nodes)
    for kny__uygui, var in enumerate(jtdx__rawdt):
        gen_getitem(var, icmtx__kcy, kny__uygui, calltypes, nodes)
    return nodes


distributed_pass.distributed_run_extensions[Sort] = sort_distributed_run


def _copy_array_nodes(var, nodes, typingctx, targetctx, typemap, calltypes):

    def _impl(arr):
        return arr.copy()
    yrn__ckzh = compile_to_numba_ir(_impl, {}, typingctx=typingctx,
        targetctx=targetctx, arg_typs=(typemap[var.name],), typemap=typemap,
        calltypes=calltypes).blocks.popitem()[1]
    replace_arg_nodes(yrn__ckzh, [var])
    nodes += yrn__ckzh.body[:-2]
    return nodes[-1].target


def get_sort_cpp_section(key_name_args, col_name_args, ascending_list,
    na_position_b, parallel_b):
    ebjxa__alps = len(key_name_args)
    lafp__pnjh = ['array_to_info({})'.format(cily__rng) for cily__rng in
        key_name_args] + ['array_to_info({})'.format(cily__rng) for
        cily__rng in col_name_args]
    qdvsn__yqa = '  info_list_total = [{}]\n'.format(','.join(lafp__pnjh))
    qdvsn__yqa += '  table_total = arr_info_list_to_table(info_list_total)\n'
    qdvsn__yqa += '  vect_ascending = np.array([{}])\n'.format(','.join('1' if
        deb__dgkex else '0' for deb__dgkex in ascending_list))
    qdvsn__yqa += '  na_position = np.array([{}])\n'.format(','.join('1' if
        deb__dgkex else '0' for deb__dgkex in na_position_b))
    qdvsn__yqa += (
        """  out_table = sort_values_table(table_total, {}, vect_ascending.ctypes, na_position.ctypes, {})
"""
        .format(ebjxa__alps, parallel_b))
    kti__iptf = 0
    fesc__sjly = []
    for cily__rng in key_name_args:
        fesc__sjly.append('info_to_array(info_from_table(out_table, {}), {})'
            .format(kti__iptf, cily__rng))
        kti__iptf += 1
    qdvsn__yqa += '  key_arrs = ({},)\n'.format(','.join(fesc__sjly))
    yvnk__eyg = []
    for cily__rng in col_name_args:
        yvnk__eyg.append('info_to_array(info_from_table(out_table, {}), {})'
            .format(kti__iptf, cily__rng))
        kti__iptf += 1
    if len(yvnk__eyg) > 0:
        qdvsn__yqa += '  data = ({},)\n'.format(','.join(yvnk__eyg))
    else:
        qdvsn__yqa += '  data = ()\n'
    qdvsn__yqa += '  delete_table(out_table)\n'
    qdvsn__yqa += '  delete_table(table_total)\n'
    return qdvsn__yqa
