"""Tools for handling bodo arrays, e.g. passing to C/C++ code
"""
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.typing.templates import signature
from numba.cpython.listobj import ListInstance
from numba.extending import intrinsic, models, register_model
from numba.np.arrayobj import _getitem_array_single_int
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, get_categories_int_type
from bodo.libs import array_ext
from bodo.libs.array_item_arr_ext import ArrayItemArrayPayloadType, ArrayItemArrayType, _get_array_item_arr_payload, define_array_item_dtor, offset_type
from bodo.libs.binary_arr_ext import binary_array_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType, int128_type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.interval_arr_ext import IntervalArrayType
from bodo.libs.map_arr_ext import MapArrayType, _get_map_arr_data_type, init_map_arr_codegen
from bodo.libs.str_arr_ext import _get_str_binary_arr_payload, char_arr_type, null_bitmap_arr_type, offset_arr_type, string_array_type
from bodo.libs.struct_arr_ext import StructArrayPayloadType, StructArrayType, StructType, _get_struct_arr_payload, define_struct_arr_dtor
from bodo.libs.tuple_arr_ext import TupleArrayType
from bodo.utils.typing import BodoError, MetaType
from bodo.utils.utils import CTypeEnum, check_and_propagate_cpp_exception, numba_to_c_type
ll.add_symbol('list_string_array_to_info', array_ext.list_string_array_to_info)
ll.add_symbol('nested_array_to_info', array_ext.nested_array_to_info)
ll.add_symbol('string_array_to_info', array_ext.string_array_to_info)
ll.add_symbol('numpy_array_to_info', array_ext.numpy_array_to_info)
ll.add_symbol('categorical_array_to_info', array_ext.categorical_array_to_info)
ll.add_symbol('nullable_array_to_info', array_ext.nullable_array_to_info)
ll.add_symbol('interval_array_to_info', array_ext.interval_array_to_info)
ll.add_symbol('decimal_array_to_info', array_ext.decimal_array_to_info)
ll.add_symbol('info_to_nested_array', array_ext.info_to_nested_array)
ll.add_symbol('info_to_list_string_array', array_ext.info_to_list_string_array)
ll.add_symbol('info_to_string_array', array_ext.info_to_string_array)
ll.add_symbol('info_to_numpy_array', array_ext.info_to_numpy_array)
ll.add_symbol('info_to_nullable_array', array_ext.info_to_nullable_array)
ll.add_symbol('info_to_interval_array', array_ext.info_to_interval_array)
ll.add_symbol('alloc_numpy', array_ext.alloc_numpy)
ll.add_symbol('alloc_string_array', array_ext.alloc_string_array)
ll.add_symbol('arr_info_list_to_table', array_ext.arr_info_list_to_table)
ll.add_symbol('info_from_table', array_ext.info_from_table)
ll.add_symbol('delete_info_decref_array', array_ext.delete_info_decref_array)
ll.add_symbol('delete_table_decref_arrays', array_ext.
    delete_table_decref_arrays)
ll.add_symbol('delete_table', array_ext.delete_table)
ll.add_symbol('shuffle_table', array_ext.shuffle_table)
ll.add_symbol('get_shuffle_info', array_ext.get_shuffle_info)
ll.add_symbol('delete_shuffle_info', array_ext.delete_shuffle_info)
ll.add_symbol('reverse_shuffle_table', array_ext.reverse_shuffle_table)
ll.add_symbol('hash_join_table', array_ext.hash_join_table)
ll.add_symbol('drop_duplicates_table', array_ext.drop_duplicates_table)
ll.add_symbol('sort_values_table', array_ext.sort_values_table)
ll.add_symbol('sample_table', array_ext.sample_table)
ll.add_symbol('shuffle_renormalization', array_ext.shuffle_renormalization)
ll.add_symbol('shuffle_renormalization_group', array_ext.
    shuffle_renormalization_group)
ll.add_symbol('groupby_and_aggregate', array_ext.groupby_and_aggregate)
ll.add_symbol('pivot_groupby_and_aggregate', array_ext.
    pivot_groupby_and_aggregate)
ll.add_symbol('get_groupby_labels', array_ext.get_groupby_labels)
ll.add_symbol('array_isin', array_ext.array_isin)
ll.add_symbol('get_search_regex', array_ext.get_search_regex)
ll.add_symbol('compute_node_partition_by_hash', array_ext.
    compute_node_partition_by_hash)
ll.add_symbol('array_info_getitem', array_ext.array_info_getitem)


class ArrayInfoType(types.Type):

    def __init__(self):
        super(ArrayInfoType, self).__init__(name='ArrayInfoType()')


array_info_type = ArrayInfoType()
register_model(ArrayInfoType)(models.OpaqueModel)


class TableTypeCPP(types.Type):

    def __init__(self):
        super(TableTypeCPP, self).__init__(name='TableTypeCPP()')


table_type = TableTypeCPP()
register_model(TableTypeCPP)(models.OpaqueModel)


@intrinsic
def array_to_info(typingctx, arr_type_t=None):
    return array_info_type(arr_type_t), array_to_info_codegen


def array_to_info_codegen(context, builder, sig, args):
    in_arr, = args
    arr_type = sig.args[0]
    if isinstance(arr_type, TupleArrayType):
        smgg__hzcay = context.make_helper(builder, arr_type, in_arr)
        in_arr = smgg__hzcay.data
        arr_type = StructArrayType(arr_type.data, ('dummy',) * len(arr_type
            .data))
    context.nrt.incref(builder, arr_type, in_arr)
    if isinstance(arr_type, ArrayItemArrayType
        ) and arr_type.dtype == string_array_type:
        tnh__zof = context.make_helper(builder, arr_type, in_arr)
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='list_string_array_to_info')
        return builder.call(nou__rocu, [tnh__zof.meminfo])
    if isinstance(arr_type, (MapArrayType, ArrayItemArrayType, StructArrayType)
        ):

        def get_types(arr_typ):
            if isinstance(arr_typ, MapArrayType):
                return get_types(_get_map_arr_data_type(arr_typ))
            elif isinstance(arr_typ, ArrayItemArrayType):
                return [CTypeEnum.LIST.value] + get_types(arr_typ.dtype)
            elif isinstance(arr_typ, (StructType, StructArrayType)):
                thu__mbhs = [CTypeEnum.STRUCT.value, len(arr_typ.names)]
                for whzeo__sykb in arr_typ.data:
                    thu__mbhs += get_types(whzeo__sykb)
                return thu__mbhs
            elif isinstance(arr_typ, (types.Array, IntegerArrayType)
                ) or arr_typ == boolean_array:
                return get_types(arr_typ.dtype)
            elif arr_typ == string_array_type:
                return [CTypeEnum.STRING.value]
            elif arr_typ == binary_array_type:
                return [CTypeEnum.BINARY.value]
            elif isinstance(arr_typ, DecimalArrayType):
                return [CTypeEnum.Decimal.value, arr_typ.precision, arr_typ
                    .scale]
            else:
                return [numba_to_c_type(arr_typ)]

        def get_lengths(arr_typ, arr):
            uvtw__hzhlf = context.compile_internal(builder, lambda a: len(a
                ), types.intp(arr_typ), [arr])
            if isinstance(arr_typ, MapArrayType):
                tmyk__njzp = context.make_helper(builder, arr_typ, value=arr)
                zjqyw__cnsm = get_lengths(_get_map_arr_data_type(arr_typ),
                    tmyk__njzp.data)
            elif isinstance(arr_typ, ArrayItemArrayType):
                jzkje__cifg = _get_array_item_arr_payload(context, builder,
                    arr_typ, arr)
                zjqyw__cnsm = get_lengths(arr_typ.dtype, jzkje__cifg.data)
                zjqyw__cnsm = cgutils.pack_array(builder, [jzkje__cifg.
                    n_arrays] + [builder.extract_value(zjqyw__cnsm,
                    zvhj__lqars) for zvhj__lqars in range(zjqyw__cnsm.type.
                    count)])
            elif isinstance(arr_typ, StructArrayType):
                jzkje__cifg = _get_struct_arr_payload(context, builder,
                    arr_typ, arr)
                zjqyw__cnsm = []
                for zvhj__lqars, whzeo__sykb in enumerate(arr_typ.data):
                    jzie__xdq = get_lengths(whzeo__sykb, builder.
                        extract_value(jzkje__cifg.data, zvhj__lqars))
                    zjqyw__cnsm += [builder.extract_value(jzie__xdq,
                        snzn__rdoz) for snzn__rdoz in range(jzie__xdq.type.
                        count)]
                zjqyw__cnsm = cgutils.pack_array(builder, [uvtw__hzhlf,
                    context.get_constant(types.int64, -1)] + zjqyw__cnsm)
            elif isinstance(arr_typ, (IntegerArrayType, DecimalArrayType,
                types.Array)) or arr_typ in (boolean_array,
                datetime_date_array_type, string_array_type, binary_array_type
                ):
                zjqyw__cnsm = cgutils.pack_array(builder, [uvtw__hzhlf])
            else:
                raise RuntimeError(
                    'array_to_info: unsupported type for subarray')
            return zjqyw__cnsm

        def get_buffers(arr_typ, arr):
            if isinstance(arr_typ, MapArrayType):
                tmyk__njzp = context.make_helper(builder, arr_typ, value=arr)
                kbw__ckm = get_buffers(_get_map_arr_data_type(arr_typ),
                    tmyk__njzp.data)
            elif isinstance(arr_typ, ArrayItemArrayType):
                jzkje__cifg = _get_array_item_arr_payload(context, builder,
                    arr_typ, arr)
                uzmpu__slj = get_buffers(arr_typ.dtype, jzkje__cifg.data)
                dnd__kxfti = context.make_array(types.Array(offset_type, 1,
                    'C'))(context, builder, jzkje__cifg.offsets)
                gha__myn = builder.bitcast(dnd__kxfti.data, lir.IntType(8).
                    as_pointer())
                xhwls__rfuh = context.make_array(types.Array(types.uint8, 1,
                    'C'))(context, builder, jzkje__cifg.null_bitmap)
                iau__hoq = builder.bitcast(xhwls__rfuh.data, lir.IntType(8)
                    .as_pointer())
                kbw__ckm = cgutils.pack_array(builder, [gha__myn, iau__hoq] +
                    [builder.extract_value(uzmpu__slj, zvhj__lqars) for
                    zvhj__lqars in range(uzmpu__slj.type.count)])
            elif isinstance(arr_typ, StructArrayType):
                jzkje__cifg = _get_struct_arr_payload(context, builder,
                    arr_typ, arr)
                uzmpu__slj = []
                for zvhj__lqars, whzeo__sykb in enumerate(arr_typ.data):
                    abkwu__fanqs = get_buffers(whzeo__sykb, builder.
                        extract_value(jzkje__cifg.data, zvhj__lqars))
                    uzmpu__slj += [builder.extract_value(abkwu__fanqs,
                        snzn__rdoz) for snzn__rdoz in range(abkwu__fanqs.
                        type.count)]
                xhwls__rfuh = context.make_array(types.Array(types.uint8, 1,
                    'C'))(context, builder, jzkje__cifg.null_bitmap)
                iau__hoq = builder.bitcast(xhwls__rfuh.data, lir.IntType(8)
                    .as_pointer())
                kbw__ckm = cgutils.pack_array(builder, [iau__hoq] + uzmpu__slj)
            elif isinstance(arr_typ, (IntegerArrayType, DecimalArrayType)
                ) or arr_typ in (boolean_array, datetime_date_array_type):
                tjzqi__myo = arr_typ.dtype
                if isinstance(arr_typ, DecimalArrayType):
                    tjzqi__myo = int128_type
                elif arr_typ == datetime_date_array_type:
                    tjzqi__myo = types.int64
                arr = cgutils.create_struct_proxy(arr_typ)(context, builder,
                    arr)
                rfeaf__qdfv = context.make_array(types.Array(tjzqi__myo, 1,
                    'C'))(context, builder, arr.data)
                xhwls__rfuh = context.make_array(types.Array(types.uint8, 1,
                    'C'))(context, builder, arr.null_bitmap)
                fokr__mtwo = builder.bitcast(rfeaf__qdfv.data, lir.IntType(
                    8).as_pointer())
                iau__hoq = builder.bitcast(xhwls__rfuh.data, lir.IntType(8)
                    .as_pointer())
                kbw__ckm = cgutils.pack_array(builder, [iau__hoq, fokr__mtwo])
            elif arr_typ in (string_array_type, binary_array_type):
                jzkje__cifg = _get_str_binary_arr_payload(context, builder,
                    arr, arr_typ)
                vmxv__ggb = context.make_helper(builder, offset_arr_type,
                    jzkje__cifg.offsets).data
                wncex__dchin = context.make_helper(builder, char_arr_type,
                    jzkje__cifg.data).data
                jhvt__jdgss = context.make_helper(builder,
                    null_bitmap_arr_type, jzkje__cifg.null_bitmap).data
                kbw__ckm = cgutils.pack_array(builder, [builder.bitcast(
                    vmxv__ggb, lir.IntType(8).as_pointer()), builder.
                    bitcast(jhvt__jdgss, lir.IntType(8).as_pointer()),
                    builder.bitcast(wncex__dchin, lir.IntType(8).as_pointer())]
                    )
            elif isinstance(arr_typ, types.Array):
                arr = context.make_array(arr_typ)(context, builder, arr)
                fokr__mtwo = builder.bitcast(arr.data, lir.IntType(8).
                    as_pointer())
                ztfvg__obg = lir.Constant(lir.IntType(8).as_pointer(), None)
                kbw__ckm = cgutils.pack_array(builder, [ztfvg__obg, fokr__mtwo]
                    )
            else:
                raise RuntimeError(
                    'array_to_info: unsupported type for subarray ' + str(
                    arr_typ))
            return kbw__ckm

        def get_field_names(arr_typ):
            gcsa__jngf = []
            if isinstance(arr_typ, StructArrayType):
                for faum__mhteg, uixzp__ajob in zip(arr_typ.dtype.names,
                    arr_typ.data):
                    gcsa__jngf.append(faum__mhteg)
                    gcsa__jngf += get_field_names(uixzp__ajob)
            elif isinstance(arr_typ, ArrayItemArrayType):
                gcsa__jngf += get_field_names(arr_typ.dtype)
            elif isinstance(arr_typ, MapArrayType):
                gcsa__jngf += get_field_names(_get_map_arr_data_type(arr_typ))
            return gcsa__jngf
        thu__mbhs = get_types(arr_type)
        hoap__ezjna = cgutils.pack_array(builder, [context.get_constant(
            types.int32, t) for t in thu__mbhs])
        qtwfm__dkpjb = cgutils.alloca_once_value(builder, hoap__ezjna)
        zjqyw__cnsm = get_lengths(arr_type, in_arr)
        lengths_ptr = cgutils.alloca_once_value(builder, zjqyw__cnsm)
        kbw__ckm = get_buffers(arr_type, in_arr)
        frdt__gtelm = cgutils.alloca_once_value(builder, kbw__ckm)
        gcsa__jngf = get_field_names(arr_type)
        if len(gcsa__jngf) == 0:
            gcsa__jngf = ['irrelevant']
        yadm__wcz = cgutils.pack_array(builder, [context.
            insert_const_string(builder.module, a) for a in gcsa__jngf])
        dmx__roe = cgutils.alloca_once_value(builder, yadm__wcz)
        if isinstance(arr_type, MapArrayType):
            xaptz__lmu = _get_map_arr_data_type(arr_type)
            uyo__yfkzb = context.make_helper(builder, arr_type, value=in_arr)
            pen__uvkwj = uyo__yfkzb.data
        else:
            xaptz__lmu = arr_type
            pen__uvkwj = in_arr
        sjkls__gxcaa = context.make_helper(builder, xaptz__lmu, pen__uvkwj)
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(32).as_pointer(), lir.IntType(8).as_pointer().
            as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='nested_array_to_info')
        orv__bpgiz = builder.call(nou__rocu, [builder.bitcast(qtwfm__dkpjb,
            lir.IntType(32).as_pointer()), builder.bitcast(frdt__gtelm, lir
            .IntType(8).as_pointer().as_pointer()), builder.bitcast(
            lengths_ptr, lir.IntType(64).as_pointer()), builder.bitcast(
            dmx__roe, lir.IntType(8).as_pointer()), sjkls__gxcaa.meminfo])
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    if arr_type in (string_array_type, binary_array_type):
        imc__dgr = context.make_helper(builder, arr_type, in_arr)
        vjx__zwr = ArrayItemArrayType(char_arr_type)
        tnh__zof = context.make_helper(builder, vjx__zwr, imc__dgr.data)
        jzkje__cifg = _get_str_binary_arr_payload(context, builder, in_arr,
            arr_type)
        vmxv__ggb = context.make_helper(builder, offset_arr_type,
            jzkje__cifg.offsets).data
        wncex__dchin = context.make_helper(builder, char_arr_type,
            jzkje__cifg.data).data
        jhvt__jdgss = context.make_helper(builder, null_bitmap_arr_type,
            jzkje__cifg.null_bitmap).data
        ghplq__yqt = builder.zext(builder.load(builder.gep(vmxv__ggb, [
            jzkje__cifg.n_arrays])), lir.IntType(64))
        rwjw__ycsi = context.get_constant(types.int32, int(arr_type ==
            binary_array_type))
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(64), lir.IntType(64), lir.IntType(8).as_pointer(), lir.
            IntType(offset_type.bitwidth).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(32)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='string_array_to_info')
        return builder.call(nou__rocu, [jzkje__cifg.n_arrays, ghplq__yqt,
            wncex__dchin, vmxv__ggb, jhvt__jdgss, tnh__zof.meminfo, rwjw__ycsi]
            )
    qpe__veno = False
    if isinstance(arr_type, CategoricalArrayType):
        context.nrt.decref(builder, arr_type, in_arr)
        wddeu__fsc = context.compile_internal(builder, lambda a: len(a.
            dtype.categories), types.intp(arr_type), [in_arr])
        in_arr = cgutils.create_struct_proxy(arr_type)(context, builder, in_arr
            ).codes
        sjelv__kbghg = get_categories_int_type(arr_type.dtype)
        arr_type = types.Array(sjelv__kbghg, 1, 'C')
        qpe__veno = True
        context.nrt.incref(builder, arr_type, in_arr)
    if isinstance(arr_type, types.Array):
        arr = context.make_array(arr_type)(context, builder, in_arr)
        assert arr_type.ndim == 1, 'only 1D array shuffle supported'
        uvtw__hzhlf = builder.extract_value(arr.shape, 0)
        xewna__cjrzv = arr_type.dtype
        lbiq__uleho = numba_to_c_type(xewna__cjrzv)
        ijdoz__buycr = cgutils.alloca_once_value(builder, lir.Constant(lir.
            IntType(32), lbiq__uleho))
        if qpe__veno:
            wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir
                .IntType(64), lir.IntType(8).as_pointer(), lir.IntType(32),
                lir.IntType(64), lir.IntType(8).as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='categorical_array_to_info')
            return builder.call(nou__rocu, [uvtw__hzhlf, builder.bitcast(
                arr.data, lir.IntType(8).as_pointer()), builder.load(
                ijdoz__buycr), wddeu__fsc, arr.meminfo])
        else:
            wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir
                .IntType(64), lir.IntType(8).as_pointer(), lir.IntType(32),
                lir.IntType(8).as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='numpy_array_to_info')
            return builder.call(nou__rocu, [uvtw__hzhlf, builder.bitcast(
                arr.data, lir.IntType(8).as_pointer()), builder.load(
                ijdoz__buycr), arr.meminfo])
    if isinstance(arr_type, (IntegerArrayType, DecimalArrayType)
        ) or arr_type in (boolean_array, datetime_date_array_type):
        arr = cgutils.create_struct_proxy(arr_type)(context, builder, in_arr)
        xewna__cjrzv = arr_type.dtype
        tjzqi__myo = xewna__cjrzv
        if isinstance(arr_type, DecimalArrayType):
            tjzqi__myo = int128_type
        if arr_type == datetime_date_array_type:
            tjzqi__myo = types.int64
        rfeaf__qdfv = context.make_array(types.Array(tjzqi__myo, 1, 'C'))(
            context, builder, arr.data)
        uvtw__hzhlf = builder.extract_value(rfeaf__qdfv.shape, 0)
        xrnm__frtp = context.make_array(types.Array(types.uint8, 1, 'C'))(
            context, builder, arr.null_bitmap)
        lbiq__uleho = numba_to_c_type(xewna__cjrzv)
        ijdoz__buycr = cgutils.alloca_once_value(builder, lir.Constant(lir.
            IntType(32), lbiq__uleho))
        if isinstance(arr_type, DecimalArrayType):
            wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir
                .IntType(64), lir.IntType(8).as_pointer(), lir.IntType(32),
                lir.IntType(8).as_pointer(), lir.IntType(8).as_pointer(),
                lir.IntType(8).as_pointer(), lir.IntType(32), lir.IntType(32)])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='decimal_array_to_info')
            return builder.call(nou__rocu, [uvtw__hzhlf, builder.bitcast(
                rfeaf__qdfv.data, lir.IntType(8).as_pointer()), builder.
                load(ijdoz__buycr), builder.bitcast(xrnm__frtp.data, lir.
                IntType(8).as_pointer()), rfeaf__qdfv.meminfo, xrnm__frtp.
                meminfo, context.get_constant(types.int32, arr_type.
                precision), context.get_constant(types.int32, arr_type.scale)])
        else:
            wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir
                .IntType(64), lir.IntType(8).as_pointer(), lir.IntType(32),
                lir.IntType(8).as_pointer(), lir.IntType(8).as_pointer(),
                lir.IntType(8).as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='nullable_array_to_info')
            return builder.call(nou__rocu, [uvtw__hzhlf, builder.bitcast(
                rfeaf__qdfv.data, lir.IntType(8).as_pointer()), builder.
                load(ijdoz__buycr), builder.bitcast(xrnm__frtp.data, lir.
                IntType(8).as_pointer()), rfeaf__qdfv.meminfo, xrnm__frtp.
                meminfo])
    if isinstance(arr_type, IntervalArrayType):
        assert isinstance(arr_type.arr_type, types.Array
            ), 'array_to_info(): only IntervalArrayType with Numpy arrays supported'
        arr = cgutils.create_struct_proxy(arr_type)(context, builder, in_arr)
        wxlyu__qlkb = context.make_array(arr_type.arr_type)(context,
            builder, arr.left)
        zkz__bov = context.make_array(arr_type.arr_type)(context, builder,
            arr.right)
        uvtw__hzhlf = builder.extract_value(wxlyu__qlkb.shape, 0)
        lbiq__uleho = numba_to_c_type(arr_type.arr_type.dtype)
        ijdoz__buycr = cgutils.alloca_once_value(builder, lir.Constant(lir.
            IntType(32), lbiq__uleho))
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(64), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(32), lir.IntType(8).as_pointer(), lir
            .IntType(8).as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='interval_array_to_info')
        return builder.call(nou__rocu, [uvtw__hzhlf, builder.bitcast(
            wxlyu__qlkb.data, lir.IntType(8).as_pointer()), builder.bitcast
            (zkz__bov.data, lir.IntType(8).as_pointer()), builder.load(
            ijdoz__buycr), wxlyu__qlkb.meminfo, zkz__bov.meminfo])
    raise BodoError(f'array_to_info(): array type {arr_type} is not supported')


def _lower_info_to_array_numpy(arr_type, context, builder, in_info):
    assert arr_type.ndim == 1, 'only 1D array supported'
    arr = context.make_array(arr_type)(context, builder)
    yywc__xwz = cgutils.alloca_once(builder, lir.IntType(64))
    fokr__mtwo = cgutils.alloca_once(builder, lir.IntType(8).as_pointer())
    kgzk__ekads = cgutils.alloca_once(builder, lir.IntType(8).as_pointer())
    wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
        as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(8).
        as_pointer().as_pointer(), lir.IntType(8).as_pointer().as_pointer()])
    nou__rocu = cgutils.get_or_insert_function(builder.module, wuuw__xmxm,
        name='info_to_numpy_array')
    builder.call(nou__rocu, [in_info, yywc__xwz, fokr__mtwo, kgzk__ekads])
    context.compile_internal(builder, lambda :
        check_and_propagate_cpp_exception(), types.none(), [])
    rcxuu__mxvq = context.get_value_type(types.intp)
    iuxz__tbb = cgutils.pack_array(builder, [builder.load(yywc__xwz)], ty=
        rcxuu__mxvq)
    wkaz__zylta = context.get_constant(types.intp, context.get_abi_sizeof(
        context.get_data_type(arr_type.dtype)))
    mllh__coim = cgutils.pack_array(builder, [wkaz__zylta], ty=rcxuu__mxvq)
    wncex__dchin = builder.bitcast(builder.load(fokr__mtwo), context.
        get_data_type(arr_type.dtype).as_pointer())
    numba.np.arrayobj.populate_array(arr, data=wncex__dchin, shape=
        iuxz__tbb, strides=mllh__coim, itemsize=wkaz__zylta, meminfo=
        builder.load(kgzk__ekads))
    return arr._getvalue()


def _lower_info_to_array_list_string_array(arr_type, context, builder, in_info
    ):
    rsgxc__zcfm = context.make_helper(builder, arr_type)
    wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
        as_pointer(), lir.IntType(8).as_pointer().as_pointer()])
    nou__rocu = cgutils.get_or_insert_function(builder.module, wuuw__xmxm,
        name='info_to_list_string_array')
    builder.call(nou__rocu, [in_info, rsgxc__zcfm._get_ptr_by_name('meminfo')])
    context.compile_internal(builder, lambda :
        check_and_propagate_cpp_exception(), types.none(), [])
    return rsgxc__zcfm._getvalue()


def nested_to_array(context, builder, arr_typ, lengths_ptr, array_infos_ptr,
    lengths_pos, infos_pos):
    amvv__tliw = context.get_data_type(array_info_type)
    if isinstance(arr_typ, ArrayItemArrayType):
        ekr__ibk = lengths_pos
        bihgi__apv = infos_pos
        xwjt__vpm, lengths_pos, infos_pos = nested_to_array(context,
            builder, arr_typ.dtype, lengths_ptr, array_infos_ptr, 
            lengths_pos + 1, infos_pos + 2)
        yahgn__uobk = ArrayItemArrayPayloadType(arr_typ)
        fqpyp__tws = context.get_data_type(yahgn__uobk)
        ylrc__duk = context.get_abi_sizeof(fqpyp__tws)
        ljk__fsqtw = define_array_item_dtor(context, builder, arr_typ,
            yahgn__uobk)
        abdlw__lms = context.nrt.meminfo_alloc_dtor(builder, context.
            get_constant(types.uintp, ylrc__duk), ljk__fsqtw)
        wgok__dctm = context.nrt.meminfo_data(builder, abdlw__lms)
        emlvc__aoib = builder.bitcast(wgok__dctm, fqpyp__tws.as_pointer())
        jzkje__cifg = cgutils.create_struct_proxy(yahgn__uobk)(context, builder
            )
        jzkje__cifg.n_arrays = builder.extract_value(builder.load(
            lengths_ptr), ekr__ibk)
        jzkje__cifg.data = xwjt__vpm
        qzp__kxco = builder.load(array_infos_ptr)
        nnlz__xzkn = builder.bitcast(builder.extract_value(qzp__kxco,
            bihgi__apv), amvv__tliw)
        jzkje__cifg.offsets = _lower_info_to_array_numpy(types.Array(
            offset_type, 1, 'C'), context, builder, nnlz__xzkn)
        xll__dzj = builder.bitcast(builder.extract_value(qzp__kxco, 
            bihgi__apv + 1), amvv__tliw)
        jzkje__cifg.null_bitmap = _lower_info_to_array_numpy(types.Array(
            types.uint8, 1, 'C'), context, builder, xll__dzj)
        builder.store(jzkje__cifg._getvalue(), emlvc__aoib)
        tnh__zof = context.make_helper(builder, arr_typ)
        tnh__zof.meminfo = abdlw__lms
        return tnh__zof._getvalue(), lengths_pos, infos_pos
    elif isinstance(arr_typ, StructArrayType):
        csfd__xkxj = []
        bihgi__apv = infos_pos
        lengths_pos += 1
        infos_pos += 1
        for fke__enq in arr_typ.data:
            xwjt__vpm, lengths_pos, infos_pos = nested_to_array(context,
                builder, fke__enq, lengths_ptr, array_infos_ptr,
                lengths_pos, infos_pos)
            csfd__xkxj.append(xwjt__vpm)
        yahgn__uobk = StructArrayPayloadType(arr_typ.data)
        fqpyp__tws = context.get_value_type(yahgn__uobk)
        ylrc__duk = context.get_abi_sizeof(fqpyp__tws)
        ljk__fsqtw = define_struct_arr_dtor(context, builder, arr_typ,
            yahgn__uobk)
        abdlw__lms = context.nrt.meminfo_alloc_dtor(builder, context.
            get_constant(types.uintp, ylrc__duk), ljk__fsqtw)
        wgok__dctm = context.nrt.meminfo_data(builder, abdlw__lms)
        emlvc__aoib = builder.bitcast(wgok__dctm, fqpyp__tws.as_pointer())
        jzkje__cifg = cgutils.create_struct_proxy(yahgn__uobk)(context, builder
            )
        jzkje__cifg.data = cgutils.pack_array(builder, csfd__xkxj
            ) if types.is_homogeneous(*arr_typ.data) else cgutils.pack_struct(
            builder, csfd__xkxj)
        qzp__kxco = builder.load(array_infos_ptr)
        xll__dzj = builder.bitcast(builder.extract_value(qzp__kxco,
            bihgi__apv), amvv__tliw)
        jzkje__cifg.null_bitmap = _lower_info_to_array_numpy(types.Array(
            types.uint8, 1, 'C'), context, builder, xll__dzj)
        builder.store(jzkje__cifg._getvalue(), emlvc__aoib)
        ppdu__atax = context.make_helper(builder, arr_typ)
        ppdu__atax.meminfo = abdlw__lms
        return ppdu__atax._getvalue(), lengths_pos, infos_pos
    elif arr_typ in (string_array_type, binary_array_type):
        qzp__kxco = builder.load(array_infos_ptr)
        uzff__naudu = builder.bitcast(builder.extract_value(qzp__kxco,
            infos_pos), amvv__tliw)
        imc__dgr = context.make_helper(builder, arr_typ)
        vjx__zwr = ArrayItemArrayType(char_arr_type)
        tnh__zof = context.make_helper(builder, vjx__zwr)
        wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer().as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='info_to_string_array')
        builder.call(nou__rocu, [uzff__naudu, tnh__zof._get_ptr_by_name(
            'meminfo')])
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        imc__dgr.data = tnh__zof._getvalue()
        return imc__dgr._getvalue(), lengths_pos + 1, infos_pos + 1
    elif isinstance(arr_typ, types.Array):
        qzp__kxco = builder.load(array_infos_ptr)
        lqe__wovd = builder.bitcast(builder.extract_value(qzp__kxco, 
            infos_pos + 1), amvv__tliw)
        return _lower_info_to_array_numpy(arr_typ, context, builder, lqe__wovd
            ), lengths_pos + 1, infos_pos + 2
    elif isinstance(arr_typ, (IntegerArrayType, DecimalArrayType)
        ) or arr_typ in (boolean_array, datetime_date_array_type):
        arr = cgutils.create_struct_proxy(arr_typ)(context, builder)
        tjzqi__myo = arr_typ.dtype
        if isinstance(arr_typ, DecimalArrayType):
            tjzqi__myo = int128_type
        elif arr_typ == datetime_date_array_type:
            tjzqi__myo = types.int64
        qzp__kxco = builder.load(array_infos_ptr)
        xll__dzj = builder.bitcast(builder.extract_value(qzp__kxco,
            infos_pos), amvv__tliw)
        arr.null_bitmap = _lower_info_to_array_numpy(types.Array(types.
            uint8, 1, 'C'), context, builder, xll__dzj)
        lqe__wovd = builder.bitcast(builder.extract_value(qzp__kxco, 
            infos_pos + 1), amvv__tliw)
        arr.data = _lower_info_to_array_numpy(types.Array(tjzqi__myo, 1,
            'C'), context, builder, lqe__wovd)
        return arr._getvalue(), lengths_pos + 1, infos_pos + 2


@intrinsic
def info_to_array(typingctx, info_type, array_type):
    arr_type = array_type.instance_type if isinstance(array_type, types.TypeRef
        ) else array_type
    assert info_type == array_info_type

    def codegen(context, builder, sig, args):
        in_info, apv__gbr = args
        if isinstance(arr_type, ArrayItemArrayType
            ) and arr_type.dtype == string_array_type:
            return _lower_info_to_array_list_string_array(arr_type, context,
                builder, in_info)
        if isinstance(arr_type, (MapArrayType, ArrayItemArrayType,
            StructArrayType, TupleArrayType)):

            def get_num_arrays(arr_typ):
                if isinstance(arr_typ, ArrayItemArrayType):
                    return 1 + get_num_arrays(arr_typ.dtype)
                elif isinstance(arr_typ, StructArrayType):
                    return 1 + sum([get_num_arrays(fke__enq) for fke__enq in
                        arr_typ.data])
                else:
                    return 1

            def get_num_infos(arr_typ):
                if isinstance(arr_typ, ArrayItemArrayType):
                    return 2 + get_num_infos(arr_typ.dtype)
                elif isinstance(arr_typ, StructArrayType):
                    return 1 + sum([get_num_infos(fke__enq) for fke__enq in
                        arr_typ.data])
                elif arr_typ in (string_array_type, binary_array_type):
                    return 1
                else:
                    return 2
            if isinstance(arr_type, TupleArrayType):
                zjm__ucf = StructArrayType(arr_type.data, ('dummy',) * len(
                    arr_type.data))
            elif isinstance(arr_type, MapArrayType):
                zjm__ucf = _get_map_arr_data_type(arr_type)
            else:
                zjm__ucf = arr_type
            ank__vthjj = get_num_arrays(zjm__ucf)
            zjqyw__cnsm = cgutils.pack_array(builder, [lir.Constant(lir.
                IntType(64), 0) for apv__gbr in range(ank__vthjj)])
            lengths_ptr = cgutils.alloca_once_value(builder, zjqyw__cnsm)
            ztfvg__obg = lir.Constant(lir.IntType(8).as_pointer(), None)
            fhmyl__ltlf = cgutils.pack_array(builder, [ztfvg__obg for
                apv__gbr in range(get_num_infos(zjm__ucf))])
            array_infos_ptr = cgutils.alloca_once_value(builder, fhmyl__ltlf)
            wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(8).
                as_pointer().as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='info_to_nested_array')
            builder.call(nou__rocu, [in_info, builder.bitcast(lengths_ptr,
                lir.IntType(64).as_pointer()), builder.bitcast(
                array_infos_ptr, lir.IntType(8).as_pointer().as_pointer())])
            context.compile_internal(builder, lambda :
                check_and_propagate_cpp_exception(), types.none(), [])
            arr, apv__gbr, apv__gbr = nested_to_array(context, builder,
                zjm__ucf, lengths_ptr, array_infos_ptr, 0, 0)
            if isinstance(arr_type, TupleArrayType):
                smgg__hzcay = context.make_helper(builder, arr_type)
                smgg__hzcay.data = arr
                context.nrt.incref(builder, zjm__ucf, arr)
                arr = smgg__hzcay._getvalue()
            elif isinstance(arr_type, MapArrayType):
                sig = signature(arr_type, zjm__ucf)
                arr = init_map_arr_codegen(context, builder, sig, (arr,))
            return arr
        if arr_type in (string_array_type, binary_array_type):
            imc__dgr = context.make_helper(builder, arr_type)
            vjx__zwr = ArrayItemArrayType(char_arr_type)
            tnh__zof = context.make_helper(builder, vjx__zwr)
            wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer(), lir.IntType(8).as_pointer().as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='info_to_string_array')
            builder.call(nou__rocu, [in_info, tnh__zof._get_ptr_by_name(
                'meminfo')])
            context.compile_internal(builder, lambda :
                check_and_propagate_cpp_exception(), types.none(), [])
            imc__dgr.data = tnh__zof._getvalue()
            return imc__dgr._getvalue()
        if isinstance(arr_type, CategoricalArrayType):
            out_arr = cgutils.create_struct_proxy(arr_type)(context, builder)
            sjelv__kbghg = get_categories_int_type(arr_type.dtype)
            ddkyp__wjnec = types.Array(sjelv__kbghg, 1, 'C')
            out_arr.codes = _lower_info_to_array_numpy(ddkyp__wjnec,
                context, builder, in_info)
            if isinstance(array_type, types.TypeRef):
                assert arr_type.dtype.categories is not None, 'info_to_array: unknown categories'
                is_ordered = arr_type.dtype.ordered
                donmt__vgsdx = pd.CategoricalDtype(arr_type.dtype.
                    categories, is_ordered).categories.values
                new_cats_tup = MetaType(tuple(donmt__vgsdx))
                int_type = arr_type.dtype.int_type
                pivl__rpewg = bodo.typeof(donmt__vgsdx)
                dwyew__yuqb = context.get_constant_generic(builder,
                    pivl__rpewg, donmt__vgsdx)
                xewna__cjrzv = context.compile_internal(builder, lambda
                    c_arr: bodo.hiframes.pd_categorical_ext.init_cat_dtype(
                    bodo.utils.conversion.index_from_array(c_arr),
                    is_ordered, int_type, new_cats_tup), arr_type.dtype(
                    pivl__rpewg), [dwyew__yuqb])
            else:
                xewna__cjrzv = cgutils.create_struct_proxy(arr_type)(context,
                    builder, args[1]).dtype
                context.nrt.incref(builder, arr_type.dtype, xewna__cjrzv)
            out_arr.dtype = xewna__cjrzv
            return out_arr._getvalue()
        if isinstance(arr_type, types.Array):
            return _lower_info_to_array_numpy(arr_type, context, builder,
                in_info)
        if isinstance(arr_type, (IntegerArrayType, DecimalArrayType)
            ) or arr_type in (boolean_array, datetime_date_array_type):
            arr = cgutils.create_struct_proxy(arr_type)(context, builder)
            tjzqi__myo = arr_type.dtype
            if isinstance(arr_type, DecimalArrayType):
                tjzqi__myo = int128_type
            elif arr_type == datetime_date_array_type:
                tjzqi__myo = types.int64
            ymqqb__ira = types.Array(tjzqi__myo, 1, 'C')
            rfeaf__qdfv = context.make_array(ymqqb__ira)(context, builder)
            zwh__utjty = types.Array(types.uint8, 1, 'C')
            ptd__kxdun = context.make_array(zwh__utjty)(context, builder)
            yywc__xwz = cgutils.alloca_once(builder, lir.IntType(64))
            iwg__pmh = cgutils.alloca_once(builder, lir.IntType(64))
            fokr__mtwo = cgutils.alloca_once(builder, lir.IntType(8).
                as_pointer())
            bpb__ggx = cgutils.alloca_once(builder, lir.IntType(8).as_pointer()
                )
            kgzk__ekads = cgutils.alloca_once(builder, lir.IntType(8).
                as_pointer())
            hrf__cdkpj = cgutils.alloca_once(builder, lir.IntType(8).
                as_pointer())
            wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(64)
                .as_pointer(), lir.IntType(8).as_pointer().as_pointer(),
                lir.IntType(8).as_pointer().as_pointer(), lir.IntType(8).
                as_pointer().as_pointer(), lir.IntType(8).as_pointer().
                as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='info_to_nullable_array')
            builder.call(nou__rocu, [in_info, yywc__xwz, iwg__pmh,
                fokr__mtwo, bpb__ggx, kgzk__ekads, hrf__cdkpj])
            context.compile_internal(builder, lambda :
                check_and_propagate_cpp_exception(), types.none(), [])
            rcxuu__mxvq = context.get_value_type(types.intp)
            iuxz__tbb = cgutils.pack_array(builder, [builder.load(yywc__xwz
                )], ty=rcxuu__mxvq)
            wkaz__zylta = context.get_constant(types.intp, context.
                get_abi_sizeof(context.get_data_type(tjzqi__myo)))
            mllh__coim = cgutils.pack_array(builder, [wkaz__zylta], ty=
                rcxuu__mxvq)
            wncex__dchin = builder.bitcast(builder.load(fokr__mtwo),
                context.get_data_type(tjzqi__myo).as_pointer())
            numba.np.arrayobj.populate_array(rfeaf__qdfv, data=wncex__dchin,
                shape=iuxz__tbb, strides=mllh__coim, itemsize=wkaz__zylta,
                meminfo=builder.load(kgzk__ekads))
            arr.data = rfeaf__qdfv._getvalue()
            iuxz__tbb = cgutils.pack_array(builder, [builder.load(iwg__pmh)
                ], ty=rcxuu__mxvq)
            wkaz__zylta = context.get_constant(types.intp, context.
                get_abi_sizeof(context.get_data_type(types.uint8)))
            mllh__coim = cgutils.pack_array(builder, [wkaz__zylta], ty=
                rcxuu__mxvq)
            wncex__dchin = builder.bitcast(builder.load(bpb__ggx), context.
                get_data_type(types.uint8).as_pointer())
            numba.np.arrayobj.populate_array(ptd__kxdun, data=wncex__dchin,
                shape=iuxz__tbb, strides=mllh__coim, itemsize=wkaz__zylta,
                meminfo=builder.load(hrf__cdkpj))
            arr.null_bitmap = ptd__kxdun._getvalue()
            return arr._getvalue()
        if isinstance(arr_type, IntervalArrayType):
            arr = cgutils.create_struct_proxy(arr_type)(context, builder)
            wxlyu__qlkb = context.make_array(arr_type.arr_type)(context,
                builder)
            zkz__bov = context.make_array(arr_type.arr_type)(context, builder)
            yywc__xwz = cgutils.alloca_once(builder, lir.IntType(64))
            hqd__odjty = cgutils.alloca_once(builder, lir.IntType(8).
                as_pointer())
            nas__mqh = cgutils.alloca_once(builder, lir.IntType(8).as_pointer()
                )
            fbs__uyh = cgutils.alloca_once(builder, lir.IntType(8).as_pointer()
                )
            azzck__puo = cgutils.alloca_once(builder, lir.IntType(8).
                as_pointer())
            wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(8).
                as_pointer().as_pointer(), lir.IntType(8).as_pointer().
                as_pointer(), lir.IntType(8).as_pointer().as_pointer(), lir
                .IntType(8).as_pointer().as_pointer()])
            nou__rocu = cgutils.get_or_insert_function(builder.module,
                wuuw__xmxm, name='info_to_interval_array')
            builder.call(nou__rocu, [in_info, yywc__xwz, hqd__odjty,
                nas__mqh, fbs__uyh, azzck__puo])
            context.compile_internal(builder, lambda :
                check_and_propagate_cpp_exception(), types.none(), [])
            rcxuu__mxvq = context.get_value_type(types.intp)
            iuxz__tbb = cgutils.pack_array(builder, [builder.load(yywc__xwz
                )], ty=rcxuu__mxvq)
            wkaz__zylta = context.get_constant(types.intp, context.
                get_abi_sizeof(context.get_data_type(arr_type.arr_type.dtype)))
            mllh__coim = cgutils.pack_array(builder, [wkaz__zylta], ty=
                rcxuu__mxvq)
            enzp__qvip = builder.bitcast(builder.load(hqd__odjty), context.
                get_data_type(arr_type.arr_type.dtype).as_pointer())
            numba.np.arrayobj.populate_array(wxlyu__qlkb, data=enzp__qvip,
                shape=iuxz__tbb, strides=mllh__coim, itemsize=wkaz__zylta,
                meminfo=builder.load(fbs__uyh))
            arr.left = wxlyu__qlkb._getvalue()
            ixdht__ibd = builder.bitcast(builder.load(nas__mqh), context.
                get_data_type(arr_type.arr_type.dtype).as_pointer())
            numba.np.arrayobj.populate_array(zkz__bov, data=ixdht__ibd,
                shape=iuxz__tbb, strides=mllh__coim, itemsize=wkaz__zylta,
                meminfo=builder.load(azzck__puo))
            arr.right = zkz__bov._getvalue()
            return arr._getvalue()
        raise BodoError(
            f'info_to_array(): array type {arr_type} is not supported')
    return arr_type(info_type, array_type), codegen


@intrinsic
def test_alloc_np(typingctx, len_typ, arr_type):
    array_type = arr_type.instance_type if isinstance(arr_type, types.TypeRef
        ) else arr_type

    def codegen(context, builder, sig, args):
        uvtw__hzhlf, apv__gbr = args
        lbiq__uleho = numba_to_c_type(array_type.dtype)
        ijdoz__buycr = cgutils.alloca_once_value(builder, lir.Constant(lir.
            IntType(32), lbiq__uleho))
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(64), lir.IntType(32)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='alloc_numpy')
        return builder.call(nou__rocu, [uvtw__hzhlf, builder.load(
            ijdoz__buycr)])
    return array_info_type(len_typ, arr_type), codegen


@intrinsic
def test_alloc_string(typingctx, len_typ, n_chars_typ):

    def codegen(context, builder, sig, args):
        uvtw__hzhlf, gsye__pma = args
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(64), lir.IntType(64)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='alloc_string_array')
        return builder.call(nou__rocu, [uvtw__hzhlf, gsye__pma])
    return array_info_type(len_typ, n_chars_typ), codegen


@intrinsic
def arr_info_list_to_table(typingctx, list_arr_info_typ=None):
    assert list_arr_info_typ == types.List(array_info_type)
    return table_type(list_arr_info_typ), arr_info_list_to_table_codegen


def arr_info_list_to_table_codegen(context, builder, sig, args):
    ydzl__iaha, = args
    jqep__mpuu = numba.cpython.listobj.ListInstance(context, builder, sig.
        args[0], ydzl__iaha)
    wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.IntType
        (8).as_pointer().as_pointer(), lir.IntType(64)])
    nou__rocu = cgutils.get_or_insert_function(builder.module, wuuw__xmxm,
        name='arr_info_list_to_table')
    return builder.call(nou__rocu, [jqep__mpuu.data, jqep__mpuu.size])


@intrinsic
def info_from_table(typingctx, table_t, ind_t):

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='info_from_table')
        return builder.call(nou__rocu, args)
    return array_info_type(table_t, ind_t), codegen


@intrinsic
def cpp_table_to_py_table(typingctx, cpp_table_t, table_idx_arr_t,
    py_table_type_t):
    assert cpp_table_t == table_type, 'invalid cpp table type'
    assert isinstance(table_idx_arr_t, types.Array
        ) and table_idx_arr_t.dtype == types.int64, 'invalid table index array'
    assert isinstance(py_table_type_t, types.TypeRef), 'invalid py table ref'
    vqs__jano = py_table_type_t.instance_type

    def codegen(context, builder, sig, args):
        uhif__bpg, flfni__mbdok, apv__gbr = args
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='info_from_table')
        oly__aaz = cgutils.create_struct_proxy(vqs__jano)(context, builder)
        oly__aaz.parent = cgutils.get_null_value(oly__aaz.parent.type)
        iwk__yqyoo = context.make_array(table_idx_arr_t)(context, builder,
            flfni__mbdok)
        budl__rzh = context.get_constant(types.int64, -1)
        mln__ykazs = context.get_constant(types.int64, 0)
        muh__mqjh = cgutils.alloca_once_value(builder, mln__ykazs)
        for t, gnds__jeuu in vqs__jano.type_to_blk.items():
            ifuvq__bzi = context.get_constant(types.int64, len(vqs__jano.
                block_to_arr_ind[gnds__jeuu]))
            apv__gbr, bldkq__iwmua = ListInstance.allocate_ex(context,
                builder, types.List(t), ifuvq__bzi)
            bldkq__iwmua.size = ifuvq__bzi
            kffzs__dwic = context.make_constant_array(builder, types.Array(
                types.int64, 1, 'C'), np.array(vqs__jano.block_to_arr_ind[
                gnds__jeuu], dtype=np.int64))
            pbd__hdu = context.make_array(types.Array(types.int64, 1, 'C'))(
                context, builder, kffzs__dwic)
            with cgutils.for_range(builder, ifuvq__bzi) as loop:
                zvhj__lqars = loop.index
                upttv__legw = _getitem_array_single_int(context, builder,
                    types.int64, types.Array(types.int64, 1, 'C'), pbd__hdu,
                    zvhj__lqars)
                psfo__vbigh = _getitem_array_single_int(context, builder,
                    types.int64, table_idx_arr_t, iwk__yqyoo, upttv__legw)
                ndxn__qieb = builder.icmp_unsigned('!=', psfo__vbigh, budl__rzh
                    )
                with builder.if_else(ndxn__qieb) as (then, orelse):
                    with then:
                        hfmrb__qsly = builder.call(nou__rocu, [uhif__bpg,
                            psfo__vbigh])
                        arr = context.compile_internal(builder, lambda info:
                            info_to_array(info, t), t(array_info_type), [
                            hfmrb__qsly])
                        bldkq__iwmua.inititem(zvhj__lqars, arr, incref=False)
                        uvtw__hzhlf = context.compile_internal(builder, lambda
                            arr: len(arr), types.int64(t), [arr])
                        builder.store(uvtw__hzhlf, muh__mqjh)
                    with orelse:
                        aas__fko = context.get_constant_null(t)
                        bldkq__iwmua.inititem(zvhj__lqars, aas__fko, incref
                            =False)
            setattr(oly__aaz, f'block_{gnds__jeuu}', bldkq__iwmua.value)
        oly__aaz.len = builder.load(muh__mqjh)
        return oly__aaz._getvalue()
    return vqs__jano(cpp_table_t, table_idx_arr_t, py_table_type_t), codegen


@intrinsic
def py_table_to_cpp_table(typingctx, py_table_t, py_table_type_t):
    assert isinstance(py_table_t, bodo.hiframes.table.TableType
        ), 'invalid py table type'
    assert isinstance(py_table_type_t, types.TypeRef), 'invalid py table ref'
    vqs__jano = py_table_type_t.instance_type

    def codegen(context, builder, sig, args):
        qkumf__otuzg, apv__gbr = args
        buljs__negsz = lir.Constant(lir.IntType(64), len(vqs__jano.arr_types))
        apv__gbr, wahqu__tuzgr = ListInstance.allocate_ex(context, builder,
            types.List(array_info_type), buljs__negsz)
        wahqu__tuzgr.size = buljs__negsz
        vio__slq = cgutils.create_struct_proxy(vqs__jano)(context, builder,
            qkumf__otuzg)
        for t, gnds__jeuu in vqs__jano.type_to_blk.items():
            ifuvq__bzi = context.get_constant(types.int64, len(vqs__jano.
                block_to_arr_ind[gnds__jeuu]))
            xwrx__ipuj = getattr(vio__slq, f'block_{gnds__jeuu}')
            ebt__hjt = ListInstance(context, builder, types.List(t), xwrx__ipuj
                )
            kffzs__dwic = context.make_constant_array(builder, types.Array(
                types.int64, 1, 'C'), np.array(vqs__jano.block_to_arr_ind[
                gnds__jeuu], dtype=np.int64))
            pbd__hdu = context.make_array(types.Array(types.int64, 1, 'C'))(
                context, builder, kffzs__dwic)
            with cgutils.for_range(builder, ifuvq__bzi) as loop:
                zvhj__lqars = loop.index
                upttv__legw = _getitem_array_single_int(context, builder,
                    types.int64, types.Array(types.int64, 1, 'C'), pbd__hdu,
                    zvhj__lqars)
                rvu__xsnx = signature(types.none, vqs__jano, types.List(t),
                    types.int64, types.int64)
                fhip__tra = qkumf__otuzg, xwrx__ipuj, zvhj__lqars, upttv__legw
                bodo.hiframes.table.ensure_column_unboxed_codegen(context,
                    builder, rvu__xsnx, fhip__tra)
                arr = ebt__hjt.getitem(zvhj__lqars)
                lrsn__pwjhn = signature(array_info_type, t)
                rfwj__hhy = arr,
                gmp__brtkh = array_to_info_codegen(context, builder,
                    lrsn__pwjhn, rfwj__hhy)
                wahqu__tuzgr.inititem(upttv__legw, gmp__brtkh, incref=False)
        yai__vmgp = wahqu__tuzgr.value
        bnly__rzunz = signature(table_type, types.List(array_info_type))
        zpt__sqe = yai__vmgp,
        uhif__bpg = arr_info_list_to_table_codegen(context, builder,
            bnly__rzunz, zpt__sqe)
        context.nrt.decref(builder, types.List(array_info_type), yai__vmgp)
        return uhif__bpg
    return table_type(vqs__jano, py_table_type_t), codegen


delete_info_decref_array = types.ExternalFunction('delete_info_decref_array',
    types.void(array_info_type))
delete_table_decref_arrays = types.ExternalFunction(
    'delete_table_decref_arrays', types.void(table_type))


@intrinsic
def delete_table(typingctx, table_t=None):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='delete_table')
        builder.call(nou__rocu, args)
    return types.void(table_t), codegen


@intrinsic
def shuffle_table(typingctx, table_t, n_keys_t, _is_parallel, keep_comm_info_t
    ):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(1), lir.
            IntType(32)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='shuffle_table')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int64, types.boolean, types.int32
        ), codegen


class ShuffleInfoType(types.Type):

    def __init__(self):
        super(ShuffleInfoType, self).__init__(name='ShuffleInfoType()')


shuffle_info_type = ShuffleInfoType()
register_model(ShuffleInfoType)(models.OpaqueModel)
get_shuffle_info = types.ExternalFunction('get_shuffle_info',
    shuffle_info_type(table_type))
delete_shuffle_info = types.ExternalFunction('delete_shuffle_info', types.
    void(shuffle_info_type))
reverse_shuffle_table = types.ExternalFunction('reverse_shuffle_table',
    table_type(table_type, shuffle_info_type))


@intrinsic
def get_null_shuffle_info(typingctx):

    def codegen(context, builder, sig, args):
        return context.get_constant_null(sig.return_type)
    return shuffle_info_type(), codegen


@intrinsic
def hash_join_table(typingctx, left_table_t, right_table_t, left_parallel_t,
    right_parallel_t, n_keys_t, n_data_left_t, n_data_right_t, same_vect_t,
    same_need_typechange_t, is_left_t, is_right_t, is_join_t,
    optional_col_t, indicator, _bodo_na_equal, cond_func, left_col_nums,
    left_col_nums_len, right_col_nums, right_col_nums_len):
    assert left_table_t == table_type
    assert right_table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(1), lir.IntType(1), lir.IntType(64), lir.IntType(64),
            lir.IntType(64), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(1), lir.IntType(1), lir.IntType(1),
            lir.IntType(1), lir.IntType(1), lir.IntType(1), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(64), lir
            .IntType(8).as_pointer(), lir.IntType(64)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='hash_join_table')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(left_table_t, right_table_t, types.boolean, types.
        boolean, types.int64, types.int64, types.int64, types.voidptr,
        types.voidptr, types.boolean, types.boolean, types.boolean, types.
        boolean, types.boolean, types.boolean, types.voidptr, types.voidptr,
        types.int64, types.voidptr, types.int64), codegen


@intrinsic
def compute_node_partition_by_hash(typingctx, table_t, n_keys_t, n_pes_t):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(64)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='compute_node_partition_by_hash')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int64, types.int64), codegen


@intrinsic
def sort_values_table(typingctx, table_t, n_keys_t, vect_ascending_t,
    na_position_b_t, parallel_t):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(1)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='sort_values_table')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int64, types.voidptr, types.voidptr,
        types.boolean), codegen


@intrinsic
def sample_table(typingctx, table_t, n_keys_t, frac_t, replace_t, parallel_t):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.DoubleType(), lir
            .IntType(1), lir.IntType(1)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='sample_table')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int64, types.float64, types.boolean,
        types.boolean), codegen


@intrinsic
def shuffle_renormalization(typingctx, table_t, random_t, random_seed_t,
    is_parallel_t):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(32), lir.IntType(64), lir.
            IntType(1)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='shuffle_renormalization')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int32, types.int64, types.boolean
        ), codegen


@intrinsic
def shuffle_renormalization_group(typingctx, table_t, random_t,
    random_seed_t, is_parallel_t, num_ranks_t, ranks_t):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(32), lir.IntType(64), lir.
            IntType(1), lir.IntType(64), lir.IntType(8).as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='shuffle_renormalization_group')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int32, types.int64, types.boolean,
        types.int64, types.voidptr), codegen


@intrinsic
def drop_duplicates_table(typingctx, table_t, parallel_t, nkey_t, keep_t,
    dropna):
    assert table_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(1), lir.IntType(64), lir.
            IntType(64), lir.IntType(1)])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='drop_duplicates_table')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.boolean, types.int64, types.int64,
        types.boolean), codegen


@intrinsic
def pivot_groupby_and_aggregate(typingctx, table_t, n_keys_t,
    dispatch_table_t, dispatch_info_t, input_has_index, ftypes,
    func_offsets, udf_n_redvars, is_parallel, is_crosstab, skipdropna_t,
    return_keys, return_index, update_cb, combine_cb, eval_cb,
    udf_table_dummy_t):
    assert table_t == table_type
    assert dispatch_table_t == table_type
    assert dispatch_info_t == table_type
    assert udf_table_dummy_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(1), lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(8).as_pointer(), lir.IntType(1), lir.IntType(1), lir.
            IntType(1), lir.IntType(1), lir.IntType(1), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='pivot_groupby_and_aggregate')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int64, table_t, table_t, types.boolean,
        types.voidptr, types.voidptr, types.voidptr, types.boolean, types.
        boolean, types.boolean, types.boolean, types.boolean, types.voidptr,
        types.voidptr, types.voidptr, table_t), codegen


@intrinsic
def groupby_and_aggregate(typingctx, table_t, n_keys_t, input_has_index,
    ftypes, func_offsets, udf_n_redvars, is_parallel, skipdropna_t,
    shift_periods_t, transform_func, head_n, return_keys, return_index,
    dropna, update_cb, combine_cb, eval_cb, general_udfs_cb, udf_table_dummy_t
    ):
    assert table_t == table_type
    assert udf_table_dummy_t == table_type

    def codegen(context, builder, sig, args):
        wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(1), lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(8).as_pointer(), lir.IntType(1), lir.IntType(1), lir.
            IntType(64), lir.IntType(64), lir.IntType(64), lir.IntType(1),
            lir.IntType(1), lir.IntType(1), lir.IntType(8).as_pointer(),
            lir.IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer()])
        nou__rocu = cgutils.get_or_insert_function(builder.module,
            wuuw__xmxm, name='groupby_and_aggregate')
        orv__bpgiz = builder.call(nou__rocu, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        return orv__bpgiz
    return table_type(table_t, types.int64, types.boolean, types.voidptr,
        types.voidptr, types.voidptr, types.boolean, types.boolean, types.
        int64, types.int64, types.int64, types.boolean, types.boolean,
        types.boolean, types.voidptr, types.voidptr, types.voidptr, types.
        voidptr, table_t), codegen


get_groupby_labels = types.ExternalFunction('get_groupby_labels', types.
    int64(table_type, types.voidptr, types.voidptr, types.boolean, types.bool_)
    )
_array_isin = types.ExternalFunction('array_isin', types.void(
    array_info_type, array_info_type, array_info_type, types.bool_))


@numba.njit
def array_isin(out_arr, in_arr, in_values, is_parallel):
    vay__qaa = array_to_info(in_arr)
    tbufw__udr = array_to_info(in_values)
    kboib__mbldd = array_to_info(out_arr)
    okqzf__bdzw = arr_info_list_to_table([vay__qaa, tbufw__udr, kboib__mbldd])
    _array_isin(kboib__mbldd, vay__qaa, tbufw__udr, is_parallel)
    check_and_propagate_cpp_exception()
    delete_table(okqzf__bdzw)


_get_search_regex = types.ExternalFunction('get_search_regex', types.void(
    array_info_type, types.bool_, types.voidptr, array_info_type))


@numba.njit
def get_search_regex(in_arr, case, pat, out_arr):
    vay__qaa = array_to_info(in_arr)
    kboib__mbldd = array_to_info(out_arr)
    _get_search_regex(vay__qaa, case, pat, kboib__mbldd)
    check_and_propagate_cpp_exception()


def _gen_row_access_intrinsic(col_dtype, c_ind):
    from llvmlite import ir as lir
    if isinstance(col_dtype, types.Number) or col_dtype in [bodo.
        datetime_date_type, bodo.datetime64ns, bodo.timedelta64ns, types.bool_
        ]:

        @intrinsic
        def getitem_func(typingctx, table_t, ind_t):

            def codegen(context, builder, sig, args):
                oly__aaz, kqi__gynbr = args
                oly__aaz = builder.bitcast(oly__aaz, lir.IntType(8).
                    as_pointer().as_pointer())
                cpowj__lqbws = lir.Constant(lir.IntType(64), c_ind)
                qgook__fxctu = builder.load(builder.gep(oly__aaz, [
                    cpowj__lqbws]))
                qgook__fxctu = builder.bitcast(qgook__fxctu, context.
                    get_data_type(col_dtype).as_pointer())
                return builder.load(builder.gep(qgook__fxctu, [kqi__gynbr]))
            return col_dtype(types.voidptr, types.int64), codegen
        return getitem_func
    if col_dtype == types.unicode_type:

        @intrinsic
        def getitem_func(typingctx, table_t, ind_t):

            def codegen(context, builder, sig, args):
                oly__aaz, kqi__gynbr = args
                wuuw__xmxm = lir.FunctionType(lir.IntType(8).as_pointer(),
                    [lir.IntType(8).as_pointer(), lir.IntType(64), lir.
                    IntType(64), lir.IntType(64).as_pointer()])
                dnfbn__dtyv = cgutils.get_or_insert_function(builder.module,
                    wuuw__xmxm, name='array_info_getitem')
                cpowj__lqbws = lir.Constant(lir.IntType(64), c_ind)
                otb__folbd = cgutils.alloca_once(builder, lir.IntType(64))
                args = oly__aaz, cpowj__lqbws, kqi__gynbr, otb__folbd
                fokr__mtwo = builder.call(dnfbn__dtyv, args)
                return context.make_tuple(builder, sig.return_type, [
                    fokr__mtwo, builder.load(otb__folbd)])
            return types.Tuple([types.voidptr, types.int64])(types.voidptr,
                types.int64), codegen
        return getitem_func
    raise BodoError(
        f"General Join Conditions with '{col_dtype}' column data type not supported"
        )


def _gen_row_na_check_intrinsic(col_array_dtype, c_ind):
    if isinstance(col_array_dtype, bodo.libs.int_arr_ext.IntegerArrayType
        ) or col_array_dtype in [bodo.libs.bool_arr_ext.boolean_array, bodo
        .libs.str_arr_ext.string_array_type] or isinstance(col_array_dtype,
        types.Array) and col_array_dtype.dtype == bodo.datetime_date_type:

        @intrinsic
        def checkna_func(typingctx, table_t, ind_t):

            def codegen(context, builder, sig, args):
                itjec__ccdd, kqi__gynbr = args
                itjec__ccdd = builder.bitcast(itjec__ccdd, lir.IntType(8).
                    as_pointer().as_pointer())
                cpowj__lqbws = lir.Constant(lir.IntType(64), c_ind)
                qgook__fxctu = builder.load(builder.gep(itjec__ccdd, [
                    cpowj__lqbws]))
                jhvt__jdgss = builder.bitcast(qgook__fxctu, context.
                    get_data_type(types.bool_).as_pointer())
                vshnq__dlno = bodo.utils.cg_helpers.get_bitmap_bit(builder,
                    jhvt__jdgss, kqi__gynbr)
                kel__qyore = builder.icmp_unsigned('!=', vshnq__dlno, lir.
                    Constant(lir.IntType(8), 0))
                return builder.sext(kel__qyore, lir.IntType(8))
            return types.int8(types.voidptr, types.int64), codegen
        return checkna_func
    elif isinstance(col_array_dtype, types.Array):
        col_dtype = col_array_dtype.dtype
        if col_dtype in [bodo.datetime64ns, bodo.timedelta64ns]:

            @intrinsic
            def checkna_func(typingctx, table_t, ind_t):

                def codegen(context, builder, sig, args):
                    oly__aaz, kqi__gynbr = args
                    oly__aaz = builder.bitcast(oly__aaz, lir.IntType(8).
                        as_pointer().as_pointer())
                    cpowj__lqbws = lir.Constant(lir.IntType(64), c_ind)
                    qgook__fxctu = builder.load(builder.gep(oly__aaz, [
                        cpowj__lqbws]))
                    qgook__fxctu = builder.bitcast(qgook__fxctu, context.
                        get_data_type(col_dtype).as_pointer())
                    vbngk__dzv = builder.load(builder.gep(qgook__fxctu, [
                        kqi__gynbr]))
                    kel__qyore = builder.icmp_unsigned('!=', vbngk__dzv,
                        lir.Constant(lir.IntType(64), pd._libs.iNaT))
                    return builder.sext(kel__qyore, lir.IntType(8))
                return types.int8(types.voidptr, types.int64), codegen
            return checkna_func
        elif isinstance(col_dtype, types.Float):

            @intrinsic
            def checkna_func(typingctx, table_t, ind_t):

                def codegen(context, builder, sig, args):
                    oly__aaz, kqi__gynbr = args
                    oly__aaz = builder.bitcast(oly__aaz, lir.IntType(8).
                        as_pointer().as_pointer())
                    cpowj__lqbws = lir.Constant(lir.IntType(64), c_ind)
                    qgook__fxctu = builder.load(builder.gep(oly__aaz, [
                        cpowj__lqbws]))
                    qgook__fxctu = builder.bitcast(qgook__fxctu, context.
                        get_data_type(col_dtype).as_pointer())
                    vbngk__dzv = builder.load(builder.gep(qgook__fxctu, [
                        kqi__gynbr]))
                    jcx__kyr = signature(types.bool_, col_dtype)
                    vshnq__dlno = numba.np.npyfuncs.np_real_isnan_impl(context,
                        builder, jcx__kyr, (vbngk__dzv,))
                    return builder.not_(builder.sext(vshnq__dlno, lir.
                        IntType(8)))
                return types.int8(types.voidptr, types.int64), codegen
            return checkna_func
    raise BodoError(
        f"General Join Conditions with '{col_array_dtype}' column type not supported"
        )
