"""Array implementation for variable-size array items.
Corresponds to Spark's ArrayType: https://spark.apache.org/docs/latest/sql-reference.html
Corresponds to Arrow's Variable-size List: https://arrow.apache.org/docs/format/Columnar.html

The values are stored in a contingous data array, while an offsets array marks the
individual arrays. For example:
value:             [[1, 2], [3], None, [5, 4, 6], []]
data:              [1, 2, 3, 5, 4, 6]
offsets:           [0, 2, 3, 3, 6, 6]
"""
import operator
import llvmlite.binding as ll
import numba
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import NativeValue, box, intrinsic, models, overload, overload_attribute, overload_method, register_model, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_type
from bodo.libs import array_ext
from bodo.utils.cg_helpers import gen_allocate_array, get_array_elem_counts, get_bitmap_bit, is_na_value, pyarray_setitem, seq_getitem, set_bitmap_bit, to_arr_obj_if_list_obj
from bodo.utils.indexing import add_nested_counts, init_nested_counts
from bodo.utils.typing import BodoError, is_iterable_type, is_list_like_index_type
ll.add_symbol('count_total_elems_list_array', array_ext.
    count_total_elems_list_array)
ll.add_symbol('array_item_array_from_sequence', array_ext.
    array_item_array_from_sequence)
ll.add_symbol('np_array_from_array_item_array', array_ext.
    np_array_from_array_item_array)
offset_type = types.uint64
np_offset_type = numba.np.numpy_support.as_dtype(offset_type)


class ArrayItemArrayType(types.ArrayCompatible):

    def __init__(self, dtype):
        assert bodo.utils.utils.is_array_typ(dtype, False)
        self.dtype = dtype
        super(ArrayItemArrayType, self).__init__(name=
            'ArrayItemArrayType({})'.format(dtype))

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    def copy(self):
        return ArrayItemArrayType(self.dtype)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


class ArrayItemArrayPayloadType(types.Type):

    def __init__(self, array_type):
        self.array_type = array_type
        super(ArrayItemArrayPayloadType, self).__init__(name=
            'ArrayItemArrayPayloadType({})'.format(array_type))

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(ArrayItemArrayPayloadType)
class ArrayItemArrayPayloadModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        tzso__jyml = [('n_arrays', types.int64), ('data', fe_type.
            array_type.dtype), ('offsets', types.Array(offset_type, 1, 'C')
            ), ('null_bitmap', types.Array(types.uint8, 1, 'C'))]
        models.StructModel.__init__(self, dmm, fe_type, tzso__jyml)


@register_model(ArrayItemArrayType)
class ArrayItemArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        payload_type = ArrayItemArrayPayloadType(fe_type)
        tzso__jyml = [('meminfo', types.MemInfoPointer(payload_type))]
        models.StructModel.__init__(self, dmm, fe_type, tzso__jyml)


def define_array_item_dtor(context, builder, array_item_type, payload_type):
    yna__ris = builder.module
    ikoie__fevo = lir.FunctionType(lir.VoidType(), [cgutils.voidptr_t])
    jvoq__gdvx = cgutils.get_or_insert_function(yna__ris, ikoie__fevo, name
        ='.dtor.array_item.{}'.format(array_item_type.dtype))
    if not jvoq__gdvx.is_declaration:
        return jvoq__gdvx
    jvoq__gdvx.linkage = 'linkonce_odr'
    builder = lir.IRBuilder(jvoq__gdvx.append_basic_block())
    ndi__ilju = jvoq__gdvx.args[0]
    falh__qtcmv = context.get_value_type(payload_type).as_pointer()
    uuoai__ifyd = builder.bitcast(ndi__ilju, falh__qtcmv)
    zzbp__mqot = context.make_helper(builder, payload_type, ref=uuoai__ifyd)
    context.nrt.decref(builder, array_item_type.dtype, zzbp__mqot.data)
    context.nrt.decref(builder, types.Array(offset_type, 1, 'C'),
        zzbp__mqot.offsets)
    context.nrt.decref(builder, types.Array(types.uint8, 1, 'C'),
        zzbp__mqot.null_bitmap)
    builder.ret_void()
    return jvoq__gdvx


def construct_array_item_array(context, builder, array_item_type, n_arrays,
    n_elems, c=None):
    payload_type = ArrayItemArrayPayloadType(array_item_type)
    muojg__efy = context.get_value_type(payload_type)
    kjzqp__fuffz = context.get_abi_sizeof(muojg__efy)
    xxuh__drou = define_array_item_dtor(context, builder, array_item_type,
        payload_type)
    rdr__srsow = context.nrt.meminfo_alloc_dtor(builder, context.
        get_constant(types.uintp, kjzqp__fuffz), xxuh__drou)
    zcnvj__skiuo = context.nrt.meminfo_data(builder, rdr__srsow)
    bshsi__wpz = builder.bitcast(zcnvj__skiuo, muojg__efy.as_pointer())
    zzbp__mqot = cgutils.create_struct_proxy(payload_type)(context, builder)
    zzbp__mqot.n_arrays = n_arrays
    npgww__tly = n_elems.type.count
    pqtx__speum = builder.extract_value(n_elems, 0)
    jdacl__mclv = cgutils.alloca_once_value(builder, pqtx__speum)
    zuao__odzao = builder.icmp_signed('==', pqtx__speum, lir.Constant(
        pqtx__speum.type, -1))
    with builder.if_then(zuao__odzao):
        builder.store(n_arrays, jdacl__mclv)
    n_elems = cgutils.pack_array(builder, [builder.load(jdacl__mclv)] + [
        builder.extract_value(n_elems, jcibw__heyee) for jcibw__heyee in
        range(1, npgww__tly)])
    zzbp__mqot.data = gen_allocate_array(context, builder, array_item_type.
        dtype, n_elems, c)
    frgvk__vkveq = builder.add(n_arrays, lir.Constant(lir.IntType(64), 1))
    ryt__dsy = bodo.utils.utils._empty_nd_impl(context, builder, types.
        Array(offset_type, 1, 'C'), [frgvk__vkveq])
    offsets_ptr = ryt__dsy.data
    builder.store(context.get_constant(offset_type, 0), offsets_ptr)
    builder.store(builder.trunc(builder.extract_value(n_elems, 0), lir.
        IntType(offset_type.bitwidth)), builder.gep(offsets_ptr, [n_arrays]))
    zzbp__mqot.offsets = ryt__dsy._getvalue()
    ytew__dvirj = builder.udiv(builder.add(n_arrays, lir.Constant(lir.
        IntType(64), 7)), lir.Constant(lir.IntType(64), 8))
    ptx__ubix = bodo.utils.utils._empty_nd_impl(context, builder, types.
        Array(types.uint8, 1, 'C'), [ytew__dvirj])
    null_bitmap_ptr = ptx__ubix.data
    zzbp__mqot.null_bitmap = ptx__ubix._getvalue()
    builder.store(zzbp__mqot._getvalue(), bshsi__wpz)
    return rdr__srsow, zzbp__mqot.data, offsets_ptr, null_bitmap_ptr


def _unbox_array_item_array_copy_data(arr_typ, arr_obj, c, data_arr,
    item_ind, n_items):
    context = c.context
    builder = c.builder
    arr_obj = to_arr_obj_if_list_obj(c, context, builder, arr_obj, arr_typ)
    arr_val = c.pyapi.to_native_value(arr_typ, arr_obj).value
    sig = types.none(arr_typ, types.int64, types.int64, arr_typ)

    def copy_data(data_arr, item_ind, n_items, arr_val):
        data_arr[item_ind:item_ind + n_items] = arr_val
    gzbb__ewuee, ozur__nqgk = c.pyapi.call_jit_code(copy_data, sig, [
        data_arr, item_ind, n_items, arr_val])
    c.context.nrt.decref(builder, arr_typ, arr_val)


def _unbox_array_item_array_generic(typ, val, c, n_arrays, data_arr,
    offsets_ptr, null_bitmap_ptr):
    context = c.context
    builder = c.builder
    qghhy__tfxd = context.insert_const_string(builder.module, 'pandas')
    dlmeo__pyfql = c.pyapi.import_module_noblock(qghhy__tfxd)
    flbrd__iettg = c.pyapi.object_getattr_string(dlmeo__pyfql, 'NA')
    huoc__osec = c.context.get_constant(offset_type, 0)
    builder.store(huoc__osec, offsets_ptr)
    sja__skqu = cgutils.alloca_once_value(builder, context.get_constant(
        types.int64, 0))
    with cgutils.for_range(builder, n_arrays) as loop:
        tdac__tgj = loop.index
        item_ind = builder.load(sja__skqu)
        builder.store(builder.trunc(item_ind, lir.IntType(offset_type.
            bitwidth)), builder.gep(offsets_ptr, [tdac__tgj]))
        arr_obj = seq_getitem(builder, context, val, tdac__tgj)
        set_bitmap_bit(builder, null_bitmap_ptr, tdac__tgj, 0)
        nkpvo__neied = is_na_value(builder, context, arr_obj, flbrd__iettg)
        cwkw__bczt = builder.icmp_unsigned('!=', nkpvo__neied, lir.Constant
            (nkpvo__neied.type, 1))
        with builder.if_then(cwkw__bczt):
            set_bitmap_bit(builder, null_bitmap_ptr, tdac__tgj, 1)
            n_items = bodo.utils.utils.object_length(c, arr_obj)
            _unbox_array_item_array_copy_data(typ.dtype, arr_obj, c,
                data_arr, item_ind, n_items)
            builder.store(builder.add(item_ind, n_items), sja__skqu)
        c.pyapi.decref(arr_obj)
    builder.store(builder.trunc(builder.load(sja__skqu), lir.IntType(
        offset_type.bitwidth)), builder.gep(offsets_ptr, [n_arrays]))
    c.pyapi.decref(dlmeo__pyfql)
    c.pyapi.decref(flbrd__iettg)


@unbox(ArrayItemArrayType)
def unbox_array_item_array(typ, val, c):
    ddako__wdxim = isinstance(typ.dtype, types.Array) and typ.dtype.dtype in (
        types.int64, types.float64, types.bool_, datetime_date_type)
    n_arrays = bodo.utils.utils.object_length(c, val)
    if ddako__wdxim:
        ikoie__fevo = lir.FunctionType(lir.IntType(64), [lir.IntType(8).
            as_pointer()])
        avl__owo = cgutils.get_or_insert_function(c.builder.module,
            ikoie__fevo, name='count_total_elems_list_array')
        n_elems = cgutils.pack_array(c.builder, [c.builder.call(avl__owo, [
            val])])
    else:
        rcsvh__hvs = get_array_elem_counts(c, c.builder, c.context, val, typ)
        n_elems = cgutils.pack_array(c.builder, [c.builder.extract_value(
            rcsvh__hvs, jcibw__heyee) for jcibw__heyee in range(1,
            rcsvh__hvs.type.count)])
    rdr__srsow, data_arr, offsets_ptr, null_bitmap_ptr = (
        construct_array_item_array(c.context, c.builder, typ, n_arrays,
        n_elems, c))
    if ddako__wdxim:
        ddq__snjk = bodo.utils.utils.numba_to_c_type(typ.dtype.dtype)
        pery__ofsj = c.context.make_array(typ.dtype)(c.context, c.builder,
            data_arr).data
        ikoie__fevo = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(
            offset_type.bitwidth).as_pointer(), lir.IntType(8).as_pointer(),
            lir.IntType(32)])
        jvoq__gdvx = cgutils.get_or_insert_function(c.builder.module,
            ikoie__fevo, name='array_item_array_from_sequence')
        c.builder.call(jvoq__gdvx, [val, c.builder.bitcast(pery__ofsj, lir.
            IntType(8).as_pointer()), offsets_ptr, null_bitmap_ptr, lir.
            Constant(lir.IntType(32), ddq__snjk)])
    else:
        _unbox_array_item_array_generic(typ, val, c, n_arrays, data_arr,
            offsets_ptr, null_bitmap_ptr)
    egfd__zrde = c.context.make_helper(c.builder, typ)
    egfd__zrde.meminfo = rdr__srsow
    nef__skc = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(egfd__zrde._getvalue(), is_error=nef__skc)


def _get_array_item_arr_payload(context, builder, arr_typ, arr):
    egfd__zrde = context.make_helper(builder, arr_typ, arr)
    payload_type = ArrayItemArrayPayloadType(arr_typ)
    zcnvj__skiuo = context.nrt.meminfo_data(builder, egfd__zrde.meminfo)
    bshsi__wpz = builder.bitcast(zcnvj__skiuo, context.get_value_type(
        payload_type).as_pointer())
    zzbp__mqot = cgutils.create_struct_proxy(payload_type)(context, builder,
        builder.load(bshsi__wpz))
    return zzbp__mqot


def _box_array_item_array_generic(typ, c, n_arrays, data_arr, offsets_ptr,
    null_bitmap_ptr):
    context = c.context
    builder = c.builder
    qghhy__tfxd = context.insert_const_string(builder.module, 'numpy')
    kiow__srq = c.pyapi.import_module_noblock(qghhy__tfxd)
    xbgci__ewiba = c.pyapi.object_getattr_string(kiow__srq, 'object_')
    sdssv__jlbe = c.pyapi.long_from_longlong(n_arrays)
    sjvg__tdbjm = c.pyapi.call_method(kiow__srq, 'ndarray', (sdssv__jlbe,
        xbgci__ewiba))
    slic__nsz = c.pyapi.object_getattr_string(kiow__srq, 'nan')
    sja__skqu = cgutils.alloca_once_value(builder, lir.Constant(lir.IntType
        (64), 0))
    with cgutils.for_range(builder, n_arrays) as loop:
        tdac__tgj = loop.index
        pyarray_setitem(builder, context, sjvg__tdbjm, tdac__tgj, slic__nsz)
        avvb__rrk = get_bitmap_bit(builder, null_bitmap_ptr, tdac__tgj)
        prie__khypx = builder.icmp_unsigned('!=', avvb__rrk, lir.Constant(
            lir.IntType(8), 0))
        with builder.if_then(prie__khypx):
            n_items = builder.sext(builder.sub(builder.load(builder.gep(
                offsets_ptr, [builder.add(tdac__tgj, lir.Constant(tdac__tgj
                .type, 1))])), builder.load(builder.gep(offsets_ptr, [
                tdac__tgj]))), lir.IntType(64))
            item_ind = builder.load(sja__skqu)
            gzbb__ewuee, uucwi__gjrm = c.pyapi.call_jit_code(lambda
                data_arr, item_ind, n_items: data_arr[item_ind:item_ind +
                n_items], typ.dtype(typ.dtype, types.int64, types.int64), [
                data_arr, item_ind, n_items])
            builder.store(builder.add(item_ind, n_items), sja__skqu)
            arr_obj = c.pyapi.from_native_value(typ.dtype, uucwi__gjrm, c.
                env_manager)
            pyarray_setitem(builder, context, sjvg__tdbjm, tdac__tgj, arr_obj)
            c.pyapi.decref(arr_obj)
    c.pyapi.decref(kiow__srq)
    c.pyapi.decref(xbgci__ewiba)
    c.pyapi.decref(sdssv__jlbe)
    c.pyapi.decref(slic__nsz)
    return sjvg__tdbjm


@box(ArrayItemArrayType)
def box_array_item_arr(typ, val, c):
    zzbp__mqot = _get_array_item_arr_payload(c.context, c.builder, typ, val)
    data_arr = zzbp__mqot.data
    offsets_ptr = c.context.make_helper(c.builder, types.Array(offset_type,
        1, 'C'), zzbp__mqot.offsets).data
    null_bitmap_ptr = c.context.make_helper(c.builder, types.Array(types.
        uint8, 1, 'C'), zzbp__mqot.null_bitmap).data
    if isinstance(typ.dtype, types.Array) and typ.dtype.dtype in (types.
        int64, types.float64, types.bool_, datetime_date_type):
        ddq__snjk = bodo.utils.utils.numba_to_c_type(typ.dtype.dtype)
        pery__ofsj = c.context.make_helper(c.builder, typ.dtype, data_arr).data
        ikoie__fevo = lir.FunctionType(c.context.get_argument_type(types.
            pyobject), [lir.IntType(64), lir.IntType(8).as_pointer(), lir.
            IntType(offset_type.bitwidth).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(32)])
        rpmyo__gjb = cgutils.get_or_insert_function(c.builder.module,
            ikoie__fevo, name='np_array_from_array_item_array')
        arr = c.builder.call(rpmyo__gjb, [zzbp__mqot.n_arrays, c.builder.
            bitcast(pery__ofsj, lir.IntType(8).as_pointer()), offsets_ptr,
            null_bitmap_ptr, lir.Constant(lir.IntType(32), ddq__snjk)])
    else:
        arr = _box_array_item_array_generic(typ, c, zzbp__mqot.n_arrays,
            data_arr, offsets_ptr, null_bitmap_ptr)
    c.context.nrt.decref(c.builder, typ, val)
    return arr


def lower_pre_alloc_array_item_array(context, builder, sig, args):
    array_item_type = sig.return_type
    rnodi__fqmq, urlk__aja, bbc__xnxp = args
    zniwf__gny = bodo.utils.transform.get_type_alloc_counts(array_item_type
        .dtype)
    ctuel__jjquz = sig.args[1]
    if not isinstance(ctuel__jjquz, types.UniTuple):
        urlk__aja = cgutils.pack_array(builder, [lir.Constant(lir.IntType(
            64), -1) for bbc__xnxp in range(zniwf__gny)])
    elif ctuel__jjquz.count < zniwf__gny:
        urlk__aja = cgutils.pack_array(builder, [builder.extract_value(
            urlk__aja, jcibw__heyee) for jcibw__heyee in range(ctuel__jjquz
            .count)] + [lir.Constant(lir.IntType(64), -1) for bbc__xnxp in
            range(zniwf__gny - ctuel__jjquz.count)])
    rdr__srsow, bbc__xnxp, bbc__xnxp, bbc__xnxp = construct_array_item_array(
        context, builder, array_item_type, rnodi__fqmq, urlk__aja)
    egfd__zrde = context.make_helper(builder, array_item_type)
    egfd__zrde.meminfo = rdr__srsow
    return egfd__zrde._getvalue()


@intrinsic
def pre_alloc_array_item_array(typingctx, num_arrs_typ, num_values_typ,
    dtype_typ=None):
    assert isinstance(num_arrs_typ, types.Integer)
    array_item_type = ArrayItemArrayType(dtype_typ.instance_type)
    num_values_typ = types.unliteral(num_values_typ)
    return array_item_type(types.int64, num_values_typ, dtype_typ
        ), lower_pre_alloc_array_item_array


def pre_alloc_array_item_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 3 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_libs_array_item_arr_ext_pre_alloc_array_item_array
    ) = pre_alloc_array_item_array_equiv


def init_array_item_array_codegen(context, builder, signature, args):
    n_arrays, qqoa__zgvb, ryt__dsy, ptx__ubix = args
    array_item_type = signature.return_type
    payload_type = ArrayItemArrayPayloadType(array_item_type)
    muojg__efy = context.get_value_type(payload_type)
    kjzqp__fuffz = context.get_abi_sizeof(muojg__efy)
    xxuh__drou = define_array_item_dtor(context, builder, array_item_type,
        payload_type)
    rdr__srsow = context.nrt.meminfo_alloc_dtor(builder, context.
        get_constant(types.uintp, kjzqp__fuffz), xxuh__drou)
    zcnvj__skiuo = context.nrt.meminfo_data(builder, rdr__srsow)
    bshsi__wpz = builder.bitcast(zcnvj__skiuo, muojg__efy.as_pointer())
    zzbp__mqot = cgutils.create_struct_proxy(payload_type)(context, builder)
    zzbp__mqot.n_arrays = n_arrays
    zzbp__mqot.data = qqoa__zgvb
    zzbp__mqot.offsets = ryt__dsy
    zzbp__mqot.null_bitmap = ptx__ubix
    builder.store(zzbp__mqot._getvalue(), bshsi__wpz)
    context.nrt.incref(builder, signature.args[1], qqoa__zgvb)
    context.nrt.incref(builder, signature.args[2], ryt__dsy)
    context.nrt.incref(builder, signature.args[3], ptx__ubix)
    egfd__zrde = context.make_helper(builder, array_item_type)
    egfd__zrde.meminfo = rdr__srsow
    return egfd__zrde._getvalue()


@intrinsic
def init_array_item_array(typingctx, n_arrays_typ, data_type, offsets_typ,
    null_bitmap_typ=None):
    assert null_bitmap_typ == types.Array(types.uint8, 1, 'C')
    fvpha__yef = ArrayItemArrayType(data_type)
    sig = fvpha__yef(types.int64, data_type, offsets_typ, null_bitmap_typ)
    return sig, init_array_item_array_codegen


@intrinsic
def get_offsets(typingctx, arr_typ=None):
    assert isinstance(arr_typ, ArrayItemArrayType)

    def codegen(context, builder, sig, args):
        arr, = args
        zzbp__mqot = _get_array_item_arr_payload(context, builder, arr_typ, arr
            )
        return impl_ret_borrowed(context, builder, sig.return_type,
            zzbp__mqot.offsets)
    return types.Array(offset_type, 1, 'C')(arr_typ), codegen


@intrinsic
def get_offsets_ind(typingctx, arr_typ, ind_t=None):
    assert isinstance(arr_typ, ArrayItemArrayType)

    def codegen(context, builder, sig, args):
        arr, ind = args
        zzbp__mqot = _get_array_item_arr_payload(context, builder, arr_typ, arr
            )
        pery__ofsj = context.make_array(types.Array(offset_type, 1, 'C'))(
            context, builder, zzbp__mqot.offsets).data
        ryt__dsy = builder.bitcast(pery__ofsj, lir.IntType(offset_type.
            bitwidth).as_pointer())
        return builder.load(builder.gep(ryt__dsy, [ind]))
    return offset_type(arr_typ, types.int64), codegen


@intrinsic
def get_data(typingctx, arr_typ=None):
    assert isinstance(arr_typ, ArrayItemArrayType)

    def codegen(context, builder, sig, args):
        arr, = args
        zzbp__mqot = _get_array_item_arr_payload(context, builder, arr_typ, arr
            )
        return impl_ret_borrowed(context, builder, sig.return_type,
            zzbp__mqot.data)
    return arr_typ.dtype(arr_typ), codegen


@intrinsic
def get_null_bitmap(typingctx, arr_typ=None):
    assert isinstance(arr_typ, ArrayItemArrayType)

    def codegen(context, builder, sig, args):
        arr, = args
        zzbp__mqot = _get_array_item_arr_payload(context, builder, arr_typ, arr
            )
        return impl_ret_borrowed(context, builder, sig.return_type,
            zzbp__mqot.null_bitmap)
    return types.Array(types.uint8, 1, 'C')(arr_typ), codegen


def alias_ext_single_array(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 1
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['get_offsets',
    'bodo.libs.array_item_arr_ext'] = alias_ext_single_array
numba.core.ir_utils.alias_func_extensions['get_data',
    'bodo.libs.array_item_arr_ext'] = alias_ext_single_array
numba.core.ir_utils.alias_func_extensions['get_null_bitmap',
    'bodo.libs.array_item_arr_ext'] = alias_ext_single_array


@intrinsic
def get_n_arrays(typingctx, arr_typ=None):
    assert isinstance(arr_typ, ArrayItemArrayType)

    def codegen(context, builder, sig, args):
        arr, = args
        zzbp__mqot = _get_array_item_arr_payload(context, builder, arr_typ, arr
            )
        return zzbp__mqot.n_arrays
    return types.int64(arr_typ), codegen


@intrinsic
def replace_data_arr(typingctx, arr_typ, data_typ=None):
    assert isinstance(arr_typ, ArrayItemArrayType
        ) and data_typ == arr_typ.dtype

    def codegen(context, builder, sig, args):
        arr, wwhpe__bkbzn = args
        egfd__zrde = context.make_helper(builder, arr_typ, arr)
        payload_type = ArrayItemArrayPayloadType(arr_typ)
        zcnvj__skiuo = context.nrt.meminfo_data(builder, egfd__zrde.meminfo)
        bshsi__wpz = builder.bitcast(zcnvj__skiuo, context.get_value_type(
            payload_type).as_pointer())
        zzbp__mqot = cgutils.create_struct_proxy(payload_type)(context,
            builder, builder.load(bshsi__wpz))
        context.nrt.decref(builder, data_typ, zzbp__mqot.data)
        zzbp__mqot.data = wwhpe__bkbzn
        context.nrt.incref(builder, data_typ, wwhpe__bkbzn)
        builder.store(zzbp__mqot._getvalue(), bshsi__wpz)
    return types.none(arr_typ, data_typ), codegen


@numba.njit(no_cpython_wrapper=True)
def ensure_data_capacity(arr, old_size, new_size):
    qqoa__zgvb = get_data(arr)
    ufakv__tseq = len(qqoa__zgvb)
    if ufakv__tseq < new_size:
        mhrut__pqkvt = max(2 * ufakv__tseq, new_size)
        wwhpe__bkbzn = bodo.libs.array_kernels.resize_and_copy(qqoa__zgvb,
            old_size, mhrut__pqkvt)
        replace_data_arr(arr, wwhpe__bkbzn)


@numba.njit(no_cpython_wrapper=True)
def trim_excess_data(arr):
    qqoa__zgvb = get_data(arr)
    ryt__dsy = get_offsets(arr)
    ccgj__atxdq = len(qqoa__zgvb)
    mqiv__eiiwd = ryt__dsy[-1]
    if ccgj__atxdq != mqiv__eiiwd:
        wwhpe__bkbzn = bodo.libs.array_kernels.resize_and_copy(qqoa__zgvb,
            mqiv__eiiwd, mqiv__eiiwd)
        replace_data_arr(arr, wwhpe__bkbzn)


@overload(len, no_unliteral=True)
def overload_array_item_arr_len(A):
    if isinstance(A, ArrayItemArrayType):
        return lambda A: get_n_arrays(A)


@overload_attribute(ArrayItemArrayType, 'shape')
def overload_array_item_arr_shape(A):
    return lambda A: (get_n_arrays(A),)


@overload_attribute(ArrayItemArrayType, 'dtype')
def overload_array_item_arr_dtype(A):
    return lambda A: np.object_


@overload_attribute(ArrayItemArrayType, 'ndim')
def overload_array_item_arr_ndim(A):
    return lambda A: 1


@overload_attribute(ArrayItemArrayType, 'nbytes')
def overload_array_item_arr_nbytes(A):
    return lambda A: get_data(A).nbytes + get_offsets(A
        ).nbytes + get_null_bitmap(A).nbytes


@overload(operator.getitem, no_unliteral=True)
def array_item_arr_getitem_array(arr, ind):
    if not isinstance(arr, ArrayItemArrayType):
        return
    if isinstance(ind, types.Integer):

        def array_item_arr_getitem_impl(arr, ind):
            if ind < 0:
                ind += len(arr)
            ryt__dsy = get_offsets(arr)
            qqoa__zgvb = get_data(arr)
            wmq__xtd = ryt__dsy[ind]
            gqwx__znm = ryt__dsy[ind + 1]
            return qqoa__zgvb[wmq__xtd:gqwx__znm]
        return array_item_arr_getitem_impl
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:
        ybtrm__zdx = arr.dtype

        def impl_bool(arr, ind):
            xqrmh__hhnmr = len(arr)
            if xqrmh__hhnmr != len(ind):
                raise IndexError(
                    'boolean index did not match indexed array along dimension 0'
                    )
            ptx__ubix = get_null_bitmap(arr)
            n_arrays = 0
            muuap__ovo = init_nested_counts(ybtrm__zdx)
            for jcibw__heyee in range(xqrmh__hhnmr):
                if ind[jcibw__heyee]:
                    n_arrays += 1
                    awjew__uicso = arr[jcibw__heyee]
                    muuap__ovo = add_nested_counts(muuap__ovo, awjew__uicso)
            sjvg__tdbjm = pre_alloc_array_item_array(n_arrays, muuap__ovo,
                ybtrm__zdx)
            kzgnm__orl = get_null_bitmap(sjvg__tdbjm)
            glac__jcyr = 0
            for rcgod__uumw in range(xqrmh__hhnmr):
                if ind[rcgod__uumw]:
                    sjvg__tdbjm[glac__jcyr] = arr[rcgod__uumw]
                    wsmp__jbwdk = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                        ptx__ubix, rcgod__uumw)
                    bodo.libs.int_arr_ext.set_bit_to_arr(kzgnm__orl,
                        glac__jcyr, wsmp__jbwdk)
                    glac__jcyr += 1
            return sjvg__tdbjm
        return impl_bool
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):
        ybtrm__zdx = arr.dtype

        def impl_int(arr, ind):
            ptx__ubix = get_null_bitmap(arr)
            xqrmh__hhnmr = len(ind)
            n_arrays = xqrmh__hhnmr
            muuap__ovo = init_nested_counts(ybtrm__zdx)
            for cmf__ocl in range(xqrmh__hhnmr):
                jcibw__heyee = ind[cmf__ocl]
                awjew__uicso = arr[jcibw__heyee]
                muuap__ovo = add_nested_counts(muuap__ovo, awjew__uicso)
            sjvg__tdbjm = pre_alloc_array_item_array(n_arrays, muuap__ovo,
                ybtrm__zdx)
            kzgnm__orl = get_null_bitmap(sjvg__tdbjm)
            for via__jqx in range(xqrmh__hhnmr):
                rcgod__uumw = ind[via__jqx]
                sjvg__tdbjm[via__jqx] = arr[rcgod__uumw]
                wsmp__jbwdk = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                    ptx__ubix, rcgod__uumw)
                bodo.libs.int_arr_ext.set_bit_to_arr(kzgnm__orl, via__jqx,
                    wsmp__jbwdk)
            return sjvg__tdbjm
        return impl_int
    if isinstance(ind, types.SliceType):

        def impl_slice(arr, ind):
            xqrmh__hhnmr = len(arr)
            vuwi__zie = numba.cpython.unicode._normalize_slice(ind,
                xqrmh__hhnmr)
            ixgp__xeht = np.arange(vuwi__zie.start, vuwi__zie.stop,
                vuwi__zie.step)
            return arr[ixgp__xeht]
        return impl_slice


@overload(operator.setitem)
def array_item_arr_setitem(A, idx, val):
    if not isinstance(A, ArrayItemArrayType):
        return
    if isinstance(idx, types.Integer):

        def impl_scalar(A, idx, val):
            ryt__dsy = get_offsets(A)
            ptx__ubix = get_null_bitmap(A)
            if idx == 0:
                ryt__dsy[0] = 0
            n_items = len(val)
            rsedg__kyp = ryt__dsy[idx] + n_items
            ensure_data_capacity(A, ryt__dsy[idx], rsedg__kyp)
            qqoa__zgvb = get_data(A)
            ryt__dsy[idx + 1] = ryt__dsy[idx] + n_items
            qqoa__zgvb[ryt__dsy[idx]:ryt__dsy[idx + 1]] = val
            bodo.libs.int_arr_ext.set_bit_to_arr(ptx__ubix, idx, 1)
        return impl_scalar
    if isinstance(idx, types.SliceType) and A.dtype == val:

        def impl_slice_elem(A, idx, val):
            vuwi__zie = numba.cpython.unicode._normalize_slice(idx, len(A))
            for jcibw__heyee in range(vuwi__zie.start, vuwi__zie.stop,
                vuwi__zie.step):
                A[jcibw__heyee] = val
        return impl_slice_elem
    if isinstance(idx, types.SliceType) and is_iterable_type(val):

        def impl_slice(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            ryt__dsy = get_offsets(A)
            ptx__ubix = get_null_bitmap(A)
            hhc__wzh = get_offsets(val)
            nhs__ylynp = get_data(val)
            hjhc__ltyj = get_null_bitmap(val)
            xqrmh__hhnmr = len(A)
            vuwi__zie = numba.cpython.unicode._normalize_slice(idx,
                xqrmh__hhnmr)
            bwpck__gswf, mwrfa__bdcjw = vuwi__zie.start, vuwi__zie.stop
            assert vuwi__zie.step == 1
            if bwpck__gswf == 0:
                ryt__dsy[bwpck__gswf] = 0
            wqs__gei = ryt__dsy[bwpck__gswf]
            rsedg__kyp = wqs__gei + len(nhs__ylynp)
            ensure_data_capacity(A, wqs__gei, rsedg__kyp)
            qqoa__zgvb = get_data(A)
            qqoa__zgvb[wqs__gei:wqs__gei + len(nhs__ylynp)] = nhs__ylynp
            ryt__dsy[bwpck__gswf:mwrfa__bdcjw + 1] = hhc__wzh + wqs__gei
            vsn__oqy = 0
            for jcibw__heyee in range(bwpck__gswf, mwrfa__bdcjw):
                wsmp__jbwdk = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                    hjhc__ltyj, vsn__oqy)
                bodo.libs.int_arr_ext.set_bit_to_arr(ptx__ubix,
                    jcibw__heyee, wsmp__jbwdk)
                vsn__oqy += 1
        return impl_slice
    raise BodoError(
        'only setitem with scalar index is currently supported for list arrays'
        )


@overload_method(ArrayItemArrayType, 'copy', no_unliteral=True)
def overload_array_item_arr_copy(A):

    def copy_impl(A):
        return init_array_item_array(len(A), get_data(A).copy(),
            get_offsets(A).copy(), get_null_bitmap(A).copy())
    return copy_impl
