"""
Implements array kernels such as median and quantile.
"""
import hashlib
import inspect
import math
import operator
import re
import warnings
from math import sqrt
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types, typing
from numba.core.imputils import lower_builtin
from numba.core.ir_utils import find_const, guard
from numba.core.typing import signature
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import overload, overload_attribute, register_jitable
from numba.np.arrayobj import make_array
from numba.np.numpy_support import as_dtype
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.datetime_timedelta_ext import datetime_timedelta_array_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, init_categorical_array
from bodo.hiframes.split_impl import string_array_split_view_type
from bodo.libs import quantile_alg
from bodo.libs.array import arr_info_list_to_table, array_to_info, delete_info_decref_array, delete_table, delete_table_decref_arrays, drop_duplicates_table, info_from_table, info_to_array, sample_table
from bodo.libs.array_item_arr_ext import ArrayItemArrayType, offset_type
from bodo.libs.bool_arr_ext import BooleanArrayType, boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.distributed_api import Reduce_Type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import str_arr_set_na, string_array_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.libs.tuple_arr_ext import TupleArrayType
from bodo.utils.indexing import add_nested_counts, init_nested_counts
from bodo.utils.shuffle import getitem_arr_tup_single
from bodo.utils.typing import BodoError, check_unsupported_args, element_type, find_common_np_dtype, get_overload_const_bool, get_overload_const_list, get_overload_const_str, is_overload_none, raise_bodo_error
from bodo.utils.utils import build_set_seen_na, check_and_propagate_cpp_exception, numba_to_c_type, unliteral_all
ll.add_symbol('quantile_sequential', quantile_alg.quantile_sequential)
ll.add_symbol('quantile_parallel', quantile_alg.quantile_parallel)
MPI_ROOT = 0
sum_op = np.int32(bodo.libs.distributed_api.Reduce_Type.Sum.value)


def isna(arr, i):
    return False


@overload(isna)
def overload_isna(arr, i):
    i = types.unliteral(i)
    if arr == string_array_type:
        return lambda arr, i: bodo.libs.str_arr_ext.str_arr_is_na(arr, i)
    if isinstance(arr, (IntegerArrayType, DecimalArrayType)) or arr in (
        boolean_array, datetime_date_array_type,
        datetime_timedelta_array_type, string_array_split_view_type):
        return lambda arr, i: not bodo.libs.int_arr_ext.get_bit_bitmap_arr(arr
            ._null_bitmap, i)
    if isinstance(arr, ArrayItemArrayType):
        return lambda arr, i: not bodo.libs.int_arr_ext.get_bit_bitmap_arr(bodo
            .libs.array_item_arr_ext.get_null_bitmap(arr), i)
    if isinstance(arr, StructArrayType):
        return lambda arr, i: not bodo.libs.int_arr_ext.get_bit_bitmap_arr(bodo
            .libs.struct_arr_ext.get_null_bitmap(arr), i)
    if isinstance(arr, TupleArrayType):
        return lambda arr, i: bodo.libs.array_kernels.isna(arr._data, i)
    if isinstance(arr, bodo.hiframes.pd_categorical_ext.CategoricalArrayType):
        return lambda arr, i: arr.codes[i] == -1
    if arr == bodo.binary_array_type:
        return lambda arr, i: not bodo.libs.int_arr_ext.get_bit_bitmap_arr(bodo
            .libs.array_item_arr_ext.get_null_bitmap(arr._data), i)
    if isinstance(arr, types.List):
        if arr.dtype == types.none:
            return lambda arr, i: True
        elif isinstance(arr.dtype, types.optional):
            return lambda arr, i: arr[i] is None
        else:
            return lambda arr, i: False
    if isinstance(arr, bodo.NullableTupleType):
        return lambda arr, i: arr._null_values[i]
    assert isinstance(arr, types.Array)
    dtype = arr.dtype
    if isinstance(dtype, types.Float):
        return lambda arr, i: np.isnan(arr[i])
    if isinstance(dtype, (types.NPDatetime, types.NPTimedelta)):
        return lambda arr, i: np.isnat(arr[i])
    return lambda arr, i: False


def setna(arr, ind, int_nan_const=0):
    arr[ind] = np.nan


@overload(setna, no_unliteral=True)
def setna_overload(arr, ind, int_nan_const=0):
    if isinstance(arr.dtype, types.Float):
        return setna
    if isinstance(arr.dtype, (types.NPDatetime, types.NPTimedelta)):
        ojhh__pvej = arr.dtype('NaT')

        def _setnan_impl(arr, ind, int_nan_const=0):
            arr[ind] = ojhh__pvej
        return _setnan_impl
    if arr == string_array_type:

        def impl(arr, ind, int_nan_const=0):
            arr[ind] = ''
            str_arr_set_na(arr, ind)
        return impl
    if arr == boolean_array:

        def impl(arr, ind, int_nan_const=0):
            arr[ind] = False
            bodo.libs.int_arr_ext.set_bit_to_arr(arr._null_bitmap, ind, 0)
        return impl
    if isinstance(arr, (IntegerArrayType, DecimalArrayType)):
        return (lambda arr, ind, int_nan_const=0: bodo.libs.int_arr_ext.
            set_bit_to_arr(arr._null_bitmap, ind, 0))
    if arr == bodo.binary_array_type:

        def impl_binary_arr(arr, ind, int_nan_const=0):
            axtj__akkj = bodo.libs.array_item_arr_ext.get_offsets(arr._data)
            axtj__akkj[ind + 1] = axtj__akkj[ind]
            bodo.libs.int_arr_ext.set_bit_to_arr(bodo.libs.
                array_item_arr_ext.get_null_bitmap(arr._data), ind, 0)
        return impl_binary_arr
    if isinstance(arr, bodo.libs.array_item_arr_ext.ArrayItemArrayType):

        def impl_arr_item(arr, ind, int_nan_const=0):
            axtj__akkj = bodo.libs.array_item_arr_ext.get_offsets(arr)
            axtj__akkj[ind + 1] = axtj__akkj[ind]
            bodo.libs.int_arr_ext.set_bit_to_arr(bodo.libs.
                array_item_arr_ext.get_null_bitmap(arr), ind, 0)
        return impl_arr_item
    if isinstance(arr, bodo.libs.struct_arr_ext.StructArrayType):

        def impl(arr, ind, int_nan_const=0):
            bodo.libs.int_arr_ext.set_bit_to_arr(bodo.libs.struct_arr_ext.
                get_null_bitmap(arr), ind, 0)
            data = bodo.libs.struct_arr_ext.get_data(arr)
            setna_tup(data, ind)
        return impl
    if isinstance(arr, TupleArrayType):

        def impl(arr, ind, int_nan_const=0):
            bodo.libs.array_kernels.setna(arr._data, ind)
        return impl
    if arr.dtype == types.bool_:

        def b_set(arr, ind, int_nan_const=0):
            arr[ind] = False
        return b_set
    if isinstance(arr, bodo.hiframes.pd_categorical_ext.CategoricalArrayType):

        def setna_cat(arr, ind, int_nan_const=0):
            arr.codes[ind] = -1
        return setna_cat
    if isinstance(arr.dtype, types.Integer):

        def setna_int(arr, ind, int_nan_const=0):
            arr[ind] = int_nan_const
        return setna_int
    if arr == datetime_date_array_type:

        def setna_datetime_date(arr, ind, int_nan_const=0):
            arr._data[ind] = (1970 << 32) + (1 << 16) + 1
            bodo.libs.int_arr_ext.set_bit_to_arr(arr._null_bitmap, ind, 0)
        return setna_datetime_date
    if arr == datetime_timedelta_array_type:

        def setna_datetime_timedelta(arr, ind, int_nan_const=0):
            bodo.libs.array_kernels.setna(arr._days_data, ind)
            bodo.libs.array_kernels.setna(arr._seconds_data, ind)
            bodo.libs.array_kernels.setna(arr._microseconds_data, ind)
            bodo.libs.int_arr_ext.set_bit_to_arr(arr._null_bitmap, ind, 0)
        return setna_datetime_timedelta
    return lambda arr, ind, int_nan_const=0: None


def setna_tup(arr_tup, ind, int_nan_const=0):
    for arr in arr_tup:
        arr[ind] = np.nan


@overload(setna_tup, no_unliteral=True)
def overload_setna_tup(arr_tup, ind, int_nan_const=0):
    temb__lhu = arr_tup.count
    toliv__fkvdm = 'def f(arr_tup, ind, int_nan_const=0):\n'
    for i in range(temb__lhu):
        toliv__fkvdm += '  setna(arr_tup[{}], ind, int_nan_const)\n'.format(i)
    toliv__fkvdm += '  return\n'
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'setna': setna}, qhcdq__abnc)
    impl = qhcdq__abnc['f']
    return impl


def setna_slice(arr, s):
    arr[s] = np.nan


@overload(setna_slice, no_unliteral=True)
def overload_setna_slice(arr, s):

    def impl(arr, s):
        vraw__xsfkg = numba.cpython.unicode._normalize_slice(s, len(arr))
        for i in range(vraw__xsfkg.start, vraw__xsfkg.stop, vraw__xsfkg.step):
            setna(arr, i)
    return impl


ll.add_symbol('median_series_computation', quantile_alg.
    median_series_computation)
_median_series_computation = types.ExternalFunction('median_series_computation'
    , types.void(types.voidptr, bodo.libs.array.array_info_type, types.
    bool_, types.bool_))


@numba.njit
def median_series_computation(res, arr, is_parallel, skipna):
    dii__prgo = array_to_info(arr)
    _median_series_computation(res, dii__prgo, is_parallel, skipna)
    check_and_propagate_cpp_exception()
    delete_info_decref_array(dii__prgo)


ll.add_symbol('autocorr_series_computation', quantile_alg.
    autocorr_series_computation)
_autocorr_series_computation = types.ExternalFunction(
    'autocorr_series_computation', types.void(types.voidptr, bodo.libs.
    array.array_info_type, types.int64, types.bool_))


@numba.njit
def autocorr_series_computation(res, arr, lag, is_parallel):
    dii__prgo = array_to_info(arr)
    _autocorr_series_computation(res, dii__prgo, lag, is_parallel)
    check_and_propagate_cpp_exception()
    delete_info_decref_array(dii__prgo)


@numba.njit
def autocorr(arr, lag=1, parallel=False):
    res = np.empty(1, types.float64)
    autocorr_series_computation(res.ctypes, arr, lag, parallel)
    return res[0]


ll.add_symbol('compute_series_monotonicity', quantile_alg.
    compute_series_monotonicity)
_compute_series_monotonicity = types.ExternalFunction(
    'compute_series_monotonicity', types.void(types.voidptr, bodo.libs.
    array.array_info_type, types.int64, types.bool_))


@numba.njit
def series_monotonicity_call(res, arr, inc_dec, is_parallel):
    dii__prgo = array_to_info(arr)
    _compute_series_monotonicity(res, dii__prgo, inc_dec, is_parallel)
    check_and_propagate_cpp_exception()
    delete_info_decref_array(dii__prgo)


@numba.njit
def series_monotonicity(arr, inc_dec, parallel=False):
    res = np.empty(1, types.float64)
    series_monotonicity_call(res.ctypes, arr, inc_dec, parallel)
    rmpl__dbgyx = res[0] > 0.5
    return rmpl__dbgyx


@numba.generated_jit(nopython=True)
def get_valid_entries_from_date_offset(index_arr, offset, initial_date,
    is_last, is_parallel=False):
    if get_overload_const_bool(is_last):
        vqk__dpfg = '-'
        uolie__shxi = 'index_arr[0] > threshhold_date'
        xxc__pifgy = '1, n+1'
        vno__pjp = 'index_arr[-i] <= threshhold_date'
        mhn__nkwlf = 'i - 1'
    else:
        vqk__dpfg = '+'
        uolie__shxi = 'index_arr[-1] < threshhold_date'
        xxc__pifgy = 'n'
        vno__pjp = 'index_arr[i] >= threshhold_date'
        mhn__nkwlf = 'i'
    toliv__fkvdm = (
        'def impl(index_arr, offset, initial_date, is_last, is_parallel=False):\n'
        )
    if types.unliteral(offset) == types.unicode_type:
        toliv__fkvdm += (
            '  with numba.objmode(threshhold_date=bodo.pd_timestamp_type):\n')
        toliv__fkvdm += (
            '    date_offset = pd.tseries.frequencies.to_offset(offset)\n')
        if not get_overload_const_bool(is_last):
            toliv__fkvdm += """    if not isinstance(date_offset, pd._libs.tslibs.Tick) and date_offset.is_on_offset(index_arr[0]):
"""
            toliv__fkvdm += """      threshhold_date = initial_date - date_offset.base + date_offset
"""
            toliv__fkvdm += '    else:\n'
            toliv__fkvdm += (
                '      threshhold_date = initial_date + date_offset\n')
        else:
            toliv__fkvdm += (
                f'    threshhold_date = initial_date {vqk__dpfg} date_offset\n'
                )
    else:
        toliv__fkvdm += (
            f'  threshhold_date = initial_date {vqk__dpfg} offset\n')
    toliv__fkvdm += '  local_valid = 0\n'
    toliv__fkvdm += f'  n = len(index_arr)\n'
    toliv__fkvdm += f'  if n:\n'
    toliv__fkvdm += f'    if {uolie__shxi}:\n'
    toliv__fkvdm += '      loc_valid = n\n'
    toliv__fkvdm += '    else:\n'
    toliv__fkvdm += f'      for i in range({xxc__pifgy}):\n'
    toliv__fkvdm += f'        if {vno__pjp}:\n'
    toliv__fkvdm += f'          loc_valid = {mhn__nkwlf}\n'
    toliv__fkvdm += '          break\n'
    toliv__fkvdm += '  if is_parallel:\n'
    toliv__fkvdm += (
        '    total_valid = bodo.libs.distributed_api.dist_reduce(loc_valid, sum_op)\n'
        )
    toliv__fkvdm += '    return total_valid\n'
    toliv__fkvdm += '  else:\n'
    toliv__fkvdm += '    return loc_valid\n'
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'bodo': bodo, 'pd': pd, 'numba': numba, 'sum_op':
        sum_op}, qhcdq__abnc)
    return qhcdq__abnc['impl']


def quantile(A, q):
    return 0


def quantile_parallel(A, q):
    return 0


@infer_global(quantile)
@infer_global(quantile_parallel)
class QuantileType(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) in [2, 3]
        return signature(types.float64, *unliteral_all(args))


@lower_builtin(quantile, types.Array, types.float64)
@lower_builtin(quantile, IntegerArrayType, types.float64)
@lower_builtin(quantile, BooleanArrayType, types.float64)
def lower_dist_quantile_seq(context, builder, sig, args):
    huymv__kile = numba_to_c_type(sig.args[0].dtype)
    oefwi__eeyj = cgutils.alloca_once_value(builder, lir.Constant(lir.
        IntType(32), huymv__kile))
    smy__rpbw = args[0]
    fbh__zek = sig.args[0]
    if isinstance(fbh__zek, (IntegerArrayType, BooleanArrayType)):
        smy__rpbw = cgutils.create_struct_proxy(fbh__zek)(context, builder,
            smy__rpbw).data
        fbh__zek = types.Array(fbh__zek.dtype, 1, 'C')
    assert fbh__zek.ndim == 1
    arr = make_array(fbh__zek)(context, builder, smy__rpbw)
    kmj__vfjh = builder.extract_value(arr.shape, 0)
    apq__cfcty = [builder.bitcast(arr.data, lir.IntType(8).as_pointer()),
        kmj__vfjh, args[1], builder.load(oefwi__eeyj)]
    ghv__eydi = [lir.IntType(8).as_pointer(), lir.IntType(64), lir.
        DoubleType(), lir.IntType(32)]
    dtq__wza = lir.FunctionType(lir.DoubleType(), ghv__eydi)
    fzhal__qfsz = cgutils.get_or_insert_function(builder.module, dtq__wza,
        name='quantile_sequential')
    cuu__gon = builder.call(fzhal__qfsz, apq__cfcty)
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return cuu__gon


@lower_builtin(quantile_parallel, types.Array, types.float64, types.intp)
@lower_builtin(quantile_parallel, IntegerArrayType, types.float64, types.intp)
@lower_builtin(quantile_parallel, BooleanArrayType, types.float64, types.intp)
def lower_dist_quantile_parallel(context, builder, sig, args):
    huymv__kile = numba_to_c_type(sig.args[0].dtype)
    oefwi__eeyj = cgutils.alloca_once_value(builder, lir.Constant(lir.
        IntType(32), huymv__kile))
    smy__rpbw = args[0]
    fbh__zek = sig.args[0]
    if isinstance(fbh__zek, (IntegerArrayType, BooleanArrayType)):
        smy__rpbw = cgutils.create_struct_proxy(fbh__zek)(context, builder,
            smy__rpbw).data
        fbh__zek = types.Array(fbh__zek.dtype, 1, 'C')
    assert fbh__zek.ndim == 1
    arr = make_array(fbh__zek)(context, builder, smy__rpbw)
    kmj__vfjh = builder.extract_value(arr.shape, 0)
    if len(args) == 3:
        jdxnq__ignln = args[2]
    else:
        jdxnq__ignln = kmj__vfjh
    apq__cfcty = [builder.bitcast(arr.data, lir.IntType(8).as_pointer()),
        kmj__vfjh, jdxnq__ignln, args[1], builder.load(oefwi__eeyj)]
    ghv__eydi = [lir.IntType(8).as_pointer(), lir.IntType(64), lir.IntType(
        64), lir.DoubleType(), lir.IntType(32)]
    dtq__wza = lir.FunctionType(lir.DoubleType(), ghv__eydi)
    fzhal__qfsz = cgutils.get_or_insert_function(builder.module, dtq__wza,
        name='quantile_parallel')
    cuu__gon = builder.call(fzhal__qfsz, apq__cfcty)
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return cuu__gon


@numba.njit
def min_heapify(arr, ind_arr, n, start, cmp_f):
    pndj__gabhf = start
    ctva__yuav = 2 * start + 1
    ckkf__qzuwk = 2 * start + 2
    if ctva__yuav < n and not cmp_f(arr[ctva__yuav], arr[pndj__gabhf]):
        pndj__gabhf = ctva__yuav
    if ckkf__qzuwk < n and not cmp_f(arr[ckkf__qzuwk], arr[pndj__gabhf]):
        pndj__gabhf = ckkf__qzuwk
    if pndj__gabhf != start:
        arr[start], arr[pndj__gabhf] = arr[pndj__gabhf], arr[start]
        ind_arr[start], ind_arr[pndj__gabhf] = ind_arr[pndj__gabhf], ind_arr[
            start]
        min_heapify(arr, ind_arr, n, pndj__gabhf, cmp_f)


def select_k_nonan(A, index_arr, m, k):
    return A[:k]


@overload(select_k_nonan, no_unliteral=True)
def select_k_nonan_overload(A, index_arr, m, k):
    dtype = A.dtype
    if isinstance(dtype, types.Integer):
        return lambda A, index_arr, m, k: (A[:k].copy(), index_arr[:k].copy
            (), k)

    def select_k_nonan_float(A, index_arr, m, k):
        cpr__cfkp = np.empty(k, A.dtype)
        ctshv__hkog = np.empty(k, index_arr.dtype)
        i = 0
        ind = 0
        while i < m and ind < k:
            if not bodo.libs.array_kernels.isna(A, i):
                cpr__cfkp[ind] = A[i]
                ctshv__hkog[ind] = index_arr[i]
                ind += 1
            i += 1
        if ind < k:
            cpr__cfkp = cpr__cfkp[:ind]
            ctshv__hkog = ctshv__hkog[:ind]
        return cpr__cfkp, ctshv__hkog, i
    return select_k_nonan_float


@numba.njit
def nlargest(A, index_arr, k, is_largest, cmp_f):
    m = len(A)
    if k == 0:
        return A[:0], index_arr[:0]
    if k >= m:
        fvcyo__jsurc = np.sort(A)
        bwq__txgt = index_arr[np.argsort(A)]
        atoep__vdgd = pd.Series(fvcyo__jsurc).notna().values
        fvcyo__jsurc = fvcyo__jsurc[atoep__vdgd]
        bwq__txgt = bwq__txgt[atoep__vdgd]
        if is_largest:
            fvcyo__jsurc = fvcyo__jsurc[::-1]
            bwq__txgt = bwq__txgt[::-1]
        return np.ascontiguousarray(fvcyo__jsurc), np.ascontiguousarray(
            bwq__txgt)
    cpr__cfkp, ctshv__hkog, start = select_k_nonan(A, index_arr, m, k)
    ctshv__hkog = ctshv__hkog[cpr__cfkp.argsort()]
    cpr__cfkp.sort()
    if not is_largest:
        cpr__cfkp = np.ascontiguousarray(cpr__cfkp[::-1])
        ctshv__hkog = np.ascontiguousarray(ctshv__hkog[::-1])
    for i in range(start, m):
        if cmp_f(A[i], cpr__cfkp[0]):
            cpr__cfkp[0] = A[i]
            ctshv__hkog[0] = index_arr[i]
            min_heapify(cpr__cfkp, ctshv__hkog, k, 0, cmp_f)
    ctshv__hkog = ctshv__hkog[cpr__cfkp.argsort()]
    cpr__cfkp.sort()
    if is_largest:
        cpr__cfkp = cpr__cfkp[::-1]
        ctshv__hkog = ctshv__hkog[::-1]
    return np.ascontiguousarray(cpr__cfkp), np.ascontiguousarray(ctshv__hkog)


@numba.njit
def nlargest_parallel(A, I, k, is_largest, cmp_f):
    chc__ozqse = bodo.libs.distributed_api.get_rank()
    wpgvo__rfg, cwd__ois = nlargest(A, I, k, is_largest, cmp_f)
    mrrb__qtmaf = bodo.libs.distributed_api.gatherv(wpgvo__rfg)
    hhiv__mcnvu = bodo.libs.distributed_api.gatherv(cwd__ois)
    if chc__ozqse == MPI_ROOT:
        res, idww__vuc = nlargest(mrrb__qtmaf, hhiv__mcnvu, k, is_largest,
            cmp_f)
    else:
        res = np.empty(k, A.dtype)
        idww__vuc = np.empty(k, I.dtype)
    bodo.libs.distributed_api.bcast(res)
    bodo.libs.distributed_api.bcast(idww__vuc)
    return res, idww__vuc


@numba.njit(no_cpython_wrapper=True, cache=True)
def nancorr(mat, cov=0, minpv=1, parallel=False):
    imrgt__vwab, cdwc__skqad = mat.shape
    kkl__gsfq = np.empty((cdwc__skqad, cdwc__skqad), dtype=np.float64)
    for mjiud__homvv in range(cdwc__skqad):
        for teuyq__ozjt in range(mjiud__homvv + 1):
            gsq__ffk = 0
            dae__mctdj = esffa__ugoek = tzdu__iso = gsrf__gxa = 0.0
            for i in range(imrgt__vwab):
                if np.isfinite(mat[i, mjiud__homvv]) and np.isfinite(mat[i,
                    teuyq__ozjt]):
                    gjho__gmhkf = mat[i, mjiud__homvv]
                    qgnv__coe = mat[i, teuyq__ozjt]
                    gsq__ffk += 1
                    tzdu__iso += gjho__gmhkf
                    gsrf__gxa += qgnv__coe
            if parallel:
                gsq__ffk = bodo.libs.distributed_api.dist_reduce(gsq__ffk,
                    sum_op)
                tzdu__iso = bodo.libs.distributed_api.dist_reduce(tzdu__iso,
                    sum_op)
                gsrf__gxa = bodo.libs.distributed_api.dist_reduce(gsrf__gxa,
                    sum_op)
            if gsq__ffk < minpv:
                kkl__gsfq[mjiud__homvv, teuyq__ozjt] = kkl__gsfq[
                    teuyq__ozjt, mjiud__homvv] = np.nan
            else:
                pqx__wzu = tzdu__iso / gsq__ffk
                bbh__pbrw = gsrf__gxa / gsq__ffk
                tzdu__iso = 0.0
                for i in range(imrgt__vwab):
                    if np.isfinite(mat[i, mjiud__homvv]) and np.isfinite(mat
                        [i, teuyq__ozjt]):
                        gjho__gmhkf = mat[i, mjiud__homvv] - pqx__wzu
                        qgnv__coe = mat[i, teuyq__ozjt] - bbh__pbrw
                        tzdu__iso += gjho__gmhkf * qgnv__coe
                        dae__mctdj += gjho__gmhkf * gjho__gmhkf
                        esffa__ugoek += qgnv__coe * qgnv__coe
                if parallel:
                    tzdu__iso = bodo.libs.distributed_api.dist_reduce(tzdu__iso
                        , sum_op)
                    dae__mctdj = bodo.libs.distributed_api.dist_reduce(
                        dae__mctdj, sum_op)
                    esffa__ugoek = bodo.libs.distributed_api.dist_reduce(
                        esffa__ugoek, sum_op)
                mvwlh__yfzg = gsq__ffk - 1.0 if cov else sqrt(dae__mctdj *
                    esffa__ugoek)
                if mvwlh__yfzg != 0.0:
                    kkl__gsfq[mjiud__homvv, teuyq__ozjt] = kkl__gsfq[
                        teuyq__ozjt, mjiud__homvv] = tzdu__iso / mvwlh__yfzg
                else:
                    kkl__gsfq[mjiud__homvv, teuyq__ozjt] = kkl__gsfq[
                        teuyq__ozjt, mjiud__homvv] = np.nan
    return kkl__gsfq


@numba.njit(no_cpython_wrapper=True)
def duplicated(data, ind_arr, parallel=False):
    if parallel:
        data, (ind_arr,) = bodo.ir.join.parallel_shuffle(data, (ind_arr,))
    data = bodo.libs.str_arr_ext.to_list_if_immutable_arr(data)
    n = len(data[0])
    out = np.empty(n, np.bool_)
    toq__ugl = dict()
    for i in range(n):
        val = getitem_arr_tup_single(data, i)
        if val in toq__ugl:
            out[i] = True
        else:
            out[i] = False
            toq__ugl[val] = 0
    return out, ind_arr


@numba.njit(no_cpython_wrapper=True)
def duplicated_array(data, parallel=False):
    if parallel:
        rmdpi__wgre = arr_info_list_to_table([array_to_info(data)])
        vcbxi__pmv = bodo.libs.array.shuffle_table(rmdpi__wgre, 1, parallel, 1)
        data = info_to_array(info_from_table(vcbxi__pmv, 0), data)
        qfg__iszr = bodo.libs.array.get_shuffle_info(vcbxi__pmv)
        bodo.libs.array.delete_table(vcbxi__pmv)
        bodo.libs.array.delete_table(rmdpi__wgre)
    data = bodo.libs.str_arr_ext.to_list_if_immutable_arr(data)
    n = len(data)
    out = np.empty(n, np.bool_)
    toq__ugl = dict()
    cdz__ioazh = False
    for i in range(n):
        if bodo.libs.array_kernels.isna(data, i):
            out[i] = cdz__ioazh
            cdz__ioazh = True
        else:
            val = data[i],
            if val in toq__ugl:
                out[i] = True
            else:
                out[i] = False
                toq__ugl[val] = 0
    if parallel:
        out = bodo.hiframes.pd_groupby_ext.reverse_shuffle(out, qfg__iszr)
    return out


def sample_table_operation(data, ind_arr, n, frac, replace, parallel=False):
    return data, ind_arr


@overload(sample_table_operation, no_unliteral=True)
def overload_sample_table_operation(data, ind_arr, n, frac, replace,
    parallel=False):
    temb__lhu = len(data)
    toliv__fkvdm = (
        'def impl(data, ind_arr, n, frac, replace, parallel=False):\n')
    toliv__fkvdm += ('  info_list_total = [{}, array_to_info(ind_arr)]\n'.
        format(', '.join('array_to_info(data[{}])'.format(x) for x in range
        (temb__lhu))))
    toliv__fkvdm += '  table_total = arr_info_list_to_table(info_list_total)\n'
    toliv__fkvdm += (
        '  out_table = sample_table(table_total, n, frac, replace, parallel)\n'
        .format(temb__lhu))
    for sjl__nhki in range(temb__lhu):
        toliv__fkvdm += (
            '  out_arr_{} = info_to_array(info_from_table(out_table, {}), data[{}])\n'
            .format(sjl__nhki, sjl__nhki, sjl__nhki))
    toliv__fkvdm += (
        '  out_arr_index = info_to_array(info_from_table(out_table, {}), ind_arr)\n'
        .format(temb__lhu))
    toliv__fkvdm += '  delete_table(out_table)\n'
    toliv__fkvdm += '  delete_table(table_total)\n'
    toliv__fkvdm += '  return ({},), out_arr_index\n'.format(', '.join(
        'out_arr_{}'.format(i) for i in range(temb__lhu)))
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'np': np, 'bodo': bodo, 'array_to_info':
        array_to_info, 'sample_table': sample_table,
        'arr_info_list_to_table': arr_info_list_to_table, 'info_from_table':
        info_from_table, 'info_to_array': info_to_array, 'delete_table':
        delete_table, 'delete_table_decref_arrays':
        delete_table_decref_arrays}, qhcdq__abnc)
    impl = qhcdq__abnc['impl']
    return impl


def drop_duplicates(data, ind_arr, ncols, parallel=False):
    return data, ind_arr


@overload(drop_duplicates, no_unliteral=True)
def overload_drop_duplicates(data, ind_arr, ncols, parallel=False):
    temb__lhu = len(data)
    toliv__fkvdm = 'def impl(data, ind_arr, ncols, parallel=False):\n'
    toliv__fkvdm += ('  info_list_total = [{}, array_to_info(ind_arr)]\n'.
        format(', '.join('array_to_info(data[{}])'.format(x) for x in range
        (temb__lhu))))
    toliv__fkvdm += '  table_total = arr_info_list_to_table(info_list_total)\n'
    toliv__fkvdm += '  keep_i = 0\n'
    toliv__fkvdm += """  out_table = drop_duplicates_table(table_total, parallel, ncols, keep_i, False)
"""
    for sjl__nhki in range(temb__lhu):
        toliv__fkvdm += (
            '  out_arr_{} = info_to_array(info_from_table(out_table, {}), data[{}])\n'
            .format(sjl__nhki, sjl__nhki, sjl__nhki))
    toliv__fkvdm += (
        '  out_arr_index = info_to_array(info_from_table(out_table, {}), ind_arr)\n'
        .format(temb__lhu))
    toliv__fkvdm += '  delete_table(out_table)\n'
    toliv__fkvdm += '  delete_table(table_total)\n'
    toliv__fkvdm += '  return ({},), out_arr_index\n'.format(', '.join(
        'out_arr_{}'.format(i) for i in range(temb__lhu)))
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'np': np, 'bodo': bodo, 'array_to_info':
        array_to_info, 'drop_duplicates_table': drop_duplicates_table,
        'arr_info_list_to_table': arr_info_list_to_table, 'info_from_table':
        info_from_table, 'info_to_array': info_to_array, 'delete_table':
        delete_table, 'delete_table_decref_arrays':
        delete_table_decref_arrays}, qhcdq__abnc)
    impl = qhcdq__abnc['impl']
    return impl


def drop_duplicates_array(data_arr, parallel=False):
    return data_arr


@overload(drop_duplicates_array, no_unliteral=True)
def overload_drop_duplicates_array(data_arr, parallel=False):

    def impl(data_arr, parallel=False):
        fybll__byf = [array_to_info(data_arr)]
        sqfg__bag = arr_info_list_to_table(fybll__byf)
        tymot__dtvi = 0
        iengs__qbbn = drop_duplicates_table(sqfg__bag, parallel, 1,
            tymot__dtvi, False)
        rspz__lpo = info_to_array(info_from_table(iengs__qbbn, 0), data_arr)
        delete_table(iengs__qbbn)
        delete_table(sqfg__bag)
        return rspz__lpo
    return impl


def dropna(data, how, thresh, subset, parallel=False):
    return data


@overload(dropna, no_unliteral=True)
def overload_dropna(data, how, thresh, subset):
    zbwj__tvrk = len(data.types)
    jxon__uvpq = [('out' + str(i)) for i in range(zbwj__tvrk)]
    bbhf__msaf = get_overload_const_list(subset)
    how = get_overload_const_str(how)
    ytf__znj = ['isna(data[{}], i)'.format(i) for i in bbhf__msaf]
    gujsm__aki = 'not ({})'.format(' or '.join(ytf__znj))
    if not is_overload_none(thresh):
        gujsm__aki = '(({}) <= ({}) - thresh)'.format(' + '.join(ytf__znj),
            zbwj__tvrk - 1)
    elif how == 'all':
        gujsm__aki = 'not ({})'.format(' and '.join(ytf__znj))
    toliv__fkvdm = 'def _dropna_imp(data, how, thresh, subset):\n'
    toliv__fkvdm += '  old_len = len(data[0])\n'
    toliv__fkvdm += '  new_len = 0\n'
    toliv__fkvdm += '  for i in range(old_len):\n'
    toliv__fkvdm += '    if {}:\n'.format(gujsm__aki)
    toliv__fkvdm += '      new_len += 1\n'
    for i, out in enumerate(jxon__uvpq):
        if isinstance(data[i], bodo.CategoricalArrayType):
            toliv__fkvdm += (
                '  {0} = bodo.utils.utils.alloc_type(new_len, data[{1}], (-1,))\n'
                .format(out, i))
        else:
            toliv__fkvdm += (
                '  {0} = bodo.utils.utils.alloc_type(new_len, t{1}, (-1,))\n'
                .format(out, i))
    toliv__fkvdm += '  curr_ind = 0\n'
    toliv__fkvdm += '  for i in range(old_len):\n'
    toliv__fkvdm += '    if {}:\n'.format(gujsm__aki)
    for i in range(zbwj__tvrk):
        toliv__fkvdm += '      if isna(data[{}], i):\n'.format(i)
        toliv__fkvdm += '        setna({}, curr_ind)\n'.format(jxon__uvpq[i])
        toliv__fkvdm += '      else:\n'
        toliv__fkvdm += '        {}[curr_ind] = data[{}][i]\n'.format(
            jxon__uvpq[i], i)
    toliv__fkvdm += '      curr_ind += 1\n'
    toliv__fkvdm += '  return {}\n'.format(', '.join(jxon__uvpq))
    qhcdq__abnc = {}
    rnq__ikac = {'t{}'.format(i): pbyq__gpksd for i, pbyq__gpksd in
        enumerate(data.types)}
    rnq__ikac.update({'isna': isna, 'setna': setna, 'init_nested_counts':
        bodo.utils.indexing.init_nested_counts, 'add_nested_counts': bodo.
        utils.indexing.add_nested_counts, 'bodo': bodo})
    exec(toliv__fkvdm, rnq__ikac, qhcdq__abnc)
    lbk__rjrw = qhcdq__abnc['_dropna_imp']
    return lbk__rjrw


def get(arr, ind):
    return pd.Series(arr).str.get(ind)


@overload(get, no_unliteral=True)
def overload_get(arr, ind):
    if isinstance(arr, ArrayItemArrayType):
        fbh__zek = arr.dtype
        uboql__wxoaf = fbh__zek.dtype

        def get_arr_item(arr, ind):
            n = len(arr)
            hmj__tipuj = init_nested_counts(uboql__wxoaf)
            for k in range(n):
                if bodo.libs.array_kernels.isna(arr, k):
                    continue
                val = arr[k]
                if not len(val) > ind >= -len(val
                    ) or bodo.libs.array_kernels.isna(val, ind):
                    continue
                hmj__tipuj = add_nested_counts(hmj__tipuj, val[ind])
            rspz__lpo = bodo.utils.utils.alloc_type(n, fbh__zek, hmj__tipuj)
            for buhry__rcrr in range(n):
                if bodo.libs.array_kernels.isna(arr, buhry__rcrr):
                    setna(rspz__lpo, buhry__rcrr)
                    continue
                val = arr[buhry__rcrr]
                if not len(val) > ind >= -len(val
                    ) or bodo.libs.array_kernels.isna(val, ind):
                    setna(rspz__lpo, buhry__rcrr)
                    continue
                rspz__lpo[buhry__rcrr] = val[ind]
            return rspz__lpo
        return get_arr_item


def _is_same_categorical_array_type(arr_types):
    from bodo.hiframes.pd_categorical_ext import _to_readonly
    if not isinstance(arr_types, types.BaseTuple) or len(arr_types) == 0:
        return False
    huppr__vhzwo = _to_readonly(arr_types.types[0])
    return all(isinstance(pbyq__gpksd, CategoricalArrayType) and 
        _to_readonly(pbyq__gpksd) == huppr__vhzwo for pbyq__gpksd in
        arr_types.types)


def concat(arr_list):
    return pd.concat(arr_list)


@overload(concat, no_unliteral=True)
def concat_overload(arr_list):
    if isinstance(arr_list, bodo.NullableTupleType):
        return lambda arr_list: bodo.libs.array_kernels.concat(arr_list._data)
    if isinstance(arr_list, (types.UniTuple, types.List)) and isinstance(
        arr_list.dtype, ArrayItemArrayType):
        treh__yuz = arr_list.dtype.dtype

        def array_item_concat_impl(arr_list):
            jtau__kzspx = 0
            zrthp__vrjws = []
            for A in arr_list:
                nkmhi__tty = len(A)
                bodo.libs.array_item_arr_ext.trim_excess_data(A)
                zrthp__vrjws.append(bodo.libs.array_item_arr_ext.get_data(A))
                jtau__kzspx += nkmhi__tty
            scz__cheqx = np.empty(jtau__kzspx + 1, offset_type)
            dzqwv__uai = bodo.libs.array_kernels.concat(zrthp__vrjws)
            jov__gscda = np.empty(jtau__kzspx + 7 >> 3, np.uint8)
            ngyln__jyj = 0
            dcd__fac = 0
            for A in arr_list:
                qqrog__vafui = bodo.libs.array_item_arr_ext.get_offsets(A)
                wwkj__feq = bodo.libs.array_item_arr_ext.get_null_bitmap(A)
                nkmhi__tty = len(A)
                tfjf__nbeeg = qqrog__vafui[nkmhi__tty]
                for i in range(nkmhi__tty):
                    scz__cheqx[i + ngyln__jyj] = qqrog__vafui[i] + dcd__fac
                    inyv__wujlo = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                        wwkj__feq, i)
                    bodo.libs.int_arr_ext.set_bit_to_arr(jov__gscda, i +
                        ngyln__jyj, inyv__wujlo)
                ngyln__jyj += nkmhi__tty
                dcd__fac += tfjf__nbeeg
            scz__cheqx[ngyln__jyj] = dcd__fac
            rspz__lpo = bodo.libs.array_item_arr_ext.init_array_item_array(
                jtau__kzspx, dzqwv__uai, scz__cheqx, jov__gscda)
            return rspz__lpo
        return array_item_concat_impl
    if isinstance(arr_list, (types.UniTuple, types.List)) and isinstance(
        arr_list.dtype, bodo.StructArrayType):
        saem__yyh = arr_list.dtype.names
        toliv__fkvdm = 'def struct_array_concat_impl(arr_list):\n'
        toliv__fkvdm += f'    n_all = 0\n'
        for i in range(len(saem__yyh)):
            toliv__fkvdm += f'    concat_list{i} = []\n'
        toliv__fkvdm += '    for A in arr_list:\n'
        toliv__fkvdm += (
            '        data_tuple = bodo.libs.struct_arr_ext.get_data(A)\n')
        for i in range(len(saem__yyh)):
            toliv__fkvdm += f'        concat_list{i}.append(data_tuple[{i}])\n'
        toliv__fkvdm += '        n_all += len(A)\n'
        toliv__fkvdm += '    n_bytes = (n_all + 7) >> 3\n'
        toliv__fkvdm += '    new_mask = np.empty(n_bytes, np.uint8)\n'
        toliv__fkvdm += '    curr_bit = 0\n'
        toliv__fkvdm += '    for A in arr_list:\n'
        toliv__fkvdm += (
            '        old_mask = bodo.libs.struct_arr_ext.get_null_bitmap(A)\n')
        toliv__fkvdm += '        for j in range(len(A)):\n'
        toliv__fkvdm += (
            '            bit = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask, j)\n'
            )
        toliv__fkvdm += """            bodo.libs.int_arr_ext.set_bit_to_arr(new_mask, curr_bit, bit)
"""
        toliv__fkvdm += '            curr_bit += 1\n'
        toliv__fkvdm += (
            '    return bodo.libs.struct_arr_ext.init_struct_arr(\n')
        lxu__bsly = ', '.join([
            f'bodo.libs.array_kernels.concat(concat_list{i})' for i in
            range(len(saem__yyh))])
        toliv__fkvdm += f'        ({lxu__bsly},),\n'
        toliv__fkvdm += '        new_mask,\n'
        toliv__fkvdm += f'        {saem__yyh},\n'
        toliv__fkvdm += '    )\n'
        qhcdq__abnc = {}
        exec(toliv__fkvdm, {'bodo': bodo, 'np': np}, qhcdq__abnc)
        return qhcdq__abnc['struct_array_concat_impl']
    if isinstance(arr_list, (types.UniTuple, types.List)
        ) and arr_list.dtype == datetime_date_array_type:

        def datetime_date_array_concat_impl(arr_list):
            eqsj__nak = 0
            for A in arr_list:
                eqsj__nak += len(A)
            rhe__uoz = (bodo.hiframes.datetime_date_ext.
                alloc_datetime_date_array(eqsj__nak))
            kssd__wfgx = 0
            for A in arr_list:
                for i in range(len(A)):
                    rhe__uoz._data[i + kssd__wfgx] = A._data[i]
                    inyv__wujlo = bodo.libs.int_arr_ext.get_bit_bitmap_arr(A
                        ._null_bitmap, i)
                    bodo.libs.int_arr_ext.set_bit_to_arr(rhe__uoz.
                        _null_bitmap, i + kssd__wfgx, inyv__wujlo)
                kssd__wfgx += len(A)
            return rhe__uoz
        return datetime_date_array_concat_impl
    if isinstance(arr_list, (types.UniTuple, types.List)
        ) and arr_list.dtype == datetime_timedelta_array_type:

        def datetime_timedelta_array_concat_impl(arr_list):
            eqsj__nak = 0
            for A in arr_list:
                eqsj__nak += len(A)
            rhe__uoz = (bodo.hiframes.datetime_timedelta_ext.
                alloc_datetime_timedelta_array(eqsj__nak))
            kssd__wfgx = 0
            for A in arr_list:
                for i in range(len(A)):
                    rhe__uoz._days_data[i + kssd__wfgx] = A._days_data[i]
                    rhe__uoz._seconds_data[i + kssd__wfgx] = A._seconds_data[i]
                    rhe__uoz._microseconds_data[i + kssd__wfgx
                        ] = A._microseconds_data[i]
                    inyv__wujlo = bodo.libs.int_arr_ext.get_bit_bitmap_arr(A
                        ._null_bitmap, i)
                    bodo.libs.int_arr_ext.set_bit_to_arr(rhe__uoz.
                        _null_bitmap, i + kssd__wfgx, inyv__wujlo)
                kssd__wfgx += len(A)
            return rhe__uoz
        return datetime_timedelta_array_concat_impl
    if isinstance(arr_list, (types.UniTuple, types.List)) and isinstance(
        arr_list.dtype, DecimalArrayType):
        owa__rnqir = arr_list.dtype.precision
        bclv__ykmn = arr_list.dtype.scale

        def decimal_array_concat_impl(arr_list):
            eqsj__nak = 0
            for A in arr_list:
                eqsj__nak += len(A)
            rhe__uoz = bodo.libs.decimal_arr_ext.alloc_decimal_array(eqsj__nak,
                owa__rnqir, bclv__ykmn)
            kssd__wfgx = 0
            for A in arr_list:
                for i in range(len(A)):
                    rhe__uoz._data[i + kssd__wfgx] = A._data[i]
                    inyv__wujlo = bodo.libs.int_arr_ext.get_bit_bitmap_arr(A
                        ._null_bitmap, i)
                    bodo.libs.int_arr_ext.set_bit_to_arr(rhe__uoz.
                        _null_bitmap, i + kssd__wfgx, inyv__wujlo)
                kssd__wfgx += len(A)
            return rhe__uoz
        return decimal_array_concat_impl
    if isinstance(arr_list, (types.UniTuple, types.List)
        ) and arr_list.dtype in [string_array_type, bodo.binary_array_type]:
        if arr_list.dtype == bodo.binary_array_type:
            lnf__rvaly = 'bodo.libs.str_arr_ext.pre_alloc_binary_array'
        elif arr_list.dtype == string_array_type:
            lnf__rvaly = 'bodo.libs.str_arr_ext.pre_alloc_string_array'
        toliv__fkvdm = 'def impl(arr_list):  # pragma: no cover\n'
        toliv__fkvdm += '    # preallocate the output\n'
        toliv__fkvdm += '    num_strs = 0\n'
        toliv__fkvdm += '    num_chars = 0\n'
        toliv__fkvdm += '    for A in arr_list:\n'
        toliv__fkvdm += '        arr = A\n'
        toliv__fkvdm += '        num_strs += len(arr)\n'
        toliv__fkvdm += (
            '        # this should work for both binary and string\n')
        toliv__fkvdm += (
            '        num_chars += bodo.libs.str_arr_ext.num_total_chars(arr)\n'
            )
        toliv__fkvdm += f'    out_arr = {lnf__rvaly}(\n'
        toliv__fkvdm += '        num_strs, num_chars\n'
        toliv__fkvdm += '    )\n'
        toliv__fkvdm += (
            '    bodo.libs.str_arr_ext.set_null_bits_to_value(out_arr, -1)\n')
        toliv__fkvdm += '    # copy data to output\n'
        toliv__fkvdm += '    curr_str_ind = 0\n'
        toliv__fkvdm += '    curr_chars_ind = 0\n'
        toliv__fkvdm += '    for A in arr_list:\n'
        toliv__fkvdm += '        arr = A\n'
        toliv__fkvdm += '        # This will probably need to be extended\n'
        toliv__fkvdm += (
            '        bodo.libs.str_arr_ext.set_string_array_range(\n')
        toliv__fkvdm += (
            '            out_arr, arr, curr_str_ind, curr_chars_ind\n')
        toliv__fkvdm += '        )\n'
        toliv__fkvdm += '        curr_str_ind += len(arr)\n'
        toliv__fkvdm += (
            '        # this should work for both binary and string\n')
        toliv__fkvdm += (
            '        curr_chars_ind += bodo.libs.str_arr_ext.num_total_chars(arr)\n'
            )
        toliv__fkvdm += '    return out_arr\n'
        miw__xaer = dict()
        exec(toliv__fkvdm, {'bodo': bodo}, miw__xaer)
        meds__iea = miw__xaer['impl']
        return meds__iea
    if isinstance(arr_list, (types.UniTuple, types.List)) and isinstance(
        arr_list.dtype, IntegerArrayType) or isinstance(arr_list, types.
        BaseTuple) and all(isinstance(pbyq__gpksd.dtype, types.Integer) for
        pbyq__gpksd in arr_list.types) and any(isinstance(pbyq__gpksd,
        IntegerArrayType) for pbyq__gpksd in arr_list.types):

        def impl_int_arr_list(arr_list):
            zrqeq__tvzg = convert_to_nullable_tup(arr_list)
            xipuu__ncfh = []
            asfni__kteh = 0
            for A in zrqeq__tvzg:
                xipuu__ncfh.append(A._data)
                asfni__kteh += len(A)
            dzqwv__uai = bodo.libs.array_kernels.concat(xipuu__ncfh)
            jvjwl__cevyf = asfni__kteh + 7 >> 3
            win__utr = np.empty(jvjwl__cevyf, np.uint8)
            ztedd__yftcb = 0
            for A in zrqeq__tvzg:
                qntf__khmgv = A._null_bitmap
                for buhry__rcrr in range(len(A)):
                    inyv__wujlo = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                        qntf__khmgv, buhry__rcrr)
                    bodo.libs.int_arr_ext.set_bit_to_arr(win__utr,
                        ztedd__yftcb, inyv__wujlo)
                    ztedd__yftcb += 1
            return bodo.libs.int_arr_ext.init_integer_array(dzqwv__uai,
                win__utr)
        return impl_int_arr_list
    if isinstance(arr_list, (types.UniTuple, types.List)
        ) and arr_list.dtype == boolean_array or isinstance(arr_list, types
        .BaseTuple) and all(pbyq__gpksd.dtype == types.bool_ for
        pbyq__gpksd in arr_list.types) and any(pbyq__gpksd == boolean_array for
        pbyq__gpksd in arr_list.types):

        def impl_bool_arr_list(arr_list):
            zrqeq__tvzg = convert_to_nullable_tup(arr_list)
            xipuu__ncfh = []
            asfni__kteh = 0
            for A in zrqeq__tvzg:
                xipuu__ncfh.append(A._data)
                asfni__kteh += len(A)
            dzqwv__uai = bodo.libs.array_kernels.concat(xipuu__ncfh)
            jvjwl__cevyf = asfni__kteh + 7 >> 3
            win__utr = np.empty(jvjwl__cevyf, np.uint8)
            ztedd__yftcb = 0
            for A in zrqeq__tvzg:
                qntf__khmgv = A._null_bitmap
                for buhry__rcrr in range(len(A)):
                    inyv__wujlo = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                        qntf__khmgv, buhry__rcrr)
                    bodo.libs.int_arr_ext.set_bit_to_arr(win__utr,
                        ztedd__yftcb, inyv__wujlo)
                    ztedd__yftcb += 1
            return bodo.libs.bool_arr_ext.init_bool_array(dzqwv__uai, win__utr)
        return impl_bool_arr_list
    if isinstance(arr_list, (types.UniTuple, types.List)) and isinstance(
        arr_list.dtype, CategoricalArrayType):

        def cat_array_concat_impl(arr_list):
            dxx__mvqq = []
            for A in arr_list:
                dxx__mvqq.append(A.codes)
            return init_categorical_array(bodo.libs.array_kernels.concat(
                dxx__mvqq), arr_list[0].dtype)
        return cat_array_concat_impl
    if _is_same_categorical_array_type(arr_list):
        uavg__otwr = ', '.join(f'arr_list[{i}].codes' for i in range(len(
            arr_list)))
        toliv__fkvdm = 'def impl(arr_list):\n'
        toliv__fkvdm += f"""    return init_categorical_array(bodo.libs.array_kernels.concat(({uavg__otwr},)), arr_list[0].dtype)
"""
        miw__xaer = {}
        exec(toliv__fkvdm, {'bodo': bodo, 'init_categorical_array':
            init_categorical_array}, miw__xaer)
        return miw__xaer['impl']
    if isinstance(arr_list, types.List) and isinstance(arr_list.dtype,
        types.Array) and arr_list.dtype.ndim == 1:
        dtype = arr_list.dtype.dtype

        def impl_np_arr_list(arr_list):
            asfni__kteh = 0
            for A in arr_list:
                asfni__kteh += len(A)
            rspz__lpo = np.empty(asfni__kteh, dtype)
            iwg__sti = 0
            for A in arr_list:
                n = len(A)
                rspz__lpo[iwg__sti:iwg__sti + n] = A
                iwg__sti += n
            return rspz__lpo
        return impl_np_arr_list
    if isinstance(arr_list, types.BaseTuple) and any(isinstance(pbyq__gpksd,
        (types.Array, IntegerArrayType)) and isinstance(pbyq__gpksd.dtype,
        types.Integer) for pbyq__gpksd in arr_list.types) and any(
        isinstance(pbyq__gpksd, types.Array) and isinstance(pbyq__gpksd.
        dtype, types.Float) for pbyq__gpksd in arr_list.types):
        return lambda arr_list: np.concatenate(astype_float_tup(arr_list))
    if isinstance(arr_list, (types.UniTuple, types.List)) and isinstance(
        arr_list.dtype, bodo.MapArrayType):

        def impl_map_arr_list(arr_list):
            eqs__wlnw = []
            for A in arr_list:
                eqs__wlnw.append(A._data)
            eauu__azz = bodo.libs.array_kernels.concat(eqs__wlnw)
            kkl__gsfq = bodo.libs.map_arr_ext.init_map_arr(eauu__azz)
            return kkl__gsfq
        return impl_map_arr_list
    for mtqd__gujrs in arr_list:
        if not isinstance(mtqd__gujrs, types.Array):
            raise_bodo_error(f'concat of array types {arr_list} not supported')
    return lambda arr_list: np.concatenate(arr_list)


def astype_float_tup(arr_tup):
    return tuple(pbyq__gpksd.astype(np.float64) for pbyq__gpksd in arr_tup)


@overload(astype_float_tup, no_unliteral=True)
def overload_astype_float_tup(arr_tup):
    assert isinstance(arr_tup, types.BaseTuple)
    temb__lhu = len(arr_tup.types)
    toliv__fkvdm = 'def f(arr_tup):\n'
    toliv__fkvdm += '  return ({}{})\n'.format(','.join(
        'arr_tup[{}].astype(np.float64)'.format(i) for i in range(temb__lhu
        )), ',' if temb__lhu == 1 else '')
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'np': np}, qhcdq__abnc)
    vhgvv__hbn = qhcdq__abnc['f']
    return vhgvv__hbn


def convert_to_nullable_tup(arr_tup):
    return arr_tup


@overload(convert_to_nullable_tup, no_unliteral=True)
def overload_convert_to_nullable_tup(arr_tup):
    if isinstance(arr_tup, (types.UniTuple, types.List)) and isinstance(arr_tup
        .dtype, (IntegerArrayType, BooleanArrayType)):
        return lambda arr_tup: arr_tup
    assert isinstance(arr_tup, types.BaseTuple)
    temb__lhu = len(arr_tup.types)
    iox__kzx = find_common_np_dtype(arr_tup.types)
    uboql__wxoaf = None
    wqnsf__suo = ''
    if isinstance(iox__kzx, types.Integer):
        uboql__wxoaf = bodo.libs.int_arr_ext.IntDtype(iox__kzx)
        wqnsf__suo = '.astype(out_dtype, False)'
    toliv__fkvdm = 'def f(arr_tup):\n'
    toliv__fkvdm += '  return ({}{})\n'.format(','.join(
        'bodo.utils.conversion.coerce_to_array(arr_tup[{}], use_nullable_array=True){}'
        .format(i, wqnsf__suo) for i in range(temb__lhu)), ',' if temb__lhu ==
        1 else '')
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'bodo': bodo, 'out_dtype': uboql__wxoaf}, qhcdq__abnc)
    myk__ikoyk = qhcdq__abnc['f']
    return myk__ikoyk


def nunique(A, dropna):
    return len(set(A))


def nunique_parallel(A, dropna):
    return len(set(A))


@overload(nunique, no_unliteral=True)
def nunique_overload(A, dropna):

    def nunique_seq(A, dropna):
        s, jhc__bdj = build_set_seen_na(A)
        return len(s) + int(not dropna and jhc__bdj)
    return nunique_seq


@overload(nunique_parallel, no_unliteral=True)
def nunique_overload_parallel(A, dropna):
    sum_op = bodo.libs.distributed_api.Reduce_Type.Sum.value

    def nunique_par(A, dropna):
        onng__kxtr = bodo.libs.array_kernels.unique(A, dropna, parallel=True)
        kphb__fjhf = len(onng__kxtr)
        return bodo.libs.distributed_api.dist_reduce(kphb__fjhf, np.int32(
            sum_op))
    return nunique_par


def unique(A, dropna=False, parallel=False):
    return np.array([orbw__jswjo for orbw__jswjo in set(A)]).astype(A.dtype)


def cummin(A):
    return A


@overload(cummin, no_unliteral=True)
def cummin_overload(A):
    if isinstance(A.dtype, types.Float):
        jmybs__ooilo = np.finfo(A.dtype(1).dtype).max
    else:
        jmybs__ooilo = np.iinfo(A.dtype(1).dtype).max

    def impl(A):
        n = len(A)
        rspz__lpo = np.empty(n, A.dtype)
        ymtie__mbyp = jmybs__ooilo
        for i in range(n):
            ymtie__mbyp = min(ymtie__mbyp, A[i])
            rspz__lpo[i] = ymtie__mbyp
        return rspz__lpo
    return impl


def cummax(A):
    return A


@overload(cummax, no_unliteral=True)
def cummax_overload(A):
    if isinstance(A.dtype, types.Float):
        jmybs__ooilo = np.finfo(A.dtype(1).dtype).min
    else:
        jmybs__ooilo = np.iinfo(A.dtype(1).dtype).min

    def impl(A):
        n = len(A)
        rspz__lpo = np.empty(n, A.dtype)
        ymtie__mbyp = jmybs__ooilo
        for i in range(n):
            ymtie__mbyp = max(ymtie__mbyp, A[i])
            rspz__lpo[i] = ymtie__mbyp
        return rspz__lpo
    return impl


@overload(unique, no_unliteral=True)
def unique_overload(A, dropna=False, parallel=False):

    def unique_impl(A, dropna=False, parallel=False):
        zykr__nnl = arr_info_list_to_table([array_to_info(A)])
        pbpal__qabh = 1
        tymot__dtvi = 0
        iengs__qbbn = drop_duplicates_table(zykr__nnl, parallel,
            pbpal__qabh, tymot__dtvi, dropna)
        rspz__lpo = info_to_array(info_from_table(iengs__qbbn, 0), A)
        delete_table(zykr__nnl)
        delete_table(iengs__qbbn)
        return rspz__lpo
    return unique_impl


def explode(arr, index_arr):
    return pd.Series(arr, index_arr).explode()


@overload(explode, no_unliteral=True)
def overload_explode(arr, index_arr):
    assert isinstance(arr, ArrayItemArrayType
        ) or arr == string_array_split_view_type
    treh__yuz = bodo.utils.typing.to_nullable_type(arr.dtype)
    ekvd__jtv = index_arr
    izd__bnqax = ekvd__jtv.dtype

    def impl(arr, index_arr):
        n = len(arr)
        hmj__tipuj = init_nested_counts(treh__yuz)
        kzv__lzamy = init_nested_counts(izd__bnqax)
        for i in range(n):
            tist__gazdj = index_arr[i]
            if isna(arr, i):
                hmj__tipuj = (hmj__tipuj[0] + 1,) + hmj__tipuj[1:]
                kzv__lzamy = add_nested_counts(kzv__lzamy, tist__gazdj)
                continue
            tgfbu__mtp = arr[i]
            if len(tgfbu__mtp) == 0:
                hmj__tipuj = (hmj__tipuj[0] + 1,) + hmj__tipuj[1:]
                kzv__lzamy = add_nested_counts(kzv__lzamy, tist__gazdj)
                continue
            hmj__tipuj = add_nested_counts(hmj__tipuj, tgfbu__mtp)
            for queq__iogyc in range(len(tgfbu__mtp)):
                kzv__lzamy = add_nested_counts(kzv__lzamy, tist__gazdj)
        rspz__lpo = bodo.utils.utils.alloc_type(hmj__tipuj[0], treh__yuz,
            hmj__tipuj[1:])
        ekq__qfg = bodo.utils.utils.alloc_type(hmj__tipuj[0], ekvd__jtv,
            kzv__lzamy)
        dcd__fac = 0
        for i in range(n):
            if isna(arr, i):
                setna(rspz__lpo, dcd__fac)
                ekq__qfg[dcd__fac] = index_arr[i]
                dcd__fac += 1
                continue
            tgfbu__mtp = arr[i]
            tfjf__nbeeg = len(tgfbu__mtp)
            if tfjf__nbeeg == 0:
                setna(rspz__lpo, dcd__fac)
                ekq__qfg[dcd__fac] = index_arr[i]
                dcd__fac += 1
                continue
            rspz__lpo[dcd__fac:dcd__fac + tfjf__nbeeg] = tgfbu__mtp
            ekq__qfg[dcd__fac:dcd__fac + tfjf__nbeeg] = index_arr[i]
            dcd__fac += tfjf__nbeeg
        return rspz__lpo, ekq__qfg
    return impl


def explode_no_index(arr):
    return pd.Series(arr).explode()


@overload(explode_no_index, no_unliteral=True)
def overload_explode_no_index(arr, counts):
    assert isinstance(arr, ArrayItemArrayType
        ) or arr == string_array_split_view_type
    treh__yuz = bodo.utils.typing.to_nullable_type(arr.dtype)

    def impl(arr, counts):
        n = len(arr)
        hmj__tipuj = init_nested_counts(treh__yuz)
        for i in range(n):
            if isna(arr, i):
                hmj__tipuj = (hmj__tipuj[0] + 1,) + hmj__tipuj[1:]
                svdzp__cqhlm = 1
            else:
                tgfbu__mtp = arr[i]
                wvz__lhniv = len(tgfbu__mtp)
                if wvz__lhniv == 0:
                    hmj__tipuj = (hmj__tipuj[0] + 1,) + hmj__tipuj[1:]
                    svdzp__cqhlm = 1
                    continue
                else:
                    hmj__tipuj = add_nested_counts(hmj__tipuj, tgfbu__mtp)
                    svdzp__cqhlm = wvz__lhniv
            if counts[i] != svdzp__cqhlm:
                raise ValueError(
                    'DataFrame.explode(): columns must have matching element counts'
                    )
        rspz__lpo = bodo.utils.utils.alloc_type(hmj__tipuj[0], treh__yuz,
            hmj__tipuj[1:])
        dcd__fac = 0
        for i in range(n):
            if isna(arr, i):
                setna(rspz__lpo, dcd__fac)
                dcd__fac += 1
                continue
            tgfbu__mtp = arr[i]
            tfjf__nbeeg = len(tgfbu__mtp)
            if tfjf__nbeeg == 0:
                setna(rspz__lpo, dcd__fac)
                dcd__fac += 1
                continue
            rspz__lpo[dcd__fac:dcd__fac + tfjf__nbeeg] = tgfbu__mtp
            dcd__fac += tfjf__nbeeg
        return rspz__lpo
    return impl


def get_arr_lens(arr, na_empty_as_one=True):
    return [len(nszrm__bfysu) for nszrm__bfysu in arr]


@overload(get_arr_lens, inline='always', no_unliteral=True)
def overload_get_arr_lens(arr, na_empty_as_one=True):
    na_empty_as_one = get_overload_const_bool(na_empty_as_one)
    assert isinstance(arr, ArrayItemArrayType
        ) or arr == string_array_split_view_type or arr == string_array_type and not na_empty_as_one
    if na_empty_as_one:
        fuey__byomy = 'np.empty(n, np.int64)'
        bzry__tuzgl = 'out_arr[i] = 1'
        frs__jhf = 'max(len(arr[i]), 1)'
    else:
        fuey__byomy = 'bodo.libs.int_arr_ext.alloc_int_array(n, np.int64)'
        bzry__tuzgl = 'bodo.libs.array_kernels.setna(out_arr, i)'
        frs__jhf = 'len(arr[i])'
    toliv__fkvdm = f"""def impl(arr, na_empty_as_one=True):
    numba.parfors.parfor.init_prange()
    n = len(arr)
    out_arr = {fuey__byomy}
    for i in numba.parfors.parfor.internal_prange(n):
        if bodo.libs.array_kernels.isna(arr, i):
            {bzry__tuzgl}
        else:
            out_arr[i] = {frs__jhf}
    return out_arr
    """
    qhcdq__abnc = {}
    exec(toliv__fkvdm, {'bodo': bodo, 'numba': numba, 'np': np}, qhcdq__abnc)
    impl = qhcdq__abnc['impl']
    return impl


def explode_str_split(arr, pat, n, index_arr):
    return pd.Series(arr, index_arr).str.split(pat, n).explode()


@overload(explode_str_split, no_unliteral=True)
def overload_explode_str_split(arr, pat, n, index_arr):
    assert arr == string_array_type
    ekvd__jtv = index_arr
    izd__bnqax = ekvd__jtv.dtype

    def impl(arr, pat, n, index_arr):
        xyfpo__ane = pat is not None and len(pat) > 1
        if xyfpo__ane:
            tsll__mvql = re.compile(pat)
            if n == -1:
                n = 0
        elif n == 0:
            n = -1
        txm__esotc = len(arr)
        ytnvo__qcdyb = 0
        pip__ulg = 0
        kzv__lzamy = init_nested_counts(izd__bnqax)
        for i in range(txm__esotc):
            tist__gazdj = index_arr[i]
            if bodo.libs.array_kernels.isna(arr, i):
                ytnvo__qcdyb += 1
                kzv__lzamy = add_nested_counts(kzv__lzamy, tist__gazdj)
                continue
            if xyfpo__ane:
                bbte__fsbqn = tsll__mvql.split(arr[i], maxsplit=n)
            else:
                bbte__fsbqn = arr[i].split(pat, n)
            ytnvo__qcdyb += len(bbte__fsbqn)
            for s in bbte__fsbqn:
                kzv__lzamy = add_nested_counts(kzv__lzamy, tist__gazdj)
                pip__ulg += bodo.libs.str_arr_ext.get_utf8_size(s)
        rspz__lpo = bodo.libs.str_arr_ext.pre_alloc_string_array(ytnvo__qcdyb,
            pip__ulg)
        ekq__qfg = bodo.utils.utils.alloc_type(ytnvo__qcdyb, ekvd__jtv,
            kzv__lzamy)
        naa__wwnn = 0
        for buhry__rcrr in range(txm__esotc):
            if isna(arr, buhry__rcrr):
                rspz__lpo[naa__wwnn] = ''
                bodo.libs.array_kernels.setna(rspz__lpo, naa__wwnn)
                ekq__qfg[naa__wwnn] = index_arr[buhry__rcrr]
                naa__wwnn += 1
                continue
            if xyfpo__ane:
                bbte__fsbqn = tsll__mvql.split(arr[buhry__rcrr], maxsplit=n)
            else:
                bbte__fsbqn = arr[buhry__rcrr].split(pat, n)
            txxvv__hfknu = len(bbte__fsbqn)
            rspz__lpo[naa__wwnn:naa__wwnn + txxvv__hfknu] = bbte__fsbqn
            ekq__qfg[naa__wwnn:naa__wwnn + txxvv__hfknu] = index_arr[
                buhry__rcrr]
            naa__wwnn += txxvv__hfknu
        return rspz__lpo, ekq__qfg
    return impl


def gen_na_array(n, arr):
    return np.full(n, np.nan)


@overload(gen_na_array, no_unliteral=True)
def overload_gen_na_array(n, arr):
    if isinstance(arr, types.TypeRef):
        arr = arr.instance_type
    dtype = arr.dtype
    if isinstance(dtype, (types.Integer, types.Float)):
        dtype = dtype if isinstance(dtype, types.Float) else types.float64

        def impl_float(n, arr):
            numba.parfors.parfor.init_prange()
            rspz__lpo = np.empty(n, dtype)
            for i in numba.parfors.parfor.internal_prange(n):
                rspz__lpo[i] = np.nan
            return rspz__lpo
        return impl_float

    def impl(n, arr):
        numba.parfors.parfor.init_prange()
        rspz__lpo = bodo.utils.utils.alloc_type(n, arr, (0,))
        for i in numba.parfors.parfor.internal_prange(n):
            setna(rspz__lpo, i)
        return rspz__lpo
    return impl


def gen_na_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


ArrayAnalysis._analyze_op_call_bodo_libs_array_kernels_gen_na_array = (
    gen_na_array_equiv)


def resize_and_copy(A, new_len):
    return A


@overload(resize_and_copy, no_unliteral=True)
def overload_resize_and_copy(A, old_size, new_len):
    zax__fgu = A
    if A == types.Array(types.uint8, 1, 'C'):

        def impl_char(A, old_size, new_len):
            rspz__lpo = bodo.utils.utils.alloc_type(new_len, zax__fgu)
            bodo.libs.str_arr_ext.str_copy_ptr(rspz__lpo.ctypes, 0, A.
                ctypes, old_size)
            return rspz__lpo
        return impl_char

    def impl(A, old_size, new_len):
        rspz__lpo = bodo.utils.utils.alloc_type(new_len, zax__fgu, (-1,))
        rspz__lpo[:old_size] = A[:old_size]
        return rspz__lpo
    return impl


@register_jitable
def calc_nitems(start, stop, step):
    awl__pcmb = math.ceil((stop - start) / step)
    return int(max(awl__pcmb, 0))


def calc_nitems_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 3 and not kws
    if guard(find_const, self.func_ir, args[0]) == 0 and guard(find_const,
        self.func_ir, args[2]) == 1:
        return ArrayAnalysis.AnalyzeResult(shape=args[1], pre=[])


ArrayAnalysis._analyze_op_call_bodo_libs_array_kernels_calc_nitems = (
    calc_nitems_equiv)


def arange_parallel_impl(return_type, *args):
    dtype = as_dtype(return_type.dtype)

    def arange_1(stop):
        return np.arange(0, stop, 1, dtype)

    def arange_2(start, stop):
        return np.arange(start, stop, 1, dtype)

    def arange_3(start, stop, step):
        return np.arange(start, stop, step, dtype)
    if any(isinstance(orbw__jswjo, types.Complex) for orbw__jswjo in args):

        def arange_4(start, stop, step, dtype):
            numba.parfors.parfor.init_prange()
            jsps__cjgvk = (stop - start) / step
            awl__pcmb = math.ceil(jsps__cjgvk.real)
            myr__cwot = math.ceil(jsps__cjgvk.imag)
            vuim__epdyz = int(max(min(myr__cwot, awl__pcmb), 0))
            arr = np.empty(vuim__epdyz, dtype)
            for i in numba.parfors.parfor.internal_prange(vuim__epdyz):
                arr[i] = start + i * step
            return arr
    else:

        def arange_4(start, stop, step, dtype):
            numba.parfors.parfor.init_prange()
            vuim__epdyz = bodo.libs.array_kernels.calc_nitems(start, stop, step
                )
            arr = np.empty(vuim__epdyz, dtype)
            for i in numba.parfors.parfor.internal_prange(vuim__epdyz):
                arr[i] = start + i * step
            return arr
    if len(args) == 1:
        return arange_1
    elif len(args) == 2:
        return arange_2
    elif len(args) == 3:
        return arange_3
    elif len(args) == 4:
        return arange_4
    else:
        raise BodoError('parallel arange with types {}'.format(args))


if bodo.numba_compat._check_numba_change:
    lines = inspect.getsource(numba.parfors.parfor.arange_parallel_impl)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'c72b0390b4f3e52dcc5426bd42c6b55ff96bae5a425381900985d36e7527a4bd':
        warnings.warn('numba.parfors.parfor.arange_parallel_impl has changed')
numba.parfors.parfor.swap_functions_map['arange', 'numpy'
    ] = arange_parallel_impl


def sort(arr, ascending, inplace):
    return np.sort(arr)


@overload(sort, no_unliteral=True)
def overload_sort(arr, ascending, inplace):

    def impl(arr, ascending, inplace):
        n = len(arr)
        data = np.arange(n),
        opmam__ceoaw = arr,
        if not inplace:
            opmam__ceoaw = arr.copy(),
        vhohv__xpah = bodo.libs.str_arr_ext.to_list_if_immutable_arr(
            opmam__ceoaw)
        adtba__moku = bodo.libs.str_arr_ext.to_list_if_immutable_arr(data, True
            )
        bodo.libs.timsort.sort(vhohv__xpah, 0, n, adtba__moku)
        if not ascending:
            bodo.libs.timsort.reverseRange(vhohv__xpah, 0, n, adtba__moku)
        bodo.libs.str_arr_ext.cp_str_list_to_array(opmam__ceoaw, vhohv__xpah)
        return opmam__ceoaw[0]
    return impl


def overload_array_max(A):
    if isinstance(A, IntegerArrayType) or A == boolean_array:

        def impl(A):
            return pd.Series(A).max()
        return impl


overload(np.max, inline='always', no_unliteral=True)(overload_array_max)
overload(max, inline='always', no_unliteral=True)(overload_array_max)


def overload_array_min(A):
    if isinstance(A, IntegerArrayType) or A == boolean_array:

        def impl(A):
            return pd.Series(A).min()
        return impl


overload(np.min, inline='always', no_unliteral=True)(overload_array_min)
overload(min, inline='always', no_unliteral=True)(overload_array_min)


def overload_array_sum(A):
    if isinstance(A, IntegerArrayType) or A == boolean_array:

        def impl(A):
            return pd.Series(A).sum()
    return impl


overload(np.sum, inline='always', no_unliteral=True)(overload_array_sum)
overload(sum, inline='always', no_unliteral=True)(overload_array_sum)


@overload(np.prod, inline='always', no_unliteral=True)
def overload_array_prod(A):
    if isinstance(A, IntegerArrayType) or A == boolean_array:

        def impl(A):
            return pd.Series(A).prod()
    return impl


def nonzero(arr):
    return arr,


@overload(nonzero, no_unliteral=True)
def nonzero_overload(A, parallel=False):
    if not bodo.utils.utils.is_array_typ(A, False):
        return

    def impl(A, parallel=False):
        n = len(A)
        if parallel:
            offset = bodo.libs.distributed_api.dist_exscan(n, Reduce_Type.
                Sum.value)
        else:
            offset = 0
        kkl__gsfq = []
        for i in range(n):
            if A[i]:
                kkl__gsfq.append(i + offset)
        return np.array(kkl__gsfq, np.int64),
    return impl


def ffill_bfill_arr(arr):
    return arr


@overload(ffill_bfill_arr, no_unliteral=True)
def ffill_bfill_overload(A, method, parallel=False):
    zax__fgu = element_type(A)
    if zax__fgu == types.unicode_type:
        null_value = '""'
    elif zax__fgu == types.bool_:
        null_value = 'False'
    elif zax__fgu == bodo.datetime64ns:
        null_value = (
            'bodo.utils.conversion.unbox_if_timestamp(pd.to_datetime(0))')
    elif zax__fgu == bodo.timedelta64ns:
        null_value = (
            'bodo.utils.conversion.unbox_if_timestamp(pd.to_timedelta(0))')
    else:
        null_value = '0'
    naa__wwnn = 'i'
    iiam__pzidv = False
    zzmo__lmue = get_overload_const_str(method)
    if zzmo__lmue in ('ffill', 'pad'):
        ksyby__bafq = 'n'
        send_right = True
    elif zzmo__lmue in ('backfill', 'bfill'):
        ksyby__bafq = 'n-1, -1, -1'
        send_right = False
        if zax__fgu == types.unicode_type:
            naa__wwnn = '(n - 1) - i'
            iiam__pzidv = True
    toliv__fkvdm = 'def impl(A, method, parallel=False):\n'
    toliv__fkvdm += '  has_last_value = False\n'
    toliv__fkvdm += f'  last_value = {null_value}\n'
    toliv__fkvdm += '  if parallel:\n'
    toliv__fkvdm += '    rank = bodo.libs.distributed_api.get_rank()\n'
    toliv__fkvdm += '    n_pes = bodo.libs.distributed_api.get_size()\n'
    toliv__fkvdm += f"""    has_last_value, last_value = null_border_icomm(A, rank, n_pes, {null_value}, {send_right})
"""
    toliv__fkvdm += '  n = len(A)\n'
    toliv__fkvdm += '  out_arr = bodo.utils.utils.alloc_type(n, A, (-1,))\n'
    toliv__fkvdm += f'  for i in range({ksyby__bafq}):\n'
    toliv__fkvdm += (
        '    if (bodo.libs.array_kernels.isna(A, i) and not has_last_value):\n'
        )
    toliv__fkvdm += (
        f'      bodo.libs.array_kernels.setna(out_arr, {naa__wwnn})\n')
    toliv__fkvdm += '      continue\n'
    toliv__fkvdm += '    s = A[i]\n'
    toliv__fkvdm += '    if bodo.libs.array_kernels.isna(A, i):\n'
    toliv__fkvdm += '      s = last_value\n'
    toliv__fkvdm += f'    out_arr[{naa__wwnn}] = s\n'
    toliv__fkvdm += '    last_value = s\n'
    toliv__fkvdm += '    has_last_value = True\n'
    if iiam__pzidv:
        toliv__fkvdm += '  return out_arr[::-1]\n'
    else:
        toliv__fkvdm += '  return out_arr\n'
    ikn__uaqxp = {}
    exec(toliv__fkvdm, {'bodo': bodo, 'numba': numba, 'pd': pd,
        'null_border_icomm': null_border_icomm}, ikn__uaqxp)
    impl = ikn__uaqxp['impl']
    return impl


@register_jitable(cache=True)
def null_border_icomm(in_arr, rank, n_pes, null_value, send_right=True):
    if send_right:
        nlycn__gej = 0
        olma__qalg = n_pes - 1
        hmjj__fdcld = np.int32(rank + 1)
        mzkzl__scxsg = np.int32(rank - 1)
        nphub__vmjs = len(in_arr) - 1
        eji__fhdyg = -1
        ftksh__mlvx = -1
    else:
        nlycn__gej = n_pes - 1
        olma__qalg = 0
        hmjj__fdcld = np.int32(rank - 1)
        mzkzl__scxsg = np.int32(rank + 1)
        nphub__vmjs = 0
        eji__fhdyg = len(in_arr)
        ftksh__mlvx = 1
    ckxuv__budm = np.int32(bodo.hiframes.rolling.comm_border_tag)
    sqvc__pqodx = np.empty(1, dtype=np.bool_)
    twgzy__cng = bodo.utils.utils.alloc_type(1, in_arr, (-1,))
    dvih__fgmdg = np.empty(1, dtype=np.bool_)
    bpw__rolr = bodo.utils.utils.alloc_type(1, in_arr, (-1,))
    ngl__jljb = False
    jsei__vosw = null_value
    for i in range(nphub__vmjs, eji__fhdyg, ftksh__mlvx):
        if not isna(in_arr, i):
            ngl__jljb = True
            jsei__vosw = in_arr[i]
            break
    if rank != nlycn__gej:
        pqaa__ihib = bodo.libs.distributed_api.irecv(sqvc__pqodx, 1,
            mzkzl__scxsg, ckxuv__budm, True)
        bodo.libs.distributed_api.wait(pqaa__ihib, True)
        ahqe__roy = bodo.libs.distributed_api.irecv(twgzy__cng, 1,
            mzkzl__scxsg, ckxuv__budm, True)
        bodo.libs.distributed_api.wait(ahqe__roy, True)
        tclt__mztlt = sqvc__pqodx[0]
        okia__qzqr = twgzy__cng[0]
    else:
        tclt__mztlt = False
        okia__qzqr = null_value
    if ngl__jljb:
        dvih__fgmdg[0] = ngl__jljb
        bpw__rolr[0] = jsei__vosw
    else:
        dvih__fgmdg[0] = tclt__mztlt
        bpw__rolr[0] = okia__qzqr
    if rank != olma__qalg:
        wgm__bugdu = bodo.libs.distributed_api.isend(dvih__fgmdg, 1,
            hmjj__fdcld, ckxuv__budm, True)
        aotd__ful = bodo.libs.distributed_api.isend(bpw__rolr, 1,
            hmjj__fdcld, ckxuv__budm, True)
    return tclt__mztlt, okia__qzqr


@overload(np.sort, inline='always', no_unliteral=True)
def np_sort(A, axis=-1, kind=None, order=None):
    if not bodo.utils.utils.is_array_typ(A, False) or isinstance(A, types.Array
        ):
        return
    fpk__wfut = {'axis': axis, 'kind': kind, 'order': order}
    ikkls__odyz = {'axis': -1, 'kind': None, 'order': None}
    check_unsupported_args('np.sort', fpk__wfut, ikkls__odyz, 'numpy')

    def impl(A, axis=-1, kind=None, order=None):
        return pd.Series(A).sort_values().values
    return impl


def repeat_kernel(A, repeats):
    return A


@overload(repeat_kernel, no_unliteral=True)
def repeat_kernel_overload(A, repeats):
    zax__fgu = A
    if isinstance(repeats, types.Integer):

        def impl_int(A, repeats):
            txm__esotc = len(A)
            rspz__lpo = bodo.utils.utils.alloc_type(txm__esotc * repeats,
                zax__fgu, (-1,))
            for i in range(txm__esotc):
                naa__wwnn = i * repeats
                if bodo.libs.array_kernels.isna(A, i):
                    for buhry__rcrr in range(repeats):
                        bodo.libs.array_kernels.setna(rspz__lpo, naa__wwnn +
                            buhry__rcrr)
                else:
                    rspz__lpo[naa__wwnn:naa__wwnn + repeats] = A[i]
            return rspz__lpo
        return impl_int

    def impl_arr(A, repeats):
        txm__esotc = len(A)
        rspz__lpo = bodo.utils.utils.alloc_type(repeats.sum(), zax__fgu, (-1,))
        naa__wwnn = 0
        for i in range(txm__esotc):
            wodb__pcx = repeats[i]
            if bodo.libs.array_kernels.isna(A, i):
                for buhry__rcrr in range(wodb__pcx):
                    bodo.libs.array_kernels.setna(rspz__lpo, naa__wwnn +
                        buhry__rcrr)
            else:
                rspz__lpo[naa__wwnn:naa__wwnn + wodb__pcx] = A[i]
            naa__wwnn += wodb__pcx
        return rspz__lpo
    return impl_arr


@overload(np.repeat, inline='always', no_unliteral=True)
def np_repeat(A, repeats):
    if not bodo.utils.utils.is_array_typ(A, False) or isinstance(A, types.Array
        ):
        return
    if not isinstance(repeats, types.Integer):
        raise BodoError(
            'Only integer type supported for repeats in np.repeat()')

    def impl(A, repeats):
        return bodo.libs.array_kernels.repeat_kernel(A, repeats)
    return impl


@overload(np.unique, inline='always', no_unliteral=True)
def np_unique(A):
    if not bodo.utils.utils.is_array_typ(A, False) or isinstance(A, types.Array
        ):
        return

    def impl(A):
        cwxrc__aea = bodo.libs.array_kernels.unique(A)
        return bodo.allgatherv(cwxrc__aea, False)
    return impl


@overload(np.union1d, inline='always', no_unliteral=True)
def overload_union1d(A1, A2):
    if not bodo.utils.utils.is_array_typ(A1, False
        ) or not bodo.utils.utils.is_array_typ(A2, False):
        return
    if A1 != A2:
        raise BodoError('Both arrays must be the same type in np.union1d()')

    def impl(A1, A2):
        ytlb__jpf = bodo.libs.array_kernels.concat([A1, A2])
        tlic__emdf = bodo.libs.array_kernels.unique(ytlb__jpf)
        return pd.Series(tlic__emdf).sort_values().values
    return impl


@overload(np.intersect1d, inline='always', no_unliteral=True)
def overload_intersect1d(A1, A2, assume_unique=False, return_indices=False):
    if not bodo.utils.utils.is_array_typ(A1, False
        ) or not bodo.utils.utils.is_array_typ(A2, False):
        return
    fpk__wfut = {'assume_unique': assume_unique, 'return_indices':
        return_indices}
    ikkls__odyz = {'assume_unique': False, 'return_indices': False}
    check_unsupported_args('np.intersect1d', fpk__wfut, ikkls__odyz, 'numpy')
    if A1 != A2:
        raise BodoError('Both arrays must be the same type in np.intersect1d()'
            )
    if A1.ndim != 1 or A2.ndim != 1:
        raise BodoError('Only 1D arrays supported in np.intersect1d()')

    def impl(A1, A2, assume_unique=False, return_indices=False):
        zii__bqkwv = bodo.libs.array_kernels.unique(A1)
        nepkw__edb = bodo.libs.array_kernels.unique(A2)
        ytlb__jpf = bodo.libs.array_kernels.concat([zii__bqkwv, nepkw__edb])
        tenq__xsb = pd.Series(ytlb__jpf).sort_values().values
        return slice_array_intersect1d(tenq__xsb)
    return impl


@register_jitable
def slice_array_intersect1d(arr):
    atoep__vdgd = arr[1:] == arr[:-1]
    return arr[:-1][atoep__vdgd]


@overload(np.setdiff1d, inline='always', no_unliteral=True)
def overload_setdiff1d(A1, A2, assume_unique=False):
    if not bodo.utils.utils.is_array_typ(A1, False
        ) or not bodo.utils.utils.is_array_typ(A2, False):
        return
    fpk__wfut = {'assume_unique': assume_unique}
    ikkls__odyz = {'assume_unique': False}
    check_unsupported_args('np.setdiff1d', fpk__wfut, ikkls__odyz, 'numpy')
    if A1 != A2:
        raise BodoError('Both arrays must be the same type in np.setdiff1d()')
    if A1.ndim != 1 or A2.ndim != 1:
        raise BodoError('Only 1D arrays supported in np.setdiff1d()')

    def impl(A1, A2, assume_unique=False):
        zii__bqkwv = bodo.libs.array_kernels.unique(A1)
        nepkw__edb = bodo.libs.array_kernels.unique(A2)
        atoep__vdgd = calculate_mask_setdiff1d(zii__bqkwv, nepkw__edb)
        return pd.Series(zii__bqkwv[atoep__vdgd]).sort_values().values
    return impl


@register_jitable
def calculate_mask_setdiff1d(A1, A2):
    atoep__vdgd = np.ones(len(A1), np.bool_)
    for i in range(len(A2)):
        atoep__vdgd &= A1 != A2[i]
    return atoep__vdgd


@overload(np.linspace, inline='always', no_unliteral=True)
def np_linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=
    None, axis=0):
    fpk__wfut = {'retstep': retstep, 'axis': axis}
    ikkls__odyz = {'retstep': False, 'axis': 0}
    check_unsupported_args('np.linspace', fpk__wfut, ikkls__odyz, 'numpy')
    nxg__beby = False
    if is_overload_none(dtype):
        zax__fgu = np.promote_types(np.promote_types(numba.np.numpy_support
            .as_dtype(start), numba.np.numpy_support.as_dtype(stop)), numba
            .np.numpy_support.as_dtype(types.float64)).type
    else:
        if isinstance(dtype.dtype, types.Integer):
            nxg__beby = True
        zax__fgu = numba.np.numpy_support.as_dtype(dtype).type
    if nxg__beby:

        def impl_int(start, stop, num=50, endpoint=True, retstep=False,
            dtype=None, axis=0):
            vlwlc__vrse = np_linspace_get_stepsize(start, stop, num, endpoint)
            numba.parfors.parfor.init_prange()
            rspz__lpo = np.empty(num, zax__fgu)
            for i in numba.parfors.parfor.internal_prange(num):
                rspz__lpo[i] = zax__fgu(np.floor(start + i * vlwlc__vrse))
            return rspz__lpo
        return impl_int
    else:

        def impl(start, stop, num=50, endpoint=True, retstep=False, dtype=
            None, axis=0):
            vlwlc__vrse = np_linspace_get_stepsize(start, stop, num, endpoint)
            numba.parfors.parfor.init_prange()
            rspz__lpo = np.empty(num, zax__fgu)
            for i in numba.parfors.parfor.internal_prange(num):
                rspz__lpo[i] = zax__fgu(start + i * vlwlc__vrse)
            return rspz__lpo
        return impl


def np_linspace_get_stepsize(start, stop, num, endpoint):
    return 0


@overload(np_linspace_get_stepsize, no_unliteral=True)
def overload_np_linspace_get_stepsize(start, stop, num, endpoint):

    def impl(start, stop, num, endpoint):
        if num < 0:
            raise ValueError('np.linspace() Num must be >= 0')
        if endpoint:
            num -= 1
        if num > 1:
            return (stop - start) / num
        return 0
    return impl


@overload(operator.contains, no_unliteral=True)
def arr_contains(A, val):
    if not (bodo.utils.utils.is_array_typ(A, False) and A.dtype == types.
        unliteral(val)):
        return

    def impl(A, val):
        numba.parfors.parfor.init_prange()
        temb__lhu = 0
        n = len(A)
        for i in numba.parfors.parfor.internal_prange(n):
            if not bodo.libs.array_kernels.isna(A, i):
                temb__lhu += A[i] == val
        return temb__lhu > 0
    return impl


@overload(np.any, inline='always', no_unliteral=True)
def np_any(A, axis=None, out=None, keepdims=None):
    if not (bodo.utils.utils.is_array_typ(A, False) and A.ndim == 1):
        return
    fpk__wfut = {'axis': axis, 'out': out, 'keepdims': keepdims}
    ikkls__odyz = {'axis': None, 'out': None, 'keepdims': None}
    check_unsupported_args('np.any', fpk__wfut, ikkls__odyz, 'numpy')

    def impl(A, axis=None, out=None, keepdims=None):
        numba.parfors.parfor.init_prange()
        temb__lhu = 0
        n = len(A)
        for i in numba.parfors.parfor.internal_prange(n):
            if not bodo.libs.array_kernels.isna(A, i):
                temb__lhu += int(bool(A[i]))
        return temb__lhu > 0
    return impl


@overload(np.all, inline='always', no_unliteral=True)
def np_all(A, axis=None, out=None, keepdims=None):
    if not (bodo.utils.utils.is_array_typ(A, False) and A.ndim == 1):
        return
    fpk__wfut = {'axis': axis, 'out': out, 'keepdims': keepdims}
    ikkls__odyz = {'axis': None, 'out': None, 'keepdims': None}
    check_unsupported_args('np.any', fpk__wfut, ikkls__odyz, 'numpy')

    def impl(A, axis=None, out=None, keepdims=None):
        numba.parfors.parfor.init_prange()
        temb__lhu = 0
        n = len(A)
        for i in numba.parfors.parfor.internal_prange(n):
            if not bodo.libs.array_kernels.isna(A, i):
                temb__lhu += int(bool(A[i]))
        return temb__lhu == n
    return impl


@overload(np.cbrt, inline='always', no_unliteral=True)
def np_cbrt(A, out=None, where=True, casting='same_kind', order='K', dtype=
    None, subok=True):
    if not (isinstance(A, types.Number) or bodo.utils.utils.is_array_typ(A,
        False) and A.ndim == 1 and isinstance(A.dtype, types.Number)):
        return
    fpk__wfut = {'out': out, 'where': where, 'casting': casting, 'order':
        order, 'dtype': dtype, 'subok': subok}
    ikkls__odyz = {'out': None, 'where': True, 'casting': 'same_kind',
        'order': 'K', 'dtype': None, 'subok': True}
    check_unsupported_args('np.cbrt', fpk__wfut, ikkls__odyz, 'numpy')
    if bodo.utils.utils.is_array_typ(A, False):
        ivkrg__pijx = np.promote_types(numba.np.numpy_support.as_dtype(A.
            dtype), numba.np.numpy_support.as_dtype(types.float32)).type

        def impl_arr(A, out=None, where=True, casting='same_kind', order=
            'K', dtype=None, subok=True):
            numba.parfors.parfor.init_prange()
            n = len(A)
            rspz__lpo = np.empty(n, ivkrg__pijx)
            for i in numba.parfors.parfor.internal_prange(n):
                if bodo.libs.array_kernels.isna(A, i):
                    bodo.libs.array_kernels.setna(rspz__lpo, i)
                    continue
                rspz__lpo[i] = np_cbrt_scalar(A[i], ivkrg__pijx)
            return rspz__lpo
        return impl_arr
    ivkrg__pijx = np.promote_types(numba.np.numpy_support.as_dtype(A),
        numba.np.numpy_support.as_dtype(types.float32)).type

    def impl_scalar(A, out=None, where=True, casting='same_kind', order='K',
        dtype=None, subok=True):
        return np_cbrt_scalar(A, ivkrg__pijx)
    return impl_scalar


@register_jitable
def np_cbrt_scalar(x, float_dtype):
    if np.isnan(x):
        return np.nan
    jiyfv__pqifw = x < 0
    if jiyfv__pqifw:
        x = -x
    res = np.power(float_dtype(x), 1.0 / 3.0)
    if jiyfv__pqifw:
        return -res
    return res


@overload(np.hstack, no_unliteral=True)
def np_hstack(tup):
    lmm__hjaac = isinstance(tup, (types.BaseTuple, types.List))
    cfeku__sgsm = isinstance(tup, (bodo.SeriesType, bodo.hiframes.
        pd_series_ext.HeterogeneousSeriesType)) and isinstance(tup.data, (
        types.BaseTuple, types.List, bodo.NullableTupleType))
    if isinstance(tup, types.BaseTuple):
        for mtqd__gujrs in tup.types:
            lmm__hjaac = lmm__hjaac and bodo.utils.utils.is_array_typ(
                mtqd__gujrs, False)
    elif isinstance(tup, types.List):
        lmm__hjaac = bodo.utils.utils.is_array_typ(tup.dtype, False)
    elif cfeku__sgsm:
        ujwg__yukm = tup.data.tuple_typ if isinstance(tup.data, bodo.
            NullableTupleType) else tup.data
        for mtqd__gujrs in ujwg__yukm.types:
            cfeku__sgsm = cfeku__sgsm and bodo.utils.utils.is_array_typ(
                mtqd__gujrs, False)
    if not (lmm__hjaac or cfeku__sgsm):
        return
    if cfeku__sgsm:

        def impl_series(tup):
            arr_tup = bodo.hiframes.pd_series_ext.get_series_data(tup)
            return bodo.libs.array_kernels.concat(arr_tup)
        return impl_series

    def impl(tup):
        return bodo.libs.array_kernels.concat(tup)
    return impl


@overload(np.random.multivariate_normal, inline='always', no_unliteral=True)
def np_random_multivariate_normal(mean, cov, size=None, check_valid='warn',
    tol=1e-08):
    fpk__wfut = {'check_valid': check_valid, 'tol': tol}
    ikkls__odyz = {'check_valid': 'warn', 'tol': 1e-08}
    check_unsupported_args('np.random.multivariate_normal', fpk__wfut,
        ikkls__odyz, 'numpy')
    if not isinstance(size, types.Integer):
        raise BodoError(
            'np.random.multivariate_normal() size argument is required and must be an integer'
            )
    if not (bodo.utils.utils.is_array_typ(mean, False) and mean.ndim == 1):
        raise BodoError(
            'np.random.multivariate_normal() mean must be a 1 dimensional numpy array'
            )
    if not (bodo.utils.utils.is_array_typ(cov, False) and cov.ndim == 2):
        raise BodoError(
            'np.random.multivariate_normal() cov must be a 2 dimensional square, numpy array'
            )

    def impl(mean, cov, size=None, check_valid='warn', tol=1e-08):
        _validate_multivar_norm(cov)
        imrgt__vwab = mean.shape[0]
        tlto__htk = size, imrgt__vwab
        fzkq__wyudd = np.random.standard_normal(tlto__htk)
        cov = cov.astype(np.float64)
        bjb__ttm, s, khequ__nhs = np.linalg.svd(cov)
        res = np.dot(fzkq__wyudd, np.sqrt(s).reshape(imrgt__vwab, 1) *
            khequ__nhs)
        zkip__lyq = res + mean
        return zkip__lyq
    return impl


def _validate_multivar_norm(cov):
    return


@overload(_validate_multivar_norm, no_unliteral=True)
def _overload_validate_multivar_norm(cov):

    def impl(cov):
        if cov.shape[0] != cov.shape[1]:
            raise ValueError(
                'np.random.multivariate_normal() cov must be a 2 dimensional square, numpy array'
                )
    return impl


def _nan_argmin(arr):
    return


@overload(_nan_argmin, no_unliteral=True)
def _overload_nan_argmin(arr):
    if isinstance(arr, IntegerArrayType) or arr in [boolean_array,
        datetime_date_array_type] or arr.dtype == bodo.timedelta64ns:

        def impl_bodo_arr(arr):
            numba.parfors.parfor.init_prange()
            bca__yyyp = bodo.hiframes.series_kernels._get_type_max_value(arr)
            ket__zdmim = typing.builtins.IndexValue(-1, bca__yyyp)
            for i in numba.parfors.parfor.internal_prange(len(arr)):
                if bodo.libs.array_kernels.isna(arr, i):
                    continue
                ipfcy__onnzh = typing.builtins.IndexValue(i, arr[i])
                ket__zdmim = min(ket__zdmim, ipfcy__onnzh)
            return ket__zdmim.index
        return impl_bodo_arr
    if isinstance(arr, CategoricalArrayType):
        assert arr.dtype.ordered, 'Categorical Array must be ordered to select an argmin'
        jju__ijv = bodo.hiframes.pd_categorical_ext.get_categories_int_type(arr
            .dtype)

        def impl_cat_arr(arr):
            irjd__xsqi = (bodo.hiframes.pd_categorical_ext.
                get_categorical_arr_codes(arr))
            numba.parfors.parfor.init_prange()
            bca__yyyp = jju__ijv(len(arr.dtype.categories) + 1)
            ket__zdmim = typing.builtins.IndexValue(-1, bca__yyyp)
            for i in numba.parfors.parfor.internal_prange(len(arr)):
                if bodo.libs.array_kernels.isna(arr, i):
                    continue
                ipfcy__onnzh = typing.builtins.IndexValue(i, irjd__xsqi[i])
                ket__zdmim = min(ket__zdmim, ipfcy__onnzh)
            return ket__zdmim.index
        return impl_cat_arr
    return lambda arr: arr.argmin()


def _nan_argmax(arr):
    return


@overload(_nan_argmax, no_unliteral=True)
def _overload_nan_argmax(arr):
    if isinstance(arr, IntegerArrayType) or arr in [boolean_array,
        datetime_date_array_type] or arr.dtype == bodo.timedelta64ns:

        def impl_bodo_arr(arr):
            n = len(arr)
            numba.parfors.parfor.init_prange()
            bca__yyyp = bodo.hiframes.series_kernels._get_type_min_value(arr)
            ket__zdmim = typing.builtins.IndexValue(-1, bca__yyyp)
            for i in numba.parfors.parfor.internal_prange(n):
                if bodo.libs.array_kernels.isna(arr, i):
                    continue
                ipfcy__onnzh = typing.builtins.IndexValue(i, arr[i])
                ket__zdmim = max(ket__zdmim, ipfcy__onnzh)
            return ket__zdmim.index
        return impl_bodo_arr
    if isinstance(arr, CategoricalArrayType):
        assert arr.dtype.ordered, 'Categorical Array must be ordered to select an argmin'
        jju__ijv = bodo.hiframes.pd_categorical_ext.get_categories_int_type(arr
            .dtype)

        def impl_cat_arr(arr):
            n = len(arr)
            irjd__xsqi = (bodo.hiframes.pd_categorical_ext.
                get_categorical_arr_codes(arr))
            numba.parfors.parfor.init_prange()
            bca__yyyp = jju__ijv(-1)
            ket__zdmim = typing.builtins.IndexValue(-1, bca__yyyp)
            for i in numba.parfors.parfor.internal_prange(n):
                if bodo.libs.array_kernels.isna(arr, i):
                    continue
                ipfcy__onnzh = typing.builtins.IndexValue(i, irjd__xsqi[i])
                ket__zdmim = max(ket__zdmim, ipfcy__onnzh)
            return ket__zdmim.index
        return impl_cat_arr
    return lambda arr: arr.argmax()


@overload_attribute(types.Array, 'nbytes', inline='always')
def overload_dataframe_index(A):
    return lambda A: A.size * bodo.io.np_io.get_dtype_size(A.dtype)
