""" Implementation of binary operators for the different types.
    Currently implemented operators:
        arith: add, sub, mul, truediv, floordiv, mod, pow
        cmp: lt, le, eq, ne, ge, gt
"""
import operator
import numba
from numba.core import types
from numba.core.imputils import lower_builtin
from numba.core.typing.builtins import machine_ints
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import overload
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type, datetime_timedelta_type
from bodo.hiframes.datetime_timedelta_ext import datetime_datetime_type, datetime_timedelta_array_type, pd_timedelta_type
from bodo.hiframes.pd_dataframe_ext import DataFrameType
from bodo.hiframes.pd_index_ext import DatetimeIndexType, HeterogeneousIndexType, is_index_type
from bodo.hiframes.pd_offsets_ext import date_offset_type, month_begin_type, month_end_type, week_type
from bodo.hiframes.pd_timestamp_ext import pd_timestamp_type
from bodo.hiframes.series_impl import SeriesType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import Decimal128Type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.utils.typing import BodoError, is_overload_bool, is_timedelta_type


class SeriesCmpOpTemplate(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 2
        lhs, rhs = args
        if cmp_timeseries(lhs, rhs) or (isinstance(lhs, DataFrameType) or
            isinstance(rhs, DataFrameType)) or not (isinstance(lhs,
            SeriesType) or isinstance(rhs, SeriesType)):
            return
        glx__xydc = lhs.data if isinstance(lhs, SeriesType) else lhs
        twlk__ekj = rhs.data if isinstance(rhs, SeriesType) else rhs
        if glx__xydc in (bodo.pd_timestamp_type, bodo.pd_timedelta_type
            ) and twlk__ekj.dtype in (bodo.datetime64ns, bodo.timedelta64ns):
            glx__xydc = twlk__ekj.dtype
        elif twlk__ekj in (bodo.pd_timestamp_type, bodo.pd_timedelta_type
            ) and glx__xydc.dtype in (bodo.datetime64ns, bodo.timedelta64ns):
            twlk__ekj = glx__xydc.dtype
        bhn__azku = glx__xydc, twlk__ekj
        mqu__jduv = (
            f'{lhs} {numba.core.utils.OPERATORS_TO_BUILTINS[self.key]} {rhs} not supported'
            )
        try:
            mzco__nlork = self.context.resolve_function_type(self.key,
                bhn__azku, {}).return_type
        except Exception as aiq__lkco:
            raise BodoError(mqu__jduv)
        if is_overload_bool(mzco__nlork):
            raise BodoError(mqu__jduv)
        fjujy__ddtat = lhs.index if isinstance(lhs, SeriesType) else rhs.index
        vnv__gen = lhs.name_typ if isinstance(lhs, SeriesType
            ) else rhs.name_typ
        tyft__aapc = types.bool_
        friz__xelr = SeriesType(tyft__aapc, mzco__nlork, fjujy__ddtat, vnv__gen
            )
        return friz__xelr(*args)


def series_cmp_op_lower(op):

    def lower_impl(context, builder, sig, args):
        szz__nbmd = bodo.hiframes.series_impl.create_binary_op_overload(op)(*
            sig.args)
        if szz__nbmd is None:
            szz__nbmd = create_overload_cmp_operator(op)(*sig.args)
        return context.compile_internal(builder, szz__nbmd, sig, args)
    return lower_impl


class SeriesAndOrTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert len(args) == 2
        assert not kws
        lhs, rhs = args
        if not (isinstance(lhs, SeriesType) or isinstance(rhs, SeriesType)):
            return
        glx__xydc = lhs.data if isinstance(lhs, SeriesType) else lhs
        twlk__ekj = rhs.data if isinstance(rhs, SeriesType) else rhs
        bhn__azku = glx__xydc, twlk__ekj
        mqu__jduv = (
            f'{lhs} {numba.core.utils.OPERATORS_TO_BUILTINS[self.key]} {rhs} not supported'
            )
        try:
            mzco__nlork = self.context.resolve_function_type(self.key,
                bhn__azku, {}).return_type
        except Exception as jdopp__irek:
            raise BodoError(mqu__jduv)
        fjujy__ddtat = lhs.index if isinstance(lhs, SeriesType) else rhs.index
        vnv__gen = lhs.name_typ if isinstance(lhs, SeriesType
            ) else rhs.name_typ
        tyft__aapc = mzco__nlork.dtype
        friz__xelr = SeriesType(tyft__aapc, mzco__nlork, fjujy__ddtat, vnv__gen
            )
        return friz__xelr(*args)


def lower_series_and_or(op):

    def lower_and_or_impl(context, builder, sig, args):
        szz__nbmd = bodo.hiframes.series_impl.create_binary_op_overload(op)(*
            sig.args)
        if szz__nbmd is None:
            lhs, rhs = sig.args
            if isinstance(lhs, DataFrameType) or isinstance(rhs, DataFrameType
                ):
                szz__nbmd = (bodo.hiframes.dataframe_impl.
                    create_binary_op_overload(op)(*sig.args))
        return context.compile_internal(builder, szz__nbmd, sig, args)
    return lower_and_or_impl


def overload_add_operator_scalars(lhs, rhs):
    if lhs == week_type or rhs == week_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_week_offset_type(lhs, rhs))
    if lhs == month_begin_type or rhs == month_begin_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_month_begin_offset_type(lhs, rhs))
    if lhs == month_end_type or rhs == month_end_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_month_end_offset_type(lhs, rhs))
    if lhs == date_offset_type or rhs == date_offset_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_date_offset_type(lhs, rhs))
    if add_timestamp(lhs, rhs):
        return bodo.hiframes.pd_timestamp_ext.overload_add_operator_timestamp(
            lhs, rhs)
    if add_dt_td_and_dt_date(lhs, rhs):
        return (bodo.hiframes.datetime_date_ext.
            overload_add_operator_datetime_date(lhs, rhs))
    if add_datetime_and_timedeltas(lhs, rhs):
        return (bodo.hiframes.datetime_timedelta_ext.
            overload_add_operator_datetime_timedelta(lhs, rhs))
    raise_error_if_not_numba_supported(operator.add, lhs, rhs)


def overload_sub_operator_scalars(lhs, rhs):
    if sub_offset_to_datetime_or_timestamp(lhs, rhs):
        return bodo.hiframes.pd_offsets_ext.overload_sub_operator_offsets(lhs,
            rhs)
    if lhs == pd_timestamp_type and rhs in [pd_timestamp_type,
        datetime_timedelta_type, pd_timedelta_type]:
        return bodo.hiframes.pd_timestamp_ext.overload_sub_operator_timestamp(
            lhs, rhs)
    if sub_dt_or_td(lhs, rhs):
        return (bodo.hiframes.datetime_date_ext.
            overload_sub_operator_datetime_date(lhs, rhs))
    if sub_datetime_and_timedeltas(lhs, rhs):
        return (bodo.hiframes.datetime_timedelta_ext.
            overload_sub_operator_datetime_timedelta(lhs, rhs))
    if lhs == datetime_datetime_type and rhs == datetime_datetime_type:
        return (bodo.hiframes.datetime_datetime_ext.
            overload_sub_operator_datetime_datetime(lhs, rhs))
    raise_error_if_not_numba_supported(operator.sub, lhs, rhs)


def create_overload_arith_op(op):

    def overload_arith_operator(lhs, rhs):
        if isinstance(lhs, DataFrameType) or isinstance(rhs, DataFrameType):
            return bodo.hiframes.dataframe_impl.create_binary_op_overload(op)(
                lhs, rhs)
        if time_series_operation(lhs, rhs) and op in [operator.add,
            operator.sub]:
            return bodo.hiframes.series_dt_impl.create_bin_op_overload(op)(lhs,
                rhs)
        if isinstance(lhs, SeriesType) or isinstance(rhs, SeriesType):
            return bodo.hiframes.series_impl.create_binary_op_overload(op)(lhs,
                rhs)
        if sub_dt_index_and_timestamp(lhs, rhs) and op == operator.sub:
            return (bodo.hiframes.pd_index_ext.
                overload_sub_operator_datetime_index(lhs, rhs))
        if operand_is_index(lhs) or operand_is_index(rhs):
            return bodo.hiframes.pd_index_ext.create_binary_op_overload(op)(lhs
                , rhs)
        if args_td_and_int_array(lhs, rhs):
            return bodo.libs.int_arr_ext.get_int_array_op_pd_td(op)(lhs, rhs)
        if isinstance(lhs, IntegerArrayType) or isinstance(rhs,
            IntegerArrayType):
            return bodo.libs.int_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if lhs == boolean_array or rhs == boolean_array:
            return bodo.libs.bool_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if op == operator.add and (lhs == string_array_type or types.
            unliteral(lhs) == string_type):
            return bodo.libs.str_arr_ext.overload_add_operator_string_array(lhs
                , rhs)
        if op == operator.add:
            return overload_add_operator_scalars(lhs, rhs)
        if op == operator.sub:
            return overload_sub_operator_scalars(lhs, rhs)
        if op == operator.mul:
            if mul_timedelta_and_int(lhs, rhs):
                return (bodo.hiframes.datetime_timedelta_ext.
                    overload_mul_operator_timedelta(lhs, rhs))
            if mul_string_arr_and_int(lhs, rhs):
                return bodo.libs.str_arr_ext.overload_mul_operator_str_arr(lhs,
                    rhs)
            if mul_date_offset_and_int(lhs, rhs):
                return (bodo.hiframes.pd_offsets_ext.
                    overload_mul_date_offset_types(lhs, rhs))
            raise_error_if_not_numba_supported(op, lhs, rhs)
        if op in [operator.truediv, operator.floordiv]:
            if div_timedelta_and_int(lhs, rhs):
                if op == operator.truediv:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_truediv_operator_pd_timedelta(lhs, rhs))
                else:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_floordiv_operator_pd_timedelta(lhs, rhs))
            if div_datetime_timedelta(lhs, rhs):
                if op == operator.truediv:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_truediv_operator_dt_timedelta(lhs, rhs))
                else:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_floordiv_operator_dt_timedelta(lhs, rhs))
            raise_error_if_not_numba_supported(op, lhs, rhs)
        if op == operator.mod:
            if mod_timedeltas(lhs, rhs):
                return (bodo.hiframes.datetime_timedelta_ext.
                    overload_mod_operator_timedeltas(lhs, rhs))
            raise_error_if_not_numba_supported(op, lhs, rhs)
        if op == operator.pow:
            raise_error_if_not_numba_supported(op, lhs, rhs)
        raise BodoError(
            f'{op} operator not supported for data types {lhs} and {rhs}.')
    return overload_arith_operator


def create_overload_cmp_operator(op):

    def overload_cmp_operator(lhs, rhs):
        if isinstance(lhs, DataFrameType) or isinstance(rhs, DataFrameType):
            return bodo.hiframes.dataframe_impl.create_binary_op_overload(op)(
                lhs, rhs)
        if cmp_timeseries(lhs, rhs):
            return bodo.hiframes.series_dt_impl.create_cmp_op_overload(op)(lhs,
                rhs)
        if isinstance(lhs, SeriesType) or isinstance(rhs, SeriesType):
            return
        if lhs == datetime_date_array_type or rhs == datetime_date_array_type:
            return bodo.hiframes.datetime_date_ext.create_cmp_op_overload_arr(
                op)(lhs, rhs)
        if (lhs == datetime_timedelta_array_type or rhs ==
            datetime_timedelta_array_type):
            szz__nbmd = (bodo.hiframes.datetime_timedelta_ext.
                create_cmp_op_overload(op))
            return szz__nbmd(lhs, rhs)
        if lhs == string_array_type or rhs == string_array_type:
            return bodo.libs.str_arr_ext.create_binary_op_overload(op)(lhs, rhs
                )
        if isinstance(lhs, Decimal128Type) and isinstance(rhs, Decimal128Type):
            return bodo.libs.decimal_arr_ext.decimal_create_cmp_op_overload(op
                )(lhs, rhs)
        if lhs == boolean_array or rhs == boolean_array:
            return bodo.libs.bool_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if isinstance(lhs, IntegerArrayType) or isinstance(rhs,
            IntegerArrayType):
            return bodo.libs.int_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if binary_array_cmp(lhs, rhs):
            return bodo.libs.binary_arr_ext.create_binary_cmp_op_overload(op)(
                lhs, rhs)
        if cmp_dt_index_to_string(lhs, rhs):
            return bodo.hiframes.pd_index_ext.overload_binop_dti_str(op)(lhs,
                rhs)
        if operand_is_index(lhs) or operand_is_index(rhs):
            return bodo.hiframes.pd_index_ext.create_binary_op_overload(op)(lhs
                , rhs)
        if lhs == datetime_date_type and rhs == datetime_date_type:
            return bodo.hiframes.datetime_date_ext.create_cmp_op_overload(op)(
                lhs, rhs)
        if can_cmp_date_datetime(lhs, rhs, op):
            return (bodo.hiframes.datetime_date_ext.
                create_datetime_date_cmp_op_overload(op)(lhs, rhs))
        if lhs == datetime_datetime_type and rhs == datetime_datetime_type:
            return bodo.hiframes.datetime_datetime_ext.create_cmp_op_overload(
                op)(lhs, rhs)
        if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:
            return bodo.hiframes.datetime_timedelta_ext.create_cmp_op_overload(
                op)(lhs, rhs)
        if cmp_timedeltas(lhs, rhs):
            szz__nbmd = (bodo.hiframes.datetime_timedelta_ext.
                pd_create_cmp_op_overload(op))
            return szz__nbmd(lhs, rhs)
        if cmp_timestamp_or_date(lhs, rhs):
            return (bodo.hiframes.pd_timestamp_ext.
                create_timestamp_cmp_op_overload(op)(lhs, rhs))
        if cmp_op_supported_by_numba(lhs, rhs):
            return
        raise BodoError(
            f'{op} operator not supported for data types {lhs} and {rhs}.')
    return overload_cmp_operator


def add_dt_td_and_dt_date(lhs, rhs):
    ihic__kaj = lhs == datetime_timedelta_type and rhs == datetime_date_type
    quwi__nuhdy = rhs == datetime_timedelta_type and lhs == datetime_date_type
    return ihic__kaj or quwi__nuhdy


def add_timestamp(lhs, rhs):
    wgj__nyu = lhs == pd_timestamp_type and is_timedelta_type(rhs)
    ygpe__dkupk = is_timedelta_type(lhs) and rhs == pd_timestamp_type
    return wgj__nyu or ygpe__dkupk


def add_datetime_and_timedeltas(lhs, rhs):
    xsmi__tzd = [datetime_timedelta_type, pd_timedelta_type]
    wtk__hwjg = [datetime_timedelta_type, pd_timedelta_type,
        datetime_datetime_type]
    muytl__dkh = lhs in xsmi__tzd and rhs in xsmi__tzd
    nss__kuvg = (lhs == datetime_datetime_type and rhs in xsmi__tzd or rhs ==
        datetime_datetime_type and lhs in xsmi__tzd)
    return muytl__dkh or nss__kuvg


def mul_string_arr_and_int(lhs, rhs):
    twlk__ekj = isinstance(lhs, types.Integer) and rhs == string_array_type
    glx__xydc = lhs == string_array_type and isinstance(rhs, types.Integer)
    return twlk__ekj or glx__xydc


def mul_timedelta_and_int(lhs, rhs):
    ihic__kaj = lhs in [pd_timedelta_type, datetime_timedelta_type
        ] and isinstance(rhs, types.Integer)
    quwi__nuhdy = rhs in [pd_timedelta_type, datetime_timedelta_type
        ] and isinstance(lhs, types.Integer)
    return ihic__kaj or quwi__nuhdy


def mul_date_offset_and_int(lhs, rhs):
    hjvmf__zidg = lhs in [week_type, month_end_type, month_begin_type,
        date_offset_type] and isinstance(rhs, types.Integer)
    dvrb__qvecb = rhs in [week_type, month_end_type, month_begin_type,
        date_offset_type] and isinstance(lhs, types.Integer)
    return hjvmf__zidg or dvrb__qvecb


def sub_offset_to_datetime_or_timestamp(lhs, rhs):
    znqht__ajlyz = [datetime_datetime_type, pd_timestamp_type,
        datetime_date_type]
    qvi__szz = [date_offset_type, month_begin_type, month_end_type, week_type]
    return rhs in qvi__szz and lhs in znqht__ajlyz


def sub_dt_index_and_timestamp(lhs, rhs):
    dcwvu__swom = isinstance(lhs, DatetimeIndexType
        ) and rhs == pd_timestamp_type
    koz__fcf = isinstance(rhs, DatetimeIndexType) and lhs == pd_timestamp_type
    return dcwvu__swom or koz__fcf


def sub_dt_or_td(lhs, rhs):
    hfpn__gvsva = lhs == datetime_date_type and rhs == datetime_timedelta_type
    wgsax__cdtz = lhs == datetime_date_type and rhs == datetime_date_type
    sbix__fyoh = (lhs == datetime_date_array_type and rhs ==
        datetime_timedelta_type)
    return hfpn__gvsva or wgsax__cdtz or sbix__fyoh


def sub_datetime_and_timedeltas(lhs, rhs):
    ytj__ubl = (is_timedelta_type(lhs) or lhs == datetime_datetime_type
        ) and is_timedelta_type(rhs)
    svt__hvuh = (lhs == datetime_timedelta_array_type and rhs ==
        datetime_timedelta_type)
    return ytj__ubl or svt__hvuh


def div_timedelta_and_int(lhs, rhs):
    muytl__dkh = lhs == pd_timedelta_type and rhs == pd_timedelta_type
    jfq__jpzd = lhs == pd_timedelta_type and isinstance(rhs, types.Integer)
    return muytl__dkh or jfq__jpzd


def div_datetime_timedelta(lhs, rhs):
    muytl__dkh = (lhs == datetime_timedelta_type and rhs ==
        datetime_timedelta_type)
    jfq__jpzd = lhs == datetime_timedelta_type and rhs == types.int64
    return muytl__dkh or jfq__jpzd


def mod_timedeltas(lhs, rhs):
    voe__hmjvt = lhs == pd_timedelta_type and rhs == pd_timedelta_type
    ktif__doa = (lhs == datetime_timedelta_type and rhs ==
        datetime_timedelta_type)
    return voe__hmjvt or ktif__doa


def cmp_dt_index_to_string(lhs, rhs):
    dcwvu__swom = isinstance(lhs, DatetimeIndexType) and types.unliteral(rhs
        ) == string_type
    koz__fcf = isinstance(rhs, DatetimeIndexType) and types.unliteral(lhs
        ) == string_type
    return dcwvu__swom or koz__fcf


def cmp_timestamp_or_date(lhs, rhs):
    mhf__iil = (lhs == pd_timestamp_type and rhs == bodo.hiframes.
        datetime_date_ext.datetime_date_type)
    nrwza__hoken = (lhs == bodo.hiframes.datetime_date_ext.
        datetime_date_type and rhs == pd_timestamp_type)
    pxjy__xava = lhs == pd_timestamp_type and rhs == pd_timestamp_type
    cvgu__txwe = lhs == pd_timestamp_type and rhs == bodo.datetime64ns
    udc__wpqy = rhs == pd_timestamp_type and lhs == bodo.datetime64ns
    return mhf__iil or nrwza__hoken or pxjy__xava or cvgu__txwe or udc__wpqy


def cmp_timeseries(lhs, rhs):
    xwol__hwp = bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs) and (bodo
        .utils.typing.is_overload_constant_str(lhs) or lhs == bodo.libs.
        str_ext.string_type or lhs == bodo.hiframes.pd_timestamp_ext.
        pd_timestamp_type)
    xiun__ggf = bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs) and (bodo
        .utils.typing.is_overload_constant_str(rhs) or rhs == bodo.libs.
        str_ext.string_type or rhs == bodo.hiframes.pd_timestamp_ext.
        pd_timestamp_type)
    evipv__yoiu = xwol__hwp or xiun__ggf
    wnxm__ykzkb = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs
        ) and lhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
    eqiq__oyi = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs
        ) and rhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
    mth__omzc = wnxm__ykzkb or eqiq__oyi
    return evipv__yoiu or mth__omzc


def cmp_timedeltas(lhs, rhs):
    muytl__dkh = [pd_timedelta_type, bodo.timedelta64ns]
    return lhs in muytl__dkh and rhs in muytl__dkh


def operand_is_index(operand):
    return is_index_type(operand) or isinstance(operand, HeterogeneousIndexType
        )


def helper_time_series_checks(operand):
    rxb__muc = bodo.hiframes.pd_series_ext.is_dt64_series_typ(operand
        ) or bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(operand
        ) or operand in [datetime_timedelta_type, datetime_datetime_type,
        pd_timestamp_type]
    return rxb__muc


def binary_array_cmp(lhs, rhs):
    return lhs == binary_array_type and rhs in [bytes_type, binary_array_type
        ] or lhs in [bytes_type, binary_array_type
        ] and rhs == binary_array_type


def can_cmp_date_datetime(lhs, rhs, op):
    return op in (operator.eq, operator.ne) and (lhs == datetime_date_type and
        rhs == datetime_datetime_type or lhs == datetime_datetime_type and 
        rhs == datetime_date_type)


def time_series_operation(lhs, rhs):
    ibq__yju = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs
        ) and rhs == datetime_timedelta_type
    boepq__klda = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs
        ) and lhs == datetime_timedelta_type
    nih__jupeu = bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs
        ) and helper_time_series_checks(rhs)
    lauxz__xjyjt = bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs
        ) and helper_time_series_checks(lhs)
    return ibq__yju or boepq__klda or nih__jupeu or lauxz__xjyjt


def args_td_and_int_array(lhs, rhs):
    pju__emock = (isinstance(lhs, IntegerArrayType) or isinstance(lhs,
        types.Array) and isinstance(lhs.dtype, types.Integer)) or (isinstance
        (rhs, IntegerArrayType) or isinstance(rhs, types.Array) and
        isinstance(rhs.dtype, types.Integer))
    qgq__bqmn = lhs in [pd_timedelta_type] or rhs in [pd_timedelta_type]
    return pju__emock and qgq__bqmn


def arith_op_supported_by_numba(op, lhs, rhs):
    if op == operator.mul:
        quwi__nuhdy = isinstance(lhs, (types.Integer, types.Float)
            ) and isinstance(rhs, types.NPTimedelta)
        ihic__kaj = isinstance(rhs, (types.Integer, types.Float)
            ) and isinstance(lhs, types.NPTimedelta)
        suv__dcqvt = quwi__nuhdy or ihic__kaj
        taiz__efmz = isinstance(rhs, types.UnicodeType) and isinstance(lhs,
            types.Integer)
        doing__opbe = isinstance(lhs, types.UnicodeType) and isinstance(rhs,
            types.Integer)
        vjb__pjizm = taiz__efmz or doing__opbe
        omg__tpl = isinstance(lhs, types.Integer) and isinstance(rhs, types
            .Integer)
        xfxkr__cli = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        kil__dcu = isinstance(lhs, types.Complex) and isinstance(rhs, types
            .Complex)
        tvcpr__bwe = omg__tpl or xfxkr__cli or kil__dcu
        kii__snrdd = isinstance(lhs, types.List) and isinstance(rhs, types.
            Integer) or isinstance(lhs, types.Integer) and isinstance(rhs,
            types.List)
        tys = types.UnicodeCharSeq, types.CharSeq, types.Bytes
        enmot__upog = isinstance(lhs, tys) or isinstance(rhs, tys)
        wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
        return (suv__dcqvt or vjb__pjizm or tvcpr__bwe or kii__snrdd or
            enmot__upog or wqj__djz)
    if op == operator.pow:
        afav__cibg = isinstance(lhs, types.Integer) and isinstance(rhs, (
            types.IntegerLiteral, types.Integer))
        iesc__htg = isinstance(lhs, types.Float) and isinstance(rhs, (types
            .IntegerLiteral, types.Float, types.Integer) or rhs in types.
            unsigned_domain or rhs in types.signed_domain)
        kil__dcu = isinstance(lhs, types.Complex) and isinstance(rhs, types
            .Complex)
        wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
        return afav__cibg or iesc__htg or kil__dcu or wqj__djz
    if op == operator.floordiv:
        xfxkr__cli = lhs in types.real_domain and rhs in types.real_domain
        omg__tpl = isinstance(lhs, types.Integer) and isinstance(rhs, types
            .Integer)
        ambw__rjz = isinstance(lhs, types.Float) and isinstance(rhs, types.
            Float)
        muytl__dkh = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
            (types.Integer, types.Float, types.NPTimedelta))
        wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
        return xfxkr__cli or omg__tpl or ambw__rjz or muytl__dkh or wqj__djz
    if op == operator.truediv:
        arg__npewc = lhs in machine_ints and rhs in machine_ints
        xfxkr__cli = lhs in types.real_domain and rhs in types.real_domain
        kil__dcu = lhs in types.complex_domain and rhs in types.complex_domain
        omg__tpl = isinstance(lhs, types.Integer) and isinstance(rhs, types
            .Integer)
        ambw__rjz = isinstance(lhs, types.Float) and isinstance(rhs, types.
            Float)
        kun__bywsr = isinstance(lhs, types.Complex) and isinstance(rhs,
            types.Complex)
        muytl__dkh = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
            (types.Integer, types.Float, types.NPTimedelta))
        wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
        return (arg__npewc or xfxkr__cli or kil__dcu or omg__tpl or
            ambw__rjz or kun__bywsr or muytl__dkh or wqj__djz)
    if op == operator.mod:
        arg__npewc = lhs in machine_ints and rhs in machine_ints
        xfxkr__cli = lhs in types.real_domain and rhs in types.real_domain
        omg__tpl = isinstance(lhs, types.Integer) and isinstance(rhs, types
            .Integer)
        ambw__rjz = isinstance(lhs, types.Float) and isinstance(rhs, types.
            Float)
        wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
        return arg__npewc or xfxkr__cli or omg__tpl or ambw__rjz or wqj__djz
    if op == operator.add or op == operator.sub:
        suv__dcqvt = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
            types.NPTimedelta)
        kmwz__ohhja = isinstance(lhs, types.NPDatetime) and isinstance(rhs,
            types.NPDatetime)
        wud__wiaa = isinstance(lhs, types.NPDatetime) and isinstance(rhs,
            types.NPTimedelta)
        akef__jmrvu = isinstance(lhs, types.Set) and isinstance(rhs, types.Set)
        omg__tpl = isinstance(lhs, types.Integer) and isinstance(rhs, types
            .Integer)
        xfxkr__cli = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        kil__dcu = isinstance(lhs, types.Complex) and isinstance(rhs, types
            .Complex)
        tvcpr__bwe = omg__tpl or xfxkr__cli or kil__dcu
        wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
        xfid__nnje = isinstance(lhs, types.BaseTuple) and isinstance(rhs,
            types.BaseTuple)
        kii__snrdd = isinstance(lhs, types.List) and isinstance(rhs, types.List
            )
        leok__gxtx = isinstance(lhs, types.UnicodeCharSeq) and isinstance(rhs,
            types.UnicodeType)
        dxq__coz = isinstance(rhs, types.UnicodeCharSeq) and isinstance(lhs,
            types.UnicodeType)
        xbf__jah = isinstance(lhs, types.UnicodeCharSeq) and isinstance(rhs,
            types.UnicodeCharSeq)
        luups__gyp = isinstance(lhs, (types.CharSeq, types.Bytes)
            ) and isinstance(rhs, (types.CharSeq, types.Bytes))
        wbxw__syb = leok__gxtx or dxq__coz or xbf__jah or luups__gyp
        vjb__pjizm = isinstance(lhs, types.UnicodeType) and isinstance(rhs,
            types.UnicodeType)
        ncw__ghhj = isinstance(lhs, types.UnicodeType) and isinstance(rhs,
            types.UnicodeCharSeq)
        oirls__yllgc = vjb__pjizm or ncw__ghhj
        qeyiv__abq = lhs == types.NPTimedelta and rhs == types.NPDatetime
        vhzj__mpbhn = (xfid__nnje or kii__snrdd or wbxw__syb or
            oirls__yllgc or qeyiv__abq)
        pkt__edsz = op == operator.add and vhzj__mpbhn
        return (suv__dcqvt or kmwz__ohhja or wud__wiaa or akef__jmrvu or
            tvcpr__bwe or wqj__djz or pkt__edsz)


def cmp_op_supported_by_numba(lhs, rhs):
    wqj__djz = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
    kii__snrdd = isinstance(lhs, types.ListType) and isinstance(rhs, types.
        ListType)
    suv__dcqvt = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
        types.NPTimedelta)
    rpbgt__zgj = isinstance(lhs, types.NPDatetime) and isinstance(rhs,
        types.NPDatetime)
    unicode_types = (types.UnicodeType, types.StringLiteral, types.CharSeq,
        types.Bytes, types.UnicodeCharSeq)
    vjb__pjizm = isinstance(lhs, unicode_types) and isinstance(rhs,
        unicode_types)
    xfid__nnje = isinstance(lhs, types.BaseTuple) and isinstance(rhs, types
        .BaseTuple)
    akef__jmrvu = isinstance(lhs, types.Set) and isinstance(rhs, types.Set)
    tvcpr__bwe = isinstance(lhs, types.Number) and isinstance(rhs, types.Number
        )
    gzjcm__zaoa = isinstance(lhs, types.Boolean) and isinstance(rhs, types.
        Boolean)
    pmxjy__hav = isinstance(lhs, types.NoneType) or isinstance(rhs, types.
        NoneType)
    gskjb__iwuca = isinstance(lhs, types.DictType) and isinstance(rhs,
        types.DictType)
    hhtnt__gaxyq = isinstance(lhs, types.EnumMember) and isinstance(rhs,
        types.EnumMember)
    lmtv__vtimu = isinstance(lhs, types.Literal) and isinstance(rhs, types.
        Literal)
    return (kii__snrdd or suv__dcqvt or rpbgt__zgj or vjb__pjizm or
        xfid__nnje or akef__jmrvu or tvcpr__bwe or gzjcm__zaoa or
        pmxjy__hav or gskjb__iwuca or wqj__djz or hhtnt__gaxyq or lmtv__vtimu)


def raise_error_if_not_numba_supported(op, lhs, rhs):
    if arith_op_supported_by_numba(op, lhs, rhs):
        return
    raise BodoError(
        f'{op} operator not supported for data types {lhs} and {rhs}.')


def _install_series_and_or():
    for op in (operator.or_, operator.and_):
        infer_global(op)(SeriesAndOrTyper)
        lower_impl = lower_series_and_or(op)
        lower_builtin(op, SeriesType, SeriesType)(lower_impl)
        lower_builtin(op, SeriesType, types.Any)(lower_impl)
        lower_builtin(op, types.Any, SeriesType)(lower_impl)


_install_series_and_or()


def _install_cmp_ops():
    for op in (operator.lt, operator.eq, operator.ne, operator.ge, operator
        .gt, operator.le):
        infer_global(op)(SeriesCmpOpTemplate)
        lower_impl = series_cmp_op_lower(op)
        lower_builtin(op, SeriesType, SeriesType)(lower_impl)
        lower_builtin(op, SeriesType, types.Any)(lower_impl)
        lower_builtin(op, types.Any, SeriesType)(lower_impl)
        xkzr__mjorn = create_overload_cmp_operator(op)
        overload(op, no_unliteral=True)(xkzr__mjorn)


_install_cmp_ops()


def install_arith_ops():
    for op in (operator.add, operator.sub, operator.mul, operator.truediv,
        operator.floordiv, operator.mod, operator.pow):
        xkzr__mjorn = create_overload_arith_op(op)
        overload(op, no_unliteral=True)(xkzr__mjorn)


install_arith_ops()
