"""Nullable boolean array that stores data in Numpy format (1 byte per value)
but nulls are stored in bit arrays (1 bit per value) similar to Arrow's nulls.
Pandas converts boolean array to object when NAs are introduced.
"""
import operator
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed, lower_constant
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import NativeValue, box, intrinsic, lower_builtin, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, type_callable, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.libs import hstr_ext
from bodo.libs.str_arr_ext import string_array_type
from bodo.utils.typing import is_list_like_index_type
ll.add_symbol('is_bool_array', hstr_ext.is_bool_array)
ll.add_symbol('is_pd_boolean_array', hstr_ext.is_pd_boolean_array)
ll.add_symbol('unbox_bool_array_obj', hstr_ext.unbox_bool_array_obj)
from bodo.utils.indexing import array_getitem_bool_index, array_getitem_int_index, array_getitem_slice_index, array_setitem_bool_index, array_setitem_int_index, array_setitem_slice_index
from bodo.utils.typing import BodoError, is_iterable_type, is_overload_false, is_overload_true, parse_dtype, raise_bodo_error


class BooleanArrayType(types.ArrayCompatible):

    def __init__(self):
        super(BooleanArrayType, self).__init__(name='BooleanArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return types.bool_

    def copy(self):
        return BooleanArrayType()


boolean_array = BooleanArrayType()


@typeof_impl.register(pd.arrays.BooleanArray)
def typeof_boolean_array(val, c):
    return boolean_array


data_type = types.Array(types.bool_, 1, 'C')
nulls_type = types.Array(types.uint8, 1, 'C')


@register_model(BooleanArrayType)
class BooleanArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        lki__xqtrt = [('data', data_type), ('null_bitmap', nulls_type)]
        models.StructModel.__init__(self, dmm, fe_type, lki__xqtrt)


make_attribute_wrapper(BooleanArrayType, 'data', '_data')
make_attribute_wrapper(BooleanArrayType, 'null_bitmap', '_null_bitmap')


class BooleanDtype(types.Number):

    def __init__(self):
        self.dtype = types.bool_
        super(BooleanDtype, self).__init__('BooleanDtype')


boolean_dtype = BooleanDtype()
register_model(BooleanDtype)(models.OpaqueModel)


@box(BooleanDtype)
def box_boolean_dtype(typ, val, c):
    yzan__mfcjf = c.context.insert_const_string(c.builder.module, 'pandas')
    dxkt__oks = c.pyapi.import_module_noblock(yzan__mfcjf)
    smou__sagz = c.pyapi.call_method(dxkt__oks, 'BooleanDtype', ())
    c.pyapi.decref(dxkt__oks)
    return smou__sagz


@unbox(BooleanDtype)
def unbox_boolean_dtype(typ, val, c):
    return NativeValue(c.context.get_dummy_value())


typeof_impl.register(pd.BooleanDtype)(lambda a, b: boolean_dtype)
type_callable(pd.BooleanDtype)(lambda c: lambda : boolean_dtype)
lower_builtin(pd.BooleanDtype)(lambda c, b, s, a: c.get_dummy_value())


@numba.njit
def gen_full_bitmap(n):
    mxot__towpr = n + 7 >> 3
    return np.full(mxot__towpr, 255, np.uint8)


def call_func_in_unbox(func, args, arg_typs, c):
    zaxp__bues = c.context.typing_context.resolve_value_type(func)
    bjfbr__euv = zaxp__bues.get_call_type(c.context.typing_context,
        arg_typs, {})
    rtywd__oymj = c.context.get_function(zaxp__bues, bjfbr__euv)
    rcrz__mbj = c.context.call_conv.get_function_type(bjfbr__euv.
        return_type, bjfbr__euv.args)
    zceb__skft = c.builder.module
    pqwq__jxuqp = lir.Function(zceb__skft, rcrz__mbj, name=zceb__skft.
        get_unique_name('.func_conv'))
    pqwq__jxuqp.linkage = 'internal'
    hkwwu__lddsd = lir.IRBuilder(pqwq__jxuqp.append_basic_block())
    nwsan__ucm = c.context.call_conv.decode_arguments(hkwwu__lddsd,
        bjfbr__euv.args, pqwq__jxuqp)
    razvs__cumfl = rtywd__oymj(hkwwu__lddsd, nwsan__ucm)
    c.context.call_conv.return_value(hkwwu__lddsd, razvs__cumfl)
    xtw__yoyxu, cjf__mvqz = c.context.call_conv.call_function(c.builder,
        pqwq__jxuqp, bjfbr__euv.return_type, bjfbr__euv.args, args)
    return cjf__mvqz


@unbox(BooleanArrayType)
def unbox_bool_array(typ, obj, c):
    pgu__osdl = c.pyapi.call_method(obj, '__len__', ())
    n = c.pyapi.long_as_longlong(pgu__osdl)
    c.pyapi.decref(pgu__osdl)
    rcrz__mbj = lir.FunctionType(lir.IntType(32), [lir.IntType(8).as_pointer()]
        )
    xsj__mesg = cgutils.get_or_insert_function(c.builder.module, rcrz__mbj,
        name='is_bool_array')
    rcrz__mbj = lir.FunctionType(lir.IntType(32), [lir.IntType(8).as_pointer()]
        )
    pqwq__jxuqp = cgutils.get_or_insert_function(c.builder.module,
        rcrz__mbj, name='is_pd_boolean_array')
    smfkh__anh = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    dkbae__gkg = c.builder.call(pqwq__jxuqp, [obj])
    adpg__bwbg = c.builder.icmp_unsigned('!=', dkbae__gkg, dkbae__gkg.type(0))
    with c.builder.if_else(adpg__bwbg) as (pd_then, pd_otherwise):
        with pd_then:
            ksl__gxd = c.pyapi.object_getattr_string(obj, '_data')
            smfkh__anh.data = c.pyapi.to_native_value(types.Array(types.
                bool_, 1, 'C'), ksl__gxd).value
            dhmu__lgb = c.pyapi.object_getattr_string(obj, '_mask')
            gbvw__wpdm = c.pyapi.to_native_value(types.Array(types.bool_, 1,
                'C'), dhmu__lgb).value
            mxot__towpr = c.builder.udiv(c.builder.add(n, lir.Constant(lir.
                IntType(64), 7)), lir.Constant(lir.IntType(64), 8))
            cyyvm__hgak = c.context.make_array(types.Array(types.bool_, 1, 'C')
                )(c.context, c.builder, gbvw__wpdm)
            aytdn__wxrx = bodo.utils.utils._empty_nd_impl(c.context, c.
                builder, types.Array(types.uint8, 1, 'C'), [mxot__towpr])
            rcrz__mbj = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
                as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(64)])
            pqwq__jxuqp = cgutils.get_or_insert_function(c.builder.module,
                rcrz__mbj, name='mask_arr_to_bitmap')
            c.builder.call(pqwq__jxuqp, [aytdn__wxrx.data, cyyvm__hgak.data, n]
                )
            smfkh__anh.null_bitmap = aytdn__wxrx._getvalue()
            c.context.nrt.decref(c.builder, types.Array(types.bool_, 1, 'C'
                ), gbvw__wpdm)
            c.pyapi.decref(ksl__gxd)
            c.pyapi.decref(dhmu__lgb)
        with pd_otherwise:
            njc__gzun = c.builder.call(xsj__mesg, [obj])
            tuufj__yoewy = c.builder.icmp_unsigned('!=', njc__gzun,
                njc__gzun.type(0))
            with c.builder.if_else(tuufj__yoewy) as (then, otherwise):
                with then:
                    smfkh__anh.data = c.pyapi.to_native_value(types.Array(
                        types.bool_, 1, 'C'), obj).value
                    smfkh__anh.null_bitmap = call_func_in_unbox(gen_full_bitmap
                        , (n,), (types.int64,), c)
                with otherwise:
                    smfkh__anh.data = bodo.utils.utils._empty_nd_impl(c.
                        context, c.builder, types.Array(types.bool_, 1, 'C'
                        ), [n])._getvalue()
                    mxot__towpr = c.builder.udiv(c.builder.add(n, lir.
                        Constant(lir.IntType(64), 7)), lir.Constant(lir.
                        IntType(64), 8))
                    smfkh__anh.null_bitmap = bodo.utils.utils._empty_nd_impl(c
                        .context, c.builder, types.Array(types.uint8, 1,
                        'C'), [mxot__towpr])._getvalue()
                    xaig__voty = c.context.make_array(types.Array(types.
                        bool_, 1, 'C'))(c.context, c.builder, smfkh__anh.data
                        ).data
                    vxh__iysn = c.context.make_array(types.Array(types.
                        uint8, 1, 'C'))(c.context, c.builder, smfkh__anh.
                        null_bitmap).data
                    rcrz__mbj = lir.FunctionType(lir.VoidType(), [lir.
                        IntType(8).as_pointer(), lir.IntType(8).as_pointer(
                        ), lir.IntType(8).as_pointer(), lir.IntType(64)])
                    pqwq__jxuqp = cgutils.get_or_insert_function(c.builder.
                        module, rcrz__mbj, name='unbox_bool_array_obj')
                    c.builder.call(pqwq__jxuqp, [obj, xaig__voty, vxh__iysn, n]
                        )
    return NativeValue(smfkh__anh._getvalue())


@box(BooleanArrayType)
def box_bool_arr(typ, val, c):
    smfkh__anh = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    data = c.pyapi.from_native_value(types.Array(typ.dtype, 1, 'C'),
        smfkh__anh.data, c.env_manager)
    towun__dmv = c.context.make_array(types.Array(types.uint8, 1, 'C'))(c.
        context, c.builder, smfkh__anh.null_bitmap).data
    pgu__osdl = c.pyapi.call_method(data, '__len__', ())
    n = c.pyapi.long_as_longlong(pgu__osdl)
    yzan__mfcjf = c.context.insert_const_string(c.builder.module, 'numpy')
    ipnhl__ureal = c.pyapi.import_module_noblock(yzan__mfcjf)
    wambw__ygfnc = c.pyapi.object_getattr_string(ipnhl__ureal, 'bool_')
    gbvw__wpdm = c.pyapi.call_method(ipnhl__ureal, 'empty', (pgu__osdl,
        wambw__ygfnc))
    lefds__ttcki = c.pyapi.object_getattr_string(gbvw__wpdm, 'ctypes')
    txqyq__gwn = c.pyapi.object_getattr_string(lefds__ttcki, 'data')
    dhqyc__cpkf = c.builder.inttoptr(c.pyapi.long_as_longlong(txqyq__gwn),
        lir.IntType(8).as_pointer())
    with cgutils.for_range(c.builder, n) as loop:
        aaac__jenax = loop.index
        lqvjl__qdmjs = c.builder.lshr(aaac__jenax, lir.Constant(lir.IntType
            (64), 3))
        ahkod__hftii = c.builder.load(cgutils.gep(c.builder, towun__dmv,
            lqvjl__qdmjs))
        qljtb__uthp = c.builder.trunc(c.builder.and_(aaac__jenax, lir.
            Constant(lir.IntType(64), 7)), lir.IntType(8))
        val = c.builder.and_(c.builder.lshr(ahkod__hftii, qljtb__uthp), lir
            .Constant(lir.IntType(8), 1))
        val = c.builder.xor(val, lir.Constant(lir.IntType(8), 1))
        izkhz__lazag = cgutils.gep(c.builder, dhqyc__cpkf, aaac__jenax)
        c.builder.store(val, izkhz__lazag)
    c.context.nrt.decref(c.builder, types.Array(types.uint8, 1, 'C'),
        smfkh__anh.null_bitmap)
    yzan__mfcjf = c.context.insert_const_string(c.builder.module, 'pandas')
    dxkt__oks = c.pyapi.import_module_noblock(yzan__mfcjf)
    xolre__gbhy = c.pyapi.object_getattr_string(dxkt__oks, 'arrays')
    smou__sagz = c.pyapi.call_method(xolre__gbhy, 'BooleanArray', (data,
        gbvw__wpdm))
    c.pyapi.decref(dxkt__oks)
    c.pyapi.decref(pgu__osdl)
    c.pyapi.decref(ipnhl__ureal)
    c.pyapi.decref(wambw__ygfnc)
    c.pyapi.decref(lefds__ttcki)
    c.pyapi.decref(txqyq__gwn)
    c.pyapi.decref(xolre__gbhy)
    c.pyapi.decref(data)
    c.pyapi.decref(gbvw__wpdm)
    return smou__sagz


@lower_constant(BooleanArrayType)
def lower_constant_bool_arr(context, builder, typ, pyval):
    n = len(pyval)
    zssq__dfflt = np.empty(n, np.bool_)
    vxoq__ppiat = np.empty(n + 7 >> 3, np.uint8)
    for aaac__jenax, s in enumerate(pyval):
        uzv__qlw = pd.isna(s)
        bodo.libs.int_arr_ext.set_bit_to_arr(vxoq__ppiat, aaac__jenax, int(
            not uzv__qlw))
        if not uzv__qlw:
            zssq__dfflt[aaac__jenax] = s
    sbaq__dwbp = context.get_constant_generic(builder, data_type, zssq__dfflt)
    arl__zrwj = context.get_constant_generic(builder, nulls_type, vxoq__ppiat)
    return lir.Constant.literal_struct([sbaq__dwbp, arl__zrwj])


def lower_init_bool_array(context, builder, signature, args):
    ntuuv__hqh, hirol__pokhl = args
    smfkh__anh = cgutils.create_struct_proxy(signature.return_type)(context,
        builder)
    smfkh__anh.data = ntuuv__hqh
    smfkh__anh.null_bitmap = hirol__pokhl
    context.nrt.incref(builder, signature.args[0], ntuuv__hqh)
    context.nrt.incref(builder, signature.args[1], hirol__pokhl)
    return smfkh__anh._getvalue()


@intrinsic
def init_bool_array(typingctx, data, null_bitmap=None):
    assert data == types.Array(types.bool_, 1, 'C')
    assert null_bitmap == types.Array(types.uint8, 1, 'C')
    sig = boolean_array(data, null_bitmap)
    return sig, lower_init_bool_array


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def get_bool_arr_data(A):
    return lambda A: A._data


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def get_bool_arr_bitmap(A):
    return lambda A: A._null_bitmap


def get_bool_arr_data_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 1 and not kws
    jyrx__lyx = args[0]
    if equiv_set.has_shape(jyrx__lyx):
        return ArrayAnalysis.AnalyzeResult(shape=jyrx__lyx, pre=[])
    return None


ArrayAnalysis._analyze_op_call_bodo_libs_bool_arr_ext_get_bool_arr_data = (
    get_bool_arr_data_equiv)


def init_bool_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    jyrx__lyx = args[0]
    if equiv_set.has_shape(jyrx__lyx):
        return ArrayAnalysis.AnalyzeResult(shape=jyrx__lyx, pre=[])
    return None


ArrayAnalysis._analyze_op_call_bodo_libs_bool_arr_ext_init_bool_array = (
    init_bool_array_equiv)


def alias_ext_dummy_func(lhs_name, args, alias_map, arg_aliases):
    assert len(args) >= 1
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)


def alias_ext_init_bool_array(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 2
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)
    numba.core.ir_utils._add_alias(lhs_name, args[1].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_bool_array',
    'bodo.libs.bool_arr_ext'] = alias_ext_init_bool_array
numba.core.ir_utils.alias_func_extensions['get_bool_arr_data',
    'bodo.libs.bool_arr_ext'] = alias_ext_dummy_func
numba.core.ir_utils.alias_func_extensions['get_bool_arr_bitmap',
    'bodo.libs.bool_arr_ext'] = alias_ext_dummy_func


@numba.njit(no_cpython_wrapper=True)
def alloc_bool_array(n):
    zssq__dfflt = np.empty(n, dtype=np.bool_)
    wkza__otf = np.empty(n + 7 >> 3, dtype=np.uint8)
    return init_bool_array(zssq__dfflt, wkza__otf)


def alloc_bool_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 1 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


ArrayAnalysis._analyze_op_call_bodo_libs_bool_arr_ext_alloc_bool_array = (
    alloc_bool_array_equiv)


@overload(operator.getitem, no_unliteral=True)
def bool_arr_getitem(A, ind):
    if A != boolean_array:
        return
    if isinstance(types.unliteral(ind), types.Integer):
        return lambda A, ind: A._data[ind]
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:

        def impl_bool(A, ind):
            ewrrb__eueqm, vrwk__mvl = array_getitem_bool_index(A, ind)
            return init_bool_array(ewrrb__eueqm, vrwk__mvl)
        return impl_bool
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):

        def impl(A, ind):
            ewrrb__eueqm, vrwk__mvl = array_getitem_int_index(A, ind)
            return init_bool_array(ewrrb__eueqm, vrwk__mvl)
        return impl
    if isinstance(ind, types.SliceType):

        def impl_slice(A, ind):
            ewrrb__eueqm, vrwk__mvl = array_getitem_slice_index(A, ind)
            return init_bool_array(ewrrb__eueqm, vrwk__mvl)
        return impl_slice
    raise BodoError(
        f'getitem for BooleanArray with indexing type {ind} not supported.')


@overload(operator.setitem, no_unliteral=True)
def bool_arr_setitem(A, idx, val):
    if A != boolean_array:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    pfa__lpv = (
        f"setitem for BooleanArray with indexing type {idx} received an incorrect 'value' type {val}."
        )
    if isinstance(idx, types.Integer):
        if types.unliteral(val) == types.bool_:

            def impl_scalar(A, idx, val):
                A._data[idx] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx, 1)
            return impl_scalar
        else:
            raise BodoError(pfa__lpv)
    if not (is_iterable_type(val) and val.dtype == types.bool_ or types.
        unliteral(val) == types.bool_):
        raise BodoError(pfa__lpv)
    if is_list_like_index_type(idx) and isinstance(idx.dtype, types.Integer):

        def impl_arr_ind_mask(A, idx, val):
            array_setitem_int_index(A, idx, val)
        return impl_arr_ind_mask
    if is_list_like_index_type(idx) and idx.dtype == types.bool_:

        def impl_bool_ind_mask(A, idx, val):
            array_setitem_bool_index(A, idx, val)
        return impl_bool_ind_mask
    if isinstance(idx, types.SliceType):

        def impl_slice_mask(A, idx, val):
            array_setitem_slice_index(A, idx, val)
        return impl_slice_mask
    raise BodoError(
        f'setitem for BooleanArray with indexing type {idx} not supported.')


@overload(len, no_unliteral=True)
def overload_bool_arr_len(A):
    if A == boolean_array:
        return lambda A: len(A._data)


@overload_attribute(BooleanArrayType, 'shape')
def overload_bool_arr_shape(A):
    return lambda A: (len(A._data),)


@overload_attribute(BooleanArrayType, 'dtype')
def overload_bool_arr_dtype(A):
    return lambda A: pd.BooleanDtype()


@overload_attribute(BooleanArrayType, 'ndim')
def overload_bool_arr_ndim(A):
    return lambda A: 1


@overload_attribute(BooleanArrayType, 'nbytes')
def bool_arr_nbytes_overload(A):
    return lambda A: A._data.nbytes + A._null_bitmap.nbytes


@overload_method(BooleanArrayType, 'copy', no_unliteral=True)
def overload_bool_arr_copy(A):
    return lambda A: bodo.libs.bool_arr_ext.init_bool_array(bodo.libs.
        bool_arr_ext.get_bool_arr_data(A).copy(), bodo.libs.bool_arr_ext.
        get_bool_arr_bitmap(A).copy())


@overload_method(BooleanArrayType, 'sum', no_unliteral=True, inline='always')
def overload_bool_sum(A):

    def impl(A):
        numba.parfors.parfor.init_prange()
        s = 0
        for aaac__jenax in numba.parfors.parfor.internal_prange(len(A)):
            val = 0
            if not bodo.libs.array_kernels.isna(A, aaac__jenax):
                val = A[aaac__jenax]
            s += val
        return s
    return impl


@overload_method(BooleanArrayType, 'astype', no_unliteral=True)
def overload_bool_arr_astype(A, dtype, copy=True):
    if dtype == types.unicode_type:
        raise_bodo_error(
            "BooleanArray.astype(): 'dtype' when passed as string must be a constant value"
            )
    if dtype == types.bool_:
        if is_overload_false(copy):
            return lambda A, dtype, copy=True: A
        elif is_overload_true(copy):
            return lambda A, dtype, copy=True: A.copy()
        else:

            def impl(A, dtype, copy=True):
                if copy:
                    return A.copy()
                else:
                    return A
            return impl
    nb_dtype = parse_dtype(dtype, 'BooleanArray.astype')
    if isinstance(nb_dtype, types.Float):

        def impl_float(A, dtype, copy=True):
            data = bodo.libs.bool_arr_ext.get_bool_arr_data(A)
            n = len(data)
            kebd__odlow = np.empty(n, nb_dtype)
            for aaac__jenax in numba.parfors.parfor.internal_prange(n):
                kebd__odlow[aaac__jenax] = data[aaac__jenax]
                if bodo.libs.array_kernels.isna(A, aaac__jenax):
                    kebd__odlow[aaac__jenax] = np.nan
            return kebd__odlow
        return impl_float
    return (lambda A, dtype, copy=True: bodo.libs.bool_arr_ext.
        get_bool_arr_data(A).astype(nb_dtype))


@overload(str, no_unliteral=True)
def overload_str_bool(val):
    if val == types.bool_:

        def impl(val):
            if val:
                return 'True'
            return 'False'
        return impl


ufunc_aliases = {'equal': 'eq', 'not_equal': 'ne', 'less': 'lt',
    'less_equal': 'le', 'greater': 'gt', 'greater_equal': 'ge'}


def create_op_overload(op, n_inputs):
    qefc__wcp = op.__name__
    qefc__wcp = ufunc_aliases.get(qefc__wcp, qefc__wcp)
    if n_inputs == 1:

        def overload_bool_arr_op_nin_1(A):
            if isinstance(A, BooleanArrayType):
                return bodo.libs.int_arr_ext.get_nullable_array_unary_impl(op,
                    A)
        return overload_bool_arr_op_nin_1
    elif n_inputs == 2:

        def overload_bool_arr_op_nin_2(lhs, rhs):
            if lhs == boolean_array or rhs == boolean_array:
                return bodo.libs.int_arr_ext.get_nullable_array_binary_impl(op,
                    lhs, rhs)
        return overload_bool_arr_op_nin_2
    else:
        raise RuntimeError(
            "Don't know how to register ufuncs from ufunc_db with arity > 2")


def _install_np_ufuncs():
    import numba.np.ufunc_db
    for sakmr__hrdvo in numba.np.ufunc_db.get_ufuncs():
        tby__bsbxx = create_op_overload(sakmr__hrdvo, sakmr__hrdvo.nin)
        overload(sakmr__hrdvo, no_unliteral=True)(tby__bsbxx)


_install_np_ufuncs()
skips = [operator.lt, operator.le, operator.eq, operator.ne, operator.gt,
    operator.ge, operator.add, operator.sub, operator.mul, operator.truediv,
    operator.floordiv, operator.pow, operator.mod, operator.or_, operator.and_]


def _install_binary_ops():
    for op in numba.core.typing.npydecl.NumpyRulesArrayOperator._op_map.keys():
        if op in skips:
            continue
        tby__bsbxx = create_op_overload(op, 2)
        overload(op, no_unliteral=True)(tby__bsbxx)


_install_binary_ops()


def _install_inplace_binary_ops():
    for op in numba.core.typing.npydecl.NumpyRulesInplaceArrayOperator._op_map.keys(
        ):
        tby__bsbxx = create_op_overload(op, 2)
        overload(op, no_unliteral=True)(tby__bsbxx)


_install_inplace_binary_ops()


def _install_unary_ops():
    for op in (operator.neg, operator.invert, operator.pos):
        tby__bsbxx = create_op_overload(op, 1)
        overload(op, no_unliteral=True)(tby__bsbxx)


_install_unary_ops()


@overload_method(BooleanArrayType, 'unique', no_unliteral=True)
def overload_unique(A):

    def impl_bool_arr(A):
        data = []
        qljtb__uthp = []
        asjcp__auh = False
        jfo__rmjmx = False
        ofq__ajxop = False
        for aaac__jenax in range(len(A)):
            if bodo.libs.array_kernels.isna(A, aaac__jenax):
                if not asjcp__auh:
                    data.append(False)
                    qljtb__uthp.append(False)
                    asjcp__auh = True
                continue
            val = A[aaac__jenax]
            if val and not jfo__rmjmx:
                data.append(True)
                qljtb__uthp.append(True)
                jfo__rmjmx = True
            if not val and not ofq__ajxop:
                data.append(False)
                qljtb__uthp.append(True)
                ofq__ajxop = True
            if asjcp__auh and jfo__rmjmx and ofq__ajxop:
                break
        ewrrb__eueqm = np.array(data)
        n = len(ewrrb__eueqm)
        mxot__towpr = 1
        vrwk__mvl = np.empty(mxot__towpr, np.uint8)
        for bqkpw__kwu in range(n):
            bodo.libs.int_arr_ext.set_bit_to_arr(vrwk__mvl, bqkpw__kwu,
                qljtb__uthp[bqkpw__kwu])
        return init_bool_array(ewrrb__eueqm, vrwk__mvl)
    return impl_bool_arr


@overload(operator.getitem, no_unliteral=True)
def bool_arr_ind_getitem(A, ind):
    if ind == boolean_array and (isinstance(A, (types.Array, bodo.libs.
        int_arr_ext.IntegerArrayType)) or isinstance(A, bodo.libs.
        struct_arr_ext.StructArrayType) or isinstance(A, bodo.libs.
        array_item_arr_ext.ArrayItemArrayType) or isinstance(A, bodo.libs.
        map_arr_ext.MapArrayType) or A in (string_array_type, bodo.hiframes
        .split_impl.string_array_split_view_type, boolean_array)):
        return lambda A, ind: A[ind._data]


@lower_cast(types.Array(types.bool_, 1, 'C'), boolean_array)
def cast_np_bool_arr_to_bool_arr(context, builder, fromty, toty, val):
    func = lambda A: bodo.libs.bool_arr_ext.init_bool_array(A, np.full(len(
        A) + 7 >> 3, 255, np.uint8))
    smou__sagz = context.compile_internal(builder, func, toty(fromty), [val])
    return impl_ret_borrowed(context, builder, toty, smou__sagz)


@overload(operator.setitem, no_unliteral=True)
def overload_np_array_setitem_bool_arr(A, idx, val):
    if isinstance(A, types.Array) and idx == boolean_array:

        def impl(A, idx, val):
            A[idx._data] = val
        return impl


def create_nullable_logical_op_overload(op):
    wde__qmjiv = op == operator.or_

    def bool_array_impl(val1, val2):
        if not is_valid_boolean_array_logical_op(val1, val2):
            return
        spg__tfih = bodo.utils.utils.is_array_typ(val1, False)
        hiqq__mkvx = bodo.utils.utils.is_array_typ(val2, False)
        jqzi__zgny = 'val1' if spg__tfih else 'val2'
        ruapf__phoy = 'def impl(val1, val2):\n'
        ruapf__phoy += f'  n = len({jqzi__zgny})\n'
        ruapf__phoy += (
            '  out_arr = bodo.utils.utils.alloc_type(n, bodo.boolean_array, (-1,))\n'
            )
        ruapf__phoy += '  for i in numba.parfors.parfor.internal_prange(n):\n'
        if spg__tfih:
            null1 = 'bodo.libs.array_kernels.isna(val1, i)\n'
            roh__ahj = 'val1[i]'
        else:
            null1 = 'False\n'
            roh__ahj = 'val1'
        if hiqq__mkvx:
            null2 = 'bodo.libs.array_kernels.isna(val2, i)\n'
            lgt__mjx = 'val2[i]'
        else:
            null2 = 'False\n'
            lgt__mjx = 'val2'
        if wde__qmjiv:
            ruapf__phoy += f"""    result, isna_val = compute_or_body({null1}, {null2}, {roh__ahj}, {lgt__mjx})
"""
        else:
            ruapf__phoy += f"""    result, isna_val = compute_and_body({null1}, {null2}, {roh__ahj}, {lgt__mjx})
"""
        ruapf__phoy += '    out_arr[i] = result\n'
        ruapf__phoy += '    if isna_val:\n'
        ruapf__phoy += '      bodo.libs.array_kernels.setna(out_arr, i)\n'
        ruapf__phoy += '      continue\n'
        ruapf__phoy += '  return out_arr\n'
        ugh__igid = {}
        exec(ruapf__phoy, {'bodo': bodo, 'numba': numba, 'compute_and_body':
            compute_and_body, 'compute_or_body': compute_or_body}, ugh__igid)
        impl = ugh__igid['impl']
        return impl
    return bool_array_impl


def compute_or_body(null1, null2, val1, val2):
    pass


@overload(compute_or_body)
def overload_compute_or_body(null1, null2, val1, val2):

    def impl(null1, null2, val1, val2):
        if null1 and null2:
            return False, True
        elif null1:
            return val2, val2 == False
        elif null2:
            return val1, val1 == False
        else:
            return val1 | val2, False
    return impl


def compute_and_body(null1, null2, val1, val2):
    pass


@overload(compute_and_body)
def overload_compute_and_body(null1, null2, val1, val2):

    def impl(null1, null2, val1, val2):
        if null1 and null2:
            return False, True
        elif null1:
            return val2, val2 == True
        elif null2:
            return val1, val1 == True
        else:
            return val1 & val2, False
    return impl


def create_boolean_array_logical_lower_impl(op):

    def logical_lower_impl(context, builder, sig, args):
        impl = create_nullable_logical_op_overload(op)(*sig.args)
        return context.compile_internal(builder, impl, sig, args)
    return logical_lower_impl


class BooleanArrayLogicalOperatorTemplate(AbstractTemplate):

    def generic(self, args, kws):
        assert len(args) == 2
        assert not kws
        if not is_valid_boolean_array_logical_op(args[0], args[1]):
            return
        kkq__qxhh = boolean_array
        return kkq__qxhh(*args)


def is_valid_boolean_array_logical_op(typ1, typ2):
    jvqhh__rrqke = (typ1 == bodo.boolean_array or typ2 == bodo.boolean_array
        ) and (bodo.utils.utils.is_array_typ(typ1, False) and typ1.dtype ==
        types.bool_ or typ1 == types.bool_) and (bodo.utils.utils.
        is_array_typ(typ2, False) and typ2.dtype == types.bool_ or typ2 ==
        types.bool_)
    return jvqhh__rrqke


def _install_nullable_logical_lowering():
    for op in (operator.and_, operator.or_):
        vxoy__sul = create_boolean_array_logical_lower_impl(op)
        infer_global(op)(BooleanArrayLogicalOperatorTemplate)
        for typ1, typ2 in [(boolean_array, boolean_array), (boolean_array,
            types.bool_), (boolean_array, types.Array(types.bool_, 1, 'C'))]:
            lower_builtin(op, typ1, typ2)(vxoy__sul)
            if typ1 != typ2:
                lower_builtin(op, typ2, typ1)(vxoy__sul)


_install_nullable_logical_lowering()
