"""CSR Matrix data type implementation for scipy.sparse.csr_matrix
"""
import operator
import numba
import numpy as np
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, typeof_impl, unbox
import bodo
from bodo.utils.typing import BodoError


class CSRMatrixType(types.ArrayCompatible):
    ndim = 2

    def __init__(self, dtype, idx_dtype):
        self.dtype = dtype
        self.idx_dtype = idx_dtype
        super(CSRMatrixType, self).__init__(name=
            f'CSRMatrixType({dtype}, {idx_dtype})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 2, 'C')

    def copy(self):
        return CSRMatrixType(self.dtype, self.idx_dtype)


@register_model(CSRMatrixType)
class CSRMatrixModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        jas__aubof = [('data', types.Array(fe_type.dtype, 1, 'C')), (
            'indices', types.Array(fe_type.idx_dtype, 1, 'C')), ('indptr',
            types.Array(fe_type.idx_dtype, 1, 'C')), ('shape', types.
            UniTuple(types.int64, 2))]
        models.StructModel.__init__(self, dmm, fe_type, jas__aubof)


make_attribute_wrapper(CSRMatrixType, 'data', 'data')
make_attribute_wrapper(CSRMatrixType, 'indices', 'indices')
make_attribute_wrapper(CSRMatrixType, 'indptr', 'indptr')
make_attribute_wrapper(CSRMatrixType, 'shape', 'shape')


@intrinsic
def init_csr_matrix(typingctx, data_t, indices_t, indptr_t, shape_t=None):
    assert isinstance(data_t, types.Array)
    assert isinstance(indices_t, types.Array) and isinstance(indices_t.
        dtype, types.Integer)
    assert indices_t == indptr_t

    def codegen(context, builder, signature, args):
        kvvvn__ppqw, erm__gjawa, kddz__yfdc, fmh__vzlns = args
        xkvrv__iuv = cgutils.create_struct_proxy(signature.return_type)(context
            , builder)
        xkvrv__iuv.data = kvvvn__ppqw
        xkvrv__iuv.indices = erm__gjawa
        xkvrv__iuv.indptr = kddz__yfdc
        xkvrv__iuv.shape = fmh__vzlns
        context.nrt.incref(builder, signature.args[0], kvvvn__ppqw)
        context.nrt.incref(builder, signature.args[1], erm__gjawa)
        context.nrt.incref(builder, signature.args[2], kddz__yfdc)
        return xkvrv__iuv._getvalue()
    xup__vrt = CSRMatrixType(data_t.dtype, indices_t.dtype)
    wzt__efr = xup__vrt(data_t, indices_t, indptr_t, types.UniTuple(types.
        int64, 2))
    return wzt__efr, codegen


if bodo.utils.utils.has_scipy():
    import scipy.sparse

    @typeof_impl.register(scipy.sparse.csr_matrix)
    def _typeof_csr_matrix(val, c):
        dtype = numba.from_dtype(val.dtype)
        idx_dtype = numba.from_dtype(val.indices.dtype)
        return CSRMatrixType(dtype, idx_dtype)


@unbox(CSRMatrixType)
def unbox_csr_matrix(typ, val, c):
    xkvrv__iuv = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    hod__psxsj = c.pyapi.object_getattr_string(val, 'data')
    uxqv__pjg = c.pyapi.object_getattr_string(val, 'indices')
    fatfl__ategj = c.pyapi.object_getattr_string(val, 'indptr')
    bklf__gjqr = c.pyapi.object_getattr_string(val, 'shape')
    xkvrv__iuv.data = c.pyapi.to_native_value(types.Array(typ.dtype, 1, 'C'
        ), hod__psxsj).value
    xkvrv__iuv.indices = c.pyapi.to_native_value(types.Array(typ.idx_dtype,
        1, 'C'), uxqv__pjg).value
    xkvrv__iuv.indptr = c.pyapi.to_native_value(types.Array(typ.idx_dtype, 
        1, 'C'), fatfl__ategj).value
    xkvrv__iuv.shape = c.pyapi.to_native_value(types.UniTuple(types.int64, 
        2), bklf__gjqr).value
    c.pyapi.decref(hod__psxsj)
    c.pyapi.decref(uxqv__pjg)
    c.pyapi.decref(fatfl__ategj)
    c.pyapi.decref(bklf__gjqr)
    geliz__ryj = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(xkvrv__iuv._getvalue(), is_error=geliz__ryj)


@box(CSRMatrixType)
def box_csr_matrix(typ, val, c):
    jsreb__tve = c.context.insert_const_string(c.builder.module, 'scipy.sparse'
        )
    szmr__vhfsm = c.pyapi.import_module_noblock(jsreb__tve)
    xkvrv__iuv = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, types.Array(typ.dtype, 1, 'C'),
        xkvrv__iuv.data)
    hod__psxsj = c.pyapi.from_native_value(types.Array(typ.dtype, 1, 'C'),
        xkvrv__iuv.data, c.env_manager)
    c.context.nrt.incref(c.builder, types.Array(typ.idx_dtype, 1, 'C'),
        xkvrv__iuv.indices)
    uxqv__pjg = c.pyapi.from_native_value(types.Array(typ.idx_dtype, 1, 'C'
        ), xkvrv__iuv.indices, c.env_manager)
    c.context.nrt.incref(c.builder, types.Array(typ.idx_dtype, 1, 'C'),
        xkvrv__iuv.indptr)
    fatfl__ategj = c.pyapi.from_native_value(types.Array(typ.idx_dtype, 1,
        'C'), xkvrv__iuv.indptr, c.env_manager)
    bklf__gjqr = c.pyapi.from_native_value(types.UniTuple(types.int64, 2),
        xkvrv__iuv.shape, c.env_manager)
    zjt__gxcc = c.pyapi.tuple_pack([hod__psxsj, uxqv__pjg, fatfl__ategj])
    kdfe__qbybe = c.pyapi.call_method(szmr__vhfsm, 'csr_matrix', (zjt__gxcc,
        bklf__gjqr))
    c.pyapi.decref(zjt__gxcc)
    c.pyapi.decref(hod__psxsj)
    c.pyapi.decref(uxqv__pjg)
    c.pyapi.decref(fatfl__ategj)
    c.pyapi.decref(bklf__gjqr)
    c.pyapi.decref(szmr__vhfsm)
    c.context.nrt.decref(c.builder, typ, val)
    return kdfe__qbybe


@overload(len, no_unliteral=True)
def overload_csr_matrix_len(A):
    if isinstance(A, CSRMatrixType):
        return lambda A: A.shape[0]


@overload_attribute(CSRMatrixType, 'ndim')
def overload_csr_matrix_ndim(A):
    return lambda A: 2


@overload_method(CSRMatrixType, 'copy', no_unliteral=True)
def overload_csr_matrix_copy(A):

    def copy_impl(A):
        return init_csr_matrix(A.data.copy(), A.indices.copy(), A.indptr.
            copy(), A.shape)
    return copy_impl


@overload(operator.getitem, no_unliteral=True)
def csr_matrix_getitem(A, idx):
    if not isinstance(A, CSRMatrixType):
        return
    bezdx__dgqgt = A.dtype
    thzl__ppw = A.idx_dtype
    if isinstance(idx, types.BaseTuple) and len(idx) == 2 and isinstance(idx
        [0], types.SliceType) and isinstance(idx[1], types.SliceType):

        def impl(A, idx):
            afo__kcv, dnulz__mjhxv = A.shape
            nclf__zthoc = numba.cpython.unicode._normalize_slice(idx[0],
                afo__kcv)
            hdazl__ujfpl = numba.cpython.unicode._normalize_slice(idx[1],
                dnulz__mjhxv)
            if nclf__zthoc.step != 1 or hdazl__ujfpl.step != 1:
                raise ValueError(
                    'CSR matrix slice getitem only supports step=1 currently')
            nlqm__wiawq = nclf__zthoc.start
            esy__xot = nclf__zthoc.stop
            bvonm__uok = hdazl__ujfpl.start
            zal__frwa = hdazl__ujfpl.stop
            ybxat__mcku = A.indptr
            lplui__ajv = A.indices
            kjc__oizu = A.data
            xloup__zgw = esy__xot - nlqm__wiawq
            apu__jfay = zal__frwa - bvonm__uok
            gon__hqodd = 0
            frjv__rnv = 0
            for saqp__zoip in range(xloup__zgw):
                sgz__kfzgv = ybxat__mcku[nlqm__wiawq + saqp__zoip]
                gvs__ecawq = ybxat__mcku[nlqm__wiawq + saqp__zoip + 1]
                for qycua__oid in range(sgz__kfzgv, gvs__ecawq):
                    if lplui__ajv[qycua__oid] >= bvonm__uok and lplui__ajv[
                        qycua__oid] < zal__frwa:
                        gon__hqodd += 1
            cidt__zgnyi = np.empty(xloup__zgw + 1, thzl__ppw)
            rva__kpcq = np.empty(gon__hqodd, thzl__ppw)
            ifxe__eapgw = np.empty(gon__hqodd, bezdx__dgqgt)
            cidt__zgnyi[0] = 0
            for saqp__zoip in range(xloup__zgw):
                sgz__kfzgv = ybxat__mcku[nlqm__wiawq + saqp__zoip]
                gvs__ecawq = ybxat__mcku[nlqm__wiawq + saqp__zoip + 1]
                for qycua__oid in range(sgz__kfzgv, gvs__ecawq):
                    if lplui__ajv[qycua__oid] >= bvonm__uok and lplui__ajv[
                        qycua__oid] < zal__frwa:
                        rva__kpcq[frjv__rnv] = lplui__ajv[qycua__oid
                            ] - bvonm__uok
                        ifxe__eapgw[frjv__rnv] = kjc__oizu[qycua__oid]
                        frjv__rnv += 1
                cidt__zgnyi[saqp__zoip + 1] = frjv__rnv
            return init_csr_matrix(ifxe__eapgw, rva__kpcq, cidt__zgnyi, (
                xloup__zgw, apu__jfay))
        return impl
    raise BodoError(
        f'getitem for CSR matrix with index type {idx} not supported yet.')
