"""
Array of intervals corresponding to IntervalArray of Pandas.
Used for IntervalIndex, which is necessary for Series.value_counts() with 'bins'
argument.
"""
import numba
import pandas as pd
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo


class IntervalType(types.Type):

    def __init__(self):
        super(IntervalType, self).__init__('IntervalType()')


class IntervalArrayType(types.ArrayCompatible):

    def __init__(self, arr_type):
        self.arr_type = arr_type
        self.dtype = IntervalType()
        super(IntervalArrayType, self).__init__(name=
            f'IntervalArrayType({arr_type})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    def copy(self):
        return IntervalArrayType(self.arr_type)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(IntervalArrayType)
class IntervalArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        ncr__gre = [('left', fe_type.arr_type), ('right', fe_type.arr_type)]
        models.StructModel.__init__(self, dmm, fe_type, ncr__gre)


make_attribute_wrapper(IntervalArrayType, 'left', '_left')
make_attribute_wrapper(IntervalArrayType, 'right', '_right')


@typeof_impl.register(pd.arrays.IntervalArray)
def typeof_interval_array(val, c):
    arr_type = bodo.typeof(val._left)
    return IntervalArrayType(arr_type)


@intrinsic
def init_interval_array(typingctx, left, right=None):
    assert left == right, 'Interval left/right array types should be the same'

    def codegen(context, builder, signature, args):
        papk__fppqy, jct__hzaly = args
        itqle__ownol = cgutils.create_struct_proxy(signature.return_type)(
            context, builder)
        itqle__ownol.left = papk__fppqy
        itqle__ownol.right = jct__hzaly
        context.nrt.incref(builder, signature.args[0], papk__fppqy)
        context.nrt.incref(builder, signature.args[1], jct__hzaly)
        return itqle__ownol._getvalue()
    uekh__ysebu = IntervalArrayType(left)
    dyvx__uru = uekh__ysebu(left, right)
    return dyvx__uru, codegen


def init_interval_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    ozrw__ssxw = []
    for xqjoj__tsdhv in args:
        mkqo__vcif = equiv_set.get_shape(xqjoj__tsdhv)
        if mkqo__vcif is not None:
            ozrw__ssxw.append(mkqo__vcif[0])
    if len(ozrw__ssxw) > 1:
        equiv_set.insert_equiv(*ozrw__ssxw)
    left = args[0]
    if equiv_set.has_shape(left):
        return ArrayAnalysis.AnalyzeResult(shape=left, pre=[])
    return None


(ArrayAnalysis._analyze_op_call_bodo_libs_interval_arr_ext_init_interval_array
    ) = init_interval_array_equiv


def alias_ext_init_interval_array(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 2
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)
    numba.core.ir_utils._add_alias(lhs_name, args[1].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_interval_array',
    'bodo.libs.int_arr_ext'] = alias_ext_init_interval_array


@box(IntervalArrayType)
def box_interval_arr(typ, val, c):
    itqle__ownol = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, typ.arr_type, itqle__ownol.left)
    xfmv__pkt = c.pyapi.from_native_value(typ.arr_type, itqle__ownol.left,
        c.env_manager)
    c.context.nrt.incref(c.builder, typ.arr_type, itqle__ownol.right)
    iyw__xfen = c.pyapi.from_native_value(typ.arr_type, itqle__ownol.right,
        c.env_manager)
    pttv__gnjx = c.context.insert_const_string(c.builder.module, 'pandas')
    jnuuh__yhpy = c.pyapi.import_module_noblock(pttv__gnjx)
    bspdh__jhdfa = c.pyapi.object_getattr_string(jnuuh__yhpy, 'arrays')
    omt__zwlbw = c.pyapi.object_getattr_string(bspdh__jhdfa, 'IntervalArray')
    udrp__rbhha = c.pyapi.call_method(omt__zwlbw, 'from_arrays', (xfmv__pkt,
        iyw__xfen))
    c.pyapi.decref(xfmv__pkt)
    c.pyapi.decref(iyw__xfen)
    c.pyapi.decref(jnuuh__yhpy)
    c.pyapi.decref(bspdh__jhdfa)
    c.pyapi.decref(omt__zwlbw)
    c.context.nrt.decref(c.builder, typ, val)
    return udrp__rbhha


@unbox(IntervalArrayType)
def unbox_interval_arr(typ, val, c):
    xfmv__pkt = c.pyapi.object_getattr_string(val, '_left')
    left = c.pyapi.to_native_value(typ.arr_type, xfmv__pkt).value
    c.pyapi.decref(xfmv__pkt)
    iyw__xfen = c.pyapi.object_getattr_string(val, '_right')
    right = c.pyapi.to_native_value(typ.arr_type, iyw__xfen).value
    c.pyapi.decref(iyw__xfen)
    itqle__ownol = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    itqle__ownol.left = left
    itqle__ownol.right = right
    gaykt__kcix = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(itqle__ownol._getvalue(), is_error=gaykt__kcix)


@overload(len, no_unliteral=True)
def overload_interval_arr_len(A):
    if isinstance(A, IntervalArrayType):
        return lambda A: len(A._left)


@overload_attribute(IntervalArrayType, 'shape')
def overload_interval_arr_shape(A):
    return lambda A: (len(A._left),)


@overload_attribute(IntervalArrayType, 'ndim')
def overload_interval_arr_ndim(A):
    return lambda A: 1


@overload_attribute(IntervalArrayType, 'nbytes')
def overload_interval_arr_nbytes(A):
    return lambda A: A._left.nbytes + A._right.nbytes


@overload_method(IntervalArrayType, 'copy', no_unliteral=True)
def overload_interval_arr_copy(A):
    return lambda A: bodo.libs.interval_arr_ext.init_interval_array(A._left
        .copy(), A._right.copy())
