"""
JIT support for Python's logging module
"""
import logging
import numba
from numba.core import types
from numba.core.imputils import lower_constant
from numba.core.typing.templates import bound_function
from numba.core.typing.templates import AttributeTemplate, infer_getattr, signature
from numba.extending import NativeValue, box, models, overload_attribute, overload_method, register_model, typeof_impl, unbox
from bodo.utils.typing import create_unsupported_overload, gen_objmode_attr_overload


class LoggingLoggerType(types.Type):

    def __init__(self, is_root=False):
        self.is_root = is_root
        super(LoggingLoggerType, self).__init__(name=
            f'LoggingLoggerType(is_root={is_root})')


@typeof_impl.register(logging.RootLogger)
@typeof_impl.register(logging.Logger)
def typeof_logging(val, c):
    if isinstance(val, logging.RootLogger):
        return LoggingLoggerType(is_root=True)
    else:
        return LoggingLoggerType(is_root=False)


register_model(LoggingLoggerType)(models.OpaqueModel)


@box(LoggingLoggerType)
def box_logging_logger(typ, val, c):
    c.pyapi.incref(val)
    return val


@unbox(LoggingLoggerType)
def unbox_logging_logger(typ, obj, c):
    c.pyapi.incref(obj)
    return NativeValue(obj)


@lower_constant(LoggingLoggerType)
def lower_constant_logger(context, builder, ty, pyval):
    lnl__lgtd = context.get_python_api(builder)
    return lnl__lgtd.unserialize(lnl__lgtd.serialize_object(pyval))


gen_objmode_attr_overload(LoggingLoggerType, 'level', None, types.int64)
gen_objmode_attr_overload(LoggingLoggerType, 'name', None, 'unicode_type')
gen_objmode_attr_overload(LoggingLoggerType, 'propagate', None, types.boolean)
gen_objmode_attr_overload(LoggingLoggerType, 'disabled', None, types.boolean)
gen_objmode_attr_overload(LoggingLoggerType, 'parent', None,
    LoggingLoggerType())
gen_objmode_attr_overload(LoggingLoggerType, 'root', None,
    LoggingLoggerType(is_root=True))


@infer_getattr
class LoggingLoggerAttribute(AttributeTemplate):
    key = LoggingLoggerType

    def _resolve_helper(self, logger_typ, args, kws):
        kws = dict(kws)
        qkhk__beay = ', '.join('e{}'.format(kgrat__svr) for kgrat__svr in
            range(len(args)))
        if qkhk__beay:
            qkhk__beay += ', '
        opo__fht = ', '.join("{} = ''".format(abtmc__nkwwa) for
            abtmc__nkwwa in kws.keys())
        wbzgx__gqsf = f'def format_stub(string, {qkhk__beay} {opo__fht}):\n'
        wbzgx__gqsf += '    pass\n'
        hhjgb__fzs = {}
        exec(wbzgx__gqsf, {}, hhjgb__fzs)
        nun__ccufx = hhjgb__fzs['format_stub']
        moxpj__kgc = numba.core.utils.pysignature(nun__ccufx)
        pjfw__xgumn = (logger_typ,) + args + tuple(kws.values())
        return signature(logger_typ, pjfw__xgumn).replace(pysig=moxpj__kgc)
    func_names = ('debug', 'warning', 'warn', 'info', 'error', 'exception',
        'critical', 'log', 'setLevel')
    for txvk__yizk in ('logging.Logger', 'logging.RootLogger'):
        for cptyt__tosu in func_names:
            tfze__ecjl = f'@bound_function("{txvk__yizk}.{cptyt__tosu}")\n'
            tfze__ecjl += (
                f'def resolve_{cptyt__tosu}(self, logger_typ, args, kws):\n')
            tfze__ecjl += (
                '    return self._resolve_helper(logger_typ, args, kws)')
            exec(tfze__ecjl)


logging_logger_unsupported_attrs = {'filters', 'handlers', 'manager'}
logging_logger_unsupported_methods = {'addHandler', 'callHandlers', 'fatal',
    'findCaller', 'getChild', 'getEffectiveLevel', 'handle', 'hasHandlers',
    'isEnabledFor', 'makeRecord', 'removeHandler'}


def _install_logging_logger_unsupported_objects():
    for ppbiz__uhx in logging_logger_unsupported_attrs:
        tbsji__enqxp = 'logging.Logger.' + ppbiz__uhx
        overload_attribute(LoggingLoggerType, ppbiz__uhx)(
            create_unsupported_overload(tbsji__enqxp))
    for juj__cpoo in logging_logger_unsupported_methods:
        tbsji__enqxp = 'logging.Logger.' + juj__cpoo
        overload_method(LoggingLoggerType, juj__cpoo)(
            create_unsupported_overload(tbsji__enqxp))


_install_logging_logger_unsupported_objects()
