"""Array implementation for map values.
Corresponds to Spark's MapType: https://spark.apache.org/docs/latest/sql-reference.html
Corresponds to Arrow's Map arrays: https://github.com/apache/arrow/blob/master/format/Schema.fbs

The implementation uses an array(struct) array underneath similar to Spark and Arrow.
For example: [{1: 2.1, 3: 1.1}, {5: -1.0}]
[[{"key": 1, "value" 2.1}, {"key": 3, "value": 1.1}], [{"key": 5, "value": -1.0}]]
"""
import operator
import llvmlite.binding as ll
import numba
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_type
from bodo.libs.array_item_arr_ext import ArrayItemArrayType, _get_array_item_arr_payload, offset_type
from bodo.libs.struct_arr_ext import StructArrayType, _get_struct_arr_payload
from bodo.utils.cg_helpers import dict_keys, dict_merge_from_seq2, dict_values, gen_allocate_array, get_array_elem_counts, get_bitmap_bit, is_na_value, pyarray_setitem, seq_getitem, set_bitmap_bit
from bodo.utils.typing import BodoError
from bodo.libs import array_ext, hdist
ll.add_symbol('count_total_elems_list_array', array_ext.
    count_total_elems_list_array)
ll.add_symbol('map_array_from_sequence', array_ext.map_array_from_sequence)
ll.add_symbol('np_array_from_map_array', array_ext.np_array_from_map_array)


class MapArrayType(types.ArrayCompatible):

    def __init__(self, key_arr_type, value_arr_type):
        self.key_arr_type = key_arr_type
        self.value_arr_type = value_arr_type
        super(MapArrayType, self).__init__(name='MapArrayType({}, {})'.
            format(key_arr_type, value_arr_type))

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return types.DictType(self.key_arr_type.dtype, self.value_arr_type.
            dtype)

    def copy(self):
        return MapArrayType(self.key_arr_type, self.value_arr_type)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


def _get_map_arr_data_type(map_type):
    vwtk__baref = StructArrayType((map_type.key_arr_type, map_type.
        value_arr_type), ('key', 'value'))
    return ArrayItemArrayType(vwtk__baref)


@register_model(MapArrayType)
class MapArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        simbo__wunew = _get_map_arr_data_type(fe_type)
        tegqc__vrwm = [('data', simbo__wunew)]
        models.StructModel.__init__(self, dmm, fe_type, tegqc__vrwm)


make_attribute_wrapper(MapArrayType, 'data', '_data')


@unbox(MapArrayType)
def unbox_map_array(typ, val, c):
    n_maps = bodo.utils.utils.object_length(c, val)
    hvrq__mkavx = all(isinstance(qdkyl__kgaoc, types.Array) and 
        qdkyl__kgaoc.dtype in (types.int64, types.float64, types.bool_,
        datetime_date_type) for qdkyl__kgaoc in (typ.key_arr_type, typ.
        value_arr_type))
    if hvrq__mkavx:
        kaya__nmxi = lir.FunctionType(lir.IntType(64), [lir.IntType(8).
            as_pointer()])
        lsex__fgo = cgutils.get_or_insert_function(c.builder.module,
            kaya__nmxi, name='count_total_elems_list_array')
        jrr__mtvhi = cgutils.pack_array(c.builder, [n_maps, c.builder.call(
            lsex__fgo, [val])])
    else:
        jrr__mtvhi = get_array_elem_counts(c, c.builder, c.context, val, typ)
    simbo__wunew = _get_map_arr_data_type(typ)
    data_arr = gen_allocate_array(c.context, c.builder, simbo__wunew,
        jrr__mtvhi, c)
    gkp__xqfro = _get_array_item_arr_payload(c.context, c.builder,
        simbo__wunew, data_arr)
    null_bitmap_ptr = c.context.make_array(types.Array(types.uint8, 1, 'C'))(c
        .context, c.builder, gkp__xqfro.null_bitmap).data
    offsets_ptr = c.context.make_array(types.Array(offset_type, 1, 'C'))(c.
        context, c.builder, gkp__xqfro.offsets).data
    wsvv__gizxz = _get_struct_arr_payload(c.context, c.builder,
        simbo__wunew.dtype, gkp__xqfro.data)
    key_arr = c.builder.extract_value(wsvv__gizxz.data, 0)
    value_arr = c.builder.extract_value(wsvv__gizxz.data, 1)
    sig = types.none(types.Array(types.uint8, 1, 'C'))
    dhhf__dqf, fam__tvb = c.pyapi.call_jit_code(lambda A: A.fill(255), sig,
        [wsvv__gizxz.null_bitmap])
    if hvrq__mkavx:
        pcqmt__obein = c.context.make_array(simbo__wunew.dtype.data[0])(c.
            context, c.builder, key_arr).data
        ugc__hixal = c.context.make_array(simbo__wunew.dtype.data[1])(c.
            context, c.builder, value_arr).data
        kaya__nmxi = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(offset_type.bitwidth).as_pointer(),
            lir.IntType(8).as_pointer(), lir.IntType(32), lir.IntType(32)])
        oaih__rqu = cgutils.get_or_insert_function(c.builder.module,
            kaya__nmxi, name='map_array_from_sequence')
        naj__jtv = bodo.utils.utils.numba_to_c_type(typ.key_arr_type.dtype)
        jbey__nkpvd = bodo.utils.utils.numba_to_c_type(typ.value_arr_type.dtype
            )
        c.builder.call(oaih__rqu, [val, c.builder.bitcast(pcqmt__obein, lir
            .IntType(8).as_pointer()), c.builder.bitcast(ugc__hixal, lir.
            IntType(8).as_pointer()), offsets_ptr, null_bitmap_ptr, lir.
            Constant(lir.IntType(32), naj__jtv), lir.Constant(lir.IntType(
            32), jbey__nkpvd)])
    else:
        _unbox_map_array_generic(typ, val, c, n_maps, key_arr, value_arr,
            offsets_ptr, null_bitmap_ptr)
    hphre__thasx = c.context.make_helper(c.builder, typ)
    hphre__thasx.data = data_arr
    fomw__wiq = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(hphre__thasx._getvalue(), is_error=fomw__wiq)


def _unbox_map_array_generic(typ, val, c, n_maps, key_arr, value_arr,
    offsets_ptr, null_bitmap_ptr):
    from bodo.libs.array_item_arr_ext import _unbox_array_item_array_copy_data
    context = c.context
    builder = c.builder
    scm__wlijn = context.insert_const_string(builder.module, 'pandas')
    febw__mrsp = c.pyapi.import_module_noblock(scm__wlijn)
    rfs__tzra = c.pyapi.object_getattr_string(febw__mrsp, 'NA')
    fjj__sqq = c.context.get_constant(offset_type, 0)
    builder.store(fjj__sqq, offsets_ptr)
    nip__uhwj = cgutils.alloca_once_value(builder, context.get_constant(
        types.int64, 0))
    with cgutils.for_range(builder, n_maps) as loop:
        whir__fszfe = loop.index
        item_ind = builder.load(nip__uhwj)
        builder.store(builder.trunc(item_ind, lir.IntType(offset_type.
            bitwidth)), builder.gep(offsets_ptr, [whir__fszfe]))
        botr__ouphw = seq_getitem(builder, context, val, whir__fszfe)
        set_bitmap_bit(builder, null_bitmap_ptr, whir__fszfe, 0)
        otd__eps = is_na_value(builder, context, botr__ouphw, rfs__tzra)
        bnvkc__mhbj = builder.icmp_unsigned('!=', otd__eps, lir.Constant(
            otd__eps.type, 1))
        with builder.if_then(bnvkc__mhbj):
            set_bitmap_bit(builder, null_bitmap_ptr, whir__fszfe, 1)
            xmta__yemkr = dict_keys(builder, context, botr__ouphw)
            lumid__eipy = dict_values(builder, context, botr__ouphw)
            n_items = bodo.utils.utils.object_length(c, xmta__yemkr)
            _unbox_array_item_array_copy_data(typ.key_arr_type, xmta__yemkr,
                c, key_arr, item_ind, n_items)
            _unbox_array_item_array_copy_data(typ.value_arr_type,
                lumid__eipy, c, value_arr, item_ind, n_items)
            builder.store(builder.add(item_ind, n_items), nip__uhwj)
            c.pyapi.decref(xmta__yemkr)
            c.pyapi.decref(lumid__eipy)
        c.pyapi.decref(botr__ouphw)
    builder.store(builder.trunc(builder.load(nip__uhwj), lir.IntType(
        offset_type.bitwidth)), builder.gep(offsets_ptr, [n_maps]))
    c.pyapi.decref(febw__mrsp)
    c.pyapi.decref(rfs__tzra)


@box(MapArrayType)
def box_map_arr(typ, val, c):
    hphre__thasx = c.context.make_helper(c.builder, typ, val)
    data_arr = hphre__thasx.data
    simbo__wunew = _get_map_arr_data_type(typ)
    gkp__xqfro = _get_array_item_arr_payload(c.context, c.builder,
        simbo__wunew, data_arr)
    null_bitmap_ptr = c.context.make_array(types.Array(types.uint8, 1, 'C'))(c
        .context, c.builder, gkp__xqfro.null_bitmap).data
    offsets_ptr = c.context.make_array(types.Array(offset_type, 1, 'C'))(c.
        context, c.builder, gkp__xqfro.offsets).data
    wsvv__gizxz = _get_struct_arr_payload(c.context, c.builder,
        simbo__wunew.dtype, gkp__xqfro.data)
    key_arr = c.builder.extract_value(wsvv__gizxz.data, 0)
    value_arr = c.builder.extract_value(wsvv__gizxz.data, 1)
    if all(isinstance(qdkyl__kgaoc, types.Array) and qdkyl__kgaoc.dtype in
        (types.int64, types.float64, types.bool_, datetime_date_type) for
        qdkyl__kgaoc in (typ.key_arr_type, typ.value_arr_type)):
        pcqmt__obein = c.context.make_array(simbo__wunew.dtype.data[0])(c.
            context, c.builder, key_arr).data
        ugc__hixal = c.context.make_array(simbo__wunew.dtype.data[1])(c.
            context, c.builder, value_arr).data
        kaya__nmxi = lir.FunctionType(c.context.get_argument_type(types.
            pyobject), [lir.IntType(64), lir.IntType(8).as_pointer(), lir.
            IntType(8).as_pointer(), lir.IntType(offset_type.bitwidth).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(32), lir
            .IntType(32)])
        vfiej__sgju = cgutils.get_or_insert_function(c.builder.module,
            kaya__nmxi, name='np_array_from_map_array')
        naj__jtv = bodo.utils.utils.numba_to_c_type(typ.key_arr_type.dtype)
        jbey__nkpvd = bodo.utils.utils.numba_to_c_type(typ.value_arr_type.dtype
            )
        arr = c.builder.call(vfiej__sgju, [gkp__xqfro.n_arrays, c.builder.
            bitcast(pcqmt__obein, lir.IntType(8).as_pointer()), c.builder.
            bitcast(ugc__hixal, lir.IntType(8).as_pointer()), offsets_ptr,
            null_bitmap_ptr, lir.Constant(lir.IntType(32), naj__jtv), lir.
            Constant(lir.IntType(32), jbey__nkpvd)])
    else:
        arr = _box_map_array_generic(typ, c, gkp__xqfro.n_arrays, key_arr,
            value_arr, offsets_ptr, null_bitmap_ptr)
    c.context.nrt.decref(c.builder, typ, val)
    return arr


def _box_map_array_generic(typ, c, n_maps, key_arr, value_arr, offsets_ptr,
    null_bitmap_ptr):
    context = c.context
    builder = c.builder
    scm__wlijn = context.insert_const_string(builder.module, 'numpy')
    bygz__thz = c.pyapi.import_module_noblock(scm__wlijn)
    wpem__qemtb = c.pyapi.object_getattr_string(bygz__thz, 'object_')
    cbc__ohu = c.pyapi.long_from_longlong(n_maps)
    aqybd__olc = c.pyapi.call_method(bygz__thz, 'ndarray', (cbc__ohu,
        wpem__qemtb))
    abm__icrvf = c.pyapi.object_getattr_string(bygz__thz, 'nan')
    zgvdo__afdo = c.pyapi.unserialize(c.pyapi.serialize_object(zip))
    nip__uhwj = cgutils.alloca_once_value(builder, lir.Constant(lir.IntType
        (64), 0))
    with cgutils.for_range(builder, n_maps) as loop:
        goslg__bqpw = loop.index
        pyarray_setitem(builder, context, aqybd__olc, goslg__bqpw, abm__icrvf)
        ahol__idimv = get_bitmap_bit(builder, null_bitmap_ptr, goslg__bqpw)
        zad__juv = builder.icmp_unsigned('!=', ahol__idimv, lir.Constant(
            lir.IntType(8), 0))
        with builder.if_then(zad__juv):
            n_items = builder.sext(builder.sub(builder.load(builder.gep(
                offsets_ptr, [builder.add(goslg__bqpw, lir.Constant(
                goslg__bqpw.type, 1))])), builder.load(builder.gep(
                offsets_ptr, [goslg__bqpw]))), lir.IntType(64))
            item_ind = builder.load(nip__uhwj)
            botr__ouphw = c.pyapi.dict_new()
            ilyza__qyrt = lambda data_arr, item_ind, n_items: data_arr[item_ind
                :item_ind + n_items]
            dhhf__dqf, lfhe__isf = c.pyapi.call_jit_code(ilyza__qyrt, typ.
                key_arr_type(typ.key_arr_type, types.int64, types.int64), [
                key_arr, item_ind, n_items])
            dhhf__dqf, bss__cta = c.pyapi.call_jit_code(ilyza__qyrt, typ.
                value_arr_type(typ.value_arr_type, types.int64, types.int64
                ), [value_arr, item_ind, n_items])
            vnk__cgfgj = c.pyapi.from_native_value(typ.key_arr_type,
                lfhe__isf, c.env_manager)
            whn__idk = c.pyapi.from_native_value(typ.value_arr_type,
                bss__cta, c.env_manager)
            tke__tmqb = c.pyapi.call_function_objargs(zgvdo__afdo, (
                vnk__cgfgj, whn__idk))
            dict_merge_from_seq2(builder, context, botr__ouphw, tke__tmqb)
            builder.store(builder.add(item_ind, n_items), nip__uhwj)
            pyarray_setitem(builder, context, aqybd__olc, goslg__bqpw,
                botr__ouphw)
            c.pyapi.decref(tke__tmqb)
            c.pyapi.decref(vnk__cgfgj)
            c.pyapi.decref(whn__idk)
            c.pyapi.decref(botr__ouphw)
    c.pyapi.decref(zgvdo__afdo)
    c.pyapi.decref(bygz__thz)
    c.pyapi.decref(wpem__qemtb)
    c.pyapi.decref(cbc__ohu)
    c.pyapi.decref(abm__icrvf)
    return aqybd__olc


def init_map_arr_codegen(context, builder, sig, args):
    data_arr, = args
    hphre__thasx = context.make_helper(builder, sig.return_type)
    hphre__thasx.data = data_arr
    context.nrt.incref(builder, sig.args[0], data_arr)
    return hphre__thasx._getvalue()


@intrinsic
def init_map_arr(typingctx, data_typ=None):
    assert isinstance(data_typ, ArrayItemArrayType) and isinstance(data_typ
        .dtype, StructArrayType)
    rdd__dcw = MapArrayType(data_typ.dtype.data[0], data_typ.dtype.data[1])
    return rdd__dcw(data_typ), init_map_arr_codegen


def alias_ext_init_map_arr(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 1
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_map_arr',
    'bodo.libs.map_arr_ext'] = alias_ext_init_map_arr


@numba.njit
def pre_alloc_map_array(num_maps, nested_counts, struct_typ):
    bbu__lztpb = bodo.libs.array_item_arr_ext.pre_alloc_array_item_array(
        num_maps, nested_counts, struct_typ)
    return init_map_arr(bbu__lztpb)


def pre_alloc_map_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 3 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis._analyze_op_call_bodo_libs_map_arr_ext_pre_alloc_map_array
    ) = pre_alloc_map_array_equiv


@overload(len, no_unliteral=True)
def overload_map_arr_len(A):
    if isinstance(A, MapArrayType):
        return lambda A: len(A._data)


@overload_attribute(MapArrayType, 'shape')
def overload_map_arr_shape(A):
    return lambda A: (len(A._data),)


@overload_attribute(MapArrayType, 'dtype')
def overload_map_arr_dtype(A):
    return lambda A: np.object_


@overload_attribute(MapArrayType, 'ndim')
def overload_map_arr_ndim(A):
    return lambda A: 1


@overload_attribute(MapArrayType, 'nbytes')
def overload_map_arr_nbytes(A):
    return lambda A: A._data.nbytes


@overload_method(MapArrayType, 'copy')
def overload_map_arr_copy(A):
    return lambda A: init_map_arr(A._data.copy())


@overload(operator.setitem, no_unliteral=True)
def map_arr_setitem(arr, ind, val):
    if not isinstance(arr, MapArrayType):
        return
    yzoxv__yixw = arr.key_arr_type, arr.value_arr_type
    if isinstance(ind, types.Integer):

        def map_arr_setitem_impl(arr, ind, val):
            jmoue__fefsj = val.keys()
            qvl__cais = bodo.libs.struct_arr_ext.pre_alloc_struct_array(len
                (val), (-1,), yzoxv__yixw, ('key', 'value'))
            for dchpc__jyy, psm__bhuss in enumerate(jmoue__fefsj):
                qvl__cais[dchpc__jyy] = bodo.libs.struct_arr_ext.init_struct((
                    psm__bhuss, val[psm__bhuss]), ('key', 'value'))
            arr._data[ind] = qvl__cais
        return map_arr_setitem_impl
    raise BodoError(
        'operator.setitem with MapArrays is only supported with an integer index.'
        )


@overload(operator.getitem, no_unliteral=True)
def map_arr_getitem(arr, ind):
    if not isinstance(arr, MapArrayType):
        return
    if isinstance(ind, types.Integer):

        def map_arr_getitem_impl(arr, ind):
            if ind < 0:
                ind += len(arr)
            nall__bri = dict()
            zfua__toorm = bodo.libs.array_item_arr_ext.get_offsets(arr._data)
            qvl__cais = bodo.libs.array_item_arr_ext.get_data(arr._data)
            uap__wtbu, jozg__ibwee = bodo.libs.struct_arr_ext.get_data(
                qvl__cais)
            qybqj__wmzm = zfua__toorm[ind]
            uyju__pce = zfua__toorm[ind + 1]
            for dchpc__jyy in range(qybqj__wmzm, uyju__pce):
                nall__bri[uap__wtbu[dchpc__jyy]] = jozg__ibwee[dchpc__jyy]
            return nall__bri
        return map_arr_getitem_impl
    raise BodoError(
        'operator.getitem with MapArrays is only supported with an integer index.'
        )
