"""Array implementation for string objects, which are usually immutable.
The characters are stored in a contingous data array, and an offsets array marks the
the individual strings. For example:
value:             ['a', 'bc', '', 'abc', None, 'bb']
data:              [a, b, c, a, b, c, b, b]
offsets:           [0, 1, 3, 3, 6, 6, 8]
"""
import glob
import operator
import llvmlite.llvmpy.core as lc
import numba
import numba.core.typing.typeof
import numpy as np
import pandas as pd
from numba.core import cgutils, types
from numba.core.imputils import RefType, impl_ret_borrowed, iternext_impl, lower_constant
from numba.core.typing.templates import signature
from numba.core.unsafe.bytes import memcpy_region
from numba.extending import NativeValue, box, intrinsic, lower_builtin, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, type_callable, typeof_impl, unbox
import bodo
from bodo.libs.array_item_arr_ext import ArrayItemArrayPayloadType, ArrayItemArrayType, _get_array_item_arr_payload, np_offset_type, offset_type
from bodo.libs.binary_arr_ext import BinaryArrayType, binary_array_type, pre_alloc_binary_array
from bodo.libs.str_ext import memcmp, string_type, unicode_to_utf8_and_len
from bodo.utils.typing import BodoError, is_list_like_index_type, is_overload_constant_int, is_overload_none, is_overload_true, parse_dtype, raise_bodo_error
use_pd_string_array = False
char_type = types.uint8
char_arr_type = types.Array(char_type, 1, 'C')
offset_arr_type = types.Array(offset_type, 1, 'C')
null_bitmap_arr_type = types.Array(types.uint8, 1, 'C')
data_ctypes_type = types.ArrayCTypes(char_arr_type)
offset_ctypes_type = types.ArrayCTypes(offset_arr_type)


class StringArrayType(types.IterableType, types.ArrayCompatible):

    def __init__(self):
        super(StringArrayType, self).__init__(name='StringArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return string_type

    @property
    def iterator_type(self):
        return StringArrayIterator()

    def copy(self):
        return StringArrayType()


string_array_type = StringArrayType()


@typeof_impl.register(pd.arrays.StringArray)
def typeof_string_array(val, c):
    return string_array_type


@register_model(BinaryArrayType)
@register_model(StringArrayType)
class StringArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        abpkm__obvfq = ArrayItemArrayType(char_arr_type)
        mpv__nmw = [('data', abpkm__obvfq)]
        models.StructModel.__init__(self, dmm, fe_type, mpv__nmw)


make_attribute_wrapper(StringArrayType, 'data', '_data')
make_attribute_wrapper(BinaryArrayType, 'data', '_data')


@intrinsic
def init_str_arr(typingctx, data_typ=None):
    assert isinstance(data_typ, ArrayItemArrayType
        ) and data_typ.dtype == types.Array(char_type, 1, 'C')

    def codegen(context, builder, sig, args):
        rtuv__uwvh, = args
        zhehf__naaeb = context.make_helper(builder, string_array_type)
        zhehf__naaeb.data = rtuv__uwvh
        context.nrt.incref(builder, data_typ, rtuv__uwvh)
        return zhehf__naaeb._getvalue()
    return string_array_type(data_typ), codegen


class StringDtype(types.Number):

    def __init__(self):
        super(StringDtype, self).__init__('StringDtype')


string_dtype = StringDtype()
register_model(StringDtype)(models.OpaqueModel)


@box(StringDtype)
def box_string_dtype(typ, val, c):
    vxelb__goy = c.context.insert_const_string(c.builder.module, 'pandas')
    vzs__gkp = c.pyapi.import_module_noblock(vxelb__goy)
    hyuc__eyrtg = c.pyapi.call_method(vzs__gkp, 'StringDtype', ())
    c.pyapi.decref(vzs__gkp)
    return hyuc__eyrtg


@unbox(StringDtype)
def unbox_string_dtype(typ, val, c):
    return NativeValue(c.context.get_dummy_value())


typeof_impl.register(pd.StringDtype)(lambda a, b: string_dtype)
type_callable(pd.StringDtype)(lambda c: lambda : string_dtype)
lower_builtin(pd.StringDtype)(lambda c, b, s, a: c.get_dummy_value())


def create_binary_op_overload(op):

    def overload_string_array_binary_op(lhs, rhs):
        if lhs == string_array_type and rhs == string_array_type:

            def impl_both(lhs, rhs):
                numba.parfors.parfor.init_prange()
                yvi__ercj = len(lhs)
                out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(yvi__ercj)
                for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                    if bodo.libs.array_kernels.isna(lhs, i
                        ) or bodo.libs.array_kernels.isna(rhs, i):
                        bodo.libs.array_kernels.setna(out_arr, i)
                        continue
                    val = op(lhs[i], rhs[i])
                    out_arr[i] = val
                return out_arr
            return impl_both
        if lhs == string_array_type and types.unliteral(rhs) == string_type:

            def impl_left(lhs, rhs):
                numba.parfors.parfor.init_prange()
                yvi__ercj = len(lhs)
                out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(yvi__ercj)
                for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                    if bodo.libs.array_kernels.isna(lhs, i):
                        bodo.libs.array_kernels.setna(out_arr, i)
                        continue
                    val = op(lhs[i], rhs)
                    out_arr[i] = val
                return out_arr
            return impl_left
        if types.unliteral(lhs) == string_type and rhs == string_array_type:

            def impl_right(lhs, rhs):
                numba.parfors.parfor.init_prange()
                yvi__ercj = len(rhs)
                out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(yvi__ercj)
                for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                    if bodo.libs.array_kernels.isna(rhs, i):
                        bodo.libs.array_kernels.setna(out_arr, i)
                        continue
                    val = op(lhs, rhs[i])
                    out_arr[i] = val
                return out_arr
            return impl_right
        raise BodoError(
            f'{op} operator not supported for data types {lhs} and {rhs}.')
    return overload_string_array_binary_op


def overload_add_operator_string_array(lhs, rhs):
    brn__rsds = lhs == string_array_type or isinstance(lhs, types.Array
        ) and lhs.dtype == string_type
    thsbo__dcjpn = rhs == string_array_type or isinstance(rhs, types.Array
        ) and rhs.dtype == string_type
    if (lhs == string_array_type and thsbo__dcjpn or brn__rsds and rhs ==
        string_array_type):

        def impl_both(lhs, rhs):
            numba.parfors.parfor.init_prange()
            l = len(lhs)
            out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(l, -1)
            for j in numba.parfors.parfor.internal_prange(l):
                if bodo.libs.array_kernels.isna(lhs, j
                    ) or bodo.libs.array_kernels.isna(rhs, j):
                    out_arr[j] = ''
                    bodo.libs.array_kernels.setna(out_arr, j)
                else:
                    out_arr[j] = lhs[j] + rhs[j]
            return out_arr
        return impl_both
    if lhs == string_array_type and types.unliteral(rhs) == string_type:

        def impl_left(lhs, rhs):
            numba.parfors.parfor.init_prange()
            l = len(lhs)
            out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(l, -1)
            for j in numba.parfors.parfor.internal_prange(l):
                if bodo.libs.array_kernels.isna(lhs, j):
                    out_arr[j] = ''
                    bodo.libs.array_kernels.setna(out_arr, j)
                else:
                    out_arr[j] = lhs[j] + rhs
            return out_arr
        return impl_left
    if types.unliteral(lhs) == string_type and rhs == string_array_type:

        def impl_right(lhs, rhs):
            numba.parfors.parfor.init_prange()
            l = len(rhs)
            out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(l, -1)
            for j in numba.parfors.parfor.internal_prange(l):
                if bodo.libs.array_kernels.isna(rhs, j):
                    out_arr[j] = ''
                    bodo.libs.array_kernels.setna(out_arr, j)
                else:
                    out_arr[j] = lhs + rhs[j]
            return out_arr
        return impl_right


def overload_mul_operator_str_arr(lhs, rhs):
    if lhs == string_array_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            numba.parfors.parfor.init_prange()
            l = len(lhs)
            out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(l, -1)
            for j in numba.parfors.parfor.internal_prange(l):
                if bodo.libs.array_kernels.isna(lhs, j):
                    out_arr[j] = ''
                    bodo.libs.array_kernels.setna(out_arr, j)
                else:
                    out_arr[j] = lhs[j] * rhs
            return out_arr
        return impl
    if isinstance(lhs, types.Integer) and rhs == string_array_type:

        def impl(lhs, rhs):
            return rhs * lhs
        return impl


class StringArrayIterator(types.SimpleIteratorType):

    def __init__(self):
        npv__ztgsg = 'iter(String)'
        smgf__lsjha = string_type
        super(StringArrayIterator, self).__init__(npv__ztgsg, smgf__lsjha)


@register_model(StringArrayIterator)
class StrArrayIteratorModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        mpv__nmw = [('index', types.EphemeralPointer(types.uintp)), (
            'array', string_array_type)]
        super(StrArrayIteratorModel, self).__init__(dmm, fe_type, mpv__nmw)


lower_builtin('getiter', string_array_type)(numba.np.arrayobj.getiter_array)


@lower_builtin('iternext', StringArrayIterator)
@iternext_impl(RefType.NEW)
def iternext_str_array(context, builder, sig, args, result):
    [angr__rxnuj] = sig.args
    [joch__vaxf] = args
    jik__npagt = context.make_helper(builder, angr__rxnuj, value=joch__vaxf)
    lrr__zhzx = signature(types.intp, string_array_type)
    fux__fgm = context.compile_internal(builder, lambda a: len(a),
        lrr__zhzx, [jik__npagt.array])
    rjtq__ewaa = builder.load(jik__npagt.index)
    hkn__vizlo = builder.icmp(lc.ICMP_SLT, rjtq__ewaa, fux__fgm)
    result.set_valid(hkn__vizlo)
    with builder.if_then(hkn__vizlo):
        wmu__hku = signature(string_type, string_array_type, types.intp)
        value = context.compile_internal(builder, lambda a, i: a[i],
            wmu__hku, [jik__npagt.array, rjtq__ewaa])
        result.yield_(value)
        xoi__octvw = cgutils.increment_index(builder, rjtq__ewaa)
        builder.store(xoi__octvw, jik__npagt.index)


def _get_str_binary_arr_payload(context, builder, arr_value, arr_typ):
    assert arr_typ == string_array_type or arr_typ == binary_array_type
    rgp__vglt = context.make_helper(builder, arr_typ, arr_value)
    abpkm__obvfq = ArrayItemArrayType(char_arr_type)
    pyxfu__uqgwe = _get_array_item_arr_payload(context, builder,
        abpkm__obvfq, rgp__vglt.data)
    return pyxfu__uqgwe


@intrinsic
def num_strings(typingctx, str_arr_typ=None):
    assert str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        return pyxfu__uqgwe.n_arrays
    return types.int64(string_array_type), codegen


def _get_num_total_chars(builder, offsets, num_strings):
    return builder.zext(builder.load(builder.gep(offsets, [num_strings])),
        lir.IntType(64))


@intrinsic
def num_total_chars(typingctx, in_arr_typ=None):
    assert in_arr_typ in [binary_array_type, string_array_type]

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        iyz__vrw = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets).data
        return _get_num_total_chars(builder, iyz__vrw, pyxfu__uqgwe.n_arrays)
    return types.uint64(in_arr_typ), codegen


@intrinsic
def get_offset_ptr(typingctx, in_arr_typ=None):
    assert in_arr_typ in [binary_array_type, string_array_type]

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        hton__zuavt = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets)
        vavhi__twvut = context.make_helper(builder, offset_ctypes_type)
        vavhi__twvut.data = builder.bitcast(hton__zuavt.data, lir.IntType(
            offset_type.bitwidth).as_pointer())
        vavhi__twvut.meminfo = hton__zuavt.meminfo
        hyuc__eyrtg = vavhi__twvut._getvalue()
        return impl_ret_borrowed(context, builder, offset_ctypes_type,
            hyuc__eyrtg)
    return offset_ctypes_type(in_arr_typ), codegen


@intrinsic
def get_data_ptr(typingctx, in_arr_typ=None):
    assert in_arr_typ in [binary_array_type, string_array_type]

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        rtuv__uwvh = context.make_helper(builder, char_arr_type,
            pyxfu__uqgwe.data)
        vavhi__twvut = context.make_helper(builder, data_ctypes_type)
        vavhi__twvut.data = rtuv__uwvh.data
        vavhi__twvut.meminfo = rtuv__uwvh.meminfo
        hyuc__eyrtg = vavhi__twvut._getvalue()
        return impl_ret_borrowed(context, builder, data_ctypes_type,
            hyuc__eyrtg)
    return data_ctypes_type(in_arr_typ), codegen


@intrinsic
def get_data_ptr_ind(typingctx, in_arr_typ, int_t=None):
    assert in_arr_typ in [binary_array_type, string_array_type]

    def codegen(context, builder, sig, args):
        vwc__vibae, ind = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            vwc__vibae, sig.args[0])
        rtuv__uwvh = context.make_helper(builder, char_arr_type,
            pyxfu__uqgwe.data)
        vavhi__twvut = context.make_helper(builder, data_ctypes_type)
        vavhi__twvut.data = builder.gep(rtuv__uwvh.data, [ind])
        vavhi__twvut.meminfo = rtuv__uwvh.meminfo
        hyuc__eyrtg = vavhi__twvut._getvalue()
        return impl_ret_borrowed(context, builder, data_ctypes_type,
            hyuc__eyrtg)
    return data_ctypes_type(in_arr_typ, types.intp), codegen


@intrinsic
def copy_single_char(typingctx, dst_ptr_t, dst_ind_t, src_ptr_t, src_ind_t=None
    ):

    def codegen(context, builder, sig, args):
        cple__uxk, mbigz__lps, mgokl__xxho, lxne__yll = args
        lxgq__yjgvf = builder.bitcast(builder.gep(cple__uxk, [mbigz__lps]),
            lir.IntType(8).as_pointer())
        swux__fjqay = builder.bitcast(builder.gep(mgokl__xxho, [lxne__yll]),
            lir.IntType(8).as_pointer())
        jrj__wauc = builder.load(swux__fjqay)
        builder.store(jrj__wauc, lxgq__yjgvf)
        return context.get_dummy_value()
    return types.void(types.voidptr, types.intp, types.voidptr, types.intp
        ), codegen


@intrinsic
def get_null_bitmap_ptr(typingctx, in_arr_typ=None):
    assert in_arr_typ in [binary_array_type, string_array_type]

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        afqgo__gjn = context.make_helper(builder, null_bitmap_arr_type,
            pyxfu__uqgwe.null_bitmap)
        vavhi__twvut = context.make_helper(builder, data_ctypes_type)
        vavhi__twvut.data = afqgo__gjn.data
        vavhi__twvut.meminfo = afqgo__gjn.meminfo
        hyuc__eyrtg = vavhi__twvut._getvalue()
        return impl_ret_borrowed(context, builder, data_ctypes_type,
            hyuc__eyrtg)
    return data_ctypes_type(in_arr_typ), codegen


@intrinsic
def getitem_str_offset(typingctx, in_arr_typ, ind_t=None):
    assert in_arr_typ in [binary_array_type, string_array_type]

    def codegen(context, builder, sig, args):
        in_str_arr, ind = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        iyz__vrw = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets).data
        return builder.load(builder.gep(iyz__vrw, [ind]))
    return offset_type(in_arr_typ, ind_t), codegen


@intrinsic
def setitem_str_offset(typingctx, str_arr_typ, ind_t, val_t=None):
    assert str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        in_str_arr, ind, val = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        offsets = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets).data
        builder.store(val, builder.gep(offsets, [ind]))
        return context.get_dummy_value()
    return types.void(string_array_type, ind_t, offset_type), codegen


@intrinsic
def getitem_str_bitmap(typingctx, in_bitmap_typ, ind_t=None):

    def codegen(context, builder, sig, args):
        csxr__vasq, ind = args
        if in_bitmap_typ == data_ctypes_type:
            vavhi__twvut = context.make_helper(builder, data_ctypes_type,
                csxr__vasq)
            csxr__vasq = vavhi__twvut.data
        return builder.load(builder.gep(csxr__vasq, [ind]))
    return char_type(in_bitmap_typ, ind_t), codegen


@intrinsic
def setitem_str_bitmap(typingctx, in_bitmap_typ, ind_t, val_t=None):

    def codegen(context, builder, sig, args):
        csxr__vasq, ind, val = args
        if in_bitmap_typ == data_ctypes_type:
            vavhi__twvut = context.make_helper(builder, data_ctypes_type,
                csxr__vasq)
            csxr__vasq = vavhi__twvut.data
        builder.store(val, builder.gep(csxr__vasq, [ind]))
        return context.get_dummy_value()
    return types.void(in_bitmap_typ, ind_t, char_type), codegen


@intrinsic
def copy_str_arr_slice(typingctx, out_str_arr_typ, in_str_arr_typ, ind_t=None):
    assert out_str_arr_typ == string_array_type and in_str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        out_str_arr, in_str_arr, ind = args
        caqz__uqkd = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        uswa__vgkxb = _get_str_binary_arr_payload(context, builder,
            out_str_arr, string_array_type)
        ngqu__btzvx = context.make_helper(builder, offset_arr_type,
            caqz__uqkd.offsets).data
        ubv__ocb = context.make_helper(builder, offset_arr_type,
            uswa__vgkxb.offsets).data
        loxp__ybr = context.make_helper(builder, char_arr_type, caqz__uqkd.data
            ).data
        nhkgn__ygee = context.make_helper(builder, char_arr_type,
            uswa__vgkxb.data).data
        psly__dagn = context.make_helper(builder, null_bitmap_arr_type,
            caqz__uqkd.null_bitmap).data
        bqjd__qggh = context.make_helper(builder, null_bitmap_arr_type,
            uswa__vgkxb.null_bitmap).data
        rfaax__idgks = builder.add(ind, context.get_constant(types.intp, 1))
        cgutils.memcpy(builder, ubv__ocb, ngqu__btzvx, rfaax__idgks)
        cgutils.memcpy(builder, nhkgn__ygee, loxp__ybr, builder.load(
            builder.gep(ngqu__btzvx, [ind])))
        jzq__jgg = builder.add(ind, lir.Constant(lir.IntType(64), 7))
        rtl__jhua = builder.lshr(jzq__jgg, lir.Constant(lir.IntType(64), 3))
        cgutils.memcpy(builder, bqjd__qggh, psly__dagn, rtl__jhua)
        return context.get_dummy_value()
    return types.void(string_array_type, string_array_type, ind_t), codegen


@intrinsic
def copy_data(typingctx, str_arr_typ, out_str_arr_typ=None):
    assert str_arr_typ == string_array_type and out_str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        out_str_arr, in_str_arr = args
        caqz__uqkd = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        uswa__vgkxb = _get_str_binary_arr_payload(context, builder,
            out_str_arr, string_array_type)
        ngqu__btzvx = context.make_helper(builder, offset_arr_type,
            caqz__uqkd.offsets).data
        loxp__ybr = context.make_helper(builder, char_arr_type, caqz__uqkd.data
            ).data
        nhkgn__ygee = context.make_helper(builder, char_arr_type,
            uswa__vgkxb.data).data
        num_total_chars = _get_num_total_chars(builder, ngqu__btzvx,
            caqz__uqkd.n_arrays)
        cgutils.memcpy(builder, nhkgn__ygee, loxp__ybr, num_total_chars)
        return context.get_dummy_value()
    return types.void(string_array_type, string_array_type), codegen


@intrinsic
def copy_non_null_offsets(typingctx, str_arr_typ, out_str_arr_typ=None):
    assert str_arr_typ == string_array_type and out_str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        out_str_arr, in_str_arr = args
        caqz__uqkd = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        uswa__vgkxb = _get_str_binary_arr_payload(context, builder,
            out_str_arr, string_array_type)
        ngqu__btzvx = context.make_helper(builder, offset_arr_type,
            caqz__uqkd.offsets).data
        ubv__ocb = context.make_helper(builder, offset_arr_type,
            uswa__vgkxb.offsets).data
        psly__dagn = context.make_helper(builder, null_bitmap_arr_type,
            caqz__uqkd.null_bitmap).data
        yvi__ercj = caqz__uqkd.n_arrays
        xmt__ynxbu = context.get_constant(offset_type, 0)
        wip__dkb = cgutils.alloca_once_value(builder, xmt__ynxbu)
        with cgutils.for_range(builder, yvi__ercj) as loop:
            mdad__gcce = lower_is_na(context, builder, psly__dagn, loop.index)
            with cgutils.if_likely(builder, builder.not_(mdad__gcce)):
                erst__ard = builder.load(builder.gep(ngqu__btzvx, [loop.index])
                    )
                coe__xfgg = builder.load(wip__dkb)
                builder.store(erst__ard, builder.gep(ubv__ocb, [coe__xfgg]))
                builder.store(builder.add(coe__xfgg, lir.Constant(context.
                    get_value_type(offset_type), 1)), wip__dkb)
        coe__xfgg = builder.load(wip__dkb)
        erst__ard = builder.load(builder.gep(ngqu__btzvx, [yvi__ercj]))
        builder.store(erst__ard, builder.gep(ubv__ocb, [coe__xfgg]))
        return context.get_dummy_value()
    return types.void(string_array_type, string_array_type), codegen


@intrinsic
def str_copy(typingctx, buff_arr_typ, ind_typ, str_typ, len_typ=None):

    def codegen(context, builder, sig, args):
        osoae__yob, ind, str, lmio__xqo = args
        osoae__yob = context.make_array(sig.args[0])(context, builder,
            osoae__yob)
        lva__unbu = builder.gep(osoae__yob.data, [ind])
        cgutils.raw_memcpy(builder, lva__unbu, str, lmio__xqo, 1)
        return context.get_dummy_value()
    return types.void(null_bitmap_arr_type, types.intp, types.voidptr,
        types.intp), codegen


@intrinsic
def str_copy_ptr(typingctx, ptr_typ, ind_typ, str_typ, len_typ=None):

    def codegen(context, builder, sig, args):
        lva__unbu, ind, cid__plxm, lmio__xqo = args
        lva__unbu = builder.gep(lva__unbu, [ind])
        cgutils.raw_memcpy(builder, lva__unbu, cid__plxm, lmio__xqo, 1)
        return context.get_dummy_value()
    return types.void(types.voidptr, types.intp, types.voidptr, types.intp
        ), codegen


@numba.njit(no_cpython_wrapper=True)
def get_str_arr_item_length(A, i):
    return np.int64(getitem_str_offset(A, i + 1) - getitem_str_offset(A, i))


@numba.njit(no_cpython_wrapper=True)
def get_str_arr_item_ptr(A, i):
    return get_data_ptr_ind(A, getitem_str_offset(A, i))


@numba.njit(no_cpython_wrapper=True)
def get_str_arr_item_copy(B, j, A, i):
    if j == 0:
        setitem_str_offset(B, 0, 0)
    cam__dkb = getitem_str_offset(A, i)
    dtf__zsr = getitem_str_offset(A, i + 1)
    fyax__wwyn = dtf__zsr - cam__dkb
    abu__jym = getitem_str_offset(B, j)
    hiyo__fuj = abu__jym + fyax__wwyn
    setitem_str_offset(B, j + 1, hiyo__fuj)
    if str_arr_is_na(A, i):
        str_arr_set_na(B, j)
    else:
        str_arr_set_not_na(B, j)
    if fyax__wwyn != 0:
        rtuv__uwvh = B._data
        bodo.libs.array_item_arr_ext.ensure_data_capacity(rtuv__uwvh, np.
            int64(abu__jym), np.int64(hiyo__fuj))
        vut__bir = get_data_ptr(B).data
        tbk__cbugc = get_data_ptr(A).data
        memcpy_region(vut__bir, abu__jym, tbk__cbugc, cam__dkb, fyax__wwyn, 1)


@numba.njit(no_cpython_wrapper=True)
def get_str_null_bools(str_arr):
    yvi__ercj = len(str_arr)
    vekru__kcrir = np.empty(yvi__ercj, np.bool_)
    for i in range(yvi__ercj):
        vekru__kcrir[i] = bodo.libs.array_kernels.isna(str_arr, i)
    return vekru__kcrir


def to_list_if_immutable_arr(arr, str_null_bools=None):
    return arr


@overload(to_list_if_immutable_arr, no_unliteral=True)
def to_list_if_immutable_arr_overload(data, str_null_bools=None):
    if data in [string_array_type, binary_array_type]:

        def to_list_impl(data, str_null_bools=None):
            yvi__ercj = len(data)
            l = []
            for i in range(yvi__ercj):
                l.append(data[i])
            return l
        return to_list_impl
    if isinstance(data, types.BaseTuple):
        ewi__khn = data.count
        iilo__jiw = ['to_list_if_immutable_arr(data[{}])'.format(i) for i in
            range(ewi__khn)]
        if is_overload_true(str_null_bools):
            iilo__jiw += ['get_str_null_bools(data[{}])'.format(i) for i in
                range(ewi__khn) if data.types[i] in [string_array_type,
                binary_array_type]]
        xoqve__hiab = 'def f(data, str_null_bools=None):\n'
        xoqve__hiab += '  return ({}{})\n'.format(', '.join(iilo__jiw), ',' if
            ewi__khn == 1 else '')
        iin__mhfrg = {}
        exec(xoqve__hiab, {'to_list_if_immutable_arr':
            to_list_if_immutable_arr, 'get_str_null_bools':
            get_str_null_bools, 'bodo': bodo}, iin__mhfrg)
        whcl__rvnp = iin__mhfrg['f']
        return whcl__rvnp
    return lambda data, str_null_bools=None: data


def cp_str_list_to_array(str_arr, str_list, str_null_bools=None):
    return


@overload(cp_str_list_to_array, no_unliteral=True)
def cp_str_list_to_array_overload(str_arr, list_data, str_null_bools=None):
    if str_arr == string_array_type:
        if is_overload_none(str_null_bools):

            def cp_str_list_impl(str_arr, list_data, str_null_bools=None):
                yvi__ercj = len(list_data)
                for i in range(yvi__ercj):
                    cid__plxm = list_data[i]
                    str_arr[i] = cid__plxm
            return cp_str_list_impl
        else:

            def cp_str_list_impl_null(str_arr, list_data, str_null_bools=None):
                yvi__ercj = len(list_data)
                for i in range(yvi__ercj):
                    cid__plxm = list_data[i]
                    str_arr[i] = cid__plxm
                    if str_null_bools[i]:
                        str_arr_set_na(str_arr, i)
                    else:
                        str_arr_set_not_na(str_arr, i)
            return cp_str_list_impl_null
    if isinstance(str_arr, types.BaseTuple):
        ewi__khn = str_arr.count
        pyg__ylub = 0
        xoqve__hiab = 'def f(str_arr, list_data, str_null_bools=None):\n'
        for i in range(ewi__khn):
            if is_overload_true(str_null_bools) and str_arr.types[i
                ] == string_array_type:
                xoqve__hiab += (
                    """  cp_str_list_to_array(str_arr[{}], list_data[{}], list_data[{}])
"""
                    .format(i, i, ewi__khn + pyg__ylub))
                pyg__ylub += 1
            else:
                xoqve__hiab += (
                    '  cp_str_list_to_array(str_arr[{}], list_data[{}])\n'.
                    format(i, i))
        xoqve__hiab += '  return\n'
        iin__mhfrg = {}
        exec(xoqve__hiab, {'cp_str_list_to_array': cp_str_list_to_array},
            iin__mhfrg)
        hduim__mtq = iin__mhfrg['f']
        return hduim__mtq
    return lambda str_arr, list_data, str_null_bools=None: None


def str_list_to_array(str_list):
    return str_list


@overload(str_list_to_array, no_unliteral=True)
def str_list_to_array_overload(str_list):
    if isinstance(str_list, types.List) and str_list.dtype == bodo.string_type:

        def str_list_impl(str_list):
            yvi__ercj = len(str_list)
            str_arr = pre_alloc_string_array(yvi__ercj, -1)
            for i in range(yvi__ercj):
                cid__plxm = str_list[i]
                str_arr[i] = cid__plxm
            return str_arr
        return str_list_impl
    return lambda str_list: str_list


def get_num_total_chars(A):
    pass


@overload(get_num_total_chars)
def overload_get_num_total_chars(A):
    if isinstance(A, types.List) and A.dtype == string_type:

        def str_list_impl(A):
            yvi__ercj = len(A)
            hyhn__odp = 0
            for i in range(yvi__ercj):
                cid__plxm = A[i]
                hyhn__odp += get_utf8_size(cid__plxm)
            return hyhn__odp
        return str_list_impl
    assert A == string_array_type
    return lambda A: num_total_chars(A)


@overload_method(StringArrayType, 'copy', no_unliteral=True)
def str_arr_copy_overload(arr):

    def copy_impl(arr):
        yvi__ercj = len(arr)
        n_chars = num_total_chars(arr)
        kxd__aowu = pre_alloc_string_array(yvi__ercj, np.int64(n_chars))
        copy_str_arr_slice(kxd__aowu, arr, yvi__ercj)
        return kxd__aowu
    return copy_impl


@overload(len, no_unliteral=True)
def str_arr_len_overload(str_arr):
    if str_arr == string_array_type:

        def str_arr_len(str_arr):
            return str_arr.size
        return str_arr_len


@overload_attribute(StringArrayType, 'size')
def str_arr_size_overload(str_arr):
    return lambda str_arr: len(str_arr._data)


@overload_attribute(StringArrayType, 'shape')
def str_arr_shape_overload(str_arr):
    return lambda str_arr: (str_arr.size,)


@overload_attribute(StringArrayType, 'nbytes')
def str_arr_nbytes_overload(str_arr):
    return lambda str_arr: str_arr._data.nbytes


@overload_method(types.Array, 'tolist', no_unliteral=True)
@overload_method(StringArrayType, 'tolist', no_unliteral=True)
def overload_to_list(arr):
    return lambda arr: list(arr)


import llvmlite.binding as ll
from llvmlite import ir as lir
from bodo.libs import array_ext, hstr_ext
ll.add_symbol('get_str_len', hstr_ext.get_str_len)
ll.add_symbol('setitem_string_array', hstr_ext.setitem_string_array)
ll.add_symbol('is_na', hstr_ext.is_na)
ll.add_symbol('string_array_from_sequence', array_ext.
    string_array_from_sequence)
ll.add_symbol('pd_array_from_string_array', hstr_ext.pd_array_from_string_array
    )
ll.add_symbol('np_array_from_string_array', hstr_ext.np_array_from_string_array
    )
ll.add_symbol('convert_len_arr_to_offset32', hstr_ext.
    convert_len_arr_to_offset32)
ll.add_symbol('convert_len_arr_to_offset', hstr_ext.convert_len_arr_to_offset)
ll.add_symbol('set_string_array_range', hstr_ext.set_string_array_range)
ll.add_symbol('str_arr_to_int64', hstr_ext.str_arr_to_int64)
ll.add_symbol('str_arr_to_float64', hstr_ext.str_arr_to_float64)
ll.add_symbol('get_utf8_size', hstr_ext.get_utf8_size)
ll.add_symbol('print_str_arr', hstr_ext.print_str_arr)
ll.add_symbol('inplace_int64_to_str', hstr_ext.inplace_int64_to_str)
inplace_int64_to_str = types.ExternalFunction('inplace_int64_to_str', types
    .void(types.voidptr, types.int64, types.int64))
convert_len_arr_to_offset32 = types.ExternalFunction(
    'convert_len_arr_to_offset32', types.void(types.voidptr, types.intp))
convert_len_arr_to_offset = types.ExternalFunction('convert_len_arr_to_offset',
    types.void(types.voidptr, types.voidptr, types.intp))
setitem_string_array = types.ExternalFunction('setitem_string_array', types
    .void(types.CPointer(offset_type), types.CPointer(char_type), types.
    uint64, types.voidptr, types.intp, offset_type, offset_type, types.intp))
_get_utf8_size = types.ExternalFunction('get_utf8_size', types.intp(types.
    voidptr, types.intp, offset_type))
_print_str_arr = types.ExternalFunction('print_str_arr', types.void(types.
    uint64, types.uint64, types.CPointer(offset_type), types.CPointer(
    char_type)))


@numba.generated_jit(nopython=True)
def empty_str_arr(in_seq):
    xoqve__hiab = 'def f(in_seq):\n'
    xoqve__hiab += '    n_strs = len(in_seq)\n'
    xoqve__hiab += '    A = pre_alloc_string_array(n_strs, -1)\n'
    xoqve__hiab += '    return A\n'
    iin__mhfrg = {}
    exec(xoqve__hiab, {'pre_alloc_string_array': pre_alloc_string_array},
        iin__mhfrg)
    huphf__hxeh = iin__mhfrg['f']
    return huphf__hxeh


@numba.generated_jit(nopython=True)
def str_arr_from_sequence(in_seq):
    if in_seq.dtype == bodo.bytes_type:
        ryaq__plr = 'pre_alloc_binary_array'
    else:
        ryaq__plr = 'pre_alloc_string_array'
    xoqve__hiab = 'def f(in_seq):\n'
    xoqve__hiab += '    n_strs = len(in_seq)\n'
    xoqve__hiab += f'    A = {ryaq__plr}(n_strs, -1)\n'
    xoqve__hiab += '    for i in range(n_strs):\n'
    xoqve__hiab += '        A[i] = in_seq[i]\n'
    xoqve__hiab += '    return A\n'
    iin__mhfrg = {}
    exec(xoqve__hiab, {'pre_alloc_string_array': pre_alloc_string_array,
        'pre_alloc_binary_array': pre_alloc_binary_array}, iin__mhfrg)
    huphf__hxeh = iin__mhfrg['f']
    return huphf__hxeh


@intrinsic
def set_all_offsets_to_0(typingctx, arr_typ=None):
    assert arr_typ in (string_array_type, binary_array_type
        ), 'set_all_offsets_to_0 requires a string or binary array'

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        vcgsx__lbm = builder.add(pyxfu__uqgwe.n_arrays, lir.Constant(lir.
            IntType(64), 1))
        gxc__rmi = builder.lshr(lir.Constant(lir.IntType(64), offset_type.
            bitwidth), lir.Constant(lir.IntType(64), 3))
        rtl__jhua = builder.mul(vcgsx__lbm, gxc__rmi)
        hnw__stenx = context.make_array(offset_arr_type)(context, builder,
            pyxfu__uqgwe.offsets).data
        cgutils.memset(builder, hnw__stenx, rtl__jhua, 0)
        return context.get_dummy_value()
    return types.none(arr_typ), codegen


@intrinsic
def set_bitmap_all_NA(typingctx, arr_typ=None):
    assert arr_typ in (string_array_type, binary_array_type
        ), 'set_bitmap_all_NA requires a string or binary array'

    def codegen(context, builder, sig, args):
        in_str_arr, = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, sig.args[0])
        cpwga__ywr = pyxfu__uqgwe.n_arrays
        rtl__jhua = builder.lshr(builder.add(cpwga__ywr, lir.Constant(lir.
            IntType(64), 7)), lir.Constant(lir.IntType(64), 3))
        xczo__fgrcr = context.make_array(null_bitmap_arr_type)(context,
            builder, pyxfu__uqgwe.null_bitmap).data
        cgutils.memset(builder, xczo__fgrcr, rtl__jhua, 0)
        return context.get_dummy_value()
    return types.none(arr_typ), codegen


@numba.njit
def pre_alloc_string_array(n_strs, n_chars):
    if n_chars is None:
        n_chars = -1
    str_arr = init_str_arr(bodo.libs.array_item_arr_ext.
        pre_alloc_array_item_array(np.int64(n_strs), (np.int64(n_chars),),
        char_arr_type))
    if n_chars == 0:
        set_all_offsets_to_0(str_arr)
    return str_arr


@register_jitable
def gen_na_str_array_lens(n_strs, total_len, len_arr):
    str_arr = pre_alloc_string_array(n_strs, total_len)
    set_bitmap_all_NA(str_arr)
    offsets = bodo.libs.array_item_arr_ext.get_offsets(str_arr._data)
    davo__oyggy = 0
    vhuxc__aqxp = len(len_arr)
    for i in range(vhuxc__aqxp):
        offsets[i] = davo__oyggy
        davo__oyggy += len_arr[i]
    offsets[vhuxc__aqxp] = davo__oyggy
    return str_arr


kBitmask = np.array([1, 2, 4, 8, 16, 32, 64, 128], dtype=np.uint8)


@numba.njit
def set_bit_to(bits, i, bit_is_set):
    fpxu__plc = i // 8
    cwkvq__ffni = getitem_str_bitmap(bits, fpxu__plc)
    cwkvq__ffni ^= np.uint8(-np.uint8(bit_is_set) ^ cwkvq__ffni) & kBitmask[
        i % 8]
    setitem_str_bitmap(bits, fpxu__plc, cwkvq__ffni)


@numba.njit
def get_bit_bitmap(bits, i):
    return getitem_str_bitmap(bits, i >> 3) >> (i & 7) & 1


@numba.njit
def copy_nulls_range(out_str_arr, in_str_arr, out_start):
    owi__dmqlq = get_null_bitmap_ptr(out_str_arr)
    mcnc__jfgf = get_null_bitmap_ptr(in_str_arr)
    for j in range(len(in_str_arr)):
        nqj__louf = get_bit_bitmap(mcnc__jfgf, j)
        set_bit_to(owi__dmqlq, out_start + j, nqj__louf)


@intrinsic
def set_string_array_range(typingctx, out_typ, in_typ, curr_str_typ,
    curr_chars_typ=None):
    assert out_typ == string_array_type and in_typ == string_array_type or out_typ == binary_array_type and in_typ == binary_array_type, 'set_string_array_range requires string or binary arrays'
    assert isinstance(curr_str_typ, types.Integer) and isinstance(
        curr_chars_typ, types.Integer
        ), 'set_string_array_range requires integer indices'

    def codegen(context, builder, sig, args):
        out_arr, vwc__vibae, rdy__buk, luahi__qycp = args
        caqz__uqkd = _get_str_binary_arr_payload(context, builder,
            vwc__vibae, string_array_type)
        uswa__vgkxb = _get_str_binary_arr_payload(context, builder, out_arr,
            string_array_type)
        ngqu__btzvx = context.make_helper(builder, offset_arr_type,
            caqz__uqkd.offsets).data
        ubv__ocb = context.make_helper(builder, offset_arr_type,
            uswa__vgkxb.offsets).data
        loxp__ybr = context.make_helper(builder, char_arr_type, caqz__uqkd.data
            ).data
        nhkgn__ygee = context.make_helper(builder, char_arr_type,
            uswa__vgkxb.data).data
        num_total_chars = _get_num_total_chars(builder, ngqu__btzvx,
            caqz__uqkd.n_arrays)
        mvf__soy = lir.FunctionType(lir.VoidType(), [lir.IntType(
            offset_type.bitwidth).as_pointer(), lir.IntType(8).as_pointer(),
            lir.IntType(offset_type.bitwidth).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(64), lir.IntType(64), lir.IntType(64),
            lir.IntType(64)])
        humsj__tsxay = cgutils.get_or_insert_function(builder.module,
            mvf__soy, name='set_string_array_range')
        builder.call(humsj__tsxay, [ubv__ocb, nhkgn__ygee, ngqu__btzvx,
            loxp__ybr, rdy__buk, luahi__qycp, caqz__uqkd.n_arrays,
            num_total_chars])
        hbtuv__zcym = context.typing_context.resolve_value_type(
            copy_nulls_range)
        jleu__wnkwa = hbtuv__zcym.get_call_type(context.typing_context, (
            string_array_type, string_array_type, types.int64), {})
        hzygf__cshur = context.get_function(hbtuv__zcym, jleu__wnkwa)
        hzygf__cshur(builder, (out_arr, vwc__vibae, rdy__buk))
        return context.get_dummy_value()
    sig = types.void(out_typ, in_typ, types.intp, types.intp)
    return sig, codegen


@box(BinaryArrayType)
@box(StringArrayType)
def box_str_arr(typ, val, c):
    assert typ in [binary_array_type, string_array_type]
    khjig__kbyv = c.context.make_helper(c.builder, typ, val)
    abpkm__obvfq = ArrayItemArrayType(char_arr_type)
    pyxfu__uqgwe = _get_array_item_arr_payload(c.context, c.builder,
        abpkm__obvfq, khjig__kbyv.data)
    rgng__yvje = c.context.get_constant(types.int32, int(typ ==
        binary_array_type))
    euefk__mdpz = 'np_array_from_string_array'
    if use_pd_string_array and typ != binary_array_type:
        euefk__mdpz = 'pd_array_from_string_array'
    mvf__soy = lir.FunctionType(c.context.get_argument_type(types.pyobject),
        [lir.IntType(64), lir.IntType(offset_type.bitwidth).as_pointer(),
        lir.IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
        IntType(32)])
    ikkty__bvn = cgutils.get_or_insert_function(c.builder.module, mvf__soy,
        name=euefk__mdpz)
    iyz__vrw = c.context.make_array(offset_arr_type)(c.context, c.builder,
        pyxfu__uqgwe.offsets).data
    sgq__mqt = c.context.make_array(char_arr_type)(c.context, c.builder,
        pyxfu__uqgwe.data).data
    xczo__fgrcr = c.context.make_array(null_bitmap_arr_type)(c.context, c.
        builder, pyxfu__uqgwe.null_bitmap).data
    arr = c.builder.call(ikkty__bvn, [pyxfu__uqgwe.n_arrays, iyz__vrw,
        sgq__mqt, xczo__fgrcr, rgng__yvje])
    c.context.nrt.decref(c.builder, typ, val)
    return arr


@intrinsic
def str_arr_is_na(typingctx, str_arr_typ, ind_typ=None):
    assert str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        in_str_arr, ind = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        xczo__fgrcr = context.make_array(null_bitmap_arr_type)(context,
            builder, pyxfu__uqgwe.null_bitmap).data
        efq__hjl = builder.lshr(ind, lir.Constant(lir.IntType(64), 3))
        tch__ncc = builder.urem(ind, lir.Constant(lir.IntType(64), 8))
        cwkvq__ffni = builder.load(builder.gep(xczo__fgrcr, [efq__hjl],
            inbounds=True))
        afldq__kcv = lir.ArrayType(lir.IntType(8), 8)
        jebth__ovzqf = cgutils.alloca_once_value(builder, lir.Constant(
            afldq__kcv, (1, 2, 4, 8, 16, 32, 64, 128)))
        ogfik__jpl = builder.load(builder.gep(jebth__ovzqf, [lir.Constant(
            lir.IntType(64), 0), tch__ncc], inbounds=True))
        return builder.icmp_unsigned('==', builder.and_(cwkvq__ffni,
            ogfik__jpl), lir.Constant(lir.IntType(8), 0))
    return types.bool_(str_arr_typ, types.intp), codegen


@intrinsic
def str_arr_set_na(typingctx, str_arr_typ, ind_typ=None):
    assert str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        in_str_arr, ind = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        efq__hjl = builder.lshr(ind, lir.Constant(lir.IntType(64), 3))
        tch__ncc = builder.urem(ind, lir.Constant(lir.IntType(64), 8))
        xczo__fgrcr = context.make_array(null_bitmap_arr_type)(context,
            builder, pyxfu__uqgwe.null_bitmap).data
        offsets = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets).data
        akdn__mduf = builder.gep(xczo__fgrcr, [efq__hjl], inbounds=True)
        cwkvq__ffni = builder.load(akdn__mduf)
        afldq__kcv = lir.ArrayType(lir.IntType(8), 8)
        jebth__ovzqf = cgutils.alloca_once_value(builder, lir.Constant(
            afldq__kcv, (1, 2, 4, 8, 16, 32, 64, 128)))
        ogfik__jpl = builder.load(builder.gep(jebth__ovzqf, [lir.Constant(
            lir.IntType(64), 0), tch__ncc], inbounds=True))
        ogfik__jpl = builder.xor(ogfik__jpl, lir.Constant(lir.IntType(8), -1))
        builder.store(builder.and_(cwkvq__ffni, ogfik__jpl), akdn__mduf)
        if str_arr_typ == string_array_type:
            pwnve__kar = builder.add(ind, lir.Constant(lir.IntType(64), 1))
            mzzw__hqnv = builder.icmp_unsigned('!=', pwnve__kar,
                pyxfu__uqgwe.n_arrays)
            with builder.if_then(mzzw__hqnv):
                builder.store(builder.load(builder.gep(offsets, [ind])),
                    builder.gep(offsets, [pwnve__kar]))
        return context.get_dummy_value()
    return types.void(str_arr_typ, types.intp), codegen


@intrinsic
def str_arr_set_not_na(typingctx, str_arr_typ, ind_typ=None):
    assert str_arr_typ == string_array_type

    def codegen(context, builder, sig, args):
        in_str_arr, ind = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        efq__hjl = builder.lshr(ind, lir.Constant(lir.IntType(64), 3))
        tch__ncc = builder.urem(ind, lir.Constant(lir.IntType(64), 8))
        xczo__fgrcr = context.make_array(null_bitmap_arr_type)(context,
            builder, pyxfu__uqgwe.null_bitmap).data
        akdn__mduf = builder.gep(xczo__fgrcr, [efq__hjl], inbounds=True)
        cwkvq__ffni = builder.load(akdn__mduf)
        afldq__kcv = lir.ArrayType(lir.IntType(8), 8)
        jebth__ovzqf = cgutils.alloca_once_value(builder, lir.Constant(
            afldq__kcv, (1, 2, 4, 8, 16, 32, 64, 128)))
        ogfik__jpl = builder.load(builder.gep(jebth__ovzqf, [lir.Constant(
            lir.IntType(64), 0), tch__ncc], inbounds=True))
        builder.store(builder.or_(cwkvq__ffni, ogfik__jpl), akdn__mduf)
        return context.get_dummy_value()
    return types.void(str_arr_typ, types.intp), codegen


@intrinsic
def set_null_bits_to_value(typingctx, arr_typ, value_typ=None):
    assert (arr_typ == string_array_type or arr_typ == binary_array_type
        ) and is_overload_constant_int(value_typ)

    def codegen(context, builder, sig, args):
        in_str_arr, value = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder,
            in_str_arr, string_array_type)
        rtl__jhua = builder.udiv(builder.add(pyxfu__uqgwe.n_arrays, lir.
            Constant(lir.IntType(64), 7)), lir.Constant(lir.IntType(64), 8))
        xczo__fgrcr = context.make_array(null_bitmap_arr_type)(context,
            builder, pyxfu__uqgwe.null_bitmap).data
        cgutils.memset(builder, xczo__fgrcr, rtl__jhua, value)
        return context.get_dummy_value()
    return types.none(arr_typ, types.int8), codegen


def _get_str_binary_arr_data_payload_ptr(context, builder, str_arr):
    ygeno__funap = context.make_helper(builder, string_array_type, str_arr)
    abpkm__obvfq = ArrayItemArrayType(char_arr_type)
    iuarp__jikoa = context.make_helper(builder, abpkm__obvfq, ygeno__funap.data
        )
    hjb__faszl = ArrayItemArrayPayloadType(abpkm__obvfq)
    gayn__rjrw = context.nrt.meminfo_data(builder, iuarp__jikoa.meminfo)
    zwyt__lze = builder.bitcast(gayn__rjrw, context.get_value_type(
        hjb__faszl).as_pointer())
    return zwyt__lze


@intrinsic
def move_str_binary_arr_payload(typingctx, to_arr_typ, from_arr_typ=None):
    assert to_arr_typ == string_array_type and from_arr_typ == string_array_type or to_arr_typ == binary_array_type and from_arr_typ == binary_array_type

    def codegen(context, builder, sig, args):
        vfotw__oxedh, lgzn__sechw = args
        nzckb__nsyve = _get_str_binary_arr_data_payload_ptr(context,
            builder, lgzn__sechw)
        ekg__rxyep = _get_str_binary_arr_data_payload_ptr(context, builder,
            vfotw__oxedh)
        qhjq__udnqf = _get_str_binary_arr_payload(context, builder,
            lgzn__sechw, sig.args[1])
        fne__rdrew = _get_str_binary_arr_payload(context, builder,
            vfotw__oxedh, sig.args[0])
        context.nrt.incref(builder, char_arr_type, qhjq__udnqf.data)
        context.nrt.incref(builder, offset_arr_type, qhjq__udnqf.offsets)
        context.nrt.incref(builder, null_bitmap_arr_type, qhjq__udnqf.
            null_bitmap)
        context.nrt.decref(builder, char_arr_type, fne__rdrew.data)
        context.nrt.decref(builder, offset_arr_type, fne__rdrew.offsets)
        context.nrt.decref(builder, null_bitmap_arr_type, fne__rdrew.
            null_bitmap)
        builder.store(builder.load(nzckb__nsyve), ekg__rxyep)
        return context.get_dummy_value()
    return types.none(to_arr_typ, from_arr_typ), codegen


dummy_use = numba.njit(lambda a: None)


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def get_utf8_size(s):
    if isinstance(s, types.StringLiteral):
        l = len(s.literal_value.encode())
        return lambda s: l

    def impl(s):
        if s is None:
            return 0
        s = bodo.utils.indexing.unoptional(s)
        if s._is_ascii == 1:
            return len(s)
        yvi__ercj = _get_utf8_size(s._data, s._length, s._kind)
        dummy_use(s)
        return yvi__ercj
    return impl


@intrinsic
def setitem_str_arr_ptr(typingctx, str_arr_t, ind_t, ptr_t, len_t=None):

    def codegen(context, builder, sig, args):
        arr, ind, lva__unbu, rqz__srvw = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder, arr,
            sig.args[0])
        offsets = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets).data
        data = context.make_helper(builder, char_arr_type, pyxfu__uqgwe.data
            ).data
        mvf__soy = lir.FunctionType(lir.VoidType(), [lir.IntType(
            offset_type.bitwidth).as_pointer(), lir.IntType(8).as_pointer(),
            lir.IntType(64), lir.IntType(8).as_pointer(), lir.IntType(64),
            lir.IntType(32), lir.IntType(32), lir.IntType(64)])
        sjrf__cfnkx = cgutils.get_or_insert_function(builder.module,
            mvf__soy, name='setitem_string_array')
        cgbzp__tfa = context.get_constant(types.int32, -1)
        wmldh__wmoq = context.get_constant(types.int32, 1)
        num_total_chars = _get_num_total_chars(builder, offsets,
            pyxfu__uqgwe.n_arrays)
        builder.call(sjrf__cfnkx, [offsets, data, num_total_chars, builder.
            extract_value(lva__unbu, 0), rqz__srvw, cgbzp__tfa, wmldh__wmoq,
            ind])
        return context.get_dummy_value()
    return types.void(str_arr_t, ind_t, ptr_t, len_t), codegen


def lower_is_na(context, builder, bull_bitmap, ind):
    mvf__soy = lir.FunctionType(lir.IntType(1), [lir.IntType(8).as_pointer(
        ), lir.IntType(64)])
    jzgp__tcea = cgutils.get_or_insert_function(builder.module, mvf__soy,
        name='is_na')
    return builder.call(jzgp__tcea, [bull_bitmap, ind])


@intrinsic
def _memcpy(typingctx, dest_t, src_t, count_t, item_size_t=None):

    def codegen(context, builder, sig, args):
        lxgq__yjgvf, swux__fjqay, ewi__khn, psbvd__vcfxs = args
        cgutils.raw_memcpy(builder, lxgq__yjgvf, swux__fjqay, ewi__khn,
            psbvd__vcfxs)
        return context.get_dummy_value()
    return types.void(types.voidptr, types.voidptr, types.intp, types.intp
        ), codegen


@numba.njit
def print_str_arr(arr):
    _print_str_arr(num_strings(arr), num_total_chars(arr), get_offset_ptr(
        arr), get_data_ptr(arr))


def inplace_eq(A, i, val):
    return A[i] == val


@overload(inplace_eq)
def inplace_eq_overload(A, ind, val):

    def impl(A, ind, val):
        rggt__msx, ymwqx__lpy = unicode_to_utf8_and_len(val)
        fjkz__uim = getitem_str_offset(A, ind)
        rxhac__aovtw = getitem_str_offset(A, ind + 1)
        swx__mvi = rxhac__aovtw - fjkz__uim
        if swx__mvi != ymwqx__lpy:
            return False
        lva__unbu = get_data_ptr_ind(A, fjkz__uim)
        return memcmp(lva__unbu, rggt__msx, ymwqx__lpy) == 0
    return impl


def str_arr_setitem_int_to_str(A, ind, value):
    A[ind] = str(value)


@overload(str_arr_setitem_int_to_str)
def overload_str_arr_setitem_int_to_str(A, ind, val):

    def impl(A, ind, val):
        fjkz__uim = getitem_str_offset(A, ind)
        swx__mvi = bodo.libs.str_ext.int_to_str_len(val)
        iwrdv__vyfzx = fjkz__uim + swx__mvi
        bodo.libs.array_item_arr_ext.ensure_data_capacity(A._data,
            fjkz__uim, iwrdv__vyfzx)
        lva__unbu = get_data_ptr_ind(A, fjkz__uim)
        inplace_int64_to_str(lva__unbu, swx__mvi, val)
        setitem_str_offset(A, ind + 1, fjkz__uim + swx__mvi)
        str_arr_set_not_na(A, ind)
    return impl


@intrinsic
def inplace_set_NA_str(typingctx, ptr_typ=None):

    def codegen(context, builder, sig, args):
        lva__unbu, = args
        plxf__rxng = context.insert_const_string(builder.module, '<NA>')
        uoryq__uih = lir.Constant(lir.IntType(64), len('<NA>'))
        cgutils.raw_memcpy(builder, lva__unbu, plxf__rxng, uoryq__uih, 1)
    return types.none(types.voidptr), codegen


def str_arr_setitem_NA_str(A, ind):
    A[ind] = '<NA>'


@overload(str_arr_setitem_NA_str)
def overload_str_arr_setitem_NA_str(A, ind):
    gjwz__qpl = len('<NA>')

    def impl(A, ind):
        fjkz__uim = getitem_str_offset(A, ind)
        iwrdv__vyfzx = fjkz__uim + gjwz__qpl
        bodo.libs.array_item_arr_ext.ensure_data_capacity(A._data,
            fjkz__uim, iwrdv__vyfzx)
        lva__unbu = get_data_ptr_ind(A, fjkz__uim)
        inplace_set_NA_str(lva__unbu)
        setitem_str_offset(A, ind + 1, fjkz__uim + gjwz__qpl)
        str_arr_set_not_na(A, ind)
    return impl


@overload(operator.getitem, no_unliteral=True)
def str_arr_getitem_int(A, ind):
    if A != string_array_type:
        return
    if isinstance(ind, types.Integer):

        def str_arr_getitem_impl(A, ind):
            if ind < 0:
                ind += A.size
            fjkz__uim = getitem_str_offset(A, ind)
            rxhac__aovtw = getitem_str_offset(A, ind + 1)
            rqz__srvw = rxhac__aovtw - fjkz__uim
            lva__unbu = get_data_ptr_ind(A, fjkz__uim)
            nws__czgaj = decode_utf8(lva__unbu, rqz__srvw)
            return nws__czgaj
        return str_arr_getitem_impl
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:

        def bool_impl(A, ind):
            ind = bodo.utils.conversion.coerce_to_ndarray(ind)
            yvi__ercj = len(A)
            n_strs = 0
            n_chars = 0
            for i in range(yvi__ercj):
                if not bodo.libs.array_kernels.isna(ind, i) and ind[i]:
                    n_strs += 1
                    n_chars += get_str_arr_item_length(A, i)
            out_arr = pre_alloc_string_array(n_strs, n_chars)
            vut__bir = get_data_ptr(out_arr).data
            tbk__cbugc = get_data_ptr(A).data
            pyg__ylub = 0
            coe__xfgg = 0
            setitem_str_offset(out_arr, 0, 0)
            for i in range(yvi__ercj):
                if not bodo.libs.array_kernels.isna(ind, i) and ind[i]:
                    fag__zghvy = get_str_arr_item_length(A, i)
                    if fag__zghvy == 1:
                        copy_single_char(vut__bir, coe__xfgg, tbk__cbugc,
                            getitem_str_offset(A, i))
                    else:
                        memcpy_region(vut__bir, coe__xfgg, tbk__cbugc,
                            getitem_str_offset(A, i), fag__zghvy, 1)
                    coe__xfgg += fag__zghvy
                    setitem_str_offset(out_arr, pyg__ylub + 1, coe__xfgg)
                    if str_arr_is_na(A, i):
                        str_arr_set_na(out_arr, pyg__ylub)
                    else:
                        str_arr_set_not_na(out_arr, pyg__ylub)
                    pyg__ylub += 1
            return out_arr
        return bool_impl
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):

        def str_arr_arr_impl(A, ind):
            yvi__ercj = len(ind)
            out_arr = pre_alloc_string_array(yvi__ercj, -1)
            pyg__ylub = 0
            for i in range(yvi__ercj):
                cid__plxm = A[ind[i]]
                out_arr[pyg__ylub] = cid__plxm
                if str_arr_is_na(A, ind[i]):
                    str_arr_set_na(out_arr, pyg__ylub)
                pyg__ylub += 1
            return out_arr
        return str_arr_arr_impl
    if isinstance(ind, types.SliceType):

        def str_arr_slice_impl(A, ind):
            yvi__ercj = len(A)
            oqbo__aynw = numba.cpython.unicode._normalize_slice(ind, yvi__ercj)
            qpeoe__tfd = numba.cpython.unicode._slice_span(oqbo__aynw)
            if oqbo__aynw.step == 1:
                fjkz__uim = getitem_str_offset(A, oqbo__aynw.start)
                rxhac__aovtw = getitem_str_offset(A, oqbo__aynw.stop)
                n_chars = rxhac__aovtw - fjkz__uim
                kxd__aowu = pre_alloc_string_array(qpeoe__tfd, np.int64(
                    n_chars))
                for i in range(qpeoe__tfd):
                    kxd__aowu[i] = A[oqbo__aynw.start + i]
                    if str_arr_is_na(A, oqbo__aynw.start + i):
                        str_arr_set_na(kxd__aowu, i)
                return kxd__aowu
            else:
                kxd__aowu = pre_alloc_string_array(qpeoe__tfd, -1)
                for i in range(qpeoe__tfd):
                    kxd__aowu[i] = A[oqbo__aynw.start + i * oqbo__aynw.step]
                    if str_arr_is_na(A, oqbo__aynw.start + i * oqbo__aynw.step
                        ):
                        str_arr_set_na(kxd__aowu, i)
                return kxd__aowu
        return str_arr_slice_impl
    raise BodoError(
        f'getitem for StringArray with indexing type {ind} not supported.')


dummy_use = numba.njit(lambda a: None)


@overload(operator.setitem)
def str_arr_setitem(A, idx, val):
    if A != string_array_type:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    ogsg__dsa = (
        f'StringArray setitem with index {idx} and value {val} not supported yet.'
        )
    if isinstance(idx, types.Integer):
        if val != string_type:
            raise BodoError(ogsg__dsa)
        rhv__hgbj = 4

        def impl_scalar(A, idx, val):
            uxob__azcv = (val._length if val._is_ascii else rhv__hgbj * val
                ._length)
            rtuv__uwvh = A._data
            fjkz__uim = np.int64(getitem_str_offset(A, idx))
            iwrdv__vyfzx = fjkz__uim + uxob__azcv
            bodo.libs.array_item_arr_ext.ensure_data_capacity(rtuv__uwvh,
                fjkz__uim, iwrdv__vyfzx)
            setitem_string_array(get_offset_ptr(A), get_data_ptr(A),
                iwrdv__vyfzx, val._data, val._length, val._kind, val.
                _is_ascii, idx)
            str_arr_set_not_na(A, idx)
            dummy_use(A)
            dummy_use(val)
        return impl_scalar
    if isinstance(idx, types.SliceType):
        if val == string_array_type:

            def impl_slice(A, idx, val):
                oqbo__aynw = numba.cpython.unicode._normalize_slice(idx, len(A)
                    )
                wpoj__ilwx = oqbo__aynw.start
                rtuv__uwvh = A._data
                fjkz__uim = np.int64(getitem_str_offset(A, wpoj__ilwx))
                iwrdv__vyfzx = fjkz__uim + np.int64(num_total_chars(val))
                bodo.libs.array_item_arr_ext.ensure_data_capacity(rtuv__uwvh,
                    fjkz__uim, iwrdv__vyfzx)
                set_string_array_range(A, val, wpoj__ilwx, fjkz__uim)
                juoel__lyrt = 0
                for i in range(oqbo__aynw.start, oqbo__aynw.stop,
                    oqbo__aynw.step):
                    if str_arr_is_na(val, juoel__lyrt):
                        str_arr_set_na(A, i)
                    else:
                        str_arr_set_not_na(A, i)
                    juoel__lyrt += 1
            return impl_slice
        elif isinstance(val, types.List) and val.dtype == string_type:

            def impl_slice_list(A, idx, val):
                fssr__iakio = str_list_to_array(val)
                A[idx] = fssr__iakio
            return impl_slice_list
        elif val == string_type:

            def impl_slice(A, idx, val):
                oqbo__aynw = numba.cpython.unicode._normalize_slice(idx, len(A)
                    )
                for i in range(oqbo__aynw.start, oqbo__aynw.stop,
                    oqbo__aynw.step):
                    A[i] = val
            return impl_slice
        else:
            raise BodoError(ogsg__dsa)
    if is_list_like_index_type(idx) and idx.dtype == types.bool_:
        if val == string_type:

            def impl_bool_scalar(A, idx, val):
                yvi__ercj = len(A)
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                out_arr = pre_alloc_string_array(yvi__ercj, -1)
                for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                    if not bodo.libs.array_kernels.isna(idx, i) and idx[i]:
                        out_arr[i] = val
                    elif bodo.libs.array_kernels.isna(A, i):
                        out_arr[i] = ''
                        str_arr_set_na(out_arr, i)
                    else:
                        get_str_arr_item_copy(out_arr, i, A, i)
                move_str_binary_arr_payload(A, out_arr)
            return impl_bool_scalar
        elif val == string_array_type or isinstance(val, types.Array
            ) and isinstance(val.dtype, types.UnicodeCharSeq):

            def impl_bool_arr(A, idx, val):
                yvi__ercj = len(A)
                idx = bodo.utils.conversion.coerce_to_array(idx,
                    use_nullable_array=True)
                out_arr = pre_alloc_string_array(yvi__ercj, -1)
                yor__zcq = 0
                for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                    if not bodo.libs.array_kernels.isna(idx, i) and idx[i]:
                        if bodo.libs.array_kernels.isna(val, yor__zcq):
                            out_arr[i] = ''
                            str_arr_set_na(out_arr, yor__zcq)
                        else:
                            out_arr[i] = str(val[yor__zcq])
                        yor__zcq += 1
                    elif bodo.libs.array_kernels.isna(A, i):
                        out_arr[i] = ''
                        str_arr_set_na(out_arr, i)
                    else:
                        get_str_arr_item_copy(out_arr, i, A, i)
                move_str_binary_arr_payload(A, out_arr)
            return impl_bool_arr
        else:
            raise BodoError(ogsg__dsa)
    raise BodoError(ogsg__dsa)


@overload_attribute(StringArrayType, 'dtype')
def overload_str_arr_dtype(A):
    return lambda A: pd.StringDtype()


@overload_attribute(StringArrayType, 'ndim')
def overload_str_arr_ndim(A):
    return lambda A: 1


@overload_method(StringArrayType, 'astype', no_unliteral=True)
def overload_str_arr_astype(A, dtype, copy=True):
    if dtype == types.unicode_type:
        raise_bodo_error(
            "StringArray.astype(): 'dtype' when passed as string must be a constant value"
            )
    if isinstance(dtype, types.Function) and dtype.key[0] == str:
        return lambda A, dtype, copy=True: A
    pdvzl__fhdcu = parse_dtype(dtype, 'StringArray.astype')
    if not isinstance(pdvzl__fhdcu, (types.Float, types.Integer)):
        raise BodoError('invalid dtype in StringArray.astype()')
    if isinstance(pdvzl__fhdcu, types.Float):

        def impl_float(A, dtype, copy=True):
            numba.parfors.parfor.init_prange()
            yvi__ercj = len(A)
            B = np.empty(yvi__ercj, pdvzl__fhdcu)
            for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                if bodo.libs.array_kernels.isna(A, i):
                    B[i] = np.nan
                else:
                    B[i] = float(A[i])
            return B
        return impl_float
    else:

        def impl_int(A, dtype, copy=True):
            numba.parfors.parfor.init_prange()
            yvi__ercj = len(A)
            B = np.empty(yvi__ercj, pdvzl__fhdcu)
            for i in numba.parfors.parfor.internal_prange(yvi__ercj):
                B[i] = int(A[i])
            return B
        return impl_int


@intrinsic
def decode_utf8(typingctx, ptr_t, len_t=None):

    def codegen(context, builder, sig, args):
        lva__unbu, rqz__srvw = args
        qrkx__gqjag = context.get_python_api(builder)
        jydb__qyeiv = qrkx__gqjag.string_from_string_and_size(lva__unbu,
            rqz__srvw)
        wdj__zblfp = qrkx__gqjag.to_native_value(string_type, jydb__qyeiv
            ).value
        adby__pct = cgutils.create_struct_proxy(string_type)(context,
            builder, wdj__zblfp)
        adby__pct.hash = adby__pct.hash.type(-1)
        qrkx__gqjag.decref(jydb__qyeiv)
        return adby__pct._getvalue()
    return string_type(types.voidptr, types.intp), codegen


def get_arr_data_ptr(arr, ind):
    return arr


@overload(get_arr_data_ptr, no_unliteral=True)
def overload_get_arr_data_ptr(arr, ind):
    assert isinstance(types.unliteral(ind), types.Integer)
    if isinstance(arr, bodo.libs.int_arr_ext.IntegerArrayType):

        def impl_int(arr, ind):
            return bodo.hiframes.split_impl.get_c_arr_ptr(arr._data.ctypes, ind
                )
        return impl_int
    assert isinstance(arr, types.Array)

    def impl_np(arr, ind):
        return bodo.hiframes.split_impl.get_c_arr_ptr(arr.ctypes, ind)
    return impl_np


def set_to_numeric_out_na_err(out_arr, out_ind, err_code):
    pass


@overload(set_to_numeric_out_na_err)
def set_to_numeric_out_na_err_overload(out_arr, out_ind, err_code):
    if isinstance(out_arr, bodo.libs.int_arr_ext.IntegerArrayType):

        def impl_int(out_arr, out_ind, err_code):
            bodo.libs.int_arr_ext.set_bit_to_arr(out_arr._null_bitmap,
                out_ind, 0 if err_code == -1 else 1)
        return impl_int
    assert isinstance(out_arr, types.Array)
    if isinstance(out_arr.dtype, types.Float):

        def impl_np(out_arr, out_ind, err_code):
            if err_code == -1:
                out_arr[out_ind] = np.nan
        return impl_np
    return lambda out_arr, out_ind, err_code: None


@numba.njit(no_cpython_wrapper=True)
def str_arr_item_to_numeric(out_arr, out_ind, str_arr, ind):
    err_code = _str_arr_item_to_numeric(get_arr_data_ptr(out_arr, out_ind),
        str_arr, ind, out_arr.dtype)
    set_to_numeric_out_na_err(out_arr, out_ind, err_code)


@intrinsic
def _str_arr_item_to_numeric(typingctx, out_ptr_t, str_arr_t, ind_t,
    out_dtype_t=None):
    assert str_arr_t == string_array_type
    assert ind_t == types.int64

    def codegen(context, builder, sig, args):
        ivll__wdfbi, arr, ind, yzgnf__mpbn = args
        pyxfu__uqgwe = _get_str_binary_arr_payload(context, builder, arr,
            string_array_type)
        offsets = context.make_helper(builder, offset_arr_type,
            pyxfu__uqgwe.offsets).data
        data = context.make_helper(builder, char_arr_type, pyxfu__uqgwe.data
            ).data
        mvf__soy = lir.FunctionType(lir.IntType(32), [ivll__wdfbi.type, lir
            .IntType(offset_type.bitwidth).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(64)])
        clyp__apq = 'str_arr_to_int64'
        if sig.args[3].dtype == types.float64:
            clyp__apq = 'str_arr_to_float64'
        else:
            assert sig.args[3].dtype == types.int64
        gxzf__rtrr = cgutils.get_or_insert_function(builder.module,
            mvf__soy, clyp__apq)
        return builder.call(gxzf__rtrr, [ivll__wdfbi, offsets, data, ind])
    return types.int32(out_ptr_t, string_array_type, types.int64, out_dtype_t
        ), codegen


@unbox(BinaryArrayType)
@unbox(StringArrayType)
def unbox_str_series(typ, val, c):
    rgng__yvje = c.context.get_constant(types.int32, int(typ ==
        binary_array_type))
    mvf__soy = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.IntType(8
        ).as_pointer(), lir.IntType(32)])
    cvaaj__kwam = cgutils.get_or_insert_function(c.builder.module, mvf__soy,
        name='string_array_from_sequence')
    vnav__fhobo = c.builder.call(cvaaj__kwam, [val, rgng__yvje])
    abpkm__obvfq = ArrayItemArrayType(char_arr_type)
    iuarp__jikoa = c.context.make_helper(c.builder, abpkm__obvfq)
    iuarp__jikoa.meminfo = vnav__fhobo
    ygeno__funap = c.context.make_helper(c.builder, typ)
    rtuv__uwvh = iuarp__jikoa._getvalue()
    ygeno__funap.data = rtuv__uwvh
    vxkw__yhl = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(ygeno__funap._getvalue(), is_error=vxkw__yhl)


@lower_constant(BinaryArrayType)
@lower_constant(StringArrayType)
def lower_constant_str_arr(context, builder, typ, pyval):
    yvi__ercj = len(pyval)
    coe__xfgg = 0
    micqz__axzvk = np.empty(yvi__ercj + 1, np_offset_type)
    kghv__mtc = []
    xgzh__eteac = np.empty(yvi__ercj + 7 >> 3, np.uint8)
    for i, s in enumerate(pyval):
        micqz__axzvk[i] = coe__xfgg
        gsmb__hbwv = pd.isna(s)
        bodo.libs.int_arr_ext.set_bit_to_arr(xgzh__eteac, i, int(not
            gsmb__hbwv))
        if gsmb__hbwv:
            continue
        bggx__ocwuf = list(s.encode()) if isinstance(s, str) else list(s)
        kghv__mtc.extend(bggx__ocwuf)
        coe__xfgg += len(bggx__ocwuf)
    micqz__axzvk[yvi__ercj] = coe__xfgg
    srnde__mkmwi = np.array(kghv__mtc, np.uint8)
    rjjqn__hsrl = context.get_constant(types.int64, yvi__ercj)
    bssy__bvfaj = context.get_constant_generic(builder, char_arr_type,
        srnde__mkmwi)
    ers__doz = context.get_constant_generic(builder, offset_arr_type,
        micqz__axzvk)
    aowu__gyu = context.get_constant_generic(builder, null_bitmap_arr_type,
        xgzh__eteac)
    pyxfu__uqgwe = lir.Constant.literal_struct([rjjqn__hsrl, bssy__bvfaj,
        ers__doz, aowu__gyu])
    pyxfu__uqgwe = cgutils.global_constant(builder, '.const.payload',
        pyxfu__uqgwe).bitcast(cgutils.voidptr_t)
    seokz__blt = context.get_constant(types.int64, -1)
    xet__vch = context.get_constant_null(types.voidptr)
    btq__qrn = lir.Constant.literal_struct([seokz__blt, xet__vch, xet__vch,
        pyxfu__uqgwe, seokz__blt])
    btq__qrn = cgutils.global_constant(builder, '.const.meminfo', btq__qrn
        ).bitcast(cgutils.voidptr_t)
    rtuv__uwvh = lir.Constant.literal_struct([btq__qrn])
    ygeno__funap = lir.Constant.literal_struct([rtuv__uwvh])
    return ygeno__funap


def pre_alloc_str_arr_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


from numba.parfors.array_analysis import ArrayAnalysis
(ArrayAnalysis._analyze_op_call_bodo_libs_str_arr_ext_pre_alloc_string_array
    ) = pre_alloc_str_arr_equiv


@overload(glob.glob, no_unliteral=True)
def overload_glob_glob(pathname, recursive=False):

    def _glob_glob_impl(pathname, recursive=False):
        with numba.objmode(l='list_str_type'):
            l = glob.glob(pathname, recursive=recursive)
        return l
    return _glob_glob_impl
