"""Array implementation for structs of values.
Corresponds to Spark's StructType: https://spark.apache.org/docs/latest/sql-reference.html
Corresponds to Arrow's Struct arrays: https://arrow.apache.org/docs/format/Columnar.html

The values are stored in contiguous data arrays; one array per field. For example:
A:             ["AA", "B", "C"]
B:             [1, 2, 4]
"""
import operator
import llvmlite.binding as ll
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import NativeValue, box, intrinsic, models, overload, overload_attribute, overload_method, register_model, unbox
from numba.parfors.array_analysis import ArrayAnalysis
from numba.typed.typedobjectutils import _cast
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_type
from bodo.libs import array_ext
from bodo.utils.cg_helpers import gen_allocate_array, get_array_elem_counts, get_bitmap_bit, is_na_value, pyarray_setitem, seq_getitem, set_bitmap_bit, to_arr_obj_if_list_obj
from bodo.utils.typing import BodoError, dtype_to_array_type, get_overload_const_int, get_overload_const_str, is_list_like_index_type, is_overload_constant_int, is_overload_constant_str, is_overload_none
ll.add_symbol('struct_array_from_sequence', array_ext.
    struct_array_from_sequence)
ll.add_symbol('np_array_from_struct_array', array_ext.
    np_array_from_struct_array)


class StructArrayType(types.ArrayCompatible):

    def __init__(self, data, names=None):
        assert isinstance(data, tuple) and len(data) > 0 and all(bodo.utils
            .utils.is_array_typ(mbxz__qpba, False) for mbxz__qpba in data)
        if names is not None:
            assert isinstance(names, tuple) and all(isinstance(mbxz__qpba,
                str) for mbxz__qpba in names) and len(names) == len(data)
        else:
            names = tuple('f{}'.format(i) for i in range(len(data)))
        self.data = data
        self.names = names
        super(StructArrayType, self).__init__(name=
            'StructArrayType({}, {})'.format(data, names))

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return StructType(tuple(mfvvh__nwve.dtype for mfvvh__nwve in self.
            data), self.names)

    @classmethod
    def from_dict(cls, d):
        assert isinstance(d, dict)
        names = tuple(str(mbxz__qpba) for mbxz__qpba in d.keys())
        data = tuple(dtype_to_array_type(mfvvh__nwve) for mfvvh__nwve in d.
            values())
        return StructArrayType(data, names)

    def copy(self):
        return StructArrayType(self.data, self.names)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


class StructArrayPayloadType(types.Type):

    def __init__(self, data):
        assert isinstance(data, tuple) and all(bodo.utils.utils.
            is_array_typ(mbxz__qpba, False) for mbxz__qpba in data)
        self.data = data
        super(StructArrayPayloadType, self).__init__(name=
            'StructArrayPayloadType({})'.format(data))

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(StructArrayPayloadType)
class StructArrayPayloadModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pxcv__vpk = [('data', types.BaseTuple.from_types(fe_type.data)), (
            'null_bitmap', types.Array(types.uint8, 1, 'C'))]
        models.StructModel.__init__(self, dmm, fe_type, pxcv__vpk)


@register_model(StructArrayType)
class StructArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        payload_type = StructArrayPayloadType(fe_type.data)
        pxcv__vpk = [('meminfo', types.MemInfoPointer(payload_type))]
        models.StructModel.__init__(self, dmm, fe_type, pxcv__vpk)


def define_struct_arr_dtor(context, builder, struct_arr_type, payload_type):
    zajp__uoxfn = builder.module
    cpdcf__ocpy = lir.FunctionType(lir.VoidType(), [cgutils.voidptr_t])
    mrh__elscg = cgutils.get_or_insert_function(zajp__uoxfn, cpdcf__ocpy,
        name='.dtor.struct_arr.{}.{}.'.format(struct_arr_type.data,
        struct_arr_type.names))
    if not mrh__elscg.is_declaration:
        return mrh__elscg
    mrh__elscg.linkage = 'linkonce_odr'
    builder = lir.IRBuilder(mrh__elscg.append_basic_block())
    iekba__spr = mrh__elscg.args[0]
    sgouk__loo = context.get_value_type(payload_type).as_pointer()
    sum__lmjr = builder.bitcast(iekba__spr, sgouk__loo)
    xahpq__hsvqp = context.make_helper(builder, payload_type, ref=sum__lmjr)
    context.nrt.decref(builder, types.BaseTuple.from_types(struct_arr_type.
        data), xahpq__hsvqp.data)
    context.nrt.decref(builder, types.Array(types.uint8, 1, 'C'),
        xahpq__hsvqp.null_bitmap)
    builder.ret_void()
    return mrh__elscg


def construct_struct_array(context, builder, struct_arr_type, n_structs,
    n_elems, c=None):
    payload_type = StructArrayPayloadType(struct_arr_type.data)
    ipgtj__gmjad = context.get_value_type(payload_type)
    wtgk__ltu = context.get_abi_sizeof(ipgtj__gmjad)
    tnw__lsewj = define_struct_arr_dtor(context, builder, struct_arr_type,
        payload_type)
    otg__lnjey = context.nrt.meminfo_alloc_dtor(builder, context.
        get_constant(types.uintp, wtgk__ltu), tnw__lsewj)
    qbl__xbo = context.nrt.meminfo_data(builder, otg__lnjey)
    pxkc__etl = builder.bitcast(qbl__xbo, ipgtj__gmjad.as_pointer())
    xahpq__hsvqp = cgutils.create_struct_proxy(payload_type)(context, builder)
    toa__mcd = []
    nrh__ylafc = 0
    for arr_typ in struct_arr_type.data:
        wtc__dhs = bodo.utils.transform.get_type_alloc_counts(arr_typ.dtype)
        yiu__fzz = cgutils.pack_array(builder, [n_structs] + [builder.
            extract_value(n_elems, i) for i in range(nrh__ylafc, nrh__ylafc +
            wtc__dhs)])
        arr = gen_allocate_array(context, builder, arr_typ, yiu__fzz, c)
        toa__mcd.append(arr)
        nrh__ylafc += wtc__dhs
    xahpq__hsvqp.data = cgutils.pack_array(builder, toa__mcd
        ) if types.is_homogeneous(*struct_arr_type.data
        ) else cgutils.pack_struct(builder, toa__mcd)
    kwwef__cfjcu = builder.udiv(builder.add(n_structs, lir.Constant(lir.
        IntType(64), 7)), lir.Constant(lir.IntType(64), 8))
    hwr__rcc = bodo.utils.utils._empty_nd_impl(context, builder, types.
        Array(types.uint8, 1, 'C'), [kwwef__cfjcu])
    null_bitmap_ptr = hwr__rcc.data
    xahpq__hsvqp.null_bitmap = hwr__rcc._getvalue()
    builder.store(xahpq__hsvqp._getvalue(), pxkc__etl)
    return otg__lnjey, xahpq__hsvqp.data, null_bitmap_ptr


def _get_C_API_ptrs(c, data_tup, data_typ, names):
    phj__fzx = []
    assert len(data_typ) > 0
    for i, arr_typ in enumerate(data_typ):
        boxk__lxh = c.builder.extract_value(data_tup, i)
        arr = c.context.make_array(arr_typ)(c.context, c.builder, value=
            boxk__lxh)
        phj__fzx.append(arr.data)
    cunp__tuz = cgutils.pack_array(c.builder, phj__fzx
        ) if types.is_homogeneous(*data_typ) else cgutils.pack_struct(c.
        builder, phj__fzx)
    hcwy__kmihd = cgutils.alloca_once_value(c.builder, cunp__tuz)
    kugw__qdvi = [c.context.get_constant(types.int32, bodo.utils.utils.
        numba_to_c_type(mbxz__qpba.dtype)) for mbxz__qpba in data_typ]
    jzg__jhdh = cgutils.alloca_once_value(c.builder, cgutils.pack_array(c.
        builder, kugw__qdvi))
    bqiga__hbog = cgutils.pack_array(c.builder, [c.context.
        insert_const_string(c.builder.module, mbxz__qpba) for mbxz__qpba in
        names])
    azoi__bzly = cgutils.alloca_once_value(c.builder, bqiga__hbog)
    return hcwy__kmihd, jzg__jhdh, azoi__bzly


@unbox(StructArrayType)
def unbox_struct_array(typ, val, c, is_tuple_array=False):
    from bodo.libs.tuple_arr_ext import TupleArrayType
    n_structs = bodo.utils.utils.object_length(c, val)
    fkoh__hlt = all(isinstance(mfvvh__nwve, types.Array) and mfvvh__nwve.
        dtype in (types.int64, types.float64, types.bool_,
        datetime_date_type) for mfvvh__nwve in typ.data)
    if fkoh__hlt:
        n_elems = cgutils.pack_array(c.builder, [], lir.IntType(64))
    else:
        rcbz__evy = get_array_elem_counts(c, c.builder, c.context, val, 
            TupleArrayType(typ.data) if is_tuple_array else typ)
        n_elems = cgutils.pack_array(c.builder, [c.builder.extract_value(
            rcbz__evy, i) for i in range(1, rcbz__evy.type.count)], lir.
            IntType(64))
    otg__lnjey, data_tup, null_bitmap_ptr = construct_struct_array(c.
        context, c.builder, typ, n_structs, n_elems, c)
    if fkoh__hlt:
        hcwy__kmihd, jzg__jhdh, azoi__bzly = _get_C_API_ptrs(c, data_tup,
            typ.data, typ.names)
        cpdcf__ocpy = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer(), lir.IntType(32), lir.IntType(8).as_pointer(), lir
            .IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(8).as_pointer(), lir.IntType(1)])
        mrh__elscg = cgutils.get_or_insert_function(c.builder.module,
            cpdcf__ocpy, name='struct_array_from_sequence')
        c.builder.call(mrh__elscg, [val, c.context.get_constant(types.int32,
            len(typ.data)), c.builder.bitcast(hcwy__kmihd, lir.IntType(8).
            as_pointer()), null_bitmap_ptr, c.builder.bitcast(jzg__jhdh,
            lir.IntType(8).as_pointer()), c.builder.bitcast(azoi__bzly, lir
            .IntType(8).as_pointer()), c.context.get_constant(types.bool_,
            is_tuple_array)])
    else:
        _unbox_struct_array_generic(typ, val, c, n_structs, data_tup,
            null_bitmap_ptr, is_tuple_array)
    llz__skiu = c.context.make_helper(c.builder, typ)
    llz__skiu.meminfo = otg__lnjey
    csdrt__icpjf = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(llz__skiu._getvalue(), is_error=csdrt__icpjf)


def _unbox_struct_array_generic(typ, val, c, n_structs, data_tup,
    null_bitmap_ptr, is_tuple_array=False):
    context = c.context
    builder = c.builder
    oscu__ighn = context.insert_const_string(builder.module, 'pandas')
    bjjcm__oct = c.pyapi.import_module_noblock(oscu__ighn)
    csrb__jeln = c.pyapi.object_getattr_string(bjjcm__oct, 'NA')
    with cgutils.for_range(builder, n_structs) as loop:
        fpvpy__yrba = loop.index
        favc__ffjfa = seq_getitem(builder, context, val, fpvpy__yrba)
        set_bitmap_bit(builder, null_bitmap_ptr, fpvpy__yrba, 0)
        for qwzj__duqb in range(len(typ.data)):
            arr_typ = typ.data[qwzj__duqb]
            data_arr = builder.extract_value(data_tup, qwzj__duqb)

            def set_na(data_arr, i):
                bodo.libs.array_kernels.setna(data_arr, i)
            sig = types.none(arr_typ, types.int64)
            eevr__xxa, doi__oevac = c.pyapi.call_jit_code(set_na, sig, [
                data_arr, fpvpy__yrba])
        glgr__stkvx = is_na_value(builder, context, favc__ffjfa, csrb__jeln)
        nroli__fpb = builder.icmp_unsigned('!=', glgr__stkvx, lir.Constant(
            glgr__stkvx.type, 1))
        with builder.if_then(nroli__fpb):
            set_bitmap_bit(builder, null_bitmap_ptr, fpvpy__yrba, 1)
            for qwzj__duqb in range(len(typ.data)):
                arr_typ = typ.data[qwzj__duqb]
                if is_tuple_array:
                    zdo__tbbe = c.pyapi.tuple_getitem(favc__ffjfa, qwzj__duqb)
                else:
                    zdo__tbbe = c.pyapi.dict_getitem_string(favc__ffjfa,
                        typ.names[qwzj__duqb])
                glgr__stkvx = is_na_value(builder, context, zdo__tbbe,
                    csrb__jeln)
                nroli__fpb = builder.icmp_unsigned('!=', glgr__stkvx, lir.
                    Constant(glgr__stkvx.type, 1))
                with builder.if_then(nroli__fpb):
                    zdo__tbbe = to_arr_obj_if_list_obj(c, context, builder,
                        zdo__tbbe, arr_typ.dtype)
                    field_val = c.pyapi.to_native_value(arr_typ.dtype,
                        zdo__tbbe).value
                    data_arr = builder.extract_value(data_tup, qwzj__duqb)

                    def set_data(data_arr, i, field_val):
                        data_arr[i] = field_val
                    sig = types.none(arr_typ, types.int64, arr_typ.dtype)
                    eevr__xxa, doi__oevac = c.pyapi.call_jit_code(set_data,
                        sig, [data_arr, fpvpy__yrba, field_val])
                    c.context.nrt.decref(builder, arr_typ.dtype, field_val)
        c.pyapi.decref(favc__ffjfa)
    c.pyapi.decref(bjjcm__oct)
    c.pyapi.decref(csrb__jeln)


def _get_struct_arr_payload(context, builder, arr_typ, arr):
    llz__skiu = context.make_helper(builder, arr_typ, arr)
    payload_type = StructArrayPayloadType(arr_typ.data)
    qbl__xbo = context.nrt.meminfo_data(builder, llz__skiu.meminfo)
    pxkc__etl = builder.bitcast(qbl__xbo, context.get_value_type(
        payload_type).as_pointer())
    xahpq__hsvqp = cgutils.create_struct_proxy(payload_type)(context,
        builder, builder.load(pxkc__etl))
    return xahpq__hsvqp


@box(StructArrayType)
def box_struct_arr(typ, val, c, is_tuple_array=False):
    xahpq__hsvqp = _get_struct_arr_payload(c.context, c.builder, typ, val)
    eevr__xxa, length = c.pyapi.call_jit_code(lambda A: len(A), types.int64
        (typ), [val])
    null_bitmap_ptr = c.context.make_helper(c.builder, types.Array(types.
        uint8, 1, 'C'), xahpq__hsvqp.null_bitmap).data
    fkoh__hlt = all(isinstance(mfvvh__nwve, types.Array) and mfvvh__nwve.
        dtype in (types.int64, types.float64, types.bool_,
        datetime_date_type) for mfvvh__nwve in typ.data)
    if fkoh__hlt:
        hcwy__kmihd, jzg__jhdh, azoi__bzly = _get_C_API_ptrs(c,
            xahpq__hsvqp.data, typ.data, typ.names)
        cpdcf__ocpy = lir.FunctionType(c.context.get_argument_type(types.
            pyobject), [lir.IntType(64), lir.IntType(32), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(1)])
        sgz__roca = cgutils.get_or_insert_function(c.builder.module,
            cpdcf__ocpy, name='np_array_from_struct_array')
        arr = c.builder.call(sgz__roca, [length, c.context.get_constant(
            types.int32, len(typ.data)), c.builder.bitcast(hcwy__kmihd, lir
            .IntType(8).as_pointer()), null_bitmap_ptr, c.builder.bitcast(
            jzg__jhdh, lir.IntType(8).as_pointer()), c.builder.bitcast(
            azoi__bzly, lir.IntType(8).as_pointer()), c.context.
            get_constant(types.bool_, is_tuple_array)])
    else:
        arr = _box_struct_array_generic(typ, c, length, xahpq__hsvqp.data,
            null_bitmap_ptr, is_tuple_array)
    c.context.nrt.decref(c.builder, typ, val)
    return arr


def _box_struct_array_generic(typ, c, length, data_arrs_tup,
    null_bitmap_ptr, is_tuple_array=False):
    context = c.context
    builder = c.builder
    oscu__ighn = context.insert_const_string(builder.module, 'numpy')
    qext__rnj = c.pyapi.import_module_noblock(oscu__ighn)
    vdx__sll = c.pyapi.object_getattr_string(qext__rnj, 'object_')
    uuqq__rxa = c.pyapi.long_from_longlong(length)
    hywz__ptd = c.pyapi.call_method(qext__rnj, 'ndarray', (uuqq__rxa, vdx__sll)
        )
    smzy__kburf = c.pyapi.object_getattr_string(qext__rnj, 'nan')
    with cgutils.for_range(builder, length) as loop:
        fpvpy__yrba = loop.index
        pyarray_setitem(builder, context, hywz__ptd, fpvpy__yrba, smzy__kburf)
        sjlia__yae = get_bitmap_bit(builder, null_bitmap_ptr, fpvpy__yrba)
        hsab__ufgd = builder.icmp_unsigned('!=', sjlia__yae, lir.Constant(
            lir.IntType(8), 0))
        with builder.if_then(hsab__ufgd):
            if is_tuple_array:
                favc__ffjfa = c.pyapi.tuple_new(len(typ.data))
            else:
                favc__ffjfa = c.pyapi.dict_new(len(typ.data))
            for i, arr_typ in enumerate(typ.data):
                if is_tuple_array:
                    c.pyapi.incref(smzy__kburf)
                    c.pyapi.tuple_setitem(favc__ffjfa, i, smzy__kburf)
                else:
                    c.pyapi.dict_setitem_string(favc__ffjfa, typ.names[i],
                        smzy__kburf)
                data_arr = c.builder.extract_value(data_arrs_tup, i)
                eevr__xxa, isdq__crjp = c.pyapi.call_jit_code(lambda
                    data_arr, ind: not bodo.libs.array_kernels.isna(
                    data_arr, ind), types.bool_(arr_typ, types.int64), [
                    data_arr, fpvpy__yrba])
                with builder.if_then(isdq__crjp):
                    eevr__xxa, field_val = c.pyapi.call_jit_code(lambda
                        data_arr, ind: data_arr[ind], arr_typ.dtype(arr_typ,
                        types.int64), [data_arr, fpvpy__yrba])
                    udxun__sdsx = c.pyapi.from_native_value(arr_typ.dtype,
                        field_val, c.env_manager)
                    if is_tuple_array:
                        c.pyapi.tuple_setitem(favc__ffjfa, i, udxun__sdsx)
                    else:
                        c.pyapi.dict_setitem_string(favc__ffjfa, typ.names[
                            i], udxun__sdsx)
                        c.pyapi.decref(udxun__sdsx)
            pyarray_setitem(builder, context, hywz__ptd, fpvpy__yrba,
                favc__ffjfa)
            c.pyapi.decref(favc__ffjfa)
    c.pyapi.decref(qext__rnj)
    c.pyapi.decref(vdx__sll)
    c.pyapi.decref(uuqq__rxa)
    c.pyapi.decref(smzy__kburf)
    return hywz__ptd


def _fix_nested_counts(nested_counts, struct_arr_type, nested_counts_type,
    builder):
    zzn__gbbfe = bodo.utils.transform.get_type_alloc_counts(struct_arr_type
        ) - 1
    if zzn__gbbfe == 0:
        return nested_counts
    if not isinstance(nested_counts_type, types.UniTuple):
        nested_counts = cgutils.pack_array(builder, [lir.Constant(lir.
            IntType(64), -1) for jut__yvqe in range(zzn__gbbfe)])
    elif nested_counts_type.count < zzn__gbbfe:
        nested_counts = cgutils.pack_array(builder, [builder.extract_value(
            nested_counts, i) for i in range(nested_counts_type.count)] + [
            lir.Constant(lir.IntType(64), -1) for jut__yvqe in range(
            zzn__gbbfe - nested_counts_type.count)])
    return nested_counts


@intrinsic
def pre_alloc_struct_array(typingctx, num_structs_typ, nested_counts_typ,
    dtypes_typ, names_typ=None):
    assert isinstance(num_structs_typ, types.Integer) and isinstance(dtypes_typ
        , types.BaseTuple)
    if is_overload_none(names_typ):
        names = tuple(f'f{i}' for i in range(len(dtypes_typ)))
    else:
        names = tuple(get_overload_const_str(mfvvh__nwve) for mfvvh__nwve in
            names_typ.types)
    xwc__dkxno = tuple(mfvvh__nwve.instance_type for mfvvh__nwve in
        dtypes_typ.types)
    struct_arr_type = StructArrayType(xwc__dkxno, names)

    def codegen(context, builder, sig, args):
        dqse__pzt, nested_counts, jut__yvqe, jut__yvqe = args
        nested_counts_type = sig.args[1]
        nested_counts = _fix_nested_counts(nested_counts, struct_arr_type,
            nested_counts_type, builder)
        otg__lnjey, jut__yvqe, jut__yvqe = construct_struct_array(context,
            builder, struct_arr_type, dqse__pzt, nested_counts)
        llz__skiu = context.make_helper(builder, struct_arr_type)
        llz__skiu.meminfo = otg__lnjey
        return llz__skiu._getvalue()
    return struct_arr_type(num_structs_typ, nested_counts_typ, dtypes_typ,
        names_typ), codegen


def pre_alloc_struct_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 4 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis._analyze_op_call_bodo_libs_struct_arr_ext_pre_alloc_struct_array
    ) = pre_alloc_struct_array_equiv


class StructType(types.Type):

    def __init__(self, data, names):
        assert isinstance(data, tuple) and len(data) > 0
        assert isinstance(names, tuple) and all(isinstance(mbxz__qpba, str) for
            mbxz__qpba in names) and len(names) == len(data)
        self.data = data
        self.names = names
        super(StructType, self).__init__(name='StructType({}, {})'.format(
            data, names))

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


class StructPayloadType(types.Type):

    def __init__(self, data):
        assert isinstance(data, tuple)
        self.data = data
        super(StructPayloadType, self).__init__(name=
            'StructPayloadType({})'.format(data))

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(StructPayloadType)
class StructPayloadModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pxcv__vpk = [('data', types.BaseTuple.from_types(fe_type.data)), (
            'null_bitmap', types.UniTuple(types.int8, len(fe_type.data)))]
        models.StructModel.__init__(self, dmm, fe_type, pxcv__vpk)


@register_model(StructType)
class StructModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        payload_type = StructPayloadType(fe_type.data)
        pxcv__vpk = [('meminfo', types.MemInfoPointer(payload_type))]
        models.StructModel.__init__(self, dmm, fe_type, pxcv__vpk)


def define_struct_dtor(context, builder, struct_type, payload_type):
    zajp__uoxfn = builder.module
    cpdcf__ocpy = lir.FunctionType(lir.VoidType(), [cgutils.voidptr_t])
    mrh__elscg = cgutils.get_or_insert_function(zajp__uoxfn, cpdcf__ocpy,
        name='.dtor.struct.{}.{}.'.format(struct_type.data, struct_type.names))
    if not mrh__elscg.is_declaration:
        return mrh__elscg
    mrh__elscg.linkage = 'linkonce_odr'
    builder = lir.IRBuilder(mrh__elscg.append_basic_block())
    iekba__spr = mrh__elscg.args[0]
    sgouk__loo = context.get_value_type(payload_type).as_pointer()
    sum__lmjr = builder.bitcast(iekba__spr, sgouk__loo)
    xahpq__hsvqp = context.make_helper(builder, payload_type, ref=sum__lmjr)
    for i in range(len(struct_type.data)):
        phl__lccma = builder.extract_value(xahpq__hsvqp.null_bitmap, i)
        hsab__ufgd = builder.icmp_unsigned('==', phl__lccma, lir.Constant(
            phl__lccma.type, 1))
        with builder.if_then(hsab__ufgd):
            val = builder.extract_value(xahpq__hsvqp.data, i)
            context.nrt.decref(builder, struct_type.data[i], val)
    builder.ret_void()
    return mrh__elscg


def _get_struct_payload(context, builder, typ, struct):
    struct = context.make_helper(builder, typ, struct)
    payload_type = StructPayloadType(typ.data)
    qbl__xbo = context.nrt.meminfo_data(builder, struct.meminfo)
    pxkc__etl = builder.bitcast(qbl__xbo, context.get_value_type(
        payload_type).as_pointer())
    xahpq__hsvqp = cgutils.create_struct_proxy(payload_type)(context,
        builder, builder.load(pxkc__etl))
    return xahpq__hsvqp, pxkc__etl


@unbox(StructType)
def unbox_struct(typ, val, c):
    context = c.context
    builder = c.builder
    oscu__ighn = context.insert_const_string(builder.module, 'pandas')
    bjjcm__oct = c.pyapi.import_module_noblock(oscu__ighn)
    csrb__jeln = c.pyapi.object_getattr_string(bjjcm__oct, 'NA')
    pryat__ofzss = []
    nulls = []
    for i, mfvvh__nwve in enumerate(typ.data):
        udxun__sdsx = c.pyapi.dict_getitem_string(val, typ.names[i])
        apkcy__emfm = cgutils.alloca_once_value(c.builder, context.
            get_constant(types.uint8, 0))
        vnseg__mzn = cgutils.alloca_once_value(c.builder, cgutils.
            get_null_value(context.get_value_type(mfvvh__nwve)))
        glgr__stkvx = is_na_value(builder, context, udxun__sdsx, csrb__jeln)
        hsab__ufgd = builder.icmp_unsigned('!=', glgr__stkvx, lir.Constant(
            glgr__stkvx.type, 1))
        with builder.if_then(hsab__ufgd):
            builder.store(context.get_constant(types.uint8, 1), apkcy__emfm)
            field_val = c.pyapi.to_native_value(mfvvh__nwve, udxun__sdsx).value
            builder.store(field_val, vnseg__mzn)
        pryat__ofzss.append(builder.load(vnseg__mzn))
        nulls.append(builder.load(apkcy__emfm))
    c.pyapi.decref(bjjcm__oct)
    c.pyapi.decref(csrb__jeln)
    otg__lnjey = construct_struct(context, builder, typ, pryat__ofzss, nulls)
    struct = context.make_helper(builder, typ)
    struct.meminfo = otg__lnjey
    csdrt__icpjf = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(struct._getvalue(), is_error=csdrt__icpjf)


@box(StructType)
def box_struct(typ, val, c):
    ocjrz__sxxs = c.pyapi.dict_new(len(typ.data))
    xahpq__hsvqp, jut__yvqe = _get_struct_payload(c.context, c.builder, typ,
        val)
    assert len(typ.data) > 0
    for i, val_typ in enumerate(typ.data):
        c.pyapi.dict_setitem_string(ocjrz__sxxs, typ.names[i], c.pyapi.
            borrow_none())
        phl__lccma = c.builder.extract_value(xahpq__hsvqp.null_bitmap, i)
        hsab__ufgd = c.builder.icmp_unsigned('==', phl__lccma, lir.Constant
            (phl__lccma.type, 1))
        with c.builder.if_then(hsab__ufgd):
            mglt__evon = c.builder.extract_value(xahpq__hsvqp.data, i)
            c.context.nrt.incref(c.builder, val_typ, mglt__evon)
            zdo__tbbe = c.pyapi.from_native_value(val_typ, mglt__evon, c.
                env_manager)
            c.pyapi.dict_setitem_string(ocjrz__sxxs, typ.names[i], zdo__tbbe)
            c.pyapi.decref(zdo__tbbe)
    c.context.nrt.decref(c.builder, typ, val)
    return ocjrz__sxxs


@intrinsic
def init_struct(typingctx, data_typ, names_typ=None):
    names = tuple(get_overload_const_str(mfvvh__nwve) for mfvvh__nwve in
        names_typ.types)
    struct_type = StructType(data_typ.types, names)

    def codegen(context, builder, sig, args):
        data, jclh__ihdca = args
        payload_type = StructPayloadType(struct_type.data)
        ipgtj__gmjad = context.get_value_type(payload_type)
        wtgk__ltu = context.get_abi_sizeof(ipgtj__gmjad)
        tnw__lsewj = define_struct_dtor(context, builder, struct_type,
            payload_type)
        otg__lnjey = context.nrt.meminfo_alloc_dtor(builder, context.
            get_constant(types.uintp, wtgk__ltu), tnw__lsewj)
        qbl__xbo = context.nrt.meminfo_data(builder, otg__lnjey)
        pxkc__etl = builder.bitcast(qbl__xbo, ipgtj__gmjad.as_pointer())
        xahpq__hsvqp = cgutils.create_struct_proxy(payload_type)(context,
            builder)
        xahpq__hsvqp.data = data
        xahpq__hsvqp.null_bitmap = cgutils.pack_array(builder, [context.
            get_constant(types.uint8, 1) for jut__yvqe in range(len(
            data_typ.types))])
        builder.store(xahpq__hsvqp._getvalue(), pxkc__etl)
        context.nrt.incref(builder, data_typ, data)
        struct = context.make_helper(builder, struct_type)
        struct.meminfo = otg__lnjey
        return struct._getvalue()
    return struct_type(data_typ, names_typ), codegen


@intrinsic
def get_struct_data(typingctx, struct_typ=None):
    assert isinstance(struct_typ, StructType)

    def codegen(context, builder, sig, args):
        struct, = args
        xahpq__hsvqp, jut__yvqe = _get_struct_payload(context, builder,
            struct_typ, struct)
        return impl_ret_borrowed(context, builder, sig.return_type,
            xahpq__hsvqp.data)
    return types.BaseTuple.from_types(struct_typ.data)(struct_typ), codegen


@intrinsic
def get_struct_null_bitmap(typingctx, struct_typ=None):
    assert isinstance(struct_typ, StructType)

    def codegen(context, builder, sig, args):
        struct, = args
        xahpq__hsvqp, jut__yvqe = _get_struct_payload(context, builder,
            struct_typ, struct)
        return impl_ret_borrowed(context, builder, sig.return_type,
            xahpq__hsvqp.null_bitmap)
    edxj__koc = types.UniTuple(types.int8, len(struct_typ.data))
    return edxj__koc(struct_typ), codegen


@intrinsic
def set_struct_data(typingctx, struct_typ, field_ind_typ, val_typ=None):
    assert isinstance(struct_typ, StructType) and is_overload_constant_int(
        field_ind_typ)
    field_ind = get_overload_const_int(field_ind_typ)

    def codegen(context, builder, sig, args):
        struct, jut__yvqe, val = args
        xahpq__hsvqp, pxkc__etl = _get_struct_payload(context, builder,
            struct_typ, struct)
        xbqci__vvioh = xahpq__hsvqp.data
        jcvwj__oys = builder.insert_value(xbqci__vvioh, val, field_ind)
        pwb__znmi = types.BaseTuple.from_types(struct_typ.data)
        context.nrt.decref(builder, pwb__znmi, xbqci__vvioh)
        context.nrt.incref(builder, pwb__znmi, jcvwj__oys)
        xahpq__hsvqp.data = jcvwj__oys
        builder.store(xahpq__hsvqp._getvalue(), pxkc__etl)
        return context.get_dummy_value()
    return types.none(struct_typ, field_ind_typ, val_typ), codegen


def _get_struct_field_ind(struct, ind, op):
    if not is_overload_constant_str(ind):
        raise BodoError(
            'structs (from struct array) only support constant strings for {}, not {}'
            .format(op, ind))
    dje__vfex = get_overload_const_str(ind)
    if dje__vfex not in struct.names:
        raise BodoError('Field {} does not exist in struct {}'.format(
            dje__vfex, struct))
    return struct.names.index(dje__vfex)


def is_field_value_null(s, field_name):
    pass


@overload(is_field_value_null, no_unliteral=True)
def overload_is_field_value_null(s, field_name):
    field_ind = _get_struct_field_ind(s, field_name, 'element access (getitem)'
        )
    return lambda s, field_name: get_struct_null_bitmap(s)[field_ind] == 0


@overload(operator.getitem, no_unliteral=True)
def struct_getitem(struct, ind):
    if not isinstance(struct, StructType):
        return
    field_ind = _get_struct_field_ind(struct, ind, 'element access (getitem)')
    return lambda struct, ind: get_struct_data(struct)[field_ind]


@overload(operator.setitem, no_unliteral=True)
def struct_setitem(struct, ind, val):
    if not isinstance(struct, StructType):
        return
    field_ind = _get_struct_field_ind(struct, ind, 'item assignment (setitem)')
    field_typ = struct.data[field_ind]
    return lambda struct, ind, val: set_struct_data(struct, field_ind,
        _cast(val, field_typ))


@overload(len, no_unliteral=True)
def overload_struct_arr_len(struct):
    if isinstance(struct, StructType):
        num_fields = len(struct.data)
        return lambda struct: num_fields


def construct_struct(context, builder, struct_type, values, nulls):
    payload_type = StructPayloadType(struct_type.data)
    ipgtj__gmjad = context.get_value_type(payload_type)
    wtgk__ltu = context.get_abi_sizeof(ipgtj__gmjad)
    tnw__lsewj = define_struct_dtor(context, builder, struct_type, payload_type
        )
    otg__lnjey = context.nrt.meminfo_alloc_dtor(builder, context.
        get_constant(types.uintp, wtgk__ltu), tnw__lsewj)
    qbl__xbo = context.nrt.meminfo_data(builder, otg__lnjey)
    pxkc__etl = builder.bitcast(qbl__xbo, ipgtj__gmjad.as_pointer())
    xahpq__hsvqp = cgutils.create_struct_proxy(payload_type)(context, builder)
    xahpq__hsvqp.data = cgutils.pack_array(builder, values
        ) if types.is_homogeneous(*struct_type.data) else cgutils.pack_struct(
        builder, values)
    xahpq__hsvqp.null_bitmap = cgutils.pack_array(builder, nulls)
    builder.store(xahpq__hsvqp._getvalue(), pxkc__etl)
    return otg__lnjey


@intrinsic
def struct_array_get_struct(typingctx, struct_arr_typ, ind_typ=None):
    assert isinstance(struct_arr_typ, StructArrayType) and isinstance(ind_typ,
        types.Integer)
    geey__yipqy = tuple(d.dtype for d in struct_arr_typ.data)
    ietqc__lgu = StructType(geey__yipqy, struct_arr_typ.names)

    def codegen(context, builder, sig, args):
        ocqlg__pnjm, ind = args
        xahpq__hsvqp = _get_struct_arr_payload(context, builder,
            struct_arr_typ, ocqlg__pnjm)
        pryat__ofzss = []
        imzmh__mmiis = []
        for i, arr_typ in enumerate(struct_arr_typ.data):
            boxk__lxh = builder.extract_value(xahpq__hsvqp.data, i)
            zze__pvzhe = context.compile_internal(builder, lambda arr, ind:
                np.uint8(0) if bodo.libs.array_kernels.isna(arr, ind) else
                np.uint8(1), types.uint8(arr_typ, types.int64), [boxk__lxh,
                ind])
            imzmh__mmiis.append(zze__pvzhe)
            nxm__vbl = cgutils.alloca_once_value(builder, context.
                get_constant_null(arr_typ.dtype))
            hsab__ufgd = builder.icmp_unsigned('==', zze__pvzhe, lir.
                Constant(zze__pvzhe.type, 1))
            with builder.if_then(hsab__ufgd):
                yjz__lhz = context.compile_internal(builder, lambda arr,
                    ind: arr[ind], arr_typ.dtype(arr_typ, types.int64), [
                    boxk__lxh, ind])
                builder.store(yjz__lhz, nxm__vbl)
            pryat__ofzss.append(builder.load(nxm__vbl))
        if isinstance(ietqc__lgu, types.DictType):
            fgxk__xgm = [context.insert_const_string(builder.module,
                kfdq__bkbbg) for kfdq__bkbbg in struct_arr_typ.names]
            nrm__aphlh = cgutils.pack_array(builder, pryat__ofzss)
            hvl__mith = cgutils.pack_array(builder, fgxk__xgm)

            def impl(names, vals):
                d = {}
                for i, kfdq__bkbbg in enumerate(names):
                    d[kfdq__bkbbg] = vals[i]
                return d
            cncnl__rugcg = context.compile_internal(builder, impl,
                ietqc__lgu(types.Tuple(tuple(types.StringLiteral(
                kfdq__bkbbg) for kfdq__bkbbg in struct_arr_typ.names)),
                types.Tuple(geey__yipqy)), [hvl__mith, nrm__aphlh])
            context.nrt.decref(builder, types.BaseTuple.from_types(
                geey__yipqy), nrm__aphlh)
            return cncnl__rugcg
        otg__lnjey = construct_struct(context, builder, ietqc__lgu,
            pryat__ofzss, imzmh__mmiis)
        struct = context.make_helper(builder, ietqc__lgu)
        struct.meminfo = otg__lnjey
        return struct._getvalue()
    return ietqc__lgu(struct_arr_typ, ind_typ), codegen


@intrinsic
def get_data(typingctx, arr_typ=None):
    assert isinstance(arr_typ, StructArrayType)

    def codegen(context, builder, sig, args):
        arr, = args
        xahpq__hsvqp = _get_struct_arr_payload(context, builder, arr_typ, arr)
        return impl_ret_borrowed(context, builder, sig.return_type,
            xahpq__hsvqp.data)
    return types.BaseTuple.from_types(arr_typ.data)(arr_typ), codegen


@intrinsic
def get_null_bitmap(typingctx, arr_typ=None):
    assert isinstance(arr_typ, StructArrayType)

    def codegen(context, builder, sig, args):
        arr, = args
        xahpq__hsvqp = _get_struct_arr_payload(context, builder, arr_typ, arr)
        return impl_ret_borrowed(context, builder, sig.return_type,
            xahpq__hsvqp.null_bitmap)
    return types.Array(types.uint8, 1, 'C')(arr_typ), codegen


@intrinsic
def init_struct_arr(typingctx, data_typ, null_bitmap_typ, names_typ=None):
    names = tuple(get_overload_const_str(mfvvh__nwve) for mfvvh__nwve in
        names_typ.types)
    struct_arr_type = StructArrayType(data_typ.types, names)

    def codegen(context, builder, sig, args):
        data, hwr__rcc, jclh__ihdca = args
        payload_type = StructArrayPayloadType(struct_arr_type.data)
        ipgtj__gmjad = context.get_value_type(payload_type)
        wtgk__ltu = context.get_abi_sizeof(ipgtj__gmjad)
        tnw__lsewj = define_struct_arr_dtor(context, builder,
            struct_arr_type, payload_type)
        otg__lnjey = context.nrt.meminfo_alloc_dtor(builder, context.
            get_constant(types.uintp, wtgk__ltu), tnw__lsewj)
        qbl__xbo = context.nrt.meminfo_data(builder, otg__lnjey)
        pxkc__etl = builder.bitcast(qbl__xbo, ipgtj__gmjad.as_pointer())
        xahpq__hsvqp = cgutils.create_struct_proxy(payload_type)(context,
            builder)
        xahpq__hsvqp.data = data
        xahpq__hsvqp.null_bitmap = hwr__rcc
        builder.store(xahpq__hsvqp._getvalue(), pxkc__etl)
        context.nrt.incref(builder, data_typ, data)
        context.nrt.incref(builder, null_bitmap_typ, hwr__rcc)
        llz__skiu = context.make_helper(builder, struct_arr_type)
        llz__skiu.meminfo = otg__lnjey
        return llz__skiu._getvalue()
    return struct_arr_type(data_typ, null_bitmap_typ, names_typ), codegen


@overload(operator.getitem, no_unliteral=True)
def struct_arr_getitem(arr, ind):
    if not isinstance(arr, StructArrayType):
        return
    if isinstance(ind, types.Integer):

        def struct_arr_getitem_impl(arr, ind):
            if ind < 0:
                ind += len(arr)
            return struct_array_get_struct(arr, ind)
        return struct_arr_getitem_impl
    pbuv__bmwh = len(arr.data)
    gnkx__qswa = 'def impl(arr, ind):\n'
    gnkx__qswa += '  data = get_data(arr)\n'
    gnkx__qswa += '  null_bitmap = get_null_bitmap(arr)\n'
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:
        gnkx__qswa += """  out_null_bitmap = get_new_null_mask_bool_index(null_bitmap, ind, len(data[0]))
"""
    elif is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):
        gnkx__qswa += """  out_null_bitmap = get_new_null_mask_int_index(null_bitmap, ind, len(data[0]))
"""
    elif isinstance(ind, types.SliceType):
        gnkx__qswa += """  out_null_bitmap = get_new_null_mask_slice_index(null_bitmap, ind, len(data[0]))
"""
    else:
        raise BodoError('invalid index {} in struct array indexing'.format(ind)
            )
    gnkx__qswa += ('  return init_struct_arr(({},), out_null_bitmap, ({},))\n'
        .format(', '.join('ensure_contig_if_np(data[{}][ind])'.format(i) for
        i in range(pbuv__bmwh)), ', '.join("'{}'".format(kfdq__bkbbg) for
        kfdq__bkbbg in arr.names)))
    gznki__qpzms = {}
    exec(gnkx__qswa, {'init_struct_arr': init_struct_arr, 'get_data':
        get_data, 'get_null_bitmap': get_null_bitmap, 'ensure_contig_if_np':
        bodo.utils.conversion.ensure_contig_if_np,
        'get_new_null_mask_bool_index': bodo.utils.indexing.
        get_new_null_mask_bool_index, 'get_new_null_mask_int_index': bodo.
        utils.indexing.get_new_null_mask_int_index,
        'get_new_null_mask_slice_index': bodo.utils.indexing.
        get_new_null_mask_slice_index}, gznki__qpzms)
    impl = gznki__qpzms['impl']
    return impl


@overload(operator.setitem, no_unliteral=True)
def struct_arr_setitem(arr, ind, val):
    if not isinstance(arr, StructArrayType):
        return
    if val == types.none or isinstance(val, types.optional):
        return
    if isinstance(ind, types.Integer):
        pbuv__bmwh = len(arr.data)
        gnkx__qswa = 'def impl(arr, ind, val):\n'
        gnkx__qswa += '  data = get_data(arr)\n'
        gnkx__qswa += '  null_bitmap = get_null_bitmap(arr)\n'
        gnkx__qswa += '  set_bit_to_arr(null_bitmap, ind, 1)\n'
        for i in range(pbuv__bmwh):
            if isinstance(val, StructType):
                gnkx__qswa += "  if is_field_value_null(val, '{}'):\n".format(
                    arr.names[i])
                gnkx__qswa += (
                    '    bodo.libs.array_kernels.setna(data[{}], ind)\n'.
                    format(i))
                gnkx__qswa += '  else:\n'
                gnkx__qswa += "    data[{}][ind] = val['{}']\n".format(i,
                    arr.names[i])
            else:
                gnkx__qswa += "  data[{}][ind] = val['{}']\n".format(i, arr
                    .names[i])
        gznki__qpzms = {}
        exec(gnkx__qswa, {'bodo': bodo, 'get_data': get_data,
            'get_null_bitmap': get_null_bitmap, 'set_bit_to_arr': bodo.libs
            .int_arr_ext.set_bit_to_arr, 'is_field_value_null':
            is_field_value_null}, gznki__qpzms)
        impl = gznki__qpzms['impl']
        return impl
    if isinstance(ind, types.SliceType):
        pbuv__bmwh = len(arr.data)
        gnkx__qswa = 'def impl(arr, ind, val):\n'
        gnkx__qswa += '  data = get_data(arr)\n'
        gnkx__qswa += '  null_bitmap = get_null_bitmap(arr)\n'
        gnkx__qswa += '  val_data = get_data(val)\n'
        gnkx__qswa += '  val_null_bitmap = get_null_bitmap(val)\n'
        gnkx__qswa += """  setitem_slice_index_null_bits(null_bitmap, val_null_bitmap, ind, len(arr))
"""
        for i in range(pbuv__bmwh):
            gnkx__qswa += '  data[{0}][ind] = val_data[{0}]\n'.format(i)
        gznki__qpzms = {}
        exec(gnkx__qswa, {'bodo': bodo, 'get_data': get_data,
            'get_null_bitmap': get_null_bitmap, 'set_bit_to_arr': bodo.libs
            .int_arr_ext.set_bit_to_arr, 'setitem_slice_index_null_bits':
            bodo.utils.indexing.setitem_slice_index_null_bits}, gznki__qpzms)
        impl = gznki__qpzms['impl']
        return impl
    raise BodoError(
        'only setitem with scalar/slice index is currently supported for struct arrays'
        )


@overload(len, no_unliteral=True)
def overload_struct_arr_len(A):
    if isinstance(A, StructArrayType):
        return lambda A: len(get_data(A)[0])


@overload_attribute(StructArrayType, 'shape')
def overload_struct_arr_shape(A):
    return lambda A: (len(get_data(A)[0]),)


@overload_attribute(StructArrayType, 'dtype')
def overload_struct_arr_dtype(A):
    return lambda A: np.object_


@overload_attribute(StructArrayType, 'ndim')
def overload_struct_arr_ndim(A):
    return lambda A: 1


@overload_attribute(StructArrayType, 'nbytes')
def overload_struct_arr_nbytes(A):
    gnkx__qswa = 'def impl(A):\n'
    gnkx__qswa += '  total_nbytes = 0\n'
    gnkx__qswa += '  data = get_data(A)\n'
    for i in range(len(A.data)):
        gnkx__qswa += f'  total_nbytes += data[{i}].nbytes\n'
    gnkx__qswa += '  total_nbytes += get_null_bitmap(A).nbytes\n'
    gnkx__qswa += '  return total_nbytes\n'
    gznki__qpzms = {}
    exec(gnkx__qswa, {'get_data': get_data, 'get_null_bitmap':
        get_null_bitmap}, gznki__qpzms)
    impl = gznki__qpzms['impl']
    return impl


@overload_method(StructArrayType, 'copy', no_unliteral=True)
def overload_struct_arr_copy(A):
    names = A.names

    def copy_impl(A):
        data = get_data(A)
        hwr__rcc = get_null_bitmap(A)
        rnu__ntscr = bodo.ir.join.copy_arr_tup(data)
        cgsir__kdbcj = hwr__rcc.copy()
        return init_struct_arr(rnu__ntscr, cgsir__kdbcj, names)
    return copy_impl
