"""
Helper functions for transformations.
"""
import itertools
import math
import operator
import types as pytypes
from collections import namedtuple
import numba
import numpy as np
import pandas as pd
from numba.core import ir, ir_utils, types
from numba.core.ir_utils import GuardException, build_definitions, compile_to_numba_ir, compute_cfg_from_blocks, find_callname, find_const, get_definition, guard, is_setitem, mk_unique_var, replace_arg_nodes, require
from numba.core.registry import CPUDispatcher
from numba.core.typing.templates import fold_arguments
import bodo
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.map_arr_ext import MapArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.struct_arr_ext import StructArrayType, StructType
from bodo.libs.tuple_arr_ext import TupleArrayType
from bodo.utils.typing import BodoConstUpdatedError, BodoError, can_literalize_type, get_literal_value, get_overload_const_bool, get_overload_const_list, is_literal_type, is_overload_constant_bool
from bodo.utils.utils import is_array_typ, is_assign, is_call, is_expr
ReplaceFunc = namedtuple('ReplaceFunc', ['func', 'arg_types', 'args',
    'glbls', 'inline_bodo_calls', 'run_full_pipeline', 'pre_nodes'])
bodo_types_with_params = {'ArrayItemArrayType', 'CSRMatrixType',
    'CategoricalArrayType', 'CategoricalIndexType', 'DataFrameType',
    'DatetimeIndexType', 'Decimal128Type', 'DecimalArrayType',
    'IntegerArrayType', 'IntervalArrayType', 'IntervalIndexType', 'List',
    'MapArrayType', 'NumericIndexType', 'PDCategoricalDtype',
    'PeriodIndexType', 'RangeIndexType', 'SeriesType', 'StringIndexType',
    'BinaryIndexType', 'StructArrayType', 'TimedeltaIndexType',
    'TupleArrayType'}
container_update_method_names = ('clear', 'pop', 'popitem', 'update', 'add',
    'difference_update', 'discard', 'intersection_update', 'remove',
    'symmetric_difference_update', 'append', 'extend', 'insert', 'reverse',
    'sort')
no_side_effect_call_tuples = {(int,), (list,), (set,), (dict,), (min,), (
    max,), (abs,), (len,), (bool,), (str,), ('ceil', math), ('init_series',
    'pd_series_ext', 'hiframes', bodo), ('get_series_data', 'pd_series_ext',
    'hiframes', bodo), ('get_series_index', 'pd_series_ext', 'hiframes',
    bodo), ('get_series_name', 'pd_series_ext', 'hiframes', bodo), (
    'get_index_data', 'pd_index_ext', 'hiframes', bodo), ('get_index_name',
    'pd_index_ext', 'hiframes', bodo), ('init_binary_str_index',
    'pd_index_ext', 'hiframes', bodo), ('init_numeric_index',
    'pd_index_ext', 'hiframes', bodo), ('init_categorical_index',
    'pd_index_ext', 'hiframes', bodo), ('_dti_val_finalize', 'pd_index_ext',
    'hiframes', bodo), ('init_datetime_index', 'pd_index_ext', 'hiframes',
    bodo), ('init_timedelta_index', 'pd_index_ext', 'hiframes', bodo), (
    'init_range_index', 'pd_index_ext', 'hiframes', bodo), (
    'init_heter_index', 'pd_index_ext', 'hiframes', bodo), (
    'get_int_arr_data', 'int_arr_ext', 'libs', bodo), ('get_int_arr_bitmap',
    'int_arr_ext', 'libs', bodo), ('init_integer_array', 'int_arr_ext',
    'libs', bodo), ('alloc_int_array', 'int_arr_ext', 'libs', bodo), (
    'inplace_eq', 'str_arr_ext', 'libs', bodo), ('get_bool_arr_data',
    'bool_arr_ext', 'libs', bodo), ('get_bool_arr_bitmap', 'bool_arr_ext',
    'libs', bodo), ('init_bool_array', 'bool_arr_ext', 'libs', bodo), (
    'alloc_bool_array', 'bool_arr_ext', 'libs', bodo), (bodo.libs.
    bool_arr_ext.compute_or_body,), (bodo.libs.bool_arr_ext.
    compute_and_body,), ('alloc_datetime_date_array', 'datetime_date_ext',
    'hiframes', bodo), ('alloc_datetime_timedelta_array',
    'datetime_timedelta_ext', 'hiframes', bodo), ('cat_replace',
    'pd_categorical_ext', 'hiframes', bodo), ('init_categorical_array',
    'pd_categorical_ext', 'hiframes', bodo), ('alloc_categorical_array',
    'pd_categorical_ext', 'hiframes', bodo), ('get_categorical_arr_codes',
    'pd_categorical_ext', 'hiframes', bodo), ('_sum_handle_nan',
    'series_kernels', 'hiframes', bodo), ('_box_cat_val', 'series_kernels',
    'hiframes', bodo), ('_mean_handle_nan', 'series_kernels', 'hiframes',
    bodo), ('_var_handle_mincount', 'series_kernels', 'hiframes', bodo), (
    '_handle_nan_count', 'series_kernels', 'hiframes', bodo), (
    '_compute_var_nan_count_ddof', 'series_kernels', 'hiframes', bodo), (
    '_sem_handle_nan', 'series_kernels', 'hiframes', bodo), ('dist_return',
    'distributed_api', 'libs', bodo), ('init_dataframe', 'pd_dataframe_ext',
    'hiframes', bodo), ('get_dataframe_data', 'pd_dataframe_ext',
    'hiframes', bodo), ('get_dataframe_table', 'pd_dataframe_ext',
    'hiframes', bodo), ('get_table_data', 'table', 'hiframes', bodo), (
    'get_dataframe_index', 'pd_dataframe_ext', 'hiframes', bodo), (
    'init_rolling', 'pd_rolling_ext', 'hiframes', bodo), ('init_groupby',
    'pd_groupby_ext', 'hiframes', bodo), ('calc_nitems', 'array_kernels',
    'libs', bodo), ('concat', 'array_kernels', 'libs', bodo), ('unique',
    'array_kernels', 'libs', bodo), ('nunique', 'array_kernels', 'libs',
    bodo), ('quantile', 'array_kernels', 'libs', bodo), ('explode',
    'array_kernels', 'libs', bodo), ('explode_no_index', 'array_kernels',
    'libs', bodo), ('get_arr_lens', 'array_kernels', 'libs', bodo), (
    'str_arr_from_sequence', 'str_arr_ext', 'libs', bodo), (
    'get_str_arr_item_length', 'str_arr_ext', 'libs', bodo), (
    'parse_datetime_str', 'pd_timestamp_ext', 'hiframes', bodo), (
    'integer_to_dt64', 'pd_timestamp_ext', 'hiframes', bodo), (
    'dt64_to_integer', 'pd_timestamp_ext', 'hiframes', bodo), (
    'timedelta64_to_integer', 'pd_timestamp_ext', 'hiframes', bodo), (
    'integer_to_timedelta64', 'pd_timestamp_ext', 'hiframes', bodo), (
    'npy_datetimestruct_to_datetime', 'pd_timestamp_ext', 'hiframes', bodo),
    ('isna', 'array_kernels', 'libs', bodo), ('copy',), (
    'from_iterable_impl', 'typing', 'utils', bodo), ('chain', itertools), (
    'groupby',), ('rolling',), (pd.CategoricalDtype,), (bodo.hiframes.
    pd_categorical_ext.get_code_for_value,), ('asarray', np), ('int32', np),
    ('int64', np), ('float64', np), ('float32', np), ('bool_', np), ('full',
    np), ('round', np), ('isnan', np), ('isnat', np), ('internal_prange',
    'parfor', numba), ('internal_prange', 'parfor', 'parfors', numba), (
    'empty_inferred', 'ndarray', 'unsafe', numba), ('_slice_span',
    'unicode', numba), ('_normalize_slice', 'unicode', numba), (
    'init_session_builder', 'pyspark_ext', 'libs', bodo), ('init_session',
    'pyspark_ext', 'libs', bodo), ('init_spark_df', 'pyspark_ext', 'libs',
    bodo), ('h5size', 'h5_api', 'io', bodo), ('pre_alloc_struct_array',
    'struct_arr_ext', 'libs', bodo), (bodo.libs.struct_arr_ext.
    pre_alloc_struct_array,), ('pre_alloc_tuple_array', 'tuple_arr_ext',
    'libs', bodo), (bodo.libs.tuple_arr_ext.pre_alloc_tuple_array,), (
    'pre_alloc_array_item_array', 'array_item_arr_ext', 'libs', bodo), (
    bodo.libs.array_item_arr_ext.pre_alloc_array_item_array,), (
    'dist_reduce', 'distributed_api', 'libs', bodo), (bodo.libs.
    distributed_api.dist_reduce,), ('pre_alloc_string_array', 'str_arr_ext',
    'libs', bodo), (bodo.libs.str_arr_ext.pre_alloc_string_array,), (
    'pre_alloc_binary_array', 'binary_arr_ext', 'libs', bodo), (bodo.libs.
    binary_arr_ext.pre_alloc_binary_array,), ('pre_alloc_map_array',
    'map_arr_ext', 'libs', bodo), (bodo.libs.map_arr_ext.
    pre_alloc_map_array,), ('prange', bodo), (bodo.prange,), ('objmode',
    bodo), (bodo.objmode,), ('get_label_dict_from_categories',
    'pd_categorial_ext', 'hiframes', bodo), (
    'get_label_dict_from_categories_no_duplicates', 'pd_categorial_ext',
    'hiframes', bodo), ('build_nullable_tuple', 'nullable_tuple_ext',
    'libs', bodo)}


def remove_hiframes(rhs, lives, call_list):
    hpqn__qufd = tuple(call_list)
    if hpqn__qufd in no_side_effect_call_tuples:
        return True
    if len(call_list) == 4 and call_list[1:] == ['conversion', 'utils', bodo]:
        return True
    if isinstance(call_list[-1], pytypes.ModuleType) and call_list[-1
        ].__name__ == 'bodosql':
        return True
    if len(call_list) == 2 and call_list[0] == 'copy':
        return True
    if call_list == ['h5read', 'h5_api', 'io', bodo] and rhs.args[5
        ].name not in lives:
        return True
    if call_list == ['move_str_binary_arr_payload', 'str_arr_ext', 'libs', bodo
        ] and rhs.args[0].name not in lives:
        return True
    if call_list == ['setna', 'array_kernels', 'libs', bodo] and rhs.args[0
        ].name not in lives:
        return True
    if call_list == ['set_table_data', 'table', 'hiframes', bodo] and rhs.args[
        0].name not in lives:
        return True
    if len(hpqn__qufd) == 1 and tuple in getattr(hpqn__qufd[0], '__mro__', ()):
        return True
    return False


numba.core.ir_utils.remove_call_handlers.append(remove_hiframes)


def compile_func_single_block(func, args, ret_var, typing_info=None,
    extra_globals=None, infer_types=True, run_untyped_pass=False, flags=
    None, replace_globals=True):
    ntjyh__awcsy = {'numba': numba, 'np': np, 'bodo': bodo, 'pd': pd,
        'math': math}
    if extra_globals is not None:
        ntjyh__awcsy.update(extra_globals)
    if not replace_globals:
        ntjyh__awcsy = func.__globals__
    loc = ir.Loc('', 0)
    if ret_var:
        loc = ret_var.loc
    if typing_info and infer_types:
        loc = typing_info.curr_loc
        f_ir = compile_to_numba_ir(func, ntjyh__awcsy, typingctx=
            typing_info.typingctx, targetctx=typing_info.targetctx,
            arg_typs=tuple(typing_info.typemap[oivx__mncx.name] for
            oivx__mncx in args), typemap=typing_info.typemap, calltypes=
            typing_info.calltypes)
    else:
        f_ir = compile_to_numba_ir(func, ntjyh__awcsy)
    assert len(f_ir.blocks
        ) == 1, 'only single block functions supported in compile_func_single_block()'
    if run_untyped_pass:
        hiefr__qqizw = tuple(typing_info.typemap[oivx__mncx.name] for
            oivx__mncx in args)
        wraj__mdkww = bodo.transforms.untyped_pass.UntypedPass(f_ir,
            typing_info.typingctx, hiefr__qqizw, {}, {}, flags)
        wraj__mdkww.run()
    ocj__fvbxy = f_ir.blocks.popitem()[1]
    replace_arg_nodes(ocj__fvbxy, args)
    cjeu__oup = ocj__fvbxy.body[:-2]
    update_locs(cjeu__oup[len(args):], loc)
    for stmt in cjeu__oup[:len(args)]:
        stmt.target.loc = loc
    if ret_var is not None:
        yiz__hvv = ocj__fvbxy.body[-2]
        assert is_assign(yiz__hvv) and is_expr(yiz__hvv.value, 'cast')
        czt__gfr = yiz__hvv.value.value
        cjeu__oup.append(ir.Assign(czt__gfr, ret_var, loc))
    return cjeu__oup


def update_locs(node_list, loc):
    for stmt in node_list:
        stmt.loc = loc
        for lqw__zne in stmt.list_vars():
            lqw__zne.loc = loc
        if is_assign(stmt):
            stmt.value.loc = loc


def get_stmt_defs(stmt):
    if is_assign(stmt):
        return set([stmt.target.name])
    if type(stmt) in numba.core.analysis.ir_extension_usedefs:
        jrfy__hqeey = numba.core.analysis.ir_extension_usedefs[type(stmt)]
        lihz__hbexy, jjbtg__zygax = jrfy__hqeey(stmt)
        return jjbtg__zygax
    return set()


def get_const_value(var, func_ir, err_msg, typemap=None, arg_types=None,
    file_info=None):
    if hasattr(var, 'loc'):
        loc = var.loc
    else:
        loc = None
    try:
        tqo__epbtu = get_const_value_inner(func_ir, var, arg_types, typemap,
            file_info=file_info)
        if isinstance(tqo__epbtu, ir.UndefinedType):
            dfgep__ohwii = func_ir.get_definition(var.name).name
            raise BodoError(f"name '{dfgep__ohwii}' is not defined", loc=loc)
    except GuardException as tvu__wxctk:
        raise BodoError(err_msg, loc=loc)
    return tqo__epbtu


def get_const_value_inner(func_ir, var, arg_types=None, typemap=None,
    updated_containers=None, file_info=None, pyobject_to_literal=False,
    literalize_args=True):
    require(isinstance(var, ir.Var))
    rgqo__upk = get_definition(func_ir, var)
    yvr__omy = None
    if typemap is not None:
        yvr__omy = typemap.get(var.name, None)
    if isinstance(rgqo__upk, ir.Arg) and arg_types is not None:
        yvr__omy = arg_types[rgqo__upk.index]
    if updated_containers and var.name in updated_containers:
        raise BodoConstUpdatedError(
            f"variable '{var.name}' is updated inplace using '{updated_containers[var.name]}'"
            )
    if is_literal_type(yvr__omy):
        return get_literal_value(yvr__omy)
    if isinstance(rgqo__upk, (ir.Const, ir.Global, ir.FreeVar)):
        tqo__epbtu = rgqo__upk.value
        return tqo__epbtu
    if literalize_args and isinstance(rgqo__upk, ir.Arg
        ) and can_literalize_type(yvr__omy, pyobject_to_literal):
        raise numba.core.errors.ForceLiteralArg({rgqo__upk.index}, loc=var.
            loc, file_infos={rgqo__upk.index: file_info} if file_info is not
            None else None)
    if is_expr(rgqo__upk, 'binop'):
        if file_info and rgqo__upk.fn == operator.add:
            try:
                lkqmb__qse = get_const_value_inner(func_ir, rgqo__upk.lhs,
                    arg_types, typemap, updated_containers, literalize_args
                    =False)
                file_info.set_concat(lkqmb__qse, True)
                rreu__tsim = get_const_value_inner(func_ir, rgqo__upk.rhs,
                    arg_types, typemap, updated_containers, file_info)
                return rgqo__upk.fn(lkqmb__qse, rreu__tsim)
            except (GuardException, BodoConstUpdatedError) as tvu__wxctk:
                pass
            try:
                rreu__tsim = get_const_value_inner(func_ir, rgqo__upk.rhs,
                    arg_types, typemap, updated_containers, literalize_args
                    =False)
                file_info.set_concat(rreu__tsim, False)
                lkqmb__qse = get_const_value_inner(func_ir, rgqo__upk.lhs,
                    arg_types, typemap, updated_containers, file_info)
                return rgqo__upk.fn(lkqmb__qse, rreu__tsim)
            except (GuardException, BodoConstUpdatedError) as tvu__wxctk:
                pass
        lkqmb__qse = get_const_value_inner(func_ir, rgqo__upk.lhs,
            arg_types, typemap, updated_containers)
        rreu__tsim = get_const_value_inner(func_ir, rgqo__upk.rhs,
            arg_types, typemap, updated_containers)
        return rgqo__upk.fn(lkqmb__qse, rreu__tsim)
    if is_expr(rgqo__upk, 'unary'):
        tqo__epbtu = get_const_value_inner(func_ir, rgqo__upk.value,
            arg_types, typemap, updated_containers)
        return rgqo__upk.fn(tqo__epbtu)
    if is_expr(rgqo__upk, 'getattr') and typemap:
        eor__rour = typemap.get(rgqo__upk.value.name, None)
        if isinstance(eor__rour, bodo.hiframes.pd_dataframe_ext.DataFrameType
            ) and rgqo__upk.attr == 'columns':
            return pd.Index(eor__rour.columns)
        if isinstance(eor__rour, types.SliceType):
            llun__ndu = get_definition(func_ir, rgqo__upk.value)
            require(is_call(llun__ndu))
            red__pmb = find_callname(func_ir, llun__ndu)
            jxs__jpnq = False
            if red__pmb == ('_normalize_slice', 'numba.cpython.unicode'):
                require(rgqo__upk.attr in ('start', 'step'))
                llun__ndu = get_definition(func_ir, llun__ndu.args[0])
                jxs__jpnq = True
            require(find_callname(func_ir, llun__ndu) == ('slice', 'builtins'))
            if len(llun__ndu.args) == 1:
                if rgqo__upk.attr == 'start':
                    return 0
                if rgqo__upk.attr == 'step':
                    return 1
                require(rgqo__upk.attr == 'stop')
                return get_const_value_inner(func_ir, llun__ndu.args[0],
                    arg_types, typemap, updated_containers)
            if rgqo__upk.attr == 'start':
                tqo__epbtu = get_const_value_inner(func_ir, llun__ndu.args[
                    0], arg_types, typemap, updated_containers)
                if tqo__epbtu is None:
                    tqo__epbtu = 0
                if jxs__jpnq:
                    require(tqo__epbtu == 0)
                return tqo__epbtu
            if rgqo__upk.attr == 'stop':
                assert not jxs__jpnq
                return get_const_value_inner(func_ir, llun__ndu.args[1],
                    arg_types, typemap, updated_containers)
            require(rgqo__upk.attr == 'step')
            if len(llun__ndu.args) == 2:
                return 1
            else:
                tqo__epbtu = get_const_value_inner(func_ir, llun__ndu.args[
                    2], arg_types, typemap, updated_containers)
                if tqo__epbtu is None:
                    tqo__epbtu = 1
                if jxs__jpnq:
                    require(tqo__epbtu == 1)
                return tqo__epbtu
    if is_expr(rgqo__upk, 'getattr'):
        return getattr(get_const_value_inner(func_ir, rgqo__upk.value,
            arg_types, typemap, updated_containers), rgqo__upk.attr)
    if is_expr(rgqo__upk, 'getitem'):
        value = get_const_value_inner(func_ir, rgqo__upk.value, arg_types,
            typemap, updated_containers)
        index = get_const_value_inner(func_ir, rgqo__upk.index, arg_types,
            typemap, updated_containers)
        return value[index]
    foe__vuv = guard(find_callname, func_ir, rgqo__upk, typemap)
    if foe__vuv is not None and len(foe__vuv) == 2 and foe__vuv[0
        ] == 'keys' and isinstance(foe__vuv[1], ir.Var):
        ulogg__tdx = rgqo__upk.func
        rgqo__upk = get_definition(func_ir, foe__vuv[1])
        tbpp__uoei = foe__vuv[1].name
        if updated_containers and tbpp__uoei in updated_containers:
            raise BodoConstUpdatedError(
                "variable '{}' is updated inplace using '{}'".format(
                tbpp__uoei, updated_containers[tbpp__uoei]))
        require(is_expr(rgqo__upk, 'build_map'))
        vals = [lqw__zne[0] for lqw__zne in rgqo__upk.items]
        tyqmh__fqtg = guard(get_definition, func_ir, ulogg__tdx)
        assert isinstance(tyqmh__fqtg, ir.Expr) and tyqmh__fqtg.attr == 'keys'
        tyqmh__fqtg.attr = 'copy'
        return [get_const_value_inner(func_ir, lqw__zne, arg_types, typemap,
            updated_containers) for lqw__zne in vals]
    if is_expr(rgqo__upk, 'build_map'):
        return {get_const_value_inner(func_ir, lqw__zne[0], arg_types,
            typemap, updated_containers): get_const_value_inner(func_ir,
            lqw__zne[1], arg_types, typemap, updated_containers) for
            lqw__zne in rgqo__upk.items}
    if is_expr(rgqo__upk, 'build_tuple'):
        return tuple(get_const_value_inner(func_ir, lqw__zne, arg_types,
            typemap, updated_containers) for lqw__zne in rgqo__upk.items)
    if is_expr(rgqo__upk, 'build_list'):
        return [get_const_value_inner(func_ir, lqw__zne, arg_types, typemap,
            updated_containers) for lqw__zne in rgqo__upk.items]
    if is_expr(rgqo__upk, 'build_set'):
        return {get_const_value_inner(func_ir, lqw__zne, arg_types, typemap,
            updated_containers) for lqw__zne in rgqo__upk.items}
    if foe__vuv == ('list', 'builtins'):
        values = get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers)
        if isinstance(values, set):
            values = sorted(values)
        return list(values)
    if foe__vuv == ('set', 'builtins'):
        return set(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('range', 'builtins') and len(rgqo__upk.args) == 1:
        return range(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('slice', 'builtins'):
        return slice(*tuple(get_const_value_inner(func_ir, lqw__zne,
            arg_types, typemap, updated_containers) for lqw__zne in
            rgqo__upk.args))
    if foe__vuv == ('str', 'builtins'):
        return str(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('bool', 'builtins'):
        return bool(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('format', 'builtins'):
        oivx__mncx = get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers)
        nsv__pix = get_const_value_inner(func_ir, rgqo__upk.args[1],
            arg_types, typemap, updated_containers) if len(rgqo__upk.args
            ) > 1 else ''
        return format(oivx__mncx, nsv__pix)
    if foe__vuv in (('init_binary_str_index', 'bodo.hiframes.pd_index_ext'),
        ('init_numeric_index', 'bodo.hiframes.pd_index_ext'), (
        'init_categorical_index', 'bodo.hiframes.pd_index_ext'), (
        'init_datetime_index', 'bodo.hiframes.pd_index_ext'), (
        'init_timedelta_index', 'bodo.hiframes.pd_index_ext'), (
        'init_heter_index', 'bodo.hiframes.pd_index_ext')):
        return pd.Index(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('str_arr_from_sequence', 'bodo.libs.str_arr_ext'):
        return np.array(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('init_range_index', 'bodo.hiframes.pd_index_ext'):
        return pd.RangeIndex(get_const_value_inner(func_ir, rgqo__upk.args[
            0], arg_types, typemap, updated_containers),
            get_const_value_inner(func_ir, rgqo__upk.args[1], arg_types,
            typemap, updated_containers), get_const_value_inner(func_ir,
            rgqo__upk.args[2], arg_types, typemap, updated_containers))
    if foe__vuv == ('len', 'builtins') and typemap and isinstance(typemap.
        get(rgqo__upk.args[0].name, None), types.BaseTuple):
        return len(typemap[rgqo__upk.args[0].name])
    if foe__vuv == ('len', 'builtins'):
        zyw__uclxp = guard(get_definition, func_ir, rgqo__upk.args[0])
        if isinstance(zyw__uclxp, ir.Expr) and zyw__uclxp.op in ('build_tuple',
            'build_list', 'build_set', 'build_map'):
            return len(zyw__uclxp.items)
        return len(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv == ('CategoricalDtype', 'pandas'):
        kws = dict(rgqo__upk.kws)
        rqv__fpi = get_call_expr_arg('CategoricalDtype', rgqo__upk.args,
            kws, 0, 'categories', '')
        vxnnp__uosy = get_call_expr_arg('CategoricalDtype', rgqo__upk.args,
            kws, 1, 'ordered', False)
        if vxnnp__uosy is not False:
            vxnnp__uosy = get_const_value_inner(func_ir, vxnnp__uosy,
                arg_types, typemap, updated_containers)
        if rqv__fpi == '':
            rqv__fpi = None
        else:
            rqv__fpi = get_const_value_inner(func_ir, rqv__fpi, arg_types,
                typemap, updated_containers)
        return pd.CategoricalDtype(rqv__fpi, vxnnp__uosy)
    if foe__vuv == ('dtype', 'numpy'):
        return np.dtype(get_const_value_inner(func_ir, rgqo__upk.args[0],
            arg_types, typemap, updated_containers))
    if foe__vuv is not None and len(foe__vuv) == 2 and foe__vuv[1
        ] == 'pandas' and foe__vuv[0] in ('Int8Dtype', 'Int16Dtype',
        'Int32Dtype', 'Int64Dtype', 'UInt8Dtype', 'UInt16Dtype',
        'UInt32Dtype', 'UInt64Dtype'):
        return getattr(pd, foe__vuv[0])()
    if foe__vuv is not None and len(foe__vuv) == 2 and isinstance(foe__vuv[
        1], ir.Var):
        tqo__epbtu = get_const_value_inner(func_ir, foe__vuv[1], arg_types,
            typemap, updated_containers)
        args = [get_const_value_inner(func_ir, lqw__zne, arg_types, typemap,
            updated_containers) for lqw__zne in rgqo__upk.args]
        kws = {tbumm__lsvox[0]: get_const_value_inner(func_ir, tbumm__lsvox
            [1], arg_types, typemap, updated_containers) for tbumm__lsvox in
            rgqo__upk.kws}
        return getattr(tqo__epbtu, foe__vuv[0])(*args, **kws)
    if foe__vuv is not None and len(foe__vuv) == 2 and foe__vuv[1
        ] == 'bodo' and foe__vuv[0] in bodo_types_with_params:
        args = tuple(get_const_value_inner(func_ir, lqw__zne, arg_types,
            typemap, updated_containers) for lqw__zne in rgqo__upk.args)
        kwargs = {dfgep__ohwii: get_const_value_inner(func_ir, lqw__zne,
            arg_types, typemap, updated_containers) for dfgep__ohwii,
            lqw__zne in dict(rgqo__upk.kws).items()}
        return getattr(bodo, foe__vuv[0])(*args, **kwargs)
    if is_call(rgqo__upk) and typemap and isinstance(typemap.get(rgqo__upk.
        func.name, None), types.Dispatcher):
        py_func = typemap[rgqo__upk.func.name].dispatcher.py_func
        require(rgqo__upk.vararg is None)
        args = tuple(get_const_value_inner(func_ir, lqw__zne, arg_types,
            typemap, updated_containers) for lqw__zne in rgqo__upk.args)
        kwargs = {dfgep__ohwii: get_const_value_inner(func_ir, lqw__zne,
            arg_types, typemap, updated_containers) for dfgep__ohwii,
            lqw__zne in dict(rgqo__upk.kws).items()}
        arg_types = tuple(bodo.typeof(lqw__zne) for lqw__zne in args)
        kw_types = {ipbu__btqkn: bodo.typeof(lqw__zne) for ipbu__btqkn,
            lqw__zne in kwargs.items()}
        require(_func_is_pure(py_func, arg_types, kw_types))
        return py_func(*args, **kwargs)
    raise GuardException('Constant value not found')


def _func_is_pure(py_func, arg_types, kw_types):
    from bodo.hiframes.pd_dataframe_ext import DataFrameType
    from bodo.hiframes.pd_series_ext import SeriesType
    from bodo.ir.csv_ext import CsvReader
    from bodo.ir.json_ext import JsonReader
    from bodo.ir.parquet_ext import ParquetReader
    from bodo.ir.sql_ext import SqlReader
    f_ir, typemap, tafws__jrxwu, tafws__jrxwu = (bodo.compiler.
        get_func_type_info(py_func, arg_types, kw_types))
    for block in f_ir.blocks.values():
        for stmt in block.body:
            if isinstance(stmt, ir.Print):
                return False
            if isinstance(stmt, (CsvReader, JsonReader, ParquetReader,
                SqlReader)):
                return False
            if is_setitem(stmt) and isinstance(guard(get_definition, f_ir,
                stmt.target), ir.Arg):
                return False
            if is_assign(stmt):
                rhs = stmt.value
                if isinstance(rhs, ir.Yield):
                    return False
                if is_call(rhs):
                    fhid__tedrp = guard(get_definition, f_ir, rhs.func)
                    if isinstance(fhid__tedrp, ir.Const) and isinstance(
                        fhid__tedrp.value, numba.core.dispatcher.
                        ObjModeLiftedWith):
                        return False
                    ffrlu__zgifz = guard(find_callname, f_ir, rhs)
                    if ffrlu__zgifz is None:
                        return False
                    func_name, vayqa__czgj = ffrlu__zgifz
                    if vayqa__czgj == 'pandas' and func_name.startswith('read_'
                        ):
                        return False
                    if ffrlu__zgifz in (('fromfile', 'numpy'), ('file_read',
                        'bodo.io.np_io')):
                        return False
                    if ffrlu__zgifz == ('File', 'h5py'):
                        return False
                    if isinstance(vayqa__czgj, ir.Var):
                        yvr__omy = typemap[vayqa__czgj.name]
                        if isinstance(yvr__omy, (DataFrameType, SeriesType)
                            ) and func_name in ('to_csv', 'to_excel',
                            'to_json', 'to_sql', 'to_pickle', 'to_parquet',
                            'info'):
                            return False
                        if isinstance(yvr__omy, types.Array
                            ) and func_name == 'tofile':
                            return False
                        if isinstance(yvr__omy, bodo.LoggingLoggerType):
                            return False
                        if str(yvr__omy).startswith('Mpl'):
                            return False
                        if (func_name in container_update_method_names and
                            isinstance(guard(get_definition, f_ir,
                            vayqa__czgj), ir.Arg)):
                            return False
                    if vayqa__czgj in ('numpy.random', 'time', 'logging',
                        'matplotlib.pyplot'):
                        return False
    return True


def fold_argument_types(pysig, args, kws):

    def normal_handler(index, param, value):
        return value

    def default_handler(index, param, default):
        return types.Omitted(default)

    def stararg_handler(index, param, values):
        return types.StarArgTuple(values)
    args = fold_arguments(pysig, args, kws, normal_handler, default_handler,
        stararg_handler)
    return args


def get_const_func_output_type(func, arg_types, kw_types, typing_context,
    target_context, is_udf=True):
    from bodo.hiframes.pd_series_ext import HeterogeneousSeriesType, SeriesType
    py_func = None
    if isinstance(func, types.MakeFunctionLiteral):
        pnr__vzpo = func.literal_value.code
        ggyk__wxwx = {'np': np, 'pd': pd, 'numba': numba, 'bodo': bodo}
        if hasattr(func.literal_value, 'globals'):
            ggyk__wxwx = func.literal_value.globals
        f_ir = numba.core.ir_utils.get_ir_of_code(ggyk__wxwx, pnr__vzpo)
        fix_struct_return(f_ir)
        typemap, autsq__yxo, mng__mencw, tafws__jrxwu = (numba.core.
            typed_passes.type_inference_stage(typing_context,
            target_context, f_ir, arg_types, None))
    elif isinstance(func, bodo.utils.typing.FunctionLiteral):
        py_func = func.literal_value
        f_ir, typemap, mng__mencw, autsq__yxo = (bodo.compiler.
            get_func_type_info(py_func, arg_types, kw_types))
    elif isinstance(func, CPUDispatcher):
        py_func = func.py_func
        f_ir, typemap, mng__mencw, autsq__yxo = (bodo.compiler.
            get_func_type_info(py_func, arg_types, kw_types))
    else:
        if not isinstance(func, types.Dispatcher):
            if isinstance(func, types.Function):
                raise BodoError(
                    f'Bodo does not support built-in functions yet, {func}')
            else:
                raise BodoError(f'Function type expected, not {func}')
        py_func = func.dispatcher.py_func
        f_ir, typemap, mng__mencw, autsq__yxo = (bodo.compiler.
            get_func_type_info(py_func, arg_types, kw_types))
    if is_udf and isinstance(autsq__yxo, types.DictType):
        fbj__zeib = guard(get_struct_keynames, f_ir, typemap)
        if fbj__zeib is not None:
            autsq__yxo = StructType((autsq__yxo.value_type,) * len(
                fbj__zeib), fbj__zeib)
    if is_udf and isinstance(autsq__yxo, (SeriesType, HeterogeneousSeriesType)
        ):
        ylqo__lhuof = numba.core.registry.cpu_target.typing_context
        bpmip__jowc = numba.core.registry.cpu_target.target_context
        fiv__nqe = bodo.transforms.series_pass.SeriesPass(f_ir, ylqo__lhuof,
            bpmip__jowc, typemap, mng__mencw, {})
        fiv__nqe.run()
        fiv__nqe.run()
        fiv__nqe.run()
        ppa__lehl = compute_cfg_from_blocks(f_ir.blocks)
        iiae__gds = [guard(_get_const_series_info, f_ir.blocks[ajzb__ram],
            f_ir, typemap) for ajzb__ram in ppa__lehl.exit_points() if
            isinstance(f_ir.blocks[ajzb__ram].body[-1], ir.Return)]
        if None in iiae__gds or len(pd.Series(iiae__gds).unique()) != 1:
            autsq__yxo.const_info = None
        else:
            autsq__yxo.const_info = iiae__gds[0]
    return autsq__yxo


def _get_const_series_info(block, f_ir, typemap):
    from bodo.hiframes.pd_series_ext import HeterogeneousSeriesType
    assert isinstance(block.body[-1], ir.Return)
    rhzv__cic = block.body[-1].value
    qfucc__baeoy = get_definition(f_ir, rhzv__cic)
    require(is_expr(qfucc__baeoy, 'cast'))
    qfucc__baeoy = get_definition(f_ir, qfucc__baeoy.value)
    require(is_call(qfucc__baeoy) and find_callname(f_ir, qfucc__baeoy) ==
        ('init_series', 'bodo.hiframes.pd_series_ext'))
    iap__vlf = qfucc__baeoy.args[1]
    imrk__gcnje = tuple(get_const_value_inner(f_ir, iap__vlf, typemap=typemap))
    if isinstance(typemap[rhzv__cic.name], HeterogeneousSeriesType):
        return len(typemap[rhzv__cic.name].data), imrk__gcnje
    gzjok__phg = qfucc__baeoy.args[0]
    nmr__gmqxj = get_definition(f_ir, gzjok__phg)
    func_name, qxtyq__swhj = find_callname(f_ir, nmr__gmqxj)
    if is_call(nmr__gmqxj) and bodo.utils.utils.is_alloc_callname(func_name,
        qxtyq__swhj):
        uffwf__joxp = nmr__gmqxj.args[0]
        jyl__nmwta = get_const_value_inner(f_ir, uffwf__joxp, typemap=typemap)
        return jyl__nmwta, imrk__gcnje
    if is_call(nmr__gmqxj) and find_callname(f_ir, nmr__gmqxj) in [(
        'asarray', 'numpy'), ('str_arr_from_sequence', 'bodo.libs.str_arr_ext')
        ]:
        gzjok__phg = nmr__gmqxj.args[0]
        nmr__gmqxj = get_definition(f_ir, gzjok__phg)
    require(is_expr(nmr__gmqxj, 'build_tuple') or is_expr(nmr__gmqxj,
        'build_list'))
    return len(nmr__gmqxj.items), imrk__gcnje


def extract_keyvals_from_struct_map(f_ir, build_map, loc, scope, typemap=None):
    kyp__olwk = []
    sjqd__vmq = []
    values = []
    for ipbu__btqkn, lqw__zne in build_map.items:
        yod__ysqn = find_const(f_ir, ipbu__btqkn)
        require(isinstance(yod__ysqn, str))
        sjqd__vmq.append(yod__ysqn)
        kyp__olwk.append(ipbu__btqkn)
        values.append(lqw__zne)
    htsue__yvkw = ir.Var(scope, mk_unique_var('val_tup'), loc)
    gemux__kgsdx = ir.Assign(ir.Expr.build_tuple(values, loc), htsue__yvkw, loc
        )
    f_ir._definitions[htsue__yvkw.name] = [gemux__kgsdx.value]
    nxyc__slfm = ir.Var(scope, mk_unique_var('key_tup'), loc)
    hffrr__jogjh = ir.Assign(ir.Expr.build_tuple(kyp__olwk, loc),
        nxyc__slfm, loc)
    f_ir._definitions[nxyc__slfm.name] = [hffrr__jogjh.value]
    if typemap is not None:
        typemap[htsue__yvkw.name] = types.Tuple([typemap[lqw__zne.name] for
            lqw__zne in values])
        typemap[nxyc__slfm.name] = types.Tuple([typemap[lqw__zne.name] for
            lqw__zne in kyp__olwk])
    return sjqd__vmq, htsue__yvkw, gemux__kgsdx, nxyc__slfm, hffrr__jogjh


def _replace_const_map_return(f_ir, block, label):
    require(isinstance(block.body[-1], ir.Return))
    lkeh__hnwa = block.body[-1].value
    nobw__bmfw = guard(get_definition, f_ir, lkeh__hnwa)
    require(is_expr(nobw__bmfw, 'cast'))
    qfucc__baeoy = guard(get_definition, f_ir, nobw__bmfw.value)
    require(is_expr(qfucc__baeoy, 'build_map'))
    require(len(qfucc__baeoy.items) > 0)
    loc = block.loc
    scope = block.scope
    sjqd__vmq, htsue__yvkw, gemux__kgsdx, nxyc__slfm, hffrr__jogjh = (
        extract_keyvals_from_struct_map(f_ir, qfucc__baeoy, loc, scope))
    iozr__aytw = ir.Var(scope, mk_unique_var('conv_call'), loc)
    ptwzf__hrd = ir.Assign(ir.Global('struct_if_heter_dict', bodo.utils.
        conversion.struct_if_heter_dict, loc), iozr__aytw, loc)
    f_ir._definitions[iozr__aytw.name] = [ptwzf__hrd.value]
    cnfik__pdbg = ir.Var(scope, mk_unique_var('struct_val'), loc)
    hktqp__krkcw = ir.Assign(ir.Expr.call(iozr__aytw, [htsue__yvkw,
        nxyc__slfm], {}, loc), cnfik__pdbg, loc)
    f_ir._definitions[cnfik__pdbg.name] = [hktqp__krkcw.value]
    nobw__bmfw.value = cnfik__pdbg
    qfucc__baeoy.items = [(ipbu__btqkn, ipbu__btqkn) for ipbu__btqkn,
        tafws__jrxwu in qfucc__baeoy.items]
    block.body = block.body[:-2] + [gemux__kgsdx, hffrr__jogjh, ptwzf__hrd,
        hktqp__krkcw] + block.body[-2:]
    return tuple(sjqd__vmq)


def get_struct_keynames(f_ir, typemap):
    ppa__lehl = compute_cfg_from_blocks(f_ir.blocks)
    uka__fusvy = list(ppa__lehl.exit_points())[0]
    block = f_ir.blocks[uka__fusvy]
    require(isinstance(block.body[-1], ir.Return))
    lkeh__hnwa = block.body[-1].value
    nobw__bmfw = guard(get_definition, f_ir, lkeh__hnwa)
    require(is_expr(nobw__bmfw, 'cast'))
    qfucc__baeoy = guard(get_definition, f_ir, nobw__bmfw.value)
    require(is_call(qfucc__baeoy) and find_callname(f_ir, qfucc__baeoy) ==
        ('struct_if_heter_dict', 'bodo.utils.conversion'))
    return get_overload_const_list(typemap[qfucc__baeoy.args[1].name])


def fix_struct_return(f_ir):
    cdn__bxlx = None
    ppa__lehl = compute_cfg_from_blocks(f_ir.blocks)
    for uka__fusvy in ppa__lehl.exit_points():
        cdn__bxlx = guard(_replace_const_map_return, f_ir, f_ir.blocks[
            uka__fusvy], uka__fusvy)
    return cdn__bxlx


def update_node_list_definitions(node_list, func_ir):
    loc = ir.Loc('', 0)
    zpfr__ajduf = ir.Block(ir.Scope(None, loc), loc)
    zpfr__ajduf.body = node_list
    build_definitions({(0): zpfr__ajduf}, func_ir._definitions)
    return


NESTED_TUP_SENTINEL = '$BODO_NESTED_TUP'


def gen_const_val_str(c):
    if isinstance(c, tuple):
        return "'{}{}', ".format(NESTED_TUP_SENTINEL, len(c)) + ', '.join(
            gen_const_val_str(lqw__zne) for lqw__zne in c)
    if isinstance(c, str):
        return "'{}'".format(c)
    if isinstance(c, (pd.Timestamp, pd.Timedelta, float)):
        return "'{}'".format(c)
    return str(c)


def gen_const_tup(vals):
    uog__qqxn = ', '.join(gen_const_val_str(c) for c in vals)
    return '({}{})'.format(uog__qqxn, ',' if len(vals) == 1 else '')


def get_const_tup_vals(c_typ):
    vals = get_overload_const_list(c_typ)
    return _get_original_nested_tups(vals)


def _get_original_nested_tups(vals):
    for uws__yfh in range(len(vals) - 1, -1, -1):
        lqw__zne = vals[uws__yfh]
        if isinstance(lqw__zne, str) and lqw__zne.startswith(
            NESTED_TUP_SENTINEL):
            ljiy__rlczw = int(lqw__zne[len(NESTED_TUP_SENTINEL):])
            return _get_original_nested_tups(tuple(vals[:uws__yfh]) + (
                tuple(vals[uws__yfh + 1:uws__yfh + ljiy__rlczw + 1]),) +
                tuple(vals[uws__yfh + ljiy__rlczw + 1:]))
    return tuple(vals)


def get_call_expr_arg(f_name, args, kws, arg_no, arg_name, default=None,
    err_msg=None, use_default=False):
    oivx__mncx = None
    if len(args) > arg_no and arg_no >= 0:
        oivx__mncx = args[arg_no]
        if arg_name in kws:
            err_msg = (
                f"{f_name}() got multiple values for argument '{arg_name}'")
            raise BodoError(err_msg)
    elif arg_name in kws:
        oivx__mncx = kws[arg_name]
    if oivx__mncx is None:
        if use_default or default is not None:
            return default
        if err_msg is None:
            err_msg = "{} requires '{}' argument".format(f_name, arg_name)
        raise BodoError(err_msg)
    return oivx__mncx


def set_call_expr_arg(var, args, kws, arg_no, arg_name, add_if_missing=False):
    if len(args) > arg_no:
        args[arg_no] = var
    elif add_if_missing or arg_name in kws:
        kws[arg_name] = var
    else:
        raise BodoError('cannot set call argument since does not exist')


def avoid_udf_inline(py_func, arg_types, kw_types):
    from bodo.hiframes.pd_dataframe_ext import DataFrameType
    f_ir = numba.core.compiler.run_frontend(py_func, inline_closures=True)
    if '_bodo_inline' in kw_types and is_overload_constant_bool(kw_types[
        '_bodo_inline']):
        return not get_overload_const_bool(kw_types['_bodo_inline'])
    if any(isinstance(t, DataFrameType) for t in arg_types + tuple(kw_types
        .values())):
        return True
    for block in f_ir.blocks.values():
        if isinstance(block.body[-1], (ir.Raise, ir.StaticRaise)):
            return True
        for stmt in block.body:
            if isinstance(stmt, ir.EnterWith):
                return True
    return False


def replace_func(pass_info, func, args, const=False, pre_nodes=None,
    extra_globals=None, pysig=None, kws=None, inline_bodo_calls=False,
    run_full_pipeline=False):
    ntjyh__awcsy = {'numba': numba, 'np': np, 'bodo': bodo, 'pd': pd}
    if extra_globals is not None:
        ntjyh__awcsy.update(extra_globals)
    func.__globals__.update(ntjyh__awcsy)
    if pysig is not None:
        pre_nodes = [] if pre_nodes is None else pre_nodes
        scope = next(iter(pass_info.func_ir.blocks.values())).scope
        loc = scope.loc

        def normal_handler(index, param, default):
            return default

        def default_handler(index, param, default):
            dorpv__eci = ir.Var(scope, mk_unique_var('defaults'), loc)
            try:
                pass_info.typemap[dorpv__eci.name] = types.literal(default)
            except:
                pass_info.typemap[dorpv__eci.name] = numba.typeof(default)
            ukjj__gjvf = ir.Assign(ir.Const(default, loc), dorpv__eci, loc)
            pre_nodes.append(ukjj__gjvf)
            return dorpv__eci
        args = numba.core.typing.fold_arguments(pysig, args, kws,
            normal_handler, default_handler, normal_handler)
    hiefr__qqizw = tuple(pass_info.typemap[lqw__zne.name] for lqw__zne in args)
    if const:
        rkrz__vrr = []
        for uws__yfh, oivx__mncx in enumerate(args):
            tqo__epbtu = guard(find_const, pass_info.func_ir, oivx__mncx)
            if tqo__epbtu:
                rkrz__vrr.append(types.literal(tqo__epbtu))
            else:
                rkrz__vrr.append(hiefr__qqizw[uws__yfh])
        hiefr__qqizw = tuple(rkrz__vrr)
    return ReplaceFunc(func, hiefr__qqizw, args, ntjyh__awcsy,
        inline_bodo_calls, run_full_pipeline, pre_nodes)


def is_var_size_item_array_type(t):
    assert is_array_typ(t, False)
    return t == string_array_type or isinstance(t, ArrayItemArrayType
        ) or isinstance(t, StructArrayType) and any(
        is_var_size_item_array_type(erhzr__rdfl) for erhzr__rdfl in t.data)


def gen_init_varsize_alloc_sizes(t):
    if t == string_array_type:
        amnq__jovy = 'num_chars_{}'.format(ir_utils.next_label())
        return f'  {amnq__jovy} = 0\n', (amnq__jovy,)
    if isinstance(t, ArrayItemArrayType):
        nov__wum, hgcfz__cariu = gen_init_varsize_alloc_sizes(t.dtype)
        amnq__jovy = 'num_items_{}'.format(ir_utils.next_label())
        return f'  {amnq__jovy} = 0\n' + nov__wum, (amnq__jovy,) + hgcfz__cariu
    return '', ()


def gen_varsize_item_sizes(t, item, var_names):
    if t == string_array_type:
        return '    {} += bodo.libs.str_arr_ext.get_utf8_size({})\n'.format(
            var_names[0], item)
    if isinstance(t, ArrayItemArrayType):
        return '    {} += len({})\n'.format(var_names[0], item
            ) + gen_varsize_array_counts(t.dtype, item, var_names[1:])
    return ''


def gen_varsize_array_counts(t, item, var_names):
    if t == string_array_type:
        return ('    {} += bodo.libs.str_arr_ext.get_num_total_chars({})\n'
            .format(var_names[0], item))
    return ''


def get_type_alloc_counts(t):
    if isinstance(t, (StructArrayType, TupleArrayType)):
        return 1 + sum(get_type_alloc_counts(erhzr__rdfl.dtype) for
            erhzr__rdfl in t.data)
    if isinstance(t, ArrayItemArrayType) or t == string_array_type:
        return 1 + get_type_alloc_counts(t.dtype)
    if isinstance(t, MapArrayType):
        return get_type_alloc_counts(t.key_arr_type) + get_type_alloc_counts(t
            .value_arr_type)
    if bodo.utils.utils.is_array_typ(t, False) or t == bodo.string_type:
        return 1
    if isinstance(t, StructType):
        return sum(get_type_alloc_counts(erhzr__rdfl) for erhzr__rdfl in t.data
            )
    if isinstance(t, types.BaseTuple):
        return sum(get_type_alloc_counts(erhzr__rdfl) for erhzr__rdfl in t.
            types)
    return 0


def find_udf_str_name(obj_dtype, func_name, typing_context, caller_name):
    gptcn__ojk = typing_context.resolve_getattr(obj_dtype, func_name)
    if gptcn__ojk is None:
        mfwhs__mhby = types.misc.Module(np)
        try:
            gptcn__ojk = typing_context.resolve_getattr(mfwhs__mhby, func_name)
        except AttributeError as tvu__wxctk:
            gptcn__ojk = None
        if gptcn__ojk is None:
            raise BodoError(
                f"{caller_name}(): No Pandas method or Numpy function found with the name '{func_name}'."
                )
    return gptcn__ojk


def get_udf_str_return_type(obj_dtype, func_name, typing_context,
    caller_name, axis=None):
    gptcn__ojk = find_udf_str_name(obj_dtype, func_name, typing_context,
        caller_name)
    if isinstance(gptcn__ojk, types.BoundFunction):
        if axis is not None:
            tymn__uefn = gptcn__ojk.get_call_type(typing_context, (), {
                'axis': axis})
        else:
            tymn__uefn = gptcn__ojk.get_call_type(typing_context, (), {})
        return tymn__uefn.return_type
    else:
        if bodo.utils.typing.is_numpy_ufunc(gptcn__ojk):
            tymn__uefn = gptcn__ojk.get_call_type(typing_context, (
                obj_dtype,), {})
            return tymn__uefn.return_type
        raise BodoError(
            f"{caller_name}(): Only Pandas methods and np.ufunc are supported as string literals. '{func_name}' not supported."
            )


def get_pandas_method_str_impl(obj_dtype, func_name, typing_context,
    caller_name, axis=None):
    gptcn__ojk = find_udf_str_name(obj_dtype, func_name, typing_context,
        caller_name)
    if isinstance(gptcn__ojk, types.BoundFunction):
        frp__hidh = gptcn__ojk.template
        if axis is not None:
            return frp__hidh._overload_func(obj_dtype, axis=axis)
        else:
            return frp__hidh._overload_func(obj_dtype)
    return None


def dict_to_const_keys_var_values_lists(dict_var, func_ir, arg_types,
    typemap, updated_containers, require_const_map, label):
    require(isinstance(dict_var, ir.Var))
    trug__wzehb = get_definition(func_ir, dict_var)
    require(isinstance(trug__wzehb, ir.Expr))
    require(trug__wzehb.op == 'build_map')
    ysik__qcb = trug__wzehb.items
    kyp__olwk = []
    values = []
    nsk__gev = False
    for uws__yfh in range(len(ysik__qcb)):
        oadqi__jyr, value = ysik__qcb[uws__yfh]
        try:
            ojy__fshec = get_const_value_inner(func_ir, oadqi__jyr,
                arg_types, typemap, updated_containers)
            kyp__olwk.append(ojy__fshec)
            values.append(value)
        except GuardException as tvu__wxctk:
            require_const_map[oadqi__jyr] = label
            nsk__gev = True
    if nsk__gev:
        raise GuardException
    return kyp__olwk, values


def _get_const_keys_from_dict(args, func_ir, build_map, err_msg, loc):
    try:
        kyp__olwk = tuple(get_const_value_inner(func_ir, t[0], args) for t in
            build_map.items)
    except GuardException as tvu__wxctk:
        raise BodoError(err_msg, loc)
    if not all(isinstance(c, (str, int)) for c in kyp__olwk):
        raise BodoError(err_msg, loc)
    return kyp__olwk


def _convert_const_key_dict(args, func_ir, build_map, err_msg, scope, loc,
    output_sentinel_tuple=False):
    kyp__olwk = _get_const_keys_from_dict(args, func_ir, build_map, err_msg,
        loc)
    lphz__ismpe = []
    lhla__hhs = [bodo.transforms.typing_pass._create_const_var(ipbu__btqkn,
        'dict_key', scope, loc, lphz__ismpe) for ipbu__btqkn in kyp__olwk]
    pjzbo__etzr = [t[1] for t in build_map.items]
    if output_sentinel_tuple:
        aoo__qwrfp = ir.Var(scope, mk_unique_var('sentinel'), loc)
        tve__unkds = ir.Var(scope, mk_unique_var('dict_tup'), loc)
        lphz__ismpe.append(ir.Assign(ir.Const('__bodo_tup', loc),
            aoo__qwrfp, loc))
        mkb__ljn = [aoo__qwrfp] + lhla__hhs + pjzbo__etzr
        lphz__ismpe.append(ir.Assign(ir.Expr.build_tuple(mkb__ljn, loc),
            tve__unkds, loc))
        return (tve__unkds,), lphz__ismpe
    else:
        evmr__qaq = ir.Var(scope, mk_unique_var('values_tup'), loc)
        tnoi__frf = ir.Var(scope, mk_unique_var('idx_tup'), loc)
        lphz__ismpe.append(ir.Assign(ir.Expr.build_tuple(pjzbo__etzr, loc),
            evmr__qaq, loc))
        lphz__ismpe.append(ir.Assign(ir.Expr.build_tuple(lhla__hhs, loc),
            tnoi__frf, loc))
        return (evmr__qaq, tnoi__frf), lphz__ismpe
