"""
Collection of utility functions. Needs to be refactored in separate files.
"""
import hashlib
import inspect
import keyword
import re
import warnings
from enum import Enum
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, ir, ir_utils, types
from numba.core.imputils import lower_builtin, lower_constant
from numba.core.ir_utils import find_callname, find_const, get_definition, guard, mk_unique_var, require
from numba.core.typing import signature
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import intrinsic, overload
from numba.np.arrayobj import get_itemsize, make_array, populate_array
import bodo
from bodo.libs.binary_arr_ext import bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import num_total_chars, pre_alloc_string_array, string_array_type
from bodo.libs.str_ext import string_type
from bodo.utils.cg_helpers import is_ll_eq
from bodo.utils.typing import NOT_CONSTANT, BodoError, BodoWarning, MetaType
int128_type = types.Integer('int128', 128)


class CTypeEnum(Enum):
    Int8 = 0
    UInt8 = 1
    Int32 = 2
    UInt32 = 3
    Int64 = 4
    UInt64 = 7
    Float32 = 5
    Float64 = 6
    Int16 = 8
    UInt16 = 9
    STRING = 10
    Bool = 11
    Decimal = 12
    Date = 13
    Datetime = 14
    Timedelta = 15
    Int128 = 16
    LIST = 18
    STRUCT = 19
    BINARY = 20


_numba_to_c_type_map = {types.int8: CTypeEnum.Int8.value, types.uint8:
    CTypeEnum.UInt8.value, types.int32: CTypeEnum.Int32.value, types.uint32:
    CTypeEnum.UInt32.value, types.int64: CTypeEnum.Int64.value, types.
    uint64: CTypeEnum.UInt64.value, types.float32: CTypeEnum.Float32.value,
    types.float64: CTypeEnum.Float64.value, types.NPDatetime('ns'):
    CTypeEnum.Datetime.value, types.NPTimedelta('ns'): CTypeEnum.Timedelta.
    value, types.bool_: CTypeEnum.Bool.value, types.int16: CTypeEnum.Int16.
    value, types.uint16: CTypeEnum.UInt16.value, int128_type: CTypeEnum.
    Int128.value}
numba.core.errors.error_extras = {'unsupported_error': '', 'typing': '',
    'reportable': '', 'interpreter': '', 'constant_inference': ''}
np_alloc_callnames = 'empty', 'zeros', 'ones', 'full'
CONST_DICT_SLOW_WARN_THRESHOLD = 100
CONST_LIST_SLOW_WARN_THRESHOLD = 100000


def unliteral_all(args):
    return tuple(types.unliteral(a) for a in args)


def get_constant(func_ir, var, default=NOT_CONSTANT):
    kcwb__vre = guard(get_definition, func_ir, var)
    if kcwb__vre is None:
        return default
    if isinstance(kcwb__vre, ir.Const):
        return kcwb__vre.value
    if isinstance(kcwb__vre, ir.Var):
        return get_constant(func_ir, kcwb__vre, default)
    return default


def numba_to_c_type(t):
    if isinstance(t, bodo.libs.decimal_arr_ext.Decimal128Type):
        return CTypeEnum.Decimal.value
    if t == bodo.hiframes.datetime_date_ext.datetime_date_type:
        return CTypeEnum.Date.value
    return _numba_to_c_type_map[t]


def is_alloc_callname(func_name, mod_name):
    return isinstance(mod_name, str) and (mod_name == 'numpy' and func_name in
        np_alloc_callnames or func_name == 'empty_inferred' and mod_name in
        ('numba.extending', 'numba.np.unsafe.ndarray') or func_name ==
        'pre_alloc_string_array' and mod_name == 'bodo.libs.str_arr_ext' or
        func_name == 'pre_alloc_binary_array' and mod_name ==
        'bodo.libs.binary_arr_ext' or func_name ==
        'alloc_random_access_string_array' and mod_name ==
        'bodo.libs.str_ext' or func_name == 'pre_alloc_array_item_array' and
        mod_name == 'bodo.libs.array_item_arr_ext' or func_name ==
        'pre_alloc_struct_array' and mod_name == 'bodo.libs.struct_arr_ext' or
        func_name == 'pre_alloc_map_array' and mod_name ==
        'bodo.libs.map_arr_ext' or func_name == 'pre_alloc_tuple_array' and
        mod_name == 'bodo.libs.tuple_arr_ext' or func_name ==
        'alloc_bool_array' and mod_name == 'bodo.libs.bool_arr_ext' or 
        func_name == 'alloc_int_array' and mod_name ==
        'bodo.libs.int_arr_ext' or func_name == 'alloc_datetime_date_array' and
        mod_name == 'bodo.hiframes.datetime_date_ext' or func_name ==
        'alloc_datetime_timedelta_array' and mod_name ==
        'bodo.hiframes.datetime_timedelta_ext' or func_name ==
        'alloc_decimal_array' and mod_name == 'bodo.libs.decimal_arr_ext' or
        func_name == 'alloc_categorical_array' and mod_name ==
        'bodo.hiframes.pd_categorical_ext' or func_name == 'gen_na_array' and
        mod_name == 'bodo.libs.array_kernels')


def find_build_tuple(func_ir, var):
    require(isinstance(var, (ir.Var, str)))
    uhkx__mhves = get_definition(func_ir, var)
    require(isinstance(uhkx__mhves, ir.Expr))
    require(uhkx__mhves.op == 'build_tuple')
    return uhkx__mhves.items


def cprint(*s):
    print(*s)


@infer_global(cprint)
class CprintInfer(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        return signature(types.none, *unliteral_all(args))


typ_to_format = {types.int32: 'd', types.uint32: 'u', types.int64: 'lld',
    types.uint64: 'llu', types.float32: 'f', types.float64: 'lf', types.
    voidptr: 's'}


@lower_builtin(cprint, types.VarArg(types.Any))
def cprint_lower(context, builder, sig, args):
    for ertn__eicgs, val in enumerate(args):
        typ = sig.args[ertn__eicgs]
        if isinstance(typ, types.ArrayCTypes):
            cgutils.printf(builder, '%p ', val)
            continue
        vcl__ajc = typ_to_format[typ]
        cgutils.printf(builder, '%{} '.format(vcl__ajc), val)
    cgutils.printf(builder, '\n')
    return context.get_dummy_value()


def is_whole_slice(typemap, func_ir, var, accept_stride=False):
    require(typemap[var.name] == types.slice2_type or accept_stride and 
        typemap[var.name] == types.slice3_type)
    ozj__rwks = get_definition(func_ir, var)
    require(isinstance(ozj__rwks, ir.Expr) and ozj__rwks.op == 'call')
    assert len(ozj__rwks.args) == 2 or accept_stride and len(ozj__rwks.args
        ) == 3
    assert find_callname(func_ir, ozj__rwks) == ('slice', 'builtins')
    aal__dmfc = get_definition(func_ir, ozj__rwks.args[0])
    mzpc__pqfq = get_definition(func_ir, ozj__rwks.args[1])
    require(isinstance(aal__dmfc, ir.Const) and aal__dmfc.value == None)
    require(isinstance(mzpc__pqfq, ir.Const) and mzpc__pqfq.value == None)
    return True


def is_slice_equiv_arr(arr_var, index_var, func_ir, equiv_set,
    accept_stride=False):
    nyl__gvpfa = get_definition(func_ir, index_var)
    require(find_callname(func_ir, nyl__gvpfa) == ('slice', 'builtins'))
    require(len(nyl__gvpfa.args) in (2, 3))
    require(find_const(func_ir, nyl__gvpfa.args[0]) in (0, None))
    require(equiv_set.is_equiv(nyl__gvpfa.args[1], arr_var.name + '#0'))
    require(accept_stride or len(nyl__gvpfa.args) == 2 or find_const(
        func_ir, nyl__gvpfa.args[2]) == 1)
    return True


def get_slice_step(typemap, func_ir, var):
    require(typemap[var.name] == types.slice3_type)
    ozj__rwks = get_definition(func_ir, var)
    require(isinstance(ozj__rwks, ir.Expr) and ozj__rwks.op == 'call')
    assert len(ozj__rwks.args) == 3
    return ozj__rwks.args[2]


def is_array_typ(var_typ, include_index_series=True):
    return is_np_array_typ(var_typ) or var_typ in (string_array_type, bodo.
        binary_array_type, bodo.hiframes.split_impl.
        string_array_split_view_type, bodo.hiframes.datetime_date_ext.
        datetime_date_array_type, bodo.hiframes.datetime_timedelta_ext.
        datetime_timedelta_array_type, boolean_array, bodo.libs.str_ext.
        random_access_string_array) or isinstance(var_typ, (
        IntegerArrayType, bodo.libs.decimal_arr_ext.DecimalArrayType, bodo.
        hiframes.pd_categorical_ext.CategoricalArrayType, bodo.libs.
        array_item_arr_ext.ArrayItemArrayType, bodo.libs.struct_arr_ext.
        StructArrayType, bodo.libs.interval_arr_ext.IntervalArrayType, bodo
        .libs.tuple_arr_ext.TupleArrayType, bodo.libs.map_arr_ext.
        MapArrayType, bodo.libs.csr_matrix_ext.CSRMatrixType)
        ) or include_index_series and (isinstance(var_typ, (bodo.hiframes.
        pd_series_ext.SeriesType, bodo.hiframes.pd_multi_index_ext.
        MultiIndexType)) or bodo.hiframes.pd_index_ext.is_pd_index_type(
        var_typ))


def is_np_array_typ(var_typ):
    return isinstance(var_typ, types.Array)


def is_distributable_typ(var_typ):
    return is_array_typ(var_typ) or isinstance(var_typ, bodo.hiframes.table
        .TableType) or isinstance(var_typ, bodo.hiframes.pd_dataframe_ext.
        DataFrameType) or isinstance(var_typ, types.List
        ) and is_distributable_typ(var_typ.dtype) or isinstance(var_typ,
        types.DictType) and is_distributable_typ(var_typ.value_type)


def is_distributable_tuple_typ(var_typ):
    return isinstance(var_typ, types.BaseTuple) and any(
        is_distributable_typ(t) or is_distributable_tuple_typ(t) for t in
        var_typ.types) or isinstance(var_typ, types.List
        ) and is_distributable_tuple_typ(var_typ.dtype) or isinstance(var_typ,
        types.DictType) and is_distributable_tuple_typ(var_typ.value_type
        ) or isinstance(var_typ, types.iterators.EnumerateType) and (
        is_distributable_typ(var_typ.yield_type[1]) or
        is_distributable_tuple_typ(var_typ.yield_type[1]))


@numba.generated_jit(nopython=True, cache=True)
def build_set_seen_na(A):

    def impl(A):
        s = dict()
        vbo__peqpl = False
        for ertn__eicgs in range(len(A)):
            if bodo.libs.array_kernels.isna(A, ertn__eicgs):
                vbo__peqpl = True
                continue
            s[A[ertn__eicgs]] = 0
        return s, vbo__peqpl
    return impl


@numba.generated_jit(nopython=True, cache=True)
def build_set(A):
    if isinstance(A, IntegerArrayType) or A in (string_array_type,
        boolean_array):

        def impl_int_arr(A):
            s = dict()
            for ertn__eicgs in range(len(A)):
                if not bodo.libs.array_kernels.isna(A, ertn__eicgs):
                    s[A[ertn__eicgs]] = 0
            return s
        return impl_int_arr
    else:

        def impl(A):
            s = dict()
            for ertn__eicgs in range(len(A)):
                s[A[ertn__eicgs]] = 0
            return s
        return impl


def to_array(A):
    return np.array(A)


@overload(to_array, no_unliteral=True)
def to_array_overload(A):
    if isinstance(A, types.DictType):
        dtype = A.key_type

        def impl(A):
            n = len(A)
            arr = alloc_type(n, dtype, (-1,))
            ertn__eicgs = 0
            for v in A.keys():
                arr[ertn__eicgs] = v
                ertn__eicgs += 1
            return arr
        return impl

    def to_array_impl(A):
        return np.array(A)
    try:
        numba.njit(to_array_impl).get_call_template((A,), {})
        return to_array_impl
    except:
        pass


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def unique(A):
    if isinstance(A, IntegerArrayType) or A == boolean_array:
        return lambda A: A.unique()
    return lambda A: to_array(build_set(A))


def empty_like_type(n, arr):
    return np.empty(n, arr.dtype)


@overload(empty_like_type, no_unliteral=True)
def empty_like_type_overload(n, arr):
    if isinstance(arr, bodo.hiframes.pd_categorical_ext.CategoricalArrayType):
        return (lambda n, arr: bodo.hiframes.pd_categorical_ext.
            alloc_categorical_array(n, arr.dtype))
    if isinstance(arr, types.Array):
        return lambda n, arr: np.empty(n, arr.dtype)
    if isinstance(arr, types.List) and arr.dtype == string_type:

        def empty_like_type_str_list(n, arr):
            return [''] * n
        return empty_like_type_str_list
    if isinstance(arr, types.List) and arr.dtype == bytes_type:

        def empty_like_type_binary_list(n, arr):
            return [b''] * n
        return empty_like_type_binary_list
    if isinstance(arr, IntegerArrayType):
        ykw__rpl = arr.dtype

        def empty_like_type_int_arr(n, arr):
            return bodo.libs.int_arr_ext.alloc_int_array(n, ykw__rpl)
        return empty_like_type_int_arr
    if arr == boolean_array:

        def empty_like_type_bool_arr(n, arr):
            return bodo.libs.bool_arr_ext.alloc_bool_array(n)
        return empty_like_type_bool_arr
    if arr == bodo.hiframes.datetime_date_ext.datetime_date_array_type:

        def empty_like_type_datetime_date_arr(n, arr):
            return bodo.hiframes.datetime_date_ext.alloc_datetime_date_array(n)
        return empty_like_type_datetime_date_arr
    if (arr == bodo.hiframes.datetime_timedelta_ext.
        datetime_timedelta_array_type):

        def empty_like_type_datetime_timedelta_arr(n, arr):
            return (bodo.hiframes.datetime_timedelta_ext.
                alloc_datetime_timedelta_array(n))
        return empty_like_type_datetime_timedelta_arr
    if isinstance(arr, bodo.libs.decimal_arr_ext.DecimalArrayType):
        precision = arr.precision
        scale = arr.scale

        def empty_like_type_decimal_arr(n, arr):
            return bodo.libs.decimal_arr_ext.alloc_decimal_array(n,
                precision, scale)
        return empty_like_type_decimal_arr
    assert arr == string_array_type

    def empty_like_type_str_arr(n, arr):
        pwuje__vyg = 20
        if len(arr) != 0:
            pwuje__vyg = num_total_chars(arr) // len(arr)
        return pre_alloc_string_array(n, n * pwuje__vyg)
    return empty_like_type_str_arr


def _empty_nd_impl(context, builder, arrtype, shapes):
    bmzgt__yaood = make_array(arrtype)
    kvibw__vfufl = bmzgt__yaood(context, builder)
    hxpvv__poaeg = context.get_data_type(arrtype.dtype)
    esf__fbmnp = context.get_constant(types.intp, get_itemsize(context,
        arrtype))
    hroxh__mzif = context.get_constant(types.intp, 1)
    yqu__tzgk = lir.Constant(lir.IntType(1), 0)
    for s in shapes:
        ajk__cmmgt = builder.smul_with_overflow(hroxh__mzif, s)
        hroxh__mzif = builder.extract_value(ajk__cmmgt, 0)
        yqu__tzgk = builder.or_(yqu__tzgk, builder.extract_value(ajk__cmmgt, 1)
            )
    if arrtype.ndim == 0:
        vcg__sinrf = ()
    elif arrtype.layout == 'C':
        vcg__sinrf = [esf__fbmnp]
        for cypgv__hrts in reversed(shapes[1:]):
            vcg__sinrf.append(builder.mul(vcg__sinrf[-1], cypgv__hrts))
        vcg__sinrf = tuple(reversed(vcg__sinrf))
    elif arrtype.layout == 'F':
        vcg__sinrf = [esf__fbmnp]
        for cypgv__hrts in shapes[:-1]:
            vcg__sinrf.append(builder.mul(vcg__sinrf[-1], cypgv__hrts))
        vcg__sinrf = tuple(vcg__sinrf)
    else:
        raise NotImplementedError(
            "Don't know how to allocate array with layout '{0}'.".format(
            arrtype.layout))
    qbe__hszga = builder.smul_with_overflow(hroxh__mzif, esf__fbmnp)
    vsrtb__qssuu = builder.extract_value(qbe__hszga, 0)
    yqu__tzgk = builder.or_(yqu__tzgk, builder.extract_value(qbe__hszga, 1))
    with builder.if_then(yqu__tzgk, likely=False):
        cgutils.printf(builder,
            'array is too big; `arr.size * arr.dtype.itemsize` is larger than the maximum possible size.'
            )
    dtype = arrtype.dtype
    efy__ukgf = context.get_preferred_array_alignment(dtype)
    dnf__josa = context.get_constant(types.uint32, efy__ukgf)
    pzhlw__ucrzz = context.nrt.meminfo_alloc_aligned(builder, size=
        vsrtb__qssuu, align=dnf__josa)
    data = context.nrt.meminfo_data(builder, pzhlw__ucrzz)
    fjd__efr = context.get_value_type(types.intp)
    fcy__rgu = cgutils.pack_array(builder, shapes, ty=fjd__efr)
    fdazg__ihpdg = cgutils.pack_array(builder, vcg__sinrf, ty=fjd__efr)
    populate_array(kvibw__vfufl, data=builder.bitcast(data, hxpvv__poaeg.
        as_pointer()), shape=fcy__rgu, strides=fdazg__ihpdg, itemsize=
        esf__fbmnp, meminfo=pzhlw__ucrzz)
    return kvibw__vfufl


if bodo.numba_compat._check_numba_change:
    lines = inspect.getsource(numba.np.arrayobj._empty_nd_impl)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'b6a998927680caa35917a553c79704e9d813d8f1873d83a5f8513837c159fa29':
        warnings.warn('numba.np.arrayobj._empty_nd_impl has changed')


def alloc_arr_tup(n, arr_tup, init_vals=()):
    xyyt__mlwa = []
    for asdi__mjz in arr_tup:
        xyyt__mlwa.append(np.empty(n, asdi__mjz.dtype))
    return tuple(xyyt__mlwa)


@overload(alloc_arr_tup, no_unliteral=True)
def alloc_arr_tup_overload(n, data, init_vals=()):
    wkij__kyql = data.count
    iolik__wub = ','.join(['empty_like_type(n, data[{}])'.format(
        ertn__eicgs) for ertn__eicgs in range(wkij__kyql)])
    if init_vals != ():
        iolik__wub = ','.join(['np.full(n, init_vals[{}], data[{}].dtype)'.
            format(ertn__eicgs, ertn__eicgs) for ertn__eicgs in range(
            wkij__kyql)])
    qcayn__cqyiv = 'def f(n, data, init_vals=()):\n'
    qcayn__cqyiv += '  return ({}{})\n'.format(iolik__wub, ',' if 
        wkij__kyql == 1 else '')
    sgbvc__hzxpb = {}
    exec(qcayn__cqyiv, {'empty_like_type': empty_like_type, 'np': np},
        sgbvc__hzxpb)
    onbx__gikag = sgbvc__hzxpb['f']
    return onbx__gikag


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def tuple_to_scalar(n):
    if isinstance(n, types.BaseTuple) and len(n.types) == 1:
        return lambda n: n[0]
    return lambda n: n


def alloc_type(n, t, s=None):
    return np.empty(n, t.dtype)


@overload(alloc_type)
def overload_alloc_type(n, t, s=None):
    typ = t.instance_type if isinstance(t, types.TypeRef) else t
    if typ == string_array_type:
        return (lambda n, t, s=None: bodo.libs.str_arr_ext.
            pre_alloc_string_array(n, s[0]))
    if typ == bodo.binary_array_type:
        return (lambda n, t, s=None: bodo.libs.binary_arr_ext.
            pre_alloc_binary_array(n, s[0]))
    if isinstance(typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        dtype = typ.dtype
        return (lambda n, t, s=None: bodo.libs.array_item_arr_ext.
            pre_alloc_array_item_array(n, s, dtype))
    if isinstance(typ, bodo.libs.struct_arr_ext.StructArrayType):
        dtypes = typ.data
        names = typ.names
        return (lambda n, t, s=None: bodo.libs.struct_arr_ext.
            pre_alloc_struct_array(n, s, dtypes, names))
    if isinstance(typ, bodo.libs.map_arr_ext.MapArrayType):
        struct_typ = bodo.libs.struct_arr_ext.StructArrayType((typ.
            key_arr_type, typ.value_arr_type), ('key', 'value'))
        return lambda n, t, s=None: bodo.libs.map_arr_ext.pre_alloc_map_array(n
            , s, struct_typ)
    if isinstance(typ, bodo.libs.tuple_arr_ext.TupleArrayType):
        dtypes = typ.data
        return (lambda n, t, s=None: bodo.libs.tuple_arr_ext.
            pre_alloc_tuple_array(n, s, dtypes))
    if isinstance(typ, bodo.hiframes.pd_categorical_ext.CategoricalArrayType):
        if isinstance(t, types.TypeRef):
            if typ.dtype.categories is None:
                raise BodoError(
                    'UDFs or Groupbys that return Categorical values must have categories known at compile time.'
                    )
            is_ordered = typ.dtype.ordered
            int_type = typ.dtype.int_type
            new_cats_arr = pd.CategoricalDtype(typ.dtype.categories, is_ordered
                ).categories.values
            new_cats_tup = MetaType(tuple(new_cats_arr))
            return (lambda n, t, s=None: bodo.hiframes.pd_categorical_ext.
                alloc_categorical_array(n, bodo.hiframes.pd_categorical_ext
                .init_cat_dtype(bodo.utils.conversion.index_from_array(
                new_cats_arr), is_ordered, int_type, new_cats_tup)))
        else:
            return (lambda n, t, s=None: bodo.hiframes.pd_categorical_ext.
                alloc_categorical_array(n, t.dtype))
    if typ.dtype == bodo.hiframes.datetime_date_ext.datetime_date_type:
        return (lambda n, t, s=None: bodo.hiframes.datetime_date_ext.
            alloc_datetime_date_array(n))
    if (typ.dtype == bodo.hiframes.datetime_timedelta_ext.
        datetime_timedelta_type):
        return (lambda n, t, s=None: bodo.hiframes.datetime_timedelta_ext.
            alloc_datetime_timedelta_array(n))
    if isinstance(typ, DecimalArrayType):
        precision = typ.dtype.precision
        scale = typ.dtype.scale
        return (lambda n, t, s=None: bodo.libs.decimal_arr_ext.
            alloc_decimal_array(n, precision, scale))
    dtype = numba.np.numpy_support.as_dtype(typ.dtype)
    if isinstance(typ, IntegerArrayType):
        return lambda n, t, s=None: bodo.libs.int_arr_ext.alloc_int_array(n,
            dtype)
    if typ == boolean_array:
        return lambda n, t, s=None: bodo.libs.bool_arr_ext.alloc_bool_array(n)
    return lambda n, t, s=None: np.empty(n, dtype)


def astype(A, t):
    return A.astype(t.dtype)


@overload(astype, no_unliteral=True)
def overload_astype(A, t):
    typ = t.instance_type if isinstance(t, types.TypeRef) else t
    dtype = typ.dtype
    if A == typ:
        return lambda A, t: A
    if isinstance(A, (types.Array, IntegerArrayType)) and isinstance(typ,
        types.Array):
        return lambda A, t: A.astype(dtype)
    if isinstance(typ, IntegerArrayType):
        return lambda A, t: bodo.libs.int_arr_ext.init_integer_array(A.
            astype(dtype), np.full(len(A) + 7 >> 3, 255, np.uint8))
    raise BodoError(f'cannot convert array type {A} to {typ}')


def full_type(n, val, t):
    return np.full(n, val, t.dtype)


@overload(full_type, no_unliteral=True)
def overload_full_type(n, val, t):
    typ = t.instance_type if isinstance(t, types.TypeRef) else t
    if isinstance(typ, types.Array):
        dtype = numba.np.numpy_support.as_dtype(typ.dtype)
        return lambda n, val, t: np.full(n, val, dtype)
    if isinstance(typ, IntegerArrayType):
        dtype = numba.np.numpy_support.as_dtype(typ.dtype)
        return lambda n, val, t: bodo.libs.int_arr_ext.init_integer_array(np
            .full(n, val, dtype), np.full(tuple_to_scalar(n) + 7 >> 3, 255,
            np.uint8))
    if typ == boolean_array:
        return lambda n, val, t: bodo.libs.bool_arr_ext.init_bool_array(np.
            full(n, val, np.bool_), np.full(tuple_to_scalar(n) + 7 >> 3, 
            255, np.uint8))
    if typ == string_array_type:

        def impl_str(n, val, t):
            klq__zdrdf = n * len(val)
            A = pre_alloc_string_array(n, klq__zdrdf)
            for ertn__eicgs in range(n):
                A[ertn__eicgs] = val
            return A
        return impl_str

    def impl(n, val, t):
        A = alloc_type(n, typ, (-1,))
        for ertn__eicgs in range(n):
            A[ertn__eicgs] = val
        return A
    return impl


@intrinsic
def get_ctypes_ptr(typingctx, ctypes_typ=None):
    assert isinstance(ctypes_typ, types.ArrayCTypes)

    def codegen(context, builder, sig, args):
        hxaew__udx, = args
        hgnw__jydt = context.make_helper(builder, sig.args[0], hxaew__udx)
        return hgnw__jydt.data
    return types.voidptr(ctypes_typ), codegen


@intrinsic
def is_null_pointer(typingctx, ptr_typ=None):

    def codegen(context, builder, signature, args):
        csr__xqvi, = args
        clo__qyomz = context.get_constant_null(ptr_typ)
        return builder.icmp_unsigned('==', csr__xqvi, clo__qyomz)
    return types.bool_(ptr_typ), codegen


@intrinsic
def is_null_value(typingctx, val_typ=None):

    def codegen(context, builder, signature, args):
        val, = args
        uljxp__ppqai = cgutils.alloca_once_value(builder, val)
        emo__xqa = cgutils.alloca_once_value(builder, context.
            get_constant_null(val_typ))
        return is_ll_eq(builder, uljxp__ppqai, emo__xqa)
    return types.bool_(val_typ), codegen


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def tuple_list_to_array(A, data, elem_type):
    elem_type = elem_type.instance_type if isinstance(elem_type, types.TypeRef
        ) else elem_type
    qcayn__cqyiv = 'def impl(A, data, elem_type):\n'
    qcayn__cqyiv += '  for i, d in enumerate(data):\n'
    if elem_type == bodo.hiframes.pd_timestamp_ext.pd_timestamp_type:
        qcayn__cqyiv += (
            '    A[i] = bodo.utils.conversion.unbox_if_timestamp(d)\n')
    else:
        qcayn__cqyiv += '    A[i] = d\n'
    sgbvc__hzxpb = {}
    exec(qcayn__cqyiv, {'bodo': bodo}, sgbvc__hzxpb)
    impl = sgbvc__hzxpb['impl']
    return impl


def object_length(c, obj):
    jxig__antn = c.context.get_argument_type(types.pyobject)
    thahx__oov = lir.FunctionType(lir.IntType(64), [jxig__antn])
    iidx__vix = cgutils.get_or_insert_function(c.builder.module, thahx__oov,
        name='PyObject_Length')
    return c.builder.call(iidx__vix, (obj,))


def sequence_getitem(c, obj, ind):
    jxig__antn = c.context.get_argument_type(types.pyobject)
    thahx__oov = lir.FunctionType(jxig__antn, [jxig__antn, lir.IntType(64)])
    iidx__vix = cgutils.get_or_insert_function(c.builder.module, thahx__oov,
        name='PySequence_GetItem')
    return c.builder.call(iidx__vix, (obj, ind))


@intrinsic
def incref(typingctx, data=None):

    def codegen(context, builder, signature, args):
        tcl__ldqw, = args
        context.nrt.incref(builder, signature.args[0], tcl__ldqw)
    return types.void(data), codegen


def gen_getitem(out_var, in_var, ind, calltypes, nodes):
    yisz__gbt = out_var.loc
    tphxi__wavt = ir.Expr.static_getitem(in_var, ind, None, yisz__gbt)
    calltypes[tphxi__wavt] = None
    nodes.append(ir.Assign(tphxi__wavt, out_var, yisz__gbt))


def is_static_getsetitem(node):
    return is_expr(node, 'static_getitem') or isinstance(node, ir.StaticSetItem
        )


def get_getsetitem_index_var(node, typemap, nodes):
    index_var = node.index_var if is_static_getsetitem(node) else node.index
    if index_var is None:
        assert is_static_getsetitem(node)
        try:
            ewrok__okibo = types.literal(node.index)
        except:
            ewrok__okibo = numba.typeof(node.index)
        index_var = ir.Var(node.value.scope, ir_utils.mk_unique_var(
            'dummy_index'), node.loc)
        typemap[index_var.name] = ewrok__okibo
        nodes.append(ir.Assign(ir.Const(node.index, node.loc), index_var,
            node.loc))
    return index_var


import copy
ir.Const.__deepcopy__ = lambda self, memo: ir.Const(self.value, copy.
    deepcopy(self.loc))


def is_call_assign(stmt):
    return isinstance(stmt, ir.Assign) and isinstance(stmt.value, ir.Expr
        ) and stmt.value.op == 'call'


def is_call(expr):
    return isinstance(expr, ir.Expr) and expr.op == 'call'


def is_var_assign(inst):
    return isinstance(inst, ir.Assign) and isinstance(inst.value, ir.Var)


def is_assign(inst):
    return isinstance(inst, ir.Assign)


def is_expr(val, op):
    return isinstance(val, ir.Expr) and val.op == op


def sanitize_varname(varname):
    if isinstance(varname, (tuple, list)):
        varname = '_'.join(sanitize_varname(v) for v in varname)
    varname = str(varname)
    cqhp__kdcd = re.sub('\\W+', '_', varname)
    if not cqhp__kdcd or not cqhp__kdcd[0].isalpha():
        cqhp__kdcd = '_' + cqhp__kdcd
    if not cqhp__kdcd.isidentifier() or keyword.iskeyword(cqhp__kdcd):
        cqhp__kdcd = mk_unique_var('new_name').replace('.', '_')
    return cqhp__kdcd


def dump_node_list(node_list):
    for n in node_list:
        print('   ', n)


def debug_prints():
    return numba.core.config.DEBUG_ARRAY_OPT == 1


@overload(reversed)
def list_reverse(A):
    if isinstance(A, types.List):

        def impl_reversed(A):
            akj__dnwoq = len(A)
            for ertn__eicgs in range(akj__dnwoq):
                yield A[akj__dnwoq - 1 - ertn__eicgs]
        return impl_reversed


@numba.njit()
def count_nonnan(a):
    return np.count_nonzero(~np.isnan(a))


@numba.njit()
def nanvar_ddof1(a):
    nin__fblks = count_nonnan(a)
    if nin__fblks <= 1:
        return np.nan
    return np.nanvar(a) * (nin__fblks / (nin__fblks - 1))


@numba.njit()
def nanstd_ddof1(a):
    return np.sqrt(nanvar_ddof1(a))


def has_supported_h5py():
    try:
        import h5py
        from bodo.io import _hdf5
    except ImportError as tdmnc__rqp:
        nzoc__kcud = False
    else:
        nzoc__kcud = h5py.version.hdf5_version_tuple[1] == 10
    return nzoc__kcud


def check_h5py():
    if not has_supported_h5py():
        raise BodoError("install 'h5py' package to enable hdf5 support")


def has_pyarrow():
    try:
        import pyarrow
    except ImportError as tdmnc__rqp:
        xzk__nqrwm = False
    else:
        xzk__nqrwm = True
    return xzk__nqrwm


def has_scipy():
    try:
        import scipy
    except ImportError as tdmnc__rqp:
        tiait__ukn = False
    else:
        tiait__ukn = True
    return tiait__ukn


@intrinsic
def check_and_propagate_cpp_exception(typingctx):

    def codegen(context, builder, sig, args):
        vucfz__xmqor = context.get_python_api(builder)
        yiw__vycra = vucfz__xmqor.err_occurred()
        mxnoz__utdwq = cgutils.is_not_null(builder, yiw__vycra)
        with builder.if_then(mxnoz__utdwq):
            builder.ret(numba.core.callconv.RETCODE_EXC)
    return types.void(), codegen


def inlined_check_and_propagate_cpp_exception(context, builder):
    vucfz__xmqor = context.get_python_api(builder)
    yiw__vycra = vucfz__xmqor.err_occurred()
    mxnoz__utdwq = cgutils.is_not_null(builder, yiw__vycra)
    with builder.if_then(mxnoz__utdwq):
        builder.ret(numba.core.callconv.RETCODE_EXC)


@numba.njit
def check_java_installation(fname):
    with numba.objmode():
        check_java_installation_(fname)


def check_java_installation_(fname):
    if not fname.startswith('hdfs://'):
        return
    import shutil
    if not shutil.which('java'):
        qjre__vjx = (
            "Java not found. Make sure openjdk is installed for hdfs. openjdk can be installed by calling 'conda install openjdk=8 -c conda-forge'."
            )
        raise BodoError(qjre__vjx)


dt_err = """
        If you are trying to set NULL values for timedelta64 in regular Python, 

        consider using np.timedelta64('nat') instead of None
        """


@lower_constant(types.List)
def lower_constant_list(context, builder, typ, pyval):
    if len(pyval) > CONST_LIST_SLOW_WARN_THRESHOLD:
        warnings.warn(BodoWarning(
            'Using large global lists can result in long compilation times. Please pass large lists as arguments to JIT functions or use arrays.'
            ))
    rdi__mimvf = []
    for a in pyval:
        if bodo.typeof(a) != typ.dtype:
            raise BodoError(
                f'Values in list must have the same data type for type stability. Expected: {typ.dtype}, Actual: {bodo.typeof(a)}'
                )
        rdi__mimvf.append(context.get_constant_generic(builder, typ.dtype, a))
    tvd__smyei = context.get_constant_generic(builder, types.int64, len(pyval))
    pos__lpnlo = context.get_constant_generic(builder, types.bool_, False)
    jga__utagl = context.get_constant_null(types.pyobject)
    ieufe__axha = lir.Constant.literal_struct([tvd__smyei, tvd__smyei,
        pos__lpnlo] + rdi__mimvf)
    ieufe__axha = cgutils.global_constant(builder, '.const.payload',
        ieufe__axha).bitcast(cgutils.voidptr_t)
    vocpd__ytpzr = context.get_constant(types.int64, -1)
    ygm__wfew = context.get_constant_null(types.voidptr)
    pzhlw__ucrzz = lir.Constant.literal_struct([vocpd__ytpzr, ygm__wfew,
        ygm__wfew, ieufe__axha, vocpd__ytpzr])
    pzhlw__ucrzz = cgutils.global_constant(builder, '.const.meminfo',
        pzhlw__ucrzz).bitcast(cgutils.voidptr_t)
    return lir.Constant.literal_struct([pzhlw__ucrzz, jga__utagl])


@lower_constant(types.Set)
def lower_constant_set(context, builder, typ, pyval):
    for a in pyval:
        if bodo.typeof(a) != typ.dtype:
            raise BodoError(
                f'Values in set must have the same data type for type stability. Expected: {typ.dtype}, Actual: {bodo.typeof(a)}'
                )
    ymwzs__pwaz = types.List(typ.dtype)
    rhxjj__leilm = context.get_constant_generic(builder, ymwzs__pwaz, list(
        pyval))
    cab__jeh = context.compile_internal(builder, lambda l: set(l), types.
        Set(typ.dtype)(ymwzs__pwaz), [rhxjj__leilm])
    return cab__jeh


def lower_const_dict_fast_path(context, builder, typ, pyval):
    from bodo.utils.typing import can_replace
    fgxu__zdj = pd.Series(pyval.keys()).values
    zqdbo__xow = pd.Series(pyval.values()).values
    nvan__eanul = bodo.typeof(fgxu__zdj)
    bem__yyxb = bodo.typeof(zqdbo__xow)
    require(nvan__eanul.dtype == typ.key_type or can_replace(typ.key_type,
        nvan__eanul.dtype))
    require(bem__yyxb.dtype == typ.value_type or can_replace(typ.value_type,
        bem__yyxb.dtype))
    ahl__gpzp = context.get_constant_generic(builder, nvan__eanul, fgxu__zdj)
    hrui__popjn = context.get_constant_generic(builder, bem__yyxb, zqdbo__xow)

    def create_dict(keys, vals):
        dgoi__eow = {}
        for k, v in zip(keys, vals):
            dgoi__eow[k] = v
        return dgoi__eow
    xubt__ajf = context.compile_internal(builder, create_dict, typ(
        nvan__eanul, bem__yyxb), [ahl__gpzp, hrui__popjn])
    return xubt__ajf


@lower_constant(types.DictType)
def lower_constant_dict(context, builder, typ, pyval):
    try:
        return lower_const_dict_fast_path(context, builder, typ, pyval)
    except:
        pass
    if len(pyval) > CONST_DICT_SLOW_WARN_THRESHOLD:
        warnings.warn(BodoWarning(
            'Using large global dictionaries can result in long compilation times. Please pass large dictionaries as arguments to JIT functions.'
            ))
    sxpw__hfmcr = typ.key_type
    bci__ylxyq = typ.value_type

    def make_dict():
        return numba.typed.Dict.empty(sxpw__hfmcr, bci__ylxyq)
    xubt__ajf = context.compile_internal(builder, make_dict, typ(), [])

    def set_dict_val(d, k, v):
        d[k] = v
    for k, v in pyval.items():
        nmlh__nswcg = context.get_constant_generic(builder, sxpw__hfmcr, k)
        kye__afux = context.get_constant_generic(builder, bci__ylxyq, v)
        context.compile_internal(builder, set_dict_val, types.none(typ,
            sxpw__hfmcr, bci__ylxyq), [xubt__ajf, nmlh__nswcg, kye__afux])
    return xubt__ajf
